/*******************************************************************************
    Copyright (c) 2013 NVidia Corporation

    Permission is hereby granted, free of charge, to any person obtaining a copy
    of this software and associated documentation files (the "Software"), to
    deal in the Software without restriction, including without limitation the
    rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
    sell copies of the Software, and to permit persons to whom the Software is
    furnished to do so, subject to the following conditions:

        The above copyright notice and this permission notice shall be
        included in all copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
    IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
    FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
    THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
    LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
    FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
    DEALINGS IN THE SOFTWARE.
*******************************************************************************/

//
//     uvmtypes.h
//
//     This file contains basic datatypes that UVM requires.
//

#ifndef _UVMTYPES_H_
#define _UVMTYPES_H_

#include "nvtypes.h"
#include "nvstatus.h"
#include "nvCpuUuid.h"

/*******************************************************************************
    UVM stream types
*******************************************************************************/

typedef enum
{
    UvmStreamTypeRegular = 0,
    UvmStreamTypeAll = 1,
    UvmStreamTypeNone = 2
} UvmStreamType;

#define UVM_STREAM_INVALID  ((UvmStream)0ULL)
#define UVM_STREAM_ALL      ((UvmStream)2ULL)
#define UVM_STREAM_NONE     ((UvmStream)3ULL)

typedef unsigned long long UvmStream;

#define UVM_MAX_GPUS         32
#define UVM_MAX_PROCESSORS   (UVM_MAX_GPUS + 1)
#define UVM_MAX_DEVICES      UVM_MAX_GPUS
#define UVM_MAX_SUBDEVICES   8

#define UVM_RANGE_GROUP_ID_NONE        ((NvU64)0)

// TODO: Nuke Uvm*Uuid and use NvProcessorUuid,
// in NvGpuOps, UVM driver and User mode driver.
typedef NvProcessorUuid UvmProcessorUuid;
typedef NvProcessorUuid UvmGpuUuid;

//------------------------------------------------------------------------------
// UVM GPU mapping types
//
// These types indicate the kinds of accesses allowed from a given GPU at the
// specified virtual address range. There are 3 basic kinds of accesses: read,
// write and atomics. Each type indicates what kinds of accesses are allowed.
// Accesses of any disallowed kind are fatal. The "Default" type specifies that
// the UVM driver should decide on the types of accesses allowed.
//------------------------------------------------------------------------------
typedef enum
{
    UvmGpuMappingTypeDefault = 0,
    UvmGpuMappingTypeReadWriteAtomic = 1,
    UvmGpuMappingTypeReadWrite = 2,
    UvmGpuMappingTypeReadOnly = 3
} UvmGpuMappingType;

//------------------------------------------------------------------------------
// UVM GPU caching types
//
// These types indicate the cacheability of the specified virtual address range
// from a given GPU. The "Default" type specifies that the UVM driver should
// set caching on or off as required to follow the UVM coherence model. The
// "ForceUncached" and "ForceCached" types will always turn caching off or on
// respectively. These two types override the cacheability specified by the UVM
// coherence model.
//------------------------------------------------------------------------------
typedef enum
{
    UvmGpuCachingTypeDefault = 0,
    UvmGpuCachingTypeForceUncached = 1,
    UvmGpuCachingTypeForceCached = 2
} UvmGpuCachingType;

/*******************************************************************************
 * TODO: Remove these flags and all their references in uvm.h
 * UvmSingleGpuAttribute.gpuAttrFlags notes:
 *
 * If UVM_GPU_ATTRIBUTE_FLAGS_MAPPED is not set for a given GPU's attribute,
 * then you may not set either UVM_GPU_ATTRIBUTE_FLAGS_GPU_CACHEABLE or
 * UVM_GPU_ATTRIBUTE_FLAGS_ATOMICS_ENABLED for that GPU.
 *
 * For the various API calls that accept GPU attributes as an input, these
 * attributes will override any previously set attributes.
 */

#define UVM_GPU_ATTRIBUTE_FLAGS_MAPPED           0x00000001
#define UVM_GPU_ATTRIBUTE_FLAGS_GPU_CACHEABLE    0x00000002
#define UVM_GPU_ATTRIBUTE_FLAGS_ATOMICS_ENABLED  0x00000004
#define UVM_GPU_ATTRIBUTE_FLAGS_READ_ONLY        0x00000008

typedef struct
{
    UvmGpuUuid gpuUuid;
    NvU32      gpuMappingType; // UvmGpuMappingType
    NvU32      gpuCachingType; // UvmGpuCachingType
    NvU32      gpuAttrFlags;   // TODO: Remove this
} UvmGpuMappingAttributes;
// TODO: Remove this once all usages of UvmGpuAttributes are removed
typedef UvmGpuMappingAttributes UvmGpuAttributes;

// forward declaration of OS-dependent structure
struct UvmGlobalState_tag;

// Platform specific parameters for UvmRegisterGpuVaSpace
typedef union
{
    struct {
        // File descriptor for RM's control file
        int ctrlFd;
        // RM client handle
        NvHandle hClient;
        // RM GPU VA space handle
        NvHandle hVaSpace;
    } rm_linux;
} UvmGpuVaSpacePlatformParams;

// Platform specific parameters for UvmRegisterChannel and UvmUnregisterChannel
typedef union
{
    struct {
        // File descriptor for RM's control file
        int ctrlFd;
        // RM client handle
        NvHandle hClient;
        // RM channel handle
        NvHandle hChannel;
    } rm_linux;
} UvmChannelPlatformParams;

// Platform specific parameters for UvmMapExternalAllocation
typedef union
{
    struct {
        // File descriptor for RM's control file
        int ctrlFd;
        // RM client handle
        NvHandle hClient;
        // RM allocation handle
        NvHandle hMemory;
    } rm_linux;
} UvmAllocationPlatformParams;

//------------------------------------------------------------------------------
//    Tools API types
//------------------------------------------------------------------------------

#define UVM_DEBUG_V1    0x00000001

typedef NvUPtr UvmDebugSession;

//------------------------------------------------------------------------------
// Counter scope: It can be one of the following:
// - Single GPU for a process (UvmCounterScopeProcessSingleGpu)
// - Aggregate of all GPUs for a process (UvmCounterScopeProcessAllGpu)
// - Single GPU system-wide (UvmCounterScopeGlobalSingleGpu)
// (UvmCounterScopeGlobalSingleGpu is not supported for CUDA 6.0)
//
// Note: The user must not assume that the counter values are equal to zero
// at the time of enabling counters.
// Difference between end state counter value and start state counter value
// should be used to find out the correct value over a given period of time.
//------------------------------------------------------------------------------
typedef enum
{
    UvmCounterScopeProcessSingleGpu = 0,
    UvmCounterScopeProcessAllGpu = 1,
    UvmCounterScopeGlobalSingleGpu = 2,
    UvmCounterScopeSize
} UvmCounterScope;

//------------------------------------------------------------------------------
// Following numbers assigned to the counter name are used to index their value
// in the counter array.
//------------------------------------------------------------------------------
typedef enum
{
    UvmCounterNameBytesXferHtD = 0,         // host to device
    UvmCounterNameBytesXferDtH = 1,         // device to host
    UvmCounterNameCpuPageFaultCount = 2,
#ifdef __windows__
    UvmCounterNameWddmBytesXferBtH = 3,     // backing store to host
    UvmCounterNameWddmBytesXferHtB = 4,     // host to backing store
    //
    // eviction (device to backing store)
    //
    UvmCounterNameWddmBytesXferDtB = 5,
    //
    // restoration (backing store to device)
    //
    UvmCounterNameWddmBytesXferBtD = 6,
#endif
    //
    // bytes prefetched host to device.
    // These bytes are also counted in
    // UvmCounterNameBytesXferHtD
    //
    UvmCounterNamePrefetchBytesXferHtD = 7,
    //
    // bytes prefetched device to host.
    // These bytes are also counted in
    // UvmCounterNameBytesXferDtH
    //
    UvmCounterNamePrefetchBytesXferDtH = 8,
    //
    // number of faults reported on the GPU
    //
    UvmCounterNameGpuPageFaultCount = 9,
    UVM_TOTAL_COUNTERS
} UvmCounterName;

#define UVM_COUNTER_NAME_FLAG_BYTES_XFER_HTD 0x1
#define UVM_COUNTER_NAME_FLAG_BYTES_XFER_DTH 0x2
#define UVM_COUNTER_NAME_FLAG_CPU_PAGE_FAULT_COUNT 0x4
#define UVM_COUNTER_NAME_FLAG_WDDM_BYTES_XFER_BTH 0x8
#define UVM_COUNTER_NAME_FLAG_WDDM_BYTES_XFER_HTB 0x10
#define UVM_COUNTER_NAME_FLAG_BYTES_XFER_DTB 0x20
#define UVM_COUNTER_NAME_FLAG_BYTES_XFER_BTD 0x40
#define UVM_COUNTER_NAME_FLAG_PREFETCH_BYTES_XFER_HTD 0x80
#define UVM_COUNTER_NAME_FLAG_PREFETCH_BYTES_XFER_DTH 0x100
#define UVM_COUNTER_NAME_FLAG_GPU_PAGE_FAULT_COUNT 0x200

//------------------------------------------------------------------------------
// UVM counter config structure
//
// - scope: Please see the UvmCounterScope  enum (above), for details.
// - name: Name of the counter. Please check UvmCounterName for list.
// - gpuid: Identifies the GPU for which the counter will be enabled/disabled
//          This parameter is ignored in AllGpu scopes.
// - state: A value of 0 will disable the counter, a value of 1 will enable
//           the counter.
//------------------------------------------------------------------------------
typedef struct
{
    NvU32      scope; //UVM_DEBUG_V1 (UvmCounterScope)
    NvU32      name;  //UVM_DEBUG_V1 (UvmCounterName)
    UvmGpuUuid gpuid; //UVM_DEBUG_V1
    NvU32      state; //UVM_DEBUG_V1
} UvmCounterConfig;

#define UVM_COUNTER_CONFIG_STATE_DISABLE_REQUESTED  0
#define UVM_COUNTER_CONFIG_STATE_ENABLE_REQUESTED   1

//------------------------------------------------------------------------------
// Bit flags used to enable/ disable events:
//------------------------------------------------------------------------------
#define UVM_EVENT_ENABLE_MEMORY_VIOLATION            0x1
#define UVM_EVENT_ENABLE_MIGRATION                   0x2
#define UVM_EVENT_ENABLE_GPU_FAULT                   0x4
#define UVM_EVENT_ENABLE_GPU_FAULT_REPLAY            0x8
#define UVM_EVENT_ENABLE_FAULT_BUFFER_OVERFLOW       0x10

typedef enum {
    UvmEventMemoryAccessTypeRead = 0,
    UvmEventMemoryAccessTypeWrite = 1,
    UvmEventMemoryAccessTypeAtomic = 2,
    UvmEventMemoryAccessTypePrefetch = 3
} UvmEventMemoryAccessType;

//------------------------------------------------------------------------------
// Event type is used to identify the event data in the event queue
//------------------------------------------------------------------------------
typedef enum
{
    UvmEventTypeMemoryViolation = 0,
    UvmEventTypeMigration = 1,
    UvmEventTypeGpuFault = 2,
    UvmEventTypeGpuFaultReplay = 3,
    UvmEventTypeFaultBufferOverflow = 4,
    UvmEventNumTypes = 5,
    UvmEventTypeInvalid = 5,
} UvmEventType;

//------------------------------------------------------------------------------
// Information associated with a memory violation event
//------------------------------------------------------------------------------
typedef struct
{
    //
    // eventType has to be 1st argument of this structure. Setting eventType to
    // UvmEventTypeMigration helps to identify event data in a queue.
    //
    NvU8 eventType;
    NvU8 accessType;          // read/write violation (UvmEventMemoryAccessType)
    //
    // This structure is shared between UVM kernel and tools.
    // Manually padding the structure so that compiler options like pragma pack
    // or malign-double will have no effect on the field offsets.
    //
    NvU16 padding16Bits;
    NvU32 padding32Bits;
    NvU64 address;            // faulting address
    NvU64 timeStamp;          // time when violation occurred
    NvU32 pid;                // process id causing violation
    NvU32 threadId;           // thread id causing violation
} UvmEventMemoryViolationInfo;

typedef enum
{
    UvmEventMigrationDirectionCpuToGpu = 0,
    UvmEventMigrationDirectionGpuToCpu = 1
} UvmEventMigrationDirection;

//------------------------------------------------------------------------------
// Information associated with a migration event
//------------------------------------------------------------------------------
typedef struct
{
    //
    // eventType has to be the 1st argument of this structure.
    // Setting eventType = UvmEventMigrationInfo helps to identify event data in
    // a queue.
    //
    NvU8 eventType;
    // direction of migration (UvmEventMigrationDirection )
    // this field is deprecated, in favor of (src|dst)Index
    NvU8 direction;    
    //
    // Indices are used for the source and destination of migration instead of
    // using gpu uuid/cpu id. This reduces the size of each event. gpuIndex to
    // gpuUuid relation can be obtained from UvmEventGetGpuUuidTable.
    // Currently we do not distinguish between CPUs so they all use index 0xFF.
    //
    NvU8 srcIndex;                 // source CPU/GPU index
    NvU8 dstIndex;                 // destination CPU/GPU index
    //
    // This structure is shared between UVM kernel and tools.
    // Manually padding the structure so that compiler options like pragma pack
    // or malign-double will have no effect on the field offsets
    //
    NvU32 padding32Bits;
    NvU64 address;                 // virtual addr used for migration
    NvU64 migratedBytes;           // number of bytes migrated
    NvU64 beginTimeStamp;          // migration start time stamp
    NvU64 endTimeStamp;            // migration end time stamp
    NvU64 streamId;                // stream causing the migration
} UvmEventMigrationInfo;


//------------------------------------------------------------------------------
// Information associated with a migration event UVM8 onwards
//------------------------------------------------------------------------------
typedef struct
{
    //
    // eventType has to be the 1st argument of this structure. Setting eventType
    // to UvmEventMigrationInfo helps to identify event data in a queue.
    //
    NvU8 eventType;
    NvU8 padding8Bits;
    //
    // Indices are used for the source and destination of migration instead of
    // using gpu uuid/cpu id. This reduces the size of each event. gpuIndex to
    // gpuUuid relation can be obtained from UvmEventGetGpuUuidTable.
    // Currently we do not distinguish between CPUs so they all use index 0.
    //
    NvU8 srcIndex;                 // source CPU/GPU index
    NvU8 dstIndex;                 // destination CPU/GPU index
    //
    // This structure is shared between UVM kernel and tools.
    // Manually padding the structure so that compiler options like pragma pack
    // or malign-double will have no effect on the field offsets
    //
    NvU32 padding32Bits;
    NvU64 address;                 // virtual addr used for migration
    NvU64 migratedBytes;           // number of bytes migrated
    NvU64 beginTimeStamp;          // migration start time stamp
    NvU64 endTimeStamp;            // migration end time stamp
    NvU64 rangeGroupId;            // range group tied with this migration
} UvmEventMigrationInfo_Version2;

typedef enum
{
    // These fault types are handled and may be "fixed" by the UVM driver
    UvmFaultTypeInvalidPde = 0,
    UvmFaultTypeInvalidPte = 1,
    UvmFaultTypeWrite = 2,
    UvmFaultTypeAtomic = 3,
    // The next fault types are fatal and cannot be fixed by the UVM driver
    UvmFaultTypeFatal = 4,
    UvmFaultTypeInvalidPdeSize = UvmFaultTypeFatal,
    UvmFaultTypeLimitViolation = 5,
    UvmFaultTypeUnboundInstBlock = 6,
    UvmFaultTypePrivViolation = 7,
    UvmFaultTypePitchMaskViolation = 8,
    UvmFaultTypeWorkCreation = 9,
    UvmFaultTypeUnsupportedAperture = 10,
    UvmFaultTypeCompressionFailure = 11,
    UvmFaultTypeUnsupportedKind = 12,
    UvmFaultTypeRegionViolation = 13,
    UvmFaultTypePoison = 14,
} UvmFaultEventType;

//------------------------------------------------------------------------------
// This info is provided per gpu fault
// This event can be treated as a start event for gpu fault handling
//------------------------------------------------------------------------------
typedef struct
{
    //
    // eventType has to be the 1st argument of this structure.
    // Setting eventType = UvmEventGpuFaultInfo helps to identify event data in
    // a queue.
    //
    NvU8 eventType;
    NvU8 faultType;       // type of gpu fault, refer UvmFaultEventType
    NvU8 accessType;      // memory access type, refer UvmEventMemoryAccessType
    NvU8 reserved8bits;
    NvU32 reserved32bits;
    NvU64 address;        // virtual address at which gpu faulted
    NvU64 timestampCpu;   // time when fault packet entry is read, this is not
    NvU64 timestampGpu;   // the gpu fault timestamp from fault packet
} UvmEventGpuFaultInfo;

//------------------------------------------------------------------------------
// This info is provided when a gpu fault is replayed.
// This event can be treated as an end event for gpu fault handling.
// Any other events eg migration events caused as a side-effect of the gpu fault
// would lie between the start and end event. Faults are serviced over a range
// Minimum range = page size (4K/64K/2MB). max would depend on prefetching logic
// Each start event would have a corresponding end event.
//------------------------------------------------------------------------------
typedef struct
{
    //
    // eventType has to be the 1st argument of this structure.
    // Setting eventType = UvmEventGpuFaultReplayInfo helps to identify event
    // data in a queue.
    //
    NvU8 eventType;
    //
    // This structure is shared between UVM kernel and tools.
    // Manually padding the structure so that compiler options like pragma pack
    // or malign-double will have no effect on the field offsets
    //
    NvU8  padding8bits;
    NvU16 padding16bits;
    NvU32 padding32bits;
    NvU64 address;          // virtual address at which gpu faulted
    NvU64 timestamp;        // time when gpu fault is detected
} UvmEventGpuFaultReplayInfo;

//------------------------------------------------------------------------------
// This info is provided when the fault buffer overflows.
// This event will get generated after the number of events is greater than the
// maxFaultsInBuffer.
//------------------------------------------------------------------------------
typedef struct
{
    //
    // eventType has to be the 1st argument of this structure.
    // Setting eventType = UvmEventGpuFaultReplayInfo helps to identify event
    // data in a queue.
    //
    NvU8 eventType;
    //
    // This structure is shared between UVM kernel and tools.
    // Manually padding the structure so that compiler options like pragma pack
    // or malign-double will have no effect on the field offsets
    //
    NvU8  padding8bits;
    NvU16 padding16bits;
    NvU32 padding32bits;

    NvU64 timeStamp;
} UvmEventFaultBufferOverflowInfo;

//------------------------------------------------------------------------------
// Entry added in the event queue buffer when an enabled event occurs. For
// compatibility with all tools ensure that this structure is 64 bit aligned.
//------------------------------------------------------------------------------
typedef struct
{
    union
    {
        NvU8 eventType;
        UvmEventMemoryViolationInfo memViolation;
        UvmEventMigrationInfo migration;
        UvmEventMigrationInfo_Version2 migrationVer2;
        UvmEventGpuFaultInfo gpuFault;
        UvmEventGpuFaultReplayInfo gpuFaultReplay;
        UvmEventFaultBufferOverflowInfo faultBufferOverflow;
    } eventData;
} UvmEventEntry;

//------------------------------------------------------------------------------
// Type of time stamp used in the event entry:
//
// On windows we support QPC type which uses RDTSC if possible else fallbacks to
// HPET.
//
// On Linux ClockGetTime provides similar functionality.
// In UvmEventTimeStampTypeAuto the system decides which time stamp best suites
// current environment.
//------------------------------------------------------------------------------
typedef enum
{
    UvmEventTimeStampTypeWin32QPC = 0,
    UvmEventTimeStampTypePosixClockGetTime = 1,
    UvmEventTimeStampTypeAuto = 2
} UvmEventTimeStampType;

//------------------------------------------------------------------------------
// An opaque queue handle is returned to the user when a queue is created.
//------------------------------------------------------------------------------
typedef NvUPtr UvmEventQueueHandle;

//------------------------------------------------------------------------------
// Setting default page size to 4k,
// this can be updated to 64k in case of power PC
//------------------------------------------------------------------------------
#define UVM_DEBUG_ACCESS_PAGE_SIZE      (1 << 12) // 4k page

typedef enum
{
    UvmDebugAccessTypeRead = 0,
    UvmDebugAccessTypeWrite = 1,
} UvmDebugAccessType;

typedef enum
{
    UvmPerfInfoTypeThrashing = 1,
    UvmPerfInfoTypePrefetch  = 2,
} UvmPerfInfoType;

// Thrashing prevention techniques implemented in the driver
typedef enum
{
    UvmPerfThrashingPreventionNone          = 0, // Do nothing
    UvmPerfThrashingPreventionThrottleSched,     // In the CPU fault handler, call sched and try again
    UvmPerfThrashingPreventionThrottleSleep,     // In the CPU fault handler, call msleep and try again
    UvmPerfThrashingPreventionPinSysmem    ,     // In the CPU fault handler, migrate the page to sysmem, pin it there
                                                 // and update the GPU page tables to point there
    UvmPerfThrashingPreventionCount
} UvmPerfThrashingPrevention;

typedef struct
{
    NvU64 time;
    NvU8  proc; // Id of the processor that caused the fault
} UvmPerfTimestamp;

#define UVM_PERF_INFO_MAX_TSTAMPS 256

// Per-page thrashing counters and statistics used for performance heuristics analyses
typedef struct
{
    // General statistics
    NvU32 atomic_faults;
    NvU32 write_faults;
    NvU32 read_faults;

    // Number of page migrations between device memories. Migrations can also happen on faults on a neighboring page in
    // the same 2 MB descriptor
    NvU32 migrations;

    // Number of detected thrashing periods
    NvU32 thrashing_total;

    // Data that depends on the used thrashing prevention approach
    union
    {
        struct
        {
            // Timestamps of the last page fault events. Used for offline analysis of access patterns
            UvmPerfTimestamp fault_tstamps[UVM_PERF_INFO_MAX_TSTAMPS];
        } none;
        struct
        {
            // Number of times the CPU was put to sleep
            NvLength naps;
        } throttle;
        struct
        {
            // Number of times the page was pinned to sysmem. They can be speculatively unpinned to avoid paying the
            // costs of remote accesses in GPU kernels
            NvLength pins;
        } pin_sysmem;
    };
} UvmPageThrashingInfo;

typedef struct UvmEventControlData_tag {
    // entries between get_ahead and get_behind are currently being read
    volatile NvU32 get_ahead;
    volatile NvU32 get_behind;
    // entries between put_ahead and put_behind are currently being written
    volatile NvU32 put_ahead;
    volatile NvU32 put_behind;

    // counter of dropped events
    NvU64 dropped[UvmEventNumTypes];
} UvmToolsEventControlData;

//------------------------------------------------------------------------------
// UVM Tools forward types (handles) definitions
//------------------------------------------------------------------------------
struct UvmToolsSession_tag;
struct UvmToolsEventQueue_tag;
struct UvmToolsCounters_tag;

typedef struct UvmToolsSession_tag UvmToolsSession;
typedef struct UvmToolsEventQueue_tag UvmToolsEventQueue;
typedef struct UvmToolsCounters_tag *UvmToolsCounters;

typedef UvmToolsSession *UvmToolsSessionHandle;
typedef UvmToolsEventQueue *UvmToolsEventQueueHandle; 
typedef UvmToolsCounters *UvmToolsCountersHandle; 

#endif // _UVMTYPES_H_
