/*******************************************************************************
    Copyright (c) 2015 NVIDIA Corporation

    Permission is hereby granted, free of charge, to any person obtaining a copy
    of this software and associated documentation files (the "Software"), to
    deal in the Software without restriction, including without limitation the
    rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
    sell copies of the Software, and to permit persons to whom the Software is
    furnished to do so, subject to the following conditions:

        The above copyright notice and this permission notice shall be
        included in all copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
    IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
    FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
    THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
    LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
    FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
    DEALINGS IN THE SOFTWARE.

*******************************************************************************/

#ifndef __UVM8_GPU_H__
#define __UVM8_GPU_H__

#include "nvtypes.h"
#include "nvmisc.h"
#include "uvmtypes.h"
#include "nv_uvm_types.h"
#include "uvm_linux.h"
#include "uvm_common.h"
#include "ctrl2080mc.h"
#include "uvm8_forward_decl.h"
#include "uvm8_pmm_gpu.h"
#include "uvm8_mmu.h"
#include "uvm8_gpu_page_fault.h"
#include "uvm8_hal_types.h"

#include <linux/bitmap.h>

// TODO: Replace by NV0080_MAX_DEVICES
#define UVM8_MAX_GPUS 32
#define UVM8_MAX_PROCESSORS (UVM8_MAX_GPUS + 1)
#define UVM8_MAX_UNIQUE_GPU_PAIRS SUM_FROM_0_TO_N(UVM8_MAX_GPUS - 1)

#if UVM_IS_NEXT()
    #include "uvm8_gpu_next.h"
#else
    struct uvm_gpu_next_data_struct
    {
        int not_used;
    };
#endif

// Static processor id for the CPU
#define UVM_CPU_ID (0)

// A unique uvm internal gpu id, guaranteed to be in the [1, UVM8_MAX_PROCESSORS) range
typedef NvU32 uvm_gpu_id_t;

// Helper to iterate over all valid gpu ids
#define for_each_gpu_id(i) for (i = 1; i < UVM8_MAX_PROCESSORS; ++i)

// Same as gpu id type but can also include the cpu id (UVM_CPU_ID)
typedef NvU32 uvm_processor_id_t;

// A collection of uvm processor ids
// Operated on with the uvm_processor_mask_*() functions below
typedef struct
{
    DECLARE_BITMAP(bitmap, UVM8_MAX_PROCESSORS);
} uvm_processor_mask_t;

// Buffer length to store uvm gpu id, RM device name and gpu uuid.
#define UVM_GPU_NICE_NAME_BUFFER_LENGTH (sizeof("ID 999: : ") + \
            UVM_GPU_NAME_LENGTH + UVM_GPU_UUID_TEXT_BUFFER_LENGTH)

typedef struct
{
    // Fault buffer information and structures provided by RM
    UvmGpuFaultInfo rm_info;

    // Maximum number of faults entries that can be stored in the buffer
    NvU32 max_faults;

    // Number of faults to be processed in batch before fetching new entries from the GPU buffer
    NvU32 fault_batch_count;

    // Cached value of the GPU GET register to minimize the round-trips over PCIe
    NvU32 cached_get;

    // Cached value of the GPU PUT register to minimize the round-trips over PCIe
    NvU32 cached_put;

    // Array of elements fetched from the GPU fault buffer. The number of elements in this array is exactly
    // fault_batch_count
    uvm_fault_buffer_entry_t *fault_cache;

    // Array of pointers to elements in fault cache used for fault preprocessing. The number of elements in this array
    // is exactly fault_batch_count
    uvm_fault_buffer_entry_t **ordered_fault_cache;

    // Policy that determines when GPU replays are issued during normal fault servicing
    uvm_perf_fault_replay_policy_t replay_policy;

    // Tracker used to aggregate replay operations, needed for fault cancel
    uvm_tracker_t replay_tracker;
} uvm_fault_buffer_info_t;

struct uvm_gpu_struct
{
    // Reference count for how many places are holding onto a GPU (internal to UVM driver).
    // This includes any GPUs we know about, not just GPUs that are registered with a VA space.
    // Most GPUs end up being registered, but there are brief periods when they are not
    // registered, such as during interrupt handling, and in add_gpu() or remove_gpu().
    struct kref gpu_kref;

    // Refcount of the gpu, i.e. how many times it has been retained. This is
    // roughly a count of how many times it has been registered with a VA space,
    // except that some paths retain the GPU temporarily without a VA space.
    //
    // While this is >0, the GPU can't be removed. This differs from gpu_kref,
    // which merely prevents the uvm_gpu_t object from being freed.
    //
    // In most cases this count is protected by the global lock: retaining a GPU
    // from a UUID and any release require the global lock to be taken. But it's
    // also useful for a caller to retain a GPU they've already retained, in
    // which case there's no need to take the global lock. This can happen when
    // an operation needs to drop the VA space lock but continue operating on a
    // GPU. This is an atomic variable to handle those cases.
    //
    // Security note: keep it as a 64-bit counter to prevent overflow cases (a
    // user can create a lot of va spaces and register the gpu with them).
    atomic64_t retained_count;

    // A unique uvm gpu id in range [1, UVM8_MAX_PROCESSORS)
    uvm_gpu_id_t id;

    // The gpu's uuid
    NvProcessorUuid uuid;

    // Nice printable name including the uvm gpu id, ascii name from RM and uuid
    char name[UVM_GPU_NICE_NAME_BUFFER_LENGTH];

    // Gpu architecture; NV2080_CTRL_MC_ARCH_INFO_ARCHITECTURE_*
    NvU32 architecture;
    // Gpu implementation; NV2080_CTRL_MC_ARCH_INFO_IMPLEMENTATION_*
    NvU32 implementation;

    // Host (gpfifo) class; *_CHANNEL_GPFIFO_*, e.g. KEPLER_CHANNEL_GPFIFO_A
    NvU32 host_class;
    // Copy engine (dma) class; *_DMA_COPY_*, e.g. KEPLER_DMA_COPY_A
    NvU32 ce_class;
    // Fault buffer class; *_FAULT_BUFFER_*, e.g. MAXWELL_FAULT_BUFFER_A
    NvU32 fault_buffer_class;

    // Hardware Abstraction Layer
    uvm_host_hal_t *host_hal;
    uvm_ce_hal_t *ce_hal;
    uvm_arch_hal_t *arch_hal;
    uvm_fault_buffer_hal_t *fault_buffer_hal;

    // The amount of video memory the GPU has in total, in bytes.
    NvU64 vidmem_size;

    struct
    {
        // Big page size used by the internal UVM VA space
        // Notably it may be different than the big page size used by a user's VA
        // space in general.
        NvU32 internal_size;

        // Whether big page mappings have the physical memory swizzled.
        // On some architectures (Kepler) physical memory mapped with a big page
        // size doesn't follow the common 1-1 mapping where each offset within the
        // big page maps to the same offset within the mapped physical memory.
        // Swizzling makes physically accessing memory mapped with big pages
        // infeasible and has to be worked around by creating virtual identity
        // mappings (see big_page_self_identity_mapping below).
        bool swizzling;

        // Big page self identity mapping
        // Used only on GPUs that have big page swizzling enabled. Notably
        // that's only Kepler which only supports one big page size at a time
        // and hence only a single mapping is needed.
        struct
        {
            // VA where the identity mapping should be mapped in the internal VA
            // space managed by address_space_tree (see below).
            NvU64 base;

            // Page tables with the mapping.
            uvm_page_table_range_vec_t *range_vec;
        } identity_mapping;
    } big_page;

    // Parameters used by the TLB batching API
    struct
    {
        // Is the targeted VA invalidate supported at all?
        NvBool va_invalidate_supported;

        // How many pages does it make sense to invalidate with the targeted VA
        // invalidate before falling back to invalidate all?
        NvU32 max_pages;
    } tlb_batch;

    // For the next chip and for any other features that are not yet ready to be
    // made public:
    uvm_gpu_next_data_t uvm_next;

    // Largest VA (exclusive) which can be used for channel buffer mappings
    NvU64 max_channel_va;

    // VA base and size of the RM managed part of the internal UVM VA space.
    //
    // The internal UVM VA is shared with RM by RM controlling some of the top
    // level PDEs and leaving the rest for UVM to control.
    // On Pascal a single top level PDE covers 128 TB of VA and given that
    // semaphores and other allocations limited to 40bit are currently allocated
    // through RM, RM needs to control the [0, 128TB) VA range at least for now.
    // On Kepler and Maxwell limit RMs VA to [0, 128GB) that should easily fit
    // all RM allocations and leave enough space for UVM.
    NvU64 rm_va_base;
    NvU64 rm_va_size;

    // Base and size of the GPU VA used for uvm_mem_t allocations mapped in the
    // internal address_space_tree.
    NvU64 uvm_mem_va_base;
    NvU64 uvm_mem_va_size;

    // RM address space handle used in many of the UVM/RM APIs
    // Represents both an RM device and a GPU VA in RM.
    uvmGpuAddressSpaceHandle rm_address_space;

    // Page tree used for the internal UVM VA space shared with RM
    uvm_page_tree_t address_space_tree;

    // Set to true during add_gpu() as soon as the RM's address space is moved
    // to the address_space_tree.
    bool rm_address_space_moved_to_page_tree;

    // ECC handling
    // In order to trap ECC errors as soon as possible the driver has the hw
    // interrupt register mapped directly. If an ECC interrupt is ever noticed
    // to be pending, then the UVM driver needs to:
    //
    //   1) ask RM to service interrupts, and then
    //   2) inspect the ECC error notifier state.
    //
    // Notably, checking for channel errors is not enough, because ECC errors
    // can be pending, even after a channel has become idle.
    //
    // See more details in uvm_gpu_check_ecc_error().
    struct
    {
        // Does the GPU have ECC enabled?
        bool enabled;

        // Direct mapping of the 32-bit part of the hw interrupt tree that has
        // the ECC bits.
        volatile NvU32 *hw_interrupt_tree_location;

        // Mask to get the ECC interrupt bits from the 32-bits above.
        NvU32 mask;

        // Set to true by RM when a fatal ECC error is encountered (requires
        // asking RM to service pending interrupts to be current).
        NvBool *error_notifier;
    } ecc;

    UvmGpuCopyEngineCaps ce_caps[UVM_COPY_ENGINE_COUNT_MAX];

    uvm_gpu_semaphore_pool_t *semaphore_pool;

    uvm_channel_manager_t *channel_manager;

    struct
    {
        // Procfs entry for the GPU directory
        struct proc_dir_entry *dir;

        // Procfs entry for the uuid symlink to the GPU directory
        struct proc_dir_entry *dir_uuid_symlink;

        // Procfs entry for the info file
        struct proc_dir_entry *info_file;
    } procfs;

    uvm_pmm_gpu_t pmm;

    // Protects against changes to the state of a GPU as it transitions from top-half to
    // bottom-half interrupt handler.
    uvm_mutex_t isr_lock;

    // There is exactly one workqueue per GPU. It is used for the ISR bottom half. Sometimes
    // it is busy, other times it is idle. Note that the Linux kernel has a dedicated kernel
    // thread per CPU, to service workqueues. So N CPUs will be servicing M GPUs, in general.
    struct work_struct workqueue;

    // This is set to true during add_gpu(), if the GPU supports replayable faults
    // (fault_buffer_hal is not NULL). It is set back to false during remove_gpu(). The
    // page_fault_interrupts_lock must be held, in order to read or write this variable.
    // This should be treated as a private variable for the interrupt handling routines.
    bool handling_replayable_faults;

    // Fault buffer info. This is only valid if supports_replayable_faults is set to true
    uvm_fault_buffer_info_t fault_buffer_info;

    // Protects the state of page fault interrupts (enabled/disabled) and
    // whether the GPU is currently handling them. Taken in both interrupt and
    // process context.
    uvm_spinlock_irqsave_t page_fault_interrupts_lock;

    // Number of times uvm_gpu_disable_replayable_faults() has been called
    // without a corresponding call to uvm_gpu_enable_replayable_faults(). If
    // this is >0, replayable page fault interrupts are disabled. This field is
    // protected by page_fault_interrupts_lock.
    NvU64 disable_intr_ref_count;

    // Number of top-half ISRs called for this GPU over its lifetime
    NvU64 interrupt_count;

    // Number of bottom-half invocations operating on this GPU over its lifetime
    NvU64 interrupt_count_bottom_half;

    // Table of all registered channels (instance pointers) under this GPU.
    // Converts from instance pointer to uvm_va_space_t. The bottom half reads
    // the table under the isr_lock, but a separate lock is necessary because
    // entries are added and removed from the table under the va_space lock, and
    // we can't take the isr_lock while holding the va_space lock.
    uvm_spinlock_t instance_ptr_table_lock;
    struct radix_tree_root instance_ptr_table;
};

typedef enum
{
    UVM_GPU_LINK_INVALID = 0,
    UVM_GPU_LINK_PCIE,
    UVM_GPU_LINK_NVLINK_1,
    UVM_GPU_LINK_MAX
} uvm_gpu_link_type_t;

struct uvm_gpu_peer_struct
{
    // Note: All the peer_caps fields in this global structure can be queried
    // if and only if the corresponding bit from "va_space.enabled_peers" bitmap is set.

    // Peer Id associated with this device w.r.t. to a peer GPU.
    // Note: peerId (A -> B) != peerId (B -> A)
    // peer_id[0] from min(gpu_id_1, gpi_id_2) -> max(gpu_id_1, gpi_id_2)
    // peer_id[1] from max(gpu_id_1, gpi_id_2) -> min(gpu_id_1, gpi_id_2)
    NvU8 peer_ids[2];

    // When this bit is set, peer-to-peer atomics between GPUs are supported
    // natively in hardware instead of being demoted to separate non-atomic
    // read-modify-write accesses.
    NvU8 atomic_supported : 1;

    // The link type between the peer GPUs, currently either PCIE or NVLINK1.
    uvm_gpu_link_type_t link_type;

    NvU64 registered_ref_count;

    // This handle gets populated when enable_peer_access successfully creates
    // an NV50_P2P object. disable_peer_access resets the same on the object deletion.
    NvHandle p2p_handle;
};

// Initialize global gpu state
NV_STATUS uvm_gpu_init(void);

// Deinitialize global state (called from module exit)
void uvm_gpu_exit(void);

// Note that there is a uvm_gpu_get() function defined in uvm_global.h to break
// a circular dep between global and gpu modules.

// Get a gpu by uuid. This returns NULL if the GPU is not present. This is the general purpose
// call that should be used normally.
//
// LOCKING: requires the global lock to be held for read, and also acquires and releases the
// gpu_table_lock.
uvm_gpu_t *uvm_gpu_get_by_uuid(NvProcessorUuid *gpu_uuid);

// Same as uvm_gpu_get_by_uuid, except that this one does not assert that the caller is holding
// the global_lock. This is a narrower purpose function, and is only intended for use by the
// top-half ISR, or other very limited cases.
uvm_gpu_t *uvm_gpu_get_by_uuid_locked(NvProcessorUuid *gpu_uuid);

// Retain a gpu by uuid
// Returns the retained uvm_gpu_t in gpu_out on success
// LOCKING: requires the global lock to be held for write
NV_STATUS uvm_gpu_retain_by_uuid(NvProcessorUuid *gpu_uuid, uvm_gpu_t **gpu_out);

// Retain a gpu which is known to already be retained. Does NOT require the
// global lock to be held.
void uvm_gpu_retain(uvm_gpu_t *gpu);

// Release a gpu
// LOCKING: requires the global lock to be held for write
void uvm_gpu_release_locked(uvm_gpu_t *gpu);

// Like uvm_gpu_release_locked, but takes and releases the global lock in write
// mode for the caller.
void uvm_gpu_release(uvm_gpu_t *gpu);

// Helper which calls uvm_gpu_retain on each GPU in mask
void uvm_gpu_retain_mask(const uvm_processor_mask_t *mask);

// Helper which calls uvm_gpu_release_locked on each GPU in mask
void uvm_gpu_release_mask_locked(const uvm_processor_mask_t *mask);

// Like uvm_gpu_release_mask_locked, but takes and releases the global lock in
// write mode for the caller.
void uvm_gpu_release_mask(const uvm_processor_mask_t *mask);

static NvU64 uvm_gpu_retained_count(uvm_gpu_t *gpu)
{
    return atomic64_read(&gpu->retained_count);
}

// Decrease the refcount on the GPU object, and actually delete the object if the refcount hits
// zero.
void uvm_gpu_kref_put(uvm_gpu_t *gpu);

// Calculates peer table index using GPU ids.
NvU32 uvm_gpu_peer_table_index(uvm_gpu_id_t gpu_id1, uvm_gpu_id_t gpu_id2);

// Retains existing peer entry. Otherwise, create a new entry.
// LOCKING: requires the global lock to be held for write
NV_STATUS uvm_gpu_retain_peer_access(uvm_gpu_t *gpu_1, uvm_gpu_t *gpu_2);

// Releases a peer entry.
// LOCKING: requires the global lock to be held for write
void uvm_gpu_release_peer_access(uvm_gpu_t *gpu_1, uvm_gpu_t *gpu_2);

// Get the aperture for local_gpu to use to map memory resident on remote_gpu.
// They must not be the same gpu.
uvm_aperture_t uvm_gpu_peer_aperture(uvm_gpu_t *local_gpu, uvm_gpu_t *remote_gpu);

// Check for ECC errors
// Notably this check cannot be performed from the top half interrupt handler
NV_STATUS uvm_gpu_check_ecc_error(uvm_gpu_t *gpu);

static bool uvm_gpu_is_gk110_plus(uvm_gpu_t *gpu)
{
    return gpu->architecture >= NV2080_CTRL_MC_ARCH_INFO_ARCHITECTURE_GK110;
}

static bool uvm_processor_uuid_eq(const NvProcessorUuid *uuid1, const NvProcessorUuid *uuid2)
{
    return memcmp(uuid1, uuid2, sizeof(*uuid1)) == 0;
}

// Copies a UUID from source (src) to destination (dst).
static void uvm_processor_uuid_copy(NvProcessorUuid *dst, NvProcessorUuid *src)
{
    memcpy(dst, src, sizeof(*dst));
}

static bool uvm_processor_mask_test(const uvm_processor_mask_t *mask, uvm_processor_id_t id)
{
    UVM_ASSERT_MSG(id < UVM8_MAX_PROCESSORS, "id %u\n", id);

    return test_bit(id, mask->bitmap);
}

// uvm_processor_mask_set and uvm_processor_mask_clear assume to modify the mask
// under a lock, so they use the non-atomic variants __set_bit and __clear_bit.
static void uvm_processor_mask_set(uvm_processor_mask_t *mask, uvm_processor_id_t id)
{
    UVM_ASSERT_MSG(id < UVM8_MAX_PROCESSORS, "id %u\n", id);

    __set_bit(id, mask->bitmap);
}

static void uvm_processor_mask_clear(uvm_processor_mask_t *mask, uvm_processor_id_t id)
{
    UVM_ASSERT_MSG(id < UVM8_MAX_PROCESSORS, "id %u\n", id);

    __clear_bit(id, mask->bitmap);
}

static void uvm_processor_mask_zero(uvm_processor_mask_t *mask)
{
    bitmap_zero(mask->bitmap, UVM8_MAX_PROCESSORS);
}

static bool uvm_processor_mask_empty(const uvm_processor_mask_t *mask)
{
    return bitmap_empty(mask->bitmap, UVM8_MAX_PROCESSORS);
}

static void uvm_processor_mask_copy(uvm_processor_mask_t *dst, const uvm_processor_mask_t *src)
{
    bitmap_copy(dst->bitmap, src->bitmap, UVM8_MAX_PROCESSORS);
}

// Returns the first id that's set in the mask or UVM8_MAX_PROCESSORS if there isn't one.
static uvm_processor_id_t uvm_processor_mask_find_first_id(const uvm_processor_mask_t *mask)
{
    return find_first_bit(mask->bitmap, UVM8_MAX_PROCESSORS);
}

// Returns the first id greater or equal to min_id that's set in the mask or
// UVM8_MAX_PROCESSORS if there isn't one.
static uvm_processor_id_t uvm_processor_mask_find_next_id(const uvm_processor_mask_t *mask,
                                                          uvm_processor_id_t min_id)
{
    return find_next_bit(mask->bitmap, UVM8_MAX_PROCESSORS, min_id);
}

// Get the number of processors set in the mask
static NvU32 uvm_processor_mask_get_count(const uvm_processor_mask_t *mask)
{
    return bitmap_weight(mask->bitmap, UVM8_MAX_PROCESSORS);
}

// Get the number of GPUs set in the mask
static NvU32 uvm_processor_mask_get_gpu_count(const uvm_processor_mask_t *mask)
{
    NvU32 gpu_count = uvm_processor_mask_get_count(mask);
    if (uvm_processor_mask_test(mask, UVM_CPU_ID))
        --gpu_count;
    return gpu_count;
}

// Helper to iterate over all processor ids set in a mask
#define for_each_id_in_mask(id, mask)                                       \
    for (id = uvm_processor_mask_find_first_id(mask);                       \
         id != UVM8_MAX_PROCESSORS;                                         \
         id = uvm_processor_mask_find_next_id(mask, id + 1))

// Helper to iterate over all GPU ids set in a mask
#define for_each_gpu_id_in_mask(gpu_id, mask)                               \
    for (gpu_id = uvm_processor_mask_find_next_id(mask, UVM_CPU_ID + 1);    \
         gpu_id != UVM8_MAX_PROCESSORS;                                     \
         gpu_id = uvm_processor_mask_find_next_id(mask, gpu_id + 1))

// Returns whether the given address is within the GPU's maximum addressable VA
// range. Warning: This only checks whether the GPU's MMU can support the given
// address. Some HW units on that GPU might only support a smaller range.
//
// The GPU must be initialized before calling this function.
bool uvm_gpu_can_address(uvm_gpu_t *gpu, NvU64 addr);

// This function is only valid after uvm_hal_init_gpu has been called on the given gpu
bool uvm_gpu_supports_replayable_faults(uvm_gpu_t *gpu);

// Debug print of GPU properties
void uvm_gpu_print(uvm_gpu_t *gpu);

// Add the given instance pointer -> va_space mapping to this GPU. The bottom
// half GPU page fault handler uses this to look up the VA space for GPU faults.
NV_STATUS uvm_gpu_add_instance_ptr(uvm_gpu_t *gpu, uvm_gpu_phys_address_t instance_ptr, uvm_va_space_t *va_space);
void uvm_gpu_remove_instance_ptr(uvm_gpu_t *gpu, uvm_gpu_phys_address_t instance_ptr);

// Looks up an entry added by uvm_gpu_add_instance_ptr, or NULL if none.
uvm_va_space_t *uvm_gpu_instance_ptr_to_va_space(uvm_gpu_t *gpu, uvm_gpu_phys_address_t instance_ptr);

// If the input list of detached user channels is not empty, this flushes the
// fault buffer on the GPU to remove stale instance pointer entries, then frees
// all channels in the list using uvm_user_channel_destroy_detached.
//
// LOCKING: Takes the GPU isr_lock.
void uvm_gpu_destroy_detached_channels(uvm_gpu_t *gpu, struct list_head *detached_channels);

// Like uvm_gpu_destroy_detached_channels, except all GPUs in the mask have
// their fault buffers flushed before any channels are destroyed. The channels
// thus do not have to come from the same GPU.
void uvm_gpu_destroy_detached_channels_mask(const uvm_processor_mask_t *mask, struct list_head *detached_channels);

#endif // __UVM8_GPU_H__
