/*******************************************************************************
    Copyright (c) 2015 NVIDIA Corporation

    Permission is hereby granted, free of charge, to any person obtaining a copy
    of this software and associated documentation files (the "Software"), to
    deal in the Software without restriction, including without limitation the
    rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
    sell copies of the Software, and to permit persons to whom the Software is
    furnished to do so, subject to the following conditions:

        The above copyright notice and this permission notice shall be
        included in all copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
    IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
    FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
    THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
    LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
    FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
    DEALINGS IN THE SOFTWARE.

*******************************************************************************/

#include "linux/sort.h"
#include "nv_uvm_interface.h"
#include "uvm8_global.h"
#include "uvm8_gpu_page_fault.h"
#include "uvm8_hal.h"
#include "uvm8_kvmalloc.h"
#include "uvm8_va_block.h"
#include "uvm8_va_range.h"
#include "uvm8_va_space.h"
#include "uvm_common.h"
#include "uvm8_next_decl.h"

static void uvm8_isr_bottom_half(struct work_struct *work);

#define UVM_PERF_FAULT_BATCH_COUNT_MIN 1
#define UVM_PERF_FAULT_BATCH_COUNT_DEFAULT 8

// Number of entries that are fetched from the GPU fault buffer and serviced in batch
static unsigned uvm_perf_fault_batch_count = UVM_PERF_FAULT_BATCH_COUNT_DEFAULT;
module_param(uvm_perf_fault_batch_count, uint, S_IRUGO);

#define UVM_PERF_FAULT_REPLAY_POLICY_DEFAULT UVM_PERF_FAULT_REPLAY_POLICY_BLOCK

// Policy that determines when to issue fault replays
static uvm_perf_fault_replay_policy_t uvm_perf_fault_replay_policy;
module_param(uvm_perf_fault_replay_policy, uint, UVM_PERF_FAULT_REPLAY_POLICY_DEFAULT);

NV_STATUS uvm_gpu_fault_buffer_init(uvm_gpu_t *gpu)
{
    NV_STATUS status = NV_OK;

    uvm_assert_mutex_locked(&g_uvm_global.global_lock);
    UVM_ASSERT(gpu->fault_buffer_hal != NULL);

    uvm_tracker_init(&gpu->fault_buffer_info.replay_tracker);

    status = uvm_rm_locked_call(nvUvmInterfaceInitFaultInfo(gpu->rm_address_space,
                                                            &gpu->fault_buffer_info.rm_info));
    if (status != NV_OK) {
        UVM_ERR_PRINT("Failed to init fault buffer info from RM: %s, GPU %s\n",
                      nvstatusToString(status), gpu->name);
        goto fail;
    }

    gpu->fault_buffer_info.max_faults = gpu->fault_buffer_info.rm_info.bufferSize /
                                        gpu->fault_buffer_hal->entry_size(gpu);

    // Check provided module parameter value
    gpu->fault_buffer_info.fault_batch_count = max(uvm_perf_fault_batch_count,
                                                   (NvU32)UVM_PERF_FAULT_BATCH_COUNT_MIN);
    gpu->fault_buffer_info.fault_batch_count = min(gpu->fault_buffer_info.fault_batch_count,
                                                   gpu->fault_buffer_info.max_faults);

    if (gpu->fault_buffer_info.fault_batch_count != uvm_perf_fault_batch_count) {
        pr_info("Invalid uvm_perf_fault_batch_count value on GPU %s: %u. Valid range [%u:%u] Using %u instead\n",
                gpu->name, uvm_perf_fault_batch_count,
                UVM_PERF_FAULT_BATCH_COUNT_MIN, gpu->fault_buffer_info.max_faults,
                gpu->fault_buffer_info.fault_batch_count);
    }

    gpu->fault_buffer_info.fault_cache = uvm_kvmalloc_zero(gpu->fault_buffer_info.max_faults *
                                                           sizeof(*gpu->fault_buffer_info.fault_cache));
    if (!gpu->fault_buffer_info.fault_cache) {
        status = NV_ERR_NO_MEMORY;
        goto fail;
    }

    gpu->fault_buffer_info.ordered_fault_cache =
        uvm_kvmalloc_zero(gpu->fault_buffer_info.max_faults * sizeof(*gpu->fault_buffer_info.ordered_fault_cache));
    if (!gpu->fault_buffer_info.ordered_fault_cache) {
        status = NV_ERR_NO_MEMORY;
        goto fail;
    }

    status = uvm_rm_locked_call(nvUvmInterfaceOwnPageFaultIntr((NvU8*)&gpu->uuid,
                                                               sizeof(gpu->uuid),
                                                               NV_TRUE));
    if (status != NV_OK) {
        UVM_ERR_PRINT("Failed to take page fault ownership from RM: %s, GPU %s\n",
                      nvstatusToString(status), gpu->name);
        goto fail;
    }

    // Read current get/put pointers as this might not be the first time we have taken control of the fault buffer
    // since the GPU was initialized
    gpu->fault_buffer_info.cached_get = UVM_READ_ONCE(*gpu->fault_buffer_info.rm_info.pFaultBufferGet);
    gpu->fault_buffer_info.cached_put = UVM_READ_ONCE(*gpu->fault_buffer_info.rm_info.pFaultBufferPut);

    if (uvm_perf_fault_replay_policy >= UVM_PERF_FAULT_REPLAY_POLICY_MAX)
        gpu->fault_buffer_info.replay_policy = UVM_PERF_FAULT_REPLAY_POLICY_DEFAULT;

    if (gpu->fault_buffer_info.replay_policy != uvm_perf_fault_replay_policy) {
        pr_info("Invalid uvm_perf_fault_replay_policy value on GPU %s: %d. Using %d instead\n",
                gpu->name, uvm_perf_fault_replay_policy, gpu->fault_buffer_info.replay_policy);
    }

    return NV_OK;

fail:
    if (gpu->fault_buffer_info.rm_info.faultBufferHandle) {
        uvm_rm_locked_call_void(nvUvmInterfaceDestroyFaultInfo(gpu->rm_address_space,
                                                               &gpu->fault_buffer_info.rm_info));

        uvm_tracker_deinit(&gpu->fault_buffer_info.replay_tracker);

        // Signal the destructor that we did not acquire the ownership of GPU faults and fault info is not initialized.
        // No need to release ownership or destroy fault info
        gpu->fault_buffer_info.rm_info.faultBufferHandle = 0;
    }

    uvm_gpu_fault_buffer_deinit(gpu);

    return status;
}

void uvm_gpu_fault_buffer_deinit(uvm_gpu_t *gpu)
{
    NV_STATUS status = NV_OK;

    uvm_assert_mutex_locked(&g_uvm_global.global_lock);

    if (gpu->fault_buffer_info.rm_info.faultBufferHandle) {
        status = uvm_rm_locked_call(nvUvmInterfaceOwnPageFaultIntr((NvU8*)&gpu->uuid, sizeof(gpu->uuid), NV_FALSE));
        UVM_ASSERT(status == NV_OK);

        uvm_rm_locked_call_void(nvUvmInterfaceDestroyFaultInfo(gpu->rm_address_space,
                                                               &gpu->fault_buffer_info.rm_info));

        UVM_ASSERT(uvm_tracker_is_completed(&gpu->fault_buffer_info.replay_tracker) ||
                   uvm_global_get_status() != NV_OK);
        uvm_tracker_deinit(&gpu->fault_buffer_info.replay_tracker);

        gpu->fault_buffer_info.rm_info.faultBufferHandle = 0;
    }

    uvm_kvfree(gpu->fault_buffer_info.fault_cache);
    uvm_kvfree(gpu->fault_buffer_info.ordered_fault_cache);
    gpu->fault_buffer_info.fault_cache = NULL;
    gpu->fault_buffer_info.ordered_fault_cache = NULL;
}


// This is called from RM's top-half ISR (see: the nvidia_isr() function), and UVM is given a
// chance to handle the interrupt, before most of the RM processing. UVM communicates what it
// did, back to RM, via the return code:
//
//     NV_OK:
//         UVM handled an interrupt.
//
//     NV_WARN_MORE_PROCESSING_REQUIRED:
//         UVM did not schedule a bottom half, because it was unable to get the locks it
//         needed, but there is still UVM work to be done. RM will return "not handled" to the
//         Linux kernel, *unless* RM handled other faults in its top half. In that case, the
//         fact that UVM did not handle its interrupt is lost. However, life and interrupt
//         processing continues anyway: the GPU will soon raise another interrupt, because
//         that's what it does when there are replayable page faults remaining (GET != PUT in
//         the fault buffer).
//
//     NV_ERR_NO_INTR_PENDING:
//         UVM did not find any work to do. Currently this is handled in RM in exactly the same
//         way as NV_WARN_MORE_PROCESSING_REQUIRED is handled. However, the extra precision is
//         available for the future. RM's interrupt handling tends to evolve as new chips and
//         new interrupts get created.

NV_STATUS uvm8_isr_top_half(NvProcessorUuid *gpu_uuid)
{
    uvm_gpu_t *gpu;
    NV_STATUS status = NV_ERR_NO_INTR_PENDING;

    if (!in_interrupt()) {
        // Early-out if not in interrupt context. This happens with
        // CONFIG_DEBUG_SHIRQ enabled where the interrupt handler is called as
        // part of its removal to make sure it's prepared for being called even
        // when it's being freed.
        // This breaks the assumption that the UVM driver is called in atomic
        // context only in the interrupt context, which
        // uvm_thread_context_retain() relies on.
        return NV_OK;
    }

    if (!gpu_uuid) {
        // This can happen early in the main GPU driver initialization, because
        // that involves testing interrupts before the GPU is fully set up.
        return status;
    }

    uvm_spin_lock_irqsave(&g_uvm_global.gpu_table_lock);

    gpu = uvm_gpu_get_by_uuid_locked(gpu_uuid);

    if (gpu == NULL) {
        uvm_spin_unlock_irqrestore(&g_uvm_global.gpu_table_lock);
        return status;
    }

    kref_get(&gpu->gpu_kref);
    uvm_spin_unlock_irqrestore(&g_uvm_global.gpu_table_lock);

    // We don't need an atomic to increment this count since only this top half
    // writes it, and only one top half can run per GPU at a time.
    ++gpu->interrupt_count;

    // Now that we got a GPU object, lock it so that it can't be removed without us noticing.
    uvm_spin_lock_irqsave(&gpu->page_fault_interrupts_lock);


    // gpu->handling_replayable_faults gets set to false during removal, so quit if the GPU is
    // in the process of being removed.
    if (!gpu->handling_replayable_faults)
        goto done_no_bottom_half;

    // TODO: add support to lockdep, for leaving this lock acquired (the bottom half eventually
    // releases it).
    if (mutex_trylock(&gpu->isr_lock.m) == 0) {
        status = NV_WARN_MORE_PROCESSING_REQUIRED;
        goto done_no_bottom_half;
    }

    if (!uvm_replayable_faults_pending(gpu)) {
        mutex_unlock(&gpu->isr_lock.m);
        goto done_no_bottom_half;
    }

    uvm_gpu_disable_replayable_faults(gpu);

    // Schedule a bottom half, but do *not* release the GPU ISR lock. The bottom half releases
    // the GPU ISR lock as part of its cleanup.
    NV_WORKQUEUE_INIT_AND_SCHEDULE(&gpu->workqueue, uvm8_isr_bottom_half);

    uvm_spin_unlock_irqrestore(&gpu->page_fault_interrupts_lock);

    // Keep the isr_lock, and the gpu_kref count, and run the bottom half:
    return NV_OK;

done_no_bottom_half:
    uvm_spin_unlock_irqrestore(&gpu->page_fault_interrupts_lock);
    uvm_gpu_kref_put(gpu);

    return status;
}

bool uvm_replayable_faults_pending(uvm_gpu_t *gpu)
{
    UVM_ASSERT(uvm_gpu_supports_replayable_faults(gpu));

    // Fast path 1: we left some faults unserviced in the buffer in the last pass
    if (gpu->fault_buffer_info.cached_get != gpu->fault_buffer_info.cached_put)
        return true;

    // Fast path 2: read the valid bit of the fault buffer entry pointed by the cached get pointer
    if (!gpu->fault_buffer_hal->entry_is_valid(gpu, gpu->fault_buffer_info.cached_get)) {
        // Slow path: read the put pointer from the GPU register via BAR0 over PCIe
        gpu->fault_buffer_info.cached_put = UVM_READ_ONCE(*gpu->fault_buffer_info.rm_info.pFaultBufferPut);

        // No interrupt pending
        if (gpu->fault_buffer_info.cached_get == gpu->fault_buffer_info.cached_put)
            return false;
    }

    return true;
}

static NV_STATUS push_replay_on_gpu(uvm_gpu_t *gpu, uvm_fault_replay_type_t type)
{
    NV_STATUS status;
    uvm_push_t push;

    status = uvm_push_begin(gpu->channel_manager, UVM_CHANNEL_TYPE_MEMOPS, &push, "Replaying faults");
    if (status != NV_OK)
        return status;

    gpu->host_hal->replay_faults(&push, type);

    uvm_push_end(&push);

    // Add this push to the GPU's replay tracker so cancel can wait on it.
    status = uvm_tracker_add_push_safe(&gpu->fault_buffer_info.replay_tracker, &push);

    return status;
}

NV_STATUS uvm_gpu_fault_buffer_flush(uvm_gpu_t *gpu)
{
    NV_STATUS status = NV_OK;

    UVM_ASSERT(uvm_gpu_supports_replayable_faults(gpu));

    // Disables replayable fault interrupts and fault servicing
    uvm_gpu_isr_lock(gpu);

    status = uvm_gpu_fault_buffer_flush_locked(gpu, true, UVM_FAULT_REPLAY_TYPE_START);

    // This will trigger the top half to start servicing faults again, if the
    // replay brought any back in
    uvm_gpu_isr_unlock(gpu);
    return status;
}

NV_STATUS uvm_gpu_fault_buffer_flush_locked(uvm_gpu_t *gpu, bool update_put, uvm_fault_replay_type_t fault_replay)
{
    NvU32 get;
    NvU32 put;

    // TODO: right now uvm locks do not support the synchronization method used by top and bottom ISR. Add uvm lock
    // assert when it's supported. Use plain mutex kernel utilities for now.
    UVM_ASSERT(uvm_gpu_supports_replayable_faults(gpu));
    UVM_ASSERT(mutex_is_locked(&gpu->isr_lock.m));

    // Read PUT pointer from the GPU if requested
    if (update_put)
        gpu->fault_buffer_info.cached_put = UVM_READ_ONCE(*gpu->fault_buffer_info.rm_info.pFaultBufferPut);

    get = gpu->fault_buffer_info.cached_get;
    put = gpu->fault_buffer_info.cached_put;

    while (get != put) {
        // waiting until valid bit is set. TODO: Do something smarter for the busy-wait (Bug 1710855)
        while (!gpu->fault_buffer_hal->entry_is_valid(gpu, get))
            cpu_relax();

        gpu->fault_buffer_hal->entry_clear_valid(gpu, get);
        ++get;
        if (get == gpu->fault_buffer_info.max_faults)
            get = 0;
    }

    gpu->fault_buffer_info.cached_get = get;

    // Update get pointer on the GPU
    UVM_WRITE_ONCE(*gpu->fault_buffer_info.rm_info.pFaultBufferGet, get);

    // Issue fault replay
    return push_replay_on_gpu(gpu, fault_replay);
}

static NvU32 fetch_fault_buffer_entries(uvm_gpu_t *gpu)
{
    NvU32 get;
    NvU32 put;
    NvU32 i;
    NvU32 cached_faults;
    uvm_fault_buffer_entry_t *fault_cache;

    // TODO: right now uvm locks do not support the synchronization method used by top and bottom ISR. Add uvm lock
    // assert when it's supported. Use plain mutex kernel utilities for now.
    UVM_ASSERT(mutex_is_locked(&gpu->isr_lock.m));
    UVM_ASSERT(uvm_gpu_supports_replayable_faults(gpu));

    fault_cache = gpu->fault_buffer_info.fault_cache;

    get = gpu->fault_buffer_info.cached_get;

    // Read put pointer from GPU and cache it
    if (get == gpu->fault_buffer_info.cached_put)
        gpu->fault_buffer_info.cached_put = UVM_READ_ONCE(*gpu->fault_buffer_info.rm_info.pFaultBufferPut);

    put = gpu->fault_buffer_info.cached_put;

    if (get == put)
        return 0;

    // Parse until get != put and have enough space to cache.
    cached_faults = 0;
    for (i = 0; (get != put) && (i < gpu->fault_buffer_info.fault_batch_count); ++i) {
        // We cannot just wait for the last entry (the one pointed by put) to become valid, we have to do it
        // individually since entries can be written out of order
        while (!gpu->fault_buffer_hal->entry_is_valid(gpu, get)) {
            // Keep waiting if no entry parsed. TODO: Do something smarter (Bug 1710855)
            if (!cached_faults)
                cpu_relax();
            else
                goto done;  // We have some entry to work on. Let's do the rest later
        }

        // Prevent later accesses being moved above the read of the valid bit
        smp_mb__after_atomic();

        // Got valid bit set. Let's cache.
        gpu->fault_buffer_hal->parse_entry(gpu, get, &fault_cache[i]);

        // Make sure that are fields in the entry are properly initialized
        fault_cache[i].waived = false;
        fault_cache[i].va_space = NULL;
        fault_cache[i].is_fatal = (fault_cache[i].fault_type >= UVM_FAULT_TYPE_FATAL);

        ++cached_faults;
        ++get;
        if (get == gpu->fault_buffer_info.max_faults)
            get = 0;
    }

done:
    gpu->fault_buffer_info.cached_get = get;

    // Update get pointer on the GPU
    UVM_WRITE_ONCE(*gpu->fault_buffer_info.rm_info.pFaultBufferGet, get);

    return cached_faults;
}

#define CMP_DEFAULT(a,b)                  \
({                                        \
    typeof(a) _a = a;                     \
    typeof(b) _b = b;                     \
    int __ret;                            \
    BUILD_BUG_ON(sizeof(a) != sizeof(b)); \
    if (_a < _b)                          \
        __ret = -1;                       \
    else if (_a > _b)                     \
        __ret = 1;                        \
    else                                  \
        __ret = 0;                        \
                                          \
    __ret;                                \
 })

// Compare two gpu physical addresses
static inline int cmp_gpu_phys_addr(uvm_gpu_phys_address_t a, uvm_gpu_phys_address_t b)
{
    int result;

    result = CMP_DEFAULT(a.aperture, b.aperture);
    if (result != 0)
        return result;

    return CMP_DEFAULT(a.address, b.address);
}

// Compare two VA spaces
static inline int cmp_va_space(uvm_va_space_t *a, uvm_va_space_t *b)
{
    return CMP_DEFAULT(a, b);
}

// Compare two virtual addresses
static inline int cmp_addr(NvU64 a, NvU64 b)
{
    return CMP_DEFAULT(a, b);
}

// Compare two fault access types
static inline int cmp_access_type(uvm_fault_access_type_t a, uvm_fault_access_type_t b)
{
    UVM_ASSERT(a > 0 && a < UVM_FAULT_ACCESS_TYPE_MAX);
    UVM_ASSERT(b > 0 && b < UVM_FAULT_ACCESS_TYPE_MAX);

    // Check that fault access type enum values are ordered by "intrusiveness"
    BUILD_BUG_ON(UVM_FAULT_ACCESS_TYPE_ATOMIC >= UVM_FAULT_ACCESS_TYPE_WRITE);
    BUILD_BUG_ON(UVM_FAULT_ACCESS_TYPE_WRITE >= UVM_FAULT_ACCESS_TYPE_READ);
    BUILD_BUG_ON(UVM_FAULT_ACCESS_TYPE_READ >= UVM_FAULT_ACCESS_TYPE_PREFETCH);

    return CMP_DEFAULT(a, b);
}

// Sort comparator for pointers to fault buffer entries that sorts by instance pointer
static int cmp_sort_fault_entry_by_instance_ptr(const void *_a, const void *_b)
{
    const uvm_fault_buffer_entry_t **a = (const uvm_fault_buffer_entry_t **)_a;
    const uvm_fault_buffer_entry_t **b = (const uvm_fault_buffer_entry_t **)_b;

    return cmp_gpu_phys_addr((*a)->instance_ptr, (*b)->instance_ptr);
}

// Sort comparator for pointers to fault buffer entries that sorts by va_space, fault address and fault access type
static int cmp_sort_fault_entry_by_va_space_address_access_type(const void *_a, const void *_b)
{
    const uvm_fault_buffer_entry_t **a = (const uvm_fault_buffer_entry_t **)_a;
    const uvm_fault_buffer_entry_t **b = (const uvm_fault_buffer_entry_t **)_b;

    int result;

    result = cmp_va_space((*a)->va_space, (*b)->va_space);
    if (result != 0)
        return result;

    result = cmp_addr((*a)->fault_address, (*b)->fault_address);
    if (result != 0)
        return result;

    return cmp_access_type((*a)->fault_access_type, (*b)->fault_access_type);
}

// Translate all instance pointers to VA spaces. Since the buffer is ordered by instance_ptr, we minimize the number of
// translations
static void translate_instance_ptrs(uvm_gpu_t *gpu, uvm_fault_buffer_entry_t **ordered_fault_cache,
                                    NvU32 cached_faults, NvU32 *fatal_faults)
{
    uvm_gpu_phys_address_t prev_instance_ptr = { 0, 0 };
    NvU32 i;

    *fatal_faults = 0;

    for (i = 0; i < cached_faults; ++i) {
        uvm_fault_buffer_entry_t *current_entry;

        current_entry = ordered_fault_cache[i];

        if (cmp_gpu_phys_addr(current_entry->instance_ptr, prev_instance_ptr) != 0) {
            // If instance_ptr is different, make a new translation
            current_entry->va_space = uvm_gpu_instance_ptr_to_va_space(gpu, current_entry->instance_ptr);
        }
        else {
            current_entry->va_space = ordered_fault_cache[i - 1]->va_space;
        }

        // If the va_space is gone just ignore the faults
        if (current_entry->va_space == NULL) {
            current_entry->is_fatal = true;
            ++(*fatal_faults);
        }
    }
}

// Fault cache preprocessing for fault coalescing
//
// This function generates an ordered view of the given fault_cache in which faults are sorted by VA space, fault
// address (aligned to 4K) and access type "intrusiveness" (atomic - write - read - prefetch). In order to minimize
// the number of instance_ptr to VA space translations we perform a first sort by instance_ptr.
//
// Current scheme:
// 1) sort by instance_ptr
// 2) translate all instance_ptrs to VA spaces
// 3) sort by va_space, fault address (GPU already reports 4K-aligned address) and access type
//
// TODO: decide what should we do with prefetch requests
static void preprocess_fault_batch(uvm_gpu_t *gpu, NvU32 cached_faults, NvU32 *fatal_faults)
{
    NvU32 i;
    uvm_fault_buffer_entry_t *fault_cache;
    uvm_fault_buffer_entry_t **ordered_fault_cache;

    UVM_ASSERT(cached_faults > 0);

    fault_cache = gpu->fault_buffer_info.fault_cache;
    ordered_fault_cache = gpu->fault_buffer_info.ordered_fault_cache;

    // Generate an ordered view of the fault cache in ordered_fault_cache. We sort the pointers, not the entries
    // in fault_cache

    // Initialize pointers before they are sorted
    for (i = 0; i < cached_faults; ++i)
        ordered_fault_cache[i] = &fault_cache[i];

    // 1) sort by instance_ptr
    sort(ordered_fault_cache, cached_faults, sizeof(*ordered_fault_cache),
         cmp_sort_fault_entry_by_instance_ptr, NULL);

    // 2) translate all instance_ptrs to VA spaces
    translate_instance_ptrs(gpu, ordered_fault_cache, cached_faults, fatal_faults);

    // 3) sort by va_space, fault address (GPU already reports 4K-aligned address) and access type
    sort(ordered_fault_cache, cached_faults, sizeof(*ordered_fault_cache),
         cmp_sort_fault_entry_by_va_space_address_access_type, NULL);
}

// We notify the fault event for all faults within the block so that the performance heuristics are updated. Then, all
// required actions for the block data are performed by the performance heuristics code
static NV_STATUS service_fault_batch_block(uvm_gpu_t *gpu, uvm_va_block_t *va_block,
                                           NvU32 first_fault_index, NvU32 batch_faults,
                                           NvU32 *block_faults, NvU32 *block_fatal_faults)
{
    NV_STATUS status = NV_OK;
    NvU32 first_fault = first_fault_index;
    NvU32 i;
    uvm_fault_buffer_entry_t **ordered_fault_cache;

    *block_faults = 0;
    *block_fatal_faults = 0;

    ordered_fault_cache = gpu->fault_buffer_info.ordered_fault_cache;

    uvm_mutex_lock(&va_block->lock);

    // Check that the va_block is still valid
    UVM_ASSERT(va_block->va_range);

    // Scan the sorted array and notify the fault event for all fault entries
    for (i = first_fault; (i < batch_faults) && (ordered_fault_cache[i]->fault_address <= va_block->end); ++i) {
        uvm_perf_event_data_t event_data;

        event_data.fault.block = va_block;
        event_data.fault.space = va_block->va_range->va_space;
        event_data.fault.proc_id = gpu->id;
        event_data.fault.gpu.gpu = gpu;
        event_data.fault.gpu.buffer_entry = ordered_fault_cache[i];

        uvm_perf_event_notify(&va_block->va_range->va_space->perf_events, UVM_PERF_EVENT_FAULT, &event_data);

        ++(*block_faults);
    }

    // TODO: implement actual fault handling

    uvm_mutex_unlock(&va_block->lock);

    return status;
}

static void push_cancel_on_gpu(uvm_gpu_t *gpu, uvm_gpu_phys_address_t instance_ptr, NvU32 gpc_id, NvU32 client_id)
{
    NV_STATUS status;
    uvm_push_t push;

    status = uvm_push_begin(gpu->channel_manager, UVM_CHANNEL_TYPE_ANY, &push,
                            "Pushing targeted cancel, GPU %s", gpu->name);
    UVM_ASSERT(status == NV_OK);

    gpu->host_hal->cancel_faults_targeted(&push, instance_ptr, gpc_id, client_id);

    uvm_push_end(&push);
}

// Scan the ordered view of faults and group them by different va_blocks. Service faults for each va_block, in batch
static NV_STATUS service_fault_batch(uvm_gpu_t *gpu, NvU32 batch_faults, NvU32 *fatal_faults, NvU32 *replays)
{
    NV_STATUS status = NV_OK;
    NvU32 i;
    uvm_va_space_t *va_space = NULL;
    uvm_fault_buffer_entry_t **ordered_fault_cache = gpu->fault_buffer_info.ordered_fault_cache;

    UVM_ASSERT(uvm_gpu_supports_replayable_faults(gpu));

    *fatal_faults = 0;
    *replays = 0;
    for (i = 0; i < batch_faults;) {
        uvm_va_block_t *va_block;
        uvm_fault_buffer_entry_t *current_entry;
        bool is_new_va_space;

        current_entry = ordered_fault_cache[i];

        UVM_ASSERT(current_entry->va_space);

        is_new_va_space = false;
        if (current_entry->va_space != va_space) {
            // Fault on a different va_space, drop the lock of the old one...
            if (va_space != NULL)
                uvm_up_read(&va_space->lock);

            // ... and take the lock of the new one
            uvm_down_read(&current_entry->va_space->lock);

            va_space = current_entry->va_space;
            is_new_va_space = true;
        }

        status = uvm_va_block_find_create(current_entry->va_space, current_entry->fault_address, &va_block);
        if (status == NV_OK) {
            NvU32 block_faults;
            NvU32 block_fatal_faults;

            status = service_fault_batch_block(gpu, va_block, i, batch_faults, &block_faults, &block_fatal_faults);

            if (status == NV_OK && gpu->fault_buffer_info.replay_policy == UVM_PERF_FAULT_REPLAY_POLICY_BLOCK) {
                status = push_replay_on_gpu(gpu, UVM_FAULT_REPLAY_TYPE_START);
                ++(*replays);
            }

            if (status != NV_OK)
                goto fail;

            *fatal_faults += block_fatal_faults;

            i += block_faults;
        }
        else {
            // Avoid dropping fault events when the VA block is not found or cannot be created
            uvm_perf_event_data_t event_data;

            event_data.fault.block = NULL;
            event_data.fault.space = va_space;
            event_data.fault.proc_id = gpu->id;
            event_data.fault.gpu.gpu = gpu;
            event_data.fault.gpu.buffer_entry = current_entry;

            uvm_perf_event_notify(&va_space->perf_events, UVM_PERF_EVENT_FAULT, &event_data);

            if (status == NV_ERR_INVALID_ADDRESS) {
                // If the VA block cannot be found, set the fatal fault flag
                current_entry->is_fatal = true;

                ++(*fatal_faults);

                ++i;

                // Not an internal error, just a fatal fault
                status = NV_OK;
            }
            else {
                // If VA block creation failed, exit early
                goto fail;
            }
        }
    }

fail:
    if (va_space != NULL)
        uvm_up_read(&va_space->lock);

    return status;
}

static void cancel_fault_batch(uvm_gpu_t *gpu, NvU32 cached_faults)
{
    NvU32 i;

    for (i = 0; i < cached_faults; ++i) {
        uvm_fault_buffer_entry_t *current_entry;

        current_entry = &gpu->fault_buffer_info.fault_cache[i];
        push_cancel_on_gpu(gpu, current_entry->instance_ptr, current_entry->fault_source.gpc_id,
                           current_entry->fault_source.client_id);
    }
}

static NV_STATUS service_fault_buffer(uvm_gpu_t *gpu)
{
    NvU32 replays = 0;
    NV_STATUS status = NV_OK;

    UVM_ASSERT(uvm_gpu_supports_replayable_faults(gpu));

    // Process all faults in the buffer
    while (1) {
        NvU32 cached_faults;
        NvU32 fatal_faults_stage;
        NvU32 fatal_faults = 0;
        NvU32 batch_replays = 0;

        cached_faults = fetch_fault_buffer_entries(gpu);

        if (cached_faults == 0)
            break;

        preprocess_fault_batch(gpu, cached_faults, &fatal_faults_stage);
        fatal_faults += fatal_faults_stage;

        // If UVM_NEXT is servicing the fault buffer, it will return something *other*
        // than NV_ERR_NOT_SUPPORTED.
        status = uvm_next_service_fault_batch(gpu, cached_faults, &fatal_faults_stage, &batch_replays);

        if (status == NV_ERR_NOT_SUPPORTED)
            status = service_fault_batch(gpu, cached_faults, &fatal_faults_stage, &batch_replays);

        if (status != NV_OK)
            break;

        replays += batch_replays;
        fatal_faults += fatal_faults_stage;

        // This is a temporary workaround to make applications gracefully fail
        cancel_fault_batch(gpu, cached_faults);

        if (gpu->fault_buffer_info.replay_policy == UVM_PERF_FAULT_REPLAY_POLICY_BATCH) {
            status = push_replay_on_gpu(gpu, UVM_FAULT_REPLAY_TYPE_START);
            ++replays;
            if (status != NV_OK)
                break;
        }
        else if (gpu->fault_buffer_info.replay_policy == UVM_PERF_FAULT_REPLAY_POLICY_BATCH_FLUSH) {
            status = uvm_gpu_fault_buffer_flush_locked(gpu, false, UVM_FAULT_REPLAY_TYPE_START);
            ++replays;
            break;
        }
    }

    // Make sure that we issue at least one replay if no replay has been issued yet to avoid dropping faults that do
    // not show up in the buffer
    if ((status == NV_OK && gpu->fault_buffer_info.replay_policy == UVM_PERF_FAULT_REPLAY_POLICY_ONCE) || replays == 0)
        status = push_replay_on_gpu(gpu, UVM_FAULT_REPLAY_TYPE_START);

    return status;
}

static void uvm8_isr_bottom_half(struct work_struct *work)
{
    uvm_gpu_t *gpu = container_of(work, uvm_gpu_t, workqueue);

    // Multiple bottom halves can be running concurrently, but only one can be
    // running here for a given GPU since we enter with the isr_lock held.
    ++gpu->interrupt_count_bottom_half;

    // TODO (here, or within service_fault_buffer): acquire and release the va_space lock.
    service_fault_buffer(gpu);

    uvm_gpu_isr_unlock(gpu);
    uvm_gpu_kref_put(gpu);
}

void uvm_gpu_disable_replayable_faults(uvm_gpu_t *gpu)
{
    uvm_assert_spinlock_locked(&gpu->page_fault_interrupts_lock);
    UVM_ASSERT(gpu->handling_replayable_faults);

    if (gpu->disable_intr_ref_count == 0)
        gpu->fault_buffer_hal->disable_replayable_faults(gpu);
    ++gpu->disable_intr_ref_count;
}

void uvm_gpu_enable_replayable_faults(uvm_gpu_t *gpu)
{
    uvm_assert_spinlock_locked(&gpu->page_fault_interrupts_lock);
    UVM_ASSERT(gpu->handling_replayable_faults);
    UVM_ASSERT(gpu->disable_intr_ref_count > 0);

    --gpu->disable_intr_ref_count;
    if (gpu->disable_intr_ref_count == 0)
        gpu->fault_buffer_hal->enable_replayable_faults(gpu);
}

void uvm_gpu_isr_lock(uvm_gpu_t *gpu)
{
    UVM_ASSERT(uvm_gpu_retained_count(gpu) > 0);

    uvm_spin_lock_irqsave(&gpu->page_fault_interrupts_lock);

    // Bump the disable ref count. This guarantees that the bottom half or
    // another thread trying to take the isr_lock won't inadvertently re-enable
    // interrupts during this locking sequence.
    if (gpu->handling_replayable_faults)
        uvm_gpu_disable_replayable_faults(gpu);

    uvm_spin_unlock_irqrestore(&gpu->page_fault_interrupts_lock);

    // Now that we know replayable fault interrupts can't get enabled, take the
    // lock. This has to be a raw call without the uvm_lock.h wrappers: although
    // this function is called from non-interrupt context, the corresponding
    // uvm_gpu_isr_unlock() function is also used by the bottom half, which
    // pairs its unlock with the raw call in the top half.
    mutex_lock(&gpu->isr_lock.m);
}

void uvm_gpu_isr_unlock(uvm_gpu_t *gpu)
{
    UVM_ASSERT(atomic_read(&gpu->gpu_kref.refcount) > 0);

    uvm_spin_lock_irqsave(&gpu->page_fault_interrupts_lock);

    // The following sequence is delicate:
    //
    //     1) Enable replayable page fault interrupts
    //     2) Unlock GPU isr_lock (mutex)
    //     3) Unlock page_fault_interrupts_lock (spin lock)
    //
    // ...because the moment that page fault interrupts are reenabled, the top half will start
    // receiving them. As the gpu->isr_lock is still held, the top half will start returning
    // NV_WARN_MORE_PROCESSING_REQUIRED, due to failing the attempted
    // mutex_trylock(&gpu->isr_lock). This can lead to an interrupt storm, which will cripple
    // the system, and often cause Linux to permanently disable the GPU's interrupt line.
    //
    // In order to avoid such an interrupt storm, the gpu->page_fault_interrupts_lock (which
    // is acquired via spinlock_irqsave, thus disabling local CPU interrupts) is held until
    // after releasing the ISR mutex. That way, once local interrupts are enabled, the mutex
    // is available for the top half to acquire. This avoids a storm on the local CPU, but
    // still allows a small window of high interrupts to occur, if another CPU handles the
    // interrupt. However, in that cause, the local CPU is not being slowed down
    // (interrupted), and you'll notice that the very next instruction after enabling page
    // fault interrupts is to unlock the ISR mutex. Such a small window may allow a few
    // interrupts, but not enough to be any sort of problem.

    if (gpu->handling_replayable_faults) {
        // Turn page fault interrupts back on, unless remove_gpu() has already removed this GPU
        // from the GPU table. remove_gpu() indicates that situation by setting
        // gpu->handling_replayable_faults to false.
        //
        // This path can only be taken from the bottom half. User threads
        // calling this function must have previously retained the GPU, so they
        // can't race with remove_gpu.
        //
        // TODO: Assert that we're in a bottom half thread, once that's tracked
        //       by the lock assertion code.
        //
        // Note that if we're in the bottom half and the GPU was removed before
        // we checked handling_replayable_faults, we won't drop our interrupt
        // disable ref ount from the corresponding top-half call to
        // uvm_gpu_disable_replayable_faults. That's ok because remove_gpu
        // ignores the refcount after waiting for the bottom half to finish.
        uvm_gpu_enable_replayable_faults(gpu);
    }

    // Raw unlock call, to correspond to the raw lock call in the top half:
    mutex_unlock(&gpu->isr_lock.m);

    uvm_spin_unlock_irqrestore(&gpu->page_fault_interrupts_lock);
}
