/*******************************************************************************
    Copyright (c) 2015 NVIDIA Corporation

    Permission is hereby granted, free of charge, to any person obtaining a copy
    of this software and associated documentation files (the "Software"), to
    deal in the Software without restriction, including without limitation the
    rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
    sell copies of the Software, and to permit persons to whom the Software is
    furnished to do so, subject to the following conditions:

        The above copyright notice and this permission notice shall be
        included in all copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
    IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
    FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
    THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
    LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
    FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
    DEALINGS IN THE SOFTWARE.

*******************************************************************************/

#ifndef __UVM8_LOCK_H__
#define __UVM8_LOCK_H__

#include "uvm8_forward_decl.h"
#include "uvm_linux.h"
#include "uvm_common.h"

// --------------------------- UVM Locking Order ---------------------------- //
//
// Any locks described here should have their locking order added to
// uvm_lock_order_t below.
//
// - mmap_sem
//      Order: UVM_LOCK_ORDER_MMAP_SEM
//      Reader/writer lock (rw_semaphore)
//
//      We're often called with the kernel already holding mmap_sem: mmap,
//      munmap, fault, etc. These operations may have to take any number of UVM
//      locks, so mmap_sem must be at the top of our lock order since it's
//      sometimes out of our control.
//
//      We need to hold mmap_sem when calling vm_insert_page, which means that
//      any time an operation (such as an ioctl) might need to install a CPU
//      mapping, it must take current->mm->mmap_sem in read mode very early on.
//
//      However, current->mm is not necessarily the owning mm of the UVM vma.
//      fork or fd passing via a UNIX doman socket can cause that. Notably, this
//      is also the case when handling GPU faults from a kernel thread. This
//      means we must lock current->mm->mmap_sem, then look up the UVM vma and
//      compare its mm before operating on that vma.
//
// - Global driver state lock (g_uvm_global.global_lock)
//      Order: UVM_LOCK_ORDER_GLOBAL
//      Exclusive lock (mutex)
//
//      This protects state associated with GPUs, such as the P2P table
//      and instance pointer mappings.
//
//      This should be taken whenever global GPU state might need to be modified.
//
// - GPU ISR lock (gpu->isr_lock)
//      Order: UVM_LOCK_ORDER_ISR
//      Reader/writer lock (rw_semaphore) per gpu
//
//      Protects:
//      - Changes to the state of a GPU as it transitions from top-half to bottom-half
//        interrupt handler. This lock is acquired for that GPU, in the ISR top-half. Then a
//        bottom-half is scheduled (to run in a workqueue). Then the bottom-half releases the
//        lock when that GPU's processing appears to be done.
//
// - VA space lock (va_space->lock)
//      Order: UVM_LOCK_ORDER_VA_SPACE
//      Reader/writer lock (rw_semaphore) per uvm_va_space (UVM struct file)
//
//      This is the UVM equivalent of mmap_sem. It protects all state under that
//      va_space, such as the VA range tree.
//
//      Read mode: Faults (CPU and GPU), mapping creation, prefetches. These
//      will be serialized at the VA block level if necessary.
//
//      Write mode: Modification of the range state such as mmap and changes to
//      logical permissions or location preferences.
//
// - VA block lock (va_block->lock)
//      Order: UVM_LOCK_ORDER_VA_BLOCK
//      Exclusive lock (mutex)
//
//      Protects:
//      - CPU and GPU page table mappings for all VAs under the block
//      - Updates to the GPU work tracker for that block (migrations)
//
//      Operations allowed while holding the lock:
//      - CPU allocation (we don't evict CPU memory)
//      - GPU memory allocation which cannot evict
//      - CPU page table mapping/unmapping
//      - Pushing work (GPU page table mapping/unmapping)
//
//      Operations not allowed while holding the lock:
//      - GPU memory allocation which can evict memory (would require nesting
//        block locks)
//
// - GPU semaphore pool lock (semaphore_pool->mutex)
//      Order: UVM_LOCK_ORDER_GPU_SEMAPHORE_POOL
//      Exclusive lock (mutex) per uvm_gpu_semaphore_pool
//
//      Protects the state of the semaphore pool.
//
// - RM API lock
//      Order: UVM_LOCK_ORDER_RM_API
//      Exclusive lock
//
//      This is an internal RM lock that's acquired by most if not all UVM-RM
//      APIs.
//      Notably this lock is also held on PMA eviction.
//
// - Physical lock (chunk->lock)
//      Order: UVM_LOCK_ORDER_PHYS
//      Exclusive lock (bit lock) per chunk (region of fixed-sized
//      naturally-aligned physical memory).
//
//      In most cases (outside eviction heuristic and PMM) it can be assumed
//      to protect the state of chunk.
//
//      Operations allowed while holding the lock:
//      - CPU allocation (we don't evict CPU memory)
//      - GPU memory allocation which cannot evict
//      - CPU page table mapping/unmapping
//      - Pushing work (GPU page table mapping/unmapping)
//
//      Operations not allowed while holding the lock:
//      - GPU memory allocation which can evict memory (would require nesting
//        physical locks)
//
// - RM GPUs lock
//      Order: UVM_LOCK_ORDER_RM_GPUS
//      Exclusive lock
//
//      This is an internal RM lock that's acquired by most if not all UVM-RM
//      APIs and disables interrupts for the GPUs.
//      Notably this lock is *not* held on PMA eviction.
//
// - Page tree lock
//      Order: UVM_LOCK_ORDER_PAGE_TREE
//      Exclusive lock per GPU page tree
//
//      This protects a page tree.  All modifications to the device's page tree
//      and the host-side cache of that tree must be done under this lock.
//      The host-side cache and device state must be consistent when this lock is released
//
//      Operations allowed while holding this lock
//      - Pushing work
//
//      Operations not allowed while holding this lock
//      - GPU memory allocation which can evict
//
// - Concurrent push semaphore
//      Order: UVM_LOCK_ORDER_PUSH
//      Semaphore (uvm_semaphore_t)
//
//      This is a semaphore limiting the amount of concurrent pushes that is
//      held for the duration of a push (between uvm_push_begin*() and
//      uvm_push_end()).
//
// - PMM GPU lock (pmm->lock)
//      Order: UVM_LOCK_ORDER_PMM
//      Exclusive lock (mutex) per uvm_pmm_gpu_t
//
//      Protects the state of PMM - internal to PMM.
//
//     Operations allowed while holding the lock:
//     - Query heuristic for page eviction
//
// - Channel lock
//      Order: UVM_LOCK_ORDER_CHANNEL
//      Spinlock (uvm_spinlock_t)
//
// - Leaf locks
//      Order: UVM_LOCK_ORDER_LEAF
//
//      All leaf locks.
//
// -------------------------------------------------------------------------- //

// Remember to add any new lock orders to uvm_lock_order_to_string() in uvm8_lock.c
typedef enum
{
    UVM_LOCK_ORDER_INVALID = 0,
    UVM_LOCK_ORDER_MMAP_SEM,
    UVM_LOCK_ORDER_GLOBAL,
    UVM_LOCK_ORDER_ISR,
    UVM_LOCK_ORDER_VA_SPACE,
    UVM_LOCK_ORDER_VA_BLOCK,
    UVM_LOCK_ORDER_GPU_SEMAPHORE_POOL,
    UVM_LOCK_ORDER_RM_API,
    UVM_LOCK_ORDER_RM_GPUS,
    UVM_LOCK_ORDER_PHYS,
    UVM_LOCK_ORDER_PAGE_TREE,
    UVM_LOCK_ORDER_PUSH,
    UVM_LOCK_ORDER_PMM,
    UVM_LOCK_ORDER_CHANNEL,
    UVM_LOCK_ORDER_LEAF,
    UVM_LOCK_ORDER_COUNT,
} uvm_lock_order_t;

const char *uvm_lock_order_to_string(uvm_lock_order_t lock_order);

// Record locking a lock of given lock_order in exclusive or shared mode
// Returns true if the recorded lock follows all the locking rules and false otherwise.
bool __uvm_record_lock(void *lock, uvm_lock_order_t lock_order, bool exclusive);

// Record unlocking a lock of given lock_order in exclusive or shared mode and
// possibly out of order.
// Returns true if the unlock follows all the locking rules and false otherwise.
bool __uvm_record_unlock(void *lock, uvm_lock_order_t lock_order, bool exclusive, bool out_of_order);

bool __uvm_record_downgrade(void *lock, uvm_lock_order_t lock_order);

// Check whether a lock of given lock_order is held in exclusive or shared mode
// by the current thread. Notably holding a lock in exclusive mode also
// satisfies holding it in shared mode.
bool __uvm_check_locked(void *lock, uvm_lock_order_t lock_order, bool exclusive);

// Check that no locks are held with the given or deeper lock order
bool __uvm_check_unlocked_order(uvm_lock_order_t lock_order);

// Check that all locks have been released in a uvm thread context
bool __uvm_check_all_unlocked(uvm_thread_context_t *uvm_context);

// Check that all locks have been released in the current uvm thread context
bool __uvm_thread_check_all_unlocked(void);

#if UVM_IS_DEBUG()
  // These macros are intended to be expanded on the call site directly and will print
  // the precise location of the violation while the __uvm_record* functions will error print the details.
  #define uvm_record_lock_raw(lock, lock_order, exclusive) \
      UVM_ASSERT_MSG(__uvm_record_lock((lock), (lock_order), (exclusive)), "Locking violation\n")
  #define uvm_record_unlock_raw(lock, lock_order, exclusive, out_of_order) \
      UVM_ASSERT_MSG(__uvm_record_unlock((lock), (lock_order), (exclusive), (out_of_order)), "Locking violation\n")
  #define uvm_record_downgrade_raw(lock, lock_order) \
      UVM_ASSERT_MSG(__uvm_record_downgrade((lock), (lock_order)), "Locking violation\n")

  // Record UVM lock (a lock that has a lock_order member) operation and assert that it's correct
  #define uvm_record_lock(lock, exclusive) uvm_record_lock_raw((lock), (lock)->lock_order, (exclusive))
  #define uvm_record_unlock(lock, exclusive) uvm_record_unlock_raw((lock), (lock)->lock_order, (exclusive), false)
  #define uvm_record_downgrade(lock) uvm_record_downgrade_raw((lock), (lock)->lock_order)

  // Check whether a UVM lock (a lock that has a lock_order member) is held in shared or exclusive mode
  #define uvm_check_locked(lock, exclusive) __uvm_check_locked((lock), (lock)->lock_order, (exclusive))

  // Helpers for recording and asserting mmap_sem state
  #define uvm_record_lock_mmap_sem_read(mmap_sem) \
          uvm_record_lock_raw((mmap_sem), UVM_LOCK_ORDER_MMAP_SEM, false)

  #define uvm_record_unlock_mmap_sem_read(mmap_sem) \
          uvm_record_unlock_raw((mmap_sem), UVM_LOCK_ORDER_MMAP_SEM, false, false)

  #define uvm_record_unlock_mmap_sem_read_out_of_order(mmap_sem) \
          uvm_record_unlock_raw((mmap_sem), UVM_LOCK_ORDER_MMAP_SEM, false, true)

  #define uvm_record_lock_mmap_sem_write(mmap_sem) \
          uvm_record_lock_raw((mmap_sem), UVM_LOCK_ORDER_MMAP_SEM, true)

  #define uvm_record_unlock_mmap_sem_write(mmap_sem) \
          uvm_record_unlock_raw((mmap_sem), UVM_LOCK_ORDER_MMAP_SEM, true, false)

  #define uvm_record_unlock_mmap_sem_write_out_of_order(mmap_sem) \
          uvm_record_unlock_raw((mmap_sem), UVM_LOCK_ORDER_MMAP_SEM, true, true)

  #define uvm_check_locked_mmap_sem(mmap_sem, exclusive) \
           __uvm_check_locked((mmap_sem), UVM_LOCK_ORDER_MMAP_SEM, (exclusive))

  #define uvm_assert_mmap_sem_locked_read(mmap_sem) ({                                      \
          typeof(mmap_sem) _sem = (mmap_sem);                                               \
          UVM_ASSERT(rwsem_is_locked(_sem) && uvm_check_locked_mmap_sem(_sem, false));      \
      })

  #define uvm_assert_mmap_sem_locked_write(mmap_sem) ({                                     \
          typeof(mmap_sem) _sem = (mmap_sem);                                               \
          UVM_ASSERT(rwsem_is_locked(_sem) && uvm_check_locked_mmap_sem(_sem, true));       \
      })

  // Helpers for recording RM API lock usage around UVM-RM interfaces
  #define uvm_record_lock_rm_api() \
          uvm_record_lock_raw((void*)UVM_LOCK_ORDER_RM_API, UVM_LOCK_ORDER_RM_API, true)
  #define uvm_record_unlock_rm_api() \
          uvm_record_unlock_raw((void*)UVM_LOCK_ORDER_RM_API, UVM_LOCK_ORDER_RM_API, true, false)

  // Helpers for recording RM GPUS lock usage around UVM-RM interfaces
  #define uvm_record_lock_rm_gpus() \
          uvm_record_lock_raw((void*)UVM_LOCK_ORDER_RM_GPUS, UVM_LOCK_ORDER_RM_GPUS, true)
  #define uvm_record_unlock_rm_gpus() \
          uvm_record_unlock_raw((void*)UVM_LOCK_ORDER_RM_GPUS, UVM_LOCK_ORDER_RM_GPUS, true, false)

  // Helpers for recording both RM locks usage around UVM-RM interfaces
  #define uvm_record_lock_rm_all() ({ uvm_record_lock_rm_api(); uvm_record_lock_rm_gpus(); })
  #define uvm_record_unlock_rm_all() ({ uvm_record_unlock_rm_gpus(); uvm_record_unlock_rm_api(); })

#else
  #define uvm_record_lock                               UVM_IGNORE_EXPR2
  #define uvm_record_unlock                             UVM_IGNORE_EXPR2
  #define uvm_record_downgrade                          UVM_IGNORE_EXPR

  static bool uvm_check_locked(void *lock, bool exclusive)
  {
      return false;
  }

  #define uvm_record_lock_mmap_sem_read                 UVM_IGNORE_EXPR
  #define uvm_record_unlock_mmap_sem_read               UVM_IGNORE_EXPR
  #define uvm_record_unlock_mmap_sem_read_out_of_order  UVM_IGNORE_EXPR
  #define uvm_record_lock_mmap_sem_write                UVM_IGNORE_EXPR
  #define uvm_record_unlock_mmap_sem_write              UVM_IGNORE_EXPR
  #define uvm_record_unlock_mmap_sem_write_out_of_order UVM_IGNORE_EXPR

  #define uvm_check_locked_mmap_sem                     uvm_check_locked

  #define uvm_assert_mmap_sem_locked_read               UVM_IGNORE_EXPR
  #define uvm_assert_mmap_sem_locked_write              UVM_IGNORE_EXPR

  #define uvm_record_lock_rm_api()
  #define uvm_record_unlock_rm_api()

  #define uvm_record_lock_rm_gpus()
  #define uvm_record_unlock_rm_gpus()

  #define uvm_record_lock_rm_all()
  #define uvm_record_unlock_rm_all()
#endif

#define uvm_thread_assert_all_unlocked() UVM_ASSERT(__uvm_thread_check_all_unlocked())
#define uvm_assert_unlocked_order(order) UVM_ASSERT(__uvm_check_unlocked_order(order))

// Helpers for locking mmap_sem and recording its usage
#define uvm_down_read_mmap_sem(mmap_sem) ({             \
        typeof(mmap_sem) _sem = (mmap_sem);             \
        uvm_record_lock_mmap_sem_read(_sem);            \
        down_read(_sem);                                \
    })

#define uvm_up_read_mmap_sem(mmap_sem) ({               \
        typeof(mmap_sem) _sem = (mmap_sem);             \
        up_read(_sem);                                  \
        uvm_record_unlock_mmap_sem_read(_sem);          \
    })

#define uvm_up_read_mmap_sem_out_of_order(mmap_sem) ({      \
        typeof(mmap_sem) _sem = (mmap_sem);                 \
        up_read(_sem);                                      \
        uvm_record_unlock_mmap_sem_read_out_of_order(_sem); \
    })

// Helper for calling a UVM-RM interface function with lock recording
#define uvm_rm_locked_call(call) ({                     \
        typeof(call) ret;                               \
        uvm_record_lock_rm_all();                       \
        ret = call;                                     \
        uvm_record_unlock_rm_all();                     \
        ret;                                            \
    })

// Helper for calling a UVM-RM interface function that returns void with lock recording
#define uvm_rm_locked_call_void(call) ({                \
        uvm_record_lock_rm_all();                       \
        call;                                           \
        uvm_record_unlock_rm_all();                     \
    })

typedef struct
{
    struct rw_semaphore sem;
#if UVM_IS_DEBUG()
    uvm_lock_order_t lock_order;
#endif
} uvm_rw_semaphore_t;

// Asserts that the semaphore is held
#define uvm_assert_rwsem_locked_read(uvm_sem) ({                                    \
        typeof(uvm_sem) _sem = (uvm_sem);                                           \
        UVM_ASSERT(rwsem_is_locked(&_sem->sem) && uvm_check_locked(_sem, false));   \
    })

#define uvm_assert_rwsem_locked_write(uvm_sem) ({                                   \
        typeof(uvm_sem) _sem = (uvm_sem);                                           \
        UVM_ASSERT(rwsem_is_locked(&_sem->sem) && uvm_check_locked(_sem, true));    \
    })

#define uvm_assert_rwsem_unlocked(uvm_sem) UVM_ASSERT(!rwsem_is_locked(&(uvm_sem)->sem))

static void uvm_init_rwsem(uvm_rw_semaphore_t *uvm_sem, uvm_lock_order_t lock_order)
{
    init_rwsem(&uvm_sem->sem);
#if UVM_IS_DEBUG()
    uvm_sem->lock_order = lock_order;
#endif
    uvm_assert_rwsem_unlocked(uvm_sem);
}

static void __uvm_down_read(uvm_rw_semaphore_t *uvm_sem)
{
    down_read(&uvm_sem->sem);
    uvm_assert_rwsem_locked_read(uvm_sem);
}

#define uvm_down_read(sem) ({                           \
        typeof(sem) _sem = (sem);                       \
        uvm_record_lock(_sem, false);                   \
        __uvm_down_read(_sem);                          \
    })

static void __uvm_up_read(uvm_rw_semaphore_t *uvm_sem)
{
    uvm_assert_rwsem_locked_read(uvm_sem);
    up_read(&uvm_sem->sem);
}
#define uvm_up_read(sem) ({                             \
        typeof(sem) _sem = (sem);                       \
        __uvm_up_read(_sem);                            \
        uvm_record_unlock(_sem, false);                 \
    })

static void __uvm_down_write(uvm_rw_semaphore_t *uvm_sem)
{
    down_write(&uvm_sem->sem);
    uvm_assert_rwsem_locked_write(uvm_sem);
}
#define uvm_down_write(sem) ({                          \
        typeof (sem) _sem = (sem);                      \
        uvm_record_lock(_sem, true);                    \
        __uvm_down_write(_sem);                         \
    })

// trylock for writing: returns 1 if successful, 0 if not:
static int __uvm_down_write_trylock(uvm_rw_semaphore_t *uvm_sem)
{
    int ret = down_write_trylock(&uvm_sem->sem);
    if (ret == 0)
        return ret;

    uvm_assert_rwsem_locked_write(uvm_sem);

    return ret;
}
#define uvm_down_write_trylock(sem) ({                  \
        typeof(sem) _sem = (sem);                       \
        int locked;                                     \
        uvm_record_lock(_sem, true);                    \
        locked = __uvm_down_write_trylock(_sem);        \
        if (locked == 0)                                \
            uvm_record_unlock(_sem, true);              \
        locked;                                         \
    })

static void __uvm_up_write(uvm_rw_semaphore_t *uvm_sem)
{
    uvm_assert_rwsem_locked_write(uvm_sem);
    up_write(&uvm_sem->sem);
}
#define uvm_up_write(sem) ({                            \
        typeof(sem) _sem = (sem);                       \
        __uvm_up_write(_sem);                           \
        uvm_record_unlock(_sem, true);                  \
    })

static void __uvm_downgrade_write(uvm_rw_semaphore_t *uvm_sem)
{
    uvm_assert_rwsem_locked_write(uvm_sem);
    downgrade_write(&uvm_sem->sem);
}
#define uvm_downgrade_write(sem) ({                     \
        typeof(sem) _sem = (sem);                       \
        __uvm_downgrade_write(_sem);                    \
        uvm_record_downgrade(_sem);                     \
    })

typedef struct
{
    struct mutex m;
#if UVM_IS_DEBUG()
    uvm_lock_order_t lock_order;
#endif
} uvm_mutex_t;

#define uvm_assert_mutex_locked(uvm_mutex) ({                                       \
        typeof(uvm_mutex) _mutex = (uvm_mutex);                                     \
        UVM_ASSERT(mutex_is_locked(&_mutex->m) && uvm_check_locked(_mutex, true));  \
    })

#define uvm_assert_mutex_unlocked(uvm_mutex) UVM_ASSERT(!mutex_is_locked(&(uvm_mutex)->m))

static void uvm_mutex_init(uvm_mutex_t *mutex, uvm_lock_order_t lock_order)
{
    mutex_init(&mutex->m);
#if UVM_IS_DEBUG()
    mutex->lock_order = lock_order;
#endif
    uvm_assert_mutex_unlocked(mutex);
}

static void __uvm_mutex_lock(uvm_mutex_t *mutex)
{
    mutex_lock(&mutex->m);
    uvm_assert_mutex_locked(mutex);
}
#define uvm_mutex_lock(mutex) ({                        \
        typeof(mutex) _mutex = (mutex);                 \
        uvm_record_lock(_mutex, true);                  \
        __uvm_mutex_lock(_mutex);                       \
    })

// Lock w/o any tracking. This should be extremely rare and *_no_tracking
// helpers will be added only as needed.
#define uvm_mutex_lock_no_tracking(mutex) mutex_lock(&(mutex)->m)

static void __uvm_mutex_unlock(uvm_mutex_t *mutex)
{
    uvm_assert_mutex_locked(mutex);
    mutex_unlock(&mutex->m);
}
#define uvm_mutex_unlock(mutex) ({                      \
        typeof(mutex) _mutex = (mutex);                 \
        __uvm_mutex_unlock(_mutex);                     \
        uvm_record_unlock(_mutex, true);                \
    })

// Unlock w/o any tracking. This should be extremely rare and *_no_tracking
// helpers will be added only as needed.
#define uvm_mutex_unlock_no_tracking(mutex) mutex_unlock(&(mutex)->m)

typedef struct
{
    struct semaphore sem;
#if UVM_IS_DEBUG()
    uvm_lock_order_t lock_order;
#endif
} uvm_semaphore_t;

static void uvm_sema_init(uvm_semaphore_t *semaphore, int val, uvm_lock_order_t lock_order)
{
    sema_init(&semaphore->sem, val);
#if UVM_IS_DEBUG()
    semaphore->lock_order = lock_order;
#endif
}

static void __uvm_down(uvm_semaphore_t *semaphore)
{
    down(&semaphore->sem);
}
#define uvm_down(sem) ({                                \
        typeof(sem) _sem = (sem);                       \
        uvm_record_lock(_sem, false);                   \
        __uvm_down(_sem);                               \
    })

static void __uvm_up(uvm_semaphore_t *semaphore)
{
    up(&semaphore->sem);
}
#define uvm_up(sem) ({                                  \
        typeof(sem) _sem = (sem);                       \
        __uvm_up(_sem);                                 \
        uvm_record_unlock(_sem, false);                 \
    })

// A regular spinlock
// Locked/unlocked with uvm_spin_lock()/uvm_spin_unlock()
typedef struct
{
    spinlock_t lock;
#if UVM_IS_DEBUG()
    uvm_lock_order_t lock_order;
#endif
} uvm_spinlock_t;

// A separate spinlock type for spinlocks that need to disable interrupts. For
// guaranteed correctness and convenience embed the saved and restored irq state
// in the lock itself.
// Locked/unlocked with uvm_spin_lock_irqsave()/uvm_spin_unlock_irqrestore()
typedef struct
{
    spinlock_t lock;
    unsigned long irq_flags;
#if UVM_IS_DEBUG()
    uvm_lock_order_t lock_order;
#endif
} uvm_spinlock_irqsave_t;

// Asserts that the spinlock is held. Notably the macros below support both
// types of spinlocks.
#define uvm_assert_spinlock_locked(spinlock) ({                                     \
        typeof(spinlock) _lock = (spinlock);                                        \
        UVM_ASSERT(spin_is_locked(&_lock->lock) && uvm_check_locked(_lock, true));  \
    })

#define uvm_assert_spinlock_unlocked(spinlock) UVM_ASSERT(!spin_is_locked(&(spinlock)->lock))

static void uvm_spin_lock_init(uvm_spinlock_t *spinlock, uvm_lock_order_t lock_order)
{
    spin_lock_init(&spinlock->lock);
#if UVM_IS_DEBUG()
    spinlock->lock_order = lock_order;
#endif
    uvm_assert_spinlock_unlocked(spinlock);
}

static void __uvm_spin_lock(uvm_spinlock_t *spinlock)
{
    spin_lock(&spinlock->lock);
    uvm_assert_spinlock_locked(spinlock);
}
#define uvm_spin_lock(lock) ({                          \
        typeof(lock) _lock = (lock);                    \
        uvm_record_lock(_lock, true);                   \
        __uvm_spin_lock(_lock);                         \
    })

static void __uvm_spin_unlock(uvm_spinlock_t *spinlock)
{
    uvm_assert_spinlock_locked(spinlock);
    spin_unlock(&spinlock->lock);
}
#define uvm_spin_unlock(lock) ({                        \
        typeof(lock) _lock = (lock);                    \
        __uvm_spin_unlock(_lock);                       \
        uvm_record_unlock(_lock, true);                 \
    })

static void uvm_spin_lock_irqsave_init(uvm_spinlock_irqsave_t *spinlock, uvm_lock_order_t lock_order)
{
    spin_lock_init(&spinlock->lock);
#if UVM_IS_DEBUG()
    spinlock->lock_order = lock_order;
#endif
    uvm_assert_spinlock_unlocked(spinlock);
}

static void __uvm_spin_lock_irqsave(uvm_spinlock_irqsave_t *spinlock)
{
    // Use a temp to not rely on flags being written after acquiring the lock.
    unsigned long irq_flags;
    spin_lock_irqsave(&(spinlock)->lock, irq_flags);
    spinlock->irq_flags = irq_flags;
    uvm_assert_spinlock_locked(spinlock);
}
#define uvm_spin_lock_irqsave(lock) ({                  \
        typeof(lock) _lock = (lock);                    \
        uvm_record_lock(_lock, true);                   \
        __uvm_spin_lock_irqsave(_lock);                 \
    })

static void __uvm_spin_unlock_irqrestore(uvm_spinlock_irqsave_t *spinlock)
{
    // Use a temp to not rely on flags being read before releasing the lock.
    unsigned long irq_flags = spinlock->irq_flags;
    uvm_assert_spinlock_locked(spinlock);
    spin_unlock_irqrestore(&(spinlock)->lock, irq_flags);
}
#define uvm_spin_unlock_irqrestore(lock) ({             \
        typeof(lock) _lock = (lock);                    \
        __uvm_spin_unlock_irqrestore(_lock);            \
        uvm_record_unlock(_lock, true);                 \
    })

// Bit locks are 'compressed' mutexes which take only 1 bit by virtue of
// using shared waitqueues.

// Asserts that the bit lock is held.
//
// TODO:
//  - assert for the right ownership (defining the owner might be tricky in
//    the kernel).
//  - how to check for locking order.
#define uvm_assert_bit_locked(word, bit)   UVM_ASSERT(test_bit((bit), (word)))
#define uvm_assert_bit_unlocked(word, bit) UVM_ASSERT(!test_bit((bit), (word)))

static void uvm_bit_lock(unsigned long *word, unsigned long bit)
{
    int res;
    UVM_ASSERT(bit < BITS_PER_LONG);
    res = UVM_WAIT_ON_BIT_LOCK(word, bit, TASK_UNINTERRUPTIBLE);
    UVM_ASSERT_MSG(res == 0, "Uninterruptible task interrupted: %d\n", res);
    uvm_assert_bit_locked(word, bit);
}

static void uvm_bit_unlock(unsigned long *word, unsigned long bit)
{
    UVM_ASSERT(bit < BITS_PER_LONG);
    uvm_assert_bit_locked(word, bit);
    clear_bit_unlock(bit, word);
    // Make sure we don't reorder release with wakeup as it would cause
    // deadlocks (other thread checking lock and adding itself to queue
    // in reversed order). clear_bit_unlock has only release semantics.
    smp_mb__after_atomic();
    wake_up_bit(word, bit);
}

#endif // __UVM8_LOCK_H__
