/*******************************************************************************
    Copyright (c) 2016 NVIDIA Corporation

    Permission is hereby granted, free of charge, to any person obtaining a copy
    of this software and associated documentation files (the "Software"), to
    deal in the Software without restriction, including without limitation the
    rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
    sell copies of the Software, and to permit persons to whom the Software is
    furnished to do so, subject to the following conditions:

        The above copyright notice and this permission notice shall be
        included in all copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
    IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
    FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
    THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
    LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
    FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
    DEALINGS IN THE SOFTWARE.

*******************************************************************************/

#include "uvm_common.h"
#include "uvm_linux.h"
#include "uvm8_forward_decl.h"
#include "uvm8_lock.h"
#include "uvm8_mmu.h"
#include "uvm8_api.h"
#include "uvm8_global.h"
#include "uvm8_gpu.h"
#include "uvm8_push.h"
#include "uvm8_va_space.h"
#include "uvm8_va_range.h"
#include "uvm8_tracker.h"
#include "uvm8_hal.h"
#include "uvm8_hal_types.h"
#include "uvm8_map_external.h"
#include "uvm8_pte_batch.h"
#include "uvm8_tlb_batch.h"
#include "nv_uvm_interface.h"

#include "uvm8_pushbuffer.h"

// Assume almost all of the push space can be used for PTEs leaving 1K of margin.
#define MAX_COPY_SIZE_PER_PUSH ((size_t)(UVM_MAX_PUSH_SIZE - 1024))


static NV_STATUS get_rm_ptes(uvm_va_range_t *va_range,
                             uvm_gpu_t *gpu,
                             NvU64 map_offset,
                             NvU64 map_size,
                             UvmGpuExternalMappingInfo *mapping_info)
{
    uvm_gpu_va_space_t *gpu_va_space = uvm_gpu_va_space_get(va_range->va_space, gpu);
    uvm_ext_gpu_map_t *ext_gpu_map;
    NV_STATUS status;

    // Both queries operate on the same UvmGpuExternalMappingInfo type
    if (va_range->type == UVM_VA_RANGE_TYPE_CHANNEL) {
        status = uvm_rm_locked_call(nvUvmInterfaceGetChannelResourcePtes(gpu_va_space->duped_gpu_va_space,
                                                                         va_range->channel.rm_descriptor,
                                                                         map_offset,
                                                                         map_size,
                                                                         mapping_info));
    }
    else {
        ext_gpu_map = uvm_va_range_ext_gpu_map(va_range, gpu);
        UVM_ASSERT(ext_gpu_map);
        status = uvm_rm_locked_call(nvUvmInterfaceGetExternalAllocPtes(gpu_va_space->duped_gpu_va_space,
                                                                       ext_gpu_map->dup_handle,
                                                                       map_offset,
                                                                       map_size,
                                                                       mapping_info));
    }

    if (status != NV_OK) {
        UVM_ERR_PRINT("Failed to get %s mappings for VA range [0x%llx, 0x%llx], offset 0x%llx, size 0x%llx: %s\n",
                      va_range->type == UVM_VA_RANGE_TYPE_CHANNEL ? "channel" : "external",
                      va_range->node.start, va_range->node.end, map_offset, map_size, nvstatusToString(status));
    }

    return status;
}

// Copies the input ptes buffer to the given physical address, with an optional
// TLB invalidate. The copy acquires the input tracker then updates it.
static NV_STATUS copy_ptes(uvm_page_tree_t *tree,
                           NvU64 page_size,
                           uvm_gpu_phys_address_t pte_addr,
                           NvU64 *ptes,
                           NvU32 num_ptes,
                           bool last_mapping,
                           NvU64 range_start,
                           NvU64 range_end,
                           uvm_tracker_t *tracker)
{
    uvm_push_t push;
    NV_STATUS status;
    NvU32 pte_size = uvm_mmu_pte_size(tree, page_size);

    UVM_ASSERT(pte_size * num_ptes <= MAX_COPY_SIZE_PER_PUSH);

    // CPU_TO_GPU because the data being transferred is within the pushbuffer
    status = uvm_push_begin_acquire(tree->gpu->channel_manager,
                                    UVM_CHANNEL_TYPE_CPU_TO_GPU,
                                    tracker,
                                    &push,
                                    "Writing %u bytes of PTEs to {%s, 0x%llx}",
                                    pte_size * num_ptes,
                                    uvm_aperture_string(pte_addr.aperture),
                                    pte_addr.address);
    if (status != NV_OK)
        return status;

    uvm_pte_batch_single_write_ptes(&push, pte_addr, ptes, pte_size, num_ptes);

    if (last_mapping) {
        // Do a TLB invalidate if this is the last mapping in the VA range
        // Membar: This is a permissions upgrade, so no post-invalidate membar
        //         is needed.
        uvm_tlb_batch_single_invalidate(tree, &push, range_start, range_end - range_start + 1,
                page_size, UVM_MEMBAR_NONE);
    }
    else {
        // For pushes prior to the last one, the PTE batch write has
        // already pushed a membar that's enough to order the PTE writes
        // with the TLB invalidate in the last push and that's all
        // that's needed.
        // If a failure happens before the push for the last mapping, it is
        // still ok as what will follow is more CE writes to unmap the PTEs and
        // those will get ordered by the membar from the PTE batch.
        uvm_push_set_flag(&push, UVM_PUSH_FLAG_CE_NEXT_MEMBAR_NONE);
    }

    uvm_push_end(&push);

    // The push acquired the tracker so it's ok to just overwrite it with
    // the entry tracking the push.
    uvm_tracker_overwrite_with_push(tracker, &push);

    return NV_OK;
}

// Map all of pt_range, which is contained with the va_range and begins at
// virtual address start. The PTE values are queried from RM and the pushed
// writes are added to the input tracker.
//
// If the mapped range ends on va_range->node.end, a TLB invalidate for upgrade
// is also issued.
static NV_STATUS map_rm_pt_range(uvm_va_range_t *va_range,
                                 uvm_page_tree_t *tree,
                                 uvm_page_table_range_t *pt_range,
                                 UvmGpuExternalMappingInfo *mapping_info,
                                 NvU64 start,
                                 NvU64 map_offset,
                                 uvm_tracker_t *tracker)
{
    uvm_gpu_phys_address_t pte_addr;
    NvU64 page_size = pt_range->page_size;
    NvU32 pte_size = uvm_mmu_pte_size(tree, page_size);
    NvU64 addr, end;
    size_t max_ptes, ptes_left, num_ptes;
    NvU64 map_size;
    NvU64 real_pte_buffer_size = mapping_info->pteBufferSize;
    bool last_mapping;
    NV_STATUS status = NV_OK;

    end = start + uvm_page_table_range_size(pt_range) - 1;

    UVM_ASSERT(va_range->type == UVM_VA_RANGE_TYPE_EXTERNAL || va_range->type == UVM_VA_RANGE_TYPE_CHANNEL);
    UVM_ASSERT(start >= va_range->node.start);
    UVM_ASSERT(end <= va_range->node.end);
    UVM_ASSERT(page_size & tree->hal->page_sizes());
    UVM_ASSERT(IS_ALIGNED(start, page_size));
    UVM_ASSERT(IS_ALIGNED(map_offset, page_size));
    UVM_ASSERT(IS_ALIGNED(mapping_info->pteBufferSize, pte_size));
    UVM_ASSERT(mapping_info->pteBufferSize <= MAX_COPY_SIZE_PER_PUSH);

    pte_addr = uvm_page_table_range_entry_address(tree, pt_range, 0);
    max_ptes = (size_t)uvm_div_pow2_64(mapping_info->pteBufferSize, pte_size);

    addr = start;
    ptes_left = (size_t)uvm_div_pow2_64(uvm_page_table_range_size(pt_range), page_size);
    while (addr < end) {
        num_ptes = min(max_ptes, ptes_left);
        map_size = num_ptes * page_size;
        UVM_ASSERT(addr + map_size <= end + 1);

        // Request only as many PTEs from RM as we can push right now. Note that
        // we'll restore the original size at function return for future calls.
        //
        // TODO: RM can determine the buffer size from the map_size parameter.
        //       Bug 1735291 tracks updating this API.
        mapping_info->pteBufferSize = num_ptes * pte_size;
        status = get_rm_ptes(va_range, tree->gpu, map_offset, map_size, mapping_info);
        if (status != NV_OK)
            goto out;

        // TODO: Remove with bug 1735291.
        UVM_ASSERT(mapping_info->pteSize == pte_size);
        UVM_ASSERT(mapping_info->numWrittenPtes == num_ptes);

        last_mapping = (addr + map_size - 1 == va_range->node.end);

        // These copies are technically independent, except for the last one
        // which issues the TLB invalidate and thus must wait for all others.
        // However, since each copy will saturate the bus anyway we force them
        // to serialize to avoid bus contention.
        status = copy_ptes(tree,
                           page_size,
                           pte_addr,
                           mapping_info->pteBuffer,
                           num_ptes,
                           last_mapping,
                           va_range->node.start,
                           va_range->node.end,
                           tracker);
        if (status != NV_OK)
            goto out;

        ptes_left -= num_ptes;
        pte_addr.address += num_ptes * pte_size;
        addr += map_size;
        map_offset += map_size;
    }

out:
    mapping_info->pteBufferSize = real_pte_buffer_size;
    return status;
}

static size_t calc_pte_buffer_size(uvm_va_range_t *va_range, uvm_page_tree_t *tree, NvU32 page_size)
{
    // We batch up PTE querying per page_table_range, so the max mapping we'll
    // query is capped by PDE coverage. Since that could still be really big,
    // clamp it for small VA ranges so we don't grossly overallocate.
    NvU64 max_map_size = min(uvm_va_range_size(va_range), uvm_mmu_pde_coverage(tree, page_size));
    size_t max_ptes = (size_t)uvm_div_pow2_64(max_map_size, page_size);
    NvU32 pte_size = uvm_mmu_pte_size(tree, page_size);

    // Further clamp the buffer by the max size we can copy down at once
    UVM_ASSERT(IS_ALIGNED(MAX_COPY_SIZE_PER_PUSH, pte_size));
    return min(max_ptes * pte_size, MAX_COPY_SIZE_PER_PUSH);
}

// Determine the appropriate membar for downgrades on a VA range with type
// UVM_VA_RANGE_TYPE_EXTERNAL or UVM_VA_RANGE_TYPE_CHANNEL.
static uvm_membar_t va_range_downgrade_membar(uvm_va_range_t *va_range, uvm_gpu_t *mapped_gpu)
{
    uvm_aperture_t aperture;

    if (va_range->type == UVM_VA_RANGE_TYPE_CHANNEL)
        aperture = va_range->channel.aperture;
    else
        aperture = uvm_va_range_ext_gpu_map(va_range, mapped_gpu)->aperture;

    if (aperture == UVM_APERTURE_VID)
        return UVM_MEMBAR_GPU;
    return UVM_MEMBAR_SYS;
}

NV_STATUS uvm_va_range_map_rm_allocation(uvm_va_range_t *va_range,
                                         uvm_gpu_t *mapping_gpu,
                                         UvmGpuMemoryInfo *mem_info,
                                         uvm_map_rm_params_t *map_rm_params)
{
    uvm_gpu_va_space_t *gpu_va_space = uvm_gpu_va_space_get(va_range->va_space, mapping_gpu);
    uvm_page_tree_t *page_tree;
    UvmGpuExternalMappingInfo mapping_info;
    uvm_page_table_range_vec_t *pt_range_vec;
    uvm_page_table_range_t *pt_range;
    NvU64 addr, size;
    NvU64 map_offset = map_rm_params->map_offset;
    size_t i;
    NV_STATUS status;

    // map_rm_params contains the tracker we need to update with the work we
    // we push. However, that's only an output tracker for the caller to use to
    // wait for the whole operation at the end. Since we need to serialize our
    // pushes but not with the caller's tracker, we track our work separately
    // then add our work at the end.
    uvm_tracker_t local_tracker = UVM_TRACKER_INIT();

    UVM_ASSERT(gpu_va_space);
    UVM_ASSERT(va_range->type == UVM_VA_RANGE_TYPE_EXTERNAL || va_range->type == UVM_VA_RANGE_TYPE_CHANNEL);
    UVM_ASSERT(uvm_gpu_can_address(mapping_gpu, va_range->node.end));
    UVM_ASSERT(IS_ALIGNED(mem_info->size, mem_info->pageSize));
    UVM_ASSERT(map_rm_params->tracker);

    page_tree = &gpu_va_space->page_tables;

    // Verify that the GPU VA space supports this page size
    if ((mem_info->pageSize & page_tree->hal->page_sizes()) == 0)
        return NV_ERR_INVALID_ADDRESS;

    // Consolidate input checks for API-level callers
    if (!IS_ALIGNED(va_range->node.start, mem_info->pageSize) ||
        !IS_ALIGNED(uvm_va_range_size(va_range), mem_info->pageSize))
        return NV_ERR_INVALID_ADDRESS;

    if (!IS_ALIGNED(map_offset, mem_info->pageSize) ||
        map_offset + uvm_va_range_size(va_range) > mem_info->size)
        return NV_ERR_INVALID_OFFSET;

    // Set the mapping info once for all loop iterations
    memset(&mapping_info, 0, sizeof(mapping_info));
    mapping_info.cachingType   = map_rm_params->caching_type;
    mapping_info.mappingType   = map_rm_params->mapping_type;
    mapping_info.pteBufferSize = calc_pte_buffer_size(va_range, page_tree, mem_info->pageSize);

    mapping_info.pteBuffer = uvm_kvmalloc(mapping_info.pteBufferSize);
    if (!mapping_info.pteBuffer)
        return NV_ERR_NO_MEMORY;

    if (va_range->type == UVM_VA_RANGE_TYPE_EXTERNAL)
        pt_range_vec = &uvm_va_range_ext_gpu_map(va_range, mapping_gpu)->pt_range_vec;
    else
        pt_range_vec = &va_range->channel.pt_range_vec;

    // Allocate all page tables for this VA range.
    //
    // TODO: Benchmark to see if we get any performance improvement from
    //       parallelizing page range allocation with writing PTEs for earlier
    //       ranges.
    status = uvm_page_table_range_vec_init(page_tree,
                                           va_range->node.start,
                                           uvm_va_range_size(va_range),
                                           mem_info->pageSize,
                                           pt_range_vec);
    if (status != NV_OK)
        goto out;

    addr = va_range->node.start;
    for (i = 0; i < pt_range_vec->range_count; i++) {
        pt_range = &pt_range_vec->ranges[i];

        status = map_rm_pt_range(va_range, page_tree, pt_range, &mapping_info, addr, map_offset, &local_tracker);
        if (status != NV_OK)
            goto out;

        size = uvm_page_table_range_size(pt_range);
        addr += size;
        map_offset += size;
    }

    status = uvm_tracker_add_tracker(map_rm_params->tracker, &local_tracker);
    if (status != NV_OK)
        goto out;

out:
    if (status != NV_OK) {
        // We could have any number of mappings in flight to these page tables,
        // so wait for everything before we clear and free them.
        if (uvm_tracker_wait(&local_tracker) != NV_OK) {
            // System-fatal error. Just leak.
            return status;
        }

        if (pt_range_vec->ranges) {
            uvm_page_table_range_vec_clear_ptes(pt_range_vec, va_range_downgrade_membar(va_range, mapping_gpu));
            uvm_page_table_range_vec_deinit(pt_range_vec);
        }
    }

    uvm_kvfree(mapping_info.pteBuffer);
    uvm_tracker_deinit(&local_tracker);
    return status;
}

static bool uvm_api_mapping_type_invalid(UvmGpuMappingType map_type)
{
    switch (map_type) {
        case UvmGpuMappingTypeDefault:
        case UvmGpuMappingTypeReadWriteAtomic:
        case UvmGpuMappingTypeReadWrite:
        case UvmGpuMappingTypeReadOnly:
            return false;
    }
    return true;
}

static bool uvm_api_caching_type_invalid(UvmGpuCachingType cache_type)
{
    switch (cache_type) {
        case UvmGpuCachingTypeDefault:
        case UvmGpuCachingTypeForceUncached:
        case UvmGpuCachingTypeForceCached:
            return false;
    }
    return true;
}

static NV_STATUS set_ext_gpu_map_location(uvm_ext_gpu_map_t *ext_gpu_map,
                                          uvm_va_space_t *va_space,
                                          uvm_gpu_t *mapping_gpu,
                                          UvmGpuMemoryInfo *mem_info)
{
    uvm_gpu_t *owning_gpu;
    uvm_aperture_t aperture;

    // If the allocation is in sysmem then it doesn't matter which GPU owns it,
    // because our dup is enough to keep it around even when the owning GPU
    // dies.
    if (mem_info->sysmem) {
        ext_gpu_map->owning_gpu = NULL;
        ext_gpu_map->aperture = UVM_APERTURE_SYS;
        return NV_OK;
    }

    // This is a local or peer allocation, so the owning GPU must have been
    // registered.
    owning_gpu = uvm_va_space_get_gpu_by_uuid(va_space, &mem_info->uuid);
    if (!owning_gpu)
        return NV_ERR_INVALID_DEVICE;

    if (owning_gpu == mapping_gpu) {
        aperture = UVM_APERTURE_VID;
    }
    else {
        // TODO: In SLI, the returned UUID may be different but a local mapping
        //       must be used. We need to query SLI groups to know that.
        if (!uvm_va_space_peer_enabled(va_space, mapping_gpu, owning_gpu))
            return NV_ERR_INVALID_DEVICE;

        aperture = uvm_gpu_peer_aperture(mapping_gpu, owning_gpu);
    }

    ext_gpu_map->owning_gpu = owning_gpu;
    ext_gpu_map->aperture = aperture;
    return NV_OK;
}

static NV_STATUS uvm_map_external_allocation_on_gpu(uvm_va_range_t *va_range,
                                                    uvm_gpu_t *mapping_gpu,
                                                    uvm_rm_user_object_t *user_rm_mem,
                                                    uvm_map_rm_params_t *map_rm_params)
{
    uvm_va_space_t *va_space = va_range->va_space;
    uvm_ext_gpu_map_t *ext_gpu_map;
    UvmGpuMemoryInfo mem_info;
    NV_STATUS status;

    uvm_assert_rwsem_locked_write(&va_space->lock);

    // Check if the GPU already has a mapping here
    if (uvm_va_range_ext_gpu_map(va_range, mapping_gpu) != NULL)
        return NV_ERR_UVM_ADDRESS_IN_USE;

    // Check if the GPU can access the VA
    if (!uvm_gpu_can_address(mapping_gpu, va_range->node.end))
        return NV_ERR_OUT_OF_RANGE;

    ext_gpu_map = uvm_kvmalloc_zero(sizeof(*ext_gpu_map));
    if (!ext_gpu_map)
        return NV_ERR_NO_MEMORY;

    // Insert the ext_gpu_map into the VA range immediately since some of the
    // below calls require it to be there.
    va_range->external.gpu_maps[mapping_gpu->id - 1] = ext_gpu_map;
    uvm_processor_mask_set(&va_range->external.mapped_gpus, mapping_gpu->id);

    // Dup the memory. This verifies the input handles, takes a ref count on the
    // physical allocation so it can't go away under us, and returns us the
    // allocation info.
    status = uvm_rm_locked_call(nvUvmInterfaceDupMemory(mapping_gpu->rm_address_space,
                                                        user_rm_mem->user_client,
                                                        user_rm_mem->user_object,
                                                        &ext_gpu_map->dup_handle,
                                                        &mem_info));
    if (status != NV_OK) {
        UVM_DBG_PRINT("Failed to dup memory handle {0x%x, 0x%x}: %s, GPU: %s\n",
                      user_rm_mem->user_client, user_rm_mem->user_object,
                      nvstatusToString(status), mapping_gpu->name);
        goto error;
    }

    status = set_ext_gpu_map_location(ext_gpu_map, va_space, mapping_gpu, &mem_info);
    if (status != NV_OK)
        goto error;

    status = uvm_va_range_map_rm_allocation(va_range, mapping_gpu, &mem_info, map_rm_params);
    if (status != NV_OK)
        goto error;

    return NV_OK;

error:
    uvm_ext_gpu_map_destroy(va_range, mapping_gpu);
    return status;
}

// Actual implementation of UvmMapExternalAllocation
static NV_STATUS uvm_map_external_allocation(uvm_va_space_t *va_space, UVM_MAP_EXTERNAL_ALLOCATION_PARAMS *params)
{
    uvm_va_range_t *va_range = NULL;
    uvm_gpu_t *mapping_gpu;
    bool range_was_allocated = false;
    NV_STATUS status = NV_OK;
    size_t i = 0, j;
    uvm_map_rm_params_t map_rm_params;
    uvm_rm_user_object_t user_rm_mem =
    {
        .rm_control_fd = params->rmCtrlFd,
        .user_client   = params->hClient,
        .user_object   = params->hMemory
    };
    uvm_tracker_t tracker = UVM_TRACKER_INIT();

    if (uvm_api_range_invalid(params->base, params->length))
        return NV_ERR_INVALID_ADDRESS;

    if (params->gpuAttributesCount == 0 || params->gpuAttributesCount > UVM_MAX_GPUS)
        return NV_ERR_INVALID_ARGUMENT;

    // TODO: Consider a va_range lock for external ranges so we can do the
    //       mappings in read mode
    uvm_down_write(&va_space->lock);

    // Check if the VA range already exists. It's ok to create a new mapping in
    // an existing VA range, as long as the new mapping matches the bounds of
    // the VA range exactly.
    va_range = uvm_va_range_find(va_space, params->base);
    if (!va_range) {
        // No range exists at base, but we still might collide in [base,
        // base+length). Creating the VA range will fail in that case.
        status = uvm_va_range_create_external(va_space, params->base, params->length, &va_range);
        if (status != NV_OK) {
            UVM_DBG_PRINT_RL("Failed to create external VA range [0x%llx, 0x%llx)\n",
                             params->base, params->base + params->length);
            goto error;
        }
        range_was_allocated = true;
    }
    else if (va_range->type != UVM_VA_RANGE_TYPE_EXTERNAL    ||
             va_range->node.start != params->base            ||
             va_range->node.end != params->base + params->length - 1) {
        status = NV_ERR_UVM_ADDRESS_IN_USE;
        goto error;
    }

    for (i = 0; i < params->gpuAttributesCount; i++) {
        if (uvm_api_mapping_type_invalid(params->perGpuAttributes[i].gpuMappingType) ||
            uvm_api_caching_type_invalid(params->perGpuAttributes[i].gpuCachingType)) {
            status = NV_ERR_INVALID_ARGUMENT;
            goto error;
        }

        mapping_gpu = uvm_va_space_get_gpu_by_uuid_with_gpu_va_space(va_space, &params->perGpuAttributes[i].gpuUuid);
        if (!mapping_gpu) {
            status = NV_ERR_INVALID_DEVICE;
            goto error;
        }

        // Use a tracker to get as much parallelization as possible among GPUs,
        // so one GPU can have its PTE writes in flight while we're working on
        // the next one.
        map_rm_params.map_offset = params->offset;
        map_rm_params.mapping_type = params->perGpuAttributes[i].gpuMappingType;
        map_rm_params.caching_type = params->perGpuAttributes[i].gpuCachingType;
        map_rm_params.tracker = &tracker;
        status = uvm_map_external_allocation_on_gpu(va_range, mapping_gpu, &user_rm_mem, &map_rm_params);
        if (status != NV_OK)
            goto error;
    }

    // Wait for outstanding page table operations to finish across all GPUs. We
    // just need to hold the VA space lock to prevent the GPUs on which we're
    // waiting from getting unregistered underneath us. We can do that in read
    // mode.
    uvm_downgrade_write(&va_space->lock);

    status = uvm_tracker_wait_deinit(&tracker);

    uvm_up_read(&va_space->lock);
    return status;

error:
    // We still have to wait for page table writes to finish, since the teardown
    // could free them.
    (void)uvm_tracker_wait_deinit(&tracker);
    if (va_range) {
        if (range_was_allocated) {
            uvm_va_range_destroy(va_range);
        }
        else {
            // Tear down only those mappings we created during this call
            for (j = 0; j < i; j++) {
                mapping_gpu = uvm_va_space_get_gpu_by_uuid_with_gpu_va_space(va_range->va_space,
                                                                             &params->perGpuAttributes[j].gpuUuid);
                UVM_ASSERT(mapping_gpu);
                uvm_ext_gpu_map_destroy(va_range, mapping_gpu);
            }
        }
    }

    uvm_up_write(&va_space->lock);
    return status;
}

NV_STATUS uvm_api_map_external_allocation(UVM_MAP_EXTERNAL_ALLOCATION_PARAMS *params, struct file *filp)
{
    uvm_va_space_t *va_space = uvm_va_space_get(filp);
    return uvm_map_external_allocation(va_space, params);
}

void uvm_ext_gpu_map_destroy(uvm_va_range_t *va_range, uvm_gpu_t *mapped_gpu)
{
    uvm_gpu_va_space_t *gpu_va_space;
    uvm_ext_gpu_map_t *ext_gpu_map = uvm_va_range_ext_gpu_map(va_range, mapped_gpu);
    NV_STATUS status;

    if (!ext_gpu_map)
        return;

    gpu_va_space = uvm_gpu_va_space_get(va_range->va_space, mapped_gpu);
    UVM_ASSERT(gpu_va_space);

    // Unmap the PTEs
    if (ext_gpu_map->pt_range_vec.ranges) {
        uvm_page_table_range_vec_clear_ptes(&ext_gpu_map->pt_range_vec,
                                            va_range_downgrade_membar(va_range, mapped_gpu));
        uvm_page_table_range_vec_deinit(&ext_gpu_map->pt_range_vec);
    }

    if (ext_gpu_map->dup_handle) {
        status = uvm_rm_locked_call(nvUvmInterfaceFreeDupedHandle(mapped_gpu->rm_address_space,
                                                                  ext_gpu_map->dup_handle));
        UVM_ASSERT(status == NV_OK);
    }

    uvm_kvfree(ext_gpu_map);

    va_range->external.gpu_maps[mapped_gpu->id - 1] = NULL;
    uvm_processor_mask_clear(&va_range->external.mapped_gpus, mapped_gpu->id);
}

// This destroys VA ranges created by UvmMapExternalAllocation and
// UvmAllocSemaphorePool *only*. VA ranges created by UvmMemMap and UvmAlloc go
// through mmap/munmap.
static NV_STATUS uvm_free(uvm_va_space_t *va_space, NvU64 base, NvU64 length)
{
    uvm_va_range_t *va_range;
    NV_STATUS status = NV_OK;

    if (uvm_api_range_invalid(base, length))
        return NV_ERR_INVALID_ADDRESS;

    uvm_down_write(&va_space->lock);

    // Non-managed ranges are defined to not require splitting, so a partial
    // free attempt is an error.
    //
    // TODO: Remove the length parameter from the ioctl
    va_range = uvm_va_range_find(va_space, base);
    if (!va_range                                    ||
        va_range->type != UVM_VA_RANGE_TYPE_EXTERNAL ||
        va_range->node.start != base                 ||
        va_range->node.end != base + length - 1) {
        status = NV_ERR_INVALID_ADDRESS;
        goto out;
    }

    uvm_va_range_destroy(va_range);

out:
    uvm_up_write(&va_space->lock);
    return status;
}

NV_STATUS uvm_api_free(UVM_FREE_PARAMS *params, struct file *filp)
{
    uvm_va_space_t *va_space = uvm_va_space_get(filp);
    return uvm_free(va_space, params->base, params->length);
}
