/*******************************************************************************
    Copyright (c) 2016 NVIDIA Corporation

    Permission is hereby granted, free of charge, to any person obtaining a copy
    of this software and associated documentation files (the "Software"), to
    deal in the Software without restriction, including without limitation the
    rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
    sell copies of the Software, and to permit persons to whom the Software is
    furnished to do so, subject to the following conditions:

        The above copyright notice and this permission notice shall be
        included in all copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
    IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
    FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
    THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
    LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
    FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
    DEALINGS IN THE SOFTWARE.

*******************************************************************************/

#include "uvm8_mem.h"
#include "uvm8_mmu.h"
#include "uvm8_gpu.h"
#include "uvm8_global.h"
#include "uvm8_kvmalloc.h"
#include "uvm8_push.h"
#include "uvm8_range_allocator.h"
#include "uvm8_hal.h"
#include "uvm_linux.h"

static uvm_range_allocator_t g_free_ranges;
static bool g_mem_initialized;

NV_STATUS uvm_mem_global_init(void)
{
    NV_STATUS status = uvm_range_allocator_init(UVM_MEM_VA_SIZE, &g_free_ranges);
    if (status != NV_OK)
        return status;

    g_mem_initialized = true;

    return NV_OK;
}

void uvm_mem_global_exit(void)
{
    if (!g_mem_initialized)
        return;

    uvm_range_allocator_deinit(&g_free_ranges);
}

uvm_chunk_sizes_mask_t uvm_mem_kernel_chunk_sizes(uvm_gpu_t *gpu)
{
    // Get the mmu mode hal directly as the internal address space tree has not
    // been created yet.
    uvm_mmu_mode_hal_t *hal = gpu->arch_hal->mmu_mode_hal(gpu->big_page.internal_size);

    return hal->page_sizes();
}

static bool uvm_mem_is_sysmem(uvm_mem_t *mem)
{
    return mem->backing_gpu == NULL;
}

static bool uvm_mem_is_vidmem(uvm_mem_t *mem)
{
    return !uvm_mem_is_sysmem(mem);
}

static bool uvm_mem_is_local_vidmem(uvm_mem_t *mem, uvm_gpu_t *gpu)
{
    return mem->backing_gpu == gpu;
}

static NvU32 pick_page_size(uvm_mem_t *mem)
{
    NvU32 biggest_page_size;

    if (uvm_mem_is_sysmem(mem))
        return PAGE_SIZE;

    biggest_page_size = uvm_mmu_biggest_page_size(&mem->backing_gpu->address_space_tree);

    if (mem->size < mem->backing_gpu->big_page.internal_size)
        return UVM_PAGE_SIZE_4K;
    else if (mem->size < biggest_page_size)
        return mem->backing_gpu->big_page.internal_size;
    else
        return biggest_page_size;
}

static void uvm_mem_free_vidmem_chunks(uvm_mem_t *mem)
{
    size_t i;

    UVM_ASSERT(uvm_mem_is_vidmem(mem));

    if (!mem->chunks)
        return;

    for (i = 0; i < mem->chunks_count; ++i) {
        if (mem->chunks[i] == NULL)
            break;
        uvm_pmm_gpu_free_kernel(&mem->backing_gpu->pmm, mem->page_size, mem->chunks[i]);
    }

    uvm_kvfree(mem->chunks);
    mem->chunks = NULL;
}

static void uvm_mem_free_sysmem_chunks(uvm_mem_t *mem)
{
    size_t i;

    UVM_ASSERT(uvm_mem_is_sysmem(mem));

    if (!mem->cpu_pages)
        return;

    for (i = 0; i < mem->chunks_count; ++i) {
        if (!mem->cpu_pages[i])
            break;
        __free_pages(mem->cpu_pages[i], 0);
    }

    uvm_kvfree(mem->cpu_pages);
    mem->cpu_pages = NULL;
}

static void uvm_mem_free_chunks(uvm_mem_t *mem)
{
    if (uvm_mem_is_sysmem(mem))
        uvm_mem_free_sysmem_chunks(mem);
    else
        uvm_mem_free_vidmem_chunks(mem);
}

static NV_STATUS uvm_mem_alloc_sysmem_chunks(uvm_mem_t *mem, size_t size)
{
    NV_STATUS status = NV_OK;
    size_t i;

    UVM_ASSERT(mem->page_size == PAGE_SIZE);

    mem->cpu_pages = uvm_kvmalloc_zero(sizeof(*mem->cpu_pages) * mem->chunks_count);
    if (!mem->cpu_pages) {
        status = NV_ERR_NO_MEMORY;
        goto error;
    }

    for (i = 0; i < mem->chunks_count; ++i) {
        mem->cpu_pages[i] = alloc_pages(NV_UVM_GFP_FLAGS, 0);
        if (!mem->cpu_pages[i]) {
            status = NV_ERR_NO_MEMORY;
            goto error;
        }
    }

    return NV_OK;

error:
    uvm_mem_free_sysmem_chunks(mem);
    return status;
}

static NV_STATUS uvm_mem_alloc_vidmem_chunks(uvm_mem_t *mem, size_t size)
{
    NV_STATUS status = NV_OK;
    size_t i;

    UVM_ASSERT(uvm_mem_is_vidmem(mem));

    mem->chunks = uvm_kvmalloc_zero(mem->chunks_count * sizeof(*mem->chunks));
    if (!mem->chunks)
        return NV_ERR_NO_MEMORY;

    status = uvm_pmm_gpu_alloc_kernel(&mem->backing_gpu->pmm, mem->chunks_count, mem->page_size, mem->chunks, UVM_PMM_ALLOC_FLAGS_NONE);
    if (status != NV_OK) {
        UVM_ERR_PRINT("pmm_gpu_alloc(count=%zd, size=0x%x) failed: %s\n", mem->chunks_count, mem->page_size, nvstatusToString(status));

        // Reset all the chunk pointers so that uvm_mem_free_vidmem_chunks()
        // doesn't try to free them.
        memset(mem->chunks, 0, mem->chunks_count * sizeof(*mem->chunks));
        return status;
    }

    for (i = 0; i < mem->chunks_count; ++i) {
        status = uvm_tracker_wait(&mem->chunks[i]->tracker);
        if (status != NV_OK)
            return status;
    }

    return NV_OK;
}

static NV_STATUS uvm_mem_alloc_chunks(uvm_mem_t *mem)
{
    if (uvm_mem_is_sysmem(mem))
        return uvm_mem_alloc_sysmem_chunks(mem, mem->physical_allocation_size);
    else
        return uvm_mem_alloc_vidmem_chunks(mem, mem->physical_allocation_size);
}

static const char *uvm_mem_physical_source(uvm_mem_t *mem)
{
    if (uvm_mem_is_vidmem(mem))
        return mem->backing_gpu->name;
    else
        return "CPU";
}

NV_STATUS uvm_mem_map(uvm_mem_t *mem, uvm_processor_mask_t *mask)
{
    uvm_gpu_t *gpu;
    NV_STATUS status;

    if (!mask)
        return NV_OK;

    if (uvm_processor_mask_test(mask, UVM_CPU_ID)) {
        status = uvm_mem_map_cpu(mem);
        if (status != NV_OK)
            return status;
    }

    for_each_gpu_in_mask(gpu, mask) {
        status = uvm_mem_map_gpu(mem, gpu);
        if (status != NV_OK)
            return status;
    }
    return NV_OK;
}

NV_STATUS uvm_mem_alloc(uvm_mem_alloc_params_t *params, uvm_mem_t **mem_out)
{
    NV_STATUS status = NV_OK;
    uvm_mem_t *mem = NULL;

    if (params->size == 0) {
        UVM_ASSERT(params->size != 0);
        return NV_ERR_INVALID_ARGUMENT;
    }

    mem = uvm_kvmalloc_zero(sizeof(*mem));
    if (mem == NULL)
        return NV_ERR_NO_MEMORY;

    mem->backing_gpu = params->backing_gpu;
    mem->size = params->size;
    mem->page_size = params->page_size;

    if (mem->page_size == UVM_PAGE_SIZE_DEFAULT) {
        mem->page_size = pick_page_size(mem);
    }

    mem->physical_allocation_size = UVM_ALIGN_UP(mem->size, mem->page_size);
    mem->chunks_count = mem->physical_allocation_size / mem->page_size;

    status = uvm_mem_alloc_chunks(mem);
    if (status != NV_OK) {
        UVM_ERR_PRINT("uvm_mem_alloc_chunks (chunk count %zu, page size %u) failed: %s, %s\n",
                mem->chunks_count, mem->page_size,
                nvstatusToString(status), uvm_mem_physical_source(mem));
        goto error;
    }

    status = uvm_range_allocator_alloc(&g_free_ranges, mem->physical_allocation_size, mem->page_size, &mem->range_alloc);
    if (status != NV_OK) {
        UVM_ERR_PRINT("Failed to allocate a free range for size %llu alignment %u\n",
                mem->physical_allocation_size, mem->page_size);
        goto error;
    }

    status = uvm_mem_map(mem, params->map_on);
    if (status != NV_OK)
        goto error;

    *mem_out = mem;

    return NV_OK;

error:
    uvm_mem_free(mem);
    return status;
}

static NvU64 reserved_gpu_va(uvm_mem_t *mem, uvm_gpu_t *gpu)
{
    UVM_ASSERT(mem->range_alloc.aligned_start + mem->physical_allocation_size < gpu->uvm_mem_va_size);

    return gpu->uvm_mem_va_base + mem->range_alloc.aligned_start;
}

static NV_STATUS uvm_mem_map_cpu_sysmem(uvm_mem_t *mem)
{
    UVM_ASSERT(mem->backing_gpu == NULL);

    mem->cpu_addr = vmap(mem->cpu_pages, mem->chunks_count, VM_MAP, PAGE_KERNEL);
    if (!mem->cpu_addr)
        return NV_ERR_NO_MEMORY;

    return NV_OK;
}

static void uvm_mem_unmap_cpu_sysmem(uvm_mem_t *mem)
{
    UVM_ASSERT(mem->backing_gpu == NULL);
    UVM_ASSERT(mem->cpu_addr != NULL);

    vunmap(mem->cpu_addr);
    mem->cpu_addr = NULL;
}

NV_STATUS uvm_mem_map_cpu(uvm_mem_t *mem)
{
    NV_STATUS status;

    UVM_ASSERT(mem);

    if (uvm_processor_mask_test(&mem->mapped_on, UVM_CPU_ID)) {
        // Already mapped
        return NV_OK;
    }

    // Only mapping sysmem on the CPU is supported
    UVM_ASSERT(uvm_mem_is_sysmem(mem));

    status = uvm_mem_map_cpu_sysmem(mem);
    if (status != NV_OK)
        return status;

    uvm_processor_mask_set(&mem->mapped_on, UVM_CPU_ID);
    return NV_OK;
}

void uvm_mem_unmap_cpu(uvm_mem_t *mem)
{
    UVM_ASSERT(mem);

    if (!uvm_processor_mask_test(&mem->mapped_on, UVM_CPU_ID)) {
        // Already unmapped
        return;
    }

    uvm_mem_unmap_cpu_sysmem(mem);

    uvm_processor_mask_clear(&mem->mapped_on, UVM_CPU_ID);
}

static uvm_gpu_phys_address_t uvm_mem_gpu_physical_vidmem(uvm_mem_t *mem, size_t offset)
{
    size_t chunk_index = uvm_div_pow2_64(offset, mem->page_size);
    size_t chunk_offset = offset & (mem->page_size - 1);

    UVM_ASSERT(uvm_mem_is_vidmem(mem));

    return uvm_gpu_phys_address(UVM_APERTURE_VID, mem->chunks[chunk_index]->address + chunk_offset);
}

static uvm_gpu_phys_address_t uvm_mem_gpu_physical_sysmem(uvm_mem_t *mem, size_t offset)
{
    struct page *page = mem->cpu_pages[offset / PAGE_SIZE];

    UVM_ASSERT(uvm_mem_is_sysmem(mem));

    return uvm_gpu_phys_address(UVM_APERTURE_SYS, page_to_phys(page) + offset % PAGE_SIZE);
}

uvm_gpu_phys_address_t uvm_mem_gpu_physical(uvm_mem_t *mem, uvm_gpu_t *gpu, NvU64 offset, NvU64 size)
{
    UVM_ASSERT(size != 0);
    UVM_ASSERT_MSG(UVM_ALIGN_DOWN(offset, mem->page_size) == UVM_ALIGN_DOWN(offset + size - 1, mem->page_size),
            "offset %llu size %llu page_size %u\n", offset, size, mem->page_size);
    UVM_ASSERT_MSG(offset / mem->page_size < mem->chunks_count, "offset %llu\n", offset);

    if (uvm_mem_is_vidmem(mem))
        return uvm_mem_gpu_physical_vidmem(mem, offset);
    else
        return uvm_mem_gpu_physical_sysmem(mem, offset);
}

static NvU64 uvm_mem_pte_maker(uvm_page_table_range_vec_t *range_vec, NvU64 offset, void *data)
{
    uvm_mem_t *mem = (uvm_mem_t *)data;
    uvm_page_tree_t *tree = range_vec->tree;
    uvm_gpu_t *gpu = tree->gpu;
    uvm_gpu_phys_address_t phys = uvm_mem_gpu_physical(mem, gpu, offset, range_vec->page_size);
    bool is_vol = (phys.aperture != UVM_APERTURE_VID);
    return tree->hal->make_pte(phys.aperture, phys.address, UVM_PROT_READ_WRITE_ATOMIC, is_vol, range_vec->page_size);
}

static void unmap_gpu(uvm_mem_t *mem, uvm_gpu_t *gpu)
{
    NV_STATUS status;
    uvm_page_table_range_vec_t *range_vec = mem->range_vecs[gpu->id - 1];
    uvm_membar_t tlb_membar = UVM_MEMBAR_SYS;

    if (uvm_mem_is_local_vidmem(mem, gpu))
        tlb_membar = UVM_MEMBAR_GPU;

    status = uvm_page_table_range_vec_clear_ptes(range_vec, tlb_membar);
    if (status != NV_OK)
        UVM_ERR_PRINT("Clearing PTEs failed: %s, GPU %s\n", nvstatusToString(status), gpu->name);

    uvm_page_table_range_vec_destroy(range_vec);
    mem->range_vecs[gpu->id - 1] = NULL;
}

static NV_STATUS map_gpu(uvm_mem_t *mem, uvm_gpu_t *gpu, NvU64 gpu_va, NvU32 page_size)
{
    NV_STATUS status;
    uvm_page_tree_t *tree = &gpu->address_space_tree;
    uvm_page_table_range_vec_t **range_vec = &mem->range_vecs[gpu->id - 1];

    status = uvm_page_table_range_vec_create(tree, gpu_va, mem->physical_allocation_size, page_size, range_vec);
    if (status != NV_OK) {
        UVM_ERR_PRINT("Failed to init page mapping at [0x%llx, 0x%llx): %s, GPU %s\n",
                gpu_va, gpu_va + mem->physical_allocation_size, nvstatusToString(status), gpu->name);
        return status;
    }

    status = uvm_page_table_range_vec_write_ptes(*range_vec, UVM_MEMBAR_NONE, uvm_mem_pte_maker, mem);
    if (status != NV_OK) {
        UVM_ERR_PRINT("Failed to write PTEs for mapping at [0x%llx, 0x%llx): %s, GPU %s\n",
                gpu_va, gpu_va + mem->physical_allocation_size, nvstatusToString(status), gpu->name);
        goto error;
    }

    return NV_OK;

error:
    unmap_gpu(mem, gpu);
    return status;
}

NV_STATUS uvm_mem_map_gpu(uvm_mem_t *mem, uvm_gpu_t *gpu)
{
    NV_STATUS status;
    NvU64 gpu_va;
    if (uvm_processor_mask_test(&mem->mapped_on, gpu->id))
        return NV_OK;

    gpu_va = reserved_gpu_va(mem, gpu);

    status = map_gpu(mem, gpu, gpu_va, mem->page_size);
    if (status != NV_OK)
        return status;

    uvm_processor_mask_set(&mem->mapped_on, gpu->id);

    return NV_OK;
}

void uvm_mem_unmap_gpu(uvm_mem_t *mem, uvm_gpu_t *gpu)
{
    UVM_ASSERT(mem);
    UVM_ASSERT(gpu);

    if (!uvm_processor_mask_test(&mem->mapped_on, gpu->id)) {
        // Already unmapped
        return;
    }

    unmap_gpu(mem, gpu);
    uvm_processor_mask_clear(&mem->mapped_on, gpu->id);
}

void uvm_mem_free(uvm_mem_t *mem)
{
    uvm_gpu_t *gpu;

    if (mem == NULL)
        return;

    if (uvm_processor_mask_test(&mem->mapped_on, UVM_CPU_ID))
        uvm_mem_unmap_cpu(mem);

    for_each_gpu_in_mask(gpu, &mem->mapped_on)
        uvm_mem_unmap_gpu(mem, gpu);

    if (mem->range_alloc.node)
        uvm_range_allocator_free(&g_free_ranges, &mem->range_alloc);

    uvm_mem_free_chunks(mem);

    uvm_kvfree(mem);
}

NV_STATUS uvm_mem_alloc_and_map_cpu(uvm_mem_alloc_params_t *params, uvm_mem_t **mem_out)
{
    uvm_mem_t *mem;
    NV_STATUS status = uvm_mem_alloc(params, &mem);
    if (status != NV_OK)
        return status;
    status = uvm_mem_map_cpu(mem);
    if (status != NV_OK)
        goto error;

    *mem_out = mem;

    return NV_OK;

error:
    uvm_mem_free(mem);
    return status;
}

NV_STATUS uvm_mem_map_all_gpus(uvm_mem_t *mem)
{
    uvm_gpu_t *gpu;

    UVM_ASSERT(mem);

    for_each_global_gpu(gpu) {
        NV_STATUS status = uvm_mem_map_gpu(mem, gpu);
        if (status != NV_OK)
            return status;
    }
    return NV_OK;
}

NV_STATUS uvm_mem_alloc_and_map_all(uvm_mem_alloc_params_t *params, uvm_mem_t **mem_out)
{
    uvm_mem_t *mem;

    NV_STATUS status = uvm_mem_alloc_and_map_cpu(params, &mem);
    if (status != NV_OK)
        return status;

    status = uvm_mem_map_all_gpus(mem);
    if (status != NV_OK)
        goto error;

    *mem_out = mem;

    return NV_OK;

error:
    uvm_mem_free(mem);
    return status;
}

void *uvm_mem_get_cpu_addr(uvm_mem_t *mem)
{
    UVM_ASSERT(uvm_processor_mask_test(&mem->mapped_on, UVM_CPU_ID));

    return mem->cpu_addr;
}

NvU64 uvm_mem_get_gpu_va(uvm_mem_t *mem, uvm_gpu_t *gpu)
{
    UVM_ASSERT_MSG(uvm_processor_mask_test(&mem->mapped_on, gpu->id), "GPU %s\n", gpu->name);

    return reserved_gpu_va(mem, gpu);
}

uvm_gpu_address_t uvm_mem_gpu_address_virtual(uvm_mem_t *mem, uvm_gpu_t *gpu)
{
    return uvm_gpu_address_virtual(uvm_mem_get_gpu_va(mem, gpu));
}

uvm_gpu_address_t uvm_mem_gpu_address_physical(uvm_mem_t *mem, uvm_gpu_t *gpu, NvU64 offset, NvU64 size)
{
    return uvm_gpu_address_from_phys(uvm_mem_gpu_physical(mem, gpu, offset, size));
}
