/*******************************************************************************
    Copyright (c) 2016 NVIDIA Corporation

    Permission is hereby granted, free of charge, to any person obtaining a copy
    of this software and associated documentation files (the "Software"), to
    deal in the Software without restriction, including without limitation the
    rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
    sell copies of the Software, and to permit persons to whom the Software is
    furnished to do so, subject to the following conditions:

        The above copyright notice and this permission notice shall be
        included in all copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
    IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
    FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
    THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
    LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
    FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
    DEALINGS IN THE SOFTWARE.

*******************************************************************************/

#include "uvm_common.h"
#include "uvm_linux.h"
#include "uvm_linux_ioctl.h"
#include "uvm8_global.h"
#include "uvm8_gpu.h"
#include "uvm8_lock.h"
#include "uvm8_va_space.h"
#include "uvm8_va_range.h"
#include "uvm8_va_block.h"
#include "uvm8_tracker.h"
#include "uvm8_api.h"

static NV_STATUS uvm_va_block_migrate(uvm_va_block_t *va_block,
                                      NvU64 base,
                                      NvU64 end,
                                      uvm_processor_id_t dest_id,
                                      uvm_tracker_t *out_tracker)
{
    uvm_processor_mask_t map_mask;
    NV_STATUS status = NV_OK;
    NV_STATUS tracker_status;

    UVM_ASSERT(base >= va_block->start);
    UVM_ASSERT(end  <= va_block->end);

    uvm_processor_mask_zero(&map_mask);

    // TODO: Only sysmem is supported right now. To provide the programming
    //       model behavior in which the destination is mapped, we populate
    //       sysmem and map the destination processor to it.

    uvm_mutex_lock(&va_block->lock);

    // TODO: Needs to handle allocation-retry
    //
    // TODO: Don't force CPU
    status = uvm_va_block_make_resident(va_block, UVM_CPU_ID, base, end);
    if (status != NV_OK)
        goto out;

    // TODO: Handle read duplication
    // TODO: Handle atomic mappings on Pascal+

    // TODO: Also add SetAccessedBy processors
    uvm_processor_mask_set(&map_mask, dest_id);
    status = uvm_va_block_map_mask(va_block, &map_mask, base, end, UVM_PROT_READ_WRITE_ATOMIC);

out:
    // Report any successfully-pushed operations to the migration tracker
    tracker_status = uvm_tracker_add_tracker_safe(out_tracker, &va_block->tracker);

    uvm_mutex_unlock(&va_block->lock);
    return status == NV_OK ? tracker_status : status;
}

static NV_STATUS uvm_va_range_migrate(uvm_va_range_t *va_range,
                                      NvU64 base,
                                      NvU64 end,
                                      uvm_processor_id_t dest_id,
                                      uvm_tracker_t *out_tracker)
{
    uvm_va_block_t *va_block;
    size_t i;
    NV_STATUS status;

    UVM_ASSERT(base >= va_range->node.start);
    UVM_ASSERT(end  <= va_range->node.end);
    UVM_ASSERT(va_range->type == UVM_VA_RANGE_TYPE_MMAP);
    uvm_assert_rwsem_locked_read(&va_range->va_space->lock);

    if (va_range->range_group)
        UVM_ASSERT(uvm_range_group_migratable(va_range->range_group));

    // Iterate over blocks, populating them if necessary
    for (i = uvm_va_range_block_index(va_range, base); i <= uvm_va_range_block_index(va_range, end); i++) {
        status = uvm_va_range_block_create(va_range, i, &va_block);
        if (status != NV_OK)
            return status;

        status = uvm_va_block_migrate(va_block,
                                      max(base, va_block->start),
                                      min(end, va_block->end),
                                      dest_id,
                                      out_tracker);
        if (status != NV_OK)
            return status;
    }

    return NV_OK;
}

static NV_STATUS uvm_migrate(uvm_va_space_t *va_space, NvU64 base, NvU64 length, NvProcessorUuid *destination_uuid)
{
    uvm_gpu_t *gpu = NULL;
    uvm_processor_id_t dest_id;
    uvm_va_range_t *va_range, *va_range_last;
    NvU64 end = base + length - 1;
    uvm_tracker_t tracker = UVM_TRACKER_INIT();
    NV_STATUS status = NV_OK;
    NV_STATUS tracker_status;
    bool skipped_migrate = false;

    if (uvm_api_range_invalid(base, length))
        return NV_ERR_INVALID_ADDRESS;

    // mmap_sem will be needed if we have to create CPU mappings
    uvm_down_read_mmap_sem(&current->mm->mmap_sem);
    uvm_down_read(&va_space->lock);

    if (uvm_uuid_is_cpu(destination_uuid)) {
        dest_id = UVM_CPU_ID;
    }
    else {
        gpu = uvm_va_space_get_gpu_by_uuid_with_gpu_va_space(va_space, destination_uuid);
        if (!gpu) {
            status = NV_ERR_INVALID_DEVICE;
            goto done;
        }

        // Check if the GPU can access the VA
        if (!uvm_gpu_can_address(gpu, base + length - 1)) {
            status = NV_ERR_OUT_OF_RANGE;
            goto done;
        }

        dest_id = gpu->id;
    }

    va_range_last = NULL;
    uvm_for_each_va_range_in_contig(va_range, va_space, base, end) {
        va_range_last = va_range;

        // Only managed ranges can be migrated
        if (va_range->type != UVM_VA_RANGE_TYPE_MMAP) {
            status = NV_ERR_INVALID_ADDRESS;
            goto done;
        }

        // Skip non-migratable VA ranges
        if (!uvm_range_group_migratable(va_range->range_group)) {
            // Only return NV_WARN_MORE_PROCESSING_REQUIRED if the pages aren't
            // already resident at dest_id.
            if (va_range->preferred_location != dest_id)
                skipped_migrate = true;
        }
        else {
            status = uvm_va_range_migrate(va_range,
                                          max(base, va_range->node.start),
                                          min(end, va_range->node.end),
                                          dest_id,
                                          &tracker);
            if (status != NV_OK)
                goto done;
        }
    }

    // Check that we were able to iterate over the entire range without any gaps
    if (!va_range_last || va_range_last->node.end < end) {
        status = NV_ERR_INVALID_ADDRESS;
        goto done;
    }

done:
    // We only need to hold mmap_sem to create new CPU mappings, so drop it
    // while we wait for the tracker to finish.
    uvm_up_read_mmap_sem_out_of_order(&current->mm->mmap_sem);

    // Regardless of whether there was a user-generated error (out of memory,
    // invalid address), we need to wait for in-flight migrations to complete.
    // We have to do this while holding the VA space lock to prevent the GPUs
    // on which we're waiting from getting unregistered underneath us.
    tracker_status = uvm_tracker_wait_deinit(&tracker);

    // Only clobber status if we didn't hit an earlier error
    if (status == NV_OK)
        status = tracker_status;

    uvm_up_read(&va_space->lock);

    if (status == NV_OK && skipped_migrate)
        return NV_WARN_MORE_PROCESSING_REQUIRED;
    return status;
}

NV_STATUS uvm_api_migrate(UVM_MIGRATE_PARAMS *params, struct file *filp)
{
    uvm_va_space_t *va_space = uvm_va_space_get(filp);
    return uvm_migrate(va_space, params->base, params->length, &params->destinationUuid);
}
