/*******************************************************************************
    Copyright (c) 2016 NVIDIA Corporation

    Permission is hereby granted, free of charge, to any person obtaining a copy
    of this software and associated documentation files (the "Software"), to
    deal in the Software without restriction, including without limitation the
    rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
    sell copies of the Software, and to permit persons to whom the Software is
    furnished to do so, subject to the following conditions:

        The above copyright notice and this permission notice shall be
        included in all copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
    IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
    FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
    THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
    LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
    FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
    DEALINGS IN THE SOFTWARE.

*******************************************************************************/

#include "uvm8_gpu.h"
#include "uvm8_hal.h"
#include "uvm8_kvmalloc.h"
#include "uvm8_mem.h"
#include "uvm8_push.h"
#include "uvm8_va_space.h"
#include "uvm8_test.h"
#include "uvm8_test_ioctl.h"

// Test that big page swizzling is happening or not based on the
// gpu->big_page_swizzling property.
// If big page swizzling is expected, verify that the big page identity mapping
// works correctly.
static NV_STATUS test_big_page_swizzling(uvm_gpu_t *gpu)
{
    NV_STATUS status;
    uvm_mem_t *gpu_mem = NULL;
    uvm_mem_t *sys_mem = NULL;
    int i;
    NvU64 *sys_verif;
    NvU32 big_page_size = gpu->big_page.internal_size;
    uvm_push_t push;
    uvm_gpu_address_t gpu_virtual;
    uvm_gpu_address_t gpu_physical;
    int use_identity_mapping;

    uvm_mem_alloc_params_t mem_params_gpu = { 0 };

    mem_params_gpu.backing_gpu = gpu;
    mem_params_gpu.size = big_page_size;
    mem_params_gpu.page_size = big_page_size;

    status = uvm_mem_alloc(&mem_params_gpu, &gpu_mem);
    TEST_CHECK_GOTO(status == NV_OK, done);

    status = uvm_mem_map_gpu(gpu_mem, gpu);
    TEST_CHECK_GOTO(status == NV_OK, done);

    gpu_virtual = uvm_mem_gpu_address_virtual(gpu_mem, gpu);
    gpu_physical = uvm_mem_gpu_address_physical(gpu_mem, gpu, 0, big_page_size);

    status = uvm_mem_alloc_sysmem_and_map_cpu(big_page_size, &sys_mem);
    TEST_CHECK_GOTO(status == NV_OK, done);

    status = uvm_mem_map_gpu(sys_mem, gpu);
    TEST_CHECK_GOTO(status == NV_OK, done);

    sys_verif = uvm_mem_get_cpu_addr(sys_mem);

    for (use_identity_mapping = 0; use_identity_mapping <= gpu->big_page.swizzling; use_identity_mapping++) {
        bool mismatch = false;
        bool should_match = !gpu->big_page.swizzling || use_identity_mapping;

        for (i = 0; i < big_page_size / sizeof(NvU64); ++i)
            sys_verif[i] = i;

        status = uvm_push_begin(gpu->channel_manager, UVM_CHANNEL_TYPE_CPU_TO_GPU, &push,
                "Accessing big page mapped physical 0x%llx with%s identity mapping",
                gpu_physical.address, use_identity_mapping ? "" : "out");
        TEST_CHECK_GOTO(status == NV_OK, done);

        if (use_identity_mapping)
            gpu_physical = uvm_mmu_gpu_address_for_big_page_physical(gpu_physical, gpu);

        // Copy known system memory into the GPU using virtual big page mapping
        // and copy it back using the physical address. If identity mappings are
        // being used, the physical address has been converted to a virtual one
        // the identity mapping above.
        gpu->ce_hal->memcopy(&push, gpu_virtual, uvm_mem_gpu_address_virtual(sys_mem, gpu), big_page_size);
        gpu->ce_hal->memcopy(&push, uvm_mem_gpu_address_virtual(sys_mem, gpu), gpu_physical, big_page_size);

        status = uvm_push_end_and_wait(&push);
        TEST_CHECK_GOTO(status == NV_OK, done);

        for (i = 0; i < big_page_size / sizeof(NvU64); ++i) {
            if (sys_verif[i] != i) {
                mismatch = true;
                break;
            }
        }
        if (should_match && mismatch) {
            UVM_TEST_PRINT("Invalid at %d = %llu instead of %llu, identity mapping %d, GPU %s\n",
                    i, sys_verif[i], (NvU64)i, use_identity_mapping, gpu->name);
            status = NV_ERR_INVALID_STATE;
            goto done;
        }
        else if (!should_match && !mismatch) {
            UVM_TEST_PRINT("Everything matchinng even though it shouldn't, GPU %s\n", gpu->name);
            status = NV_ERR_INVALID_STATE;
            goto done;
        }
    }

done:
    if (gpu_mem)
        uvm_mem_free(gpu_mem);
    if (sys_mem)
        uvm_mem_free(sys_mem);

    return status;
}

static NV_STATUS test_all_gpus(uvm_va_space_t *va_space)
{
    uvm_gpu_t *gpu;
    for_each_va_space_gpu(gpu, va_space)
        TEST_CHECK_RET(test_big_page_swizzling(gpu) == NV_OK);

    return NV_OK;
}

NV_STATUS uvm8_test_mmu_sanity(UVM_TEST_MMU_SANITY_PARAMS *params, struct file *filp)
{
    NV_STATUS status;
    uvm_va_space_t *va_space = uvm_va_space_get(filp);

    uvm_down_read(&va_space->lock);

    status = test_all_gpus(va_space);

    uvm_up_read(&va_space->lock);

    return status;
}
