/*******************************************************************************
    Copyright (c) 2016 NVIDIA Corporation

    Permission is hereby granted, free of charge, to any person obtaining a copy
    of this software and associated documentation files (the "Software"), to
    deal in the Software without restriction, including without limitation the
    rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
    sell copies of the Software, and to permit persons to whom the Software is
    furnished to do so, subject to the following conditions:

        The above copyright notice and this permission notice shall be
        included in all copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
    IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
    FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
    THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
    LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
    FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
    DEALINGS IN THE SOFTWARE.

*******************************************************************************/

#include "uvm8_perf_events.h"
#include "uvm8_va_block.h"
#include "uvm8_va_range.h"
#include "uvm8_va_space.h"
#include "uvm8_kvmalloc.h"
#include "uvm8_test.h"

static void callback_destroy(uvm_perf_event_t event_id, uvm_perf_event_data_t *event_data)
{
    // In block creation initializes the pointer reserved for perf stats with the address of test_value
    // In block destruction it unsets the pointer
    int *data;

    data = event_data->block_destroy.block->perf_modules_data[UVM_PERF_MODULE_TYPE_STATS].data;

    if (data) {
        uvm_kvfree(data);
        event_data->block_destroy.block->perf_modules_data[UVM_PERF_MODULE_TYPE_STATS].data = NULL;
    }
}

static int *get_or_alloc_block_data(uvm_va_block_t *block)
{
    int **value;
    value = (int **)&block->perf_modules_data[UVM_PERF_MODULE_TYPE_STATS].data;
    if (!(*value))
        *value = (int *)uvm_kvmalloc_zero(sizeof(int));

    return *value;
}

static void callback_inc_1(uvm_perf_event_t event_id, uvm_perf_event_data_t *event_data)
{
    int *value = get_or_alloc_block_data(event_data->fault.block);
    if (value)
        ++(*value);
}

static void callback_inc_2(uvm_perf_event_t event_id, uvm_perf_event_data_t *event_data)
{
    int *value = get_or_alloc_block_data(event_data->fault.block);
    if (value)
        *value += 2;
}

static NV_STATUS test_events(uvm_va_space_t *va_space)
{
    NV_STATUS status;
    uvm_perf_event_data_t event_data;

    uvm_va_block_t block;

    memset(block.perf_modules_data, 0, sizeof(block.perf_modules_data));

    // va_space write lock is required for event callback registration
    uvm_down_write(&va_space->lock);

    // Register a callback for block destruction
    status = uvm_perf_register_event_callback(&va_space->perf_events, UVM_PERF_EVENT_BLOCK_DESTROY, callback_destroy);
    MEM_NV_CHECK_RET(status, NV_OK);
    // Register a callback for page fault
    status = uvm_perf_register_event_callback(&va_space->perf_events, UVM_PERF_EVENT_FAULT, callback_inc_1);
    MEM_NV_CHECK_RET(status, NV_OK);
    // Register a callback for page fault
    status = uvm_perf_register_event_callback(&va_space->perf_events, UVM_PERF_EVENT_FAULT, callback_inc_2);
    MEM_NV_CHECK_RET(status, NV_OK);

    uvm_up_write(&va_space->lock);

    // va_space read lock is required for page fault event notification
    uvm_down_read(&va_space->lock);

    // Notify (fake) page fault. The two registered callbacks for this event increment the value of test_value
    event_data.fault.block = &block;
    uvm_perf_event_notify(&va_space->perf_events, UVM_PERF_EVENT_FAULT, &event_data);
    // Module-allocated integer was initialized to zero. It should have been incremented by 1 and 2, respectively in
    // the callbacks
    TEST_CHECK_RET(block.perf_modules_data[UVM_PERF_MODULE_TYPE_STATS].data != NULL);
    MEM_NV_CHECK_RET(*(int *)block.perf_modules_data[UVM_PERF_MODULE_TYPE_STATS].data, 3);

    // Notify (fake) block destruction. It just unsets the pointer
    event_data.block_destroy.block = &block;
    uvm_perf_event_notify(&va_space->perf_events, UVM_PERF_EVENT_BLOCK_DESTROY, &event_data);
    MEM_NV_CHECK_RET(status, NV_OK);

    uvm_up_read(&va_space->lock);

    // va_space write lock is required for event callback unregister
    uvm_down_write(&va_space->lock);

    // Unregister all callbacks
    uvm_perf_unregister_event_callback(&va_space->perf_events, UVM_PERF_EVENT_FAULT, callback_inc_1);
    uvm_perf_unregister_event_callback(&va_space->perf_events, UVM_PERF_EVENT_FAULT, callback_inc_2);
    uvm_perf_unregister_event_callback(&va_space->perf_events, UVM_PERF_EVENT_BLOCK_DESTROY, callback_destroy);

    uvm_up_write(&va_space->lock);

    return NV_OK;
}

NV_STATUS uvm8_test_perf_events_sanity(UVM_TEST_PERF_EVENTS_SANITY_PARAMS *params, struct file *filp)
{
    NV_STATUS status;
    uvm_va_space_t *va_space;

    va_space = uvm_va_space_get(filp);

    status = test_events(va_space);
    if (status != NV_OK)
        goto done;

done:
    return status;
}

