/*******************************************************************************
    Copyright (c) 2015 NVIDIA Corporation

    Permission is hereby granted, free of charge, to any person obtaining a copy
    of this software and associated documentation files (the "Software"), to
    deal in the Software without restriction, including without limitation the
    rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
    sell copies of the Software, and to permit persons to whom the Software is
    furnished to do so, subject to the following conditions:

        The above copyright notice and this permission notice shall be
        included in all copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
    IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
    FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
    THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
    LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
    FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
    DEALINGS IN THE SOFTWARE.

*******************************************************************************/

// Below the Physical Memory Manager (PMM) is implemented. It manages
// the life cycle of GPU physical memory including the allocation,
// freeing, and returning to PMA after freeing or after eviction.
//
// The unit of management is a memory chunk in one of the sizes given
// by MMU HAL.

#include "uvm_common.h"
#include "nv_uvm_interface.h"
#include "uvm8_gpu.h"
#include "uvm8_pmm_gpu.h"
#include "uvm8_mem.h"
#include "uvm8_mmu.h"
#include "uvm8_global.h"
#include "uvm_linux.h"

#define for_each_chunk_size_from(size, chunk_sizes)              \
    for (; (size) != UVM_CHUNK_SIZE_INVALID;                     \
         (size) = find_next_size((chunk_sizes), (size)))
#define for_each_chunk_size_rev_from(size, chunk_size)           \
    for (; (size) != UVM_CHUNK_SIZE_INVALID;                     \
         (size) = find_prev_size((chunk_sizes), (size)))

// Helper type for refcounting cache
typedef struct
{
    // Cache for given split size
    struct kmem_cache *cache;

    // Number of GPUs using given split size
    NvU32 refcount;

    // Name of cache
    char name[32];
} kmem_cache_ref_t;

struct uvm_pmm_gpu_chunk_suballoc_struct
{
    // Number of allocated chunks
    size_t allocated;

    // Array of all child subchunks
    // TODO: Can the array be inlined when we put the uvm_gpu_chunk_t
    //   on diet? It could save the parent pointer.
    uvm_gpu_chunk_t *subchunks[0];
};

// Cache for allocation of uvm_pmm_gpu_chunk_suballoc_t. At index n it stores
// a suballoc structure for size 2**n.
//
// For convenience of init/deinit code level 0 is for allocation of chunks
static kmem_cache_ref_t chunk_split_cache[UVM_PMM_CHUNK_SPLIT_CACHE_SIZES];
#define CHUNK_CACHE chunk_split_cache[0].cache

static char *uvm_pmm_mode = NULL;
static uvm_pmm_mode_t pmm_mode = UVM_PMM_MODE_INVALID;
module_param(uvm_pmm_mode, charp, S_IRUGO);
MODULE_PARM_DESC(uvm_pmm_mode, "Set the mode of allocation");

uvm_pmm_mode_t get_pmm_mode(void)
{
    return pmm_mode;
}

const char *get_pmm_mode_string(void)
{
    switch (pmm_mode) {
        UVM_ENUM_STRING_CASE(UVM_PMM_MODE_PMA);
        UVM_ENUM_STRING_CASE(UVM_PMM_MODE_RM);
        UVM_ENUM_STRING_DEFAULT();
    }
}

static NV_STATUS alloc_chunk(uvm_pmm_gpu_t *pmm,
                             uvm_pmm_gpu_memory_type_t type,
                             uvm_chunk_size_t chunk_size,
                             uvm_gpu_chunk_t **chunk,
                             uvm_pmm_alloc_flags_t flags);
static NV_STATUS alloc_phys(uvm_pmm_gpu_t *pmm,
                            uvm_pmm_gpu_memory_type_t type,
                            uvm_chunk_size_t chunk_size,
                            uvm_gpu_chunk_t **chunk);
static void free_phys(uvm_pmm_gpu_t *pmm,
                      uvm_chunk_size_t chunk_size,
                      UvmGpuPointer pa);
static NV_STATUS split_gpu_chunk(uvm_pmm_gpu_t *pmm,
                                 uvm_pmm_gpu_memory_type_t type,
                                 uvm_gpu_chunk_t *chunk,
                                 uvm_chunk_size_t chunk_size,
                                 uvm_chunk_size_t subchunk_size);
static void free_chunk(uvm_pmm_gpu_t *pmm,
                       uvm_chunk_size_t chunk_size,
                       uvm_pmm_gpu_memory_type_t mem_type,
                       uvm_gpu_chunk_t *chunk);
static uvm_gpu_chunk_t *find_next_chunk_to_free(uvm_pmm_gpu_t *pmm,
                                                uvm_pmm_gpu_memory_type_t *out_memory_type,
                                                uvm_chunk_size_t *out_chunk_size);
static void delayed_free(struct work_struct *work);
static NV_STATUS init_chunk_split_cache(uvm_pmm_gpu_t *pmm);
static NV_STATUS init_chunk_split_cache_level(uvm_pmm_gpu_t *pmm, size_t level);
static void deinit_chunk_split_cache(uvm_pmm_gpu_t *pmm);
static uvm_chunk_size_t find_first_size(uvm_chunk_sizes_mask_t chunk_sizes);
static uvm_chunk_size_t find_last_size(uvm_chunk_sizes_mask_t chunk_sizes);
static uvm_chunk_size_t find_next_size(uvm_chunk_sizes_mask_t sizes, uvm_chunk_size_t chunk_size);
static uvm_chunk_size_t find_prev_size(uvm_chunk_sizes_mask_t sizes, uvm_chunk_size_t chunk_size);
static struct list_head *find_free_list(uvm_pmm_gpu_t *pmm, uvm_pmm_gpu_memory_type_t type, uvm_chunk_size_t chunk_size);
static void init_pmm_mode(void);

NV_STATUS uvm_pmm_gpu_init(uvm_gpu_t *gpu, uvm_pmm_gpu_t *pmm)
{
    typedef uvm_chunk_sizes_mask_t (*chunk_size_init_func_t)(uvm_gpu_t *gpu);
    static const chunk_size_init_func_t chunk_size_init_func[][UVM_PMM_GPU_MEMORY_TYPE_COUNT] = {
        {uvm_mmu_user_chunk_sizes, uvm_mmu_kernel_chunk_sizes},
        {NULL, uvm_mem_kernel_chunk_sizes},
    };
    static const uvm_chunk_sizes_mask_t PMM_CHUNK_SIZES = UVM_CHUNK_SIZE_2M | UVM_CHUNK_SIZE_128K | UVM_CHUNK_SIZE_64K;
    uvm_chunk_size_t max_chunk_size;
    NV_STATUS status = NV_OK;
    size_t i, j;

    init_pmm_mode();

    // UVM_CHUNK_SIZE_INVALID is UVM_CHUNK_SIZE_MAX shifted left by 1. This protects
    // UVM_CHUNK_SIZE_INVALID from being negative
    BUILD_BUG_ON(UVM_CHUNK_SIZE_MAX >= UVM_CHUNK_SIZE_INVALID);

    uvm_assert_mutex_locked(&g_uvm_global.global_lock);

    memset(pmm, 0, sizeof(*pmm));

    for (i = 0; i < ARRAY_SIZE(pmm->free_list); i++) {
        for (j = 0; j < ARRAY_SIZE(pmm->free_list[i]); j++)
            INIT_LIST_HEAD(&pmm->free_list[i][j]);
    }

    INIT_WORK(&pmm->delayed_free, delayed_free);

    uvm_mutex_init(&pmm->lock, UVM_LOCK_ORDER_PMM);

    // The biggest chunk size must be the common allocation size to make code
    // responsible for eviction easier. We need to find it and mark as possible
    // allocation for all types.
    // For simplicity, the eviction path always evicts/allocates the single
    // largest supported chunk size across all types. Doing so requires
    // finding that chunk size and forcing all types to support it as
    // a common possible size.
    max_chunk_size = 0;
    for (i = 0; i < UVM_PMM_GPU_MEMORY_TYPE_COUNT; i++) {
        pmm->chunk_sizes[i] = 0;
        for (j = 0; j < ARRAY_SIZE(chunk_size_init_func); j++) {
            if (chunk_size_init_func[j][i])
                pmm->chunk_sizes[i] |= chunk_size_init_func[j][i](gpu);
        }
        UVM_ASSERT(pmm->chunk_sizes[i] < UVM_CHUNK_SIZE_INVALID);
        max_chunk_size = find_last_size(max_chunk_size | pmm->chunk_sizes[i]);
    }

    // Round to next size supported by PMM. Note that find_next_size requires
    // the second argument to be in the bitmask passed in first one so we need to
    // or it.
    max_chunk_size = find_next_size(PMM_CHUNK_SIZES | (max_chunk_size >> 1), max_chunk_size >> 1);
    UVM_ASSERT(max_chunk_size != UVM_CHUNK_SIZE_INVALID);

    for (i = 0; i < UVM_PMM_GPU_MEMORY_TYPE_COUNT; i++) {
        pmm->chunk_sizes[i] |= max_chunk_size;
        UVM_ASSERT_MSG(hweight_long(pmm->chunk_sizes[i]) <= UVM_MAX_CHUNK_SIZES, "PMM does not support more levels than UVM_MAX_CHUNK_SIZES");
    }

    status = init_chunk_split_cache(pmm);
    if (status != NV_OK)
        goto cleanup;

    switch (pmm_mode) {
        case UVM_PMM_MODE_RM:
            pmm->rm_address_space = gpu->rm_address_space;
            break;
        case UVM_PMM_MODE_PMA:
            status = nvUvmInterfaceGetPmaObject(&gpu->uuid, &pmm->pma);
            if (status != NV_OK)
                goto cleanup;
            break;
        default:
            UVM_ASSERT(0);
    }

    return NV_OK;
cleanup:
    uvm_pmm_gpu_deinit(pmm);
    return status;
}

void uvm_pmm_gpu_deinit(uvm_pmm_gpu_t *pmm)
{
    size_t i, j;

    if (!pmm)
        return;

    // TODO: On Pascal+ we or heuristic should assert that non-free lists
    //   are empty and free remaining work
    // TODO: Handle ECC/RC
    if (!bitmap_empty(pmm->chunk_split_cache_initialized, UVM_PMM_CHUNK_SPLIT_CACHE_SIZES)) {
        flush_scheduled_work();
        for (i = 0; i < ARRAY_SIZE(pmm->free_list); i++) {
            for (j = 0; j < ARRAY_SIZE(pmm->free_list[i]); j++)
                UVM_ASSERT(list_empty(&pmm->free_list[i][j]));
        }
    }

    deinit_chunk_split_cache(pmm);
}

NV_STATUS uvm_pmm_gpu_alloc(uvm_pmm_gpu_t *pmm,
                            size_t num_chunks,
                            uvm_chunk_size_t chunk_size,
                            uvm_pmm_gpu_memory_type_t mem_type,
                            uvm_gpu_chunk_t **chunks,
                            uvm_pmm_alloc_flags_t flags)
{
    NV_STATUS status;
    uvm_chunk_sizes_mask_t chunk_sizes = pmm->chunk_sizes[mem_type];
    size_t i;

    UVM_ASSERT((unsigned)mem_type < UVM_PMM_GPU_MEMORY_TYPE_COUNT);
    UVM_ASSERT(is_power_of_2(chunk_size) && (chunk_size & chunk_sizes));
    UVM_ASSERT(num_chunks == 0 || chunks);
    UVM_ASSERT((flags & UVM_PMM_ALLOC_FLAGS_MASK) == flags);

    if (flags & UVM_PMM_ALLOC_FLAGS_EVICT) {
        // If eviction is requested then no va block lock or below should be held.
        uvm_assert_unlocked_order(UVM_LOCK_ORDER_VA_BLOCK);
    }

    uvm_mutex_lock(&pmm->lock);
    for (i = 0; i < num_chunks; i++) {
        status = alloc_chunk(pmm, mem_type, chunk_size, &chunks[i], flags);
        if (status != NV_OK)
            goto cleanup_under_lock;
    }
    uvm_mutex_unlock(&pmm->lock);

    return NV_OK;

cleanup_under_lock:
    while (i-- > 0)
        free_chunk(pmm, chunk_size, mem_type, chunks[i]);
    uvm_mutex_unlock(&pmm->lock);
    return status;
}

NV_STATUS uvm_pmm_gpu_alloc_kernel(uvm_pmm_gpu_t *pmm,
                                   size_t num_chunks,
                                   uvm_chunk_size_t chunk_size,
                                   uvm_gpu_chunk_t **chunks,
                                   uvm_pmm_alloc_flags_t flags)
{
    size_t i;

    NV_STATUS status = uvm_pmm_gpu_alloc(pmm, num_chunks, chunk_size, UVM_PMM_GPU_MEMORY_TYPE_KERNEL, chunks, flags);
    if (status != NV_OK)
        return status;

    for (i = 0; i < num_chunks; ++i)
        chunks[i]->state = UVM_PMM_GPU_CHUNK_STATE_ALLOCATED;

    return NV_OK;
}

void uvm_pmm_gpu_unpin_temp(uvm_pmm_gpu_t *pmm, uvm_gpu_chunk_t *chunk)
{
    uvm_assert_chunk_locked(pmm, chunk);

    UVM_ASSERT(chunk->state == UVM_PMM_GPU_CHUNK_STATE_TEMP_PINNED);

    chunk->state = UVM_PMM_GPU_CHUNK_STATE_ALLOCATED;

    // TODO: If chunk is evicted we need to wake up eviction queue.
}

void uvm_pmm_gpu_free(uvm_pmm_gpu_t *pmm,
                      uvm_chunk_size_t chunk_size,
                      uvm_pmm_gpu_memory_type_t mem_type,
                      uvm_gpu_chunk_t *chunk)
{
    uvm_assert_chunk_locked(pmm, chunk);

    UVM_ASSERT(chunk->state == UVM_PMM_GPU_CHUNK_STATE_ALLOCATED ||
               chunk->state == UVM_PMM_GPU_CHUNK_STATE_TEMP_PINNED);

    // Note that order here is tricky - we don't know at this moment
    // if memory of chunk will be freed so we need to unlock its lock before
    // calling free_chunk. However if we unlock before PMM lock another
    // thread may select this page for eviction and attempt to free it.
    uvm_mutex_lock(&pmm->lock);
    chunk->state = UVM_PMM_GPU_CHUNK_STATE_FREE;
    uvm_pmm_gpu_unlock_chunk(pmm, chunk);
    free_chunk(pmm, chunk_size, mem_type, chunk);
    uvm_mutex_unlock(&pmm->lock);
}

void uvm_pmm_gpu_free_kernel(uvm_pmm_gpu_t *pmm, uvm_chunk_size_t chunk_size, uvm_gpu_chunk_t *chunk)
{
    uvm_pmm_gpu_lock_chunk(pmm, chunk);
    uvm_pmm_gpu_free(pmm, chunk_size, UVM_PMM_GPU_MEMORY_TYPE_KERNEL, chunk);
}

// Allocates a single chunk of a given size. If needed splits a chunk of bigger size
// or, if that is not possible, allocates from PMA or evicts.
//
// PMM lock must be held when calling this function but it may be temporarily dropped
// during eviction.
NV_STATUS alloc_chunk(uvm_pmm_gpu_t *pmm,
                      uvm_pmm_gpu_memory_type_t type,
                      uvm_chunk_size_t chunk_size,
                      uvm_gpu_chunk_t **out_chunk,
                      uvm_pmm_alloc_flags_t flags)
{
    uvm_chunk_size_t cur_size;
    NV_STATUS status;
    uvm_chunk_sizes_mask_t chunk_sizes = pmm->chunk_sizes[type];

    uvm_assert_mutex_locked(&pmm->lock);

    cur_size = chunk_size;
    for_each_chunk_size_from(cur_size, chunk_sizes) {
        if (likely(!list_empty(find_free_list(pmm, type, cur_size))))
            break;
    }

    if (unlikely(cur_size == UVM_CHUNK_SIZE_INVALID)) {
        uvm_gpu_chunk_t *chunk;
        cur_size = find_last_size(chunk_sizes);

        status = alloc_phys(pmm, type, cur_size, &chunk);
        if (status != NV_OK) {
            // TODO: Perform selection and eviction here if
            // MMU and flags indicate it's possible
            return status;
        }

        list_add_tail(&chunk->list, find_free_list(pmm, type, cur_size));
    }

    for_each_chunk_size_rev_from(cur_size, chunk_sizes) {
        uvm_gpu_chunk_t *chunk;
        uvm_chunk_size_t prev_size = find_prev_size(chunk_sizes, cur_size);

        UVM_ASSERT(!list_empty(find_free_list(pmm, type, cur_size)));
        chunk = list_first_entry(find_free_list(pmm, type, cur_size), uvm_gpu_chunk_t, list);
        list_del(&chunk->list);
#ifdef DEBUG
        UVM_ASSERT(chunk->size == cur_size);
        UVM_ASSERT(chunk->type == type);
#endif
        UVM_ASSERT(chunk->state == UVM_PMM_GPU_CHUNK_STATE_FREE);
        chunk->state = UVM_PMM_GPU_CHUNK_STATE_TEMP_PINNED;

        if (chunk->parent) {
            UVM_ASSERT(chunk->parent->suballoc);
#ifdef DEBUG
            UVM_ASSERT(chunk->parent->size == find_next_size(chunk_sizes, cur_size));
            UVM_ASSERT(chunk->parent->type == type);
#endif
            UVM_ASSERT(chunk->parent->suballoc->allocated < find_next_size(chunk_sizes, cur_size) / cur_size);

            chunk->parent->suballoc->allocated++;
        }

        if (likely(cur_size == chunk_size)) {
            *out_chunk = chunk;
            return NV_OK;
        }

        UVM_ASSERT(prev_size != UVM_CHUNK_SIZE_INVALID);
        status = split_gpu_chunk(pmm, type, chunk, cur_size, prev_size);
        if (status != NV_OK) {
            free_chunk(pmm, cur_size, type, chunk);
            return status;
        }
    }
    UVM_PANIC();
}

// Wrapper function around nvUvmInterfacePmaAllocPages converting from/to PMA types
//
// In RM Mode it allocates from RM.
NV_STATUS alloc_phys(uvm_pmm_gpu_t *pmm,
                     uvm_pmm_gpu_memory_type_t type,
                     uvm_chunk_size_t chunk_size,
                     uvm_gpu_chunk_t **out_chunk)
{
    NV_STATUS status;
    uvm_gpu_chunk_t *chunk;
    UvmGpuPointer pa = ~(UvmGpuPointer)0;
    UvmGpuAllocInfo allocInfo = {0};
    UvmPmaAllocationOptions options = {0};

    switch (pmm_mode) {
        case UVM_PMM_MODE_RM:
            allocInfo.bContiguousPhysAlloc = NV_TRUE;
            allocInfo.alignment = chunk_size;

            // Normally it should mark the RM API lock being locked but
            // the hierarchy in uvm_lock.h is prepared to be used with
            // PMA. As there is no evicition callback this is safe
            // (RM API lock just changes its position in hierarchy)
            status = nvUvmInterfaceMemoryAllocGpuPa(pmm->rm_address_space,
                                                    chunk_size,
                                                    &pa,
                                                    &allocInfo);
            if (status != NV_OK)
                return status;

            UVM_ASSERT((pa % chunk_size) == 0);
            break;

        case UVM_PMM_MODE_PMA:
            options.flags = UVM_PMA_ALLOCATE_DONT_EVICT;
            switch (type) {
                case UVM_PMM_GPU_MEMORY_TYPE_USER:
                    // TODO: Query MMU HAL if GPU supports eviction
                    options.flags |= UVM_PMA_ALLOCATE_PINNED;
                    break;
                case UVM_PMM_GPU_MEMORY_TYPE_KERNEL:
                    options.flags |= UVM_PMA_ALLOCATE_PINNED;
                    break;
                default:
                    UVM_ASSERT(0);
            }

            UVM_ASSERT(chunk_size == UVM_CHUNK_SIZE_2M ||
                       chunk_size == UVM_CHUNK_SIZE_128K ||
                       chunk_size == UVM_CHUNK_SIZE_64K);
            status = nvUvmInterfacePmaAllocPages(pmm->pma, 1, chunk_size, &options, &pa);
            if (status != NV_OK)
                return status;
            break;
        default:
            UVM_ASSERT(0);
            return NV_ERR_NO_MEMORY;
    }

    // TODO: Allocate a flat array for top-level chunks
    chunk = kmem_cache_alloc(CHUNK_CACHE, NV_UVM_GFP_FLAGS);
    if (chunk == NULL) {
        status = NV_ERR_NO_MEMORY;
        goto cleanup;
    }
    chunk->address = pa;
    chunk->lock = 0;
    uvm_tracker_init(&chunk->tracker);
    chunk->parent = NULL;
    chunk->suballoc = NULL;
#ifdef DEBUG
    chunk->size = chunk_size;
    chunk->type = type;
#endif
    chunk->state = UVM_PMM_GPU_CHUNK_STATE_FREE;

    *out_chunk = chunk;
    return NV_OK;
cleanup:
    free_phys(pmm, chunk_size, pa);
    return status;
}

// Frees memory to PMA. In RM mode it frees to RM.
void free_phys(uvm_pmm_gpu_t *pmm,
               uvm_chunk_size_t chunk_size,
               UvmGpuPointer pa)
{
    switch (pmm_mode) {
        case UVM_PMM_MODE_RM:
            // Normally it should mark the RM API lock being locked but
            // the hierarchy in uvm_lock.h is prepared to be used with
            // PMA. As there is no evicition callback this is safe
            // (RM API lock just changes its position in hierarchy)
            nvUvmInterfaceMemoryFreePa(pmm->rm_address_space, pa);
            break;
        case UVM_PMM_MODE_PMA:
            nvUvmInterfacePmaFreePages(pmm->pma, &pa, 1, chunk_size, 0);
            break;
        default:
            UVM_ASSERT(0);
            break;
    }
}

// Splits passed chunk into subchunks of given size and adds them to
// free list of appropriate level.
//
// PMM lock must be held when calling this function.
NV_STATUS split_gpu_chunk(uvm_pmm_gpu_t *pmm,
                          uvm_pmm_gpu_memory_type_t type,
                          uvm_gpu_chunk_t *chunk,
                          uvm_chunk_size_t chunk_size,
                          uvm_chunk_size_t subchunk_size)
{
    uvm_chunk_sizes_mask_t chunk_sizes = pmm->chunk_sizes[type];
    size_t cache_idx = ilog2(chunk_size / subchunk_size);
    int i;
    NV_STATUS status;
    uvm_pmm_gpu_chunk_suballoc_t *suballoc;

    uvm_assert_mutex_locked(&pmm->lock);

    UVM_ASSERT(subchunk_size & chunk_sizes);
    UVM_ASSERT(find_next_size(chunk_sizes, subchunk_size) == chunk_size);
    UVM_ASSERT(chunk_split_cache[cache_idx].cache != NULL);
#ifdef DEBUG
    UVM_ASSERT(chunk->size == chunk_size);
    UVM_ASSERT(chunk->type == type);
#endif

    uvm_tracker_remove_completed(&chunk->tracker);

    suballoc = kmem_cache_zalloc(chunk_split_cache[cache_idx].cache, NV_UVM_GFP_FLAGS);
    if (suballoc == NULL)
        return NV_ERR_NO_MEMORY;

    for (i = 0; i < (chunk_size / subchunk_size); i++) {
        suballoc->subchunks[i] = kmem_cache_alloc(CHUNK_CACHE, NV_UVM_GFP_FLAGS);
        if (suballoc->subchunks[i] == NULL) {
            status = NV_ERR_NO_MEMORY;
            goto cleanup;
        }

        suballoc->subchunks[i]->address = chunk->address + i * subchunk_size;
        suballoc->subchunks[i]->lock = 0;
        list_add_tail(&suballoc->subchunks[i]->list, find_free_list(pmm, type, subchunk_size));
        suballoc->subchunks[i]->parent = chunk;
        suballoc->subchunks[i]->suballoc = NULL;
#ifdef DEBUG
        suballoc->subchunks[i]->size = subchunk_size;
        suballoc->subchunks[i]->type = type;
#endif
        suballoc->subchunks[i]->state = UVM_PMM_GPU_CHUNK_STATE_FREE;

        status = uvm_tracker_init_from(&suballoc->subchunks[i]->tracker, &chunk->tracker);
        if (status != NV_OK)
            goto cleanup;
    }
    chunk->state = UVM_PMM_GPU_CHUNK_STATE_IS_SPLIT;
    chunk->suballoc = suballoc;

    return NV_OK;
cleanup:
    for (i = 0; i < (chunk_size / subchunk_size); i++) {
        if (suballoc->subchunks[i] == NULL)
            break;
        list_del(&suballoc->subchunks[i]->list);
        uvm_tracker_deinit(&suballoc->subchunks[i]->tracker);
        kmem_cache_free(CHUNK_CACHE, suballoc->subchunks[i]);
        suballoc->subchunks[i] = NULL;
    }
    kmem_cache_free(chunk_split_cache[cache_idx].cache, suballoc);
    return status;
}

// Frees a chunk - in other words it markes chunk as free and, if eviction is not
// supported it frees it asynchronously.
void free_chunk(uvm_pmm_gpu_t *pmm,
                uvm_chunk_size_t chunk_size,
                uvm_pmm_gpu_memory_type_t mem_type,
                uvm_gpu_chunk_t *chunk)
{
    uvm_chunk_sizes_mask_t chunk_sizes = pmm->chunk_sizes[mem_type];

    uvm_assert_mutex_locked(&pmm->lock);

    while (1) {
        size_t num_subchunks;
        int i;
        uvm_gpu_chunk_t *parent;
        uvm_chunk_size_t parent_size;

#ifdef DEBUG
        UVM_ASSERT(chunk->size == chunk_size);
        UVM_ASSERT(chunk->type == mem_type);
#endif

        parent = chunk->parent;
        parent_size = find_next_size(chunk_sizes, chunk_size);

        chunk->state = UVM_PMM_GPU_CHUNK_STATE_FREE;
        list_add_tail(&chunk->list, find_free_list(pmm, mem_type, chunk_size));

        if (!parent) {
            UVM_ASSERT_MSG(parent_size > UVM_CHUNK_SIZE_MAX, "Chunk without parent must have largest size\n");
            // This should happen only after waiting on semaphore and on non-evictable memory
            // only in separate workqueue.
            //
            // TODO This should not be performed on Pascal+ once eviction is implemented
            // as we can just wait till eviction to free the memory.
            // TODO See a final impact on speed of scheduling when the tracker is
            // finished. Depending on ordering and speed it might be better to free the
            // chunk immediately.
            schedule_work(&pmm->delayed_free);
            return;
        }
        UVM_ASSERT_MSG(parent_size <= UVM_CHUNK_SIZE_MAX, "Chunk without parent must not have largest size\n");

        UVM_ASSERT(parent->state == UVM_PMM_GPU_CHUNK_STATE_IS_SPLIT);
        UVM_ASSERT(parent->suballoc);
#ifdef DEBUG
        UVM_ASSERT(parent->size == parent_size);
        UVM_ASSERT(parent->type == mem_type);
#endif

        UVM_ASSERT(parent->suballoc->allocated > 0);
        if (--parent->suballoc->allocated)
            return;

        num_subchunks = parent_size / chunk_size;

        for (i = 0; i < num_subchunks; i++) {
            NV_STATUS status;
            status = uvm_tracker_add_tracker_safe(&parent->tracker, &parent->suballoc->subchunks[i]->tracker);
            if (unlikely(status != NV_OK)) {
                // This is a system-fatal error. Just leak.
                UVM_ASSERT(uvm_global_get_status() != NV_OK);
                return;
            }

            list_del(&parent->suballoc->subchunks[i]->list);
            uvm_tracker_deinit(&parent->suballoc->subchunks[i]->tracker);
            kmem_cache_free(CHUNK_CACHE, parent->suballoc->subchunks[i]);
        }
        kmem_cache_free(chunk_split_cache[ilog2(num_subchunks)].cache, parent->suballoc);

        chunk_size = parent_size;
        chunk = parent;
    }
}

// Finds next chunk that needs freeing
uvm_gpu_chunk_t *find_next_chunk_to_free(uvm_pmm_gpu_t *pmm,
                                         uvm_pmm_gpu_memory_type_t *out_memory_type,
                                         uvm_chunk_size_t *out_chunk_size)
{
    uvm_pmm_gpu_memory_type_t memory_type;
    uvm_chunk_size_t chunk_size;
    uvm_gpu_chunk_t *result;

    uvm_mutex_lock(&pmm->lock);
    for (memory_type = 0; memory_type < UVM_PMM_GPU_MEMORY_TYPE_COUNT; memory_type++) {
        chunk_size = find_last_size(pmm->chunk_sizes[memory_type]);
        result = list_first_entry_or_null(find_free_list(pmm, memory_type, chunk_size), uvm_gpu_chunk_t, list);
        if (result != NULL) {
            list_del(&result->list);
#ifdef DEBUG
            UVM_ASSERT(result->size == chunk_size);
            UVM_ASSERT(result->type == memory_type);
#endif
            break;
        }
    }
    uvm_mutex_unlock(&pmm->lock);

    if (result) {
        *out_memory_type = memory_type;
        *out_chunk_size = chunk_size;
    }

    return result;
}

void delayed_free(struct work_struct *work)
{
    uvm_pmm_gpu_t *pmm = container_of(work, uvm_pmm_gpu_t, delayed_free);
    uvm_gpu_chunk_t *chunk;
    uvm_pmm_gpu_memory_type_t memory_type;
    uvm_chunk_size_t chunk_size;

    // TODO: Assert non-evictable gpu

    while ((chunk = find_next_chunk_to_free(pmm, &memory_type, &chunk_size)) != NULL) {
        NV_STATUS status;

        // TODO: We should scrub here as per PMA requirements

        status = uvm_tracker_wait_deinit(&chunk->tracker);
        if (status != NV_OK) {
            // TODO: Handle RC/ECC
            UVM_ASSERT(uvm_global_get_status() != NV_OK);
            return;
        }

        free_phys(pmm, chunk_size, chunk->address);
        kmem_cache_free(CHUNK_CACHE, chunk);
    }
}

// Arrays of all possible sizes for user and kernel allocations.
// TODO: Remove them when MMU HAL will provide correct information
static const uvm_chunk_size_t chunk_sizes_user[] = {
    UVM_CHUNK_SIZE_4K,
    UVM_CHUNK_SIZE_64K,
    UVM_CHUNK_SIZE_128K,
    UVM_CHUNK_SIZE_2M
};
static const uvm_chunk_size_t chunk_sizes_kernel[] = {
    UVM_CHUNK_SIZE_128,
    UVM_CHUNK_SIZE_4K,
    UVM_CHUNK_SIZE_2M
};

// Get free list for given chunk size
struct list_head *find_free_list(uvm_pmm_gpu_t *pmm, uvm_pmm_gpu_memory_type_t type, uvm_chunk_size_t chunk_size)
{
    uvm_chunk_sizes_mask_t chunk_sizes = pmm->chunk_sizes[type];
    size_t idx = hweight_long(chunk_sizes & (chunk_size - 1));
    UVM_ASSERT(is_power_of_2(chunk_size));
    UVM_ASSERT(chunk_size & chunk_sizes);
    return &pmm->free_list[type][idx];
}

// Finds first (smallest) size
uvm_chunk_size_t find_first_size(uvm_chunk_sizes_mask_t chunk_sizes)
{
    UVM_ASSERT(chunk_sizes);
    return (uvm_chunk_size_t)1 << __ffs(chunk_sizes);
}

// Finds last (biggest) size
uvm_chunk_size_t find_last_size(uvm_chunk_sizes_mask_t chunk_sizes)
{
    UVM_ASSERT(chunk_sizes);
    return (uvm_chunk_size_t)1 << __fls(chunk_sizes);
}

// Finds next (bigger) size. If there is no such value return UVM_CHUNK_SIZE_INVALID.
uvm_chunk_size_t find_next_size(uvm_chunk_sizes_mask_t chunk_sizes, uvm_chunk_size_t chunk_size)
{
    UVM_ASSERT(is_power_of_2(chunk_size));
    UVM_ASSERT(chunk_sizes & chunk_size);
    BUILD_BUG_ON(sizeof(chunk_sizes) > sizeof(unsigned long));
    return (uvm_chunk_size_t)1 << __ffs((chunk_sizes & ~((chunk_size << 1) - 1)) | UVM_CHUNK_SIZE_INVALID);
}

// Finds next (smaller) size. If there is no such value return UVM_CHUNK_SIZE_INVALID.
uvm_chunk_size_t find_prev_size(uvm_chunk_sizes_mask_t chunk_sizes, uvm_chunk_size_t chunk_size)
{
    UVM_ASSERT(is_power_of_2(chunk_size));
    UVM_ASSERT(chunk_sizes & chunk_size);
    chunk_sizes = chunk_sizes & (chunk_size - 1);
    if (!chunk_sizes)
        return UVM_CHUNK_SIZE_INVALID;
    return (uvm_chunk_size_t)1 << __fls(chunk_sizes);
}

// Initializes the split cache for given GPU.
//
// It walks through all memory splits - in other words all ratios of neighboring
// pairs of sizes - and allocates kmem cache for them, unless they are already
// allocated.
//
// It also bumps the refcount if this GPU did not use such split yet.
NV_STATUS init_chunk_split_cache(uvm_pmm_gpu_t *pmm)
{
    NV_STATUS status;
    uvm_pmm_gpu_memory_type_t type;

    uvm_assert_mutex_locked(&g_uvm_global.global_lock);

    for (type = 0; type < UVM_PMM_GPU_MEMORY_TYPE_COUNT; type++) {
        uvm_chunk_size_t prev_size, cur_size;
        uvm_chunk_sizes_mask_t chunk_sizes = pmm->chunk_sizes[type];
        // Iterate over each pair of neighboring sizes. Note that same level
        // may be visited multiple times and it is handled internally by
        // init_chunk_split_cache_level
        prev_size = find_first_size(chunk_sizes);
        cur_size = find_next_size(chunk_sizes, prev_size);
        for_each_chunk_size_from(cur_size, chunk_sizes) {
            size_t subchunk_count = cur_size / prev_size;
            size_t level = ilog2(subchunk_count);
            status = init_chunk_split_cache_level(pmm, level);
            if (status != NV_OK)
                goto cleanup;
            prev_size = cur_size;
        }
    }

    status = init_chunk_split_cache_level(pmm, 0);
    if (status != NV_OK)
        goto cleanup;

    return NV_OK;
cleanup:
    deinit_chunk_split_cache(pmm);
    return status;
}

NV_STATUS init_chunk_split_cache_level(uvm_pmm_gpu_t *pmm, size_t level)
{
    uvm_assert_mutex_locked(&g_uvm_global.global_lock);

    if (!test_bit(level, pmm->chunk_split_cache_initialized)) {
        if (!chunk_split_cache[level].cache) {
            size_t size;
            size_t align;
            if (level == 0) {
                strncpy(chunk_split_cache[level].name, "uvm_gpu_chunk_t", sizeof(chunk_split_cache[level].name) - 1);
                size = sizeof(uvm_gpu_chunk_t);
                align = __alignof__(uvm_gpu_chunk_t);
            } else {
                snprintf(chunk_split_cache[level].name,
                         sizeof(chunk_split_cache[level].name),
                         "uvm_gpu_chunk_%u", (unsigned)level);
                size = sizeof(uvm_pmm_gpu_chunk_suballoc_t) + (sizeof(uvm_gpu_chunk_t *) << level);
                align = __alignof__(uvm_pmm_gpu_chunk_suballoc_t);
            }
            chunk_split_cache[level].cache = NV_KMEM_CACHE_CREATE_FULL(chunk_split_cache[level].name, size, align, 0, NULL);
            if (!chunk_split_cache[level].cache)
                return NV_ERR_NO_MEMORY;
            chunk_split_cache[level].refcount = 1;
        } else {
            chunk_split_cache[level].refcount++;
            UVM_ASSERT_MSG(chunk_split_cache[level].refcount != 0, "Overflow of refcount\n");
        }
        __set_bit(level, pmm->chunk_split_cache_initialized);
    }
    return NV_OK;
}

void deinit_chunk_split_cache(uvm_pmm_gpu_t *pmm)
{
    unsigned long subchunk_count_log2;

    uvm_assert_mutex_locked(&g_uvm_global.global_lock);

    for_each_set_bit(subchunk_count_log2, pmm->chunk_split_cache_initialized, UVM_PMM_CHUNK_SPLIT_CACHE_SIZES) {
        UVM_ASSERT(chunk_split_cache[subchunk_count_log2].refcount > 0);
        if (--chunk_split_cache[subchunk_count_log2].refcount == 0) {
            kmem_cache_destroy(chunk_split_cache[subchunk_count_log2].cache);
            chunk_split_cache[subchunk_count_log2].cache = NULL;
        }
        __clear_bit(subchunk_count_log2, pmm->chunk_split_cache_initialized);
    }
}

// Initializes mode of allocation. This has effects only on initialization of first
// GPU so it does not need any locks.
static void init_pmm_mode(void)
{
    const static uvm_pmm_mode_t default_mode = UVM_PMM_MODE_RM;

    if (pmm_mode == UVM_PMM_MODE_INVALID) {
        if (uvm_pmm_mode == NULL) {
            pmm_mode = default_mode;
        } else if (strcmp(uvm_pmm_mode, "pma") == 0) {
            pmm_mode = UVM_PMM_MODE_PMA;
        } else if (strcmp(uvm_pmm_mode, "rm") == 0) {
            pmm_mode = UVM_PMM_MODE_RM;
        } else {
            UVM_ERR_PRINT("Unknown PMM mode \"%p\" - fallback to default\n", uvm_pmm_mode);
            pmm_mode = default_mode;
        }
    }
}
