/*******************************************************************************
    Copyright (c) 2015 NVIDIA Corporation

    Permission is hereby granted, free of charge, to any person obtaining a copy
    of this software and associated documentation files (the "Software"), to
    deal in the Software without restriction, including without limitation the
    rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
    sell copies of the Software, and to permit persons to whom the Software is
    furnished to do so, subject to the following conditions:

        The above copyright notice and this permission notice shall be
        included in all copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
    IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
    FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
    THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
    LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
    FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
    DEALINGS IN THE SOFTWARE.

*******************************************************************************/

#ifndef __UVM8_PMM_GPU_H__
#define __UVM8_PMM_GPU_H__

// PMM, Physical Memory Manager, module responsible for allocation, freeing and
// eviction of physical memory.

#include "uvm8_forward_decl.h"
#include "uvm8_lock.h"
#include "uvm8_tracker.h"
#include "uvm_linux.h"
#include "uvmtypes.h"
#include "nv_uvm_types.h"

typedef enum __attribute__ ((__packed__))
{
    UVM_CHUNK_SIZE_1       =           1ULL,
    UVM_CHUNK_SIZE_2       =           2ULL,
    UVM_CHUNK_SIZE_4       =           4ULL,
    UVM_CHUNK_SIZE_8       =           8ULL,
    UVM_CHUNK_SIZE_16      =          16ULL,
    UVM_CHUNK_SIZE_32      =          32ULL,
    UVM_CHUNK_SIZE_64      =          64ULL,
    UVM_CHUNK_SIZE_128     =         128ULL,
    UVM_CHUNK_SIZE_256     =         256ULL,
    UVM_CHUNK_SIZE_512     =         512ULL,
    UVM_CHUNK_SIZE_1K      =        1024ULL,
    UVM_CHUNK_SIZE_2K      =      2*1024ULL,
    UVM_CHUNK_SIZE_4K      =      4*1024ULL,
    UVM_CHUNK_SIZE_8K      =      8*1024ULL,
    UVM_CHUNK_SIZE_16K     =     16*1024ULL,
    UVM_CHUNK_SIZE_32K     =     32*1024ULL,
    UVM_CHUNK_SIZE_64K     =     64*1024ULL,
    UVM_CHUNK_SIZE_128K    =    128*1024ULL,
    UVM_CHUNK_SIZE_256K    =    256*1024ULL,
    UVM_CHUNK_SIZE_512K    =    512*1024ULL,
    UVM_CHUNK_SIZE_1M      =   1024*1024ULL,
    UVM_CHUNK_SIZE_2M      = 2*1024*1024ULL,
    UVM_CHUNK_SIZE_MAX     = UVM_CHUNK_SIZE_2M,
    UVM_CHUNK_SIZE_INVALID = UVM_CHUNK_SIZE_MAX * 2ULL
} uvm_chunk_size_t;

typedef enum __attribute__ ((__packed__))
{
    // Memory type for backing user pages. On Pascal+ it can be evicted.
    UVM_PMM_GPU_MEMORY_TYPE_USER,

    // Memory type for internal UVM allocations. It cannot be evicted.
    UVM_PMM_GPU_MEMORY_TYPE_KERNEL,

    // Number of types - MUST BE LAST.
    UVM_PMM_GPU_MEMORY_TYPE_COUNT
} uvm_pmm_gpu_memory_type_t;

typedef enum __attribute__ ((__packed__))
{
    // Chunk belongs to PMA. Code outside PMM should not have access to
    // it and it is likely a bug in UVM code (either in PMM or outside)
    // if that happens.
    UVM_PMM_GPU_CHUNK_STATE_PMA_OWNED,

    // Chunk is on free list. That is it can be reused or returned to PMA
    // as soon as its tracker is done. Code outside PMM should not have
    // access to this chunk and it is likely a bug in UVM code (either in
    // PMM or outside) if that happens.
    //
    // Please note that this is the only state requiring a list lock to
    // access the page.
    UVM_PMM_GPU_CHUNK_STATE_FREE,

    // Chunk is split into subchunks. Code outside PMM should not have
    // access to this chunk and it is likely a bug in UVM code (either in
    // PMM or outside) if that happens.
    UVM_PMM_GPU_CHUNK_STATE_IS_SPLIT,

    // Chunk is temporarily pinned. I.e. it has been just allocated and
    // it is about to be mapped. It should be released by either
    // uvm_pmm_gpu_unpin_temp if it backs some VA Chunk or
    // uvm_pmm_gpu_free if allocation was unnecessary.
    //
    // Please note that chunk in this state MUST NOT back an unlocked
    // page.
    UVM_PMM_GPU_CHUNK_STATE_TEMP_PINNED,

    // Chunk is allocated. That is it is backing some VA Chunk.
    UVM_PMM_GPU_CHUNK_STATE_ALLOCATED,

    // Number of states - MUST BE LAST
    UVM_PMM_GPU_CHUNK_STATE_COUNT
} uvm_pmm_gpu_chunk_state_t;

typedef enum
{
    // No flags passed
    UVM_PMM_ALLOC_FLAGS_NONE,

    // If there is no free memory, allocation may evict chunks
    // instead of returning error immediately. Therefore it must not be called
    // under VAB or chunk lock.
    UVM_PMM_ALLOC_FLAGS_EVICT = (1 << 0),

    UVM_PMM_ALLOC_FLAGS_MASK = (1 << 1) - 1
} uvm_pmm_alloc_flags_t;

// Maximum chunk sizes per type of allocation in single GPU.
// The worst case today is Maxwell with 4 allocations sizes for page tables and
// 2 page sizes used by uvm_mem_t.
#define UVM_MAX_CHUNK_SIZES 6

// This specifies a maximum GAP between 2 allocation levels.
#define UVM_PMM_MAX_SUBCHUNKS UVM_CHUNK_SIZE_MAX

#define UVM_PMM_CHUNK_SPLIT_CACHE_SIZES (ilog2(UVM_PMM_MAX_SUBCHUNKS) + 1)
#define UVM_CHUNK_SIZE_MASK_SIZE (ilog2(UVM_CHUNK_SIZE_MAX) + 1)

typedef uvm_chunk_size_t uvm_chunk_sizes_mask_t;

typedef struct uvm_pmm_gpu_chunk_suballoc_struct uvm_pmm_gpu_chunk_suballoc_t;

typedef struct uvm_gpu_chunk_struct uvm_gpu_chunk_t;

struct uvm_gpu_chunk_struct
{
    // Physical address of GPU chunk. This may be removed to save memory
    // if we will be able to get it from reverse map and changed
    // into smaller index for subchunks.
    NvU64 address;

    // Chunk lock protecting most of chunk. This is a bit lock.
    // TODO: Move it out of struct to avoid wasting bits.
    unsigned long lock;

    // List entry. This is protected by the pmm lock
    struct list_head list;

    // Tracker for the last GPU action performed on the chunk.
    uvm_tracker_t tracker;

    // If this is subchunk it points to the parent - in other words
    // chunk of bigger size which contains this chunk.
    uvm_gpu_chunk_t *parent;

    // Array describing suballocations
    uvm_pmm_gpu_chunk_suballoc_t *suballoc;

#ifdef DEBUG
    // Size of chunk
    uvm_chunk_size_t size;

    // Type of chunk
    uvm_pmm_gpu_memory_type_t type;
#endif

    // State of chunk
    uvm_pmm_gpu_chunk_state_t state;
};

typedef struct
{
    // Sizes of the MMU
    uvm_chunk_sizes_mask_t chunk_sizes[UVM_PMM_GPU_MEMORY_TYPE_COUNT];

    // Address space handle for RM backend allocations
    //
    // TODO: REMOVE: this should be using gpu->rm_address_space. There should
    // not be a separate copy here:
    uvmGpuAddressSpaceHandle rm_address_space;

    // PMA (Physical Memory Allocator) opaque handle
    void *pma;

    // Lock protecting lists
    uvm_mutex_t lock;

    // Free chunk lists
    struct list_head free_list[UVM_PMM_GPU_MEMORY_TYPE_COUNT][UVM_MAX_CHUNK_SIZES];

    // The mask of the initialized chunk sizes
    DECLARE_BITMAP(chunk_split_cache_initialized, UVM_PMM_CHUNK_SPLIT_CACHE_SIZES);

    // Delayed free work item
    struct work_struct delayed_free;
} uvm_pmm_gpu_t;

// Initialize PMM on GPU
NV_STATUS uvm_pmm_gpu_init(uvm_gpu_t *gpu, uvm_pmm_gpu_t *pmm);

// Deinitialize the PMM on GPU
void uvm_pmm_gpu_deinit(uvm_pmm_gpu_t *pmm);

// Allocates num_chunks chunks of size chunk_size in caller-supplied array (chunks).
//
// Returned chunks are in the TEMP_PINNED state, requiring a call to either
// uvm_pmm_gpu_unpin_temp or uvm_pmm_gpu_free. Each chunk's tracker is not
// guaranteed to be complete, so the caller must acquire or otherwise wait
// for it before using that chunk.
NV_STATUS uvm_pmm_gpu_alloc(uvm_pmm_gpu_t *pmm,
                            size_t num_chunks,
                            uvm_chunk_size_t chunk_size,
                            uvm_pmm_gpu_memory_type_t mem_type,
                            uvm_gpu_chunk_t **chunks,
                            uvm_pmm_alloc_flags_t flags);

// Helper for allocating kernel memory
//
// Internally calls uvm_pmm_gpu_alloc() and sets the state of all chunks to
// allocated on success.
NV_STATUS uvm_pmm_gpu_alloc_kernel(uvm_pmm_gpu_t *pmm,
                                   size_t num_chunks,
                                   uvm_chunk_size_t chunk_size,
                                   uvm_gpu_chunk_t **chunks,
                                   uvm_pmm_alloc_flags_t flags);

#define uvm_assert_chunk_locked(pmm, chunk) uvm_assert_bit_locked(&(chunk)->lock, 0)

static inline void uvm_pmm_gpu_lock_chunk(uvm_pmm_gpu_t *pmm, uvm_gpu_chunk_t *chunk)
{
    uvm_bit_lock(&chunk->lock, 0);
    uvm_assert_chunk_locked(pmm, chunk);
}

static inline void uvm_pmm_gpu_unlock_chunk(uvm_pmm_gpu_t *pmm, uvm_gpu_chunk_t *chunk)
{
    uvm_assert_chunk_locked(pmm, chunk);
    uvm_bit_unlock(&chunk->lock, 0);
}

// Unpins a temporarily pinned chunk. The chunk must be locked
// and assigned to some page. Caller must unlock chunks later.
void uvm_pmm_gpu_unpin_temp(uvm_pmm_gpu_t *pmm, uvm_gpu_chunk_t *chunk);

// Frees the chunk. The chunk must be locked. This also unpins the chunk if it is
// temporarily pinned. The tracker doesn't have to be completed.
void uvm_pmm_gpu_free(uvm_pmm_gpu_t *pmm,
                      uvm_chunk_size_t chunk_size,
                      uvm_pmm_gpu_memory_type_t mem_type,
                      uvm_gpu_chunk_t *chunk);

// Helper for freeing a kernel memory chunk that locks the chunk and calls uvm_pmm_gpu_free().
void uvm_pmm_gpu_free_kernel(uvm_pmm_gpu_t *pmm, uvm_chunk_size_t chunk_size, uvm_gpu_chunk_t *chunk);

// Is PMM using RM, or PMA, to satisfy vidmem physical allocation requests? This
// is a global, not per-GPU, setting.
//
// TODO: Using PMA is the long-term goal, so once PMA stabilizes, the whole "RM or PMA"
// infrastructure can be removed.

typedef enum {
    UVM_PMM_MODE_INVALID,

    // Default, fallback, mode for PMM until PMA is working
    UVM_PMM_MODE_RM,

    // Use new PMM interface
    UVM_PMM_MODE_PMA
} uvm_pmm_mode_t;

uvm_pmm_mode_t get_pmm_mode(void);

// Returns a string-ized version of the current uvm_pmm_mode_t: "UVM_PMM_MODE_PMA", for example.
const char *get_pmm_mode_string(void);

#endif
