/*******************************************************************************
    Copyright (c) 2015 NVIDIA Corporation

    Permission is hereby granted, free of charge, to any person obtaining a copy
    of this software and associated documentation files (the "Software"), to
    deal in the Software without restriction, including without limitation the
    rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
    sell copies of the Software, and to permit persons to whom the Software is
    furnished to do so, subject to the following conditions:

        The above copyright notice and this permission notice shall be
        included in all copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
    IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
    FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
    THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
    LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
    FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
    DEALINGS IN THE SOFTWARE.

*******************************************************************************/

#include "uvm_ioctl.h"
#include "uvm8_va_space.h"
#include "uvm8_va_range.h"
#include "uvm8_api.h"

static bool preferred_location_is_va_range_split_needed(uvm_va_range_t *va_range, void *data)
{
    uvm_processor_id_t processor_id;
    UVM_ASSERT(data);
    processor_id = *(uvm_processor_id_t*)data;
    return (processor_id != va_range->preferred_location);
}

static NV_STATUS preferred_location_set(uvm_va_space_t *va_space,
                                        NvU64 base,
                                        NvU64 length,
                                        uvm_processor_id_t preferred_location,
                                        uvm_va_range_t **first_va_range_to_migrate)
{
    uvm_va_range_t *va_range, *va_range_last;
    NvU64 last_address = base + length - 1;
    NV_STATUS status = NV_OK;

    uvm_assert_rwsem_locked_write(&va_space->lock);

    if (first_va_range_to_migrate)
        *first_va_range_to_migrate = NULL;

    status = uvm_va_space_split_span_as_needed(va_space,
                                               base,
                                               last_address + 1,
                                               preferred_location_is_va_range_split_needed,
                                               &preferred_location);
    if (status != NV_OK)
        return status;

    va_range_last = NULL;
    uvm_for_each_mmap_va_range_in_contig(va_range, va_space, base, last_address) {
        va_range_last = va_range;

        // If we didn't split the ends, check that they match
        if (va_range->node.start < base || va_range->node.end > last_address)
            UVM_ASSERT(va_range->preferred_location == preferred_location);

        uvm_va_range_set_preferred_location(va_range, preferred_location);

        // Return the first VA range that needs to be migrated so the caller
        // function doesn't need to traverse the tree again
        if (first_va_range_to_migrate && !(*first_va_range_to_migrate) &&
            preferred_location != UVM8_MAX_PROCESSORS &&
            !uvm_range_group_migratable(va_range->range_group))
            *first_va_range_to_migrate = va_range;

    }

    // Check that we were able to iterate over the entire range without any gaps
    if (!va_range_last || va_range_last->node.end < last_address)
        return NV_ERR_INVALID_ADDRESS;

    return NV_OK;
}

NV_STATUS uvm_api_set_preferred_location(UVM_SET_PREFERRED_LOCATION_PARAMS *params, struct file *filp)
{
    uvm_va_space_t *va_space = uvm_va_space_get(filp);
    uvm_va_range_t *va_range = NULL;
    uvm_va_range_t *first_va_range_to_migrate = NULL;
    uvm_processor_id_t preferred_location_id;
    NV_STATUS status = NV_OK;
    bool has_va_space_write_lock;

    UVM_ASSERT(va_space);

    // Check address and length alignment
    if (uvm_api_range_invalid(params->requestedBase, params->length))
        return NV_ERR_INVALID_ADDRESS;

    uvm_down_write(&va_space->lock);
    has_va_space_write_lock = true;

    // If the CPU is the preferred location, we don't have to find the associated uvm_gpu_t
    if (uvm_uuid_is_cpu(&params->preferredLocation)) {
        preferred_location_id = UVM_CPU_ID;
    }
    else {
        // Translate preferredLocation into a live GPU ID:
        status = uvm_registered_uuid_to_id(va_space, &params->preferredLocation, &preferred_location_id);
        if (status != NV_OK)
            goto done;
    }

    status = preferred_location_set(va_space, params->requestedBase, params->length, preferred_location_id,
                                    &first_va_range_to_migrate);
    if (status != NV_OK)
        goto done;

    // No VA range to migrate, early exit
    if (!first_va_range_to_migrate)
        goto done;

    uvm_downgrade_write(&va_space->lock);
    has_va_space_write_lock = false;

    // No need to check for holes in the VA ranges span here, this was checked by preferred_location_set
    for (va_range = first_va_range_to_migrate; va_range;
         va_range = uvm_va_space_iter_next(va_range, params->requestedBase + params->length - 1)) {
        if (!uvm_range_group_migratable(va_range->range_group)) {
            // TODO: migrate VA range to its preferred location
            // TODO: update range_groups_migration test to remove manual unmaps
        }
    }

done:
    if (has_va_space_write_lock)
        uvm_up_write(&va_space->lock);
    else
        uvm_up_read(&va_space->lock);
    return status;
}

NV_STATUS uvm_api_unset_preferred_location(UVM_UNSET_PREFERRED_LOCATION_PARAMS *params, struct file *filp)
{
    uvm_va_space_t *va_space = uvm_va_space_get(filp);
    NV_STATUS status = NV_OK;

    UVM_ASSERT(va_space);

    // Check address and length alignment
    if (uvm_api_range_invalid(params->requestedBase, params->length))
         return NV_ERR_INVALID_ADDRESS;

    uvm_down_write(&va_space->lock);
    status = preferred_location_set(va_space, params->requestedBase, params->length, UVM8_MAX_PROCESSORS, NULL);
    uvm_up_write(&va_space->lock);
    return status;
}

typedef struct
{
    uvm_processor_id_t processor_id;
    bool set_bit;
} accessed_by_split_params_t;

static bool accessed_by_is_va_range_split_needed(uvm_va_range_t *va_range, void *data)
{
    accessed_by_split_params_t *params = (accessed_by_split_params_t*)data;
    UVM_ASSERT(params);

    return (uvm_processor_mask_test(&va_range->accessed_by, params->processor_id) != params->set_bit);
}

static NV_STATUS accessed_by_set(uvm_va_space_t *va_space,
                                 NvU64 base,
                                 NvU64 length,
                                 NvProcessorUuid *processor_uuid,
                                 bool set_bit)
{
    uvm_processor_id_t processor_id;
    uvm_va_range_t *va_range, *va_range_last;
    NvU64 last_address = base + length - 1;
    accessed_by_split_params_t split_params;
    NV_STATUS status = NV_OK;

    UVM_ASSERT(va_space);

    // Check address and length alignment
    if (uvm_api_range_invalid(base, length))
        return NV_ERR_INVALID_ADDRESS;

    uvm_down_write(&va_space->lock);

    // If the CPU is the accessed by processor, we don't have to find the associated uvm_gpu_t
    if (uvm_uuid_is_cpu(processor_uuid)) {
        processor_id = UVM_CPU_ID;
    }
    else {
        // Translate processor_uuid into a live GPU ID:
        status = uvm_registered_uuid_to_id(va_space, processor_uuid, &processor_id);
        if (status != NV_OK)
            goto done;
    }

    split_params.processor_id = processor_id;
    split_params.set_bit = set_bit;
    status = uvm_va_space_split_span_as_needed(va_space,
                                               base,
                                               last_address + 1,
                                               accessed_by_is_va_range_split_needed,
                                               &split_params);
    if (status != NV_OK)
        goto done;

    va_range_last = NULL;
    uvm_for_each_mmap_va_range_in_contig(va_range, va_space, base, last_address) {
        va_range_last = va_range;

        // If we didn't split the ends, check that they match
        if (va_range->node.start < base || va_range->node.end > last_address)
            UVM_ASSERT(uvm_processor_mask_test(&va_range->accessed_by, processor_id) == set_bit);

        if (set_bit) {
            uvm_processor_mask_set(&va_range->accessed_by, processor_id);

            // TODO: Add mappings of this VA range on that processor
        }
        else {
            uvm_processor_mask_clear(&va_range->accessed_by, processor_id);
        }
    }

    // Check that we were able to iterate over the entire range without any gaps
    if (!va_range_last || va_range_last->node.end < last_address)
        status = NV_ERR_INVALID_ADDRESS;

done:
    uvm_up_write(&va_space->lock);
    return status;
}

NV_STATUS uvm_api_set_accessed_by(UVM_SET_ACCESSED_BY_PARAMS *params, struct file *filp)
{
    uvm_va_space_t *va_space = uvm_va_space_get(filp);

    return accessed_by_set(va_space, params->requestedBase, params->length, &params->accessedByUuid, true);
}

NV_STATUS uvm_api_unset_accessed_by(UVM_UNSET_ACCESSED_BY_PARAMS *params, struct file *filp)
{
    uvm_va_space_t *va_space = uvm_va_space_get(filp);

    return accessed_by_set(va_space, params->requestedBase, params->length, &params->accessedByUuid, false);
}

static bool read_duplication_is_va_range_split_needed(uvm_va_range_t *va_range, void *data)
{
    uvm_read_duplication_policy_t new_policy;
    UVM_ASSERT(data);
    new_policy = *(uvm_read_duplication_policy_t *)data;
    return va_range->read_duplication != new_policy;
}

static NV_STATUS read_duplication_set(uvm_va_space_t *va_space,
                                      NvU64 base,
                                      NvU64 length,
                                      bool enable)
{
    uvm_va_range_t *va_range, *va_range_last;
    NvU64 last_address = base + length - 1;
    NV_STATUS status = NV_OK;

    // Note that we never set the policy back to UNSET
    uvm_read_duplication_policy_t new_policy = enable ? UVM_READ_DUPLICATION_ENABLED : UVM_READ_DUPLICATION_DISABLED;

    UVM_ASSERT(va_space);

    // Check address and length alignment
    if (uvm_api_range_invalid(base, length))
        return NV_ERR_INVALID_ADDRESS;

    uvm_down_write(&va_space->lock);

    status = uvm_va_space_split_span_as_needed(va_space,
                                               base,
                                               last_address + 1,
                                               read_duplication_is_va_range_split_needed,
                                               &new_policy);
    if (status != NV_OK)
        goto done;

    va_range_last = NULL;
    uvm_for_each_mmap_va_range_in_contig(va_range, va_space, base, last_address) {
        va_range_last = va_range;

        // If we didn't split the ends, check that they match
        if (va_range->node.start < base || va_range->node.end > last_address)
            UVM_ASSERT(va_range->read_duplication == new_policy);

        va_range->read_duplication = new_policy;
    }

    // Check that we were able to iterate over the entire range without any gaps
    if (!va_range_last || va_range_last->node.end < last_address)
        status = NV_ERR_INVALID_ADDRESS;

done:
    uvm_up_write(&va_space->lock);
    return status;
}

NV_STATUS uvm_api_enable_read_duplication(UVM_ENABLE_READ_DUPLICATION_PARAMS *params, struct file *filp)
{
    uvm_va_space_t *va_space = uvm_va_space_get(filp);

    return read_duplication_set(va_space, params->requestedBase, params->length, true);
}

NV_STATUS uvm_api_disable_read_duplication(UVM_DISABLE_READ_DUPLICATION_PARAMS *params, struct file *filp)
{
    uvm_va_space_t *va_space = uvm_va_space_get(filp);

    return read_duplication_set(va_space, params->requestedBase, params->length, false);
}
