/*******************************************************************************
    Copyright (c) 2015 NVIDIA Corporation

    Permission is hereby granted, free of charge, to any person obtaining a copy
    of this software and associated documentation files (the "Software"), to
    deal in the Software without restriction, including without limitation the
    rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
    sell copies of the Software, and to permit persons to whom the Software is
    furnished to do so, subject to the following conditions:

        The above copyright notice and this permission notice shall be
        included in all copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
    IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
    FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
    THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
    LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
    FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
    DEALINGS IN THE SOFTWARE.

*******************************************************************************/

#include "uvm8_range_group.h"
#include "uvm8_va_space.h"
#include "uvm8_va_range.h"
#include "uvm_ioctl.h"
#include "uvm8_api.h"

static struct kmem_cache *g_uvm_range_group_cache __read_mostly;

NV_STATUS uvm_range_group_init(void)
{
    g_uvm_range_group_cache = NV_KMEM_CACHE_CREATE("uvm_range_group_t", uvm_range_group_t);
    if (!g_uvm_range_group_cache)
        return NV_ERR_NO_MEMORY;
    return NV_OK;
}

void uvm_range_group_exit(void)
{
    kmem_cache_destroy_safe(&g_uvm_range_group_cache);
}

NV_STATUS uvm_api_create_range_group(UVM_CREATE_RANGE_GROUP_PARAMS *params, struct file *filp)
{
    uvm_va_space_t *va_space = uvm_va_space_get(filp);
    uvm_range_group_t *range_group = NULL;
    NV_STATUS status = NV_OK;
    int ret;

    range_group = kmem_cache_zalloc(g_uvm_range_group_cache, NV_UVM_GFP_FLAGS);
    if (!range_group)
        return NV_ERR_NO_MEMORY;

    range_group->id = atomic64_inc_return(&va_space->range_group_id_counter);
    UVM_ASSERT(range_group->id != UVM_RANGE_GROUP_ID_NONE);

    atomic_set(&range_group->allow_migration, 1);
    INIT_LIST_HEAD(&range_group->va_ranges);

    uvm_down_write(&va_space->lock);

    ret = radix_tree_insert(&va_space->range_groups, range_group->id, range_group);
    status = errno_to_nv_status(ret);
    if (status != NV_OK) {
        kmem_cache_free(g_uvm_range_group_cache, range_group);
        goto done;
    }

    params->rangeGroupId = range_group->id;

done:
    uvm_up_write(&va_space->lock);

    return status;
}

static void uvm_range_group_destroy(uvm_range_group_t *range_group)
{
    uvm_va_range_t *va_range, *tmp;

    list_for_each_entry_safe(va_range, tmp, &range_group->va_ranges, range_group_node) {
        UVM_ASSERT(va_range->range_group == range_group);

        // Associate va_range with UVM_RANGE_GROUP_ID_NONE
        va_range->range_group = NULL;

        // use list_del_init so uvm_va_range_destroy can use list_del later on
        list_del_init(&va_range->range_group_node);
    }

    kmem_cache_free(g_uvm_range_group_cache, range_group);
}

NV_STATUS uvm_api_destroy_range_group(UVM_DESTROY_RANGE_GROUP_PARAMS *params, struct file *filp)
{
    uvm_va_space_t *va_space = uvm_va_space_get(filp);
    uvm_range_group_t *range_group = NULL;
    NV_STATUS status = NV_OK;

    uvm_down_write(&va_space->lock);

    range_group = radix_tree_delete(&va_space->range_groups, params->rangeGroupId);
    if (!range_group) {
        status = NV_ERR_OBJECT_NOT_FOUND;
        goto done;
    }

    uvm_range_group_destroy(range_group);

done:
    uvm_up_write(&va_space->lock);
    return status;
}

void uvm_range_group_radix_tree_destroy(struct radix_tree_root *root)
{
    uvm_range_group_t *range_group = NULL;
    NvU64 index = 1;

    while (radix_tree_gang_lookup(root, (void**)&range_group, index, 1)) {
        UVM_ASSERT(range_group);
        radix_tree_delete(root, range_group->id);
        index = range_group->id + 1;
        uvm_range_group_destroy(range_group);
    }
}

static bool uvm_range_group_is_va_range_split_needed(uvm_va_range_t *va_range, void *data)
{
    return (data != va_range->range_group);
}

NV_STATUS uvm_api_set_range_group(UVM_SET_RANGE_GROUP_PARAMS *params, struct file *filp)
{
    uvm_va_space_t *va_space = uvm_va_space_get(filp);
    uvm_range_group_t *range_group = NULL;
    uvm_va_range_t *va_range, *va_range_last;
    unsigned long long last_address = params->requestedBase + params->length - 1;
    NV_STATUS status = NV_OK;
    bool has_va_space_write_lock;

    UVM_ASSERT(va_space);

    // Check address and length alignment
    if (uvm_api_range_invalid(params->requestedBase, params->length))
        return NV_ERR_INVALID_ADDRESS;

    uvm_down_write(&va_space->lock);
    has_va_space_write_lock = true;

    // Check that range group exists
    range_group = radix_tree_lookup(&va_space->range_groups, params->rangeGroupId);
    if (!range_group && (params->rangeGroupId != UVM_RANGE_GROUP_ID_NONE)) {
        status = NV_ERR_OBJECT_NOT_FOUND;
        goto done;
    }

    status = uvm_va_space_split_span_as_needed(va_space,
                                               params->requestedBase,
                                               last_address + 1,
                                               uvm_range_group_is_va_range_split_needed,
                                               range_group);
    if (status != NV_OK)
        goto done;

    va_range_last = NULL;
    uvm_for_each_mmap_va_range_in_contig(va_range, va_space, params->requestedBase, last_address) {
        va_range_last = va_range;

        // If we didn't split the ends, check that they match
        if (va_range->node.start < params->requestedBase || va_range->node.end > last_address)
            UVM_ASSERT(va_range->range_group == range_group);

        if (!uvm_range_group_migratable(range_group) &&
            va_range->preferred_location == UVM8_MAX_PROCESSORS) {
            status = NV_ERR_INVALID_ADDRESS;
            goto done;
        }

        uvm_va_range_set_range_group(va_range, range_group);
    }

    // Check that we were able to iterate over the entire range without any gaps
    if (!va_range_last || va_range_last->node.end < last_address) {
        status = NV_ERR_INVALID_ADDRESS;
        goto done;
    }

    // Early exit if no need for migration
    if (uvm_range_group_migratable(range_group))
        goto done;

    uvm_downgrade_write(&va_space->lock);
    has_va_space_write_lock = false;

    // Already checked for gaps above
    uvm_for_each_va_range_in(va_range, va_space, params->requestedBase, last_address) {
        // TODO: migrate VA range to its preferred location
    }

done:
    if (has_va_space_write_lock)
        uvm_up_write(&va_space->lock);
    else
        uvm_up_read(&va_space->lock);
    return status;
}

static NV_STATUS uvm_range_group_set_migration_policy(uvm_range_group_t *range_group,
                                                      uvm_va_space_t *va_space,
                                                      bool allow_migration)
{
    uvm_va_range_t *va_range;
    NV_STATUS status = NV_OK;

    UVM_ASSERT(range_group);
    UVM_ASSERT(va_space);

    uvm_assert_rwsem_locked_read(&va_space->lock);

    atomic_set(&range_group->allow_migration, !!allow_migration);

    // Nothing else to do if we're allowing migrations
    if (allow_migration)
        goto done;

    list_for_each_entry(va_range, &range_group->va_ranges, range_group_node) {
        UVM_ASSERT(va_range->range_group == range_group);

        // VA ranges need to have a preferred location set in order for their
        // range group to be set to non-migratable.
        if (va_range->preferred_location == UVM8_MAX_PROCESSORS) {
            status = NV_ERR_INVALID_ARGUMENT;
            goto done;
        }

        // TODO: migrate VA range to its preferred location
        // TODO: update range_groups_migration test to remove manual unmaps
    }

done:
    return status;
}

static NV_STATUS uvm_range_groups_set_migration_policy(uvm_va_space_t *va_space,
                                                       NvU64 *range_group_ids,
                                                       NvU64 num_group_ids,
                                                       bool allow_migration)
{
    NV_STATUS status = NV_OK;
    NvU64 i;

    UVM_ASSERT(va_space);

    if (!range_group_ids || num_group_ids == 0 || num_group_ids > UVM_MAX_RANGE_GROUPS_PER_IOCTL_CALL)
        return NV_ERR_INVALID_ARGUMENT;

    uvm_down_read(&va_space->lock);

    for (i = 0; i < num_group_ids; ++i) {
        uvm_range_group_t *range_group = radix_tree_lookup(&va_space->range_groups, range_group_ids[i]);
        if (!range_group) {
            status = NV_ERR_OBJECT_NOT_FOUND;
            goto done;
        }

        status = uvm_range_group_set_migration_policy(range_group, va_space, allow_migration);
        if (status != NV_OK)
            goto done;
    }

done:
    uvm_up_read(&va_space->lock);
    return status;
}

NV_STATUS uvm_api_prevent_migration_range_groups(UVM_PREVENT_MIGRATION_RANGE_GROUPS_PARAMS *params, struct file *filp)
{
    uvm_va_space_t *va_space = uvm_va_space_get(filp);

    return uvm_range_groups_set_migration_policy(va_space, params->rangeGroupIds, params->numGroupIds, false);
}

NV_STATUS uvm_api_allow_migration_range_groups(UVM_ALLOW_MIGRATION_RANGE_GROUPS_PARAMS *params, struct file *filp)
{
    uvm_va_space_t *va_space = uvm_va_space_get(filp);

    return uvm_range_groups_set_migration_policy(va_space, params->rangeGroupIds, params->numGroupIds, true);
}
