/*******************************************************************************
    Copyright (c) 2016 NVIDIA Corporation

    Permission is hereby granted, free of charge, to any person obtaining a copy
    of this software and associated documentation files (the "Software"), to
    deal in the Software without restriction, including without limitation the
    rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
    sell copies of the Software, and to permit persons to whom the Software is
    furnished to do so, subject to the following conditions:

        The above copyright notice and this permission notice shall be
        included in all copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
    IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
    FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
    THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
    LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
    FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
    DEALINGS IN THE SOFTWARE.

*******************************************************************************/

#include "uvm_common.h"
#include "uvm_linux.h"
#include "uvm8_va_space.h"
#include "uvm8_va_range.h"
#include "uvm8_lock.h"
#include "uvm8_hal_types.h"
#include "uvm8_mmu.h"
#include "uvm8_user_channel.h"
#include "uvm8_kvmalloc.h"
#include "uvm8_api.h"
#include "uvm8_gpu.h"
#include "uvm8_tracker.h"
#include "uvm8_map_external.h"
#include "nv_uvm_interface.h"

#include <linux/sort.h>

static NV_STATUS uvm_user_channel_create(uvm_gpu_va_space_t *gpu_va_space,
                                         uvm_rm_user_object_t *user_rm_channel,
                                         uvm_user_channel_t **out_channel)
{
    UvmGpuChannelInstanceInfo channel_info;
    uvm_user_channel_t *user_channel = NULL;
    NV_STATUS status;
    NvU32 rm_client = user_rm_channel->user_client;
    NvU32 rm_channel = user_rm_channel->user_object;

    user_channel = uvm_kvmalloc_zero(sizeof(*user_channel));
    if (!user_channel)
        return NV_ERR_NO_MEMORY;

    user_channel->gpu_va_space = gpu_va_space;
    list_add(&user_channel->list_node, &gpu_va_space->registered_channels);
    memcpy(&user_channel->user_rm_channel, user_rm_channel, sizeof(*user_rm_channel));

    // Convert the user channel handles into a handle safe for kernel use.
    // This also takes a ref-count on the instance pointer, though not on
    // the channel ID.
    //
    // TODO: This interface needs to use rm_control_fd to do validation
    memset(&channel_info, 0, sizeof(channel_info));
    (void)user_channel->user_rm_channel.rm_control_fd;
    status = uvm_rm_locked_call(nvUvmInterfaceRetainChannel(gpu_va_space->duped_gpu_va_space,
                                                            rm_client,
                                                            rm_channel,
                                                            &channel_info));
    if (status != NV_OK) {
        UVM_DBG_PRINT("Failed to retain channel {0x%x, 0x%x}: %s, GPU: %s\n",
                      rm_client, rm_channel, nvstatusToString(status), gpu_va_space->gpu->name);
        goto error;
    }

    if (channel_info.sysmem)
        user_channel->instance_ptr.aperture = UVM_APERTURE_SYS;
    else
        user_channel->instance_ptr.aperture = UVM_APERTURE_VID;
    user_channel->instance_ptr.address = channel_info.base;
    user_channel->instance_ptr_desc    = channel_info.instanceDescriptor;
    user_channel->hw_channel_id        = channel_info.chId;
    user_channel->num_resource_ranges  = channel_info.resourceCount;

    *out_channel = user_channel;
    return NV_OK;

error:
    if (user_channel) {
        if (user_channel->instance_ptr_desc)
            uvm_rm_locked_call_void(nvUvmInterfaceReleaseChannel(user_channel->instance_ptr_desc));
        list_del(&user_channel->list_node);
        uvm_kvfree(user_channel);
    }
    return status;
}

// Sort channel resources from highest to lowest alignments
static int resource_align_high_cmp(const void *a, const void *b)
{
    const UvmGpuChannelResourceInfo *resource_a = a;
    const UvmGpuChannelResourceInfo *resource_b = b;

    if (resource_a->alignment > resource_b->alignment)
        return -1;
    if (resource_a->alignment < resource_b->alignment)
        return 1;
    return 0;
}

static NV_STATUS get_rm_channel_resources(uvm_user_channel_t *user_channel,
                                          UvmGpuChannelResourceInfo **out_resource_info)
{
    UvmGpuChannelResourceInfo *resource_info = NULL;
    uvm_gpu_va_space_t *gpu_va_space = user_channel->gpu_va_space;
    NV_STATUS status;
    NvU32 i, num_resources = user_channel->num_resource_ranges;

    resource_info = uvm_kvmalloc_zero(num_resources * sizeof(resource_info[0]));
    if (!resource_info)
        return NV_ERR_NO_MEMORY;

    status = uvm_rm_locked_call(nvUvmInterfaceRetainChannelResources(gpu_va_space->duped_gpu_va_space,
                                                                     user_channel->instance_ptr_desc,
                                                                     num_resources,
                                                                     resource_info));
    if (status != NV_OK) {
        UVM_DBG_PRINT("Failed to get channel resources for {0x%x, 0x%x}: %s, GPU: %s\n",
                      user_channel->user_rm_channel.user_client, user_channel->user_rm_channel.user_object,
                      nvstatusToString(status), gpu_va_space->gpu->name);
        goto error;
    }

    // Info fix-up
    for (i = 0; i < num_resources; i++) {
        UvmGpuMemoryInfo *mem_info = &resource_info[i].resourceInfo;

        // RM can return alignments of 0, so make sure it's at least page size
        // before we start using it.
        resource_info[i].alignment = max(resource_info[i].alignment, (NvU64)mem_info->pageSize);

        // RM tracks logical size, so the size might not be a multiple of page
        // size. This would cause problems in our tracking.
        mem_info->size = UVM_ALIGN_UP(mem_info->size, mem_info->pageSize);
    }

    // Sort the resources from highest to lowest alignment. This should
    // guarantee that they fit in the provided VA space, regardless of the order
    // used to calculate the total size.
    sort(resource_info, num_resources, sizeof(resource_info[0]), resource_align_high_cmp, NULL);

    *out_resource_info = resource_info;
    return NV_OK;

error:
    uvm_kvfree(resource_info);
    return status;
}

static uvm_user_channel_t *find_user_channel(uvm_gpu_va_space_t *gpu_va_space,
                                             uvm_rm_user_object_t *user_rm_channel)
{
    uvm_user_channel_t *user_channel;
    list_for_each_entry(user_channel, &gpu_va_space->registered_channels, list_node) {
        if (user_channel->user_rm_channel.user_client == user_rm_channel->user_client &&
            user_channel->user_rm_channel.user_object == user_rm_channel->user_object)
            return user_channel;
    }
    return NULL;
}

static uvm_va_range_t *find_local_range(uvm_gpu_va_space_t *gpu_va_space, NvP64 rm_descriptor)
{
    uvm_va_range_t *range;
    list_for_each_entry(range, &gpu_va_space->local_channel_ranges, channel.local_node) {
        UVM_ASSERT(range->type == UVM_VA_RANGE_TYPE_CHANNEL);
        UVM_ASSERT(range->channel.ref_count > 0);
        UVM_ASSERT(range->channel.is_local);
        if (range->channel.rm_descriptor == rm_descriptor)
            return range;
    }
    return NULL;
}

// Allocate the VA range for the given channel resource and map it. The VA used
// is the first VA in the range [base, end] which has alignment appropriate to
// the given resource. Note that base is an in/out parameter: this function
// updates base if it actually allocates a range, which it may not for local
// ranges.
//
// If an error is returned, the caller is responsible for dropping resource's RM
// ref count. Otherwise, the resource will be freed by uvm_va_range_destroy.
static NV_STATUS uvm_user_channel_map_resource(uvm_user_channel_t *user_channel,
                                               NvU64 *base,
                                               NvU64 end,
                                               UvmGpuChannelResourceInfo *resource_info,
                                               uvm_va_range_t **out_va_range,
                                               uvm_tracker_t *tracker)
{
    uvm_gpu_va_space_t *gpu_va_space = user_channel->gpu_va_space;
    UvmGpuMemoryInfo *mem_info = &resource_info->resourceInfo;
    uvm_va_range_t *range = NULL;
    NvU64 curr_base, curr_end;
    uvm_aperture_t aperture;
    uvm_map_rm_params_t map_rm_params;
    NV_STATUS status;

    uvm_assert_rwsem_locked_write(&gpu_va_space->va_space->lock);

    if (mem_info->sysmem)
        aperture = UVM_APERTURE_SYS;
    else
        aperture = UVM_APERTURE_VID;

    // If the resource is local, see if we've already mapped it
    if (!resource_info->globalResource) {
        range = find_local_range(gpu_va_space, resource_info->resourceDescriptor);
        if (range) {
            // We've already mapped this resource, so just bump the ref count
            UVM_ASSERT(IS_ALIGNED(range->node.start, resource_info->alignment));
            UVM_ASSERT(uvm_va_range_size(range) >= mem_info->size);
            UVM_ASSERT(range->channel.aperture == aperture);
            UVM_ASSERT(range->channel.is_local);

            // Drop the extra RM ref-count we took when querying the handle
            // since we're tracking that ourselves.
            uvm_rm_locked_call_void(nvUvmInterfaceReleaseChannelResources(&resource_info->resourceDescriptor, 1));

            ++range->channel.ref_count;
            *out_va_range = range;
            return NV_OK;
        }
    }

    // Dumb allocator which just keeps advancing the region
    curr_base = UVM_ALIGN_UP(*base, resource_info->alignment);
    curr_end = curr_base + mem_info->size - 1;

    // Check for exceeding end and for arithmetic overflow
    if (curr_base < *base || curr_base > end ||
        curr_end  < *base || curr_end  > end ||
        curr_base > curr_end) {
        UVM_DBG_PRINT("Range exceeded: allowed [0x%llx, 0x%llx], align: 0x%llx size: 0x%llx\n",
                      *base, end, resource_info->alignment, mem_info->size, curr_base, curr_end);
        return NV_ERR_INVALID_ADDRESS;
    }

    // TODO: RM computes alignments incorrectly. Need to fix bug 1734586.

    status = uvm_va_range_create_channel(gpu_va_space->va_space, curr_base, curr_end, &range);
    if (status != NV_OK)
        goto error;

    range->channel.gpu_va_space = gpu_va_space;
    range->channel.aperture = aperture;
    range->channel.rm_descriptor = resource_info->resourceDescriptor;
    range->channel.is_local = !resource_info->globalResource;
    if (range->channel.is_local)
        list_add(&range->channel.local_node, &gpu_va_space->local_channel_ranges);
    range->channel.ref_count = 1;

    // Perform the mapping. Some of these need to be privileged and/or read-
    // only, so use default types to let RM set those fields.
    map_rm_params.map_offset = 0;
    map_rm_params.mapping_type = UvmGpuMappingTypeDefault;
    map_rm_params.caching_type = UvmGpuCachingTypeDefault;
    map_rm_params.tracker = tracker;
    status = uvm_va_range_map_rm_allocation(range, gpu_va_space->gpu, mem_info, &map_rm_params);
    if (status != NV_OK)
        goto error;

    *base = curr_end + 1;
    *out_va_range = range;
    return NV_OK;

error:
    // On error the caller can't know how far we got, so the caller will always
    // free the resource descriptor. That means we have to prevent this range
    // destroy from doing that itself.
    if (range) {
        range->channel.rm_descriptor = 0;
        range->channel.ref_count = 0; // Destroy assumes this
        uvm_va_range_destroy(range);
    }
    return status;
}

// "Binding" the resouces tells RM the virtual address of each allocation so it
// can in turn tell the HW where they are.
static NV_STATUS bind_channel_resources(uvm_user_channel_t *user_channel)
{
    UvmGpuChannelResourceBindParams *resource_va_list = NULL;
    uvm_gpu_va_space_t *gpu_va_space = user_channel->gpu_va_space;
    uvm_va_range_t *resource_range;
    NV_STATUS status = NV_OK;
    NvU32 i;
    NvU32 num_resources = user_channel->num_resource_ranges;

    resource_va_list = uvm_kvmalloc_zero(num_resources * sizeof(resource_va_list[0]));
    if (!resource_va_list) {
        status = NV_ERR_NO_MEMORY;
        goto out;
    }

    for (i = 0; i < num_resources; i++) {
        resource_range = user_channel->resource_ranges[i];
        UVM_ASSERT(resource_range);
        UVM_ASSERT(resource_range->type == UVM_VA_RANGE_TYPE_CHANNEL);

        resource_va_list[i].resourceDescriptor = resource_range->channel.rm_descriptor;
        resource_va_list[i].resourceVa         = resource_range->node.start;

        // TODO: Bug 1732494: RM should be able to figure this out on its own
        resource_va_list[i].globalResource     = !resource_range->channel.is_local;
    }

    status = uvm_rm_locked_call(nvUvmInterfaceBindChannelResources(gpu_va_space->duped_gpu_va_space,
                                                                   user_channel->instance_ptr_desc,
                                                                   num_resources,
                                                                   resource_va_list));
    if (status != NV_OK) {
        UVM_DBG_PRINT("Failed to bind channel resources for {0x%x, 0x%x}: %s, GPU: %s\n",
                      user_channel->user_rm_channel.user_client,
                      user_channel->user_rm_channel.user_object,
                      nvstatusToString(status),
                      gpu_va_space->gpu->name);
        goto out;
    }

out:
    uvm_kvfree(resource_va_list);
    return status;
}

// Helper for uvm_user_channel_map_resources's error path. Releases each
// resource of the input array which is not present as a VA range in the given
// channel.
static void cleanup_unmapped_resources(uvm_user_channel_t *user_channel,
                                       UvmGpuChannelResourceInfo *resource_info)
{
    size_t i;

    if (!user_channel || !resource_info)
        return;

    // If the resource is already in the user channel, it will get cleaned up by
    // channel/range destroy. If not, nobody else has the resource so we need to
    // drop the ref count with RM.
    for (i = 0; i < user_channel->num_resource_ranges; i++) {
        if (!resource_info[i].resourceDescriptor)
            continue;

        if (user_channel->resource_ranges[i])
            UVM_ASSERT(user_channel->resource_ranges[i]->channel.rm_descriptor == resource_info[i].resourceDescriptor);
        else
            uvm_rm_locked_call_void(nvUvmInterfaceReleaseChannelResources(&resource_info[i].resourceDescriptor, 1));
    }
}

// Channels need virtual allocations to operate, but we don't know about them.
// This function:
// - Queries RM to find out the physical allocations which need virtual mappings
// - Carves out a chunk within [base, end] for each allocation
// - Gets the PTEs for each allocation from RM and writes them
// - Tells RM the virtual addresses of each allocation so it can write them to
//   HW state.
//
// If an error is returned, the caller must call uvm_user_channel_detach and
// uvm_gpu_destroy_detached_channels.
static NV_STATUS uvm_user_channel_map_resources(uvm_user_channel_t *user_channel, NvU64 base, NvU64 end)
{
    UvmGpuChannelResourceInfo *resource_info = NULL;
    NvU64 running_base;
    uvm_tracker_t tracker = UVM_TRACKER_INIT();
    NvU32 i;
    NV_STATUS status, tracker_status;

    status = get_rm_channel_resources(user_channel, &resource_info);
    if (status != NV_OK)
        goto out;

    user_channel->resource_ranges = uvm_kvmalloc_zero(user_channel->num_resource_ranges *
                                                      sizeof(user_channel->resource_ranges[0]));
    if (!user_channel->resource_ranges) {
        status = NV_ERR_NO_MEMORY;
        goto out;
    }

    // Map all resources
    running_base = base;
    for (i = 0; i < user_channel->num_resource_ranges; i++) {
        // Check if we're out of space
        if (running_base > end) {
            status = NV_ERR_INVALID_ADDRESS;
            goto out;
        }

        status = uvm_user_channel_map_resource(user_channel,
                                               &running_base,
                                               end,
                                               &resource_info[i],
                                               &user_channel->resource_ranges[i],
                                               &tracker);
        if (status != NV_OK)
            goto out;
    }

    // TODO: Move this down to be the last operation done as part of channel
    //       registration. See bug 1737765.
    status = bind_channel_resources(user_channel);
    if (status != NV_OK)
        goto out;

out:
    // Always wait for the tracker even on error so we don't have any pending
    // map operations happening during the subsequent destroy.
    tracker_status = uvm_tracker_wait_deinit(&tracker);

    // Only clobber status if we didn't hit an earlier error
    if (status == NV_OK)
        status = tracker_status;

    // On error, it's up to the caller to call uvm_user_channel_detach and
    // uvm_gpu_destroy_detached_channels to tear down the instance pointer and
    // mappings we've created in here. It's up to us to drop RM's ref count on
    // resources which didn't get mapped.
    if (status != NV_OK)
        cleanup_unmapped_resources(user_channel, resource_info);

    uvm_kvfree(resource_info);
    return status;
}

// TODO: Temporary workaround tracked in bug 1739252. Switch this back to
//       uvm_api_range_invalid once CUDA passes in a system-page-aligned range.
static bool uvm_channel_api_range_invalid(NvU64 base, NvU64 length)
{
    return !IS_ALIGNED(base, UVM_PAGE_SIZE_4K)   ||
           !IS_ALIGNED(length, UVM_PAGE_SIZE_4K) ||
           length == 0           ||
           base + length < base; // Overflow
}

static NV_STATUS uvm_register_channel(uvm_va_space_t *va_space,
                                      NvProcessorUuid *uuid,
                                      uvm_rm_user_object_t *user_rm_channel,
                                      NvU64 base,
                                      NvU64 length)
{
    NV_STATUS status = NV_OK;
    uvm_gpu_t *gpu;
    uvm_gpu_va_space_t *gpu_va_space;
    NvU64 end = base + length - 1;
    uvm_user_channel_t *user_channel = NULL;
    LIST_HEAD(detached_channel);

    if (uvm_channel_api_range_invalid(base, length))
        return NV_ERR_INVALID_ADDRESS;

    uvm_down_write(&va_space->lock);

    gpu = uvm_va_space_get_gpu_by_uuid_with_gpu_va_space(va_space, uuid);
    if (!gpu) {
        status = NV_ERR_INVALID_DEVICE;
        goto done;
    }

    gpu_va_space = uvm_gpu_va_space_get(va_space, gpu);
    UVM_ASSERT(gpu_va_space);

    if (end >= gpu->max_channel_va) {
        status = NV_ERR_OUT_OF_RANGE;
        goto done;
    }

    if (!uvm_va_space_range_empty(va_space, base, end)) {
        status = NV_ERR_UVM_ADDRESS_IN_USE;
        goto done;
    }

    // Verify that this handle pair wasn't already registered. This is just to
    // keep our internal state consistent, since a match doesn't mean that the
    // handles still represent the same channel. See the comment in
    // uvm_user_channel.h.
    if (find_user_channel(gpu_va_space, user_rm_channel)) {
        status = NV_ERR_INVALID_CHANNEL;
        goto done;
    }

    // TODO: Check that this handle pair also wasn't already registered on other
    //       GPUs in the GPU's SLI group.

    status = uvm_user_channel_create(gpu_va_space, user_rm_channel, &user_channel);
    if (status != NV_OK)
        goto done;

    status = uvm_user_channel_map_resources(user_channel, base, end);
    if (status != NV_OK)
        goto done;

    // Tell the GPU page fault handler about this instance_ptr -> va_space
    // mapping
    status = uvm_gpu_add_instance_ptr(gpu, user_channel->instance_ptr, va_space);
    if (status != NV_OK)
        goto done;

done:
    if (!user_channel || status == NV_OK) {
        uvm_up_write(&va_space->lock);
        return status;
    }

    // We have to detach because the channel may have been put on the runlist
    // already, depending on which error brought us here.
    uvm_user_channel_detach(user_channel);
    list_add(&user_channel->list_node, &detached_channel);
    uvm_gpu_retain(gpu);

    uvm_up_write(&va_space->lock);

    uvm_gpu_destroy_detached_channels(gpu, &detached_channel);
    uvm_gpu_release(gpu);

    return status;
}

NV_STATUS uvm_api_register_channel(UVM_REGISTER_CHANNEL_PARAMS *params, struct file *filp)
{
    uvm_va_space_t *va_space = uvm_va_space_get(filp);
    uvm_rm_user_object_t user_rm_channel =
    {
        .rm_control_fd = params->rmCtrlFd,
        .user_client   = params->hClient,
        .user_object   = params->hChannel
    };
    return uvm_register_channel(va_space, &params->gpuUuid, &user_rm_channel, params->base, params->length);
}

void uvm_user_channel_detach(uvm_user_channel_t *user_channel)
{
    uvm_va_space_t *va_space;
    uvm_va_range_t *resource_range;
    NvU32 i;

    if (!user_channel)
        return;

    UVM_ASSERT(user_channel->gpu_va_space);
    va_space = user_channel->gpu_va_space->va_space;
    UVM_ASSERT(va_space);
    uvm_assert_rwsem_locked_write(&va_space->lock);

    if (user_channel->instance_ptr_desc != 0) {
        // Prevent the bottom half from servicing faults on this channel. Note
        // that this only prevents new faults from being serviced. It doesn't
        // flush out faults currently being serviced, nor prior faults still
        // pending in the fault buffer. Those are handled separately.
        uvm_gpu_remove_instance_ptr(user_channel->gpu_va_space->gpu, user_channel->instance_ptr);

        // Tell RM to kill the channel before we start unmapping its
        // allocations. This is to prevent spurious MMU faults during teardown.
        //
        // TODO: Bug 1737765. This doesn't stop the user from putting the
        //       channel back on the runlist, which could put stale instance
        //       pointers back in the fault buffer.
        uvm_rm_locked_call_void(nvUvmInterfaceStopChannel(user_channel->gpu_va_space->duped_gpu_va_space,
                                                          (void *)user_channel->instance_ptr_desc));

        // We can't free the instance pointer here since that requires flushing
        // the GPU fault buffer, which means taking the GPU isr_lock. The caller
        // is required to do that.
    }

    // Re-init so the list_del in uvm_user_channel_destroy_detached works
    // regardless of whether this channel was put on a different list.
    list_del_init(&user_channel->list_node);

    if (user_channel->resource_ranges) {
        for (i = 0; i < user_channel->num_resource_ranges; i++) {
            resource_range = user_channel->resource_ranges[i];
            if (!resource_range)
                continue;

            UVM_ASSERT(resource_range->type == UVM_VA_RANGE_TYPE_CHANNEL);

            // Drop the ref count on each each range
            if (resource_range->channel.is_local)
                UVM_ASSERT(resource_range->channel.ref_count > 0);
            else
                UVM_ASSERT(resource_range->channel.ref_count == 1);

            --resource_range->channel.ref_count;

            if (resource_range->channel.ref_count == 0)
                uvm_va_range_destroy(resource_range);
        }
        uvm_kvfree(user_channel->resource_ranges);
        user_channel->resource_ranges = NULL;
    }

    user_channel->gpu_va_space = NULL;
}

void uvm_user_channel_destroy_detached(uvm_user_channel_t *user_channel)
{
    // Check that this channel was already detached
    UVM_ASSERT(user_channel->gpu_va_space == NULL);

    list_del(&user_channel->list_node);

    if (user_channel->instance_ptr_desc)
        uvm_rm_locked_call_void(nvUvmInterfaceReleaseChannel(user_channel->instance_ptr_desc));

    uvm_kvfree(user_channel);
}

static NV_STATUS uvm_unregister_channel(uvm_va_space_t *va_space,
                                        uvm_rm_user_object_t *user_rm_channel)
{
    uvm_gpu_va_space_t *gpu_va_space;
    uvm_gpu_t *gpu = NULL;
    uvm_user_channel_t *user_channel = NULL;
    NV_STATUS status = NV_OK;
    LIST_HEAD(detached_channel);

    uvm_down_write(&va_space->lock);

    // This is a pretty naive search but it's unlikely to show up in a perf-
    // critical path. We could optimize it in the future with a table lookup
    // instead, if it becomes a problem.
    for_each_gpu_va_space(gpu_va_space, va_space) {
        user_channel = find_user_channel(gpu_va_space, user_rm_channel);
        if (user_channel)
            break;
    }

    if (user_channel) {
        gpu = user_channel->gpu_va_space->gpu;
        uvm_user_channel_detach(user_channel);
        list_add(&user_channel->list_node, &detached_channel);
        uvm_gpu_retain(gpu);
    }
    else {
        status = NV_ERR_INVALID_CHANNEL;
    }

    uvm_up_write(&va_space->lock);

    if (status == NV_OK) {
        uvm_gpu_destroy_detached_channels(gpu, &detached_channel);
        uvm_gpu_release(gpu);
    }

    return status;
}

NV_STATUS uvm_api_unregister_channel(UVM_UNREGISTER_CHANNEL_PARAMS *params, struct file *filp)
{
    uvm_va_space_t *va_space = uvm_va_space_get(filp);
    uvm_rm_user_object_t user_rm_channel =
    {
        .rm_control_fd = -1, // Not needed for a UVM-internal handle lookup
        .user_client   = params->hClient,
        .user_object   = params->hChannel
    };
    return uvm_unregister_channel(va_space, &user_rm_channel);
}
