/*******************************************************************************
    Copyright (c) 2016 NVIDIA Corporation

    Permission is hereby granted, free of charge, to any person obtaining a copy
    of this software and associated documentation files (the "Software"), to
    deal in the Software without restriction, including without limitation the
    rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
    sell copies of the Software, and to permit persons to whom the Software is
    furnished to do so, subject to the following conditions:

        The above copyright notice and this permission notice shall be
        included in all copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
    IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
    FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
    THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
    LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
    FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
    DEALINGS IN THE SOFTWARE.

*******************************************************************************/

#ifndef __UVM8_USER_CHANNEL_H__
#define __UVM8_USER_CHANNEL_H__

#include "uvm8_forward_decl.h"
#include "uvm8_va_space.h"
#include "uvm8_hal_types.h"

typedef struct
{
    // Parent GPU VA space
    uvm_gpu_va_space_t *gpu_va_space;

    // RM handles used to register this channel. We store them for UVM-internal
    // purposes to look up the uvm_user_channel_t for unregistration, rather
    // than introducing a new "UVM channel handle" object for user space.
    //
    // DO NOT TRUST THESE VALUES AFTER UVM_REGISTER_CHANNEL. They are passed by
    // user-space at channel registration time to validate the channel with RM,
    // but afterwards the user could free and reallocate either of the client or
    // object handles, so we can't pass them to RM trusting that they still
    // represent this channel.
    //
    // That's ok because we never pass these handles to RM again after
    // registration.
    uvm_rm_user_object_t user_rm_channel;

    // Array of all VA ranges associated with this channel. Global ranges (those
    // associated with multiple channels in the GPU VA space) are also included
    // here. Channel destroy will drop their ref counts.
    uvm_va_range_t **resource_ranges;
    size_t num_resource_ranges;

    // Physical instance pointer. There is a 1:1 mapping between instance
    // pointer and channel. GPU faults report an instance pointer, and the GPU
    // fault handler converts this instance pointer into the parent
    // uvm_va_space_t.
    uvm_gpu_phys_address_t instance_ptr;

    // Descriptor to pass to RM for this channel
    NvP64 instance_ptr_desc;

    // Hardware channel ID. This is only for debugging, since we aren't
    // preventing the user from freeing their channel, just the instance pointer
    // allocation.
    NvU32 hw_channel_id;

    // For most of the channel's life this is a node in the owning
    // gpu_va_space's registered_channels list. Once the channel is detached the
    // channel is moved onto a temporary list pending destruction.
    struct list_head list_node;
} uvm_user_channel_t;

// User-facing APIs (uvm_api_register_channel, uvm_api_unregister_channel) are
// declared in uvm8_api.h.

// First phase of user channel destroy.
//
// This detaches the channel from the parent gpu_va_space. The channel is
// stopped, forcibly if necessary, and all virtual mappings associated with it
// are torn down. The user_channel object and the instance pointer it contains
// are not destroyed. The caller must use uvm_user_channel_destroy_detached to
// do that.
//
// This two-phase approach allows the caller to drop the VA space lock and flush
// the fault buffer before removing the instance pointer.
// See uvm_gpu_destroy_detached_channels.
void uvm_user_channel_detach(uvm_user_channel_t *user_channel);

// Second phase of user channel destroy which frees the user_channel object and
// the corresponding instance pointer. The channel must have been detached
// first.
void uvm_user_channel_destroy_detached(uvm_user_channel_t *user_channel);

#endif // __UVM8_USER_CHANNEL_H__
