/*******************************************************************************
    Copyright (c) 2015 NVIDIA Corporation

    Permission is hereby granted, free of charge, to any person obtaining a copy
    of this software and associated documentation files (the "Software"), to
    deal in the Software without restriction, including without limitation the
    rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
    sell copies of the Software, and to permit persons to whom the Software is
    furnished to do so, subject to the following conditions:

        The above copyright notice and this permission notice shall be
        included in all copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
    IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
    FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
    THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
    LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
    FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
    DEALINGS IN THE SOFTWARE.

*******************************************************************************/

#include "uvm_linux.h"
#include "uvm_common.h"
#include "uvm8_va_space.h"
#include "uvm8_va_range.h"
#include "uvm8_va_block.h"
#include "uvm8_hal_types.h"
#include "uvm8_kvmalloc.h"
#include "uvm8_tools.h"
#include "uvm8_push.h"
#include "uvm8_hal.h"

static struct kmem_cache *g_uvm_va_block_cache __read_mostly;
static struct kmem_cache *g_uvm_va_block_gpu_state_cache __read_mostly;

NV_STATUS uvm_va_block_init(void)
{
    g_uvm_va_block_cache = NV_KMEM_CACHE_CREATE("uvm_va_block_t", uvm_va_block_t);
    if (!g_uvm_va_block_cache)
        return NV_ERR_NO_MEMORY;

    g_uvm_va_block_gpu_state_cache = NV_KMEM_CACHE_CREATE("uvm_va_block_gpu_state_t", uvm_va_block_gpu_state_t);
    if (!g_uvm_va_block_gpu_state_cache)
        return NV_ERR_NO_MEMORY;

    return NV_OK;
}

void uvm_va_block_exit(void)
{
    kmem_cache_destroy_safe(&g_uvm_va_block_gpu_state_cache);
    kmem_cache_destroy_safe(&g_uvm_va_block_cache);
}

static size_t block_num_cpu_pages(uvm_va_block_t *block)
{
    return uvm_va_block_size(block) / PAGE_SIZE;
}

static size_t block_cpu_page_index(uvm_va_block_t *block, NvU64 addr)
{
    UVM_ASSERT(addr >= block->start);
    UVM_ASSERT(addr <= block->end);
    return (addr - block->start) / PAGE_SIZE;
}

NV_STATUS uvm_va_block_create(uvm_va_range_t *va_range,
                              NvU64 start,
                              NvU64 end,
                              uvm_va_block_t **out_block)
{
    uvm_va_block_t *block = NULL;
    NvU64 size = end - start + 1;
    NV_STATUS status;

    UVM_ASSERT(PAGE_ALIGNED(start));
    UVM_ASSERT(PAGE_ALIGNED(end + 1));
    UVM_ASSERT(PAGE_ALIGNED(size));
    UVM_ASSERT(size > 0);
    UVM_ASSERT(size <= UVM_VA_BLOCK_SIZE);
    UVM_ASSERT(start >= va_range->node.start);
    UVM_ASSERT(end <= va_range->node.end);
    UVM_ASSERT(va_range->type == UVM_VA_RANGE_TYPE_MMAP);

    // Blocks can't span a block alignment boundary
    UVM_ASSERT(UVM_VA_BLOCK_ALIGN_DOWN(start) == UVM_VA_BLOCK_ALIGN_DOWN(end));

    block = kmem_cache_zalloc(g_uvm_va_block_cache, NV_UVM_GFP_FLAGS);
    if (!block) {
        status = NV_ERR_NO_MEMORY;
        goto error;
    }

    kref_init(&block->kref);
    uvm_mutex_init(&block->lock, UVM_LOCK_ORDER_VA_BLOCK);
    block->start = start;
    block->end = end;
    block->va_range = va_range;
    uvm_tracker_init(&block->tracker);

    block->cpu.pages = uvm_kvmalloc_zero((size / PAGE_SIZE) * sizeof(block->cpu.pages[0]));
    if (!block->cpu.pages) {
        status = NV_ERR_NO_MEMORY;
        goto error;
    }

    *out_block = block;
    return NV_OK;

error:
    uvm_va_block_release(block);
    return status;
}

// Retrieves the gpu_state for the given GPU, allocating it if it doesn't exist
static uvm_va_block_gpu_state_t *block_gpu_state_get_alloc(uvm_va_block_t *block, uvm_gpu_t *gpu)
{
    uvm_va_block_gpu_state_t *gpu_state = block->gpus[gpu->id - 1];

    uvm_assert_mutex_locked(&block->lock);

    if (!gpu_state) {
        gpu_state = kmem_cache_zalloc(g_uvm_va_block_gpu_state_cache, NV_UVM_GFP_FLAGS);
        if (!gpu_state)
            return NULL;

        block->gpus[gpu->id - 1] = gpu_state;
    }

    return gpu_state;
}

// Allocates the input page in the block, if it doesn't already exist
static NV_STATUS block_populate_page_cpu(uvm_va_block_t *block, size_t page_index)
{
    struct page *page;
    gfp_t gfp_flags;

    if (block->cpu.pages[page_index])
        return NV_OK;

    UVM_ASSERT(!test_bit(page_index, block->cpu.resident));

    // TODO: Only zero the page if it isn't yet resident on another processor
    gfp_flags = NV_UVM_GFP_FLAGS | GFP_HIGHUSER | __GFP_ZERO;

    page = alloc_pages(gfp_flags, 0);
    if (!page)
        return NV_ERR_NO_MEMORY;

    block->cpu.pages[page_index] = page;
    return NV_OK;
}


static NV_STATUS uvm_va_block_make_resident_cpu(uvm_va_block_t *va_block, NvU64 start, NvU64 end)
{
    NV_STATUS status;
    size_t page_index  = block_cpu_page_index(va_block, start);
    size_t outer_index = block_cpu_page_index(va_block, end) + 1;

    for_each_clear_bit_from(page_index, va_block->cpu.resident, outer_index) {
        status = block_populate_page_cpu(va_block, page_index);
        if (status != NV_OK)
            return status;

        // TODO: Migrate the data from the currently-resident page and clear
        //       the other processors' residency state

        uvm_processor_mask_set(&va_block->resident, UVM_CPU_ID);
        __set_bit(page_index, va_block->cpu.resident);
    }

    return NV_OK;
}

NV_STATUS uvm_va_block_make_resident(uvm_va_block_t *va_block, uvm_processor_id_t dest_id, NvU64 start, NvU64 end)
{
    uvm_assert_mutex_locked(&va_block->lock);
    UVM_ASSERT(va_block->va_range);
    UVM_ASSERT(va_block->va_range->type == UVM_VA_RANGE_TYPE_MMAP);
    UVM_ASSERT(PAGE_ALIGNED(start));
    UVM_ASSERT(PAGE_ALIGNED(end + 1));
    UVM_ASSERT(start >= va_block->start);
    UVM_ASSERT(end <= va_block->end);

    // TODO: Unmap mappings on other processors

    // TODO: Break read-duplication

    if (dest_id == UVM_CPU_ID)
        return uvm_va_block_make_resident_cpu(va_block, start, end);

    // TODO: We're only handling CPU pages right now
    UVM_ASSERT(0);
    return NV_ERR_INVALID_STATE;
}

// Looks up the current CPU mapping state of addr from the
// block->cpu.pte_bits bitmaps. If write access is enabled,
// UVM_PROT_READ_WRITE_ATOMIC is returned instead of UVM_PROT_READ_WRITE, since
// write access implies atomic access for CPUs.
static uvm_prot_t block_addr_prot_cpu(uvm_va_block_t *block, NvU64 addr)
{
    size_t page_index = block_cpu_page_index(block, addr);
    uvm_prot_t prot;

    UVM_ASSERT(block->va_range);
    UVM_ASSERT(block->va_range->type == UVM_VA_RANGE_TYPE_MMAP);
    uvm_assert_mutex_locked(&block->lock);

    if (test_bit(page_index, block->cpu.pte_bits[UVM_PTE_BITS_CPU_WRITE]))
        prot = UVM_PROT_READ_WRITE_ATOMIC;
    else if (test_bit(page_index, block->cpu.pte_bits[UVM_PTE_BITS_CPU_READ]))
        prot = UVM_PROT_READ_ONLY;
    else
        prot = UVM_PROT_NONE;

    // Sanity checks
    if (prot != UVM_PROT_NONE) {
        UVM_ASSERT(uvm_processor_mask_test(&block->mapped, UVM_CPU_ID));

        // The CPU can't map the GPU directly, so the CPU must have a coherent
        // copy resident if mapped.
        UVM_ASSERT(uvm_processor_mask_test(&block->resident, UVM_CPU_ID));
        UVM_ASSERT(test_bit(page_index, block->cpu.resident));
        UVM_ASSERT(block->cpu.pages[page_index]);

        if (prot == UVM_PROT_READ_WRITE_ATOMIC) {
            // If write is set, read should be too
            UVM_ASSERT(test_bit(page_index, block->cpu.pte_bits[UVM_PTE_BITS_CPU_READ]));

            // If the prot is writable, then the CPU's page must be the only
            // coherent copy resident anywhere in the system.
            //
            // TODO: Assert that this page is not resident on any GPU.
        }
    }

    return prot;
}

// Looks up the current GPU mapping state of addr from the
// block->gpus[i]->pte_bits bitmaps.
static uvm_prot_t block_addr_prot_gpu(uvm_va_block_t *block, uvm_gpu_t *gpu, NvU64 addr)
{
    size_t page_index = block_cpu_page_index(block, addr);
    uvm_va_block_gpu_state_t *gpu_state = block_gpu_state_get_alloc(block, gpu);
    uvm_prot_t prot;

    UVM_ASSERT(block->va_range);
    UVM_ASSERT(block->va_range->type == UVM_VA_RANGE_TYPE_MMAP);
    uvm_assert_mutex_locked(&block->lock);

    if (test_bit(page_index, gpu_state->pte_bits[UVM_PTE_BITS_GPU_ATOMIC]))
        prot = UVM_PROT_READ_WRITE_ATOMIC;
    else if (test_bit(page_index, gpu_state->pte_bits[UVM_PTE_BITS_GPU_WRITE]))
        prot = UVM_PROT_READ_WRITE;
    else if (test_bit(page_index, gpu_state->pte_bits[UVM_PTE_BITS_GPU_READ]))
        prot = UVM_PROT_READ_ONLY;
    else
        prot = UVM_PROT_NONE;

    // Sanity checks
    if (prot != UVM_PROT_NONE) {
        UVM_ASSERT(uvm_processor_mask_test(&block->mapped, gpu->id));

        UVM_ASSERT(uvm_processor_mask_get_count(&block->resident) > 0);

        // TODO: We only support sysmem mappings right now
        UVM_ASSERT(uvm_processor_mask_test(&block->resident, UVM_CPU_ID));
        UVM_ASSERT(test_bit(page_index, block->cpu.resident));
        UVM_ASSERT(block->cpu.pages[page_index]);

        if (prot == UVM_PROT_READ_WRITE_ATOMIC) {
            // Atomic implies write
            UVM_ASSERT(test_bit(page_index, gpu_state->pte_bits[UVM_PTE_BITS_GPU_WRITE]));

            // TODO: Assert that no other processor has a writable mapping.
            //       Exception 1: Both processors support native HW atomics and
            //                    the memory is resident on one of the
            //                    processors in that group.
            //       Exception 2: UVM-Lite
        }
        if (prot == UVM_PROT_READ_WRITE) {
            // Write implies read
            UVM_ASSERT(test_bit(page_index, gpu_state->pte_bits[UVM_PTE_BITS_GPU_READ]));

            // TODO: Assert that there is only one resident copy of this page in
            //       all processors, except if we're doing UVM-Lite.
        }
    }

    return prot;
}

// Returns true if the block has any valid CPU PTE mapping in the range given by
// [addr, addr + size).
//
// TODO: This might need to turn into a more comprehensive set of iterators,
//       such as for_each_valid_cpu_pte, for_each_writable_cpu_pte, etc.
static bool block_has_valid_mapping_cpu(uvm_va_block_t *block, NvU64 addr, size_t size)
{
    NvU64 valid_addr, outer_addr = addr + size;
    size_t addr_index, outer_index, valid_index;

    UVM_ASSERT(addr >= block->start);
    UVM_ASSERT(outer_addr - 1 <= block->end);

    // Early-out: check whether any address in this block has a CPU mapping
    if (!uvm_processor_mask_test(&block->mapped, UVM_CPU_ID)) {
        UVM_ASSERT(bitmap_empty(block->cpu.pte_bits[UVM_PTE_BITS_CPU_READ], PAGES_PER_UVM_VA_BLOCK));
        UVM_ASSERT(bitmap_empty(block->cpu.pte_bits[UVM_PTE_BITS_CPU_WRITE], PAGES_PER_UVM_VA_BLOCK));
        return false;
    }

    addr_index = block_cpu_page_index(block, addr);

    // block_cpu_page_index will assert when outer_addr exceeds block->end
    outer_index = (outer_addr - block->start) / PAGE_SIZE;

    // All valid mappings have at least read permissions so we only need to
    // inspect the read bits.
    valid_index = find_next_bit(block->cpu.pte_bits[UVM_PTE_BITS_CPU_READ], outer_index, addr_index);
    if (valid_index == outer_index)
        return false;

    valid_addr = block->start + valid_index * PAGE_SIZE;
    UVM_ASSERT(block_addr_prot_cpu(block, valid_addr) != UVM_PROT_NONE);
    return true;
}

// Returns true if the block has any valid PTE mapping for this GPU in the range
// given by [addr, addr + size).
static bool block_has_valid_mapping_gpu(uvm_va_block_t *block,
                                        uvm_gpu_t *gpu,
                                        NvU64 addr,
                                        size_t size)
{
    uvm_va_block_gpu_state_t *gpu_state = block->gpus[gpu->id - 1];
    NvU64 valid_addr, outer_addr = addr + size;
    size_t addr_index, outer_index, valid_index;

    UVM_ASSERT(addr >= block->start);
    UVM_ASSERT(outer_addr - 1 <= block->end);

    // Early-outs: check whether this GPU maps any address in this block
    if (!gpu_state)
        return false;

    if (!uvm_processor_mask_test(&block->mapped, gpu->id)) {
        UVM_ASSERT(bitmap_empty(gpu_state->pte_bits[UVM_PTE_BITS_GPU_READ], PAGES_PER_UVM_VA_BLOCK));
        UVM_ASSERT(bitmap_empty(gpu_state->pte_bits[UVM_PTE_BITS_GPU_WRITE], PAGES_PER_UVM_VA_BLOCK));
        UVM_ASSERT(bitmap_empty(gpu_state->pte_bits[UVM_PTE_BITS_GPU_ATOMIC], PAGES_PER_UVM_VA_BLOCK));
        return false;
    }

    addr_index = block_cpu_page_index(block, addr);
    outer_index = (outer_addr - block->start) / PAGE_SIZE;

    // All valid mappings have at least read permissions so we only need to
    // inspect the read bits.
    valid_index = find_next_bit(gpu_state->pte_bits[UVM_PTE_BITS_GPU_READ], outer_index, addr_index);
    if (valid_index == outer_index)
        return false;

    valid_addr = block->start + valid_index * PAGE_SIZE;
    UVM_ASSERT(block_addr_prot_gpu(block, gpu, valid_addr) != UVM_PROT_NONE);
    return true;
}

// See the comments on uvm_va_block_unmap
static void uvm_va_block_unmap_cpu(uvm_va_block_t *block, NvU64 start, NvU64 end)
{
    uvm_va_range_t *va_range = block->va_range;
    size_t start_index, num_pages;
    size_t size = end + 1 - start;
    uvm_pte_bits_cpu_t pte_bit;

    start_index = block_cpu_page_index(block, start);

    // Early-out if nothing in the range is mapped
    if (!block_has_valid_mapping_cpu(block, start, size))
        return;

    unmap_mapping_range(&va_range->va_space->mapping, start, size, 1);

    // Clear block PTE state
    num_pages = size / PAGE_SIZE;
    for (pte_bit = 0; pte_bit < UVM_PTE_BITS_CPU_MAX; pte_bit++)
        bitmap_clear(block->cpu.pte_bits[pte_bit], start_index, num_pages);

    // Check whether the block has any more mappings
    if (bitmap_empty(block->cpu.pte_bits[UVM_PTE_BITS_CPU_READ], PAGES_PER_UVM_VA_BLOCK)) {
        UVM_ASSERT(bitmap_empty(block->cpu.pte_bits[UVM_PTE_BITS_CPU_WRITE], PAGES_PER_UVM_VA_BLOCK));
        uvm_processor_mask_clear(&block->mapped, UVM_CPU_ID);
    }
}

// See the comments on uvm_va_block_unmap
static NV_STATUS uvm_va_block_unmap_gpu(uvm_va_block_t *block,
                                        uvm_gpu_t *gpu,
                                        NvU64 start,
                                        NvU64 end,
                                        uvm_tracker_t *out_tracker)
{
    uvm_va_block_gpu_state_t *gpu_state = block->gpus[gpu->id - 1];
    uvm_gpu_va_space_t *gpu_va_space;
    uvm_push_t push;
    NV_STATUS status;
    size_t size = end + 1 - start;
    size_t num_pages, num_ptes, pte_size;
    NvU64 empty_pte_val;
    uvm_gpu_phys_address_t pte_addr;
    uvm_membar_t tlb_membar;
    uvm_pte_bits_gpu_t pte_bit;

    // We have to check gpu_state before looking at any VA space state like our
    // gpu_va_space, because we could be on the eviction path where we don't
    // have a lock on that state. However, since remove_gpu_va_space walks each
    // block to unmap the GPU before destroying the gpu_va_space, we're
    // guaranteed that if this GPU has an active mapping, the gpu_va_space can't
    // go away while we're holding the block lock.
    if (!gpu_state)
        return NV_OK;

    if (!block_has_valid_mapping_gpu(block, gpu, start, size))
        return NV_OK;

    // TODO: Handle other PTE sizes. That may require splitting PTEs, which may
    //       in turn require PTE allocation.
    UVM_ASSERT(gpu_state->page_table_range_4k.table);

    // TODO: This asserts that the VA space lock is held, which will not be the
    //       case on the eviction path (see comment block above). We'll need a
    //       lookup which doesn't assert that the lock is held.
    gpu_va_space = uvm_gpu_va_space_get(block->va_range->va_space, gpu);

    status = uvm_push_begin_acquire(gpu->channel_manager,
                                    UVM_CHANNEL_TYPE_MEMOPS,
                                    &block->tracker,
                                    &push,
                                    "Unmapping [0x%llx, 0x%llx)",
                                    start,
                                    end);
    if (status != NV_OK)
        return status;

    // Just clear out all PTEs in the range for now.
    //
    // TODO: Count the number of valid PTEs in this range to strike a balance
    //       between clearing everything (better for densely-populated PDEs) and
    //       clearing individual entries (better for sparsely-popualted PDEs).

    // TODO: Use the batch API.

    num_ptes = size / UVM_PAGE_SIZE_4K;
    pte_size = uvm_mmu_pte_size(&gpu_va_space->page_tables, UVM_PAGE_SIZE_4K);
    empty_pte_val = gpu_va_space->page_tables.hal->empty_pte(UVM_PAGE_SIZE_4K);

    pte_addr = uvm_page_table_range_entry_address(&gpu_va_space->page_tables,
                                                  &gpu_state->page_table_range_4k,
                                                  (start - block->start) / UVM_PAGE_SIZE_4K);

    // We need a membar to make sure MMU sees the PTE writes prior to the TLB
    // invalidate. By default a MEMBAR_SYS will be pushed.
    if (pte_addr.aperture == UVM_APERTURE_VID)
        uvm_push_set_flag(&push, UVM_PUSH_FLAG_CE_NEXT_MEMBAR_GPU);

    gpu->ce_hal->memset_8(&push, uvm_gpu_address_from_phys(pte_addr), empty_pte_val, num_ptes * pte_size);

    gpu->host_hal->wait_for_idle(&push);

    // TODO: These are all sysmem mappings for now. In the future we'll need to
    //       inspect gpu_state to determine whether any of these mappings
    //       pointed to remote memory.
    tlb_membar = UVM_MEMBAR_SYS;

    // Invalidate all PTEs in the GPU.
    // Depth: We're only writing PTEs, so we don't need to invalidate PDEs.
    // Membar: This is a downgrade so we always need some sort of membar.
    gpu->host_hal->tlb_invalidate_all(&push,
                                      uvm_page_tree_pdb(&gpu_va_space->page_tables)->addr,
                                      gpu_va_space->page_tables.hal->page_table_depth(UVM_PAGE_SIZE_4K),
                                      tlb_membar);

    // Let uvm_push_end do its default MEMBAR_SYS so the tracker is
    // guaranteed to wait for these writes in case we have a failure later.
    uvm_push_end(&push);

    // Clear block PTE state
    num_pages = size / PAGE_SIZE;
    for (pte_bit = 0; pte_bit < UVM_PTE_BITS_GPU_MAX; pte_bit++)
        bitmap_clear(gpu_state->pte_bits[pte_bit], block_cpu_page_index(block, start), num_pages);

    // Check whether this GPU has any more mapping in this block
    if (bitmap_empty(gpu_state->pte_bits[UVM_PTE_BITS_GPU_READ], PAGES_PER_UVM_VA_BLOCK)) {
        UVM_ASSERT(bitmap_empty(gpu_state->pte_bits[UVM_PTE_BITS_GPU_WRITE], PAGES_PER_UVM_VA_BLOCK));
        UVM_ASSERT(bitmap_empty(gpu_state->pte_bits[UVM_PTE_BITS_GPU_ATOMIC], PAGES_PER_UVM_VA_BLOCK));
        uvm_processor_mask_clear(&block->mapped, gpu->id);
    }

    return uvm_tracker_add_push_safe(out_tracker, &push);
}

NV_STATUS uvm_va_block_unmap(uvm_va_block_t *va_block,
                             uvm_processor_id_t id,
                             NvU64 start,
                             NvU64 end,
                             uvm_tracker_t *out_tracker)
{
    uvm_va_range_t *va_range = va_block->va_range;

    UVM_ASSERT(va_range);
    UVM_ASSERT(va_range->type == UVM_VA_RANGE_TYPE_MMAP);
    UVM_ASSERT(start >= va_block->start);
    UVM_ASSERT(end <= va_block->end);
    UVM_ASSERT(PAGE_ALIGNED(start));
    UVM_ASSERT(PAGE_ALIGNED(end + 1));
    uvm_assert_mutex_locked(&va_block->lock);

    if (id == UVM_CPU_ID) {
       uvm_va_block_unmap_cpu(va_block, start, end);
       return NV_OK;
    }

    return uvm_va_block_unmap_gpu(va_block, uvm_gpu_get(id), start, end, out_tracker);
}

// This function essentially works as a wrapper around vm_insert_page (hence
// the similar function prototype). This is needed since vm_insert_page
// doesn't take permissions as input, but uses vma->vm_page_prot instead.
// Since we may have multiple VA blocks under one VMA which need to map
// with different permissions, we have to manually change vma->vm_page_prot for
// each call to vm_insert_page. Multiple fauts under one VMA in separate
// blocks can be serviced concurrently, so the VMA wrapper lock is used
// to protect access to vma->vm_page_prot.
static NV_STATUS uvm_cpu_insert_page(struct vm_area_struct *vma,
                                     NvU64 addr,
                                     struct page *page,
                                     uvm_prot_t new_prot)
{
    uvm_vma_wrapper_t *vma_wrapper;
    unsigned long target_flags;
    pgprot_t target_pgprot;
    int ret;

    UVM_ASSERT(vma);
    UVM_ASSERT(vma->vm_private_data);

    vma_wrapper = vma->vm_private_data;
    target_flags = vma->vm_flags;

    if (new_prot == UVM_PROT_READ_ONLY)
        target_flags &= ~VM_WRITE;

    target_pgprot = vm_get_page_prot(target_flags);

    // Take VMA wrapper lock to check vma->vm_page_prot
    uvm_down_read(&vma_wrapper->lock);

    // Take a write lock if we need to modify the VMA vm_page_prot
    // - vma->vm_page_prot creates writable PTEs but new prot is RO
    // - vma->vm_page_prot creates read-only PTEs but new_prot is RW
    if (pgprot_val(vma->vm_page_prot) != pgprot_val(target_pgprot)) {
        uvm_up_read(&vma_wrapper->lock);
        uvm_down_write(&vma_wrapper->lock);

        vma->vm_page_prot = target_pgprot;

        uvm_downgrade_write(&vma_wrapper->lock);
    }

    ret = vm_insert_page(vma, addr, page);
    uvm_up_read(&vma_wrapper->lock);
    if (ret) {
        UVM_ASSERT_MSG(ret == -ENOMEM, "ret: %d\n", ret);
        return errno_to_nv_status(ret);
    }

    return NV_OK;
}

// Creates or modifies a CPU mapping for the given virtual address and updates
// the block's mapping and pte_bits bitmaps to reflect the new mapping.
//
// new_prot must not be UVM_PROT_NONE. uvm_va_block_unmap should be used
// instead.
//
// If the existing mapping already is new_prot, the behavior depends on
// force_remap. If force_remap is false this is a no-op. If force_remap is true
// the existing CPU mapping is unmapped first and the identical mapping is re-
// inserted. This handles cases in which the block tracking goes stale because
// the kernel unmapped the CPU behind our back.
//
// It is the caller's responsibility to:
//  - Guarantee that vm_insert_page is safe to use (vma->vm_mm has a reference
//    and mmap_sem is held in at least read mode)
//  - Ensure that the CPU page exists in block->cpu.pages
//  - Manage the block's residency bitmap
//  - Decide whether force_remap is appropriate
//  - Ensure that the block hasn't been killed (block->va_range is present)
static NV_STATUS block_map_cpu_page(uvm_va_block_t *block,
                                    NvU64 addr,
                                    uvm_prot_t new_prot,
                                    bool force_remap)
{
    uvm_va_range_t *va_range = block->va_range;
    uvm_prot_t curr_prot;
    size_t page_index;
    struct vm_area_struct *vma;
    NV_STATUS status;

    UVM_ASSERT(va_range);
    UVM_ASSERT(va_range->type == UVM_VA_RANGE_TYPE_MMAP);
    UVM_ASSERT(addr >= block->start);
    UVM_ASSERT(addr <= block->end);
    UVM_ASSERT(PAGE_ALIGNED(addr));
    UVM_ASSERT(new_prot != UVM_PROT_NONE);
    UVM_ASSERT(new_prot < UVM_PROT_MAX);

    // It's possible here that current->mm != vma->vm_mm. That can happen for
    // example due to access_process_vm (ptrace) or get_user_pages from another
    // driver. In those cases the caller has taken care of ref counting
    // vma->vm_mm for us, so we can safely operate on the vma, but we can't use
    // uvm_va_range_vma_current.
    vma = uvm_va_range_vma(va_range);
    uvm_assert_mmap_sem_locked_read(&vma->vm_mm->mmap_sem);

    uvm_assert_mutex_locked(&block->lock);

    page_index = block_cpu_page_index(block, addr);
    curr_prot = block_addr_prot_cpu(block, addr);

    UVM_ASSERT(block->cpu.pages[page_index]);

    // For the CPU, write implies atomic
    if (new_prot == UVM_PROT_READ_WRITE)
        new_prot = UVM_PROT_READ_WRITE_ATOMIC;

    // Check for existing VMA permissions
    // They can be modified after the initial mmap by mprotect
    if (!(vma->vm_flags & VM_READ))
        return NV_ERR_INVALID_ACCESS_TYPE;
    if (new_prot == UVM_PROT_READ_WRITE_ATOMIC && !(vma->vm_flags & VM_WRITE))
        return NV_ERR_INVALID_ACCESS_TYPE;

    if (!force_remap && curr_prot == new_prot)
        return NV_OK;

    // This unmap handles both the force_remap case and a special case required
    // by vm_insert_page. vm_insert_page returns -EBUSY when there's already a
    // mapping present at fault_addr, so we have to unmap first anyway when
    // upgrading from RO -> RW or downgrading from RW -> RO.
    if (curr_prot != UVM_PROT_NONE)
        unmap_mapping_range(&va_range->va_space->mapping, addr, PAGE_SIZE, 1);

    // Don't map the CPU until prior copies and GPU PTE updates finish,
    // otherwise we might not stay coherent.
    status = uvm_tracker_wait(&block->tracker);
    if (status != NV_OK)
        return status;

    status = uvm_cpu_insert_page(vma, addr, block->cpu.pages[page_index], new_prot);
    if (status != NV_OK)
        return status;

    // Update block mapping state
    if (curr_prot != new_prot) {
        // Transitioning from Invalid -> RO or Invalid -> RW
        if (curr_prot == UVM_PROT_NONE) {
            uvm_processor_mask_set(&block->mapped, UVM_CPU_ID);
            __set_bit(page_index, block->cpu.pte_bits[UVM_PTE_BITS_CPU_READ]);
        }

        // Transitioning from Invalid -> RW or RO -> RW
        if (new_prot == UVM_PROT_READ_WRITE_ATOMIC)
            __set_bit(page_index, block->cpu.pte_bits[UVM_PTE_BITS_CPU_WRITE]);

        // On RW -> RO transition, clear UVM_PTE_BITS_CPU_WRITE bit
        if (curr_prot == UVM_PROT_READ_WRITE_ATOMIC && new_prot == UVM_PROT_READ_ONLY)
            __clear_bit(page_index, block->cpu.pte_bits[UVM_PTE_BITS_CPU_WRITE]);
    }

    return NV_OK;
}

// See the comments on uvm_va_block_map
static NV_STATUS uvm_va_block_map_cpu(uvm_va_block_t *va_block, NvU64 start, NvU64 end, uvm_prot_t new_prot)
{
    NvU64 page_addr;
    NV_STATUS status;
    size_t page_index  = block_cpu_page_index(va_block, start);
    size_t outer_index = block_cpu_page_index(va_block, end) + 1;

    // TODO: Check if optimizing the unmap_mapping_range calls within
    //       block_map_cpu_page by doing them once here is helpful

    // TODO: Could optimize this substantially for already-mapped regions by
    //       scanning the CPU bit vectors for pages without the intended
    //       permissions.

    for_each_set_bit_from(page_index, va_block->cpu.resident, outer_index) {
        page_addr = va_block->start + page_index * PAGE_SIZE;

        // Don't set force_remap because we'll take a CPU fault if we get this
        // wrong, and the CPU fault handler will take care of it.
        status = block_map_cpu_page(va_block, page_addr, new_prot, false);
        if (status != NV_OK)
            return status;
    }

    return NV_OK;
}

// Helper for uvm_va_block_map_gpu. Sets the gpu_state mapping bits on a
// transition for index from curr_prot -> new_prot. new_prot must not be
// UVM_PROT_NONE.
//
// TODO: Refactor this to handle CPUs, all prot types, and ranges of pages.
static void block_update_gpu_mapping_state(uvm_va_block_t *va_block,
                                           uvm_gpu_t *gpu,
                                           size_t page_index,
                                           uvm_prot_t curr_prot,
                                           uvm_prot_t new_prot)
{
    uvm_va_block_gpu_state_t *gpu_state = va_block->gpus[gpu->id - 1];

    UVM_ASSERT(gpu_state);
    UVM_ASSERT(new_prot != UVM_PROT_NONE);

    if (new_prot < curr_prot) {
        // Permissions downgrade

        // RWA -> {RO, RW}
        if (curr_prot == UVM_PROT_READ_WRITE_ATOMIC)
            __clear_bit(page_index, gpu_state->pte_bits[UVM_PTE_BITS_GPU_ATOMIC]);

        // RWA -> RO or RW -> RO
        if (new_prot == UVM_PROT_READ_ONLY)
            __clear_bit(page_index, gpu_state->pte_bits[UVM_PTE_BITS_GPU_WRITE]);
    }
    else if (new_prot > curr_prot) {
        // Permissions upgrade

        // Invalid -> {RO, RW, RWA}
        if (curr_prot == UVM_PROT_NONE) {
            uvm_processor_mask_set(&va_block->mapped, gpu->id);
            __set_bit(page_index, gpu_state->pte_bits[UVM_PTE_BITS_GPU_READ]);
        }

        // Invalid -> {RW, RWA}, RO -> {RW, RWA}
        if (curr_prot < UVM_PROT_READ_WRITE && new_prot >= UVM_PROT_READ_WRITE)
            __set_bit(page_index, gpu_state->pte_bits[UVM_PTE_BITS_GPU_WRITE]);

        // Invalid -> RWA, RO -> RWA, RW -> RWA
        if (new_prot == UVM_PROT_READ_WRITE_ATOMIC)
            __set_bit(page_index, gpu_state->pte_bits[UVM_PTE_BITS_GPU_ATOMIC]);
    }
}

// See the comments on uvm_va_block_map
//
// TODO: This function is big and needs to be refactored. That will happen as we
//       bring vidmem mappings online.
static NV_STATUS uvm_va_block_map_gpu(uvm_va_block_t *va_block,
                                      uvm_gpu_t *gpu,
                                      NvU64 start,
                                      NvU64 end,
                                      uvm_prot_t new_prot,
                                      uvm_tracker_t *out_tracker)
{
    uvm_va_block_gpu_state_t *gpu_state = block_gpu_state_get_alloc(va_block, gpu);
    uvm_gpu_va_space_t *gpu_va_space = uvm_gpu_va_space_get(va_block->va_range->va_space, gpu);
    uvm_push_t push;
    NV_STATUS status = NV_OK;
    size_t num_written_ptes = 0;
    size_t page_index;
    uvm_membar_t pte_membar = UVM_MEMBAR_GPU;
    uvm_membar_t tlb_membar = UVM_MEMBAR_NONE;

    UVM_ASSERT(gpu_va_space);

    if (!gpu_state)
        return NV_ERR_NO_MEMORY;

    if (!gpu_state->page_table_range_4k.table) {
        // TODO: Need a retry-with-eviction mechanism
        // TODO: This writes PDEs, so we'll need to use uvm_page_tree_get_entry
        //       instead for multiple PTE sizes
        status = uvm_page_tree_get_ptes(&gpu_va_space->page_tables,
                                        UVM_PAGE_SIZE_4K,
                                        va_block->start,
                                        uvm_va_block_size(va_block),
                                        UVM_PMM_ALLOC_FLAGS_NONE,
                                        &gpu_state->page_table_range_4k);
        if (status != NV_OK)
            return status;
    }

    // TODO: Scan bitmaps to optimize already-mapped regions
    for (page_index = block_cpu_page_index(va_block, start);
         page_index <= block_cpu_page_index(va_block, end);
         page_index++) {

        uvm_prot_t curr_prot;
        size_t i, ptes_per_page;
        NvU64 addr, pte_val;
        bool is_vol;
        uvm_gpu_phys_address_t pte_addr, page_addr;
        struct page *page;

        addr = va_block->start + page_index * PAGE_SIZE;
        curr_prot = block_addr_prot_gpu(va_block, gpu, addr);

        // Skip mapping if:
        // 1) The page is not resident
        // 2) The mapping is remote and this GPU is the preferred location
        // 3) The desired mapping already exists
        if (!test_bit(page_index, va_block->cpu.resident) ||
            va_block->va_range->preferred_location == gpu->id  ||
            new_prot == curr_prot)
            continue;

        // TODO: Use PTE batch API

        // Get the push only once we know we need to do something. Otherwise the
        // push will still acquire the tracker and do a semaphore release even
        // with no copies.
        if (num_written_ptes == 0) {
            status = uvm_push_begin_acquire(gpu->channel_manager,
                                            UVM_CHANNEL_TYPE_MEMOPS,
                                            &va_block->tracker,
                                            &push,
                                            "Mapping [0x%llx, 0x%llx] as %s",
                                            start,
                                            end,
                                            uvm_prot_string(new_prot));
            if (status != NV_OK)
                return status;
        }

        // TODO: Handle non-sysmem pages
        page = va_block->cpu.pages[page_index];
        UVM_ASSERT(page);

        page_addr.aperture = UVM_APERTURE_SYS;
        page_addr.address  = page_to_phys(page);

        // Allow L2 to cache only local memory
        is_vol = (page_addr.aperture != UVM_APERTURE_VID);

        // Handle PAGE_SIZE > GPU PTE size
        ptes_per_page = PAGE_SIZE / UVM_PAGE_SIZE_4K;
        for (i = 0; i < ptes_per_page; i++) {
            pte_val = gpu_va_space->page_tables.hal->make_pte(page_addr.aperture,
                                                              page_addr.address + i*UVM_PAGE_SIZE_4K,
                                                              new_prot,
                                                              is_vol,
                                                              UVM_PAGE_SIZE_4K);

            pte_addr = uvm_page_table_range_entry_address(&gpu_va_space->page_tables,
                                                          &gpu_state->page_table_range_4k,
                                                          page_index * ptes_per_page + i);

            // We'll need a membar of some sort to make sure MMU sees the PTE
            // writes prior to the TLB invalidate. If the PTEs are in GPU memory
            // then we just need a GPU-local membar. If they're in sysmem, then
            // we need a sysmembar.
            if (pte_addr.aperture != UVM_APERTURE_VID)
                pte_membar = UVM_MEMBAR_SYS;

            ++num_written_ptes;

            // The PTE writes never overlap with each other
            uvm_push_set_flag(&push, UVM_PUSH_FLAG_CE_NEXT_PIPELINED);

            // We'll push one membar for all memsets in this loop
            uvm_push_set_flag(&push, UVM_PUSH_FLAG_CE_NEXT_MEMBAR_NONE);

            gpu->ce_hal->memset_8(&push, uvm_gpu_address_from_phys(pte_addr), pte_val, sizeof(pte_val));
        }

        block_update_gpu_mapping_state(va_block, gpu, page_index, curr_prot, new_prot);

        if (new_prot < curr_prot) {
            // Permissions downgrades need membars on TLB invalidate. If the
            // aperture was pointing to remote memory we need a sysmembar.
            // Otherwise we can use a GPU-local membar unless an earlier
            // iteration needed a sysmembar.
            if (page_addr.aperture != UVM_APERTURE_VID)
                tlb_membar = UVM_MEMBAR_SYS;
            else if (tlb_membar == UVM_MEMBAR_NONE)
                tlb_membar = UVM_MEMBAR_GPU;
        }
    }

    if (num_written_ptes) {
        gpu->host_hal->wait_for_idle(&push);
        uvm_hal_membar(gpu, &push, pte_membar);

        // Invalidate all PTEs in the GPU. TODO: Use batch API for this.
        // Depth: We're only writing PTEs, so we don't need to invalidate PDEs.
        gpu->host_hal->tlb_invalidate_all(&push,
                                          uvm_page_tree_pdb(&gpu_va_space->page_tables)->addr,
                                          gpu_va_space->page_tables.hal->page_table_depth(UVM_PAGE_SIZE_4K),
                                          tlb_membar);

        // Let uvm_push_end do its default MEMBAR_SYS so the tracker is
        // guaranteed to wait for these writes in case we have a failure later.
        uvm_push_end(&push);

        status = uvm_tracker_add_push_safe(out_tracker, &push);
    }

    return status;
}

NV_STATUS uvm_va_block_map(uvm_va_block_t *va_block,
                           uvm_processor_id_t id,
                           NvU64 start,
                           NvU64 end,
                           uvm_prot_t new_prot,
                           uvm_tracker_t *out_tracker)
{
    uvm_va_range_t *va_range = va_block->va_range;
    uvm_gpu_t *gpu;

    uvm_assert_mutex_locked(&va_block->lock);
    UVM_ASSERT(start >= va_block->start);
    UVM_ASSERT(end  <= va_block->end);
    UVM_ASSERT(PAGE_ALIGNED(start));
    UVM_ASSERT(PAGE_ALIGNED(end + 1));
    UVM_ASSERT(new_prot != UVM_PROT_NONE);
    UVM_ASSERT(new_prot < UVM_PROT_MAX);

    // TODO: The eviction path doesn't have a lock on the VA space, so it can't
    //       simply check range state nor inspect the GPU VA space registration.
    //       Need to handle this another way. We might have to skip this step if
    //       we're on the eviction path, which means we'll need to know that
    //       here.

    if (id == UVM_CPU_ID) {
        va_range = va_block->va_range;
        UVM_ASSERT(va_range);
        if (uvm_va_range_vma_current(va_range))
            return uvm_va_block_map_cpu(va_block, start, end, new_prot);
        return NV_OK;
    }

    gpu = uvm_gpu_get(id);

    // Although this GPU UUID is registered in the VA space, it might not have a
    // GPU VA space registered.
    if (uvm_gpu_va_space_get(va_range->va_space, gpu))
        return uvm_va_block_map_gpu(va_block, gpu, start, end, new_prot, out_tracker);

    return NV_OK;
}

NV_STATUS uvm_va_block_map_mask(uvm_va_block_t *va_block,
                                uvm_processor_mask_t *map_mask,
                                NvU64 start,
                                NvU64 end,
                                uvm_prot_t new_prot)
{
    uvm_tracker_t local_tracker = UVM_TRACKER_INIT();
    NV_STATUS status = NV_OK;
    NV_STATUS tracker_status;
    uvm_processor_id_t id;

    for_each_id_in_mask(id, map_mask) {
        status = uvm_va_block_map(va_block, id, start, end, new_prot, &local_tracker);
        if (status != NV_OK)
            break;
    }

    // Regardless of error, add the successfully-pushed mapping operations into
    // the block's tracker. Note that we can't overwrite the tracker because we
    // aren't guaranteed that the map actually pushed anything (in which case it
    // would've acquired the block tracker first).
    tracker_status = uvm_tracker_add_tracker_safe(&va_block->tracker, &local_tracker);
    uvm_tracker_deinit(&local_tracker);

    return status == NV_OK ? tracker_status : status;
}

NV_STATUS uvm_va_block_unmap_mask(uvm_va_block_t *va_block,
                                  uvm_processor_mask_t *unmap_mask,
                                  NvU64 start,
                                  NvU64 end)
{
    uvm_tracker_t local_tracker = UVM_TRACKER_INIT();
    NV_STATUS status = NV_OK;
    NV_STATUS tracker_status;
    uvm_processor_id_t id;

    // Watch out, unmap_mask could change during iteration since it could be
    // va_block->mapped.
    for_each_id_in_mask(id, unmap_mask) {
        // Errors could either be a system-fatal error (ECC) or an allocation
        // retry due to PTE splitting. In either case we should stop after
        // hitting the first one.
        status = uvm_va_block_unmap(va_block, id, start, end, &local_tracker);
        if (status != NV_OK)
            break;
    }

    // See the comment in uvm_va_block_map_mask for adding to the tracker.
    tracker_status = uvm_tracker_add_tracker_safe(&va_block->tracker, &local_tracker);
    uvm_tracker_deinit(&local_tracker);

    return status == NV_OK ? tracker_status : status;
}

// Unmaps all GPU mappings under this block, frees the page tables, and frees
// all the GPU chunks. This simply drops the chunks on the floor, so the caller
// must take care of copying the data elsewhere if it needs to remain intact.
//
// This serializes on the block tracker since it must unmap page tables.
static void block_destroy_gpu_state(uvm_va_block_t *block, uvm_gpu_id_t id)
{
    uvm_va_block_gpu_state_t *gpu_state = block->gpus[id - 1];
    uvm_gpu_va_space_t *gpu_va_space;
    uvm_gpu_t *gpu;

    if (!gpu_state)
        return;

    // Unmap PTEs and free page tables
    gpu = uvm_gpu_get(id);
    gpu_va_space = uvm_gpu_va_space_get(block->va_range->va_space, gpu);
    if (gpu_va_space)
        uvm_va_block_remove_gpu_va_space(block, gpu_va_space);
    UVM_ASSERT(!uvm_processor_mask_test(&block->mapped, id));

    // TODO: Free GPU chunks
    UVM_ASSERT(!uvm_processor_mask_test(&block->resident, id));

    kmem_cache_free(g_uvm_va_block_gpu_state_cache, gpu_state);
}

void uvm_va_block_remove_gpu_va_space(uvm_va_block_t *va_block, uvm_gpu_va_space_t *gpu_va_space)
{
    uvm_gpu_t *gpu = gpu_va_space->gpu;
    uvm_va_block_gpu_state_t *gpu_state = va_block->gpus[gpu->id - 1];
    NV_STATUS status;

    uvm_tracker_t local_tracker = UVM_TRACKER_INIT();

    if (!gpu_state)
        return;

    uvm_assert_mutex_locked(&va_block->lock);

    // Unmapping the whole block won't cause a page table split, so this should
    // only fail if we have a system-fatal error.
    status = uvm_va_block_unmap(va_block, gpu->id, va_block->start, va_block->end, &local_tracker);
    if (status != NV_OK) {
        UVM_ASSERT(status == uvm_global_get_status());
        return; // Just leak
    }

    UVM_ASSERT(!uvm_processor_mask_test(&va_block->mapped, gpu->id));

    // The unmap must finish before we free the page tables
    status = uvm_tracker_wait_deinit(&local_tracker);
    if (status != NV_OK)
        return; // System-fatal error, just leak

    if (gpu_state->page_table_range_4k.table) {
        uvm_page_tree_put_ptes(&gpu_va_space->page_tables, &gpu_state->page_table_range_4k);
        memset(&gpu_state->page_table_range_4k, 0, sizeof(gpu_state->page_table_range_4k));
    }
}

void uvm_va_block_disable_peer(uvm_va_block_t *va_block, uvm_gpu_t *gpu0, uvm_gpu_t *gpu1)
{
    uvm_assert_mutex_locked(&va_block->lock);

    // TODO: Unmap all mappings in this block from gpu0 -> gpu1 and from gpu1
    //       -> gpu0.
}

void uvm_va_block_unregister_gpu(uvm_va_block_t *va_block, uvm_gpu_t *gpu)
{
    uvm_va_block_gpu_state_t *gpu_state = va_block->gpus[gpu->id - 1];

    if (!gpu_state)
        return;

    uvm_assert_mutex_locked(&va_block->lock);

    // The mappings should've already been torn down by GPU VA space unregister
    UVM_ASSERT(!uvm_processor_mask_test(&va_block->mapped, gpu->id));
    UVM_ASSERT(bitmap_empty(gpu_state->pte_bits[UVM_PTE_BITS_GPU_READ], PAGES_PER_UVM_VA_BLOCK));
    UVM_ASSERT(gpu_state->page_table_range_4k.table == NULL);

    // TODO: Migrate any remaining resident allocations on this GPU to sysmem
    //       (thus unmapping any still-mapped peers), since
    //       block_destroy_gpu_state will just drop them.

    block_destroy_gpu_state(va_block, gpu->id);
}

// Tears down everything within the block, but doesn't free the block itself.
// Note that when uvm_va_block_kill is called, this is called twice: once for
// the initial kill itself, then again when the block's ref count is eventually
// destroyed. block->va_range is used to track whether the block has already
// been killed.
static void block_kill(uvm_va_block_t *block)
{
    uvm_va_range_t *va_range = block->va_range;
    uvm_perf_event_data_t event_data;
    uvm_gpu_id_t id;
    NV_STATUS status;
    size_t i;

    if (!va_range)
        return;

    UVM_ASSERT(va_range->type == UVM_VA_RANGE_TYPE_MMAP);

    event_data.block_destroy.block = block;
    uvm_perf_event_notify(&va_range->va_space->perf_events, UVM_PERF_EVENT_BLOCK_DESTROY, &event_data);

    // Unmap all processors in parallel first. Unmapping the whole block won't
    // cause a page table split, so this should only fail if we have a system-
    // fatal error.
    status = uvm_va_block_unmap_mask(block, &block->mapped, block->start, block->end);
    UVM_ASSERT(status == uvm_global_get_status());

    UVM_ASSERT(uvm_processor_mask_empty(&block->mapped));

    // Free the GPU page tables and chunks
    for_each_gpu_id(id)
        block_destroy_gpu_state(block, id);

    // Wait for the GPU PTE unmaps before freeing CPU memory
    uvm_tracker_wait_deinit(&block->tracker);

    // Free CPU pages
    if (block->cpu.pages) {
        for (i = 0; i < block_num_cpu_pages(block); i++) {
            if (block->cpu.pages[i])
                __free_page(block->cpu.pages[i]);
            else
                UVM_ASSERT(!test_bit(i, block->cpu.resident));
        }
        uvm_kvfree(block->cpu.pages);
    }

    block->va_range = NULL;
}

// Called when the block's ref count drops to 0
void uvm_va_block_destroy(struct kref *kref)
{
    uvm_va_block_t *block = container_of(kref, uvm_va_block_t, kref);

    // Nobody else should have a reference when freeing
    uvm_assert_mutex_unlocked(&block->lock);

    uvm_mutex_lock(&block->lock);
    block_kill(block);
    uvm_mutex_unlock(&block->lock);
    kmem_cache_free(g_uvm_va_block_cache, block);
}

void uvm_va_block_kill(uvm_va_block_t *va_block)
{
    uvm_mutex_lock(&va_block->lock);
    block_kill(va_block);
    uvm_mutex_unlock(&va_block->lock);

    // May call block_kill again
    uvm_va_block_release(va_block);
}

static NV_STATUS uvm_va_block_split_preallocate_gpu(uvm_va_block_t *existing, uvm_va_block_t *new, uvm_gpu_id_t id)
{
    uvm_va_block_gpu_state_t *existing_gpu_state, *new_gpu_state;
    uvm_gpu_t *gpu;

    existing_gpu_state = existing->gpus[id - 1];
    if (!existing_gpu_state)
        return NV_OK;

    // TODO: This code can't be tested until range split tests are updated to
    //       make GPU accesses.
    UVM_ASSERT_MSG(0, "TODO: Splitting GPU state is not supported yet\n");
    return NV_ERR_NOT_SUPPORTED;

    gpu = uvm_gpu_get(id);

    new_gpu_state = block_gpu_state_get_alloc(new, gpu);
    if (!new_gpu_state)
        return NV_ERR_NO_MEMORY;

    // TODO: Allocate PTEs under new, possibly allocate smaller PTEs under
    //       existing

    // TODO: Allocate GPU memory for new, but we can't evict since that might
    //       try to take existing's lock

    return NV_OK;
}

// Pre-allocate everything on both existing and new which will be needed to
// handle a split. If this fails, existing must remain functionally unmodified.
static NV_STATUS uvm_va_block_split_preallocate(uvm_va_block_t *existing, uvm_va_block_t *new)
{
    uvm_gpu_id_t id;
    NV_STATUS status;

    // Blocks don't have any CPU state to pre-allocate

    for_each_gpu_id(id) {
        status = uvm_va_block_split_preallocate_gpu(existing, new, id);
        if (status != NV_OK)
            return status;
    }

    if (existing->va_range->inject_split_error) {
        existing->va_range->inject_split_error = false;
        return NV_ERR_NO_MEMORY;
    }

    return NV_OK;
}

// Re-calculate the block's top-level processor masks:
//   - block->mapped
//   - block->resident
//
// This is called on block split.
static void block_set_processor_masks(uvm_va_block_t *block)
{
    size_t num_pages = block_num_cpu_pages(block);
    uvm_va_block_gpu_state_t *gpu_state;
    uvm_gpu_id_t id;

    if (bitmap_empty(block->cpu.pte_bits[UVM_PTE_BITS_CPU_READ], num_pages)) {
        UVM_ASSERT(bitmap_empty(block->cpu.pte_bits[UVM_PTE_BITS_CPU_WRITE], num_pages));
        uvm_processor_mask_clear(&block->mapped, UVM_CPU_ID);
    }
    else {
        uvm_processor_mask_set(&block->mapped, UVM_CPU_ID);
    }

    if (bitmap_empty(block->cpu.resident, num_pages)) {
        UVM_ASSERT(!uvm_processor_mask_test(&block->mapped, UVM_CPU_ID));
        uvm_processor_mask_clear(&block->resident, UVM_CPU_ID);
    }
    else {
        uvm_processor_mask_set(&block->resident, UVM_CPU_ID);
    }

    for_each_gpu_id(id) {
        gpu_state = block->gpus[id - 1];
        if (!gpu_state)
            continue;

        // TODO: Update GPU mapping and residency bitmaps
    }
}

// Split the CPU state within the existing block. existing's start is correct
// but its end has not yet been adjusted.
static void uvm_va_block_split_cpu(uvm_va_block_t *existing, uvm_va_block_t *new)
{
    struct page **temp_pages;
    size_t existing_pages, new_pages = block_num_cpu_pages(new);
    uvm_pte_bits_cpu_t pte_bit;

    UVM_ASSERT(existing->va_range->type == UVM_VA_RANGE_TYPE_MMAP);
    UVM_ASSERT(existing->va_range->type == new->va_range->type);
    UVM_ASSERT(existing->start < new->start);
    UVM_ASSERT(existing->end == new->end);
    UVM_ASSERT(existing->cpu.pages);
    UVM_ASSERT(new->cpu.pages);

    // We don't have to unmap the CPU since its virtual -> physical mappings
    // don't change.

    UVM_ASSERT(PAGE_ALIGNED(new->start));
    UVM_ASSERT(PAGE_ALIGNED(existing->start));
    existing_pages = (new->start - existing->start) / PAGE_SIZE;

    // Move pages from the top of existing down into new
    memcpy(&new->cpu.pages[0],
           &existing->cpu.pages[existing_pages],
           block_num_cpu_pages(new) * sizeof(new->cpu.pages[0]));

    // Attempt to shrink existing's pages allocation. If the realloc fails, just
    // keep on using the old larger one.
    temp_pages = uvm_kvrealloc(existing->cpu.pages, existing_pages * sizeof(existing->cpu.pages[0]));
    if (temp_pages)
        existing->cpu.pages = temp_pages;

    // The new block is always in the upper region of existing, so shift the bit
    // vectors down.
    //
    // Note that bitmap_shift_right requires both dst and src to be the same
    // size. That's ok since we don't scale them by block size.
    bitmap_shift_right(new->cpu.resident, existing->cpu.resident, existing_pages, PAGES_PER_UVM_VA_BLOCK);
    bitmap_clear(existing->cpu.resident, existing_pages, new_pages);

    for (pte_bit = 0; pte_bit < UVM_PTE_BITS_CPU_MAX; pte_bit++) {
        bitmap_shift_right(new->cpu.pte_bits[pte_bit],
                           existing->cpu.pte_bits[pte_bit],
                           existing_pages,
                           PAGES_PER_UVM_VA_BLOCK);
        bitmap_clear(existing->cpu.pte_bits[pte_bit], existing_pages, new_pages);
    }
}

NV_STATUS uvm_va_block_split(uvm_va_block_t *existing_va_block,
                             NvU64 new_end,
                             uvm_va_block_t **new_va_block,
                             uvm_va_range_t *new_va_range)
{
    uvm_va_space_t *va_space;
    uvm_va_block_t *new_block = NULL;
    NV_STATUS status = NV_OK;

    va_space = new_va_range->va_space;
    UVM_ASSERT(existing_va_block->va_range);
    UVM_ASSERT(existing_va_block->va_range->va_space == va_space);

    // External range types can't be split
    UVM_ASSERT(existing_va_block->va_range->type == UVM_VA_RANGE_TYPE_MMAP);
    UVM_ASSERT(new_va_range->type == UVM_VA_RANGE_TYPE_MMAP);
    uvm_assert_rwsem_locked_write(&va_space->lock);

    UVM_ASSERT(new_end > existing_va_block->start);
    UVM_ASSERT(new_end < existing_va_block->end);
    UVM_ASSERT(PAGE_ALIGNED(new_end + 1));

    status = uvm_va_block_create(new_va_range, new_end + 1, existing_va_block->end, &new_block);
    if (status != NV_OK)
        return status;

    // We're protected from other splits and faults by the va_space lock being
    // held in write mode, but that doesn't stop the reverse mapping (eviction
    // path) from inspecting the existing block. Stop them by taking the block
    // lock. When a reverse mapping thread takes this lock afterwards, it will
    // have to re-inspect state and may see that it should use the newly-split
    // block instead.
    uvm_mutex_lock(&existing_va_block->lock);

    // As soon as we update existing's reverse mappings to point to the newly-
    // split block, the eviction path could try to operate on the new block.
    // Lock that out too until new is ready.
    //
    // Note that we usually shouldn't nest block locks, but it's ok here because
    // we just created new_block so no other thead could possibly take it out of
    // order with existing's lock.
    uvm_mutex_lock_no_tracking(&new_block->lock);

    // The split has to be transactional, meaning that if we fail, the existing
    // block must not be modified. Handle that by pre-allocating everything we
    // might need under both existing and new at the start so we only have a
    // single point of failure.
    status = uvm_va_block_split_preallocate(existing_va_block, new_block);
    if (status != NV_OK)
        goto out;

    // We'll potentially be freeing page tables, so we need to wait for any
    // outstanding work before we start
    status = uvm_tracker_wait(&existing_va_block->tracker);
    if (status != NV_OK)
        goto out;

    uvm_va_block_split_cpu(existing_va_block, new_block);

    // TODO: Split and move GPU page tables and pages within block

    // block_set_processor_masks calculates the new masks based on the size of
    // the block, which in the case of existing has not been updated yet. That's
    // ok because we cleared out the now-excess bits during the split.
    block_set_processor_masks(existing_va_block);
    block_set_processor_masks(new_block);

    // Update existing's state only once we're past all failure points
    existing_va_block->end = new_end;

out:
    uvm_mutex_unlock_no_tracking(&new_block->lock);
    uvm_mutex_unlock(&existing_va_block->lock);

    if (status != NV_OK)
        uvm_va_block_release(new_block);
    else if (new_va_block)
        *new_va_block = new_block;

    return status;
}

NV_STATUS uvm_va_block_cpu_fault(uvm_va_block_t *va_block, NvU64 fault_addr, bool is_write)
{
    uvm_va_range_t *va_range = va_block->va_range;
    NV_STATUS status = NV_OK;
    bool force_remap;
    uvm_prot_t page_prot = UVM_PROT_READ_WRITE_ATOMIC;
    uvm_perf_event_data_t event_data;

    UVM_ASSERT(va_range);
    uvm_assert_rwsem_locked_read(&va_range->va_space->lock);
    UVM_ASSERT(va_range->type == UVM_VA_RANGE_TYPE_MMAP);

    UVM_ASSERT(fault_addr >= va_block->start);
    UVM_ASSERT(fault_addr <= va_block->end);

    // notify event to tools/performance heuristics
    event_data.fault.proc_id = UVM_CPU_ID;
    event_data.fault.space = va_range->va_space;
    event_data.fault.cpu.fault_va = fault_addr;
    event_data.fault.cpu.is_write = is_write;

    uvm_perf_event_notify(&va_range->va_space->perf_events, UVM_PERF_EVENT_FAULT, &event_data);

    // CPU faults to non-migratable range groups are always fatal, even if the
    // page is already on the CPU.
    if (!uvm_range_group_migratable(va_range->range_group))
        return NV_ERR_INVALID_OPERATION;

    // If the VA range is marked with read duplication enabled and this is a read
    // access, map the range as read-only.
    // TODO: this should be moved to a heuristics function eventually
    if (va_range->read_duplication == UVM_READ_DUPLICATION_ENABLED && !is_write)
        page_prot = UVM_PROT_READ_ONLY;

    // We have to use vm_insert_page instead of handing the page to the kernel
    // and letting it insert the mapping, and we must do that while holding the
    // lock on this VA block. Otherwise there will be a window in which we think
    // we've mapped the page but the CPU mapping hasn't actually been created
    // yet. During that window a GPU fault event could arrive and claim
    // ownership of that VA, "unmapping" it. Then later the kernel would
    // eventually establish the mapping, and we'd end up with both CPU and GPU
    // thinking they each owned the page.
    //
    // This function must only be called when it's safe to call vm_insert_page.
    // That is, there's a reference held on the vma's vm_mm and vm_mm->mmap_sem
    // is held in at least read mode, but current->mm might not be vma->vm_mm.

    uvm_mutex_lock(&va_block->lock);

    // Move the page to the CPU
    status = uvm_va_block_make_resident(va_block, UVM_CPU_ID, fault_addr, fault_addr + PAGE_SIZE - 1);
    if (status != NV_OK)
        goto out;

    // TODO: The kernel can downgrade our CPU mappings at any time without
    //       notifying us, which means our PTE state could be stale. We handle
    //       this for now by always forcibly re-mapping the CPU PTE, but we
    //       could instead track if we get a bunch of unexpected faults, and
    //       only forcibly unmap once we've exceeded some threshold.
    //
    //       A CPU fault is unexpected if:
    //       curr_prot == RW || (!is_write && curr_prot == RO)
    //
    //       But, note that we could also get "unexpected" faults if multiple
    //       CPU threads fault on the same address simultaneously and race to
    //       create the mapping.
    force_remap = true;

    // Note: don't use uvm_va_block_map because that will check
    // uvm_va_range_vma_current to see if the current thread is allowed to call
    // vm_insert_page. current->mm might not be vma->vm_mm, but that's ok
    // because the kernel won't trigger a fault in vma->vm_mm without holding a
    // reference on it.
    status = block_map_cpu_page(va_block, fault_addr, page_prot, force_remap);
    if (status != NV_OK) {
        // No need to free page. It'll get freed on teardown of this block
        goto out;
    }

    // TODO: Map SetAccessedBy processors, except for UVM-Lite

out:
    uvm_mutex_unlock(&va_block->lock);
    return status;
}

NV_STATUS uvm_va_block_find(uvm_va_space_t *va_space, NvU64 addr, uvm_va_block_t **out_block)
{
    uvm_va_range_t *va_range;
    uvm_va_block_t *block;
    size_t index;

    va_range = uvm_va_range_find(va_space, addr);
    if (!va_range || va_range->type != UVM_VA_RANGE_TYPE_MMAP)
        return NV_ERR_INVALID_ADDRESS;

    index = uvm_va_range_block_index(va_range, addr);
    block = uvm_va_range_block(va_range, index);
    if (!block)
        return NV_ERR_OBJECT_NOT_FOUND;

    *out_block = block;
    return NV_OK;
}

NV_STATUS uvm_va_block_find_create(uvm_va_space_t *va_space, NvU64 addr, uvm_va_block_t **out_block)
{
    uvm_va_range_t *va_range;
    NvU64 index;

    va_range = uvm_va_range_find(va_space, addr);
    if (!va_range || va_range->type != UVM_VA_RANGE_TYPE_MMAP)
        return NV_ERR_INVALID_ADDRESS;

    index = uvm_va_range_block_index(va_range, addr);
    return uvm_va_range_block_create(va_range, index, out_block);
}

NV_STATUS uvm8_test_change_pte_mapping(UVM_TEST_CHANGE_PTE_MAPPING_PARAMS *params, struct file *filp)
{
    uvm_va_space_t *va_space = uvm_va_space_get(filp);
    uvm_va_block_t *block;
    NV_STATUS status = NV_OK;
    size_t page_index;
    NvU64 start = params->va;
    NvU64 end = params->va + PAGE_SIZE - 1;
    uvm_prot_t new_prot;
    uvm_gpu_t *gpu = NULL;
    uvm_processor_id_t id;
    uvm_tracker_t local_tracker;

    if (!PAGE_ALIGNED(start))
        return NV_ERR_INVALID_ADDRESS;

    switch (params->mapping) {
        case UVM_TEST_PTE_MAPPING_INVALID:
            new_prot = UVM_PROT_NONE;
            break;
        case UVM_TEST_PTE_MAPPING_READ_ONLY:
            new_prot = UVM_PROT_READ_ONLY;
            break;
        case UVM_TEST_PTE_MAPPING_READ_WRITE:
            new_prot = UVM_PROT_READ_WRITE;
            break;
        case UVM_TEST_PTE_MAPPING_READ_WRITE_ATOMIC:
            new_prot = UVM_PROT_READ_WRITE_ATOMIC;
            break;
        default:
            return NV_ERR_INVALID_ARGUMENT;
    }

    // mmap_sem isn't needed for invalidating CPU mappings, but it will be
    // needed for inserting them.
    uvm_down_read_mmap_sem(&current->mm->mmap_sem);
    uvm_down_read(&va_space->lock);

    if (uvm_uuid_is_cpu(&params->uuid)) {
        id = UVM_CPU_ID;
    }
    else {
        gpu = uvm_va_space_get_gpu_by_uuid_with_gpu_va_space(va_space, &params->uuid);
        if (!gpu) {
            status = NV_ERR_INVALID_DEVICE;
            goto out;
        }

        // Check if the GPU can access the VA
        if (!uvm_gpu_can_address(gpu, start)) {
            status = NV_ERR_OUT_OF_RANGE;
            goto out;
        }

        id = gpu->id;
    }

    status = uvm_va_block_find_create(va_space, start, &block);
    if (status != NV_OK)
        goto out;

    // Creating new CPU mappings requires that we hold a reference on
    // vma->vm_mm, but we can only do that if current->mm == vma->vm_mm.
    if (id == UVM_CPU_ID && !uvm_va_range_vma_current(block->va_range)) {
        status = NV_ERR_INVALID_STATE;
        goto out;
    }

    uvm_mutex_lock(&block->lock);

    if (new_prot == UVM_PROT_NONE) {
        status = uvm_va_block_unmap(block, id, start, end, &block->tracker);
    }
    else {
        // The page must be resident and accessible to map
        page_index = block_cpu_page_index(block, start);
        if (!test_bit(page_index, block->cpu.resident)) {
            status = NV_ERR_INVALID_OPERATION;
        }
        else {
            // Use force_remap if we're mapping the CPU
            if (id == UVM_CPU_ID)
                status = block_map_cpu_page(block, start, new_prot, true);
            else
                status = uvm_va_block_map(block, id, start, end, new_prot, &block->tracker);
        }
    }

    if (status == NV_OK)
        status = uvm_tracker_init_from(&local_tracker, &block->tracker);

    uvm_mutex_unlock(&block->lock);

    if (status == NV_OK)
        status = uvm_tracker_wait_deinit(&local_tracker);

out:
    uvm_up_read(&va_space->lock);
    uvm_up_read_mmap_sem(&current->mm->mmap_sem);
    return status;
}

NV_STATUS uvm8_test_va_block_info(UVM_TEST_VA_BLOCK_INFO_PARAMS *params, struct file *filp)
{
    uvm_va_space_t *va_space = uvm_va_space_get(filp);
    uvm_va_block_t *va_block;
    NV_STATUS status = NV_OK;

    BUILD_BUG_ON(UVM_TEST_VA_BLOCK_SIZE != UVM_VA_BLOCK_SIZE);

    uvm_down_read_mmap_sem(&current->mm->mmap_sem);
    uvm_down_read(&va_space->lock);

    status = uvm_va_block_find(va_space, params->lookup_address, &va_block);
    if (status != NV_OK)
        goto out;

    if (!uvm_va_range_vma_current(va_block->va_range)) {
        // We aren't in the same mm as the one which owns the vma
        status = NV_ERR_INVALID_STATE;
        goto out;
    }

    params->va_block_start = va_block->start;
    params->va_block_end   = va_block->end;

out:
    uvm_up_read(&va_space->lock);
    uvm_up_read_mmap_sem(&current->mm->mmap_sem);
    return status;
}
