/*******************************************************************************
    Copyright (c) 2015 NVIDIA Corporation

    Permission is hereby granted, free of charge, to any person obtaining a copy
    of this software and associated documentation files (the "Software"), to
    deal in the Software without restriction, including without limitation the
    rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
    sell copies of the Software, and to permit persons to whom the Software is
    furnished to do so, subject to the following conditions:

        The above copyright notice and this permission notice shall be
        included in all copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
    IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
    FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
    THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
    LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
    FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
    DEALINGS IN THE SOFTWARE.

*******************************************************************************/

#include "uvm8_va_space.h"
#include "uvm8_va_range.h"
#include "uvm8_lock.h"
#include "uvm8_global.h"
#include "uvm8_kvmalloc.h"
#include "uvm8_user_channel.h"
#include "uvm_common.h"
#include "nv_uvm_interface.h"

static void disable_peers(uvm_va_space_t *va_space, uvm_gpu_t *gpu0, uvm_gpu_t *gpu1);
static void remove_gpu_va_space(uvm_gpu_va_space_t *gpu_va_space, struct list_head *detached_channels);

static void init_tools_data(uvm_va_space_t *va_space)
{
    int i;

    for (i = 0; i < ARRAY_SIZE(va_space->counters); i++)
        INIT_LIST_HEAD(va_space->counters + i);
    for (i = 0; i < ARRAY_SIZE(va_space->queues); i++)
        INIT_LIST_HEAD(va_space->queues + i);   
}

NV_STATUS uvm_va_space_create(struct inode *inode, struct file *filp)
{
    NV_STATUS status;
    uvm_va_space_t *va_space;
    va_space = uvm_kvmalloc_zero(sizeof(*va_space));

    if (!va_space)
        return NV_ERR_NO_MEMORY;

    uvm_init_rwsem(&va_space->lock, UVM_LOCK_ORDER_VA_SPACE);
    uvm_range_tree_init(&va_space->va_range_tree);

    // By default all struct files on the same inode share the same
    // address_space structure (the inode's) across all processes. This means
    // unmap_mapping_range would unmap virtual mappings across all processes on
    // that inode.
    //
    // Since the UVM driver uses the mapping offset as the VA of the file's
    // process, we need to isolate the mappings to each process.
    address_space_init_once(&va_space->mapping);
    va_space->mapping.host = inode;

    // Some paths in the kernel, for example force_page_cache_readahead which
    // can be invoked from user-space via madvise MADV_WILLNEED and fadvise
    // POSIX_FADV_WILLNEED, check the function pointers within
    // file->f_mapping->a_ops for validity. However, those paths assume that a_ops
    // itself is always valid. Handle that by using the inode's a_ops pointer,
    // which is what f_mapping->a_ops would point to anyway if we weren't re-
    // assigning f_mapping.
    va_space->mapping.a_ops = inode->i_mapping->a_ops;

#if defined(NV_ADDRESS_SPACE_HAS_BACKING_DEV_INFO)
    va_space->mapping.backing_dev_info = inode->i_mapping->backing_dev_info;
#endif

    // Init to 0 since we rely on atomic_inc_return behavior to return 1 as the first ID
    atomic64_set(&va_space->range_group_id_counter, 0);

    INIT_RADIX_TREE(&va_space->range_groups, NV_UVM_GFP_FLAGS);

    bitmap_zero(va_space->enabled_peers, UVM8_MAX_UNIQUE_GPU_PAIRS);

    filp->private_data = va_space;
    filp->f_mapping = &va_space->mapping;

    init_tools_data(va_space);

    status = uvm_perf_init_va_space_events(va_space, &va_space->perf_events);
    if (status != NV_OK)
        goto fail;


    return NV_OK;

fail:
    uvm_kvfree(va_space);

    return status;
}

// This does *not* release the GPU
static void unregister_gpu(uvm_va_space_t *va_space, uvm_gpu_t *gpu, struct list_head *detached_channels)
{
    uvm_gpu_t *peer_gpu;
    uvm_va_range_t *va_range;

    // If this GPU has any peer-to-peer pair that was explicitly enabled, but
    // not explicitly disabled, disable it.
    for_each_gpu_in_mask(peer_gpu, &va_space->registered_gpus) {
        if (gpu != peer_gpu)
            disable_peers(va_space, gpu, peer_gpu);
    }

    // If a GPU VA Space was explicitly registered, but not explicitly unregistered,
    // unregister it. This also detaches and returns all channels under the GPU VA
    // space.
    remove_gpu_va_space(uvm_gpu_va_space_get(va_space, gpu), detached_channels);

    uvm_for_each_va_range(va_range, va_space)
        uvm_va_range_unregister_gpu(va_range, gpu);

    uvm_processor_mask_clear(&va_space->registered_gpus, gpu->id);
}

void uvm_va_space_destroy(struct file *filp)
{
    uvm_va_space_t *va_space = uvm_va_space_get(filp);
    uvm_gpu_va_space_t *gpu_va_space;
    uvm_va_range_t *va_range, *va_range_next;
    uvm_gpu_t *gpu;
    uvm_processor_mask_t retained_gpus;
    LIST_HEAD(detached_channels);

    // TODO: Bug 1739262. The global lock is required because unregister_gpu
    //       might disable peer access. Use ref counts for peer pairings instead
    //       so we don't have to hold the global lock while tearing down GPU
    //       mappings in this VA space.
    uvm_mutex_lock(&g_uvm_global.global_lock);

    // The bottom half GPU page fault handler(s) could still look up and use
    // this va_space via the GPU's instance_ptr_table. Lock them out while we
    // tear down. Once we're done, the bottom half will fail to find any
    // registered GPUs in the VA space, so those faults will be canceled.
    uvm_down_write(&va_space->lock);

    uvm_processor_mask_copy(&retained_gpus, &va_space->registered_gpus);

    // The teardown sequence is:
    // 1) Unregister all channels before unmapping anything. This kills the
    //    channels and prevents spurious MMU fault from being generated
    //    (bug 1722021).
    //
    // 2) Destroy all VA ranges. We do this before unregistering the GPUs for
    //    performance, since GPU unregister will walk all VA ranges in the VA
    //    space multiple times.
    //
    // 3) Unregister all GPUs in the VA space

    for_each_gpu_va_space(gpu_va_space, va_space)
        uvm_gpu_va_space_detach_all_channels(gpu_va_space, &detached_channels);

    uvm_for_each_va_range_safe(va_range, va_range_next, va_space) {
        // All channel ranges should've been destroyed by the channel unregister
        // above
        UVM_ASSERT(va_range->type != UVM_VA_RANGE_TYPE_CHANNEL);
        uvm_va_range_destroy(va_range);
    }

    uvm_range_group_radix_tree_destroy(&va_space->range_groups);

    // Note that this does not release the GPUs. We do that below.
    //
    // All channels should've been detached already above.
    for_each_va_space_gpu(gpu, va_space)
        unregister_gpu(va_space, gpu, NULL);

    uvm_perf_destroy_va_space_events(&va_space->perf_events);

    uvm_up_write(&va_space->lock);
    uvm_mutex_unlock(&g_uvm_global.global_lock);

    UVM_ASSERT(uvm_processor_mask_empty(&va_space->registered_gpus));
    UVM_ASSERT(uvm_processor_mask_empty(&va_space->registered_gpu_va_spaces));

    // The instance pointer mappings for this VA space have been removed so no
    // new bottom halves can get to this VA space, but there could still be
    // bottom halves running from before we removed the mapping. Rather than
    // ref-count the VA space, just wait for them to finish.
    flush_scheduled_work();

    // We have to destroy detached channels outside of the VA space lock. We
    // have the GPUs still retained from the unregister_gpu call above.
    uvm_gpu_destroy_detached_channels_mask(&retained_gpus, &detached_channels);
    uvm_gpu_release_mask(&retained_gpus);

    filp->private_data = NULL;
    filp->f_mapping = NULL;

    uvm_kvfree(va_space);
}

uvm_gpu_t *uvm_va_space_get_gpu_by_uuid(uvm_va_space_t *va_space, NvProcessorUuid *gpu_uuid)
{
    uvm_gpu_t *gpu;


    for_each_va_space_gpu(gpu, va_space) {
        if (uvm_processor_uuid_eq(&gpu->uuid, gpu_uuid))
            return gpu;
    }

    return NULL;
}

uvm_gpu_t *uvm_va_space_get_gpu_by_uuid_with_gpu_va_space(uvm_va_space_t *va_space,
                                                          NvProcessorUuid *gpu_uuid)
{
    uvm_gpu_t *gpu;

    gpu = uvm_va_space_get_gpu_by_uuid(va_space, gpu_uuid);
    if (!gpu || !uvm_processor_mask_test(&va_space->registered_gpu_va_spaces, gpu->id))
        return NULL;

    return gpu;
}

// Note that the "VA space" in the function name refers to a UVM per-process VA space.
// (This is different from a per-GPU VA space.)
NV_STATUS uvm_va_space_register_gpu(uvm_va_space_t *va_space, NvProcessorUuid *gpu_uuid)
{
    NV_STATUS status;
    uvm_gpu_t *gpu;

    uvm_mutex_lock(&g_uvm_global.global_lock);
    uvm_down_write(&va_space->lock);

    status = uvm_gpu_retain_by_uuid(gpu_uuid, &gpu);
    if (status != NV_OK)
        goto done;

    // Make sure the gpu hasn't been already registered in this va space
    if (uvm_processor_mask_test(&va_space->registered_gpus, gpu->id)) {
        status = NV_ERR_INVALID_DEVICE;
        uvm_gpu_release_locked(gpu);
        goto done;
    }

    uvm_processor_mask_set(&va_space->registered_gpus, gpu->id);

done:
    uvm_up_write(&va_space->lock);
    uvm_mutex_unlock(&g_uvm_global.global_lock);

    return status;
}

NV_STATUS uvm_va_space_unregister_gpu(uvm_va_space_t *va_space, NvProcessorUuid *gpu_uuid)
{
    NV_STATUS status = NV_OK;
    uvm_gpu_t *gpu;
    LIST_HEAD(detached_channels);

    // TODO: Bug 1739262. The global lock is required because unregister_gpu
    //       might disable peer access. Use ref counts for peer pairings instead
    //       so we don't have to hold the global lock while tearing down GPU
    //       mappings in this VA space.
    uvm_mutex_lock(&g_uvm_global.global_lock);
    uvm_down_write(&va_space->lock);

    gpu = uvm_va_space_get_gpu_by_uuid(va_space, gpu_uuid);
    if (!gpu) {
        status = NV_ERR_INVALID_DEVICE;
        goto done;
    }

    unregister_gpu(va_space, gpu, &detached_channels);

done:
    uvm_up_write(&va_space->lock);
    uvm_mutex_unlock(&g_uvm_global.global_lock);

    if (status == NV_OK) {
        uvm_gpu_destroy_detached_channels(gpu, &detached_channels);
        uvm_gpu_release(gpu);
    }

    return status;
}

static void disable_peers(uvm_va_space_t *va_space, uvm_gpu_t *gpu0, uvm_gpu_t *gpu1)
{
    NvU32 table_index = uvm_gpu_peer_table_index(gpu0->id, gpu1->id);
    uvm_va_range_t *va_range;

    if (!test_bit(table_index, va_space->enabled_peers))
        return;

    // Unmap all page tables in this VA space which have peer mappings between
    // these two GPUs.
    uvm_for_each_va_range(va_range, va_space)
        uvm_va_range_disable_peer(va_range, gpu0, gpu1);

    uvm_gpu_release_peer_access(gpu0, gpu1);

    __clear_bit(table_index, va_space->enabled_peers);
}

NV_STATUS uvm_va_space_enable_peer_access(uvm_va_space_t *va_space,
                                          NvProcessorUuid *gpu_uuid_1,
                                          NvProcessorUuid *gpu_uuid_2)
{
    NV_STATUS status = NV_OK;
    uvm_gpu_t *gpu[2] = { 0 };
    NvU32 table_index = 0;

    uvm_mutex_lock(&g_uvm_global.global_lock);
    uvm_down_write(&va_space->lock);

    gpu[0] = uvm_va_space_get_gpu_by_uuid(va_space, gpu_uuid_1);
    gpu[1] = uvm_va_space_get_gpu_by_uuid(va_space, gpu_uuid_2);

    if (!gpu[0] || !gpu[1]) {
        status = NV_ERR_INVALID_DEVICE;
        goto done;
    }

    if (gpu[0]->id == gpu[1]->id) {
        status = NV_ERR_INVALID_DEVICE;
        goto done;
    }

    table_index = uvm_gpu_peer_table_index(gpu[0]->id, gpu[1]->id);

    // If this peer to peer pair is already enabled for a given VA, error out.
    if (test_bit(table_index, va_space->enabled_peers)) {
        status = NV_ERR_INVALID_DEVICE;
        goto done;
    }

    status = uvm_gpu_retain_peer_access(gpu[0], gpu[1]);
    if (status != NV_OK)
        goto done;

    __set_bit(table_index, va_space->enabled_peers);

    // TODO: Tell the VA ranges that they can map these peers, if they need to

done:
    uvm_up_write(&va_space->lock);
    uvm_mutex_unlock(&g_uvm_global.global_lock);
    return status;
}

NV_STATUS uvm_va_space_disable_peer_access(uvm_va_space_t *va_space,
                                           NvProcessorUuid *gpu_uuid_1,
                                           NvProcessorUuid *gpu_uuid_2)
{
    NV_STATUS status = NV_OK;
    uvm_gpu_t *gpu[2] = { 0 };

    // TODO: Bug 1739262. The global lock is required because disable_peers
    //       might drop the peer table ref count to 0. Delay dropping the ref
    //       count until after disable_peers is done so we don't have to hold
    //       the global lock while tearing down GPU mappings in this VA space.
    uvm_mutex_lock(&g_uvm_global.global_lock);
    uvm_down_write(&va_space->lock);

    gpu[0] = uvm_va_space_get_gpu_by_uuid(va_space, gpu_uuid_1);
    gpu[1] = uvm_va_space_get_gpu_by_uuid(va_space, gpu_uuid_2);

    if (!gpu[0] || !gpu[1]) {
        status = NV_ERR_INVALID_DEVICE;
        goto done;
    }

    if (gpu[0]->id == gpu[1]->id) {
        status = NV_ERR_INVALID_DEVICE;
        goto done;
    }

    if (!uvm_va_space_peer_enabled(va_space, gpu[0], gpu[1])) {
        status = NV_ERR_INVALID_DEVICE;
        goto done;
    }

    disable_peers(va_space, gpu[0], gpu[1]);

done:
    uvm_up_write(&va_space->lock);
    uvm_mutex_unlock(&g_uvm_global.global_lock);
    return status;
}


static NV_STATUS add_gpu_va_space(uvm_gpu_t *gpu,
                                  uvm_va_space_t *va_space,
                                  uvm_rm_user_object_t *user_rm_va_space)
{
    NV_STATUS status = NV_OK;
    uvm_gpu_va_space_t *gpu_va_space;
    NvU32 big_page_size = 0;
    uvm_gpu_phys_address_t pdb_phys;
    NvU64 num_pdes;

    UVM_ASSERT(gpu);
    UVM_ASSERT(va_space);
    uvm_assert_rwsem_locked_write(&va_space->lock);

    gpu_va_space = &va_space->gpu_va_spaces[gpu->id - 1];
    memset(gpu_va_space, 0, sizeof(*gpu_va_space));

    gpu_va_space->va_space = va_space;
    gpu_va_space->gpu = gpu;
    INIT_LIST_HEAD(&gpu_va_space->registered_channels);
    INIT_LIST_HEAD(&gpu_va_space->local_channel_ranges);

    // TODO:- This interface needs to use controlFd to do validation checks.
    (void)user_rm_va_space->rm_control_fd;
    status = uvm_rm_locked_call(nvUvmInterfaceDupAddressSpace(g_uvm_global.rm_session_handle,
                                                              gpu->uuid.uuid,
                                                              user_rm_va_space->user_client,
                                                              user_rm_va_space->user_object,
                                                              &gpu_va_space->duped_gpu_va_space));
    if (status != NV_OK) {
        UVM_DBG_PRINT("failed to dup address space with error: %s, for GPU:%s \n",
                nvstatusToString(status), gpu->name);
        goto cleanup;
    }

    status = uvm_rm_locked_call(nvUvmInterfaceGetBigPageSize(gpu_va_space->duped_gpu_va_space, &big_page_size));
    if (status != NV_OK) {
        UVM_DBG_PRINT("failed to get a big_page_size with error: %s, for GPU:%s \n",
                nvstatusToString(status), gpu->name);
        goto cleanup;
    }

    UVM_ASSERT(big_page_size);

    // Set up this GPU's page tables
    UVM_ASSERT(gpu_va_space->page_tables.root == NULL);
    status = uvm_page_tree_init(gpu, big_page_size, UVM_APERTURE_VID, &gpu_va_space->page_tables);
    if (status != NV_OK) {
        UVM_ERR_PRINT("Initializing the page tree failed: %s, GPU %s\n", nvstatusToString(status), gpu->name);
        goto cleanup;
    }

    // Replace the existing PDB, if present, with the new one allocated by UVM.
    // This will fail if nvUvmInterfaceSetPageDirectory has already been called
    // on the RM VA space object, which prevents the user from registering twice
    // and corrupting our state.
    //
    // TODO: RM needs to preempt and disable channels during this operation. See
    //       bug 1733664.
    pdb_phys = uvm_page_tree_pdb(&gpu_va_space->page_tables)->addr;
    num_pdes = uvm_mmu_page_tree_entries(&gpu_va_space->page_tables, 0, UVM_PAGE_SIZE_AGNOSTIC);
    status = uvm_rm_locked_call(nvUvmInterfaceSetPageDirectory(gpu_va_space->duped_gpu_va_space,
                                                               pdb_phys.address,
                                                               num_pdes,
                                                               pdb_phys.aperture == UVM_APERTURE_VID));
    if (status != NV_OK) {
        UVM_DBG_PRINT("nvUvmInterfaceSetPageDirectory() failed: %s, GPU %s\n", nvstatusToString(status), gpu->name);
        goto cleanup;
    }

    uvm_processor_mask_set(&va_space->registered_gpu_va_spaces, gpu->id);

    // TODO: Tell the VA ranges that they can map this GPU, if they need to

    return NV_OK;

cleanup:
    // We can't have any channels in the GPU VA space yet
    remove_gpu_va_space(gpu_va_space, NULL);
    return status;
}

NV_STATUS uvm_va_space_register_gpu_va_space(uvm_va_space_t *va_space,
                                             uvm_rm_user_object_t *user_rm_va_space,
                                             NvProcessorUuid *gpu_uuid)
{
    NV_STATUS status = NV_OK;
    uvm_gpu_t *gpu = NULL;

    uvm_down_write(&va_space->lock);

    gpu = uvm_va_space_get_gpu_by_uuid(va_space, gpu_uuid);
    if (!gpu) {
        status = NV_ERR_INVALID_DEVICE;
        goto done;
    }

    // Error out if the GPU VA space is already registered.
    if (uvm_processor_mask_test(&va_space->registered_gpu_va_spaces, gpu->id)) {
        status = NV_ERR_INVALID_DEVICE;
        goto done;
    }

    status = add_gpu_va_space(gpu, va_space, user_rm_va_space);

done:
    uvm_up_write(&va_space->lock);
    return status;
}

static void remove_gpu_va_space(uvm_gpu_va_space_t *gpu_va_space, struct list_head *detached_channels)
{
    uvm_va_space_t *va_space;
    uvm_va_range_t *va_range;
    NV_STATUS status;

    if (!gpu_va_space)
        return;

    UVM_ASSERT(gpu_va_space->gpu);
    va_space = gpu_va_space->va_space;
    UVM_ASSERT(va_space);
    uvm_assert_rwsem_locked_write(&va_space->lock);

    // Detaching all channels will forcibly kill the channels and any work still
    // pending under them. This should be done before unmapping allocations so
    // that the user doesn't get a bunch of noisy spurious GPU MMU faults during
    // unclean teardown.
    uvm_gpu_va_space_detach_all_channels(gpu_va_space, detached_channels);

    // Removing all registered channels should've removed all local ranges
    // used by those channels.
    UVM_ASSERT(list_empty(&gpu_va_space->local_channel_ranges));

    // Unmap all page tables in this VA space on this GPU.
    uvm_for_each_va_range(va_range, va_space)
        uvm_va_range_remove_gpu_va_space(va_range, gpu_va_space);

    if (gpu_va_space->page_tables.root) {
        status = uvm_rm_locked_call(nvUvmInterfaceUnsetPageDirectory(gpu_va_space->duped_gpu_va_space));
        UVM_ASSERT_MSG(status == NV_OK, "nvUvmInterfaceUnsetPageDirectory() failed: %s, GPU %s\n",
                       nvstatusToString(status), gpu_va_space->gpu->name);
        uvm_page_tree_deinit(&gpu_va_space->page_tables);
    }

    if (gpu_va_space->duped_gpu_va_space)
        uvm_rm_locked_call_void(nvUvmInterfaceAddressSpaceDestroy(gpu_va_space->duped_gpu_va_space));

    uvm_processor_mask_clear(&va_space->registered_gpu_va_spaces, gpu_va_space->gpu->id);
}

NV_STATUS uvm_va_space_unregister_gpu_va_space(uvm_va_space_t *va_space, NvProcessorUuid *gpu_uuid)
{
    NV_STATUS status = NV_OK;
    uvm_gpu_t *gpu = NULL;
    LIST_HEAD(detached_channels);

    UVM_ASSERT(va_space);

    uvm_down_write(&va_space->lock);

    gpu = uvm_va_space_get_gpu_by_uuid_with_gpu_va_space(va_space, gpu_uuid);
    if (!gpu) {
        status = NV_ERR_INVALID_DEVICE;
        goto done;
    }

    // Retain the GPU so we can destroy channels after dropping va_space lock
    uvm_gpu_retain(gpu);

    remove_gpu_va_space(uvm_gpu_va_space_get(va_space, gpu), &detached_channels);

done:
    uvm_up_write(&va_space->lock);

    if (status == NV_OK) {
        uvm_gpu_destroy_detached_channels(gpu, &detached_channels);
        uvm_gpu_release(gpu);
    }

    return status;
}

void uvm_gpu_va_space_detach_all_channels(uvm_gpu_va_space_t *gpu_va_space, struct list_head *detached_channels)
{
    uvm_user_channel_t *user_channel, *next_channel;
    list_for_each_entry_safe(user_channel, next_channel, &gpu_va_space->registered_channels, list_node) {
        uvm_user_channel_detach(user_channel);
        list_add(&user_channel->list_node, detached_channels);
    }
}

bool uvm_va_space_peer_enabled(uvm_va_space_t *va_space, uvm_gpu_t *gpu1, uvm_gpu_t *gpu2)
{
    size_t table_index;

    UVM_ASSERT(uvm_processor_mask_test(&va_space->registered_gpus, gpu1->id));
    UVM_ASSERT(uvm_processor_mask_test(&va_space->registered_gpus, gpu2->id));

    table_index = uvm_gpu_peer_table_index(gpu1->id, gpu2->id);
    return !!test_bit(table_index, va_space->enabled_peers);
}
