/*******************************************************************************
    Copyright (c) 2013 NVIDIA Corporation

    Permission is hereby granted, free of charge, to any person obtaining a copy
    of this software and associated documentation files (the "Software"), to
    deal in the Software without restriction, including without limitation the
    rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
    sell copies of the Software, and to permit persons to whom the Software is
    furnished to do so, subject to the following conditions:

        The above copyright notice and this permission notice shall be
        included in all copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
    IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
    FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
    THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
    LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
    FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
    DEALINGS IN THE SOFTWARE.

*******************************************************************************/

/*
 * uvm_common.c
 *
 * This file contains code that is common to all variants of the (Linux) UVM
 * kernel module.
 *
 * Initially, UVM-Lite and Uvm-Next work is proceeding somewhat
 * independently, but the long- and short-term plan is for them to share the
 * same kernel module (this one). That's why, at least initially, you'll find
 * files for UVM-Lite, for Uvm-Next, and for common code. We'll refactor as
 * necessary, to achieve the right balance of sharing code and decoupling the
 * features.
 *
 * --jhubbard
 *
 */

#include "uvm_common.h"
#include "uvm_linux.h"

// TODO: Remove this when the GPU event stubs are no longer needed
#include "nv_uvm_interface.h"
#include "uvm_lite.h"

#include "uvm_channel_mgmt.h"

#include "uvm_full.h"
#include "uvm_full_device_mgmt.h"

#include "uvm8_init.h"
#include "uvm8_forward_decl.h"

static dev_t g_uvmBaseDev;
struct UvmOpsUvmEvents g_exportedUvmOps;

static char* uvm_driver_mode = "lite";

// UVM-Full behavior is activated through insmod argument: (default is "lite")
//      insmod nvidia-uvm.ko uvm_driver_mode="full"
//
// Similarly, the UVM-8 production driver can be activated with:
//      insmod nvidia-uvm.ko uvm_driver_mode="8"
module_param(uvm_driver_mode, charp, S_IRUGO);
MODULE_PARM_DESC(uvm_driver_mode,
                "Set the uvm kernel driver mode. Choices include: 8, full, lite.");

typedef enum
{
    UVM_DRIVER_MODE_LITE,
    UVM_DRIVER_MODE_FULL,
    UVM_DRIVER_MODE_8,
} UvmDriverMode;

static const char * uvm_driver_mode_to_string(UvmDriverMode uvmDriverMode)
{
    switch (uvmDriverMode)
    {
        case UVM_DRIVER_MODE_LITE:
            return "lite";

        case UVM_DRIVER_MODE_FULL:
            return "full";

        case UVM_DRIVER_MODE_8:
            return "8";
    }
    return "invalid";
}

static UvmDriverMode uvm_get_mode(void)
{
    static NvBool bUvmDriverModeChecked = NV_FALSE;
    static UvmDriverMode uvmDriverMode;

    if (!bUvmDriverModeChecked)
    {
        if (strcmp("8", uvm_driver_mode) == 0)
            uvmDriverMode = UVM_DRIVER_MODE_8;
        else if (strcmp("full", uvm_driver_mode) == 0)
            uvmDriverMode = UVM_DRIVER_MODE_FULL;
        else
            uvmDriverMode = UVM_DRIVER_MODE_LITE;

        bUvmDriverModeChecked = NV_TRUE;
    }

    return uvmDriverMode;
}

static NvBool uvm8_activated(void)
{
    return uvm_get_mode() == UVM_DRIVER_MODE_8;
}

static NvBool uvmfull_activated(void)
{
    return uvm_get_mode() == UVM_DRIVER_MODE_FULL;
}

static NvBool uvmlite_activated(void)
{
    return uvm_get_mode() == UVM_DRIVER_MODE_LITE;
}

NV_STATUS uvm_api_initialize(UVM_INITIALIZE_PARAMS *params, struct file *filp)
{
    params->rmStatus = NV_OK;
    return NV_OK;
}

// This function serves to 'stub' out functionality by being a No-Op and
// returning NV_OK early.
NV_STATUS uvm_api_stub(void *pParams, struct file *filp)
{
    return NV_OK;
}

// This function serves to identify functionality that isn't supported
// in this UVM driver, by returning NV_ERR_NOT_SUPPORTED early.
NV_STATUS uvm_api_unsupported(void *pParams, struct file *filp)
{
    return NV_ERR_NOT_SUPPORTED;
}

// TODO: This would be easier if RM allowed for multiple registrations, since we
//       could register UVM-Lite and UVM-Full separately (bug 1372835).
static NV_STATUS uvm_gpu_event_start_device(UvmGpuUuid *gpuUuidStruct)
{
    NV_STATUS nvStatus = NV_OK;

    if (uvmlite_activated())
        nvStatus = uvmlite_gpu_event_start_device(gpuUuidStruct);

    return nvStatus;
}

static NV_STATUS uvm_gpu_event_stop_device(UvmGpuUuid *gpuUuidStruct)
{
    NV_STATUS nvStatus = NV_OK;

    if (uvmlite_activated())
        nvStatus = uvmlite_gpu_event_stop_device(gpuUuidStruct);

    return nvStatus;
}

static NV_STATUS uvmSetupGpuProvider(void)
{
    NV_STATUS status = NV_OK;

    g_exportedUvmOps.startDevice = uvm_gpu_event_start_device;
    g_exportedUvmOps.stopDevice  = uvm_gpu_event_stop_device;
    if (uvmfull_activated())
        g_exportedUvmOps.isrTopHalf = uvmfull_isr_top_half;

    // call RM to exchange the function pointers.
    status = nvUvmInterfaceRegisterUvmCallbacks(&g_exportedUvmOps);
    return status;
}

static int uvm_not8_init(dev_t uvmBaseDev)
{
    int ret = 0;
    int uvmliteInitialized = 0;
    int uvmfullInitialized = 0;
    int registered = 0;
    int channelApiInitialized = 0;
    NV_STATUS status;

    status = uvm_initialize_channel_mgmt_api();
    if (NV_OK != status)
    {
        UVM_ERR_PRINT_NV_STATUS("Could not initialize channel mgmt api.\n",
                                status);
        // TODO: Need a helper function to convert from NV_STATUS to errno
        ret = -ENOMEM;
        goto error;
    }
    channelApiInitialized = 1;

    if (uvmfull_activated())
    {
        ret = uvmfull_init(uvmBaseDev);
        if (ret)
            goto error;
        uvmfullInitialized = 1;
    }
    else
    {
        ret = uvmlite_init(uvmBaseDev);
        if (ret)
            goto error;
        uvmliteInitialized = 1;
    }

    // Register with RM
    //
    // TODO: This should probably happen before calling uvmlite_init and
    //       uvmfull_init, so those routines can make RM calls as part of their
    //       setup. But doing that means we could get event callbacks before
    //       those init routines are called, which is scary. Ideally this could
    //       be solved by having RM handle multiple registrations (bug 1372835).
    status = uvmSetupGpuProvider();
    if (NV_OK != status)
    {
        UVM_ERR_PRINT("Cannot register ops with NVIDIA driver. UVM error: 0x%x\n",
                        status);
        // TODO: Need a helper function to convert from NV_STATUS to errno
        ret = -EINVAL;
        goto error;
    }
    registered = 1;

    // Query GPU list from RM
    //
    // TODO: This should be called within uvmlite_init but the RM calls aren't
    //       enabled until after the uvmSetupGpuProvider call above. This can be
    //       moved once multiple RM registrations are allowed (bug 1372835).
    //       Until then we need a dummy symbol.
    if(uvmliteInitialized && registered)
    {
        ret = uvmlite_setup_gpu_list();
        if (ret)
        {
            UVM_ERR_PRINT("uvm_lite_setup_gpu_list failed\n");
            goto error;
        }
    }

    return 0;

error:
    if (registered)
        nvUvmInterfaceDeRegisterUvmOps();

    if (channelApiInitialized)
        uvm_deinitialize_channel_mgmt_api();

    if (uvmliteInitialized)
        uvmlite_exit();

    if (uvmfullInitialized)
        uvmfull_destroy();

    return ret;
}

static int __init uvm_init(void)
{
    NvBool allocated_dev = NV_FALSE;

    // The various helper init routines will create their own minor devices, so
    // we only need to create space for them here.
    int ret = alloc_chrdev_region(&g_uvmBaseDev,
                              0,
                              NVIDIA_UVM_NUM_MINOR_DEVICES,
                              NVIDIA_UVM_DEVICE_NAME);
    if (ret != 0) {
        UVM_ERR_PRINT("alloc_chrdev_region failed: %d\n", ret);
        goto error;
    }
    allocated_dev = NV_TRUE;

    if (uvm8_activated())
        ret = uvm8_init(g_uvmBaseDev);
    else
        ret = uvm_not8_init(g_uvmBaseDev);

    if (ret != 0) {
        UVM_ERR_PRINT("uvm init failed: %d\n", ret);
        goto error;
    }

    pr_info("Loaded the UVM driver in %s mode, major device number %d\n",
            uvm_driver_mode_to_string(uvm_get_mode()), MAJOR(g_uvmBaseDev));

    if (uvm_enable_builtin_tests)
        pr_info("Built-in UVM tests are enabled. This is a security risk.\n");

    return 0;

error:
    if (allocated_dev)
        unregister_chrdev_region(g_uvmBaseDev, NVIDIA_UVM_NUM_MINOR_DEVICES);

    return ret;
}

static void uvm_not8_exit(void)
{
    // this will cleanup registered interfaces
    nvUvmInterfaceDeRegisterUvmOps();

    uvm_deinitialize_channel_mgmt_api();

    if (uvmfull_activated())
        uvmfull_destroy();
    else
        uvmlite_exit();
}

static void __exit uvm_exit(void)
{
    if (uvm8_activated())
        uvm8_exit();
    else
        uvm_not8_exit();

    unregister_chrdev_region(g_uvmBaseDev, NVIDIA_UVM_NUM_MINOR_DEVICES);

    pr_info("Unloaded the UVM driver in %s mode\n", uvm_driver_mode_to_string(uvm_get_mode()));
}

//
// Convert kernel errno codes to corresponding NV_STATUS
//
NV_STATUS errno_to_nv_status(int errnoCode)
{
    if (errnoCode < 0)
        errnoCode = -errnoCode;

    switch (errnoCode)
    {
        case 0:
            return NV_OK;

        case E2BIG:
        case EINVAL:
            return NV_ERR_INVALID_ARGUMENT;

        case EACCES:
            return NV_ERR_INVALID_ACCESS_TYPE;

        case EADDRINUSE:
        case EADDRNOTAVAIL:
            return NV_ERR_UVM_ADDRESS_IN_USE;

        case EFAULT:
            return NV_ERR_INVALID_ADDRESS;

        case EINTR:
        case EBUSY:
            return NV_ERR_BUSY_RETRY;

        case ENXIO:
        case ENODEV:
            return NV_ERR_MODULE_LOAD_FAILED;

        case ENOMEM:
            return NV_ERR_NO_MEMORY;

        case EPERM:
            return NV_ERR_INSUFFICIENT_PERMISSIONS;

        case ESRCH:
            return NV_ERR_PID_NOT_FOUND;

        case ETIMEDOUT:
            return NV_ERR_TIMEOUT;

        case EEXIST:
            return NV_ERR_IN_USE;

        case ENOSYS:
            return NV_ERR_NOT_SUPPORTED;

        case ENOENT:
            return NV_ERR_NO_VALID_PATH;

        default:
            return NV_ERR_GENERIC;
    };
}

// Returns POSITIVE errno
int nv_status_to_errno(NV_STATUS status)
{
    switch (status) {
        case NV_OK:
            return 0;

        case NV_ERR_BUSY_RETRY:
            return EBUSY;

        case NV_ERR_INSUFFICIENT_PERMISSIONS:
            return EPERM;

        case NV_ERR_GPU_UUID_NOT_FOUND:
            return ENODEV;

        case NV_ERR_INSUFFICIENT_RESOURCES:
        case NV_ERR_NO_MEMORY:
            return ENOMEM;

        case NV_ERR_INVALID_ACCESS_TYPE:
            return EACCES;

        case NV_ERR_INVALID_ADDRESS:
            return EFAULT;

        case NV_ERR_INVALID_ARGUMENT:
        case NV_ERR_INVALID_DEVICE:
        case NV_ERR_INVALID_PARAMETER:
        case NV_ERR_INVALID_REQUEST:
        case NV_ERR_INVALID_STATE:
            return EINVAL;

        case NV_ERR_NOT_SUPPORTED:
            return ENOSYS;

        case NV_ERR_MODULE_LOAD_FAILED:
            return ENXIO;

        case NV_ERR_OVERLAPPING_UVM_COMMIT:
        case NV_ERR_UVM_ADDRESS_IN_USE:
            return EADDRINUSE;

        case NV_ERR_PID_NOT_FOUND:
            return ESRCH;

        case NV_ERR_TIMEOUT:
        case NV_ERR_TIMEOUT_RETRY:
            return ETIMEDOUT;

        case NV_ERR_IN_USE:
            return EEXIST;

        case NV_ERR_NO_VALID_PATH:
            return ENOENT;

        default:
            UVM_ASSERT_MSG(0, "No errno conversion set up for NV_STATUS %s\n", nvstatusToString(status));
            return EINVAL;
    }
}

//
// This routine retrieves the process ID of current, but makes no attempt to
// refcount or lock the pid in place, because that capability is only available
// to GPL-licenses device drivers.
//
// TODO: use the GPL-protected routines if and when we are able to change over
// to a dual MIT/GPL license (bug 1483843).
//
unsigned uvm_get_stale_process_id(void)
{
    return (unsigned) current->tgid;
}

unsigned uvm_get_stale_thread_id(void)
{
    return (unsigned) current->pid;
}

//
// A simple security rule for allowing access to UVM user space memory: if you
// are the same user as the owner of the memory, or if you are root, then you
// are granted access. The idea is to allow debuggers and profilers to work, but
// without opening up any security holes.
//
NvBool uvm_user_id_security_check(uid_t euidTarget)
{
    return (NV_CURRENT_EUID() == euidTarget) ||
           (UVM_ROOT_UID == euidTarget);
}

// Locking: you must hold mm->mmap_sem write lock before calling this function
NV_STATUS uvm_map_page(struct vm_area_struct *vma, struct page *page,
                       unsigned long addr)
{
    if (vm_insert_page(vma, addr, page) == 0)
        return NV_OK;

    return NV_ERR_INVALID_ARGUMENT;
}

void on_uvm_assert(void)
{
    (void)NULL;
}

module_init(uvm_init);
module_exit(uvm_exit);

// This parameter allows a program in user mode to call the kernel tests
// defined in this module. This parameter should only be used for testing and
// must not be set to true otherwise since it breaks security when it is
// enabled. By default and for safety reasons this parameter is set to false.
int uvm_enable_builtin_tests = 0;
module_param(uvm_enable_builtin_tests , int, S_IRUGO|S_IWUSR);
MODULE_PARM_DESC(uvm_enable_builtin_tests,
                 "Enable the UVM built-in tests. (This is a security risk)");

MODULE_LICENSE("MIT");
MODULE_INFO(supported, "external");
