/*******************************************************************************
    Copyright (c) 2014-2015 NVIDIA Corporation

    Permission is hereby granted, free of charge, to any person obtaining a copy
    of this software and associated documentation files (the "Software"), to
    deal in the Software without restriction, including without limitation the
    rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
    sell copies of the Software, and to permit persons to whom the Software is
    furnished to do so, subject to the following conditions:

        The above copyright notice and this permission notice shall be
        included in all copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
    IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
    FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
    THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
    LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
    FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
    DEALINGS IN THE SOFTWARE.

*******************************************************************************/

#include "uvm_common.h"
#include "uvm_linux.h"
#include "uvm_full.h"
#include "uvm_full_api.h"
#include "nv_uvm_interface.h"
#include "uvm_full_device_mgmt.h"
#include "uvm_full_pagetbl_mgmt.h"
#include "uvm_full_test.h"
#include "uvm_full_ctx_mgmt.h"
#include "uvm_channel_mgmt.h"
#include "uvm_full_fault_handler.h"
#include "uvm_full_pa_mgmt.h"
#include "uvm_full_identity_map.h"
#include "uvm_full_perf.h"

static struct cdev g_uvmfull_cdev;

#if defined(NV_VM_OPERATIONS_STRUCT_HAS_FAULT)
static int _vma_fault(struct vm_area_struct *vma, struct vm_fault *vmf)
{
    NvU64 faultingAddress = (NvU64)vmf->virtual_address;
    struct page * page = NULL;
    int retval;

    retval = uvmfull_cpu_fault_handler(vma, faultingAddress, &page, vmf->flags);
    vmf->page = page;
    return retval;
}
#else
static struct page * _vma_fault_old_style(struct vm_area_struct *vma,
                                          unsigned long address, int *type)
{
    NvU64 faultingAddress = address;
    struct page * page = NULL;
    *type = uvmfull_cpu_fault_handler(vma, faultingAddress, &page,
                                        FAULT_FLAG_FROM_OLD_KERNEL);
    return page;
}
#endif

static void uvmfull_vma_open(struct vm_area_struct *vma)
{
    // TODO consider fork usecase
}

static void uvmfull_vma_close(struct vm_area_struct *vma)
{
    UvmVaspace *vaspace;
    UvmVatContext query;
    Uvm2MbVaDesc *vaDesc;
    NvU64 vaddr;
    NvU32 remPages;
    UvmVatL0Entry *parentLevel = NULL;
    NvU32 numPages = 0;
    UvmGlobalState *uvm = uvmfull_get_global();
    vaspace = (UvmVaspace *) (vma->vm_file->private_data);
    down_read(&uvm->lock);
    down_read(&vaspace->lock);

    vaddr = UVM_ALIGN_ADDR(vma->vm_start, PAGE_SIZE_4K);
    remPages = DIV_ROUND_UP((vma->vm_end - vma->vm_start), PAGE_SIZE_4K);

    // limit to PDE page size
    while (remPages)
    {
        query.accessType = UvmVatAccess_GenericReclaim;
        query.address = vaddr;
        query.vaspace = vaspace;
        query.private.level0Entry = NULL;

        // calculate number of pages that fall within this PDE
        numPages = GPU_PDE_MAX_PAGES - GET_PDE_PTE_OFFSET(vaddr, PAGE_SHIFT_4K);
        numPages = numPages > remPages ? remPages : numPages;

        vaDesc = uvmfull_vat_acquire_leaf(&vaspace->vat, &query);
        if (vaDesc)
        {
            if (uvmfull_reclaim_va_desc(vaDesc, &parentLevel, vaddr,
                                        numPages) == 0)
            {
                query.private.level0Entry = parentLevel;
            }

            uvmfull_vat_release_leaf(&vaspace->vat, &query);
        }
        // best effort cleanup in an incorrect state
        vaddr += (numPages*PAGE_SIZE_4K);
        remPages -= numPages;
    }

    up_read(&vaspace->lock);
    up_read(&uvm->lock);
}

static struct vm_operations_struct uvmfull_vma_ops =
{
    .open            = uvmfull_vma_open,
    .close           = uvmfull_vma_close,
#if defined(NV_VM_OPERATIONS_STRUCT_HAS_FAULT)
    .fault           = _vma_fault
#else
    .nopage          = _vma_fault_old_style
#endif
};

static int uvmfull_open(struct inode *inode, struct file *filp)
{
    int ret = 0;
    if (NV_OK != uvmfull_vaspace_init(inode, filp))
        ret = -ENOMEM;
    return ret;
}

static int uvmfull_close(struct inode *inode, struct file *filp)
{
    int ret = 0;

    if (NV_OK != uvmfull_vaspace_peermap_destroy(filp))
        ret = -EINVAL;

    if (NV_OK != uvmfull_vaspace_destroy(inode, filp))
        ret = -EINVAL;

    return ret;
}

static int uvmfull_mmap(struct file * filp, struct vm_area_struct * vma)
{
    int ret = 0;
    UvmVaspace *vaspace;
    NV_STATUS status;
    UvmGlobalState *uvm;
    UvmPteHintsParams pteHintsParams={0};

    uvm = uvmfull_get_global();
    vaspace = (UvmVaspace *)filp->private_data;
    down_read(&uvm->lock);
    down_read(&vaspace->lock);

    if ((vma->vm_flags & VM_WRITE) ||
            (vma->vm_flags & VM_MAYWRITE))
    {
        pteHintsParams.updateFlags |= UVM_GPU_VA_PTE_HINTS_FLAG_SET_PROT_WRITE;
        pteHintsParams.pte.va_prot_write = 1;
    }
    if ((vma->vm_flags & VM_READ) ||
            (vma->vm_flags & VM_MAYREAD))
    {
        pteHintsParams.updateFlags |= UVM_GPU_VA_PTE_HINTS_FLAG_SET_PROT_READ;
        pteHintsParams.pte.va_prot_read = 1;
    }

    pteHintsParams.updateFlags |= UVM_GPU_VA_PTE_HINTS_FLAG_SET_MANAGED_VA;
    pteHintsParams.pte.managedVa = 1;

    status = uvmfull_alloc_va_range(vaspace,
                                    vma->vm_start,
                                    (vma->vm_end - vma->vm_start),
                                    &pteHintsParams);
    if (status != NV_OK)
    {
        ret = -ENOMEM;
        goto done;
    }

    // Do not copy vma by default on fork().
    // User can override this in madvise with MADV_DOFORK flag.
    vma->vm_flags |= VM_DONTCOPY;
    // Prohibt mremap() that would expand the vma.
    vma->vm_flags |= VM_DONTEXPAND;
    vma->vm_ops = &uvmfull_vma_ops;

    UVM_DBG_PRINT_RL("vma 0x%p [0x%p, 0x%p) pgoff 0x%p\n",
                     vma, (void*)vma->vm_start, (void*)vma->vm_end,
                     (void*)vma->vm_pgoff);

done:
    vma->vm_flags |= VM_MIXEDMAP;
    up_read(&vaspace->lock);
    up_read(&uvm->lock);
    return ret;
}

static long
uvmfull_unlocked_ioctl(struct file *filp, unsigned int cmd, unsigned long arg)
{
    switch (cmd)
    {
        case UVM_DEINITIALIZE:
            UVM_DBG_PRINT("cmd: UVM_DEINITIALIZE\n");
            return 0;
        UVM_FULL_ROUTE_CMD_STACK(UVM_INITIALIZE,              uvm_api_initialize);
        UVM_FULL_ROUTE_CMD_STACK(UVM_IS_8_SUPPORTED,          uvmfull_is_8_supported);
        UVM_FULL_ROUTE_CMD_STACK(UVM_RESERVE_VA,              uvmfull_api_reserve_va_legacy);
        UVM_FULL_ROUTE_CMD_STACK(UVM_RELEASE_VA,              uvmfull_api_release_va_legacy);
        UVM_FULL_ROUTE_CMD_STACK(UVM_REGISTER_GPU_VASPACE,    uvmfull_api_register_gpu_vaspace);
        UVM_FULL_ROUTE_CMD_STACK(UVM_UNREGISTER_GPU_VASPACE,  uvmfull_api_unregister_gpu_vaspace);
        UVM_FULL_ROUTE_CMD_STACK(UVM_MAP_EXTERNAL_ALLOCATION, uvmfull_api_map_rm_allocation);
        UVM_FULL_ROUTE_CMD_STACK(UVM_REGISTER_CHANNEL,        uvmfull_api_register_channel);
        UVM_FULL_ROUTE_CMD_STACK(UVM_UNREGISTER_CHANNEL,      uvmfull_api_unregister_channel);
        UVM_FULL_ROUTE_CMD_STACK(UVM_ENABLE_PEER_ACCESS,      uvmfull_api_enable_peer_access);
        UVM_FULL_ROUTE_CMD_STACK(UVM_DISABLE_PEER_ACCESS,     uvmfull_api_disable_peer_access);

        // Events API
        UVM_FULL_ROUTE_CMD_STACK(UVM_ADD_SESSION,             uvmfull_api_add_session);
        UVM_FULL_ROUTE_CMD_STACK(UVM_REMOVE_SESSION,          uvmfull_api_remove_session);
        UVM_FULL_ROUTE_CMD_STACK(UVM_MAP_COUNTER,             uvmfull_api_map_counter);
        UVM_FULL_ROUTE_CMD_STACK(UVM_ENABLE_COUNTERS,         uvmfull_api_enable_counters);
        UVM_FULL_ROUTE_CMD_STACK(UVM_CREATE_EVENT_QUEUE,      uvmfull_api_create_event_queue);
        UVM_FULL_ROUTE_CMD_STACK(UVM_MAP_EVENT_QUEUE,         uvmfull_api_map_event_queue);
        UVM_FULL_ROUTE_CMD_STACK(UVM_REMOVE_EVENT_QUEUE,      uvmfull_api_remove_event_queue);
        UVM_FULL_ROUTE_CMD_STACK(UVM_EVENT_CTRL,              uvmfull_api_event_ctrl);
        UVM_FULL_ROUTE_CMD_STACK(UVM_DEBUG_ACCESS_MEMORY,     uvmfull_api_debug_access_memory);

        // Skip APIs that are only defined for UVM8 (and not defined for UVM-Full) and return NV_OK
        UVM_FULL_ROUTE_CMD_STACK(UVM_REGISTER_GPU,                      uvm_api_stub);
        UVM_FULL_ROUTE_CMD_STACK(UVM_UNREGISTER_GPU,                    uvm_api_stub);
        UVM_FULL_ROUTE_CMD_STACK(UVM_ALLOC,                             uvm_api_stub);
        UVM_FULL_ROUTE_CMD_STACK(UVM_CREATE_RANGE_GROUP,                uvm_api_stub);
        UVM_FULL_ROUTE_CMD_STACK(UVM_DESTROY_RANGE_GROUP,               uvm_api_stub);
        UVM_FULL_ROUTE_CMD_STACK(UVM_FREE,                              uvm_api_stub);
        UVM_FULL_ROUTE_CMD_STACK(UVM_GET_GPU_UUID_TABLE,                uvm_api_stub);
        UVM_FULL_ROUTE_CMD_STACK(UVM_MEM_MAP,                           uvm_api_stub);
        UVM_FULL_ROUTE_CMD_STACK(UVM_MIGRATE_TO_GPU,                    uvm_api_stub);
        UVM_FULL_ROUTE_CMD_STACK(UVM_REGION_COMMIT,                     uvm_api_stub);
        UVM_FULL_ROUTE_CMD_STACK(UVM_REGION_DECOMMIT,                   uvm_api_stub);
        UVM_FULL_ROUTE_CMD_STACK(UVM_REGION_SET_STREAM,                 uvm_api_stub);
        UVM_FULL_ROUTE_CMD_STACK(UVM_REGISTER_MPS_CLIENT,               uvm_api_stub);
        UVM_FULL_ROUTE_CMD_STACK(UVM_REGISTER_MPS_SERVER,               uvm_api_stub);
        UVM_FULL_ROUTE_CMD_STACK(UVM_SET_RANGE_GROUP,                   uvm_api_stub);
        UVM_FULL_ROUTE_CMD_STACK(UVM_SET_STREAM_RUNNING,                uvm_api_stub);
        UVM_FULL_ROUTE_CMD_STACK(UVM_SET_STREAM_STOPPED,                uvm_api_stub);
        UVM_FULL_ROUTE_CMD_STACK(UVM_PAGEABLE_MEM_ACCESS,               uvm_api_stub);
        UVM_FULL_ROUTE_CMD_STACK(UVM_PREVENT_MIGRATION_RANGE_GROUPS,    uvm_api_stub);
        UVM_FULL_ROUTE_CMD_STACK(UVM_ALLOW_MIGRATION_RANGE_GROUPS,      uvm_api_stub);

        // Performance heuristics API
        UVM_FULL_ROUTE_CMD_ALLOC(UVM_PERF_GET_THRASHING_INFO, uvmfull_perf_get_thrashing_info);

        // Return NV_ERR_NOT_SUPPORTED for UVM8 functionality that UVM-Full does not support
        UVM_FULL_ROUTE_CMD_STACK(UVM_SET_PREFERRED_LOCATION,          uvm_api_unsupported);
        UVM_FULL_ROUTE_CMD_STACK(UVM_UNSET_PREFERRED_LOCATION,        uvm_api_unsupported);
        UVM_FULL_ROUTE_CMD_STACK(UVM_ENABLE_READ_DUPLICATION,         uvm_api_unsupported);
        UVM_FULL_ROUTE_CMD_STACK(UVM_DISABLE_READ_DUPLICATION,        uvm_api_unsupported);
        UVM_FULL_ROUTE_CMD_STACK(UVM_SET_ACCESSED_BY,                 uvm_api_unsupported);
        UVM_FULL_ROUTE_CMD_STACK(UVM_UNSET_ACCESSED_BY,               uvm_api_unsupported);
    }

    return uvmfull_test_ioctl(filp, cmd, arg);
}

static const struct file_operations uvmfull_fops =
{
    .open            = uvmfull_open,
    .release         = uvmfull_close,
    .mmap            = uvmfull_mmap,
    .unlocked_ioctl  = uvmfull_unlocked_ioctl,
#if NVCPU_IS_X86_64 && defined(NV_FILE_OPERATIONS_HAS_COMPAT_IOCTL)
    .compat_ioctl    = uvmfull_unlocked_ioctl,
#endif
    .owner           = THIS_MODULE,
};

int uvmfull_init(dev_t uvmBaseDev)
{
    NV_STATUS status;
    dev_t uvmfullDev;
    int ret = 0;

    status = uvmfull_ctx_mgmt_slab_init();
    UVM_PANIC_ON(status != NV_OK);

    status = uvmfull_pa_mgmt_slab_init();
    UVM_PANIC_ON(status != NV_OK);

    status = uvmfull_pagetbl_mgmt_slab_init();
    UVM_PANIC_ON(status != NV_OK);

    status = uvmfull_device_mgmt_slab_init();
    UVM_PANIC_ON(status != NV_OK);

    status = uvmfull_init_fault_cancel_cache();
    UVM_PANIC_ON(status != NV_OK);

    status = uvmfull_va_trie_slab_init();
    UVM_PANIC_ON(status != NV_OK);

    status = uvm_initialize_events_api();
    UVM_PANIC_ON(status != NV_OK);

    status = uvm_initialize_counters_api();
    UVM_PANIC_ON(status != NV_OK);

    status = uvmfull_identity_map_cache_init();
    UVM_PANIC_ON(status != NV_OK);

    status = uvm_perf_init();
    UVM_PANIC_ON(status != NV_OK);

    // Now initialize the device
    uvmfull_init_global_state();

    uvmfullDev = MKDEV(MAJOR(uvmBaseDev),
                               NVIDIA_UVM_PRIMARY_MINOR_NUMBER);

    cdev_init(&g_uvmfull_cdev, &uvmfull_fops);
    g_uvmfull_cdev.owner = THIS_MODULE;

    ret = cdev_add(&g_uvmfull_cdev, uvmfullDev, 1);
    if (ret)
    {
        UVM_ERR_PRINT("cdev_add (major %u, minor %u) failed: %d\n",
                      MAJOR(uvmfullDev), MINOR(uvmfullDev), ret);
    }
    return ret;
}

int uvmfull_destroy(void)
{
    UvmGlobalState *uvm = uvmfull_get_global();
    cdev_del(&g_uvmfull_cdev);
    // TODO: channel mgmt destroy needs to be re-added to this function.
    flush_scheduled_work();
    UVM_PANIC_ON(!list_empty(&uvm->uvmMirrors));
    UVM_PANIC_ON(!list_empty(&uvm->uvmVaspaces));

    uvm_perf_exit();

    uvm_deinitialize_counters_api();
    uvm_deinitialize_events_api();

    uvmfull_va_trie_slab_destroy();
    uvmfull_device_mgmt_slab_destroy();

    uvmfull_pagetbl_mgmt_slab_destroy();
    uvmfull_pa_mgmt_slab_destroy();
    uvmfull_ctx_mgmt_slab_destroy();

    uvmfull_destroy_fault_cancel_cache();
    uvmfull_identity_map_cache_destroy();
    return 0;
}
