/*******************************************************************************
    Copyright (c) 2014-2015 NVIDIA Corporation

    Permission is hereby granted, free of charge, to any person obtaining a copy
    of this software and associated documentation files (the "Software"), to
    deal in the Software without restriction, including without limitation the
    rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
    sell copies of the Software, and to permit persons to whom the Software is
    furnished to do so, subject to the following conditions:

        The above copyright notice and this permission notice shall be
        included in all copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
    IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
    FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
    THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
    LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
    FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
    DEALINGS IN THE SOFTWARE.

*******************************************************************************/


//
// uvm_full.h
//
// This file contains data-structure definitions for UVM-Pascal
// (UVM with GPU faulting) code.
//

#ifndef _UVM_FULL_H_
#define _UVM_FULL_H_

#include "uvmtypes.h"
#include "uvm_full_types.h"
#include "uvm_linux.h"
#include "uvm_linux_ioctl.h"
#include "uvm_channel_mgmt.h"
#include "uvm_full_fault_buffer.h"
#include "uvm_hashmap.h"
#include "mmu/gmmu_fmt.h"
#include "uvm_full_va_trie.h"

#include "uvm_kernel_events.h"
#include "uvm_kernel_counters.h"
#include "uvm_debug_session.h"

#define _UVM_FULL_UNMANAGED_VA_SUPPORTED_

#include "uvm_events.h"

#ifdef CONFIG_HMM
#include <linux/hmm.h>
#else
struct hmm_device
{
    NvU32 info;
};

struct hmm_mirror
{
    NvU32 info;
};
#endif  // CONFIG_HMM

#define UVM_ALIGN_ADDR(addr, size)      ((NvU64)(addr) & ~((NvU64)((size) - 1)))

#define UVM_FULL_MAX_GPUS                                 (64)
#define UVM_MAX_GPU_PEER_ID                                (8)

#define PAGE_SIZE_USE_DEFAULT           0
#define PAGE_SIZE_4K                    (4*1024)
#define PAGE_SIZE_64K                   (64*1024)
#define PAGE_SIZE_2M                    (2*1024*1024)
#define PAGE_SHIFT_4K                   12
#define PAGE_SHIFT_64K                  16
#define PAGE_SHIFT_2M                   21
#define PAGE_MASK_4K                    (~((1UL<<PAGE_SHIFT_4K)-1))
#define PAGE_MASK_64K                   (~((1UL<<PAGE_SHIFT_64K)-1))
#define PAGE_MASK_2M                    (~((1UL<<PAGE_SHIFT_2M)-1))

#define PFN_SHIFT                       12
#define PFN_TO_PA(pfn)                  (((NvU64)(pfn)) << PFN_SHIFT)
#define PA_TO_PFN(pa)                   ((pa) >> PFN_SHIFT)

#define NUM_4K_PER_2M                   ((PAGE_SIZE_2M)/(PAGE_SIZE_4K))
#define NUM_64K_PER_2M                  ((PAGE_SIZE_2M)/(PAGE_SIZE_64K))
#define NUM_4K_PER_64K                  ((PAGE_SIZE_64K)/(PAGE_SIZE_4K))

#define UVM_VA_DESC_SIZE                PAGE_SIZE_2M
#define UVM_VA_DESC_SHIFT               PAGE_SHIFT_2M
#define GPU_PDE_SIZE                    PAGE_SIZE_2M
#define GPU_PDE_SHIFT                   PAGE_SHIFT_2M
#define GPU_PDE_MASK                    PAGE_MASK_2M
#define GPU_PDE_MAX_PAGES               GET_GPU_PDE_NUM_PAGES(PAGE_SIZE_4K)

#define GET_PAGE_SHIFT_MASK(pgSize)     \
        (((pgSize)==PAGE_SIZE_4K)?PAGE_SHIFT_4K:\
         ((pgSize)==PAGE_SIZE_64K)?PAGE_SHIFT_64K:\
          ((pgSize)==PAGE_SIZE_2M?PAGE_SHIFT_2M:0))

// Get the 4K pte offset within a pde (2M region)
#define GET_PDE_PTE_OFFSET(addr, pgShift)   (((addr) & ~GPU_PDE_MASK)>>(pgShift))
#define GET_GPU_PTECACHE_OFFSET(addr, pgSize)   \
        GET_PDE_PTE_OFFSET(UVM_ALIGN_ADDR(addr, pgSize), PAGE_SHIFT_4K)

#define GET_PDE_PTE(pGpuPde, addr, pgSize)      \
           ((pGpuPde)->pageTable.ptes[(GPU_PDE_PTE_OFFSET((addr),GET_PAGE_SHIFT_MASK((pgSize))))])

#define GET_VA2M_HINT(pVa2m, addr, pgSize)      \
           (&(pVa2m)->hints[(GET_PDE_PTE_OFFSET((addr),GET_PAGE_SHIFT_MASK((pgSize))))])

#define GET_GPU_PDE_NUM_PAGES(pgSize)   \
           ((GPU_PDE_SIZE/(pgSize)))

#define UVM_CHECK_VALID_PAGE_SIZE(pgSize, retval)   \
           if ((pgSize)!=PAGE_SIZE_4K && (pgSize)!=PAGE_SIZE_64K && (pgSize)!=PAGE_SIZE_2M) \
           { UVM_PANIC_ON(1); return (retval);}

#define  GET_UNIQUE_PEER_LINK_IDX(gpuId0, gpuId1)   \
            (min(gpuId0, gpuId1) * UVM_FULL_MAX_GPUS) + max(gpuId0, gpuId1)

#define IS_MULTIGPU(pVaspace) ((pVaspace)->nActiveGpus > 1)

#define UVM_INVALID_PFN     (~0UL)

#define UVM_DEFAULT_GPU_PAGE_SIZE       PAGE_SIZE_64K

#define UVM_INVALID_FAULT_BUFFER_PTR    (~0U)

#define UVM_VA_PROP_PROT_NONE            0x0
#define UVM_VA_PROP_PROT_READ            0x1
#define UVM_VA_PROP_PROT_WRITE           0x2
#define UVM_VA_PROP_PROT_ATOMIC          0x4

#define UVM_CPU_MIN_PAGE_SIZE            PAGE_SIZE
#define UVM_CPU_MIN_PAGE_SHIFT           PAGE_SHIFT
/**
 * UVM-Full locking order:
 *
 * 1. UvmGpu::UvmGpuFaultBufferInfo::lock
 * 2. UvmGlobalState::lock
 * 3. UvmVaspace::lock
 * 4. UvmVaspace::rmMemoryMapLock
 * 5. UvmVaspace::UvmVat::lock(s)
 * 6. UvmVaspace::Uvm2MbVaDesc::lock
 */

/**
 * @struct UvmGpuFaultBufferInfo
 * @brief This structure defines the fault buffer information associated
 * with a given GPU
 */
struct UvmGpuFaultBufferInfo_tag
{
    NvBool bAlive;                    //!< The faultbuffer/GET,PUT are ready
    NvU32 faultBufferClass;           //!< HW fault buffer class number
    NvU8 *faultBufferPtr;             //!< Fault buffer cpu mapped address.
                                      //   If NULL gpu is not capable of
                                      //   handling gpu-faults
    NvBool bBufferOverflow;           // Is sw cached buffer overflow flag
    NvBool bOnlyReplaysNoResolves;     // Is set to true by the test IOCTL when
                                      // faults will always replay but not resolved
    //! the following memory addresses should not be accessed if bValid is false
    volatile NvU32 *pFaultBufferGet;        //!< cpu mapped gpu bar0 get pointer
    volatile NvU32 *pFaultBufferPut;        //!< cpu mapped gpu bar0 put pointer
    volatile NvU32 *pFaultBufferInfo;       //!< cpu mapped gpu bar0 fault info ptr
    volatile NvU32 *pReplay;                //!< (verif-only)cpu va bar0
                                            //    replay register
    volatile NvU32 *pPmcIntr;
    volatile NvU32 *pPmcIntrEnSet;
    volatile NvU32 *pPmcIntrEnClear;
    volatile NvU32 *pPrefetchCtrl;      //!< Throttle the prefetch access faults
    NvU32 replayableFaultMask;
    NvU32 faultBufferSize;              //!< fault-buffer size
    NvU32 maxFaultsInBuffer;            //!< Max faults in the fault-buffer
    NvU32 faultEntrySizeInBytes;        //!< Size of fault-buffer
    NvU32 cachedFaultBufferGet;         //!< cpuva fault-buffer get index
    NvU32 cachedFaultBufferPut;         //!< cpuva fault-buffer put index
    UvmFaultBufferOps faultOps;         //!< Fault buffer Operations

    struct list_head pendingCancelList; //!< Collection of pending fault cancels
    NvBool bCancelWindowValid;          //!< Whether a cancel window is opened
    NvU32 cancelWindowStart;            //!< Start of the cancel window
    NvU32 cancelWindowEnd;              //!< End of the cancel window
    UvmTracker replayTracker;           //!< Replays should not stomp on each other

    struct rw_semaphore lock;           //!< To be used in Trylock mode

    NvU64 nPrefetches;                  //!< Total num of prefetches since start
    NvU32 prefetchPenalty;              //!< High penalty count would disable
                                        //   prefetch accesses coming to the
                                        //   faultbuffer.
    NvBool bPrefetchThrottled;          //!< Whether prefetches have been throttled.
    NvHandle faultBufferHandle;         //!< Handle to the RM fault buffer object.

};

/**
 * @struct UvmGpuCancelFault
 * @brief  Represents a pending fault which has been cancelled for a
 *         given gpu client/tlb. Further faults on this clients that
 *         can be resolved, needs to be resolved before cancelling
 */

struct UvmGpuCancelFaultInfo_tag
{
    NvU32                   gpcId;
    NvU32                   clientId;
    NvU64                   instPtr;
    UvmFaultInstLoc         aperture;
    struct list_head        node;
};


/**
 * @struct UvmGpuDevice
 * @brief Describes a single GPU
 */
struct UvmGpu_tag
{
    NvBool                  bValid;

    //! Internal gpu_id. This is also the index into the global device array.
    UvmGpuId                gpuId;
    UvmGpuUuid              gpuUuid;
    UvmGpuCaps              gpuCaps;
    NvU32                   gpuArch;
    NvU32                   faultBufferClass;
    NvBool                  bUseMemops;  // TODO: remove once memops verified.
    NvBool                  bZeroFbGpu;

    //! The CE channel manager associated with this GPU.
    UvmChannelManager       *channelManager;

    UvmGpuFaultBufferInfo   faultBufferInfo;

    //! The fault packet provides the instance ptr to identify the context.
    //! The faulting gpu(UvmGpu) is identified by the uuid.
    //! Here we use an instancePtr table to provide us the UvmChannel
    //! which caused the fault.
    UvmHashMap              instPtrMap;

    // Uvm driver's per gpu session with the RM
    uvmGpuSessionHandle        gpuSession;

    // Per gpu vaspace handle with ops to do fault buffer operations.
    uvmGpuAddressSpaceHandle   gpuHandle;

    NvU64                  zeroed2MPa : 63;
    NvU64                  zeroPageAllocated : 1;

    // TODO: Add the pfn tree here.
    struct
    {
        NvBool              bProviderWritesPtes;
        NvBool              bUseInlineCopies;
    } policies;

    // FixMe: These lists should go away when a SW trie
    // is implemented to track GPU PA.
    // List to track pages allocated for indetity map
    struct list_head        gpuPageList;
    // List to track CPU mem used to create an identity
    // mapping dump
    struct list_head        cpuPageList;

    NvU64                   testZfodPfn;
    NvU32                   testZfodPattern;
    NvBool                  bTestZfodPfnValid;
};

struct UvmUuidMapEntry_tag
{
    UvmGpuUuid    uuid;
    UvmGpuId      uvmId;
    //! An active gpu would have a non zero reference.
    //! Gpu register/unregister threads need to evaluate ref.
    NvU64         ref;
    //! This device is hooked up to service interrupts
    NvBool        bIntrServiceable;
};

struct UvmUuidMap_tag
{
    //! num of valid entries in the map
    NvU32             nEntries;

    //! Each entry houses an UUID and the corresponding UvmId.
    UvmUuidMapEntry    uvmUuidMap[UVM_FULL_MAX_GPUS];

    //! This structure is also accessed by the ISR and has to be protected by
    //! a spinlock.
    spinlock_t         lock;
};

/**
 * @struct UvmGpuPeerCaps
 * @brief Peer Caps for a specific GPU. The structure contains
 *        the Peer ID from a specific source GPU and physical
 *        caps
 */
struct UvmGpuPeerCaps_tag
{
    UvmGpuPeerId  gpuPeerId;            // PeerID of GPU w.r.t source GPU
    NvU32         peerMapValid : 1;     // Set if there is a valid Peer mapping
    NvU32         writeSupported : 1;   // Peer Writes are supported
    NvU32         readSupported : 1;    // Peer Reads are supported
    NvU32         propSupported : 1;    // Peer to Peer PROP supported
    NvU32         nvlinkSupported : 1;  // NvLink is supported between Peers
    NvU32         atomicSupported : 1;  // Peer Atomics are supported
};

/**
 * @struct UvmGlobalState
 * @brief Abstracts the Nvidia uvm capable device(s) and the
 *        associated device driver in the system. The UvmDevice exposed to
 *        HMM/OS can represent multiple GPU physical devices internally.
 */
struct UvmGlobalState_tag
{
    //! hmm device associated with the Uvm device
    struct hmm_device       hmmDevice;

    //! list of all mirrors associated with this hmm device (per process)
    struct list_head        uvmMirrors;

    //! The map from uuid to uvmid
    UvmUuidMap              uuidMap;

    //! Array of GPUs in the system
    UvmGpu                  uvmGpus[UVM_FULL_MAX_GPUS];

    //! list of all vaspaces associated with this hmm device (per process)
    struct list_head        uvmVaspaces;

    //! Peer caps for all GPUs. We are keeping a MAX_GPU x MAX_GPU caps
    //! array to ease searching. The First index is Source GPU index and
    //! next index is target peer mapped GPU index
    UvmGpuPeerCaps          uvmPeerCaps[UVM_FULL_MAX_GPUS][UVM_FULL_MAX_GPUS];

    //! Reference count on unique peer link between two GPUs. Link A->B is
    //! considered same as link B->A and only single reference count is
    //! maintained on them.
    NvU64            uvmPeerLinkRef[(UVM_FULL_MAX_GPUS * UVM_FULL_MAX_GPUS)];

    //! lock protecting the global state.
    //! This lock is almost always taken in read-mode except for the case:
    //!  - A new gpu getting added.
    //!  - A new process opening the uvm driver.
    //!  - A new hmm mirror getting created.
    struct rw_semaphore     lock;
};

/**
 * @struct UvmGpuVaspace
 * @brief Abstracts the GPU virtual address space
 *        associated with a UvmVaspace.
 *
 */
struct UvmGpuVaspace_tag
{
    //! The container of this gpu vaspace
    UvmVaspace              * uvmVaspace;

    //! Physical address of the Page-directory-base for this GPU va space
    //! All channels in this vaspace will bind to the same PDB.
    NvU64                   gpuPDB;
    UvmMemType              pdbAperture;

    //!< The GPU device for which this VA space is associated with
    UvmGpuId                gpuId;

    //! List of channels/instance pointers on this gpu vaspace
    struct list_head        channels;

    //! Also a fast map from hUvmChannel to UvmChannel
    UvmHashMap              hChannelMap;

    //! Map of all the engine ctx buffers keys by their handles.
    UvmHashMap              engCtxMap;

    //! The correct place for this is in UvmGpu but the ops to query this
    //! requests a gpu address space!
    NvBool                  bZeroFb;

    struct
    {
       NvBool                           bValid;

       // Handles passed in by the client
       NvHandle                         hGpuVaspace;

       // Per gpu vaspace session to do vaspace operations.
       uvmGpuSessionHandle              uvmGpuSession;
       // The duped device/vaspace to be used for pte/gpuvaspace manipulations.
        uvmGpuAddressSpaceHandle        dupGpuVaspace;
        GMMU_FMT                        *pGmmuFmt;
        NvU64                           bigPageSizeMask;
    } client;

    struct
    {
        NvU32                           entry[4];
        NvU32                           entrySize;
    } pteDefaults;

    struct
    {
        NvU32                           pde0EntrySize;
    } pdeDefaults;

    struct
    {
        // bForceReplayStartAckAll: (INPUT)
        //  When set to true, changes the replay policy to always use
        //  start ack all as replay type.
        NvBool                          bForceReplayStartAckAll;

        // bForcePteMemsets: (INPUT)
        //  When set to true, changes PTE update policy to use memsets,
        //  else when false, set policy to inline copy.
        NvBool                          bForcePteMemsets;

        // mappingSize: (INPUT)
        //  Specifies the page size to be used for mapping.
        NvU32                           mappingSize;

        // faultServiceRange: (INPUT)
        //  Specifies the range of memory to be mapped when servicing fault.
        NvU32                           faultServiceRange;
        // bForceMigration: (INPUT)
        // when set to true, fault handler migrates the memory
        // from CPU to GPU on fault. Else, only migrated in other
        // conditions like prefetch and user initiated migration.
        NvBool                          bForceMigration;
        // bForceP2PMigration: (INPUT)
        // when set to true, fault handler migrates the memory
        // from (Peer GPU/CPU) to the GPU on fault.
        NvBool                          bForceP2PMigration;
        // Randomize pte aperture
        NvBool                          bRandomizePteAperture;
        NvBool                          bPteApertureSysmem;
    } policy;

    // If the gpu vaspace has zfod enabled or not
    NvBool                              bTestZfodEnabled;
};

/**
 * @struct UvmClientChannel
 * @brief Represents the GPU instance ptr for a given client/CUDA channel
 *        bound to a gpu and vaspace. An instance ptr is provided in the
 *        fault buffer entry to specify the fault context.
 */
struct UvmClientChannel_tag
{
    UvmGpuId                gpuId;
    NvHandle                hChannel;
    NvU64                   instancePtr;
    UvmFaultInstLoc         instancePtrLoc;
    //! Handle provided by RM for instance ptr allocation
    NvP64                   hInstancePtr;
    UvmVaspace              *vaspace;
    //! The tracker for all activity on this instancePtr
    UvmTracker              tracker;
    struct list_head        node;
    UvmHashEntry            instptrMapEntry;  // instptr map entry

    //! entry in channel handle map in UvmGpuVaspace
    UvmHashEntry            hChannelMapEntry;

    NvU32                   nCtxBuffers;
    //! Array of all the engine Ctx buffer handles in use by this channel.
    NvU64                   *ctxBufferHandleArray;
};

/**
 * @struct UvmEngCtx
 * @brief Represents the Compute engine context (buffer) used by a
 *        client/CUDA channel bound to a gpu and vaspace. A context can be
 *        shared by multiple channels if it is global.
 */
struct UvmEngCtx_tag
{
    NvU64                   ctxBufferHandle;
    UvmGpuId                gpuId;
    UvmVaspace              *vaspace;
    NvU64                   vaStart;
    NvU64                   length;
    NvBool                  bGlobalBuffer;
    NvU32                   usageCount;
    //! Entry in the ctxBufferHashMap
    UvmHashEntry            ctxBufferMapEntry;
};

/**
 * @struct UvmRmMemory
 * @brief Represents the externally allocated memory(allocated using RM)
 *        mapped into the UVM gpuVaspace.
 */
struct UvmRmMemory_tag
{
    //! Handle passed in by the client
    NvHandle                hRmMemory;

    //! entry in the hRmMemory map in UvmVaspace
    UvmHashEntry            hRmMemoryMapEntry;

    //! List of mappings
    struct list_head        rmMemRegions;
};

/**
 * @struct UvmRmMemoryRegion
 * @brief Represents the va region mapped to externally allocated memory
 *        (allocated using RM).
 */
struct UvmRmMemoryRegion_tag
{
    NvU64                   vaStart;
    NvU64                   length;
    //! Dups of the handle in the gpu vaspaces interested in mapping to.
    NvHandle                hRmMemoryDups[UVM_FULL_MAX_GPUS];

    //! Node in rmMemRegions list
    struct list_head        rmMemRegionNode;
};

/**
  * @struct UvmVaspace
  * @brief Abstracts the per-process virtual address space.
  *        HMM mirror associated with the UvmDevice is contained within this
  *        structure.
  */
struct UvmVaspace_tag
{
    //! Array of gpu vaspaces registered to this process.
    //! Use UvmGpuId to index into this.
    UvmGpuVaspace           gpuVaspaces[UVM_FULL_MAX_GPUS];

    //! Mask of registered gpus in the process.
    NvU64                   activeGpuMask;

    //! Mask of peer enabled gpus in the process.
    NvU64                   validPeerMask[UVM_FULL_MAX_GPUS];

    //! Global loopback mode enabled only for testing.
    NvBool                  bLoopback;

    //! Num of active gpus
    NvU8                    nActiveGpus;

    //! Indicates if we are going to use HMM for this process
    NvBool                  bUsingHmm;

    //! UvmVaspace can be obtained by doing a container of on hmm_mirror
    struct hmm_mirror       hmmMirror;

    //! This mirror node in global mirror list
    struct list_head        uvmMirrorListNode;

    //! To denote if the VA Space information is valid
    NvBool                  bValidRmClient;

    //! Cuda client handle for this process
    NvHandle                hRmClient;

    struct rw_semaphore     lock;

    struct file             *filep;

    //! Map of all the memory handles imported into the gpu vaspace.
    UvmHashMap              hRmMemoryMap;

    //! Adding an RM allocation changes the RMMemory HashMap but it is similar
    //! to mapping a memory range(which does not take the vaspace lock in
    //! write mode). So the hRmMemoryMap needs its own lock.
    struct rw_semaphore     rmMemoryMapLock;

    //! Needed untill rm starts using the fd
    NvU32                   pid;

    struct list_head        vaSpaceNode;

    UvmVat                  vat;

    // UVM Full Events

    //! Per Process Debug Session array.
    UvmSessionInfo sessionInfoArray[UVM_MAX_SESSIONS_PER_PROCESS];
    UvmCounterContainer *pCounterContainer;
    UvmEventContainer   *pEventContainer;

    //! effective user id of the process owning the vaspace, for security check
    uid_t euid;
};

/**
 * @struct UvmPteHints
 * @brief Describes the UVM-specific properties pertaining to the region.
 */
struct UvmPteHints_tag
{
    NvU64                   hmmFlags             :8;

    NvU64                   va_prot_read         :1;
    NvU64                   va_prot_write        :1;

    //! Home processor (CPU/GPU) TODO: Use the processorId from uvm.h
    NvU64                   homeProcessor        :6;

    //! Is this part of a driver mmap-ed region
    NvU64                   managedVa            :1;

    //! Is this memory imported from RM. The cpuPfn is the hMemory in that case.
    NvU64                   rmMemory             :1;

    //! cannot migrate from current region
    NvU64                   notMigratable        :1;

    //! Memory is pinned. (see uvm8.h)
    // NvU64                   pinned            :1;
    NvU64                   pinToCpu             :1;

    NvU64                   testPolicyEnabled    :1;
    NvU64                   testPolicyReadWrite  :1;
    NvU64                   testPolicyAtomic     :1;
    NvU64                   testPolicyRdDuplicate:1;
    NvU64                   testPolicySysMemOnly :1;
    NvU64                   testPolicyCancel     :1;

    NvU64                   cpuPfnValid          :1;
    NvU64                   cpuPfnMapped         :1;
    NvU64                   cpuPfnDirty          :1;
    NvU64                   cpuPfn               :35; // 4K aligned.

};

struct UvmPteHintsParams_tag
{
    NvU32  updateFlags;
    struct UvmPteHints_tag pte;
};

#define UVM_GPU_VA_PTE_HINTS_FLAG_SET_HOME_PROCESSOR     (NVBIT(1))
#define UVM_GPU_VA_PTE_HINTS_FLAG_SET_MANAGED_VA         (NVBIT(2))
#define UVM_GPU_VA_PTE_HINTS_FLAG_SET_NOT_MIGRATABLE     (NVBIT(3))
#define UVM_GPU_VA_PTE_HINTS_FLAG_SET_PROT_READ          (NVBIT(4))
#define UVM_GPU_VA_PTE_HINTS_FLAG_SET_PROT_WRITE         (NVBIT(5))
#define UVM_GPU_VA_PTE_HINTS_FLAG_SET_FLT_PLCY_RW        (NVBIT(6))
#define UVM_GPU_VA_PTE_HINTS_FLAG_SET_FLT_PLCY_ATOMIC    (NVBIT(7))
#define UVM_GPU_VA_PTE_HINTS_FLAG_SET_FLT_PLCY_RD_DUP    (NVBIT(8))
#define UVM_GPU_VA_PTE_HINTS_FLAG_SET_FLT_PLCY_SYSONLY   (NVBIT(9))
#define UVM_GPU_VA_PTE_HINTS_FLAG_SET_FLT_PLCY_CANCEL    (NVBIT(10))
#define UVM_GPU_VA_PTE_HINTS_FLAG_SET_CPU_PFN_VALID      (NVBIT(11))
#define UVM_GPU_VA_PTE_HINTS_FLAG_SET_CPU_PFN_VALUE      (NVBIT(12))
#define UVM_GPU_VA_PTE_HINTS_FLAG_SET_CPU_PFN_MAPPED     (NVBIT(13))
#define UVM_GPU_VA_PTE_HINTS_FLAG_SET_CPU_PFN_DIRTY      (NVBIT(14))
#define UVM_GPU_VA_PTE_HINTS_FLAGS_SET_RM_MEMORY         (NVBIT(15))
#define UVM_GPU_VA_PTE_HINTS_FLAGS_SET_ZFOD_MAPPED       (NVBIT(16))


#define NV_COMPARE_HINTS_ON_FAULT \
             UVM_GPU_VA_PTE_HINTS_FLAG_SET_HOME_PROCESSOR     |\
             UVM_GPU_VA_PTE_HINTS_FLAG_SET_MANAGED_VA         |\
             UVM_GPU_VA_PTE_HINTS_FLAG_SET_NOT_MIGRATABLE     |\
             UVM_GPU_VA_PTE_HINTS_FLAG_SET_PROT_READ          |\
             UVM_GPU_VA_PTE_HINTS_FLAG_SET_PROT_WRITE         |\
             UVM_GPU_VA_PTE_HINTS_FLAG_SET_FLT_PLCY_RW        |\
             UVM_GPU_VA_PTE_HINTS_FLAG_SET_FLT_PLCY_ATOMIC    |\
             UVM_GPU_VA_PTE_HINTS_FLAG_SET_FLT_PLCY_RD_DUP    |\
             UVM_GPU_VA_PTE_HINTS_FLAG_SET_FLT_PLCY_SYSONLY   |\
             UVM_GPU_VA_PTE_HINTS_FLAG_SET_FLT_PLCY_CANCEL    |\
             UVM_GPU_VA_PTE_HINTS_FLAGS_SET_RM_MEMORY         |\
             UVM_GPU_VA_PTE_HINTS_FLAGS_SET_ZFOD_MAPPED

/**
 * @struct UvmGpuPte
 * @brief This structure describes the cached information for the given
 *        4k or 64k or 2Mb pte.
 * Size: 8 NvU8s
 */
struct UvmGpuPte_tag
{
    //! Pte SW cache is valid
    NvU64                   swcachevalid:1;
    //!< sw operations in flight.
    NvU64                   swtransient:1;

    //! Pte is valid (this is the HW valid)
    NvU64                   valid:1;
    NvU64                   readonly:1;
    NvU64                   atomic:1;

    //! vidmem, peer, sysmem, sysmem-non-coherent
    NvU64                   aperture:2;
    //! Relative gpu peer id
    NvU64                   peerid:3;

    //! Volatile Pte
    NvU64                   vol:1;

    //! The 4K aligned physical address in the pte
    NvU64                   pfn:35;
    NvU64                   pfnvalid:1;

    //! Pte hints (TODO: to be expanded)
    NvU64                   hints:8;

    // The pte entry was generated by RM and we just blit it to HW.
    // swcachevalid would be true.
    NvU64                   rmPte:1;

    // The pte has been mapped to the ZFOD page on GPU
    NvU64                   zfodMapped:1;

    NvU64                   rsvd:7;
};

/*
 * These flags are used to compare pte params for pte merge
 */
#define UVM_GPU_SW_PTE_FLAGS_SWVALID             NVBIT(1)
#define UVM_GPU_SW_PTE_FLAGS_TRANSIENT           NVBIT(2)
#define UVM_GPU_SW_PTE_FLAGS_PTE_VALID           NVBIT(3)
#define UVM_GPU_SW_PTE_FLAGS_READ_ONLY           NVBIT(4)
#define UVM_GPU_SW_PTE_FLAGS_ATOMIC              NVBIT(5)
#define UVM_GPU_SW_PTE_FLAGS_APERTURE            NVBIT(6)
#define UVM_GPU_SW_PTE_FLAGS_PEERID              NVBIT(7)
#define UVM_GPU_SW_PTE_FLAGS_VOLATILE            NVBIT(8)
#define UVM_GPU_SW_PTE_FLAGS_PFN                 NVBIT(9)
#define UVM_GPU_SW_PTE_FLAGS_PFN_VALID           NVBIT(10)
#define UVM_GPU_SW_PTE_FLAGS_HINTS               NVBIT(11)
#define UVM_GPU_SW_PTE_FLAGS_RMPTE               NVBIT(12)
#define UVM_GPU_SW_PTE_FLAGS_ZFOD_MAPPED         NVBIT(13)
#define UVM_GPU_SW_PTE_FALGS_RSVD                NVBIT(14)

#define NV_COMPARE_GPU_PTE_ON_MERGE \
        UVM_GPU_SW_PTE_FLAGS_SWVALID       |\
        UVM_GPU_SW_PTE_FLAGS_TRANSIENT     |\
        UVM_GPU_SW_PTE_FLAGS_PTE_VALID     |\
        UVM_GPU_SW_PTE_FLAGS_READ_ONLY     |\
        UVM_GPU_SW_PTE_FLAGS_ATOMIC        |\
        UVM_GPU_SW_PTE_FLAGS_APERTURE      |\
        UVM_GPU_SW_PTE_FLAGS_PEERID        |\
        UVM_GPU_SW_PTE_FLAGS_VOLATILE      |\
        UVM_GPU_SW_PTE_FLAGS_PFN_VALID     |\
        UVM_GPU_SW_PTE_FLAGS_HINTS         |\
        UVM_GPU_SW_PTE_FLAGS_RMPTE         |\
        UVM_GPU_SW_PTE_FLAGS_ZFOD_MAPPED

typedef struct
{
    NvU32 read_faults;
    NvU32 write_faults;
    NvU32 atomic_faults;

    NvU32 migrations;

    NvBool thrashing;

    UvmPerfTimestamp fault_tstamps[UVM_PERF_INFO_MAX_TSTAMPS];
    unsigned short tstamps_head; // Assume UVM_PERF_INFO_MAX_TSTAMPS < 2^16
    unsigned short tstamps_tail;

    NvU8  thrashing_faults;
    NvU32 thrashing_total;

    union
    {
        struct
        {
            NvU32 naps;
        } throttle;
        struct
        {
            NvU32 pins;
        } pin_sysmem;
    };
} UvmPerfPageCounters;

/**
 * @struct Uvm2MbVaDesc
 * @brief Describes the 2MB VA space for the UVM process.
 */
struct Uvm2MbVaDesc_tag
{
    //! List of gpu_pdes.for the gpus participating in this va region.
    struct list_head        gpuPDEs; // TODO: have this as a gpuid->gpuPde map

    //! HMM flags and policy per 4k page in this PDE
    UvmPteHints             hints[NUM_4K_PER_2M];

    //! A 2MB granular tracker for the va range.
    UvmTracker              tracker;

    //! the starting VA of this 2MB block
    NvU64                   addr;

    UvmVaspace              *pVaspace;

    //! Pointer to the parent in the va trie. Parent has the lock.
    UvmVatL0Entry           *parent;

    //! Num of 4k chunks tracked in this 2Mb range. Used in reclaim.
    NvU32                   nValid4k;

    UvmPerfPageCounters     perf[NUM_4K_PER_2M];
};

/*
 * During a pte update, set the appropriate flags to specify
 * which fields to update
 */
#define UVM_GPU_SW_PTE_UPDATE_FLAGS_SWVALID             (1<<0)
#define UVM_GPU_SW_PTE_UPDATE_FLAGS_TRANSIENT           (1<<1)
#define UVM_GPU_SW_PTE_UPDATE_FLAGS_PTE_VALID           (1<<2)
#define UVM_GPU_SW_PTE_UPDATE_FLAGS_READ_ONLY           (1<<3)
#define UVM_GPU_SW_PTE_UPDATE_FLAGS_ATOMIC              (1<<4)
#define UVM_GPU_SW_PTE_UPDATE_FLAGS_APERTURE            (1<<5)
#define UVM_GPU_SW_PTE_UPDATE_FLAGS_PEERID              (1<<6)
#define UVM_GPU_SW_PTE_UPDATE_FLAGS_VOLATILE            (1<<7)
#define UVM_GPU_SW_PTE_UPDATE_FLAGS_PFN                 (1<<8)
#define UVM_GPU_SW_PTE_UPDATE_FLAGS_PFN_VALID           (1<<9)
#define UVM_GPU_SW_PTE_UPDATE_FLAGS_HINTS               (1<<10)
#define UVM_GPU_SW_PTE_UPDATE_FLAGS_PFN_AUTOINCREMENT   (1<<11)
#define UVM_GPU_SW_PTE_UPDATE_FLAGS_IMPLICIT_PFN        (1<<12)
#define UVM_GPU_SW_PTE_UPDATE_FLAGS_SPLIT_PFN           (1<<13)
#define UVM_GPU_SW_PTE_UPDATE_FLAGS_PFN_CONSTANT        (1<<14)
#define UVM_GPU_SW_PTE_UPDATE_FLAGS_ZFOD_MAPPED         (1<<15)

/*
 * During a pte update, provide a set of fields on how
 * pte updates need to be done in HW
 */
#define UVM_GPU_HW_PTE_UPDATE_FLAGS_WRITE_PTE              (1<<0)
#define UVM_GPU_HW_PTE_UPDATE_FLAGS_NO_FLUSH_AFTER_PTE_WR  (1<<1)
#define UVM_GPU_HW_PTE_UPDATE_FLAGS_NO_REPLAY              (1<<2)
#define UVM_GPU_HW_PTE_UPDATE_FLAGS_NO_REPLAY_USING_REG    (1<<3)

struct UvmGpuPteUpdateParams_tag
{
    //! will trigger the pte updates on the GPU.
    NvBool                  commitPteUpdates;

    //! Which of the below flags needs to be updated
    NvU32                   updateSwFlags;

    //! Flags corresponding to pte update in hardware
    NvU32                   updateHwFlags;

    //! Pte SW cache is valid
    NvBool                  swcachevalid;
    //! Pte is valid (this is the HW valid)
    NvBool                  valid;
    NvBool                  readonly;
    NvBool                  atomic;

    //! vidmem, peer, sysmem, sysmem-non-coherent
    UvmMemType              aperture;
    //! Relative gpu peer id
    NvU32                   peerid;

    //! Volatile Pte
    NvBool                  vol;

    //! Pte hints (TODO: to be expanded)
    NvU32                   hints;

    NvBool                  pfnvalid;
    NvBool                  zfodMapped;
};

struct UvmGpuPageTablePhyAddrDesc_tag
{
    NvU64                   addr;
    UvmMemType              aperture;
};

/**
 * @struct UvmGpuPageTable
 * @brief This structure contains all the cached PTEs associated with the
 *        VA region.
 *
 */
struct UvmGpuPageTable_tag
{
    //! Bits set in this bitmap indicates that the page associated
    //! is a 64k Page (bitmap is conveniently 32 bits).
    NvU32                      bitmap64KPte;

    //! It is possible to have the 64K and 4K pages simultaneously.
    //! The first entry points to 64K pte physical address and the next
    //! points to 4k pte physical address
#define UVM_GPU_PT_INDEX_BIG          0
#define UVM_GPU_PT_INDEX_SMALL        1

    UvmGpuPageTablePhyAddrDesc pageTablePhysAddr[2];

    //! This describes ptes at 4k granularity.If bitmap64KPte bit is set;
    //! the first 4K pte of the corresponding 64K region would have the
    //! relevant info.
    UvmGpuPte                  ptes[NUM_4K_PER_2M];
};

/**
 * @struct UvmGpuPde
 * @brief Describes the 2M VA space for a given GPU.
 */
struct UvmGpuPde_tag
{
    //! Gpu Id of the owner.
    NvU32                    hostGpuId              : 6;

    //! Indicates that its a 2MB page and the first pte can be used for info.
    NvU32                    is2MbPte               : 1;
    NvU32                    memBackingPresent      : 1;
    //! Does 4k/64k page table exist for this pde? (allocated)
    NvU32                   hasBigPageTable         : 1;
    NvU32                   hasSmallPageTable       : 1;
    //! Hw state of PDE0/parent directory entry
    NvU32                   bParentEntry2Mb         : 1;
    NvU32                   bParentEntryBigValid    : 1;
    NvU32                   bParentEntrySmallValid  : 1;
    NvU32                   rsvd                    : 19;

    //! The physical address of the the PDE pointing to this last level.
    NvU64                   parentAddr;
    UvmMemType              parentAperture;
    NvU32                   parentEntryId;

    NvU64                   gpuFirstPfn;

    //! The cached set of PTEs for a the given GPU device
    UvmGpuPageTable         pageTable;

    //! This is the link entry in Uvm2MbVaDesc.gpuPDEs list.
    struct list_head        pdeNode;
};

#define NV_UVM_SPLIT_MERGE_FLAGS_SPLIT_FIXED_PAGE_SIZE          1
#define NV_UVM_SPLIT_MERGE_FLAGS_MERGE_ONLY_IF_PHYS_CONTIG      2

int uvmfull_init(dev_t uvmBaseDev);
int uvmfull_destroy(void);

#endif // _UVM_FULL_H_
