/*******************************************************************************
    Copyright (c) 2014-2015 NVIDIA Corporation

    Permission is hereby granted, free of charge, to any person obtaining a copy
    of this software and associated documentation files (the "Software"), to
    deal in the Software without restriction, including without limitation the
    rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
    sell copies of the Software, and to permit persons to whom the Software is
    furnished to do so, subject to the following conditions:

        The above copyright notice and this permission notice shall be
        included in all copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
    IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
    FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
    THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
    LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
    FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
    DEALINGS IN THE SOFTWARE.

*******************************************************************************/

#include "uvm_full.h"
#include "uvm_common.h"
#include "uvm_gpu_ops_tests.h"
#include "uvm_full_device_mgmt.h"
#include "uvm_full_api.h"
#include "uvm_full_types.h"
#include "uvm_full_ctx_mgmt.h"
#include "uvm_full_device_mgmt.h"

#include "uvmtypes.h"
#include "uvm_kernel_events.h"
#include "uvm_kernel_counters.h"
#include "uvm_debug_session.h"
#include "uvm_full_pagetbl_mgmt.h"
#include "uvm_unit_test.h"

#include "uvm_common_test.h"

// The locking order for these function is:
//   1. uvm global lock
//   2. vaspace lock

NV_STATUS uvmfull_is_8_supported(UVM_IS_8_SUPPORTED_PARAMS *pParams,
                                      struct file *filp)
{
    pParams->is8Supported = 1;
    return NV_OK;
}

NV_STATUS uvmfull_api_register_gpu_vaspace(
                                   UVM_REGISTER_GPU_VASPACE_PARAMS* param,
                                   struct file* filep)
{
    return uvmfull_register_gpu_vaspace(filep, &param->gpuUuid,
                        param->hClient, param->hVaSpace);
}

NV_STATUS uvmfull_api_unregister_gpu_vaspace(
                                      UVM_UNREGISTER_GPU_VASPACE_PARAMS* param,
                                      struct file* filep)
{
    return uvmfull_unregister_gpu_vaspace(filep, &param->gpuUuid);
}

// UVM_RESERVE/RELEASE_VA would be deprecated and replaced by UVM_FULL versions.
// See UvmReserveVa in uvm.h
NV_STATUS
uvmfull_api_reserve_va_legacy(UVM_RESERVE_VA_PARAMS *pParams, struct file *filp)
{
    UVM_DBG_PRINT_RL("requestedBase: 0x%llx, length: 0x%llx\n",
                     pParams->requestedBase, pParams->length);
    //
    // There is nothing required here yet. The mmap() call in user space handles
    // everything.
    //
    return NV_OK;
}

NV_STATUS
uvmfull_api_release_va_legacy(UVM_RELEASE_VA_PARAMS *pParams, struct file *filp)
{
    UVM_DBG_PRINT_RL("requestedBase: 0x%llx, length: 0x%llx\n",
                     pParams->requestedBase, pParams->length);
    //
    // There is nothing required here yet. The mmap() call in user space handles
    // everything.
    //
    return NV_OK;
}

// UVM Event API
// You must hold a lock on  uvmfull_get_global()->lock to call it
static UvmVaspace* _uvm_find_vaspace_from_pid(unsigned pid)
{
    UvmGlobalState *uvm = uvmfull_get_global();
    UvmVaspace *vaspace = NULL;

    list_for_each_entry(vaspace, &uvm->uvmVaspaces, vaSpaceNode)
    {
        if (vaspace->pid == pid)
            return vaspace;
    }

    return NULL;
}

//
// On success, this routine increments refcount on UmvCounterContainer and on
// UvmEventContainer before returning them.
//
NV_STATUS uvmfull_secure_get_process_containers(
                                    unsigned pidTarget,
                                    UvmCounterContainer **ppCounterContainer,
                                    UvmEventContainer **ppEventContainer,
                                    uid_t *pEuid)
{
    UvmVaspace *vaspace = NULL;
    UvmGlobalState *uvm = uvmfull_get_global();

    down_read(&uvm->lock);

    vaspace = _uvm_find_vaspace_from_pid(pidTarget);
    if (vaspace == NULL)
    {
        up_read(&uvm->lock);
        return NV_ERR_PID_NOT_FOUND;
    }

    down_read(&vaspace->lock);

    *pEuid = vaspace->euid;

    if (!uvm_user_id_security_check(*pEuid))
    {
        up_read(&vaspace->lock);
        up_read(&uvm->lock);
        return NV_ERR_INSUFFICIENT_PERMISSIONS;
    }

    *ppCounterContainer = vaspace->pCounterContainer;
    *ppEventContainer = vaspace->pEventContainer;
    uvm_ref_counter_container(*ppCounterContainer);
    uvm_ref_event_container(*ppEventContainer);

    up_read(&vaspace->lock);
    up_read(&uvm->lock);

    return NV_OK;
}

NV_STATUS uvmfull_api_add_session(UVM_ADD_SESSION_PARAMS *pParams,
                                  struct file *filp)
{
    NV_STATUS status = NV_OK;
    UvmVaspace *vaspace = NULL;
    UvmCounterContainer *pCounterContainer = NULL;
    UvmEventContainer *pEventContainer = NULL;
    UvmGlobalState *uvm = uvmfull_get_global();
    struct vm_area_struct *countersVma = NULL;
    uid_t euid = UVM_ROOT_UID;

    unsigned long countersBaseAddress =
        (unsigned long) pParams->countersBaseAddress;

    down_read(&uvm->lock);

    vaspace = (UvmVaspace*)filp->private_data;
    if (!vaspace)
    {
        up_read(&uvm->lock);
        return NV_ERR_INVALID_REQUEST;
    }

    status = uvmfull_secure_get_process_containers(pParams->pidTarget,
                                                   &pCounterContainer,
                                                   &pEventContainer,
                                                   &euid);

    if (status != NV_OK)
    {
        up_read(&uvm->lock);
        return status;
    }

    down_write(&vaspace->lock);
    status = uvm_add_session_info(euid,
                                  pParams->pidTarget,
                                  &pParams->sessionIndex,
                                  pCounterContainer,
                                  pEventContainer,
                                  countersBaseAddress,
                                  vaspace->sessionInfoArray);

    if (status != NV_OK)
        goto cleanup;

    down_write(&current->mm->mmap_sem);

    countersVma =
        find_vma(current->mm, (unsigned long long)countersBaseAddress);
    status = uvm_map_counters_pages(pCounterContainer,
                                    (NvP64)countersBaseAddress,
                                    countersVma);

    up_write(&current->mm->mmap_sem);

    if (status != NV_OK)
    {
        uvm_remove_session_info(pParams->sessionIndex,
                                vaspace->sessionInfoArray);
    }

cleanup:
    if (status != NV_OK)
    {
        uvm_unref_counter_container(pCounterContainer);
        uvm_unref_event_container(pEventContainer);
    }
    up_write(&vaspace->lock);
    up_read(&uvm->lock);

    return status;
}

NV_STATUS uvmfull_api_remove_session(UVM_REMOVE_SESSION_PARAMS *pParams,
                                     struct file *filp)
{
    NV_STATUS status;
    UvmVaspace *vaspace = NULL;
    UvmGlobalState *uvm = uvmfull_get_global();
    UvmSessionInfo *pSessionInfo;

    down_read(&uvm->lock);

    vaspace = (UvmVaspace*)filp->private_data;
    if (!vaspace)
    {
        up_read(&uvm->lock);
        return NV_ERR_INVALID_REQUEST;
    }

    down_read(&vaspace->lock);
    status = uvm_get_session_info(pParams->sessionIndex,
                                  vaspace->sessionInfoArray,
                                  &pSessionInfo);

    if (NV_OK == status)
    {
        uvm_unref_counter_container(pSessionInfo->pCounterContainer);
        uvm_unref_event_container(pSessionInfo->pEventContainer);
        status = uvm_remove_session_info(pParams->sessionIndex,
                                         vaspace->sessionInfoArray);
    }

    up_read(&vaspace->lock);
    up_read(&uvm->lock);
    return status;
}

NV_STATUS uvmfull_api_map_counter(UVM_MAP_COUNTER_PARAMS *pParams,
                                  struct file *filp)
{
    NV_STATUS status = NV_OK;
    UvmVaspace *vaspace = NULL;
    UvmGlobalState *uvm = uvmfull_get_global();
    UvmSessionInfo *pSessionInfo;
    UvmGpu *pGpu;
    unsigned gpuIndex = 0;

    down_read(&uvm->lock);

    vaspace = (UvmVaspace*)filp->private_data;
    if (!vaspace)
    {
        up_read(&uvm->lock);
        return NV_ERR_INVALID_REQUEST;
    }

    down_read(&vaspace->lock);
    status = uvm_get_session_info(pParams->sessionIndex,
                                  vaspace->sessionInfoArray,
                                  &pSessionInfo);

    if (NV_OK == status)
    {
        if (pParams->scope == UvmCounterScopeProcessSingleGpu)
        {
            pGpu = uvmfull_uuid_to_gpu(&pParams->gpuUuid);
            if (!pGpu)
            {
                UVM_ERR_PRINT_UUID("failed to find gpu index ",
                                   &pParams->gpuUuid);
                status = NV_ERR_INVALID_ARGUMENT;
                goto cleanup;
            }
            gpuIndex = pGpu->gpuId;
        }
        status = uvm_map_counter(pSessionInfo,
                                 pParams->scope,
                                 pParams->counterName,
                                 gpuIndex,
                                 (NvUPtr*)&pParams->addr);
    }

cleanup:
    up_read(&vaspace->lock);
    up_read(&uvm->lock);
    return status;
}

NV_STATUS uvmfull_api_enable_counters(UVM_ENABLE_COUNTERS_PARAMS *pParams,
                                      struct file *filp)
{
    NV_STATUS status = NV_OK;
    UvmVaspace *vaspace = NULL;
    UvmGlobalState *uvm = uvmfull_get_global();
    UvmSessionInfo *pSessionInfo;

    down_read(&uvm->lock);

    vaspace = (UvmVaspace*)filp->private_data;
    if (!vaspace)
    {
        up_read(&uvm->lock);
        return NV_ERR_INVALID_REQUEST;
    }

    down_read(&vaspace->lock);
    status = uvm_get_session_info(pParams->sessionIndex,
                                  vaspace->sessionInfoArray,
                                  &pSessionInfo);

    if (NV_OK == status)
    {
        status = uvm_counter_state_atomic_update(pSessionInfo,
                                                 pParams->config,
                                                 pParams->count);
    }

    up_read(&vaspace->lock);
    up_read(&uvm->lock);
    return status;
}

NV_STATUS uvmfull_api_create_event_queue(UVM_CREATE_EVENT_QUEUE_PARAMS *pParams,
                                         struct file *filp)
{
    NV_STATUS status = NV_OK;
    UvmVaspace *vaspace = NULL;
    UvmSessionInfo *pSessionInfo = NULL;
    UvmGlobalState *uvm = uvmfull_get_global();
    unsigned eventQueueIndex;

    if (pParams->queueSize <= 0)
    {
        UVM_ERR_PRINT("invalid queue size %lld\n", pParams->queueSize);
        return NV_ERR_INVALID_ARGUMENT;
    }

   down_read(&uvm->lock);

    vaspace = (UvmVaspace*)filp->private_data;
    if (!vaspace)
    {
        up_read(&uvm->lock);
        return NV_ERR_INVALID_REQUEST;
    }

    down_read(&vaspace->lock);

    status = uvm_get_session_info(pParams->sessionIndex,
                                  vaspace->sessionInfoArray,
                                  &pSessionInfo);

    if (NV_OK != status)
    {
        UVM_ERR_PRINT("failed to locate session %d: 0x%x\n",
        pParams->sessionIndex, status);
        goto done;
    }

    down_write(&pSessionInfo->eventQueueInfoListLock);

    status = uvm_create_event_queue(pSessionInfo,
                                    &eventQueueIndex,
                                    pParams->queueSize,
                                    pParams->notificationCount,
                                    pParams->timeStampType);

    up_write(&pSessionInfo->eventQueueInfoListLock);

    if (NV_OK != status)
    {
        UVM_ERR_PRINT("failed to create event queue: 0x%x\n", status);
        goto done;
    }

    pParams->eventQueueIndex = eventQueueIndex;

done:
    up_read(&vaspace->lock);
    up_read(&uvm->lock);
    return status;
}

NV_STATUS uvmfull_api_map_event_queue(UVM_MAP_EVENT_QUEUE_PARAMS *pParams,
                                      struct file *filp)
{
    NV_STATUS status = NV_OK;
    UvmVaspace *vaspace = NULL;
    UvmSessionInfo *pSessionInfo = NULL;
    UvmGlobalState *uvm = uvmfull_get_global();
    UvmEventQueueInfo *pEventQueueInfo = NULL;
    NvLength numQueuePages;
    struct vm_area_struct *roVma, *rwVma;

    down_write(&current->mm->mmap_sem);

    rwVma = find_vma(current->mm,
                     (unsigned long long)pParams->userRWDataAddr);
    roVma = find_vma(current->mm,
                     (unsigned long long)pParams->userRODataAddr);

    if (!rwVma || !roVma)
    {
        status = NV_ERR_INVALID_ARGUMENT;
        goto unlock_mmap_sem;
    }

    down_read(&uvm->lock);

    vaspace = (UvmVaspace*)filp->private_data;
    if (!vaspace)
    {
        up_read(&uvm->lock);
        goto unlock_mmap_sem;
    }

    down_read(&vaspace->lock);

    status = uvm_get_session_info(pParams->sessionIndex,
                                  vaspace->sessionInfoArray,
                                  &pSessionInfo);
    if (NV_OK != status)
    {
        UVM_ERR_PRINT("failed to locate session %d: 0x%x\n",
                      pParams->sessionIndex, status);
        goto unlock_session;
    }

    //
    // Write lock here since we're modifying members of eventQueueInfoList
    // An alternative is to have one lock per list member, but this operation
    // isn't in a speed path.
    //
    down_write(&pSessionInfo->eventQueueInfoListLock);

    status = uvm_get_event_queue(pSessionInfo,
                                 &pEventQueueInfo,
                                 pParams->eventQueueIndex);

    if (NV_OK != status)
    {
        UVM_ERR_PRINT("failed to retrieve event queue: 0x%x\n", status);
        goto unlock_all;
    }

    numQueuePages = pEventQueueInfo->numQueuePages;

    status = uvm_map_event_queue(pEventQueueInfo,
                                 pParams->userRODataAddr,
                                 pParams->userRWDataAddr,
                                 &pParams->readIndexAddr,
                                 &pParams->writeIndexAddr,
                                 &pParams->queueBufferAddr,
                                 roVma,
                                 rwVma,
                                 filp);

unlock_all:
    up_write(&pSessionInfo->eventQueueInfoListLock);

unlock_session:
    up_read(&vaspace->lock);
    up_read(&uvm->lock);

unlock_mmap_sem:
    up_write(&current->mm->mmap_sem);

    return status;
}

NV_STATUS uvmfull_api_remove_event_queue(UVM_REMOVE_EVENT_QUEUE_PARAMS *pParams,
                                         struct file *filp)
{
    NV_STATUS status = NV_OK;
    UvmVaspace *vaspace = NULL;
    UvmSessionInfo *pSessionInfo = NULL;
    UvmGlobalState *uvm = uvmfull_get_global();
    UvmEventQueueInfo *pEventQueueInfo = NULL;

   down_read(&uvm->lock);

    vaspace = (UvmVaspace*)filp->private_data;
    if (!vaspace)
    {
        up_read(&uvm->lock);
        return NV_ERR_INVALID_REQUEST;
    }

    down_read(&vaspace->lock);
    status = uvm_get_session_info(pParams->sessionIndex,
                                  vaspace->sessionInfoArray,
                                  &pSessionInfo);

    if (NV_OK != status)
    {
        UVM_ERR_PRINT("failed to locate session %d\n", pParams->sessionIndex);
        goto unlock_session;
    }

    down_write(&pSessionInfo->eventQueueInfoListLock);

    status = uvm_get_event_queue(pSessionInfo,
                                 &pEventQueueInfo,
                                 pParams->eventQueueIndex);

    if (NV_OK != status)
    {
        UVM_ERR_PRINT("failed to retrieve event queue: status: 0x%x\n",
                      status);
        goto unlock_event;
    }

    uvm_remove_event_queue(pSessionInfo, pEventQueueInfo);

unlock_event:
    up_write(&pSessionInfo->eventQueueInfoListLock);
unlock_session:
    up_read(&vaspace->lock);
    up_read(&uvm->lock);

    return status;
}

NV_STATUS uvmfull_api_event_ctrl(UVM_EVENT_CTRL_PARAMS *pParams,
                                 struct file *filp)
{
    NV_STATUS status = NV_OK;
    UvmVaspace *vaspace = NULL;
    UvmSessionInfo *pSessionInfo = NULL;
    UvmEventQueueInfo *pEventQueueInfo = NULL;
    UvmGlobalState *uvm = uvmfull_get_global();

    if (pParams->eventType >= UvmEventNumTypes)
        return NV_ERR_INVALID_ARGUMENT;

    down_read(&uvm->lock);

    vaspace = (UvmVaspace*)filp->private_data;
    if (!vaspace)
    {
        up_read(&uvm->lock);
        return NV_ERR_INVALID_REQUEST;
    }

    down_read(&vaspace->lock);
    status = uvm_get_session_info(pParams->sessionIndex,
                                  vaspace->sessionInfoArray,
                                  &pSessionInfo);

    if (NV_OK != status)
        goto unlock_session;

    //
    // Write lock here since we're modifying members of eventQueueInfoList.
    // An alternative is to have one lock per list member, but this operation
    // isn't in a speed path.
    //
    down_write(&pSessionInfo->eventQueueInfoListLock);

    status = uvm_get_event_queue(pSessionInfo,
                                 &pEventQueueInfo,
                                 pParams->eventQueueIndex);
    if (NV_OK != status)
        goto unlock_event;

    if (pParams->enable)
        status = uvm_enable_event(pEventQueueInfo,
                                  pParams->eventType,
                                  pSessionInfo->pEventContainer);
    else
        status = uvm_disable_event(pEventQueueInfo,
                                   pParams->eventType,
                                   pSessionInfo->pEventContainer);

unlock_event:
    up_write(&pSessionInfo->eventQueueInfoListLock);
unlock_session:
    up_read(&vaspace->lock);
    up_read(&uvm->lock);

    return status;
}

NV_STATUS
uvmfull_api_map_rm_allocation(UVM_MAP_EXTERNAL_ALLOCATION_PARAMS *pParams,
                              struct file *filp)
{
    return uvmfull_map_rm_allocation(filp, pParams->base,
                        pParams->length, pParams->perGpuAttributes,
                        pParams->gpuAttributesCount, pParams->hClientFromRm,
                        pParams->hMemoryFromRm);
}

NV_STATUS
uvmfull_api_register_channel(UVM_REGISTER_CHANNEL_PARAMS *pParams,
                            struct file *filp)
{
    return uvmfull_gpu_vaspace_add_channel(filp, &pParams->gpuUuid,
                                pParams->hClient, pParams->hChannel,
                                pParams->base, pParams->length);
}

NV_STATUS
uvmfull_api_unregister_channel(UVM_UNREGISTER_CHANNEL_PARAMS *pParams,
                                struct file *filp)
{
    return uvmfull_gpu_vaspace_remove_channel(filp, &pParams->gpuUuid,
                                                    pParams->hChannel);
}

static NV_STATUS _uvm_run_channel_test(UvmTests test,
                                       UvmGpuUuid *pGpuUuidStruct,
                                       UvmVaspace *pVaspace,
                                       UvmGpuUuid *pPeerGpuUuidStruct,
                                       NvU32 peerId)
{
    UvmChannelManager           *pChannelManager = NULL;
    UvmChannelManager           *pPeerChannelManager = NULL;
    UvmGpu                      *pGpu = NULL;
    UvmGlobalState              *pUvm;
    NV_STATUS                   status = NV_OK;

    enum UvmtestMemblockFlag page_4k_to_128k[] =
        { PAGE_4K, PAGE_64K, PAGE_128K, 0 };
    enum UvmtestMemblockFlag page_4k_to_2m[] =
        { PAGE_4K, PAGE_64K, PAGE_128K, PAGE_2M, 0 };

    enum UvmtestMemblockFlag page_4k_to_128k_contig[] =
        { PAGE_4K | CONTIGUOUS, PAGE_64K | CONTIGUOUS,
          PAGE_128K | CONTIGUOUS, 0 };
    enum UvmtestMemblockFlag page_4k_to_2m_contig[] =
        { PAGE_4K  | CONTIGUOUS, PAGE_64K  | CONTIGUOUS,
          PAGE_128K  | CONTIGUOUS, PAGE_2M  | CONTIGUOUS, 0 };

    if (!pVaspace || !pGpuUuidStruct)
        return NV_ERR_INVALID_ARGUMENT;

    // Prevent the channel manager to be destroyed while we are running channel
    // tests
    pUvm = uvmfull_get_global();
    down_read(&pUvm->lock);
    down_read(&pVaspace->lock);

    pGpu = uvmfull_uuid_to_gpu(pGpuUuidStruct);
    if (pGpu)
        pChannelManager = pGpu->channelManager;
    if (!pChannelManager)
    {
        status = NV_ERR_INVALID_STATE;
        UVM_ERR_PRINT_NV_STATUS("Could not find a channel manager.", status);
        goto cleanup;
    }

    switch (test)
    {
        case UVM_CHANNEL_DIRECTED_TEST:
            status = uvmtest_channel_directed(pChannelManager);
            break;

        case UVM_CHANNEL_MGMT_API_BASIC_MIGRATION_TEST:
            status = uvmtest_channel_basic_migration(pChannelManager);
            break;

        case UVM_CHANNEL_MGMT_API_PUSHBUFFER_SIMPLE_SANITY_TEST:
            status = uvmtest_channel_pushbuffer_sanity(pChannelManager);
            break;

        case UVM_CHANNEL_MGMT_API_INLINE_REGION_SANITY_TEST:
            status = uvmtest_channel_pushbuffer_inline(pChannelManager);
            break;

        case UVM_CHANNEL_PHYSICAL_MEMCOPY_TEST:
            status = uvmtest_channel_physical_migration(pChannelManager);
            break;

        case UVM_CHANNEL_PAGESIZE_4K_TO_128K_DIRECTED_TEST:
            status = uvmtest_channel_pagesize_directed(pChannelManager,
                                                       page_4k_to_128k);
            break;

        case UVM_CHANNEL_PAGESIZE_4K_TO_2M_DIRECTED_TEST:
            status = uvmtest_channel_pagesize_directed(pChannelManager,
                                                       page_4k_to_2m);
            break;

        case UVM_CHANNEL_PAGESIZE_4K_TO_128K_CONTIGUOUS_DIRECTED_TEST:
            status = uvmtest_channel_pagesize_directed(pChannelManager,
                                                       page_4k_to_128k_contig);
            break;

        case UVM_CHANNEL_PAGESIZE_4K_TO_2M_CONTIGUOUS_DIRECTED_TEST:
            status = uvmtest_channel_pagesize_directed(pChannelManager,
                                                       page_4k_to_2m_contig);
            break;

        case UVM_CHANNEL_P2P_MEMCOPY_TEST:

            if (!pPeerGpuUuidStruct)
            {
                UVM_ERR_PRINT_NV_STATUS("No peer GPU uuid provided.", status);
                status = NV_ERR_INVALID_ARGUMENT;
                goto cleanup;
            }

            // Query peer GPU
            pGpu = uvmfull_uuid_to_gpu(pPeerGpuUuidStruct);
            if (pGpu)
                pPeerChannelManager = pGpu->channelManager;
            if (!pChannelManager)
            {
                status = NV_ERR_INVALID_STATE;
                UVM_ERR_PRINT_NV_STATUS("Could not find a channel manager.", status);
                goto cleanup;
            }

            status = uvmtest_channel_p2p_migration(pChannelManager, pPeerChannelManager, peerId);
            break;

        default:
            UVM_INFO_PRINT("bad channel test: 0x%x\n", test);
            status = NV_ERR_INVALID_ARGUMENT;
    }

cleanup:
    up_read(&pVaspace->lock);
    up_read(&pUvm->lock);
    return status;
}

NV_STATUS uvmfull_run_test(UVM_RUN_TEST_PARAMS *pParams, struct file *filp)
{
    NV_STATUS status = NV_OK;
    UvmVaspace *pVaspace = NULL;

    pVaspace = (UvmVaspace*)filp->private_data;

    switch (pParams->test)
    {
        case UVM_CHANNEL_MGMT_API_BASIC_MIGRATION_TEST:
        case UVM_CHANNEL_MGMT_API_PUSHBUFFER_SIMPLE_SANITY_TEST:
        case UVM_CHANNEL_MGMT_API_INLINE_REGION_SANITY_TEST:
        case UVM_CHANNEL_DIRECTED_TEST:
        case UVM_CHANNEL_PHYSICAL_MEMCOPY_TEST:
        case UVM_CHANNEL_PAGESIZE_4K_TO_128K_DIRECTED_TEST:
        case UVM_CHANNEL_PAGESIZE_4K_TO_2M_DIRECTED_TEST:
        case UVM_CHANNEL_PAGESIZE_4K_TO_128K_CONTIGUOUS_DIRECTED_TEST:
        case UVM_CHANNEL_PAGESIZE_4K_TO_2M_CONTIGUOUS_DIRECTED_TEST:
            status = _uvm_run_channel_test(pParams->test,
                                           &pParams->gpuUuid,
                                           pVaspace,
                                           0,
                                           0);
            break;
        // MultiGPU test
        case UVM_CHANNEL_P2P_MEMCOPY_TEST:
            status = _uvm_run_channel_test(pParams->test,
                                           &pParams->gpuUuid,
                                           pVaspace,
                                           &pParams->multiGpu.peerGpuUuid,
                                           pParams->multiGpu.peerId);
            break;
        default:
            UVM_INFO_PRINT("bad test: 0x%x\n", pParams->test);
            status = NV_ERR_INVALID_ARGUMENT;
    }

    return status;
}

NV_STATUS
uvmfull_api_enable_peer_access(UVM_ENABLE_PEER_ACCESS_PARAMS *pParams,
                               struct file *filp)
{
    return uvmfull_enable_peer_access(filp, &pParams->gpuUuidA,
                                          &pParams->gpuUuidB);
}

NV_STATUS
uvmfull_api_disable_peer_access(UVM_DISABLE_PEER_ACCESS_PARAMS *pParams,
                               struct file *filp)
{
    return uvmfull_disable_peer_access(filp, &pParams->gpuUuidA,
                                           &pParams->gpuUuidB);
}

NV_STATUS uvmfull_api_debug_access_memory(UVM_DEBUG_ACCESS_MEMORY_PARAMS *pParams,
                                 struct file *filp)
{
    NV_STATUS status = NV_OK;
    UvmVaspace *vaspace = NULL;
    UvmVaspace *vaspaceTarget = NULL;
    UvmSessionInfo *pSessionInfo = NULL;
    UvmGlobalState *uvm = uvmfull_get_global();
    NvU64 bitmaskSize;
    NvU64 *bitmask = NULL;

    down_read(&uvm->lock);

    vaspace = (UvmVaspace*)filp->private_data;
    if (!vaspace)
    {
        up_read(&uvm->lock);
        return NV_ERR_INVALID_REQUEST;
    }

    down_read(&vaspace->lock);

    status = uvm_get_session_info(pParams->sessionIndex,
                                  vaspace->sessionInfoArray,
                                  &pSessionInfo);
    if (NV_OK != status)
    {
        up_read(&vaspace->lock);
        goto unlock_global;
    }

    vaspaceTarget = _uvm_find_vaspace_from_pid(pSessionInfo->pidTarget);
    if (vaspaceTarget == NULL)
    {
        up_read(&vaspace->lock);
        status =  NV_ERR_PID_NOT_FOUND;
        goto unlock_global;
    }

    up_read(&vaspace->lock);

    down_read(&vaspaceTarget->lock);

    if (!uvm_user_id_security_check(vaspaceTarget->euid))
    {
        status = NV_ERR_INSUFFICIENT_PERMISSIONS;
        goto unlock_vaspaceTarget;
    }

    bitmaskSize = UVM_ALIGN_UP (pParams->sizeInBytes / PAGE_SIZE_4K, sizeof(NvU64));
    bitmask = kzalloc(bitmaskSize, NV_UVM_GFP_FLAGS);

    if(!bitmask)
    {
        status = NV_ERR_NO_MEMORY;
        goto unlock_vaspaceTarget;
    }

    status = uvmfull_access_memory(vaspaceTarget, pParams->buffer, pParams->baseAddress,
        pParams->sizeInBytes, pParams->accessType, bitmask, &pParams->isBitmaskSet);

    if (status == NV_OK && pParams->isBitmaskSet)
    {
        if (copy_to_user((void __user *)pParams->bitmask, bitmask, bitmaskSize) != 0)
            status = NV_ERR_GENERIC;
    }

unlock_vaspaceTarget:

    if(bitmask)
        kfree(bitmask);

    up_read(&vaspaceTarget->lock);

unlock_global:
    up_read(&uvm->lock);

    return status;
}

