/*******************************************************************************
    Copyright (c) 2014-2015 NVIDIA Corporation

    Permission is hereby granted, free of charge, to any person obtaining a copy
    of this software and associated documentation files (the "Software"), to
    deal in the Software without restriction, including without limitation the
    rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
    sell copies of the Software, and to permit persons to whom the Software is
    furnished to do so, subject to the following conditions:

        The above copyright notice and this permission notice shall be
        included in all copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
    IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
    FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
    THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
    LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
    FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
    DEALINGS IN THE SOFTWARE.

*******************************************************************************/
#ifndef __UVM_FULL_API__
#define __UVM_FULL_API__

#include "uvm_linux.h"
#include "uvmtypes.h"
#include "uvm_linux_ioctl.h"

// This weird number comes from UVM_PREVENT_MIGRATION_RANGE_GROUPS_PARAMS in
// uvm8. That ioctl is called frequently so we don't want to allocate a copy
// every time. It's a little over 256 bytes in size.
#define UVM_MAX_IOCTL_PARAM_STACK_SIZE 288

// If the BUILD_BUG_ON fires, use UVM_FULL_ROUTE_CMD_ALLOC instead.
#define UVM_FULL_ROUTE_CMD_STACK(cmd,functionName)                      \
    case cmd:                                                           \
    {                                                                   \
        cmd##_PARAMS params;                                            \
        if (copy_from_user(&params, (void __user*)arg, sizeof(params))) \
            return -EFAULT;                                             \
                                                                        \
        params.rmStatus = functionName(&params, filp);                  \
        if (copy_to_user((void __user*)arg, &params, sizeof(params)))   \
            return -EFAULT;                                             \
                                                                        \
        return 0;                                                       \
    }

// If the BUILD_BUG_ON fires, use UVM_FULL_ROUTE_CMD_STACK instead
#define UVM_FULL_ROUTE_CMD_ALLOC(cmd, functionName)                         \
    case cmd:                                                               \
    {                                                                       \
        int ret = 0;                                                        \
        cmd##_PARAMS *params = kmalloc(sizeof(*params), NV_UVM_GFP_FLAGS);  \
        if (!params)                                                        \
            return -ENOMEM;                                                 \
        BUILD_BUG_ON(sizeof(*params) <= UVM_MAX_IOCTL_PARAM_STACK_SIZE);    \
        if (copy_from_user(params, (void __user*)arg, sizeof(*params))) {   \
            kfree(params);                                                  \
            return -EFAULT;                                                 \
        }                                                                   \
                                                                            \
        params->rmStatus = functionName(params, filp);                      \
        if (copy_to_user((void __user*)arg, params, sizeof(*params)))       \
            ret = -EFAULT;                                                  \
                                                                            \
        kfree(params);                                                      \
        return ret;                                                         \
    }


NV_STATUS uvmfull_is_8_supported(UVM_IS_8_SUPPORTED_PARAMS *pParams,
                                      struct file *filp);

NV_STATUS uvmfull_api_register_gpu_vaspace(
                                   UVM_REGISTER_GPU_VASPACE_PARAMS* param,
                                   struct file* filep);

NV_STATUS uvmfull_api_unregister_gpu_vaspace(
                                      UVM_UNREGISTER_GPU_VASPACE_PARAMS* param,
                                      struct file* filep);

NV_STATUS
uvmfull_api_map_rm_allocation(UVM_MAP_EXTERNAL_ALLOCATION_PARAMS *pParams,
                              struct file *filp);

NV_STATUS
uvmfull_api_reserve_va_legacy(UVM_RESERVE_VA_PARAMS *pParams, struct file *filp);

NV_STATUS
uvmfull_api_release_va_legacy(UVM_RELEASE_VA_PARAMS *pParams, struct file *filp);

NV_STATUS
uvmfull_api_register_channel(UVM_REGISTER_CHANNEL_PARAMS *pParams,
                            struct file *filp);

NV_STATUS
uvmfull_api_unregister_channel(UVM_UNREGISTER_CHANNEL_PARAMS *pParams,
                                struct file *filp);

NV_STATUS
uvmfull_api_enable_peer_access(UVM_ENABLE_PEER_ACCESS_PARAMS *pParams,
                               struct file *filp);

NV_STATUS
uvmfull_api_disable_peer_access(UVM_DISABLE_PEER_ACCESS_PARAMS *pParams,
                                struct file *filp);

// UVM Events API
NV_STATUS uvmfull_api_add_session(UVM_ADD_SESSION_PARAMS *pParams,
                                  struct file *filp);
NV_STATUS uvmfull_api_remove_session(UVM_REMOVE_SESSION_PARAMS *pParams,
                                     struct file *filp);
NV_STATUS uvmfull_api_map_counter(UVM_MAP_COUNTER_PARAMS *pParams,
                                  struct file *filp);
NV_STATUS uvmfull_api_enable_counters(UVM_ENABLE_COUNTERS_PARAMS *pParams,
                                      struct file *filp);
NV_STATUS uvmfull_api_create_event_queue(UVM_CREATE_EVENT_QUEUE_PARAMS *pParams,
                                         struct file *filp);
NV_STATUS uvmfull_api_map_event_queue(UVM_MAP_EVENT_QUEUE_PARAMS *pParams,
                                      struct file *filp);
NV_STATUS uvmfull_api_remove_event_queue(UVM_REMOVE_EVENT_QUEUE_PARAMS *pParams,
                                         struct file *filp);
NV_STATUS uvmfull_api_event_ctrl(UVM_EVENT_CTRL_PARAMS *pParams,
                                 struct file *filp);
NV_STATUS uvmfull_api_debug_access_memory(UVM_DEBUG_ACCESS_MEMORY_PARAMS *pParams,
                                 struct file *filp);

// UVM Kernel Tests API
NV_STATUS uvmfull_run_test(UVM_RUN_TEST_PARAMS *pParams, struct file *filp);

#endif // __UVM_FULL_API__
