/*******************************************************************************
    Copyright (c) 2014-2015 NVIDIA Corporation

    Permission is hereby granted, free of charge, to any person obtaining a copy
    of this software and associated documentation files (the "Software"), to
    deal in the Software without restriction, including without limitation the
    rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
    sell copies of the Software, and to permit persons to whom the Software is
    furnished to do so, subject to the following conditions:

        The above copyright notice and this permission notice shall be
        included in all copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
    IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
    FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
    THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
    LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
    FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
    DEALINGS IN THE SOFTWARE.

*******************************************************************************/
#include "uvm_full.h"
#include "uvm_common.h"
#include "uvm_utils.h"
#include "nv_uvm_interface.h"
#include "uvm_full_device_mgmt.h"
#include "uvm_full_ctx_mgmt.h"
#include "uvm_full_pagetbl_mgmt.h"
#include "uvm_full_fault_handler.h"
#include "uvm_hashmap.h"
#include <linux/sort.h>
#include <linux/bitops.h>

#define HCHANNEL_HASHMAP_DEFAULT_SLOTS_LOG2         (7)
#define INSTPTR_KEY_LOC_VIDMEM                       0
#define INSTPTR_KEY_LOC_SYSMEM                       1

static struct kmem_cache * g_uvmClientChannelCache  __read_mostly = NULL;
static struct kmem_cache * g_uvmVaspaceCache        __read_mostly = NULL;
static struct kmem_cache * g_uvmRmMemoryCache       __read_mostly = NULL;
static struct kmem_cache * g_uvmRmMemoryRegionCache __read_mostly = NULL;
static struct kmem_cache * g_uvmEngCtxCache         __read_mostly = NULL;
static struct kmem_cache * g_uvmMappingCache        __read_mostly = NULL;

NV_STATUS uvmfull_ctx_mgmt_slab_init(void)
{
    NV_STATUS status;
    g_uvmClientChannelCache = NV_KMEM_CACHE_CREATE("uvm_channel_info",
                                                   UvmClientChannel);
    if (!g_uvmClientChannelCache)
    {
        status = NV_ERR_NO_MEMORY;
        goto cleanup;
    }
    g_uvmVaspaceCache = NV_KMEM_CACHE_CREATE("uvm_va_space", UvmVaspace);
    if (!g_uvmVaspaceCache)
    {
        status = NV_ERR_NO_MEMORY;
        goto cleanup;
    }
    g_uvmRmMemoryCache = NV_KMEM_CACHE_CREATE("uvm_rm_memory", UvmRmMemory);
    if (!g_uvmRmMemoryCache)
    {
        status = NV_ERR_NO_MEMORY;
        goto cleanup;
    }
    g_uvmRmMemoryRegionCache = NV_KMEM_CACHE_CREATE("uvm_rm_memory_region",
                                                    UvmRmMemoryRegion);
    if (!g_uvmRmMemoryRegionCache)
    {
        status = NV_ERR_NO_MEMORY;
        goto cleanup;
    }
    g_uvmEngCtxCache = NV_KMEM_CACHE_CREATE("uvm_eng_ctx", UvmEngCtx);
    if (!g_uvmEngCtxCache)
    {
        status = NV_ERR_NO_MEMORY;
        goto cleanup;
    }
    g_uvmMappingCache = NV_KMEM_CACHE_CREATE("uvm_mapping", struct address_space);
    if (!g_uvmMappingCache)
    {
        status = NV_ERR_NO_MEMORY;
        goto cleanup;
    }

    return NV_OK;

cleanup:
    uvmfull_ctx_mgmt_slab_destroy();
    return status;
}

void uvmfull_ctx_mgmt_slab_destroy(void)
{
    if (g_uvmClientChannelCache)
        kmem_cache_destroy_safe(&g_uvmClientChannelCache);
    if (g_uvmVaspaceCache)
        kmem_cache_destroy_safe(&g_uvmVaspaceCache);
    if (g_uvmRmMemoryCache)
        kmem_cache_destroy_safe(&g_uvmRmMemoryCache);
    if (g_uvmRmMemoryRegionCache)
        kmem_cache_destroy_safe(&g_uvmRmMemoryRegionCache);
    if (g_uvmEngCtxCache)
        kmem_cache_destroy_safe(&g_uvmEngCtxCache);
    if (g_uvmMappingCache)
        kmem_cache_destroy_safe(&g_uvmMappingCache);
}

inline NvBool uvmfull_gpu_vaspace_active(UvmVaspace *vaspace, UvmGpuId gpuId)
{
    return !!(vaspace->activeGpuMask & (NVBIT64(gpuId)));
}

static UvmEngCtx*
_get_eng_ctx_for_handle(UvmGpuVaspace *gpuVaspace, NvU64 handle)
{

    UvmHashEntry *e = uvm_hashmap_query(&gpuVaspace->engCtxMap, handle);
    if (!e)
        return NULL;
    return container_of(e, UvmEngCtx, ctxBufferMapEntry);
}

// Helper to linux sort function. Decreasing order of alignment.
static int
_cmp_buffer_aligments(const void *a, const void *b)
{
    UvmGpuChannelCtxBufferInfo *infoa;
    UvmGpuChannelCtxBufferInfo *infob;

    infoa = *((UvmGpuChannelCtxBufferInfo **)a);
    infob = *((UvmGpuChannelCtxBufferInfo **)b);

    if (infoa->alignment > infob->alignment)
        return -1;
    if (infoa->alignment < infob->alignment)
        return 1;
    return 0;
}

static void
_swap_buffer_pointers(void *a, void* b, int size)
{
    UvmGpuChannelCtxBufferInfo **pptra = a;
    UvmGpuChannelCtxBufferInfo **pptrb = b;
    UvmGpuChannelCtxBufferInfo *t;

    t = *pptra;
    *pptra = *pptrb;
    *pptrb = t;
}

static NV_STATUS
_map_ctx_buffers_for_channel(UvmClientChannel *channel, NvU64 base,
                                    NvLength length, NvU32 *nCount)
{
    NV_STATUS status;
    NvU32 bufCount = 0;
    NvU32 i;
    NvU64 va;
    NvU64 bufSize;
    NvU64 *physArray = NULL;
    UvmGpuId gpuId = channel->gpuId;
    UvmGpuVaspace *gpuVaspace = &channel->vaspace->gpuVaspaces[gpuId];
    uvmGpuAddressSpaceHandle dupGpuVaspace = gpuVaspace->client.dupGpuVaspace;
    UvmGpuChannelCtxBufferInfo *ctxBufInfos = NULL;
    UvmGpuChannelCtxBufferInfo **bufInfoPtrs = NULL;

    *nCount = 0;

    // The assumption here is that the client will always align the base to
    // big page size. TODO: this should be the max(cpu_page_size, max(gpu_page_size))
    if (base & (~PAGE_MASK_64K))
    {
        status = NV_ERR_INVALID_REQUEST;
        goto cleanup;
    }
    // Get number of ctx buffers
    status = nvUvmInterfaceGetCtxBufferCount(dupGpuVaspace, &bufCount);
    if (status != NV_OK)
        goto cleanup;
    // buffer count can be zero for a CE channel
    if (bufCount == 0)
        return status;

    ctxBufInfos = vmalloc(bufCount * sizeof(UvmGpuChannelCtxBufferInfo));
    if (!ctxBufInfos)
    {
        status = NV_ERR_NO_MEMORY;
        goto cleanup;
    }

    // also allocate space to store handles for these buffers in the channel
    channel->ctxBufferHandleArray = kzalloc(bufCount * sizeof(NvU64),
                                            NV_UVM_GFP_FLAGS);
    if (!channel->ctxBufferHandleArray)
    {
        status = NV_ERR_NO_MEMORY;
        goto cleanup;
    }

    // Get information about the buffers
    status = nvUvmInterfaceGetCtxBufferInfo(dupGpuVaspace,
                            channel->vaspace->hRmClient, channel->hChannel,
                            bufCount, ctxBufInfos);
    if (status != NV_OK)
        goto cleanup;

    // Check if va provided by the user is sufficient
    bufSize = 0;
    for (i = 0; i < bufCount; i++)
        bufSize += ctxBufInfos[i].size;   // size is after alignment.

    // Maybe the above check is not enough since the starting va may be
    // misaligned. TODO: Improve checking.
    if (bufSize > length)
    {
        status = NV_ERR_NO_MEMORY;
        goto cleanup;
    }

    // Arrange the buffers in decreasing order of alignment.
    // We need to fit them in the va range provided.
    {
        bufInfoPtrs = vmalloc(bufCount * sizeof(UvmGpuChannelCtxBufferInfo*));
        if (!bufInfoPtrs)
        {
            status = NV_ERR_NO_MEMORY;
            goto cleanup;
        }
        // start with the buffers as same order RM provided
        for (i = 0; i < bufCount; i++)
            bufInfoPtrs[i] = &ctxBufInfos[i];

        // Use the linux kernel's heapsort
        sort(bufInfoPtrs, bufCount, sizeof(UvmGpuChannelCtxBufferInfo*),
             _cmp_buffer_aligments, _swap_buffer_pointers);
    }

    // Now either map each buffer or just increment usage count if already mapped
    va = base;
    for (i = 0; i < bufCount; i++)
    {
        NvU64 *ctxHandle = &channel->ctxBufferHandleArray[i];
        UvmEngCtx *ctxBuffer = NULL;
        NvU8 gpu = (NvU8)gpuId;
        UvmGpuSurfaceMappingInfo surfMapInfo;
        UvmGpuMappingAttributes gpuAttrib;
        UvmGpuChannelCtxBufferInfo *info = bufInfoPtrs[i];
        NvU8 pteTemplate[16];
        UvmMemType aperture;

        memset(&surfMapInfo, 0x0, sizeof(surfMapInfo));
        memset(&gpuAttrib, 0x0, sizeof(gpuAttrib));

        ctxBuffer = _get_eng_ctx_for_handle(gpuVaspace,
                                                (NvU64)info->bufferHandle);
        if (ctxBuffer)
        {
            WARN_ON(ctxBuffer->gpuId != gpuId);
            UVM_PANIC_ON(ctxBuffer->ctxBufferHandle !=
                                                (NvU64)info->bufferHandle);
            WARN_ON(ctxBuffer->vaspace != channel->vaspace);
        }
        else
        {
            // allocate a new one and track it
            ctxBuffer = kmem_cache_zalloc(g_uvmEngCtxCache, NV_UVM_GFP_FLAGS);
            if (!ctxBuffer)
            {
                status = NV_ERR_NO_MEMORY;
                goto cleanup;   // TODO:update cleanup to undo transaction
            }

            // Now query the physical backing for the ctx
            physArray = vmalloc(info->pageCount*sizeof(NvU64));
            if (!physArray)
            {
                kmem_cache_free(g_uvmEngCtxCache, ctxBuffer);
                status = NV_ERR_NO_MEMORY;
                goto cleanup;
            }
            status = nvUvmInterfaceGetCtxBufferPhysInfo((void*)info->bufferHandle,
                                                    info->pageCount,
                                                    physArray);
            if (status != NV_OK)
            {
                kmem_cache_free(g_uvmEngCtxCache, ctxBuffer);
                goto cleanup;
            }

            if (va >= base+length)
            {
                kmem_cache_free(g_uvmEngCtxCache, ctxBuffer);
                status = NV_ERR_NO_MEMORY;
                goto cleanup;
            }

            // If an alignment is provided; make sure the va is usable.
            if (info->alignment < UVM_PAGE_SIZE_4K)
                info->alignment = UVM_PAGE_SIZE_4K;

            // TODO: For each gpuVaspace; qualify the alignment with legal page
            // sizes.

            surfMapInfo.mappingLength =
                                      UVM_ALIGN_UP(info->size, info->alignment);
            // page size should match the alignment
            surfMapInfo.pageSize = info->alignment;
            if (info->bIsContigous)
                surfMapInfo.pageCount = 1;
            else
            {
                // RM always returns 4k pfns
                surfMapInfo.pageCount = surfMapInfo.mappingLength/PAGE_SIZE_4K;
            }
            surfMapInfo.mappingOffset = 0;
            surfMapInfo.pteArray = physArray;

            // create a pte template from the buffer properties.
            // FIXIT !! TODO: fix the aperture transaltion hardcoding
            aperture = info->aperture == 1 ? UvmMemType_sysmem_coh:
                                                UvmMemType_vidmem;
            if (!create_pte_template(gpuVaspace, pteTemplate, 16,
                                    NV_TRUE /*bValid*/,
                                    NV_TRUE /*bVol*/,
                                    aperture,
                                    NV_FALSE /*bAtomic*/,
                                    NV_FALSE /*bReadOnly*/))
            {
                kmem_cache_free(g_uvmEngCtxCache, ctxBuffer);
                status = NV_ERR_NO_MEMORY;
                goto cleanup;
            }

            surfMapInfo.pteTemplate = pteTemplate;
            gpuAttrib.gpuMappingType = UvmGpuMappingTypeReadWriteAtomic;
            status = uvmfull_map_gpus_unmanaged_va(channel->vaspace, va, &gpu,
                                     &surfMapInfo, &gpuAttrib, 1);
            if (status != NV_OK)
            {
                kmem_cache_free(g_uvmEngCtxCache, ctxBuffer);
                status = NV_ERR_NO_MEMORY;
                goto cleanup;
            }
            ctxBuffer->bGlobalBuffer = info->bGlobalBuffer;
            ctxBuffer->ctxBufferHandle = (NvU64)info->bufferHandle;
            ctxBuffer->vaspace = channel->vaspace;
            ctxBuffer->gpuId = gpuId;
            ctxBuffer->length = surfMapInfo.mappingLength;
            ctxBuffer->vaStart = va;
            uvm_hashmap_insert(&gpuVaspace->engCtxMap,
                               &ctxBuffer->ctxBufferMapEntry,
                               ctxBuffer->ctxBufferHandle);
            va += surfMapInfo.mappingLength;
            if (physArray)
            {
                vfree(physArray);
                physArray = NULL;
            }
        }
        // increment the usageCount
        ctxBuffer->usageCount++;
        // update the handle in the channel tracking
        *ctxHandle = (NvU64)info->bufferHandle;
    }  // for each buffer
    *nCount = bufCount;
cleanup:
    if (physArray)
        vfree(physArray);
    if (ctxBufInfos)
        vfree(ctxBufInfos);
    if (bufInfoPtrs)
        vfree(bufInfoPtrs);
    if (status != NV_OK && channel->ctxBufferHandleArray)
    {
        kfree(channel->ctxBufferHandleArray);
        channel->ctxBufferHandleArray = NULL;
    }
    return status;
}

NV_STATUS uvmfull_gpu_vaspace_add_channel(struct file* filp,
                                          UvmGpuUuid *uuid,
                                          NvHandle hClient,
                                          NvHandle hChannel,
                                          NvU64 base,
                                          NvLength length)
{
    NV_STATUS status;
    UvmGpuVaspace *gpuVaspace;
    UvmVaspace *vaspace;
    UvmGpu *gpu;
    UvmGpuId gpuId;
    NvU64 key;
    NvU32 bufCount, i;
    UvmGpuChannelBufferVa *bufferArray = NULL;
    NvBool bFbLockTaken = NV_FALSE;
    UvmClientChannel *channel = NULL;
    UvmGpuChannelPhysInfo channelPhysInfo = {0};
    UvmGlobalState *uvm = uvmfull_get_global();

    gpu = uvmfull_intr_uuid_to_active_gpu(uuid);
    if (!gpu)
    {
        // This might be a non faulting gpu afterall
        gpu = uvmfull_uuid_to_active_gpu(uuid);
        if (!gpu)
            return NV_ERR_INVALID_ARGUMENT;
    }
    else
    {
        // Take the faultbuffer lock as we are going to query it.
        down_write(&gpu->faultBufferInfo.lock);
        bFbLockTaken = NV_TRUE;
    }

    // Take a write lock as we are going to update the instptr map in gpu
    UVM_DBG_PRINT_UUID("add_channel", uuid);
    down_write(&uvm->lock);
    gpu = uvmfull_uuid_to_active_gpu(uuid);
    if (!gpu || !gpu->bValid)
    {
        status = NV_ERR_INVALID_REQUEST;
        goto cleanup_global;
    }

    vaspace = (UvmVaspace*)filp->private_data;
    if (!vaspace)
    {
        status = NV_ERR_INVALID_REQUEST;
        goto cleanup_global;
    }

    down_write(&vaspace->lock);

    if (vaspace->filep != filp)
    {
        status = NV_ERR_INVALID_REQUEST;
        goto cleanup_vaspace;
    }
    gpuId = gpu->gpuId;
    gpuVaspace = &vaspace->gpuVaspaces[gpuId];
    // If a client was not registered earlier; error out
    if (!vaspace->bValidRmClient            ||
        vaspace->hRmClient != hClient       ||
        gpuVaspace->uvmVaspace == NULL      ||
        !gpuVaspace->client.bValid)
    {
        status = NV_ERR_INVALID_REQUEST;
        goto cleanup_vaspace;
    }

    channel = kmem_cache_zalloc(g_uvmClientChannelCache, NV_UVM_GFP_FLAGS);
    if (!channel)
    {
        status = NV_ERR_NO_MEMORY;
        goto cleanup_vaspace;
    }

    channel->hChannel = hChannel;
    channel->vaspace = vaspace;
    channel->gpuId = gpuId;
    uvm_init_tracker(&channel->tracker);

    // Verify that the channel belongs to the registered hVaspace
    status = nvUvmInterfaceValidateChannel(gpuVaspace->client.dupGpuVaspace,
                            gpuVaspace->client.hGpuVaspace, vaspace->hRmClient,
                            hChannel);
    if (status != NV_OK)
        goto cleanup_vaspace;

    // Map the channel buffers now
    status = _map_ctx_buffers_for_channel(channel, base, length, &bufCount);
    if (status != NV_OK)
        goto cleanup_vaspace;

    channel->nCtxBuffers = bufCount;
    if (bufCount)
    {
        bufferArray = vmalloc(bufCount * sizeof(UvmGpuChannelBufferVa));
        if (!bufferArray)
            goto cleanup_vaspace;   //TODO: destroy list in channel

        // Now that the buffers have been mapped; bind the channel
        for (i = 0; i < channel->nCtxBuffers; i++)
        {
            NvU64 bufferHandle = channel->ctxBufferHandleArray[i];
            UvmEngCtx *ctxBuffer =
                            _get_eng_ctx_for_handle(gpuVaspace, bufferHandle);
            UVM_PANIC_ON(!ctxBuffer);
            bufferArray[i].bIsGlobalBuffer = ctxBuffer->bGlobalBuffer;
            bufferArray[i].bufferHandle = (NvP64)bufferHandle;
            bufferArray[i].bufferVa = (NvP64)ctxBuffer->vaStart;
        }
    }

    status = nvUvmInterfaceBindChannel(gpuVaspace->client.dupGpuVaspace,
                                    vaspace->hRmClient, channel->hChannel,
                                    channel->nCtxBuffers, bufferArray);
    if (status != NV_OK)
    {
        status = NV_ERR_INVALID_REQUEST;
        goto cleanup_vaspace;
    }

    status = nvUvmInterfaceGetChannelPhysInfo(hClient, hChannel,
                                              &channelPhysInfo);
    if (status != NV_OK)
        goto cleanup_vaspace;

    channel->hInstancePtr = channelPhysInfo.memHandle;
    channel->instancePtr = channelPhysInfo.instPtr;
    channel->instancePtrLoc = channelPhysInfo.bInstPtrLocVidmem ?
                                  uvmFaultInstLoc_vidmem :
                                  uvmFaultInstLoc_sysmem_coh;

    // Activate the code below when GetChannelPhysInfo is fixed.
    /*
    if (gpuVaspace->gpuPDB != channelPhysInfo.pdb)
    {
        status = NV_ERR_INVALID_REQUEST;
        goto cleanup_vaspace;
    }
    */

    list_add(&channel->node, &gpuVaspace->channels);

    uvm_hashmap_insert(&gpuVaspace->hChannelMap, &channel->hChannelMapEntry,
                            channel->hChannel);

    key = uvmfull_instptr_key(channel->instancePtr, channel->instancePtrLoc);
    uvm_hashmap_insert(&gpu->instPtrMap, &channel->instptrMapEntry, key);

    UVM_ERR_PRINT("UVM channel added: hChannel 0x%x,\
                  instPtr 0x%llx, pdb 0x%llx, gpu %d\n",
                  channel->hChannel, channel->instancePtr,
                  gpuVaspace->gpuPDB, gpuId);

cleanup_vaspace:
    if (status != NV_OK && channel)
        kmem_cache_free(g_uvmClientChannelCache, channel);
    if (bufferArray)
        vfree(bufferArray);
    up_write(&vaspace->lock);
cleanup_global:
    up_write(&uvm->lock);
    if (bFbLockTaken)
    {
        up_write(&gpu->faultBufferInfo.lock);
        // Do we need to call the top half handler in case we missed any intrs?
    }
    return status;
}

static NV_STATUS
_unmap_ctx_buffers_for_channel(UvmClientChannel *channel)
{
    NV_STATUS status;
    NvU32 i;
    UvmGpuId gpuId = channel->gpuId;
    UvmGpuVaspace *gpuVaspace = &channel->vaspace->gpuVaspaces[gpuId];

    // Now either unmap each buffer or decrement usage count if still in use
    for (i = 0; i < channel->nCtxBuffers; i++)
    {
        UvmEngCtx *ctxBuffer = NULL;
        NvU8 gpu = (NvU8)gpuId;
        NvU64 bufferHandle = (NvU64)channel->ctxBufferHandleArray[i];
        ctxBuffer = _get_eng_ctx_for_handle(gpuVaspace, bufferHandle);
        if (!ctxBuffer)
        {
            WARN_ON(!ctxBuffer);
            continue;
        }
        WARN_ON(ctxBuffer->usageCount == 0);
        if (ctxBuffer->usageCount)
        {
            ctxBuffer->usageCount--;
            if (ctxBuffer->usageCount == 0)
            {
                status = uvmfull_unmap_on_gpus(channel->vaspace,
                                   ctxBuffer->vaStart, ctxBuffer->length,
                                   &gpu, 1/*on one gpu*/,
                                   NV_FALSE/*bSkipInvalidate*/);
                WARN_ON (status != NV_OK);
                uvm_hashmap_remove(&gpuVaspace->engCtxMap,
                                                &ctxBuffer->ctxBufferMapEntry);
                kmem_cache_free(g_uvmEngCtxCache, ctxBuffer);
            }
        }
    }  // for each handle
    return NV_OK;
}

// Caller needs to have the uvm lock and the vaspace lock in write mode.
static NV_STATUS
_gpu_vaspace_remove_channel(UvmVaspace *vaspace,
                            UvmGpuVaspace *gpuVaspace,
                            NvHandle hUvmChannel)
{
    UvmClientChannel *channel;
    NV_STATUS status;
    UvmHashEntry *hashEntry = NULL;
    UvmGlobalState *uvm = uvmfull_get_global();
    UvmGpu *gpu = &uvm->uvmGpus[gpuVaspace->gpuId];

    // Make sure a gpu vaspace has already been registered.
    if (gpuVaspace->uvmVaspace == NULL || !gpuVaspace->client.bValid)
        return NV_ERR_INVALID_REQUEST;

    // Find the channel using the hChannelUvm
    hashEntry = uvm_hashmap_query(&gpuVaspace->hChannelMap, (NvU64)hUvmChannel);
    if (hashEntry == NULL)
    {
        UVM_ERR_PRINT("hUvmChannel 0x%x not registered with UVM.\n",
                        hUvmChannel);
        return NV_ERR_INVALID_REQUEST;
    }
    channel = container_of(hashEntry, UvmClientChannel, hChannelMapEntry);
    UVM_PANIC_ON(vaspace != channel->vaspace);
    UVM_PANIC_ON(gpuVaspace->gpuId != channel->gpuId);

    // Wait for all channel activity to complete
    status = uvm_wait_for_tracker(&channel->tracker);
    if (status != NV_OK)
        return status;

    // Stop the channel
    nvUvmInterfaceStopChannel(gpuVaspace->client.dupGpuVaspace,
                                channel->hInstancePtr);

    // Clear all the intancePtr entries related to the expired channel
    status = uvmfull_flush_faultbuffer_for_channel(channel);
    WARN_ON (status != NV_OK);

    // Now unmap ctx buffers
    status = _unmap_ctx_buffers_for_channel(channel);
    WARN_ON (status != NV_OK);

    if (channel->nCtxBuffers)
    {
        WARN_ON(channel->ctxBufferHandleArray == NULL);
        if (channel->ctxBufferHandleArray)
        {
            // Free the ctx buffers
            nvUvmInterfaceFreeMemHandles(
                (void**)channel->ctxBufferHandleArray, channel->nCtxBuffers);
            // Free the array
            kfree(channel->ctxBufferHandleArray);
        }
    }

    // Also free the instanc ptr
    nvUvmInterfaceFreeMemHandles((void**)&channel->hInstancePtr, 1);

    // unlink the references from the hash maps.
    uvm_hashmap_remove(&gpuVaspace->hChannelMap, &channel->hChannelMapEntry);
    uvm_hashmap_remove(&gpu->instPtrMap, &channel->instptrMapEntry);

    // unlink the channel from the gpu vaspace
    list_del(&channel->node);
    kmem_cache_free(g_uvmClientChannelCache, channel);
    UVM_ERR_PRINT("UVM channel removed: hUvmChannel 0x%x,\
                  instPtr 0x%llx, gpu %d\n",
                  hUvmChannel, channel->instancePtr, gpu->gpuId);
    return status;
}

NV_STATUS uvmfull_gpu_vaspace_remove_channel(struct file* filp,
                                             UvmGpuUuid *uuid,
                                             NvHandle hUvmChannel)
{
    NV_STATUS status;
    UvmVaspace *vaspace;
    UvmGpu *gpu;
    UvmGpuVaspace *gpuVaspace;
    NvBool bFbLockTaken = NV_FALSE;
    UvmGlobalState *uvm = uvmfull_get_global();

    gpu = uvmfull_intr_uuid_to_active_gpu(uuid);
    if (!gpu)
    {
        // This might be a non faulting gpu afterall
        gpu = uvmfull_uuid_to_active_gpu(uuid);
        if (!gpu)
            return NV_ERR_INVALID_ARGUMENT;
    }
    else
    {
        // Take the faultbuffer lock as we are going to query it.
        down_write(&gpu->faultBufferInfo.lock);
        bFbLockTaken = NV_TRUE;
    }
    // Take a write lock as we are going to update the instptr map in gpu
    down_write(&uvm->lock);
    gpu = uvmfull_uuid_to_active_gpu(uuid);
    if (!gpu || !gpu->bValid)
    {
        status = NV_ERR_INVALID_REQUEST;
        goto cleanup_global;
    }

    vaspace = (UvmVaspace*)filp->private_data;
    if (!vaspace)
    {
        status = NV_ERR_INVALID_REQUEST;
        goto cleanup_global;
    }

    down_write(&vaspace->lock);

    if (vaspace->filep != filp)
    {
        status = NV_ERR_INVALID_REQUEST;
        goto cleanup_vaspace;
    }

    gpuVaspace = &vaspace->gpuVaspaces[gpu->gpuId];
    if (!gpuVaspace->client.bValid)
    {
        status = NV_ERR_INVALID_REQUEST;
        goto cleanup_vaspace;
    }
    status = _gpu_vaspace_remove_channel(vaspace, gpuVaspace, hUvmChannel);
cleanup_vaspace:
    up_write(&vaspace->lock);
cleanup_global:
    up_write(&uvm->lock);
    if (bFbLockTaken)
    {
        up_write(&gpu->faultBufferInfo.lock);
        // Do we need to call the top half handler in case we missed any intrs?
    }
    return status;
}

static NV_STATUS
_gpu_vaspace_init(UvmVaspace *vaspace, UvmGpuId gpuId)
{
    NV_STATUS status;
    UvmGpuVaspace *gpuVaspace;

    gpuVaspace = &vaspace->gpuVaspaces[gpuId];
    // Already initialized!!
    UVM_PANIC_ON(gpuVaspace->uvmVaspace);
    // Link with parent.
    gpuVaspace->uvmVaspace = vaspace;
    gpuVaspace->gpuId = gpuId;
    vaspace->activeGpuMask |= NVBIT64(gpuId);
    vaspace->nActiveGpus++;
    INIT_LIST_HEAD(&gpuVaspace->channels);

    status = uvm_hashmap_init(&gpuVaspace->engCtxMap,
                                        HCHANNEL_HASHMAP_DEFAULT_SLOTS_LOG2);
    WARN_ON(status != NV_OK);
    status = uvm_hashmap_init(&gpuVaspace->hChannelMap,
                                    HCHANNEL_HASHMAP_DEFAULT_SLOTS_LOG2);
    return status;
}

static NV_STATUS
_gpu_vaspace_destroy(UvmVaspace *vaspace, UvmGpuId gpuId)
{
    UvmClientChannel *channel;
    UvmGpuVaspace *gpuVaspace = &vaspace->gpuVaspaces[gpuId];
    NV_STATUS status = NV_OK;

    nvUvmInterfaceStopVaspaceChannels(gpuVaspace->client.uvmGpuSession,
        gpuVaspace->client.dupGpuVaspace, vaspace->hRmClient, gpuVaspace->client.hGpuVaspace);

    while (!list_empty(&gpuVaspace->channels))
    {
        channel = list_first_entry(&gpuVaspace->channels,
                                   UvmClientChannel,
                                   node);
        status = _gpu_vaspace_remove_channel(vaspace, gpuVaspace,
                                                channel->hChannel);
        // ? what to do if channel removal failed. (wait for tracker)
    }

    if (gpuVaspace->gpuPDB)
    {
        if (gpuVaspace->client.dupGpuVaspace)
            nvUvmInterfaceUnsetPageDirectory(gpuVaspace->client.dupGpuVaspace);
        uvmfull_free_pagetbl_memory(PA_TO_PFN(gpuVaspace->gpuPDB),
                                    PAGE_SIZE_4K,
                                    gpuVaspace->pdbAperture, gpuVaspace->gpuId);
    }

    if (gpuVaspace->client.dupGpuVaspace)
        nvUvmInterfaceAddressSpaceDestroy(gpuVaspace->client.dupGpuVaspace);

    if (gpuVaspace->client.uvmGpuSession)
        nvUvmInterfaceSessionDestroy(gpuVaspace->client.uvmGpuSession);

    memset(&gpuVaspace->client, 0x0, sizeof(gpuVaspace->client));

    if (vaspace->nActiveGpus)
    {

        WARN_ON(gpuVaspace->hChannelMap.nItems);
        uvm_hashmap_destroy(&gpuVaspace->hChannelMap);
        WARN_ON(gpuVaspace->engCtxMap.nItems);
        uvm_hashmap_destroy(&gpuVaspace->engCtxMap);

        // UnLink with parent.
        gpuVaspace->uvmVaspace = NULL;
        gpuVaspace->gpuId = UVM_FULL_MAX_GPUS;
        vaspace->activeGpuMask &= ~NVBIT64(gpuId);
        vaspace->nActiveGpus--;
    }
    return status;
}

NV_STATUS uvmfull_vaspace_peermap_destroy(struct file *fp)
{
    NV_STATUS status = NV_OK;
    UvmVaspace *vaspace = NULL;
    UvmGlobalState *uvm = uvmfull_get_global();

    down_write(&uvm->lock);
    
    vaspace = fp->private_data;
    if (!vaspace)
    {
        up_write(&uvm->lock);
        return NV_ERR_INVALID_REQUEST;
    }

    down_write(&vaspace->lock);

    // Walk thru all valid peers and disable them
    // TO -DO - Check if we have atleast 2 GPUs in system in production
    // Currently enabling loopback for testing
    if (vaspace->nActiveGpus > 1 || vaspace->bLoopback)
    {
        UvmGpu *pGpu1, *pGpu2;
        NvU64 gpuId1, gpuId2;
        UvmGpuPeerCaps *pPeerCaps;
        NvU64 tempActiveGpuMask = vaspace->activeGpuMask;
        while (tempActiveGpuMask)
        {
            NvU64 innerLoopActiveGpuMask, uniqueLinkIdx;

            gpuId1 = find_first_bit((unsigned long *)&tempActiveGpuMask,
                                    UVM_FULL_MAX_GPUS);

            if (!vaspace->bLoopback)
                tempActiveGpuMask &= ~NVBIT64(gpuId1);

            innerLoopActiveGpuMask = tempActiveGpuMask;

            // Start looking for all gpuIds in activeMask
            while (innerLoopActiveGpuMask)
            {
                gpuId2 = find_first_bit((unsigned long *)&innerLoopActiveGpuMask,
                                         UVM_FULL_MAX_GPUS);

                innerLoopActiveGpuMask &= ~NVBIT64(gpuId2);

                // Disable peer accesses on GPUs if enabled
                if (NV_OK != uvmfull_disable_peer_on_vaspace(vaspace,
                                                             gpuId1, gpuId2))
                    continue;

                // Decrement refcount on unique peerlink
                uniqueLinkIdx = GET_UNIQUE_PEER_LINK_IDX(gpuId1, gpuId2);

                if (--uvm->uvmPeerLinkRef[uniqueLinkIdx] == 0)
                {
                    // Un-Set peer caps for GPU1 and GPU2 if this is the last
                    // process keeping them alive
                    pPeerCaps = &uvm->uvmPeerCaps[gpuId1][gpuId2];
                    memset(pPeerCaps, 0, sizeof(UvmGpuPeerCaps));
            
                    pPeerCaps = &uvm->uvmPeerCaps[gpuId2][gpuId1];
                    memset(pPeerCaps, 0, sizeof(UvmGpuPeerCaps));
                }

                pGpu1 = uvmfull_gpuid_to_gpu(gpuId1);
                pGpu2 = uvmfull_gpuid_to_gpu(gpuId2);
           
                up_write(&vaspace->lock);
                up_write(&uvm->lock);
            
                //
                // Disable both GPUs in case this is the last process holding refCount
                // Don't bother about failure in stopGpu call as we still need to decrement
                // the refCount on both GPUs.
                //
                uvmfull_stop_gpu(&pGpu1->gpuUuid);
                uvmfull_stop_gpu(&pGpu2->gpuUuid);

                down_write(&uvm->lock);
                down_write(&vaspace->lock);
            }
        }
    }

    up_write(&vaspace->lock);
    up_write(&uvm->lock);

    return status;
}

// This function does the following:
// 1. Refcount the gpu being called for. If the gpu was not present; create it.
// 2. Start owning the gpu vaspace
NV_STATUS uvmfull_register_gpu_vaspace(struct file* filp, UvmGpuUuid *uuid,
                                       NvHandle hClient, NvHandle hVaspace)
{
    NV_STATUS status;
    status = uvmfull_start_gpu(uuid);
    if (status != NV_OK)
        return status;

    status = uvmfull_gpu_vaspace_activate(filp, uuid, hClient, hVaspace);
    if (status != NV_OK)
        uvmfull_stop_gpu(uuid);

    return status;
}

NV_STATUS uvmfull_unregister_gpu_vaspace(struct file* filp, UvmGpuUuid *uuid)
{
    NV_STATUS status;
    status = uvmfull_gpu_vaspace_deactivate(filp, uuid);
    if (status != NV_OK)
        return status;

    status = uvmfull_stop_gpu(uuid);
    return status;
}

NV_STATUS uvmfull_gpu_vaspace_activate(struct file* filp, UvmGpuUuid *uuid,
                                       NvHandle hClient, NvHandle hVaspace)
{
    UvmVaspace *vaspace;
    UvmGpuId gpuId;
    UvmGpu *gpu;
    NvU64 pfn;
    UvmMemType pdbAperture;
    UvmGpuFbInfo gpuFbInfo;
    MMU_FMT_LEVEL *pageTableFmt;
    NvBool bFirstRegistration = NV_FALSE;
    UvmGpuVaspace *gpuVaspace = NULL;
    UvmGlobalState *uvm = uvmfull_get_global();
    NV_STATUS status = NV_OK;

    down_read(&uvm->lock);
    gpu = uvmfull_uuid_to_active_gpu(uuid);
    if (gpu == NULL)
    {
        status = NV_ERR_INVALID_REQUEST;
        goto cleanup_one_lock;
    }

    vaspace = (UvmVaspace*)filp->private_data;
    if (!vaspace)
    {
        status = NV_ERR_INVALID_REQUEST;
        goto cleanup_one_lock;
    }

    down_write(&vaspace->lock);

    if (vaspace->filep != filp)
    {
        status = NV_ERR_INVALID_REQUEST;
        goto cleanup_two_locks;
    }
    gpuId = gpu->gpuId;
    gpuVaspace = &vaspace->gpuVaspaces[gpuId];
    if (gpuVaspace->uvmVaspace == NULL)
    {
        status = _gpu_vaspace_init(vaspace, gpuId);
        if (status != NV_OK)
        {
            status = NV_ERR_INVALID_REQUEST;
            goto cleanup;
        }
    }

    if (vaspace->bValidRmClient)
    {
        if (vaspace->hRmClient != hClient)
        {
            status = NV_ERR_INVALID_REQUEST;
            goto cleanup;
        }
    }
    else
    {
        bFirstRegistration = NV_TRUE;
    }
    if (gpuVaspace->client.bValid)
    {
        // Double register
        status = NV_ERR_INVALID_REQUEST;
        goto cleanup;
    }

    status = nvUvmInterfaceSessionCreate(&gpuVaspace->client.uvmGpuSession);
    if (status != NV_OK)
    {
        UVM_ERR_PRINT("ERROR: could not create a session\n");
        goto cleanup;
    }

    status = nvUvmInterfaceDupAddressSpace(gpuVaspace->client.uvmGpuSession,
                                           (NvU8*)uuid,
                                           hClient,
                                           hVaspace,
                                           &gpuVaspace->client.dupGpuVaspace);
    if (status != NV_OK)
    {
        UVM_ERR_PRINT("ERROR: could dup address space\n");
        goto cleanup;
    }

    memset(&gpuFbInfo, 0x0, sizeof(gpuFbInfo));
    status = nvUvmInterfaceGetFbInfo(gpuVaspace->client.dupGpuVaspace,
                                    &gpuFbInfo);
    if (status != NV_OK)
    {
        goto cleanup;
    }

    gpuVaspace->bZeroFb = gpuFbInfo.bZeroFb;

    // Get the gmmu utils template for this gpu vaspace
    status = nvUvmInterfaceGetGmmuFmt(gpuVaspace->client.dupGpuVaspace,
                                     (void**)&gpuVaspace->client.pGmmuFmt);
    if (status != NV_OK)
    {
        UVM_ERR_PRINT("ERROR: could not get gmmu fmt\n");
        goto cleanup;
    }

    // Lets initialize the pte defaults
    // pteFmt = gpuVaspace->client.pGmmuFmt->pPte;
    pageTableFmt = (MMU_FMT_LEVEL*)mmuFmtFindLevelWithPageShift(
                            gpuVaspace->client.pGmmuFmt->pRoot, PAGE_SHIFT_4K);
    if (!pageTableFmt)
    {
        UVM_ERR_PRINT("ERROR: Could not find pte size!\n");
        status = NV_ERR_INVALID_STATE;
        goto cleanup;
    }
    memset(&gpuVaspace->pteDefaults, 0x0, sizeof(gpuVaspace->pteDefaults));
    gpuVaspace->pteDefaults.entrySize = pageTableFmt->entrySize;


    pageTableFmt = (MMU_FMT_LEVEL*)mmuFmtFindLevelWithPageShift(
                            gpuVaspace->client.pGmmuFmt->pRoot, PAGE_SHIFT_2M);
    if (!pageTableFmt)
    {
        UVM_ERR_PRINT("ERROR: Could not find pte size!\n");
        status = NV_ERR_INVALID_STATE;
        goto cleanup;
    }
    gpuVaspace->pdeDefaults.pde0EntrySize = pageTableFmt->entrySize;


    // Now set the defaults
    UVM_PANIC_ON(gpuVaspace->pteDefaults.entrySize >
                        sizeof(gpuVaspace->pteDefaults.entry));


    // For UVM all the pte defaults would be zero. (no compression etc)

    // Allocate the PDB and set it.
    pdbAperture = gpuVaspace->bZeroFb ? UvmMemType_sysmem_coh:
                                        UvmMemType_vidmem;
    // TODO 1: use the gmmu_fmt to calculate the correct size of PDE top level
    // TODO 2: Top level pascal needs only 32 bytes. Suballocate from a 4k page
    status = uvmfull_get_pagetbl_memory(PAGE_SIZE_4K, &pfn, pdbAperture,
                                        gpuVaspace->gpuId);
    if (status != NV_OK)
    {
        UVM_ERR_PRINT("ERROR: could not allocate PDB for address space\n");
        goto cleanup;
    }

    // Scrub if vidmem is allocated
    if (pdbAperture == UvmMemType_vidmem)
    {
        UvmPbOps *pbOps = NULL;
        UvmTracker tracker;
        UvmPushbuffer *pb = NULL;
        uvm_init_tracker(&tracker);

        pbOps = uvm_pbops_start_on_gpu(gpuId);
        if (!pbOps)
            goto cleanup;

        pb = uvm_pbops_gpu_pb(pbOps, gpuId);
        uvm_push_memzero_vidmem(gpuId, pb, PFN_TO_PA(pfn), PAGE_SIZE_4K);
        uvm_pbops_submit(pbOps, gpuId, pb, &tracker);
        uvm_move_tracker(&tracker, pbOps->t);
        status = uvm_wait_for_tracker(&tracker);
        uvm_pbops_end(pbOps);
        if (status != NV_OK)
            goto cleanup;
    }
    gpuVaspace->gpuPDB = PFN_TO_PA(pfn);
    gpuVaspace->pdbAperture = pdbAperture;

    status = nvUvmInterfaceSetPageDirectory(gpuVaspace->client.dupGpuVaspace,
                    gpuVaspace->gpuPDB,
                    mmuFmtLevelEntryCount(gpuVaspace->client.pGmmuFmt->pRoot),
                    (gpuVaspace->pdbAperture == UvmMemType_vidmem));

    if (status != NV_OK)
    {
        UVM_ERR_PRINT("ERROR: could not set PDB for address space\n");
        goto cleanup;
    }
    // Now mark the client as valid
    gpuVaspace->client.hGpuVaspace = hVaspace;
    gpuVaspace->client.bValid = NV_TRUE;
    if (bFirstRegistration)
    {
        vaspace->hRmClient = hClient;
        vaspace->bValidRmClient = NV_TRUE;
    }
    memset(&gpuVaspace->policy, 0 , sizeof(gpuVaspace->policy));
    gpuVaspace->policy.bForceMigration = NV_TRUE;
    gpuVaspace->policy.mappingSize = NV_UVM_GPU_DEFAULT_FAULT_IN_SIZE;
    gpuVaspace->policy.faultServiceRange = NV_UVM_GPU_DEFAULT_FAULT_IN_SIZE;
cleanup:
    if (status != NV_OK)
    {
        if (gpuVaspace->uvmVaspace)
            _gpu_vaspace_destroy(vaspace, gpuId);
    }
cleanup_two_locks:
    up_write(&vaspace->lock);
cleanup_one_lock:
    up_read(&uvm->lock);
    return status;
}

NV_STATUS uvmfull_gpu_vaspace_deactivate(struct file* filp, UvmGpuUuid *uuid)
{
    UvmGpuVaspace *gpuVaspace;
    UvmVaspace *vaspace;
    UvmGpuId gpuId;
    UvmGpu *gpu;
    NvBool bFbLockTaken = NV_FALSE;
    NV_STATUS status = NV_OK;
    UvmGlobalState *uvm = uvmfull_get_global();

    gpu = uvmfull_intr_uuid_to_active_gpu(uuid);
    if (!gpu)
    {
        // This might be a non faulting gpu afterall
        gpu = uvmfull_uuid_to_active_gpu(uuid);
        if (!gpu)
            return NV_ERR_INVALID_ARGUMENT;
    }
    else
    {
        // Take the faultbuffer lock as we are going to query it.
        down_write(&gpu->faultBufferInfo.lock);
        bFbLockTaken = NV_TRUE;
    }
    // We may be removing channels from gpu. So take in write mode.
    down_write(&uvm->lock);
    gpu = uvmfull_uuid_to_active_gpu(uuid);
    if (gpu == NULL)
    {
        status = NV_ERR_INVALID_REQUEST;
        goto cleanup_one_lock;
    }

    vaspace = (UvmVaspace*)filp->private_data;
    if (!vaspace)
    {
        status = NV_ERR_INVALID_REQUEST;
        goto cleanup_one_lock;
    }

    down_write(&vaspace->lock);

    if (vaspace->filep != filp)
    {
        status = NV_ERR_INVALID_REQUEST;
        goto cleanup;
    }
    gpuId = gpu->gpuId;
    gpuVaspace = &vaspace->gpuVaspaces[gpuId];
    if (gpuVaspace->uvmVaspace == NULL ||
        gpuVaspace->uvmVaspace != vaspace)
    {
        status = NV_ERR_INVALID_REQUEST;
        goto cleanup;
    }

    status = _gpu_vaspace_destroy(vaspace, gpuId);

cleanup:
    up_write(&vaspace->lock);
cleanup_one_lock:
    up_write(&uvm->lock);
    if (bFbLockTaken)
    {
        up_write(&gpu->faultBufferInfo.lock);
        // Do we need to call the top half handler in case we missed any intrs?
    }
    return status;
}

NV_STATUS uvmfull_vaspace_init(struct inode *inode, struct file *fp)
{
    unsigned pid;
    NV_STATUS status = NV_OK;
    UvmVaspace *vaspace = NULL;
    struct address_space *mapping = NULL;
    UvmGlobalState *uvm = uvmfull_get_global();

    if (!fp)
        return NV_ERR_INVALID_REQUEST;

    UVM_PANIC_ON(!g_uvmVaspaceCache);
    vaspace = kmem_cache_zalloc(g_uvmVaspaceCache, NV_UVM_GFP_FLAGS);
    
    if (!vaspace)
    {
        status = NV_ERR_NO_MEMORY;
        goto fail;
    }

    mapping = kmem_cache_alloc(g_uvmMappingCache, NV_UVM_GFP_FLAGS);
    if (!mapping)
    {
        status = NV_ERR_NO_MEMORY;
        goto fail;
    }

    address_space_init_once(mapping);
    mapping->host = inode;
#if defined(NV_ADDRESS_SPACE_HAS_BACKING_DEV_INFO)
    mapping->backing_dev_info = inode->i_mapping->backing_dev_info;
#endif

    init_rwsem(&vaspace->lock);
    UVM_PANIC_ON(fp==NULL);
    vaspace->filep = fp;
    vaspace->pCounterContainer = NULL;
    vaspace->pEventContainer = NULL;
    pid = uvm_get_stale_process_id();
    vaspace->pid = pid;

    vaspace->euid = NV_CURRENT_EUID();
    memset(vaspace->sessionInfoArray, 0, sizeof (*vaspace->sessionInfoArray));
    status = uvm_alloc_counter_container(&vaspace->pCounterContainer);
    if (status != NV_OK)
       goto fail;
    status = uvm_alloc_event_container(&vaspace->pEventContainer);
    if (status != NV_OK)
       goto fail;

    init_rwsem(&vaspace->rmMemoryMapLock);
    uvm_hashmap_init(&vaspace->hRmMemoryMap, 128);
    uvmfull_vat_init(&vaspace->vat);
    // Add this va space to global va space list in the uvm-device
    down_write(&uvm->lock);

    list_add(&vaspace->vaSpaceNode, &uvm->uvmVaspaces);

    // update the private data of the file pointer with va space
    fp->private_data = vaspace;
    fp->f_mapping = mapping;

    up_write(&uvm->lock);
    return NV_OK;
fail:
    up_write(&uvm->lock);
    if (mapping)
    {
        kmem_cache_free(g_uvmMappingCache, mapping);
        fp->f_mapping = 0;
    }

    if (vaspace)
    {
        uvm_unref_counter_container(vaspace->pCounterContainer);
        uvm_unref_event_container(vaspace->pEventContainer);
    }
    uvmfull_vaspace_destroy(inode, fp);
    return status;
}

NV_STATUS uvmfull_vaspace_destroy(struct inode *inode, struct file *fp)
{
    unsigned i;
    NvU64 gpuMask = 0;
    NvU64 faultBufferGpuMask = 0;
    NV_STATUS status = NV_OK;
    UvmVaspace *vaspace = NULL;
    UvmGpu *pGpu = NULL;
    UvmGlobalState *uvm = uvmfull_get_global();

    down_read(&uvm->lock);
    // TODO: query if fp/pid is registered with the driver.
    vaspace = fp->private_data;
    if (!vaspace)
    {
        up_read(&uvm->lock);
        return NV_ERR_INVALID_REQUEST;
    }

    down_read(&vaspace->lock);
    if (vaspace->filep != fp)
    {
        up_read(&vaspace->lock);
        up_read(&uvm->lock);
        return NV_ERR_INVALID_REQUEST;
    }
    gpuMask = vaspace->activeGpuMask;
    up_read(&vaspace->lock);
    up_read(&uvm->lock);

    // Now take fault buffer locks for all active gpus if applicable
    for (i=0; i < UVM_FULL_MAX_GPUS; i++)
    {
        if (gpuMask & NVBIT64(i))
        {
            // Check if gpu is fault capable
            pGpu = uvmfull_intr_gpuid_to_active_gpu(i);
            if (!pGpu)
            {
                // check if gpu is alive even though not fault capable
                pGpu = uvmfull_gpuid_to_active_gpu(i);
                if (!pGpu)
                    gpuMask &= ~(NVBIT64(i));
            }
            else
            {
                faultBufferGpuMask |= NVBIT64(i);
                down_write(&pGpu->faultBufferInfo.lock);
            }
        }
    }

    // Now take the global and vaspace locks
    down_write(&uvm->lock);
    down_write(&vaspace->lock);

    // remove the entry from the device va space list
    list_del(&vaspace->vaSpaceNode);

    // Stop all channel activity on all gpu vaspaces.
    for (i=0; i < UVM_FULL_MAX_GPUS; i++)
    {
        if (gpuMask & NVBIT64(i))
        {
            UvmGpuVaspace *gpuVaspace = &vaspace->gpuVaspaces[i];
            UvmClientChannel *channel = NULL;
            while (!list_empty(&gpuVaspace->channels))
            {
                channel = list_first_entry(&gpuVaspace->channels,
                                   UvmClientChannel,
                                   node);
                status = _gpu_vaspace_remove_channel(vaspace, gpuVaspace,
                                                channel->hChannel);
                // ? what to do if channel removal failed. (wait for tracker)
            }
        }
    }

    // Unmap and reclaim all memory in this process
    down_write(&vaspace->rmMemoryMapLock);
    uvmfull_vat_destroy(&vaspace->vat);
    uvm_hashmap_destroy(&vaspace->hRmMemoryMap);
    up_write(&vaspace->rmMemoryMapLock);

    // loop through all gpu va spaces and delete each one of them
    for (i=0; i < UVM_FULL_MAX_GPUS; i++)
    {
        if (gpuMask & NVBIT64(i))
            _gpu_vaspace_destroy(vaspace, i);
    }

    uvm_unref_counter_container(vaspace->pCounterContainer);
    uvm_unref_event_container(vaspace->pEventContainer);

    up_write(&vaspace->lock);
    fp->private_data = 0;
    kmem_cache_free(g_uvmVaspaceCache, vaspace);
    if (fp->f_mapping)
    {
        kmem_cache_free(g_uvmMappingCache, fp->f_mapping);
        fp->f_mapping = 0;
    }

    up_write(&uvm->lock);
    // Drop the fault buffer locks for all active gpus if applicable
    for (i = UVM_FULL_MAX_GPUS; i > 0; i--)
    {
        if (faultBufferGpuMask & NVBIT64(i-1))
        {
            pGpu = uvmfull_gpuid_to_gpu(i-1);
            up_write(&pGpu->faultBufferInfo.lock);
        }
    }

    // Finally unref all active gpus
    for (i=0; i < UVM_FULL_MAX_GPUS; i++)
    {
        if (NVBIT64(i) & gpuMask)
            uvmfull_stop_gpuid(i);
    }
    return status;
}

NvU64 uvmfull_instptr_key(NvU64 instptr, UvmFaultInstLoc loc)
{
    NvU64 location;
    NvU64 key;
    location = (loc == uvmFaultInstLoc_vidmem) ? INSTPTR_KEY_LOC_VIDMEM:
                                                 INSTPTR_KEY_LOC_SYSMEM;
    key = instptr & PAGE_MASK_4K;
    key >>= (PAGE_SHIFT_4K - 1);
    key |= location;
    return key;
}

// caller needs to have taken the device lock.
UvmVaspace *uvmfull_get_vaspace_from_instptr(UvmGpuId gpuid, NvU64 instptr,
                                             UvmFaultInstLoc loc)
{
    UvmClientChannel    *uvmChannel;
    UvmGpu              *pGpu;
    UvmHashEntry        *entry = NULL;
    UvmGlobalState      *uvm = uvmfull_get_global();
    NvU64               key = uvmfull_instptr_key(instptr, loc);

    pGpu = &uvm->uvmGpus[gpuid];

    entry = uvm_hashmap_query(&pGpu->instPtrMap, key);
    if (entry == NULL)
    {
        // Failed to find a channel associated with this inst pointer
        return NULL;
    }
    uvmChannel = container_of(entry, UvmClientChannel, instptrMapEntry);
    return uvmChannel->vaspace;
}

static void
_cleanup_rm_memory_region(UvmVaspace *vaspace, UvmRmMemoryRegion *rmMemRegion)
{
    NvU32 i;
    for (i = 0; i < UVM_FULL_MAX_GPUS; i++)
    {
        if (rmMemRegion->hRmMemoryDups[i])
        {
            if (vaspace->gpuVaspaces[i].client.bValid)
            {
                nvUvmInterfaceFreeDupedHandle(
                                vaspace->gpuVaspaces[i].client.dupGpuVaspace,
                                rmMemRegion->hRmMemoryDups[i]);
                rmMemRegion->hRmMemoryDups[i] = 0;
            }
        }
    }
    kmem_cache_free(g_uvmRmMemoryRegionCache, rmMemRegion);
}

NV_STATUS uvmfull_map_rm_allocation(struct file* filp, NvU64 base,
                            NvU64 length, UvmGpuMappingAttributes *perGpuAttribs,
                            NvLength attribCount, NvHandle hClientFromRM,
                            NvHandle hMemoryFromRm)
{
    NV_STATUS status;
    UvmVaspace *vaspace;
    NvU32 i;
    UvmGpuVaspace *gpuVaspace = NULL;
    UvmRmMemory *rmMem = NULL;
    UvmGpuSurfaceMappingInfo surfMapInfo;
    NvU8 gpuIds[UVM_FULL_MAX_GPUS];
    NvU64 pteTemplate = 0;
    UvmHashEntry *entry = NULL;
    UvmRmMemoryRegion *region = NULL;
    NvBool bFirstMapping = NV_FALSE;

    UvmGlobalState *uvm = uvmfull_get_global();
    if (attribCount != 1)
        return NV_ERR_INVALID_ARGUMENT;

    if (base & (~PAGE_MASK_4K))
        return NV_ERR_INVALID_ARGUMENT;

    down_read(&uvm->lock);

    vaspace = (UvmVaspace*)filp->private_data;
    if (!vaspace)
    {
        status = NV_ERR_INVALID_ARGUMENT;
        goto cleanup_one_lock;
    }

    down_read(&vaspace->lock);
    if (vaspace->filep != filp)
    {
        status = NV_ERR_INVALID_ARGUMENT;
        goto cleanup_two_locks;
    }

    // Check if a client was previous registered
    if (!vaspace->bValidRmClient ||
         vaspace->hRmClient != hClientFromRM)
    {
        status = NV_ERR_INVALID_ARGUMENT;
        goto cleanup_two_locks;
    }

    // Lock the rmMemory hashmap
    down_write(&vaspace->rmMemoryMapLock);

    // Basic sanity checks passed.
    // Find the rmMemory or just create it in the hashmap.
    entry = uvm_hashmap_query(&vaspace->hRmMemoryMap, hMemoryFromRm);
    if (entry)
    {
        rmMem = container_of(entry, UvmRmMemory, hRmMemoryMapEntry);
        UVM_PANIC_ON(rmMem->hRmMemory != hMemoryFromRm);
    }
    else
    {
        rmMem = kmem_cache_zalloc(g_uvmRmMemoryCache, NV_UVM_GFP_FLAGS);
        if (rmMem == NULL)
        {
            status = NV_ERR_NO_MEMORY;
            goto cleanup_map_unlock;
        }
        INIT_LIST_HEAD(&rmMem->rmMemRegions);
        bFirstMapping = NV_TRUE;
    }

    // Now create a region for this particular mapping call.
    region = kmem_cache_zalloc(g_uvmRmMemoryRegionCache, NV_UVM_GFP_FLAGS);
    if (region == NULL)
    {
        status = NV_ERR_NO_MEMORY;
        goto cleanup_map_unlock;
    }

    // Now create dups of memory handle into all the interested gpuvaspaces.
    for (i = 0; i < attribCount; i++)
    {
        NvHandle hDupMem = 0;
        UvmGpu *gpu = uvmfull_uuid_to_active_gpu(&perGpuAttribs[i].gpuUuid);
        if (!gpu)
        {
            status = NV_ERR_INVALID_ARGUMENT;
            goto cleanup_map_unlock;
        }
        gpuVaspace = &vaspace->gpuVaspaces[gpu->gpuId];
        if (gpuVaspace->uvmVaspace != vaspace || !gpuVaspace->client.bValid)
        {
            status = NV_ERR_INVALID_ARGUMENT;
            goto cleanup_map_unlock;
        }

        status = nvUvmInterfaceDupMemory(gpuVaspace->client.dupGpuVaspace,
                                         hClientFromRM,
                                         hMemoryFromRm,
                                         &hDupMem,
                                         NULL);
        if (status != NV_OK)
        {
            status = NV_ERR_INVALID_ARGUMENT;
            goto cleanup_map_unlock;
        }
        region->hRmMemoryDups[gpu->gpuId] = hDupMem;
        gpuIds[i] = (NvU8)gpu->gpuId;
    }

    surfMapInfo.hSourceClient = hClientFromRM;
    surfMapInfo.hSourceMemory = hMemoryFromRm;
    surfMapInfo.mappingOffset = 0;
    surfMapInfo.mappingLength = UVM_ALIGN_UP(length, PAGE_SIZE_4K);
    surfMapInfo.pageCount = surfMapInfo.mappingLength/PAGE_SIZE_4K;
    surfMapInfo.pteTemplate = (NvU8*)&pteTemplate;
    surfMapInfo.pteArray = vmalloc(surfMapInfo.pageCount * sizeof(NvU64));
    if (!surfMapInfo.pteArray)
    {
        status = NV_ERR_NO_MEMORY;
        goto cleanup_map_unlock;
    }

    // Get pteTemplate for the memory by using any one gpuvaspace
    status = nvUvmInterfaceGetSurfaceMapInfo(gpuVaspace->client.dupGpuVaspace,
                                             &surfMapInfo);
    if (status != NV_OK)
    {
        status = NV_ERR_INVALID_ARGUMENT;
        vfree(surfMapInfo.pteArray);
        goto cleanup_map_unlock;
    }

    status = uvmfull_map_gpus_unmanaged_va(vaspace, base,
                                            gpuIds, &surfMapInfo, perGpuAttribs,
                                            attribCount);
    // Free the pfn array in any case
    vfree(surfMapInfo.pteArray);
    if (status != NV_OK)
    {
        status = NV_ERR_INVALID_ARGUMENT;
        goto cleanup_map_unlock;
    }
    // Now insert the rm memory tracker or just add a region.
    if (bFirstMapping)
    {
        rmMem->hRmMemory = hMemoryFromRm;
        uvm_hashmap_insert(&vaspace->hRmMemoryMap, &rmMem->hRmMemoryMapEntry,
                                        rmMem->hRmMemory);
    }
    // Also insert the region in the tracking list
    region->vaStart = base;
    region->length = surfMapInfo.mappingLength;
    list_add(&region->rmMemRegionNode, &rmMem->rmMemRegions);

cleanup_map_unlock:
    if (status != NV_OK && region)
        _cleanup_rm_memory_region(vaspace, region);
    if (status != NV_OK && bFirstMapping)
    {
        uvm_hashmap_remove(&vaspace->hRmMemoryMap, &rmMem->hRmMemoryMapEntry);
        kmem_cache_free(g_uvmRmMemoryCache, rmMem);
    }
    up_write(&vaspace->rmMemoryMapLock);
cleanup_two_locks:
    up_read(&vaspace->lock);
cleanup_one_lock:
    up_read(&uvm->lock);
    return status;
}

NV_STATUS uvmfull_enable_peer_on_vaspace(UvmVaspace *vaspace,
                                         UvmGpuId gpuId1,
                                         UvmGpuId gpuId2)
{
    NvU64 uniqueLinkIdx, uniqueLinkMask;
    NvBool validPeer = NV_FALSE;

    UVM_PANIC_ON(!vaspace);

    uniqueLinkIdx = GET_UNIQUE_PEER_LINK_IDX(gpuId1, gpuId2);
    uniqueLinkMask = vaspace->validPeerMask[uniqueLinkIdx / UVM_FULL_MAX_GPUS];

    // If duplicate register, flag error
    validPeer = uniqueLinkMask & NVBIT64(uniqueLinkIdx % UVM_FULL_MAX_GPUS);
    if (validPeer)
        return NV_ERR_INVALID_REQUEST;

    // Set valid reference for these peers in process context
    vaspace->validPeerMask[uniqueLinkIdx / UVM_FULL_MAX_GPUS] |= 
            NVBIT64(uniqueLinkIdx % UVM_FULL_MAX_GPUS);

    // Loopback mode is a global mode & is allowed only on single GPU
    if (gpuId1 == gpuId2)
    {
        if (!vaspace->bLoopback)
        {
            vaspace->bLoopback = NV_TRUE;
        }
        else
        {
            WARN_ON(vaspace->bLoopback);
            return NV_ERR_INVALID_REQUEST;
        }
    }

    return NV_OK;
}

NV_STATUS uvmfull_disable_peer_on_vaspace(UvmVaspace *vaspace,
                                          UvmGpuId gpuId1,
                                          UvmGpuId gpuId2)
{
    NvU64 uniqueLinkIdx, uniqueLinkMask;
    NvBool validPeer = NV_FALSE;

    UVM_PANIC_ON(!vaspace);

    uniqueLinkIdx = GET_UNIQUE_PEER_LINK_IDX(gpuId1, gpuId2);
    uniqueLinkMask = vaspace->validPeerMask[uniqueLinkIdx / UVM_FULL_MAX_GPUS];

    // If not registered before flag error as this is used for refCounting also
    validPeer = uniqueLinkMask & NVBIT64(uniqueLinkIdx % UVM_FULL_MAX_GPUS);
    if (!validPeer)
        return NV_ERR_INVALID_REQUEST;

    // UnSet valid reference for these peers in process context
    vaspace->validPeerMask[uniqueLinkIdx / UVM_FULL_MAX_GPUS] &= 
            ~NVBIT64(uniqueLinkIdx % UVM_FULL_MAX_GPUS);

    if (gpuId1 == gpuId2)
        vaspace->bLoopback = NV_FALSE;
    return NV_OK;
}

NvBool uvmfull_vaspace_is_peeraccess_valid(UvmVaspace *vaspace,
                                           UvmGpuId gpuId1,
                                           UvmGpuId gpuId2)
{
    NvU64 uniqueLinkIdx, uniqueLinkMask;

    UVM_PANIC_ON(!vaspace);

    uniqueLinkIdx = GET_UNIQUE_PEER_LINK_IDX(gpuId1, gpuId2);
    uniqueLinkMask = vaspace->validPeerMask[uniqueLinkIdx / UVM_FULL_MAX_GPUS];

    return ((uniqueLinkMask & NVBIT64(uniqueLinkIdx % UVM_FULL_MAX_GPUS)) ||
            vaspace->bLoopback);
}

NV_STATUS uvmfull_unmap_rm_allocation(struct file* filp, NvU64 base)
{
    // query the trie for va being the start of an rm allocation
    //   Check flag. If start, cpuPfn would be a valid handle value.
    // If yes, release the desc. grab the rm hashmap lock, lookup range,
    // for each 2mb range; lookup desc and unmap/free gpu memory.
    return NV_ERR_NOT_SUPPORTED;
}
