/*******************************************************************************
    Copyright (c) 2014-2015 NVIDIA Corporation

    Permission is hereby granted, free of charge, to any person obtaining a copy
    of this software and associated documentation files (the "Software"), to
    deal in the Software without restriction, including without limitation the
    rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
    sell copies of the Software, and to permit persons to whom the Software is
    furnished to do so, subject to the following conditions:

        The above copyright notice and this permission notice shall be
        included in all copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
    IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
    FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
    THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
    LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
    FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
    DEALINGS IN THE SOFTWARE.

*******************************************************************************/

#include "uvm_full.h"
#include "uvm_common.h"
#include "uvm_utils.h"
#include "nv_uvm_interface.h"
#include "uvm_full_device_mgmt.h"
#include "uvm_full_ctx_mgmt.h"
#include "uvm_hashmap.h"
#include "ctrl2080mc.h"
#include "uvm_full_pagetbl_mgmt.h"
#include "uvm_full_fault_handler.h"
#include "uvm_full_identity_map.h"

#define INSPTR_HASHMAP_DEFAULT_SLOTS_LOG2     (9)

// statics defined here
static UvmGlobalState gUvmState;

// Cache for UVM Bottom Half Work structs
static struct kmem_cache *g_uvmPendingFaultCache   __read_mostly = NULL;

NV_STATUS
uvmfull_device_mgmt_slab_init(void)
{
    g_uvmPendingFaultCache = NV_KMEM_CACHE_CREATE("uvm_gpu_pending_faults",
                                                  UvmPendingFault);
    if (!g_uvmPendingFaultCache)
        return NV_ERR_NO_MEMORY;

    return NV_OK;
}

void uvmfull_device_mgmt_slab_destroy(void)
{
    if (g_uvmPendingFaultCache)
        kmem_cache_destroy_safe(&g_uvmPendingFaultCache);
}

void uvmfull_init_global_state()
{
    NvU32 i;
    UvmGlobalState *uvm = uvmfull_get_global();
    UvmGpu *gpu = NULL;

    memset(uvm, 0x0, sizeof(UvmGlobalState));
    init_rwsem(&uvm->lock);
    spin_lock_init(&uvm->uuidMap.lock);
    INIT_LIST_HEAD(&uvm->uvmVaspaces);
    INIT_LIST_HEAD(&uvm->uvmMirrors);

    for (i = 0; i < UVM_FULL_MAX_GPUS; i++)
    {
        gpu = &uvm->uvmGpus[i];
        init_rwsem(&gpu->faultBufferInfo.lock);
        // Init page LISTs
        INIT_LIST_HEAD(&gpu->gpuPageList);
        INIT_LIST_HEAD(&gpu->cpuPageList);
    }
}

// Caller should hold the device lock before calling this.
NvBool uvmfull_is_gpu_fault_capable(UvmGpuId gpuId)
{
    NvU32 arch = gUvmState.uvmGpus[gpuId].gpuArch;

    if (arch >= NV2080_CTRL_MC_ARCH_INFO_ARCHITECTURE_GP100 &&
        arch < NV2080_CTRL_MC_ARCH_INFO_ARCHITECTURE_T13X)
        return NV_TRUE;

    return NV_FALSE;
}

UvmGlobalState* uvmfull_get_global()
{
    return &gUvmState;
}

NvS32 uvm_gpu_uuid_cmp(UvmGpuUuid *u1, UvmGpuUuid *u2)
{
    return memcmp(u1, u2, sizeof(UvmGpuUuid));
}

// Caller needs to take the spinlock before calling
static inline UvmGpuId _get_free_uvmId(void)
{
    return gUvmState.uuidMap.nEntries;
}

// This call will take the map spinlock.
static UvmGpuId _insert_uuid_in_map(UvmGpuUuid *uuid)
{
    UvmGpuId freeId;
    UvmUuidMap *map = &gUvmState.uuidMap;
    unsigned long oldIrq;

    spin_lock_irqsave(&map->lock, oldIrq);
    freeId = _get_free_uvmId();
    UVM_PANIC_ON(freeId >= UVM_FULL_MAX_GPUS);
    map->uvmUuidMap[freeId].uvmId = freeId;
    memcpy(&(map->uvmUuidMap[freeId].uuid), uuid, sizeof(UvmGpuUuid));
    map->nEntries++;
    spin_unlock_irqrestore(&map->lock, oldIrq);

    return freeId;
}

// This function assumes the spinlock for map is taken by the caller.
static UvmGpuId _gpu_uuid_to_id(UvmGpuUuid *pGpuUuid)
{
    size_t i;
    UvmGpuId id = UVM_FULL_MAX_GPUS;
    UvmUuidMap *map = &gUvmState.uuidMap;

    // TODO: Change this to a bst lookup on a sorted uuid map.
    for(i=0; i<map->nEntries; i++)
    {
        if (!uvm_gpu_uuid_cmp(pGpuUuid, &map->uvmUuidMap[i].uuid))
        {
            id = map->uvmUuidMap[i].uvmId;
            break;
        }
    }
    return id;
}

static inline UvmUuidMap* _get_uuid_map(void)
{
    return &gUvmState.uuidMap;
}

// Returns NV_TRUE if gpu ref toggles to non zero
static NvBool _ref_gpuid(UvmGpuId id)
{
    NvBool bGoingToNonZero = NV_FALSE;
    UvmUuidMap *map = _get_uuid_map();
    unsigned long oldIrq;

    UVM_PANIC_ON(id >= UVM_FULL_MAX_GPUS);
    spin_lock_irqsave(&map->lock, oldIrq);

    if (map->uvmUuidMap[id].ref == 0)
        bGoingToNonZero = NV_TRUE;

    map->uvmUuidMap[id].ref++;
    spin_unlock_irqrestore(&map->lock, oldIrq);
    return bGoingToNonZero;
}

// Returns NV_TRUE if gpu ref toggles to zero.
static NvBool _unref_gpuid(UvmGpuId id)
{
    NvBool bGoingToZero = NV_FALSE;
    UvmUuidMap *map = _get_uuid_map();
    unsigned long oldIrq;

    UVM_PANIC_ON(id >= UVM_FULL_MAX_GPUS);

    spin_lock_irqsave(&map->lock, oldIrq);
    WARN_ON(map->uvmUuidMap[id].ref == 0);
    if (map->uvmUuidMap[id].ref)
    {
        map->uvmUuidMap[id].ref--;
        bGoingToZero = map->uvmUuidMap[id].ref == 0;
    }
    spin_unlock_irqrestore(&map->lock, oldIrq);
    return bGoingToZero;
}

static void _toggle_gpu_intr_service(UvmGpu *gpu, NvBool bEnable)
{
    UvmGpuId id = gpu->gpuId;
    UvmUuidMap *map = _get_uuid_map();
    unsigned long oldIrq;

    UVM_PANIC_ON(id >= UVM_FULL_MAX_GPUS);
    spin_lock_irqsave(&map->lock, oldIrq);
    map->uvmUuidMap[id].bIntrServiceable = bEnable;
    spin_unlock_irqrestore(&map->lock, oldIrq);
}

UvmGpu* uvmfull_gpuid_to_gpu(UvmGpuId id)
{
    return &gUvmState.uvmGpus[id];
}

UvmGpu* uvmfull_gpuid_to_active_gpu(UvmGpuId id)
{
    UvmGpu *gpu = NULL;
    UvmUuidMap *map = _get_uuid_map();
    unsigned long oldIrq;

    spin_lock_irqsave(&map->lock, oldIrq);

    if (map->uvmUuidMap[id].ref)
        gpu = &gUvmState.uvmGpus[id];

    spin_unlock_irqrestore(&map->lock, oldIrq);
    return gpu;
}

UvmGpu* uvmfull_uuid_to_gpu(UvmGpuUuid *pGpuUuid)
{
    UvmGpuId id;
    UvmGpu* gpu = NULL;
    UvmUuidMap* map = _get_uuid_map();
    unsigned long oldIrq;

    spin_lock_irqsave(&map->lock, oldIrq);
    id = _gpu_uuid_to_id(pGpuUuid);
    if (id < UVM_FULL_MAX_GPUS)
    {
        gpu = &gUvmState.uvmGpus[id];
    }
    spin_unlock_irqrestore(&map->lock, oldIrq);
    return gpu;
}

UvmGpu* uvmfull_intr_gpuid_to_active_gpu(UvmGpuId id)
{
    UvmGpu* gpu = NULL;
    UvmUuidMap *map = _get_uuid_map();
    unsigned long oldIrq;

    spin_lock_irqsave(&map->lock, oldIrq);

    if (id < UVM_FULL_MAX_GPUS &&
        map->uvmUuidMap[id].ref &&
        map->uvmUuidMap[id].bIntrServiceable)
        gpu = &gUvmState.uvmGpus[id];

    spin_unlock_irqrestore(&map->lock, oldIrq);
    return gpu;
}

UvmGpu* uvmfull_uuid_to_active_gpu(UvmGpuUuid *uuid)
{
    UvmGpuId id;
    UvmGpu* gpu = NULL;
    UvmUuidMap *map = _get_uuid_map();
    unsigned long oldIrq;

    spin_lock_irqsave(&map->lock, oldIrq);
    id = _gpu_uuid_to_id(uuid);

    if (id < UVM_FULL_MAX_GPUS && map->uvmUuidMap[id].ref)
        gpu = &gUvmState.uvmGpus[id];

    spin_unlock_irqrestore(&map->lock, oldIrq);
    return gpu;
}

UvmGpu* uvmfull_intr_uuid_to_active_gpu(UvmGpuUuid *uuid)
{
    UvmGpuId id;
    UvmGpu* gpu = NULL;
    UvmUuidMap *map = _get_uuid_map();
    unsigned long oldIrq;

    spin_lock_irqsave(&map->lock, oldIrq);
    id = _gpu_uuid_to_id(uuid);

    if (id < UVM_FULL_MAX_GPUS &&
        map->uvmUuidMap[id].ref &&
        map->uvmUuidMap[id].bIntrServiceable)
        gpu = &gUvmState.uvmGpus[id];

    spin_unlock_irqrestore(&map->lock, oldIrq);
    return gpu;
}

NV_STATUS uvmfull_enable_peer_access(struct file* filp,
                                     UvmGpuUuid *gpuUuid1,
                                     UvmGpuUuid *gpuUuid2)
{
    UvmGpu *pGpu1, *pGpu2;
    UvmVaspace *pVaspace;
    UvmGpuP2PCapsParams opsPeerCaps;
    UvmGpuPeerCaps *pPeerCaps;
    NvU64 uniqueLinkIdx;
    NV_STATUS status = NV_OK;
    UvmGlobalState *uvm = uvmfull_get_global();

    // Enable both GPUs in case they are not enabled
    status = uvmfull_start_gpu(gpuUuid1);
    if (status != NV_OK)
        return status;

    status = uvmfull_start_gpu(gpuUuid2);
    if (status != NV_OK)
    {
        uvmfull_stop_gpu(gpuUuid1);
        return status;
    }

    pGpu1 = uvmfull_uuid_to_active_gpu(gpuUuid1);
    pGpu2 = uvmfull_uuid_to_active_gpu(gpuUuid2);
    WARN_ON(!pGpu1 || !pGpu2);

    down_write(&uvm->lock);

    // Set valid bit in process context
    pVaspace = (UvmVaspace*)filp->private_data;
    if (!pVaspace)
    {
        status =  NV_ERR_INVALID_REQUEST;
        goto cleanup;
    }

    down_write(&pVaspace->lock);
    status = uvmfull_enable_peer_on_vaspace(pVaspace,
                                            pGpu1->gpuId, pGpu2->gpuId);
    up_write(&pVaspace->lock);
    if (status != NV_OK)
        goto cleanup;

    // Set peer caps for link gpu1->gpu2 if not already set
    uniqueLinkIdx = GET_UNIQUE_PEER_LINK_IDX(pGpu1->gpuId, pGpu2->gpuId);

    // Check if we already have a peer set between these GPUs due to another
    // process
    if (uvm->uvmPeerLinkRef[uniqueLinkIdx] == 0)
    {
        pPeerCaps = &uvm->uvmPeerCaps[pGpu1->gpuId][pGpu2->gpuId]; 
        memset(&opsPeerCaps, 0, sizeof(opsPeerCaps));
        opsPeerCaps.pUuids[0] = (NvU8 *) gpuUuid1;
        opsPeerCaps.pUuids[1] = (NvU8 *) gpuUuid2;
        status = nvUvmInterfaceGetP2PCaps(&opsPeerCaps);
        if (status != NV_OK)
            goto cleanup;
        pPeerCaps->gpuPeerId = opsPeerCaps.peerIds[0];
        pPeerCaps->writeSupported  = opsPeerCaps.writeSupported;
        pPeerCaps->readSupported  = opsPeerCaps.readSupported;
        pPeerCaps->propSupported  = opsPeerCaps.propSupported;
        pPeerCaps->nvlinkSupported  = opsPeerCaps.nvlinkSupported;
        pPeerCaps->atomicSupported  = opsPeerCaps.atomicSupported;
        pPeerCaps->peerMapValid = NV_TRUE;

        // Set peer caps for link between gpu2->gpu1
        pPeerCaps = &uvm->uvmPeerCaps[pGpu2->gpuId][pGpu1->gpuId];
        pPeerCaps->gpuPeerId = opsPeerCaps.peerIds[1];
        pPeerCaps->writeSupported  = opsPeerCaps.writeSupported;
        pPeerCaps->readSupported  = opsPeerCaps.readSupported;
        pPeerCaps->propSupported  = opsPeerCaps.propSupported;
        pPeerCaps->nvlinkSupported  = opsPeerCaps.nvlinkSupported;
        pPeerCaps->atomicSupported  = opsPeerCaps.atomicSupported;
        pPeerCaps->peerMapValid = NV_TRUE;
    }

    // Bump up the refcount on unique peer link
    uvm->uvmPeerLinkRef[uniqueLinkIdx]++;

cleanup:
    up_write(&uvm->lock);

    if (status != NV_OK)
    {
        // Decrement refCount on GPUs if we have failed this call.
        uvmfull_stop_gpu(gpuUuid1);
        uvmfull_stop_gpu(gpuUuid2);
    }

    return status;
}

NV_STATUS uvmfull_disable_peer_access(struct file* filp,
                                      UvmGpuUuid *gpuUuid1,
                                      UvmGpuUuid *gpuUuid2)
{
    UvmGpu *pGpu1, *pGpu2;
    UvmVaspace *pVaspace;
    UvmGpuPeerCaps *pPeerCaps;
    NvU64 uniqueLinkIdx;
    NV_STATUS status = NV_OK;
    UvmGlobalState *uvm = uvmfull_get_global();

    pGpu1 = uvmfull_uuid_to_active_gpu(gpuUuid1);
    pGpu2 = uvmfull_uuid_to_active_gpu(gpuUuid2);

    if (!pGpu1 || !pGpu2)
        return NV_ERR_INVALID_ARGUMENT;

    down_write(&uvm->lock);

    // Unset valid bit in process context
    pVaspace = (UvmVaspace*)filp->private_data;
    if (!pVaspace)
    {
        status =  NV_ERR_INVALID_REQUEST;
        goto cleanup;
    }

    down_write(&pVaspace->lock);
    status = uvmfull_disable_peer_on_vaspace(pVaspace,
                                             pGpu1->gpuId, pGpu2->gpuId);
    up_write(&pVaspace->lock);
    if (status != NV_OK)
        goto cleanup;

    // Drop the reference from unique peer link
    uniqueLinkIdx = GET_UNIQUE_PEER_LINK_IDX(pGpu1->gpuId, pGpu2->gpuId);

    if (--uvm->uvmPeerLinkRef[uniqueLinkIdx] == 0)
    {
        // Un-Set peer caps for GPU1 and GPU2 if this is the last
        // process keeping them alive
        pPeerCaps = &uvm->uvmPeerCaps[pGpu1->gpuId][pGpu2->gpuId];
        memset(pPeerCaps, 0, sizeof(UvmGpuPeerCaps));

        pPeerCaps = &uvm->uvmPeerCaps[pGpu2->gpuId][pGpu1->gpuId];
        memset(pPeerCaps, 0, sizeof(UvmGpuPeerCaps));
    }

    up_write(&uvm->lock);

    //
    // Disable both GPUs in case this is the last process holding refCount
    // Don't bother about failure in stopGpu call as we still need to decrement
    // the refCount on both GPUs.
    //
    uvmfull_stop_gpu(gpuUuid1);
    uvmfull_stop_gpu(gpuUuid2);
    return status;

cleanup:
    up_write(&uvm->lock);
    return status;
}

UvmGpu* uvmfull_peerid_to_gpu(UvmGpuId fromGpuId, NvU32 peerId)
{
    UvmGpuPeerCaps *pPeerCaps = NULL;
    NvU32 i = 0;
    UvmGlobalState *uvm = uvmfull_get_global();

    UVM_PANIC_ON(fromGpuId >= UVM_FULL_MAX_GPUS ||
                 peerId >= UVM_MAX_GPU_PEER_ID);

    // Get the gloabl structure and see what's which GPUID maps to provided peer
    // id w.r.t source GPU Id
    for (i = 0; i < UVM_FULL_MAX_GPUS; i++)
    {
        pPeerCaps = &uvm->uvmPeerCaps[fromGpuId][i];
        if (pPeerCaps->peerMapValid && pPeerCaps->gpuPeerId == peerId)
            return uvmfull_gpuid_to_gpu(i);
    }
    return NULL;
}

// NOTE: The function takes the global lock in write mode.
// If there is no mapping for this uuid; this is the first time this device is
// being started. Assign an uvm id and activate the corresponding
// gpu array entry.
NV_STATUS uvmfull_start_gpu(UvmGpuUuid *pGpuUuid)
{
    UvmGpu *gpu;
    UvmGpuFbInfo gpuFbInfo;
    NvBool bRegisteredWithRm = NV_FALSE;
    UvmGpuInfo gpuInfo = {{0}};
    NV_STATUS status = NV_OK;
    struct page* cpuPage = NULL;
    UvmGlobalState *uvm = uvmfull_get_global();

    down_write(&uvm->lock);
    gpu = uvmfull_uuid_to_gpu(pGpuUuid);
    if (!gpu) // UUID seen for the first time.
    {
        UvmGpuId id;
        status = nvUvmInterfaceRegisterGpu(pGpuUuid);
        if (status != NV_OK)    // fake or incorrect uuid
            goto release_lock;

        id = _insert_uuid_in_map(pGpuUuid);
        if (id >= UVM_FULL_MAX_GPUS)
        {
            status = NV_ERR_INVALID_REQUEST;
            nvUvmInterfaceUnregisterGpu(pGpuUuid);
            goto release_lock;
        }

        gpu = &uvm->uvmGpus[id];
        gpu->gpuId = id;
        memcpy(&gpu->gpuUuid, pGpuUuid, sizeof(*pGpuUuid));
        // Fetch this GPU's architecture
        status = nvUvmInterfaceGetGpuInfo(&gpu->gpuUuid, &gpuInfo);
        if (status != NV_OK)
        {
            nvUvmInterfaceUnregisterGpu(pGpuUuid);
            goto release_lock;
        }

        gpu->gpuArch = gpuInfo.gpuArch;
        gpu->faultBufferClass = gpuInfo.faultBufferClass;
        // TODO: Get gpu caps.
        gpu->bValid = NV_TRUE;
        bRegisteredWithRm = NV_TRUE;
    }

    // Increment refcount and see if it was already alive
    if (NV_FALSE == _ref_gpuid(gpu->gpuId))
        goto release_lock;        // The gpu was already started and registered.

    if (!bRegisteredWithRm)
    {
        status = nvUvmInterfaceRegisterGpu(&gpu->gpuUuid);
        if (status != NV_OK)
            goto cleanup;
    }

    // Create a RM session on first touch.
    if (!gpu->gpuSession)
    {
        status = nvUvmInterfaceSessionCreate(&gpu->gpuSession);
        if (status != NV_OK)
        {
            UVM_ERR_PRINT("ERROR:Failed session create for uvm.n\n");
            goto cleanup;
        }
    }

    if (!gpu->gpuHandle)
    {
        // Allocate a temporary va space in order to obtain the fault buffer
        status = nvUvmInterfaceAddressSpaceCreate(gpu->gpuSession,
                                                  &gpu->gpuUuid,
                                                  &gpu->gpuHandle,
                                                  0,
                                                  0);// Let RM pick the vaend
        if (status != NV_OK)
        {
            UVM_ERR_PRINT("ERROR:Failed address space create for gpu.n\n");
            goto cleanup;
        }
    }

    if( !gpu->channelManager )
    {
        status = uvm_create_channel_manager(pGpuUuid, &gpu->channelManager);
        if (status != NV_OK)
            goto cleanup;
        gpu->bUseMemops = NV_TRUE;
    }

    if (!gpu->instPtrMap.nSlots)
    {
        status = uvm_hashmap_init(
                          &gpu->instPtrMap, INSPTR_HASHMAP_DEFAULT_SLOTS_LOG2);
        if (status != NV_OK)
            goto cleanup;
    }

    memset(&gpuFbInfo, 0x0, sizeof(gpuFbInfo));
    status = nvUvmInterfaceGetFbInfo(gpu->gpuHandle, &gpuFbInfo);
    if (status != NV_OK)
        goto cleanup;

    gpu->bZeroFbGpu = gpuFbInfo.bZeroFb;
    if (!gpu->bZeroFbGpu)
    {
        NvU64 pfn = 0;
        NvU32 i;
        UvmPbOps *pbOps = NULL;
        UvmTracker tracker;
        UvmPushbuffer *pb = NULL;
        uvm_init_tracker(&tracker);

        status = uvmfull_allocate_phys_vid_mem(gpu, PAGE_SIZE_2M, &pfn);
        if (status != NV_OK)
            goto cleanup;
        gpu->zeroed2MPa = PFN_TO_PA(pfn);
        gpu->zeroPageAllocated = 0x1;

        // allocate zeroed cpu memory to use as src
        cpuPage = alloc_pages(NV_UVM_GFP_FLAGS | GFP_HIGHUSER | __GFP_ZERO, 0);
        if (!cpuPage)
            goto cleanup;
        pfn = page_to_pfn(cpuPage);

        pbOps = uvm_pbops_start_on_gpu(gpu->gpuId);
        if (!pbOps)
        {
            status = NV_ERR_GENERIC;
            goto cleanup;
        }
        pb = uvm_pbops_gpu_pb(pbOps, gpu->gpuId);
        if (!pb)
        {
            status = NV_ERR_GENERIC;
            goto cleanup;
        }
        for (i = 0; i < PAGE_SIZE_2M / PAGE_SIZE_4K; i++)
        {
            uvm_push_cpu_to_gpu_copy(pb, PFN_TO_PA(pfn),
                                    (NvU64)gpu->zeroed2MPa + i*PAGE_SIZE_4K,
                                    PAGE_SIZE_4K,
                                    NV_UVM_COPY_SRC_TYPE_PHYSICAL |
                                    NV_UVM_COPY_DST_TYPE_PHYSICAL);
        }
        uvm_pbops_submit(pbOps, gpu->gpuId, pb, &tracker);
        uvm_move_tracker(&tracker, pbOps->t);
        status = uvm_wait_for_tracker(&tracker);
        uvm_pbops_end(pbOps);
        __free_page(cpuPage);
        cpuPage = NULL;
        if (status != NV_OK)
            goto cleanup;
    }

    // TODO: register with RM for interrupts if fault_capable.
    if (uvmfull_is_gpu_fault_capable(gpu->gpuId))
    {
        // Setup Identity mapping
        // This is a synchronous (blocking) call

        status = uvmfull_create_identity_map(gpu);
        if (status != NV_OK)
            goto cleanup;

        status = uvmfull_init_gpu_faultbuffer(gpu);
        if (status != NV_OK)
            goto cleanup;
        _toggle_gpu_intr_service(gpu, NV_TRUE);
    }

    goto release_lock;

cleanup:
    if (gpu->zeroPageAllocated)
    {
        uvmfull_free_phys_vid_mem(gpu, PA_TO_PFN(gpu->zeroed2MPa));
        gpu->zeroed2MPa = 0;
        gpu->zeroPageAllocated = 0;
    }
    if (gpu->channelManager)
    {
        uvm_destroy_channel_manager(gpu->channelManager);
        gpu->channelManager = NULL;
    }
    uvm_hashmap_destroy(&gpu->instPtrMap);
    _unref_gpuid(gpu->gpuId);
    if (gpu->gpuHandle)
    {
        nvUvmInterfaceAddressSpaceDestroy(gpu->gpuHandle);
        gpu->gpuHandle = NULL;
    }
    if (gpu->gpuSession)
    {
        nvUvmInterfaceSessionDestroy(gpu->gpuSession);
        gpu->gpuSession = NULL;
    }
    if (cpuPage)
        __free_page(cpuPage);

    nvUvmInterfaceUnregisterGpu(&gpu->gpuUuid);

release_lock:
    up_write(&uvm->lock);
    return status;
}

static NV_STATUS _stop_gpu(UvmGpu *gpu)
{
    NV_STATUS status = NV_OK;
    if (!gpu)
        return NV_ERR_INVALID_OPERATION;

    // TODO: Wait for all activity to complete for all channels

    if (uvmfull_is_gpu_fault_capable(gpu->gpuId))
    {
        // Cleanup identity mapping
        uvmfull_destroy_identity_map(gpu);

        // Toggle uvm's notion of fault servicing to off.
        _toggle_gpu_intr_service(gpu, NV_FALSE);
        status = uvmfull_destroy_gpu_faultbuffer(gpu);
    }

    if (gpu->zeroPageAllocated)
    {
        uvmfull_free_phys_vid_mem(gpu, PA_TO_PFN(gpu->zeroed2MPa));
        gpu->zeroed2MPa = 0;
        gpu->zeroPageAllocated = 0;
    }

    if (gpu->channelManager)
    {
        uvm_destroy_channel_manager(gpu->channelManager);
        gpu->channelManager = NULL;
    }
    uvm_hashmap_destroy(&gpu->instPtrMap);

    if (gpu->gpuHandle)
    {
        nvUvmInterfaceAddressSpaceDestroy(gpu->gpuHandle);
        gpu->gpuHandle = NULL;
    }
    if (gpu->gpuSession)
    {
        nvUvmInterfaceSessionDestroy(gpu->gpuSession);
        gpu->gpuSession = NULL;
    }
    // Now unregister the gpu with RM.
    nvUvmInterfaceUnregisterGpu(&gpu->gpuUuid);

    return status;
}

// Caller should not hold the global lock
NV_STATUS uvmfull_stop_gpuid(UvmGpuId id)
{
    NV_STATUS status = NV_OK;
    UvmGlobalState *uvm = uvmfull_get_global();
    UvmGpu *gpu = NULL;
    NvBool bFbLockTaken = NV_FALSE;

    // Check if the gpu is fault capable; in which case take the faultbuffer
    // lock.
    gpu = uvmfull_intr_gpuid_to_active_gpu(id);
    if (!gpu)
    {
        // This might be a non faulting gpu afterall
        gpu = uvmfull_gpuid_to_active_gpu(id);
        if (!gpu)
            return NV_ERR_INVALID_ARGUMENT;
    }
    else
    {
        // Take the faultbuffer lock as we are going to query it.
        down_write(&gpu->faultBufferInfo.lock);
        bFbLockTaken = NV_TRUE;
    }

    down_write(&uvm->lock);
    if (gpu->bValid)
    {
        // No need to stop gpu if ref != 0. Ref takes spinlock.
        if(!_unref_gpuid(gpu->gpuId))
            status = NV_OK;
        else
        {
            status = _stop_gpu(gpu);
            WARN_ON(status != NV_OK);
        }
    }
    up_write(&uvm->lock);


    if (bFbLockTaken)
        up_write(&gpu->faultBufferInfo.lock);

    return status;
}

NV_STATUS uvmfull_stop_gpu(UvmGpuUuid *pGpuUuid)
{
    UvmGpu *gpu;
    NV_STATUS status = NV_OK;
    UvmGlobalState *uvm = uvmfull_get_global();
    NvBool bFbLockTaken = NV_FALSE;

    // Check if the gpu is fault capable; in which case take the faultbuffer
    // lock.
    gpu = uvmfull_intr_uuid_to_active_gpu(pGpuUuid);
    if (!gpu)
    {
        // This might be a non faulting gpu afterall
        gpu = uvmfull_uuid_to_active_gpu(pGpuUuid);
        if (!gpu)
            return NV_ERR_INVALID_ARGUMENT;
    }
    else
    {
        // Take the faultbuffer lock as we are going to query it.
        down_write(&gpu->faultBufferInfo.lock);
        bFbLockTaken = NV_TRUE;
    }

    down_write(&uvm->lock);
    if (gpu->bValid)
    {
        // No need to stop gpu if ref != 0. Ref takes spinlock.
        if(!_unref_gpuid(gpu->gpuId))
            status = NV_OK;
        else
        {
            status = _stop_gpu(gpu);
            WARN_ON(status != NV_OK);
        }
    }
    up_write(&uvm->lock);

    if (gpu->bTestZfodPfnValid)
    {
        UVM_PANIC_ON(gpu->bZeroFbGpu); // sysmem ZFOD page
        nvUvmInterfaceMemoryFreePa(gpu->gpuHandle,
                            PFN_TO_PA(gpu->testZfodPfn));
        gpu->testZfodPfn = 0;
        gpu->bTestZfodPfnValid = NV_FALSE;
    }

    if (bFbLockTaken)
        up_write(&gpu->faultBufferInfo.lock);

    return status;
}

NV_STATUS uvmfull_init_gpu_faultbuffer(UvmGpu *pGpu)
{
    UvmGpuFaultInfo uvmRmFaultBufferInfo;
    NV_STATUS       status;
    UvmGpuFaultBufferInfo *info = &pGpu->faultBufferInfo;

    down_write(&info->lock);
    status = nvUvmInterfaceInitFaultInfo(pGpu->gpuHandle,
                                         &uvmRmFaultBufferInfo);
    if (status != NV_OK)
    {
        up_write(&info->lock);
        return status;
    }

    // Hardcode the class untill RM starts returning.
    info->faultBufferClass       = pGpu->faultBufferClass;
    info->faultBufferPtr         = uvmRmFaultBufferInfo.bufferAddress;
    info->faultBufferSize        = uvmRmFaultBufferInfo.bufferSize;
    info->pFaultBufferGet        = uvmRmFaultBufferInfo.pFaultBufferGet;
    info->pFaultBufferPut        = uvmRmFaultBufferInfo.pFaultBufferPut;
    info->pFaultBufferInfo       = uvmRmFaultBufferInfo.pFaultBufferInfo;
    info->pReplay                = NULL; // Should not be used
    info->pPmcIntr               = uvmRmFaultBufferInfo.pPmcIntr;
    info->pPmcIntrEnSet          = uvmRmFaultBufferInfo.pPmcIntrEnSet;
    info->pPmcIntrEnClear        = uvmRmFaultBufferInfo.pPmcIntrEnClear;
    info->replayableFaultMask    = uvmRmFaultBufferInfo.replayableFaultMask;
    info->bBufferOverflow        = NV_FALSE;    
    info->bOnlyReplaysNoResolves = NV_FALSE;
    info->faultBufferHandle      = uvmRmFaultBufferInfo.faultBufferHandle;
    // Initialize the list-head for the fault-cancels
    INIT_LIST_HEAD(&info->pendingCancelList);
    uvm_init_tracker(&info->replayTracker);

    status = uvmfull_fault_buffer_init(info->faultBufferClass, &info->faultOps);
    UVM_PANIC_ON(status != NV_OK);

    info->faultEntrySizeInBytes = info->faultOps.getFaultPacketSize();
    UVM_PANIC_ON(info->faultEntrySizeInBytes == 0);

    info->maxFaultsInBuffer    =
                         info->faultBufferSize / info->faultEntrySizeInBytes;

    // Call RM and start owning the fault buffer interrupt
    // RM should be disabling the interrupt and should not be touching GET
    // after this call.
    status = nvUvmInterfaceOwnPageFaultIntr((NvU8*)&pGpu->gpuUuid,
                                            sizeof(UvmGpuUuid),
                                            NV_TRUE);
    if (status != NV_OK)
    {
        up_write(&info->lock);
        return status;
    }

    // Safe to cache the get now.
    info->cachedFaultBufferGet = MEM_RD32(info->pFaultBufferGet);
    UVM_PANIC_ON(info->bAlive);
    info->bAlive               = NV_TRUE;
    up_write(&info->lock);

    // enable replayable faults
    info->faultOps.setFaultIntrBit((NvU32*)info->pPmcIntrEnSet);

    return status;
}

NV_STATUS uvmfull_destroy_gpu_faultbuffer(UvmGpu *pGpu)
{
    UvmGpuFaultInfo uvmRmFaultBufferInfo;
    NV_STATUS status = NV_OK;
    UvmGpuFaultBufferInfo *info;

    info = &pGpu->faultBufferInfo;

    if (!info->bAlive)
        return NV_OK;     // A error here may not help much

    // Wait for replay activity to complete.
    uvm_wait_for_tracker(&info->replayTracker);

    // Disable fault interrupts
    info->faultOps.setFaultIntrBit((NvU32*)info->pPmcIntrEnClear);

    // Call RM and start owning the fault buffer interrupt
    // RM should be disabling the interrupt and should not be touching GET
    // after this call.
    status = nvUvmInterfaceOwnPageFaultIntr((NvU8*)&pGpu->gpuUuid,
                                           sizeof(UvmGpuUuid),
                                           NV_FALSE);
    WARN_ON(status != NV_OK);
    // Even if status is not OK; we mark the fault buffer not usable internally.

    uvmRmFaultBufferInfo.bufferAddress = info->faultBufferPtr;
    uvmRmFaultBufferInfo.faultBufferHandle = info->faultBufferHandle;
    nvUvmInterfaceDestroyFaultInfo(pGpu->gpuHandle, &uvmRmFaultBufferInfo);

    info->faultBufferPtr = NULL;
    info->bAlive = NV_FALSE;
    info->bBufferOverflow = NV_FALSE;
    info->bOnlyReplaysNoResolves = NV_FALSE;
    info->faultBufferHandle = 0;
    return status;
}

static void uvmfull_isr_bottom_half(struct work_struct *work)
{
    UvmGpuUuid uuid;
    UvmPendingFault *fault;
    
    UVM_PANIC_ON(!work);
    fault = container_of(work, UvmPendingFault, work);
    memcpy(&uuid, &fault->uuid, sizeof(UvmGpuUuid));
    kmem_cache_free(g_uvmPendingFaultCache, fault);
    UVM_DBG_PRINT_UUID("\nBottom half", &uuid);

    // UVM interrupt enable and releasing the fault buffer lock are
    // done atomically. See uvmfull_handle_gpu_faults().
    uvmfull_handle_gpu_faults(&uuid);
}

// bSynchronous is used when the bottom half checks for interrupts inline.
NV_STATUS
uvmfull_isr_top_half_impl(UvmGpuUuid* pUuid, NvBool bSynchronous,
                          NvBool *pbBufferLockAcquired)
{
    UvmGpu *pGpu;
    UvmPendingFault *fault;
    UvmGpuFaultBufferInfo *faultBuffer;
    NV_STATUS status = NV_OK;

    if (!pUuid)
        return NV_ERR_NO_INTR_PENDING;

    pGpu = uvmfull_intr_uuid_to_active_gpu(pUuid);
    if (!pGpu)
        return NV_ERR_NO_INTR_PENDING;

    faultBuffer = &pGpu->faultBufferInfo;
    // Fault interrupt handling algorithm is defined in
    // //sw/docs/resman/components/compute/UVM/UVM-Full/UVM-fault-buffer-design.docx

    *pbBufferLockAcquired = NV_FALSE;
    // Try taking the lock. 0 means contention.
    if (0 == down_write_trylock(&faultBuffer->lock))
        goto ce_intr;

    // Sanity checks.
    if (!faultBuffer->bAlive || !faultBuffer->faultBufferPtr)
        goto fault_intr_release_lock;

    // Read valid bit on cached GET
    if (NV_FALSE == faultBuffer->faultOps.isFaultBufferEntryValid(
                                        (NvU64)faultBuffer->faultBufferPtr,
                                        faultBuffer->cachedFaultBufferGet))
    {
        // Read the PUT
        faultBuffer->cachedFaultBufferPut =
                              MEM_RD32(faultBuffer->pFaultBufferPut);
        if (faultBuffer->cachedFaultBufferGet ==
            faultBuffer->cachedFaultBufferPut)
        {
            // No interrupt pending
            goto fault_intr_release_lock;
        }
    }

    if (bSynchronous == NV_FALSE)
    {
        // If we reached here; atleast one fault packet was valid.
        // Disable the fault interrupt and schedule a bottom half.
        fault = kmem_cache_zalloc(g_uvmPendingFaultCache, NV_GFP_ATOMIC);
        if (!fault)
        {
            status = NV_ERR_NO_MEMORY;
            goto fault_intr_release_lock;
        }

        memcpy(&fault->uuid, pUuid, sizeof(UvmGpuUuid));

        NV_WORKQUEUE_INIT_AND_SCHEDULE(&fault->work, uvmfull_isr_bottom_half);
    }

    faultBuffer->faultOps.setFaultIntrBit((NvU32*)faultBuffer->pPmcIntrEnClear);
    *pbBufferLockAcquired = NV_TRUE;

    // Leave the lock acquired between top and bottom halves.
    return status;

fault_intr_release_lock:
    up_write(&faultBuffer->lock);
ce_intr:
    // TODO: read non stall interrupt tree
    return status;
}

/*
 * isr_top_half:
 *
 * This is the top of the interrupt event handler for the UVM driver.
 * This is actually a callback function that RM calls in order to let
 * the UVM driver know that an interrupt has occurred.
 * NV_OK means that the interrupt was for us, so use
 * NV_ERR_NO_INTR_PENDING to tell the caller that we didn't do anything.
 */
NV_STATUS uvmfull_isr_top_half(UvmGpuUuid* pUuid)
{
    NvBool bBufferLockAcquired;
    return uvmfull_isr_top_half_impl(pUuid, NV_FALSE, &bBufferLockAcquired);
}

