/*******************************************************************************
    Copyright (c) 2014-2015 NVIDIA Corporation

    Permission is hereby granted, free of charge, to any person obtaining a copy
    of this software and associated documentation files (the "Software"), to
    deal in the Software without restriction, including without limitation the
    rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
    sell copies of the Software, and to permit persons to whom the Software is
    furnished to do so, subject to the following conditions:

        The above copyright notice and this permission notice shall be
        included in all copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
    IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
    FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
    THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
    LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
    FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
    DEALINGS IN THE SOFTWARE.

*******************************************************************************/

#include "uvm_common.h"
#include "uvm_utils.h"
#include "nv_uvm_interface.h"
#include "uvm_full.h"
#include "uvm_full_pagetbl_mgmt.h"
#include "uvm_full_device_mgmt.h"
#include "uvm_full_fault_handler.h"
#include "uvm_full_ctx_mgmt.h"
#include "uvm_full_fault_buffer.h"
#include "uvm_full_identity_map.h"
#include "uvm_full_perf.h"

#define FAULTS_PER_ITER    4

#define NV_UVM_GPU_DEFAULT_FAULT_IN_SIZE    PAGE_SIZE_64K
#define NV_UVM_NUM_PAGES_TO_FAULT_IN(pgSize)     \
                (NV_UVM_GPU_DEFAULT_FAULT_IN_SIZE/(pgSize))
#define SET_BIT(bit) (1<<(bit))

#define UVM_MIN_FAULT_BUFFER_SIZE       4*1024

static NV_STATUS _downgrade_atomic_peers(Uvm2MbVaDesc *pUvm2MbVaDesc,
                                         UvmGpuPde    *pGpuPde,
                                         NvU64 vaddr,
                                         NvU32 pgSize,
                                         UvmPbOps * pbOps,
                                         NvU64 *gpuSubmitMask,
                                         NvBool bRevokeOnlyAtomic);

static struct kmem_cache *g_uvmFaultCancelEntryCache     __read_mostly = NULL;

static void _init_cancel_info(UvmGpuCancelFaultInfo *pCancelInfo)
{
    memset(pCancelInfo, 0 , sizeof(UvmGpuCancelFaultInfo));
    INIT_LIST_HEAD(&pCancelInfo->node);
}

static UvmGpuCancelFaultInfo* _alloc_cancel_info(void)
{
    UvmGpuCancelFaultInfo *pCancel = NULL;

    pCancel = kmem_cache_alloc(g_uvmFaultCancelEntryCache, NV_UVM_GFP_FLAGS);
    if (pCancel)
    {
        _init_cancel_info(pCancel);
    }
    return pCancel;
}

static NV_STATUS _uvmfull_record_gpu_fault(UvmVaspace* pVaspace,
                                           UvmFaultBufferEntry *pFault)
{
    UvmFaultEventType faultType;
    UvmEventMemoryAccessType memoryAccessType;

    if (!pVaspace || !pFault)
        return NV_ERR_INVALID_ARGUMENT;
    if (!uvm_is_event_enabled(pVaspace->pEventContainer, UvmEventTypeGpuFault))
        return NV_OK;

    memoryAccessType = (NvU8)pFault->uvmFaultAccessType;
    faultType = (NvU8) pFault->uvmFaultType;

    return uvm_record_gpu_fault(pVaspace->pEventContainer,
                                faultType,
                                memoryAccessType,
                                pFault->uvmFaultAddress,
                                NV_GETTIME(),
                                pFault->uvmFaultTimestamp);
}

static NV_STATUS _uvmfull_record_gpu_fault_replay(UvmVaspace* vaspace,
                                                  NvU64 address)
{
    if (!vaspace)
        return NV_ERR_INVALID_ARGUMENT;
    if (!uvm_is_event_enabled(vaspace->pEventContainer,
                              UvmEventTypeGpuFaultReplay))
        return NV_OK;
    return uvm_record_gpu_fault_replay(vaspace->pEventContainer,
                                       address,
                                       NV_GETTIME());
}

static NvBool
_is_inside_cancel_window(UvmGpu *pGpu, NvU32 get, NvBool *bEndOfWindow)
{
    UvmGpuFaultBufferInfo *pInfo = &pGpu->faultBufferInfo;
    NvBool bPresent = NV_FALSE;
    *bEndOfWindow = NV_FALSE;

    if (!pInfo->bCancelWindowValid)
        return NV_FALSE;

    if (pInfo->cancelWindowEnd > pInfo->cancelWindowStart)
    {
        bPresent = get >= pInfo->cancelWindowStart &&
                   get <= pInfo->cancelWindowEnd;
    }
    else
    {
        bPresent = get >= pInfo->cancelWindowStart ||
                   get <= pInfo->cancelWindowEnd;
    }

    *bEndOfWindow = (get == pInfo->cancelWindowEnd);
    return bPresent;
}

static UvmGpuCancelFaultInfo*
_find_cancel_info(UvmGpu *pGpu, NvU32 gpcId, NvU32 clientId, NvU64 instPtr, UvmFaultInstLoc aperture)
{
    UvmGpuCancelFaultInfo *pCancel = NULL;
    list_for_each_entry(pCancel, &pGpu->faultBufferInfo.pendingCancelList, node)
    {
        if (pCancel->gpcId == gpcId && pCancel->clientId == clientId && 
            pCancel->instPtr == instPtr && pCancel->aperture == aperture)
            return pCancel;
    }
    return NULL;
}

static NV_STATUS
_add_cancel_info(UvmGpu *pGpu, NvU32 gpcId, NvU32 clientId, NvU64 instPtr,  UvmFaultInstLoc aperture)
{
    UvmGpuCancelFaultInfo *pCancel = _find_cancel_info(pGpu, gpcId, clientId, instPtr, aperture);

    if (pCancel == NULL)
    {
        pCancel = _alloc_cancel_info();
        if (!pCancel)
            return NV_ERR_NO_MEMORY;

        pCancel->gpcId = gpcId;
        pCancel->clientId = clientId;
        pCancel->instPtr = instPtr;
        pCancel->aperture = aperture;
        list_add(&pCancel->node, &pGpu->faultBufferInfo.pendingCancelList);
    }

    return NV_OK;
}

static void
_push_targeted_cancel(UvmGpu *pGpu, UvmPushbuffer *pb,
                        NvU32 clientId, NvU32 gpcId,
                        NvU64 instancePtr, UvmFaultInstLoc faultAperture)
{
    NvU32 methods = 0;
    UvmFaultApperture aperture;

    if (faultAperture == uvmFaultInstLoc_vidmem)
    {
        aperture = UvmFaultApperture_vidmem;
    }
    else if (faultAperture == uvmFaultInstLoc_sysmem_coh)
    {
        aperture = UvmFaultApperture_sysmemCoh;
    }
    else if (faultAperture == uvmFaultInstLoc_sysmem_ncoh)
    {
        aperture = UvmFaultApperture_sysmemNcoh;
    }
    else
    {
        // unsupported aperture
        UVM_PANIC_ON (NV_TRUE);
    }

    methods = pb->channel->memOps.faultCancelSwMethod(
        (unsigned **)&pb->pbOffset,
        (unsigned*)(pb->cpuBegin + UVM_PUSHBUFFER_RESERVATION_SIZE),
        gpcId, clientId, instancePtr, aperture);

    UVM_PANIC_ON(!methods);

    pb->curOffset += methods;
}

NV_STATUS uvmfull_init_fault_cancel_cache()
{
    g_uvmFaultCancelEntryCache = NV_KMEM_CACHE_CREATE("uvm_fault_cancel_info",
                                                      UvmGpuCancelFaultInfo);
    if (!g_uvmFaultCancelEntryCache)
        return NV_ERR_NO_MEMORY;
    return NV_OK;
}

void uvmfull_destroy_fault_cancel_cache()
{
    kmem_cache_destroy_safe(&g_uvmFaultCancelEntryCache);
    g_uvmFaultCancelEntryCache = NULL;
}

#define UVM_MIN_FAULT_BUFFER_SIZE       4*1024
NV_STATUS uvmfull_init_fake_fault_buffer(UvmGpuUuid *pGpuUuid, NvU32 bufferSize)
{
    UvmGpu *pGpu = uvmfull_uuid_to_gpu(pGpuUuid);
    NvU32 *pFaultBuffer = (NvU32*) vmalloc(bufferSize);
    NV_STATUS status = NV_OK;

    if (!pGpu || (bufferSize < UVM_MIN_FAULT_BUFFER_SIZE))
    {
        status = NV_ERR_INVALID_ARGUMENT;
        goto cleanup;
    }

    if (!pFaultBuffer)
    {
        status = NV_ERR_NO_MEMORY;
        goto cleanup;
    }
    memset(pFaultBuffer, 0, bufferSize);

    memset(&pGpu->faultBufferInfo, 0, sizeof(UvmGpuFaultBufferInfo));

    pGpu->faultBufferInfo.faultBufferClass = MAXWELL_FAULT_BUFFER_A; //Hardcode for now
    pGpu->faultBufferInfo.pFaultBufferGet = &pFaultBuffer[0];
    pGpu->faultBufferInfo.pFaultBufferPut = &pFaultBuffer[1];
    pGpu->faultBufferInfo.pFaultBufferInfo = &pFaultBuffer[2];
    pGpu->faultBufferInfo.pReplay = &pFaultBuffer[3];
    pGpu->faultBufferInfo.pPmcIntr = &pFaultBuffer[4];
    pGpu->faultBufferInfo.pPmcIntrEnSet = &pFaultBuffer[5];
    pGpu->faultBufferInfo.pPmcIntrEnClear = &pFaultBuffer[6];
    pGpu->faultBufferInfo.faultBufferPtr = (NvU8*) &pFaultBuffer[8]; // Fault entry is 8 dword aligned

    pGpu->faultBufferInfo.faultBufferSize = bufferSize - 32;

    // Initialize the hal structures for this class
    status = uvmfull_fault_buffer_init(
            pGpu->faultBufferInfo.faultBufferClass,
            &pGpu->faultBufferInfo.faultOps);
    if (status == NV_OK)
    {
        pGpu->faultBufferInfo.faultEntrySizeInBytes =
                pGpu->faultBufferInfo.faultOps.getFaultPacketSize();
        UVM_PANIC_ON(pGpu->faultBufferInfo.faultEntrySizeInBytes == 0);

        pGpu->faultBufferInfo.maxFaultsInBuffer =
                pGpu->faultBufferInfo.faultBufferSize
                        / pGpu->faultBufferInfo.faultEntrySizeInBytes;

        pGpu->faultBufferInfo.faultBufferSize = NV_ALIGN_UP(
                pGpu->faultBufferInfo.faultBufferSize,
                pGpu->faultBufferInfo.faultEntrySizeInBytes);

        pGpu->faultBufferInfo.bAlive           = NV_TRUE;
    }
    init_rwsem(&pGpu->faultBufferInfo.lock);

cleanup:
    if (status != NV_OK && pFaultBuffer)
        vfree(pFaultBuffer);
    return status;
}

void uvmfull_destroy_fake_fault_buffer(UvmGpuUuid *pGpuUuid)
{
    UvmGpu *pGpu = uvmfull_uuid_to_gpu(pGpuUuid);

    if (!pGpu)
        return;

    if (pGpu->faultBufferInfo.pFaultBufferGet)
        vfree((void*) pGpu->faultBufferInfo.pFaultBufferGet);
    memset(&pGpu->faultBufferInfo, 0, sizeof(UvmGpuFaultBufferInfo));
}

NV_STATUS uvmfull_inject_fault_in_buffer(UvmGpuUuid *pGpuUuid,
        UvmFaultBufferEntry *pFault)
{
    NvU32 curPutIdx = 0;
    NvU32 curGetIdx = 0;
    NvU32 faultEntrySize = 0;
    NvU32 maxFbEntries = 0;
    UvmGpu *pGpu = uvmfull_uuid_to_gpu(pGpuUuid);
    NvU8 *pFPktAddr = NULL;
    NV_STATUS status = NV_OK;

    if (!pGpu)
        return NV_ERR_INVALID_ARGUMENT;

    down_write(&pGpu->faultBufferInfo.lock);

    curPutIdx = *pGpu->faultBufferInfo.pFaultBufferPut;
    curGetIdx = *pGpu->faultBufferInfo.pFaultBufferGet;
    maxFbEntries = pGpu->faultBufferInfo.maxFaultsInBuffer;
    faultEntrySize = pGpu->faultBufferInfo.faultEntrySizeInBytes;

    if (((curPutIdx + 1) % maxFbEntries) == curGetIdx)
    {
        // This means the buffer is full.
        return NV_ERR_INVALID_ARGUMENT; //TODO: FIXIT with appropriate retry code
    }

    pFPktAddr = &pGpu->faultBufferInfo.faultBufferPtr[(curPutIdx
            * faultEntrySize)];
    status = pGpu->faultBufferInfo.faultOps.writeFaultBufferPacket(
            pFault, pFPktAddr);
    curPutIdx += 1;
    if (curPutIdx >= maxFbEntries)
        curPutIdx = 0;

    // Now update the put pointer value
    if (status == NV_OK)
    {
        *pGpu->faultBufferInfo.pFaultBufferPut = curPutIdx;
    }

    up_write(&pGpu->faultBufferInfo.lock);

    return status;
}

static NV_STATUS
_trigger_replay(UvmGpu *pGpu, NvU64 pdb, UvmMemType pdbAperture,
                        NvBool bAckAll)
{
    UvmTracker tracker;
    UvmPbOps *pbOps;
    UvmPushbuffer *pb;
    NV_STATUS status = NV_OK;
    UvmTlbInvalidateVaParams invalParams = {0};

    uvm_init_tracker(&tracker);
    pbOps = uvm_pbops_start_on_gpu(pGpu->gpuId);
    UVM_PANIC_ON(!pbOps);

    invalParams.invalidatelevel = UvmTlbInvalidateLevel_all;
    pb = uvm_pbops_gpu_pb(pbOps, pGpu->gpuId);
    UVM_PANIC_ON(!pb);
    uvm_push_replay(pGpu->gpuId, pb,
                    bAckAll,
                    invalParams,
                    pdbAperture,
                    pdb >> 12);

    uvm_pbops_submit(pbOps, pGpu->gpuId, pb, &tracker);
    uvm_move_tracker(&tracker, pbOps->t);
    status = uvm_wait_for_tracker(&tracker);
    uvm_pbops_end(pbOps);
    return status;
}

static void
_invalidate_fault_buffer_entries(UvmGpu *pGpu, UvmGpuFaultBufferInfo *buf,
                                 NvU32 start, NvU32 end)
{
    NvU32 get, put;
    get = start;
    put = end;

    while (get != put)
    {
        // Wait for entry to become valid
        while (NV_FALSE == buf->faultOps.isFaultBufferEntryValid(
                                    (NvU64)buf->faultBufferPtr, get))
        {
            cpu_relax();
        }
        // invalidate the entry
        buf->faultOps.setFaultBufferEntryValid((NvU64) buf->faultBufferPtr,
                                                        get, NV_FALSE);
        // advance get
        get++;
        if (get >= buf->maxFaultsInBuffer)
            get = 0;
    }
}

static NV_STATUS
_checkpoint_gpu_faultbuffer(UvmGpu *pGpu, UvmGpuFaultBufferInfo *buf,
                            NvU64 pdb, UvmMemType pdbAperture)
{
    NvU32 get, put;
    NV_STATUS status = NV_OK;

    do
    {
        get = buf->cachedFaultBufferGet;
        put = MEM_RD32(buf->pFaultBufferPut);

        _invalidate_fault_buffer_entries(pGpu, buf, get, put);
        // update get on HW to match put
        buf->cachedFaultBufferGet = put;
        MEM_WR32(buf->pFaultBufferGet, put);

        // Issue an ack_all
        status = _trigger_replay(pGpu, pdb, pdbAperture, NV_TRUE);
        if (status != NV_OK)
        {
            // something drastic happened. Gpu should be cleaned up now.
            goto cleanup;
        }
    } while (put = MEM_RD32(buf->pFaultBufferPut),
             ((put+ 1) % buf->maxFaultsInBuffer) == buf->cachedFaultBufferGet);
    // The above loop keeps doing ack_all untill there is no probable
    // faultbuffer overflow: PUT + 1 == GET. (PUT is just behind GET by 1 that
    // it may not have progressed.

cleanup:
    buf->cachedFaultBufferPut = put;
    return status;
}

static NV_STATUS
_open_cancel_window(UvmGpu *pGpu, UvmGpuFaultBufferInfo *buf, NvU32 gpcId, NvU32 clientId,
    NvU64 instPtr,  UvmFaultInstLoc aperture, NvBool *bNewWindow)
{
    NV_STATUS status = NV_OK;
    *bNewWindow = NV_FALSE;

    // Check if a cancel window is already opened, return early.
    if (buf->bCancelWindowValid)
        return status;

    // Barrier on all previous replays to complete
    uvm_wait_for_tracker(&buf->replayTracker);

    status = _checkpoint_gpu_faultbuffer(pGpu, buf, 0,
                                        pGpu->bZeroFbGpu ?
                                        UvmMemType_sysmem_coh :
                                        UvmMemType_vidmem);
    if (status != NV_OK)
        return status;

    if (buf->cachedFaultBufferGet == buf->cachedFaultBufferPut)
        return status;

    // Add this cancel info to the cancel list in the gpu-device
    status = _add_cancel_info(pGpu, gpcId, clientId, instPtr, aperture);

    // Register the cancel window
    buf->bCancelWindowValid = NV_TRUE;
    *bNewWindow = NV_TRUE;
    // GET,PUT should have been cached by the checkpointing.
    buf->cancelWindowStart = buf->cachedFaultBufferGet;
    buf->cancelWindowEnd = buf->cachedFaultBufferPut == 0 ?
                           buf->maxFaultsInBuffer - 1:
                           buf->cachedFaultBufferPut - 1;
    return status;
}

// Returns true if a new cancel window is opened.
static NvBool
_handle_gpu_fault_cancellation(UvmGpu *gpu, UvmFaultItem *item,
                            NvBool bStartCancel, NvU64 pdb,
                            UvmMemType pdbAperture)
{
    UvmGpuCancelFaultInfo *pCancel, *pos;
    UvmTracker tracker;
    UvmPushbuffer *pb = NULL;
    NvBool bSubmit = NV_FALSE;
    UvmPbOps *pbOps = NULL;
    UvmGpuFaultBufferInfo *info = &gpu->faultBufferInfo;
    NvBool bNewWindow = NV_FALSE;

    uvm_init_tracker(&tracker);

    if (bStartCancel)
    {
        // we will trigger an ack_all and record GET and PUT.
        // The ack_all makes sure all tlbs report any translations in flight.
        // The PUT read after ack_all is the barrier to any new translations
        // untill the current window of faults are resolved/ or cancelled.
        _open_cancel_window(gpu, info,
                            item->entry.uvmFaultGpcId,
                            item->entry.uvmFaultClientId,
                            item->entry.uvmFaultInstance,
                            item->entry.uvmFaultInstLoc,
                            &bNewWindow);
    }


    if (info->bCancelWindowValid && (item->get == info->cancelWindowEnd))
    {
        // Walk through cancels and finish them.
        list_for_each_entry_safe(pCancel, pos, &info->pendingCancelList, node)
        {
            // Get a new pb and send down a targeted cancel if so.
            if (!pbOps)
                pbOps = uvm_pbops_start_on_gpu(gpu->gpuId);

            UVM_PANIC_ON(!pbOps);
            pb = uvm_pbops_gpu_pb(pbOps, gpu->gpuId);

            _push_targeted_cancel(gpu, pb,
                                pCancel->clientId,
                                pCancel->gpcId,
                                pCancel->instPtr,
                                pCancel->aperture);

            bSubmit = NV_TRUE;
            // Remove the entry from the cancel list
            list_del(&pCancel->node);
            kmem_cache_free(g_uvmFaultCancelEntryCache, pCancel);
        }
        info->bCancelWindowValid = NV_FALSE;
        info->cancelWindowEnd = info->cancelWindowStart = 0;
    }

    if (bSubmit)
    {
        uvm_pbops_submit(pbOps, gpu->gpuId, pb, &tracker);
        uvm_move_tracker(&tracker, pbOps->t);
        uvm_wait_for_tracker(&tracker);
    }

    if (pbOps)
        uvm_pbops_end(pbOps);

    return bNewWindow;
}

/// Check for overflow condition and hardware flag for overflow.
// if both are true, record overflow event. else, clear the overflow bit
// and software buffer overflopw flag
static void
_check_and_record_fault_buffer_overflow( UvmGpu *pGpu,
                                         UvmGpuFaultBufferInfo *faultBuffer)
{
    UvmFaultBufferRegisters testFaultBufferInfo = {faultBuffer->pFaultBufferGet,
                                                  faultBuffer->pFaultBufferPut,
                                                  faultBuffer->pFaultBufferInfo
                                                  };
   // read pointer == ( write pointer + 1 ) % (buf.length)
   if (faultBuffer->cachedFaultBufferGet ==
               (faultBuffer->cachedFaultBufferPut + 1) % faultBuffer->maxFaultsInBuffer)
   {
        if(faultBuffer->faultOps.testFaultBufferOverflow(testFaultBufferInfo)
               && !faultBuffer->bBufferOverflow )
        {
            UvmFaultBufferEntry entry;
            UvmVaspace *pVaspace = NULL;
            faultBuffer->faultOps.parseFaultBufferEntry((NvU64)faultBuffer->faultBufferPtr,
                                                         faultBuffer->cachedFaultBufferGet, &entry);
            pVaspace = uvmfull_get_vaspace_from_instptr(pGpu->gpuId,
                                                        entry.uvmFaultInstance,
                                                        entry.uvmFaultInstLoc);

            // If FaultBufferOverflow events are on then we have to record it
            if (uvm_is_event_enabled(pVaspace->pEventContainer,
                                     UvmEventTypeFaultBufferOverflow))
                uvm_record_fault_buffer_overflow_event(
                                               pVaspace->pEventContainer,
                                               NV_GETTIME());

            faultBuffer->bBufferOverflow = NV_TRUE;
        }
   }
   else //resetting the buffer overflow flag when it is "not full"
   {
       if(faultBuffer->bBufferOverflow)
       {
           if (faultBuffer->faultOps.testFaultBufferOverflow(testFaultBufferInfo))
           {
               faultBuffer->faultOps.clearFaultBufferOverflow(testFaultBufferInfo);
               faultBuffer->bBufferOverflow = NV_FALSE;
           }
       }
   }
}
// Parse the faultbuffer entries, cache them in array, mark parsed entries
// invalid in buffer, advance GET.
static NV_STATUS
_fetch_gpu_faults(UvmGpu *pGpu,
                   UvmFaultItem *items, NvU32 nItems, NvU32 *nCached)
{
    NvU32 get, put, i;
    UvmGpuFaultBufferInfo *buf;
    NV_STATUS status = NV_OK;
    NvU32 count = 0;

    buf = &pGpu->faultBufferInfo;
    UVM_PANIC_ON(!rwsem_is_locked(&buf->lock));
    UVM_PANIC_ON(!buf->bAlive);

    get = buf->cachedFaultBufferGet;
    put = MEM_RD32(buf->pFaultBufferPut);
    buf->cachedFaultBufferPut = put;

    if (get == put || nItems == 0)
    {
        *nCached = 0;
        return NV_OK;
    }
    _check_and_record_fault_buffer_overflow(pGpu, buf);
    // clear the cache
    memset(items, 0x0, nItems*sizeof(UvmFaultItem));

    // Parse until get != put and have enough space to cache
    // If atleast one fault is in the cache; bypass poll for valid.
    for (i = 0; (get != put) && (i < nItems); i++)
    {
        if (NV_FALSE == buf->faultOps.isFaultBufferEntryValid(
                                             (NvU64)buf->faultBufferPtr,get))
        {
            // Keep waiting if no entry parsed
            if (!count)
            {
                // Check for a timeout here.
                // The timeout would need to be gpu-platform dependent.
                while (NV_FALSE == buf->faultOps.isFaultBufferEntryValid(
                                            (NvU64)buf->faultBufferPtr, get))
                {
                    cpu_relax();
                }
            }
            else
                break;  // We have some entry to work on. Lets do the rest later
        }

        // Got valid bit set. Lets cache.
        status = buf->faultOps.parseFaultBufferEntry(
                              (NvU64)buf->faultBufferPtr, get, &items[i].entry);
        if (NV_OK != status)
            break;

        items[i].get = get;
        items[i].bValid = NV_TRUE;

        // Invalidate the entry in fault buffer
        buf->faultOps.setFaultBufferEntryValid((NvU64) buf->faultBufferPtr,
                                                get, NV_FALSE);

        count++;
        get++;
        if (get >= buf->maxFaultsInBuffer)
            get = 0;
    }

    *nCached = count;
    buf->cachedFaultBufferGet = get;
    MEM_WR32(buf->pFaultBufferGet, get);

    _check_and_record_fault_buffer_overflow(pGpu, buf);

    return status;
}

static void
_preprocess_gpu_faults(UvmGpu *pGpu, UvmFaultItem *items, NvU32 nItems)
{
    // TODO: Run through the array of faults and mark the duplicate/redundant
    // entries as sw-invalid. (coalesce)
    // Also tag the fault items with cancel information

    NvU32 i;
    UvmFaultItem *item;
    NvBool bEndOfWindow = NV_FALSE;
    UvmGpuFaultBufferInfo *info = &pGpu->faultBufferInfo;

    for (i = 0; i < nItems; i++)
    {
        item = &items[i];
        if (!item->bValid)
            continue;

        // Prefetches:
        // For now the policy is to just neglect the prefetch accesses.
        // Each prefetch access increases the penalty count by "weight".
        // Each non-prefetch decreases the penalty by 1.
        // When penalty reaches a "threshold", prefetches are throttled so as
        // to not flood the fault buffer.
        // Prefetches not happening for a long time will eventually cause the
        // throttling to be undone when penalty reaches zero again.
        if (UvmAccessType_prefetch == item->entry.uvmFaultAccessType)
        {
            // By default lets just neglect all prefetches.
            item->bPrefetch = NV_TRUE;

            info->nPrefetches++;
            info->prefetchPenalty += UVM_PREFETCH_PENALTY_WEIGHT;

            // See if there are too many already
            if (!info->bPrefetchThrottled &&
                info->prefetchPenalty >= UVM_PREFETCH_THROTTLE_THRESHOLD)
            {
                // allow one in 16^2
                info->faultOps.controlPrefetch(info->pPrefetchCtrl,
                                              uvmPrefetchThrottle_1_in_Npower2);
                info->bPrefetchThrottled = NV_TRUE;
            }
        }
        else if (info->prefetchPenalty)  // Generic fault: reduce penalty
        {
            info->prefetchPenalty--;
            if (info->bPrefetchThrottled && info->prefetchPenalty == 0)
            {
                // allow all
                info->faultOps.controlPrefetch(info->pPrefetchCtrl,
                                               uvmPrefetchThrottle_allow_all);
                info->bPrefetchThrottled = NV_FALSE;
            }
        }

        if (item->bPrefetch)
            continue;

        // Update cancel information. If we encounter a new gpcId, ClientId,
        // we add a new cancel node in the list.
        if (_is_inside_cancel_window(pGpu, item->get, &bEndOfWindow))
        {
            item->bCancelWindowEnd = bEndOfWindow;
            item->bInCancelWindow = NV_TRUE;

            // The following function is no-op if the gpcId, clientId entry
            // exists already.
            _add_cancel_info(pGpu, item->entry.uvmFaultGpcId,
                                   item->entry.uvmFaultClientId,
                                   item->entry.uvmFaultInstance,
                                   item->entry.uvmFaultInstLoc);
        }
    }
}

// Handle the cached faults.
static NV_STATUS
_service_gpu_faults(UvmGpu *pGpu, UvmFaultItem *items, NvU32 nItems,
                    NvBool *pbReplayed)
{
    NvU32 i;
    UvmVaspace *pVaspace;
    NV_STATUS status = NV_OK;
    NvBool bReCacheFaults = NV_FALSE;
    NvU32 nServiceable = nItems;
    NvU32 nServiced = 0;

    for (i = 0; i < nItems; i++)
    {
        if (!items[i].bValid)   // Can happen if it was coalesced
        {
            nServiceable--;
            continue;
        }

        // get the vaspace associated with this fault. We have the device lock.
        pVaspace = uvmfull_get_vaspace_from_instptr(pGpu->gpuId,
                                            items[i].entry.uvmFaultInstance,
                                            items[i].entry.uvmFaultInstLoc);
        if (!pVaspace)
        {
            if (items[i].bPrefetch)
            {
                nServiceable--;
                continue;
            }

            items[i].bFatalFault = NV_TRUE;
            bReCacheFaults = _handle_gpu_fault_cancellation(pGpu, &items[i],
                                             NV_TRUE,
                                             0, pGpu->bZeroFbGpu ?
                                                UvmMemType_sysmem_coh :
                                                UvmMemType_vidmem);
        }
        else
        {
            down_read(&pVaspace->lock);
            // The below case can only happen if the instancePtr has been reused.
            // The hashtable has association with old channel but the vaspace
            // info got updated behind the back. Driver-CUDA interaction bug!!
            if (pVaspace != pVaspace->gpuVaspaces[pGpu->gpuId].uvmVaspace)
            {
                UVM_ERR_PRINT("Gpu fault for an unregistered vaspace 0x%p.",
                                pVaspace);

                // A cancel is not required if prefetch.
                if (items[i].bPrefetch)
                {
                    nServiceable--;
                    continue;
                }

                items[i].bFatalFault = NV_TRUE;
                up_read(&pVaspace->lock);
                // open a cancel window and return early from the loop
                bReCacheFaults = _handle_gpu_fault_cancellation(pGpu, &items[i],
                                                      NV_TRUE,
                                                      0, pGpu->bZeroFbGpu ?
                                                         UvmMemType_sysmem_coh :
                                                         UvmMemType_vidmem);
            }
            else if (items[i].bPrefetch)
            {
                // Just send an event and ignore resolving the fault
                _uvmfull_record_gpu_fault(pVaspace, &items[i].entry);
                nServiceable--;
            }
            else
            {
                status =
                    uvmfull_handle_gpu_single_fault(pGpu, pVaspace, &items[i]);
                WARN_ON(status != NV_OK); // FIXIT: Add more debug information

                if (items[i].bReplayed)
                    nServiced++;

                bReCacheFaults =
                        _handle_gpu_fault_cancellation(pGpu, &items[i],
                                items[i].bFatalFault,
                                pVaspace->gpuVaspaces[pGpu->gpuId].gpuPDB,
                                pVaspace->gpuVaspaces[pGpu->gpuId].pdbAperture);
            }
            up_read(&pVaspace->lock);
        }

        // If a fatal fault was encountered and a new cancel window was opened.
        if (bReCacheFaults)
            break;
    }

    UVM_PANIC_ON(nServiceable > nItems);

    if (bReCacheFaults)          // Recaching triggers an ack_all
        *pbReplayed = NV_TRUE;
    else if (nServiceable == 0)  // All prefetches. No explicit replay needed for them.
        *pbReplayed = NV_TRUE;
    else                         // Check if no replays were pushed.
        *pbReplayed = (nServiced > 0);

    return status;
}

static NvBool
_faultbuffer_overflow_test_only(UvmGpu *pGpu, UvmGpuFaultBufferInfo *faultBuffer)
{
    if (faultBuffer->bOnlyReplaysNoResolves)
    {
        UvmTracker tracker;
        UvmPbOps   *pbOps = NULL;
        UvmPushbuffer *pb =NULL;
        UvmTlbInvalidateMembarType membar;
        UvmTlbInvalidateVaParams invalParams = {0};
        UvmFaultBufferEntry entry;
        UvmVaspace *pVaspace = NULL;
        UvmGpuVaspace *gpuVaspace = NULL;
        NvBool bReplayStartAckAll;

        faultBuffer->cachedFaultBufferPut = MEM_RD32(faultBuffer->pFaultBufferPut);
        faultBuffer->faultOps.parseFaultBufferEntry((NvU64)faultBuffer->faultBufferPtr,
                                                     faultBuffer->cachedFaultBufferGet, &entry);
        pVaspace = uvmfull_get_vaspace_from_instptr(pGpu->gpuId,
                                                    entry.uvmFaultInstance,
                                                    entry.uvmFaultInstLoc);
        gpuVaspace = &pVaspace->gpuVaspaces[pGpu->gpuId];
        bReplayStartAckAll = gpuVaspace->policy.bForceReplayStartAckAll;
        membar = gpuVaspace->pdbAperture == UvmMemType_vidmem ?
                                  UvmTlbInvalidateMemBar_local:
                                  UvmTlbInvalidateMemBar_sys;
        uvm_init_tracker(&tracker);
        pbOps = uvm_pbops_start_on_gpu(pGpu->gpuId);
        UVM_PANIC_ON(!pbOps);

        pb = uvm_pbops_gpu_pb(pbOps, pGpu->gpuId);
        UVM_PANIC_ON(!pb);

        if (bReplayStartAckAll)
            invalParams.invalidatelevel = UvmTlbInvalidateLevel_all;

        uvm_push_replay(pGpu->gpuId,
                        pb,
                        bReplayStartAckAll,
                        invalParams,
                        gpuVaspace->pdbAperture == UvmMemType_vidmem?
                               UvmTlbInvalidatePdbAperture_vidmem :
                               UvmTlbInvalidatePdbAperture_sysmemCoh,
                        0 /*targetPdb = ALL*/);

        // Submit the pushbuffer.
        uvm_pbops_submit(pbOps, pGpu->gpuId, pb, &tracker);
        uvm_move_tracker(&tracker, pbOps->t);
        uvm_wait_for_tracker(&tracker);
        uvm_pbops_end(pbOps);
        _check_and_record_fault_buffer_overflow(pGpu, faultBuffer);
    }
    return faultBuffer->bOnlyReplaysNoResolves;
}

// _consume, _preprocess, _service
NV_STATUS uvmfull_handle_gpu_faults(UvmGpuUuid *pGpuUuid)
{
    UvmGlobalState *uvm;
    UvmGpu *pGpu;
    UvmFaultItem uvmFaultItems[FAULTS_PER_ITER];
    NV_STATUS status = NV_OK;
    NvU32 faultsCached = 0;
    NvBool bReplayedAtleastOnce = NV_FALSE;
    unsigned long oldIrq;

    // This is awkward. We have the faultbuffer lock already but we are
    // acquiring the global lock midway. We will make sure we release the fault
    // buffer lock after releasing the global lock.
    uvm = uvmfull_get_global();
    down_read(&uvm->lock);

    pGpu = uvmfull_uuid_to_active_gpu(pGpuUuid);
    if (!pGpu || !pGpu->bValid)
    {
        WARN_ON(!pGpu || !pGpu->bValid);
        status = NV_ERR_INVALID_ARGUMENT;
        goto handler_end;
    }

    if (_faultbuffer_overflow_test_only(pGpu, &pGpu->faultBufferInfo))
    {
        // skip service GPU faults if the fault buffer is configured
        // to do just replays.
        // check for overflow and record overflow event
        goto handler_end;
    }

    while (status = _fetch_gpu_faults(pGpu, uvmFaultItems, FAULTS_PER_ITER,
                                      &faultsCached),
           (status == NV_OK && faultsCached != 0))
    {
        _preprocess_gpu_faults(pGpu, uvmFaultItems, faultsCached);

        status = _service_gpu_faults(pGpu, uvmFaultItems, faultsCached,
                                     &bReplayedAtleastOnce);
        if (status != NV_OK)
            break;    // How to handle this? TODO: Cancel all
    }

    // If the last iteration did not push a replay and there are no more faults,
    // a replay is pushed to make sure we dont miss faults in tlb.
    if (status == NV_OK && !bReplayedAtleastOnce)
        status = _trigger_replay(pGpu, 0, UvmMemType_vidmem, NV_FALSE);

handler_end:
    // Enable the gpu faults and release fault buffer lock atomically.
    local_irq_save(oldIrq);
    preempt_disable();

    if (pGpu && pGpu->bValid)
        pGpu->faultBufferInfo.faultOps.setFaultIntrBit(
                                   (NvU32*)pGpu->faultBufferInfo.pPmcIntrEnSet);
    up_read(&uvm->lock);
    up_write(&pGpu->faultBufferInfo.lock);

    preempt_enable();
    local_irq_restore(oldIrq);

    return status;
}

inline UvmPteHints* uvmfull_2mva_get_pte_hints(Uvm2MbVaDesc *pUvm2MbVaDesc,
        NvU64 vaddr)
{
    NvU32 offset = GET_PDE_PTE_OFFSET(vaddr, PAGE_SHIFT_4K);
    UVM_PANIC_ON(offset > NUM_4K_PER_2M);
    return &pUvm2MbVaDesc->hints[offset];
}

NvBool uvmfull_is_migration_needed(UvmPteHints *pPteHints,
        UvmFaultBufferEntry *pFault, UvmGpuPde *pGpuPde)
{
    // TODO: check if migration is needed here
    //       based on current settings and new target
    return NV_FALSE;
}

static NvBool _check_early_fault_cancel(UvmPteHints *pPteHints,
        UvmFaultBufferEntry *pFault, UvmGpuPde *pGpuPde)
{
    if (pPteHints->managedVa == 0x0)
        return NV_TRUE;

    if (pPteHints->testPolicyEnabled)
    {
        if (pFault->uvmFaultAccessType == UvmAccessType_write
                            && !pPteHints->testPolicyReadWrite)
            return NV_TRUE;

        if (pFault->uvmFaultAccessType == UvmAccessType_atomic
                            && !pPteHints->testPolicyAtomic)
            return NV_TRUE;

        if (pPteHints->testPolicyCancel)
            return NV_TRUE;
    }

    if (pPteHints->va_prot_write == 0x0 && pPteHints->va_prot_read == 0x0)
        return NV_TRUE;

    if (pFault->uvmFaultAccessType ==  UvmAccessType_write &&
        pPteHints->va_prot_write == 0x0)
        return NV_TRUE;

    if (pFault->uvmFaultAccessType ==  UvmAccessType_atomic &&
        pPteHints->va_prot_write == 0x0)
        return NV_TRUE;

    if (pGpuPde && pPteHints->notMigratable
            && uvmfull_is_migration_needed(pPteHints, pFault, pGpuPde))
        return NV_TRUE;

    return NV_FALSE;
}

NvBool _compare_pte_hints
(
    UvmPteHints *pPte1,
    UvmPteHints *pPte2,
    NvU32       pteHintFlags
)
{
    // Passing 0 means do not compare anything, programming error?
    UVM_PANIC_ON(pteHintFlags==0);

    if (pteHintFlags & UVM_GPU_VA_PTE_HINTS_FLAG_SET_HOME_PROCESSOR)
            if (pPte1->homeProcessor != pPte2->homeProcessor) return NV_FALSE;
    if (pteHintFlags & UVM_GPU_VA_PTE_HINTS_FLAG_SET_MANAGED_VA)
            if (pPte1->managedVa != pPte2->managedVa) return NV_FALSE;
    if (pteHintFlags & UVM_GPU_VA_PTE_HINTS_FLAG_SET_NOT_MIGRATABLE)
            if (pPte1->notMigratable != pPte2->notMigratable) return NV_FALSE;
    if (pteHintFlags & UVM_GPU_VA_PTE_HINTS_FLAG_SET_PROT_READ)
            if (pPte1->va_prot_read  != pPte2->va_prot_read ) return NV_FALSE;
    if (pteHintFlags & UVM_GPU_VA_PTE_HINTS_FLAG_SET_PROT_WRITE)
            if (pPte1->va_prot_write  != pPte2->va_prot_write ) return NV_FALSE;
    if (pteHintFlags & UVM_GPU_VA_PTE_HINTS_FLAG_SET_FLT_PLCY_RW)
            if (pPte1->testPolicyReadWrite  != pPte2->testPolicyReadWrite )  return NV_FALSE;
    if (pteHintFlags & UVM_GPU_VA_PTE_HINTS_FLAG_SET_FLT_PLCY_ATOMIC)
            if (pPte1->testPolicyAtomic  != pPte2->testPolicyAtomic )  return NV_FALSE;
    if (pteHintFlags & UVM_GPU_VA_PTE_HINTS_FLAG_SET_FLT_PLCY_RD_DUP)
            if (pPte1->testPolicyRdDuplicate  != pPte2->testPolicyRdDuplicate )  return NV_FALSE;
    if (pteHintFlags & UVM_GPU_VA_PTE_HINTS_FLAG_SET_FLT_PLCY_SYSONLY)
            if (pPte1->testPolicySysMemOnly  != pPte2->testPolicySysMemOnly )  return NV_FALSE;
    if (pteHintFlags & UVM_GPU_VA_PTE_HINTS_FLAG_SET_FLT_PLCY_CANCEL)
            if (pPte1->testPolicyCancel  != pPte2->testPolicyCancel )  return NV_FALSE;
    if (pteHintFlags & UVM_GPU_VA_PTE_HINTS_FLAG_SET_CPU_PFN_VALID)
            if (pPte1->cpuPfnValid  != pPte2->cpuPfnValid )  return NV_FALSE;
    if (pteHintFlags & UVM_GPU_VA_PTE_HINTS_FLAG_SET_CPU_PFN_VALUE)
            if (pPte1->cpuPfn  != pPte2->cpuPfn )  return NV_FALSE;

    return NV_TRUE;
}

static NvBool
_compare_gpu_ptes(UvmGpuPte *pPte1, UvmGpuPte *pPte2, NvU32 gpuPteFlags)
{
    UVM_PANIC_ON(gpuPteFlags == 0);

    if (gpuPteFlags & UVM_GPU_SW_PTE_FLAGS_SWVALID)
        if (pPte1->swcachevalid != pPte2->swcachevalid) return NV_FALSE;
    if(gpuPteFlags & UVM_GPU_SW_PTE_FLAGS_TRANSIENT)
        if (pPte1->swtransient != pPte2->swtransient) return NV_FALSE;
    if(gpuPteFlags & UVM_GPU_SW_PTE_FLAGS_PTE_VALID)
        if (pPte1->valid != pPte2->valid) return NV_FALSE;
    if(gpuPteFlags & UVM_GPU_SW_PTE_FLAGS_READ_ONLY)
         if (pPte1->readonly != pPte2->readonly) return NV_FALSE;
    if(gpuPteFlags & UVM_GPU_SW_PTE_FLAGS_ATOMIC)
         if (pPte1->atomic != pPte2->atomic) return NV_FALSE;
    if(gpuPteFlags & UVM_GPU_SW_PTE_FLAGS_APERTURE)
         if (pPte1->aperture != pPte2->aperture) return NV_FALSE;
    if(gpuPteFlags & UVM_GPU_SW_PTE_FLAGS_PEERID)
         if (pPte1->peerid != pPte2->peerid) return NV_FALSE;
    if(gpuPteFlags & UVM_GPU_SW_PTE_FLAGS_VOLATILE)
         if (pPte1->vol != pPte2->vol) return NV_FALSE;
    if(gpuPteFlags & UVM_GPU_SW_PTE_FLAGS_PFN)
         if (pPte1->pfn != pPte2->pfn) return NV_FALSE;
    if(gpuPteFlags & UVM_GPU_SW_PTE_FLAGS_PFN_VALID)
         if (pPte1->pfnvalid != pPte2->pfnvalid) return NV_FALSE;
    if(gpuPteFlags & UVM_GPU_SW_PTE_FLAGS_HINTS)
         if (pPte1->hints != pPte2->hints) return NV_FALSE;
    if(gpuPteFlags & UVM_GPU_SW_PTE_FLAGS_RMPTE)
         if (pPte1->rmPte != pPte2->rmPte) return NV_FALSE;
    if(gpuPteFlags & UVM_GPU_SW_PTE_FLAGS_ZFOD_MAPPED)
         if (pPte1->zfodMapped != pPte2->zfodMapped) return NV_FALSE;

    return NV_TRUE;
}


NvBool
uvmfull_check_page_merge(Uvm2MbVaDesc *pUvm2MbVaDesc, UvmGpuPde *pGpuPde,
                         NvU64 vaddr,UvmGpuPte *pPte, NvU32 curPgSize,
                         NvU32 mergePageSize, NvU64* startPfn
                        )
{
    NvU32 start;
    NvU32 end;
    NvU32 i =0;
    NvBool bMerge = NV_TRUE;
    NvU64 prevPfn = 0;
    NvU64 startAddr,tempOffset;
    UvmGpuPte *tempPte = NULL;
    NvU32 pgSize = 0;
    NvU32 prevPgSize = 0;
    NvBool bIsContig = NV_TRUE;

    if (curPgSize == mergePageSize)
        return NV_FALSE;

    // start with next upper page size
    start = GET_PDE_PTE_OFFSET(UVM_ALIGN_ADDR(vaddr, mergePageSize),
                         UVM_CPU_MIN_PAGE_SHIFT);
    end =
        GET_PDE_PTE_OFFSET(UVM_ALIGN_ADDR(vaddr, mergePageSize) + mergePageSize - 1,
                        UVM_CPU_MIN_PAGE_SHIFT);
    for (i = (start + 1); i < end; i++)
    {
        if (!_compare_pte_hints(&pUvm2MbVaDesc->hints[start],
                                &pUvm2MbVaDesc->hints[i],
                                NV_COMPARE_HINTS_ON_FAULT))
        {
            bMerge = NV_FALSE;
            break;
        }
    }

    if (!bMerge)
        return NV_FALSE;

    startAddr = UVM_ALIGN_ADDR(vaddr, mergePageSize);
    // loop over all the pages in this page and make sure if they are contiguous
    tempOffset = 0;
    i = 0;
    while(tempOffset < GPU_PDE_SIZE)
    {
        tempPte = uvmfull_get_gpupde_current_pte(pGpuPde,
                                                 (startAddr + tempOffset),
                                                 &pgSize);
        if (!tempPte || !tempPte->swcachevalid)
        {
            // We cannot merge a page without a valid PTEs
            bMerge = NV_FALSE;
            goto cleanup;
        }

        if (!tempPte->pfnvalid)
        {
            // We cannot merge a page without backing pages
            bMerge = NV_FALSE;
            goto cleanup;
        }

        if (!_compare_gpu_ptes(pPte, tempPte, NV_COMPARE_GPU_PTE_ON_MERGE))
        {
            // we cannot merge into a page if PTEs differ
            bMerge = NV_FALSE;
            goto cleanup;
        }

        if ( i > 0)
        {
            if ((prevPfn + prevPgSize) == tempPte->pfn)
                bIsContig = NV_FALSE;
        }
        else
            *startPfn = tempPte->pfn;
        prevPfn = tempPte->pfn;
        prevPgSize = pgSize;
        tempOffset += pgSize;
        i++;
    }
    if (!bIsContig)
       // caller asked for a merge only if phys backing is contiguous
        bMerge = NV_FALSE;

cleanup:
    return bMerge;
}

void
uvmfull_check_page_split(Uvm2MbVaDesc *pUvm2MbVaDesc, NvU64 vaddr,
        NvU32 curPageSize, NvU32 *pSuggestedPgSize)
{
    NvU32 start;
    NvU32 end;
    NvU32 numSameHints = 0;
    NvU32 i = 0;
    NvBool bMismatch = NV_FALSE;

    UVM_PANIC_ON(curPageSize != PAGE_SIZE_4K &&
                 curPageSize != PAGE_SIZE_64K &&
                 curPageSize != PAGE_SIZE_2M);

    // Start with current
    *pSuggestedPgSize = curPageSize;

    start = GET_PDE_PTE_OFFSET(UVM_ALIGN_ADDR(vaddr, curPageSize),
                                      PAGE_SHIFT_4K);
    end =
        GET_PDE_PTE_OFFSET(UVM_ALIGN_ADDR(vaddr, curPageSize) + curPageSize - 1,
                                      PAGE_SHIFT_4K);
    for (i = (start + 1); i < end; i++)
    {
        if (!_compare_pte_hints(&pUvm2MbVaDesc->hints[start],
                                &pUvm2MbVaDesc->hints[i],
                                NV_COMPARE_HINTS_ON_FAULT))
        {
            bMismatch = NV_TRUE;
            //TODO: Fix flags that we are interested in comparing
            break;
        }
        numSameHints++;
    }

    if (bMismatch)
    {
        if (curPageSize == PAGE_SIZE_2M)
            if (numSameHints >= NUM_4K_PER_64K)
                *pSuggestedPgSize = PAGE_SIZE_64K;
            else
                *pSuggestedPgSize = PAGE_SIZE_4K;
        else
            *pSuggestedPgSize = PAGE_SIZE_4K;
    }
}

static void _create_gpte_params_from_hints(UvmFaultBufferEntry *pFault,
        UvmGpuVaspace *gpuVaspace, Uvm2MbVaDesc *pUvm2MbVaDesc,
        UvmGpuPde *pGpuPde, UvmPteHints *pPteHints,
        UvmGpuPteUpdateParams *pFaultPteParams, NvU32 *pgSize)
{
    NvU32 preferredPageSize;
    NvU32 splitPgSize;
    NvU64 vaddr = pFault->uvmFaultAddress;

    if (gpuVaspace->bZeroFb || pPteHints->pinToCpu)
    {
        pFaultPteParams->aperture = UvmMemType_sysmem_coh;
        pFaultPteParams->updateSwFlags |= UVM_GPU_SW_PTE_UPDATE_FLAGS_APERTURE;
        preferredPageSize = UVM_CPU_MIN_PAGE_SIZE;
    }
    else
    {
        pFaultPteParams->aperture = UvmMemType_vidmem;
        pFaultPteParams->updateSwFlags |= UVM_GPU_SW_PTE_UPDATE_FLAGS_APERTURE;
        preferredPageSize =  gpuVaspace->policy.mappingSize;
    }

    if (pPteHints->testPolicyEnabled)
    {
        pFaultPteParams->readonly = !pPteHints->testPolicyReadWrite;
        pFaultPteParams->updateSwFlags |= UVM_GPU_SW_PTE_UPDATE_FLAGS_READ_ONLY;
        if (pPteHints->testPolicySysMemOnly)
        {
            pFaultPteParams->aperture = UvmMemType_sysmem_coh;
            pFaultPteParams->updateSwFlags |= UVM_GPU_SW_PTE_UPDATE_FLAGS_APERTURE;
            preferredPageSize = UVM_CPU_MIN_PAGE_SIZE;
        }
    }
    else
    {
        if (pPteHints->va_prot_write == 0x0)
        {
            pFaultPteParams->readonly = NV_TRUE;
            pFaultPteParams->updateSwFlags |= UVM_GPU_SW_PTE_UPDATE_FLAGS_READ_ONLY;
        }
    }

    //
    // check if the va hints are the same for all the sub-pages
    // if not, we would have to split the pages down and update
    // ptes accordingly
    //
    uvmfull_check_page_split(pUvm2MbVaDesc,
                                vaddr, preferredPageSize, &splitPgSize);
    if (splitPgSize != preferredPageSize)
    {
        WARN_ON(splitPgSize > preferredPageSize);
        preferredPageSize = splitPgSize;
    }

    if (pgSize)
        *pgSize = preferredPageSize;
}

static NV_STATUS
_gpu_single_fault_epilogue(UvmGpu *pGpu, UvmPbOps *pbOps,
                         Uvm2MbVaDesc *pVaDesc, UvmFaultItem *pFaultItem)
{
    UvmTlbInvalidateMembarType membar;
    UvmVaspace *pVaspace = pVaDesc->pVaspace;
    UvmGpuVaspace *pGpuVaspace = &pVaspace->gpuVaspaces[pGpu->gpuId];
    UvmFaultBufferEntry *pFault = &pFaultItem->entry;
    UvmPushbuffer *pb = uvm_pbops_gpu_pb(pbOps, pGpu->gpuId);
    NvBool bTrackerWait = NV_FALSE;
    NvBool bRegBasedReplay = !pGpu->bUseMemops;
    NV_STATUS status = NV_OK;

    UVM_PANIC_ON(!pFaultItem->bValid);

    // If the item is a fatal fault, dont submit replay.
    // If the item is in cancel window, dont replay but submit the other methods
    // synchronously.
    if (pFaultItem->bFatalFault)
        return status;

    _uvmfull_record_gpu_fault(pVaspace, pFault);

    if (pFaultItem->bInCancelWindow)
        bTrackerWait = NV_TRUE;
    else
    {
        UvmTlbInvalidateVaParams invalParams;
        NvBool bReplayStartAckAll = NV_FALSE;

        if (bRegBasedReplay)
        {
            // Not supported any more
            UVM_PANIC();
            //uvm_trigger_replay_priv(pGpu->gpuId, pFault,
            //                  UvmTlbInvalidateReplayType_start,
            //                  0/*pdb aperture*/, 0);
        }

        memset(&invalParams, 0x0, sizeof(UvmTlbInvalidateVaParams));
        membar = pGpuVaspace->pdbAperture == UvmMemType_vidmem ?
                                            UvmTlbInvalidateMemBar_local:
                                            UvmTlbInvalidateMemBar_sys;

        if (pFault->uvmFaultType == uvmFault_atomic ||
            pGpuVaspace->policy.bForceReplayStartAckAll)
        {
            // Ack_all is needed to prevent a live-lock.
            // Refer to UVM FD section 5.2.5.3 to get more details
            // on why an ack_all is needed.
            bTrackerWait = NV_TRUE;
            bReplayStartAckAll = NV_TRUE;
        }

        invalParams.invalidatelevel = UvmTlbInvalidateLevel_all;
        uvm_push_replay(pGpu->gpuId, pb,
                        bReplayStartAckAll,
                        invalParams,
                        pGpuVaspace->pdbAperture,
                        pGpuVaspace->gpuPDB >> 12);
        pFaultItem->bReplayed = 0x1;
    }

    _uvmfull_record_gpu_fault_replay(pVaspace, pFault->uvmFaultAddress);
    // Submit the pushbuffer.
    uvm_pbops_submit(pbOps, pGpu->gpuId, pb, &pVaDesc->tracker);
    uvm_move_tracker(&pVaDesc->tracker, pbOps->t);
    if (bTrackerWait)
        status = uvm_wait_for_tracker(&pVaDesc->tracker);

    return status;
}

NV_STATUS uvmfull_handle_gpu_single_fault(UvmGpu *pGpu,
                                UvmVaspace *pVaspace, UvmFaultItem *pFaultItem)
{
    UvmFaultBufferEntry *pFault = &pFaultItem->entry;
    NvU64 vaddr = pFault->uvmFaultAddress;
    Uvm2MbVaDesc *pUvm2MbVaDesc = NULL;
    UvmGpuPde *pGpuPde = NULL;
    UvmPteHints *pPteHints = NULL;
    NV_STATUS status = NV_OK;
    UvmVatContext query;
    pFaultItem->bFatalFault = NV_FALSE;

    // The query will allocate PDEs while traversing.
    query.accessType = UvmVatAccess_GPUTouch;
    query.address = pFault->uvmFaultAddress;
    query.gpuId = pGpu->gpuId;
    query.vaspace = pVaspace;
    pUvm2MbVaDesc = uvmfull_vat_acquire_leaf(&pVaspace->vat, &query);
    if (!pUvm2MbVaDesc)
    {
        pFaultItem->bFatalFault = NV_TRUE;
        goto cleanup;
    }

    pPteHints = uvmfull_2mva_get_pte_hints(pUvm2MbVaDesc, vaddr);
    UVM_PANIC_ON(pPteHints == NULL);

    // The query would have already allocated the gpupde struct while traversing
    pGpuPde = uvmfull_get_gpupde_in_vadesc(pUvm2MbVaDesc, pGpu->gpuId);
    if (!pGpuPde)
    {
        pFaultItem->bFatalFault = NV_TRUE;
        goto cleanup;
    }

    if (_check_early_fault_cancel(pPteHints, pFault, pGpuPde))
    {
        pFaultItem->bFatalFault = NV_TRUE;
        goto cleanup;
    }

    switch (pFault->uvmFaultType)
    {
    // PDE and PTE faults will be handled the same way
    case uvmFault_invalidPde:
        // The trie traversal allocates and links all PDE levels except
        // linking the PDE0 with the pagetables. The pagetables are maintained
        // in the 2Mb vadesc. We do the linking here. The ptes would start empty
        // so we would expect to get a pte fault. (which we handle inline)
        // TODO: The pte fault handler below may ignore handling some of the
        // faults for which it sees that the ptes are already in the target
        // state. This behavior ignores the fact that mapping pdes may also
        // need a replay. PDE mapping is done by the va-trie traversal and the
        // function below. Both should return a flag whether a replay is needed.
        status = uvmfull_handle_gpu_pde_fault(pGpu, pUvm2MbVaDesc,
                                                pGpuPde, pFaultItem);
        break;
    case uvmFault_invalidPte:
        status = uvmfull_handle_gpu_pte_fault(pGpu, pUvm2MbVaDesc,
                                                pGpuPde, pFaultItem, NULL);
        break;
    case uvmFault_write:
        status = uvmfull_handle_gpu_write_fault(pGpu, pUvm2MbVaDesc,
                                                pGpuPde, pFaultItem);
        break;
    case uvmFault_atomic:
        status = uvmfull_handle_gpu_atomic_fault(pGpu, pUvm2MbVaDesc,
                                                pGpuPde, pFaultItem);
        break;
    default:
        UVM_PANIC_ON(1);
        break;
    }

    if (status != NV_OK)
    {
        // Failed to handle the fault, cancel the fault
        pFaultItem->bFatalFault = NV_TRUE;
    }

cleanup:
    if (pUvm2MbVaDesc)
        uvmfull_vat_release_leaf(&pVaspace->vat, &query);
    return status;
}

NV_STATUS uvmfull_alloc_backing_gpu_pages(UvmGpu *pGpu, UvmGpuPde *pGpuPde,
                                            NvU32 pgSize, NvU32 numPages)
{
    NV_STATUS status = NV_OK;
    NvU64 pfn;

    if (pGpuPde->memBackingPresent)
        return NV_OK;

    /*
     * For the first version, we will pre-allocate 2Mb each time
     * so a 2Mb va cannot have a mixed gpu/sysmem allocations
     */
    status = uvmfull_allocate_phys_vid_mem(pGpu, PAGE_SIZE_2M, &pfn);
    if (status != NV_OK)
        return status;

    pGpuPde->memBackingPresent = NV_TRUE;
    pGpuPde->gpuFirstPfn = pfn;

    return NV_OK;
}

void uvmfull_free_backing_gpu_pages(UvmGpu *pGpu, UvmGpuPde *pGpuPde)
{
    nvUvmInterfaceMemoryFreePa(pGpu->gpuHandle,
                               PFN_TO_PA(pGpuPde->gpuFirstPfn));

    pGpuPde->memBackingPresent = NV_FALSE;
    pGpuPde->gpuFirstPfn = 0;
}

NV_STATUS uvmfull_free_backing_sysmem_pages(UvmVaspace *pVaspace,
                                            Uvm2MbVaDesc *pUvm2MbVaDesc,
                                            NvU64 addr, NvU32 numPages)
{
    NvU32 strOffset, i;
    struct page *pg = NULL;
    if (!pVaspace || !pUvm2MbVaDesc)
        return NV_ERR_INVALID_ARGUMENT;

    strOffset = GET_PDE_PTE_OFFSET(addr, PAGE_SHIFT_4K);
    for (i = strOffset; i < numPages; i++)
    {
        if (!pUvm2MbVaDesc->hints[i].cpuPfnValid)
        {
            pg = __pfn_to_page(pUvm2MbVaDesc->hints[i].cpuPfn);
            UVM_PANIC_ON(!pg);
            if (pg)
                __free_page(pg);
            pUvm2MbVaDesc->hints[i].cpuPfn = 0;
            pUvm2MbVaDesc->hints[i].cpuPfnValid = 0;
            pUvm2MbVaDesc->hints[i].cpuPfnMapped = 0;
        }
    }
    return NV_OK;
}

NV_STATUS uvmfull_alloc_backing_sysmem_pages(UvmGpu *pGpu,
        UvmVaspace *pVaspace, Uvm2MbVaDesc *pUvm2MbVaDesc, NvU64 addr,
        NvU32 numPages)
{
    NvU32 strOffset, i;
    NV_STATUS status = NV_OK;
    struct page *pg = NULL;

    if (!pGpu || !pVaspace || !pUvm2MbVaDesc)
        return NV_ERR_INVALID_ARGUMENT;

    strOffset = GET_PDE_PTE_OFFSET(addr, PAGE_SHIFT_4K);
    WARN_ON(strOffset + numPages > GET_GPU_PDE_NUM_PAGES(PAGE_SIZE_4K));

    for (i = strOffset; i < strOffset + numPages; i++)
    {
        if (pUvm2MbVaDesc->hints[i].cpuPfnValid)
            continue;
        pg = alloc_pages(NV_UVM_GFP_FLAGS | GFP_HIGHUSER | __GFP_ZERO, 0);
        if (!pg)
        {
            status = NV_ERR_NO_MEMORY;
            goto cleanup;
        }
        pUvm2MbVaDesc->hints[i].cpuPfn = page_to_pfn(pg);
        pUvm2MbVaDesc->hints[i].cpuPfnValid = 1;
        pUvm2MbVaDesc->hints[i].cpuPfnMapped = 0;
    }
    return NV_OK;
    cleanup: uvmfull_free_backing_sysmem_pages(pVaspace, pUvm2MbVaDesc, addr,
                                               numPages);
    return status;
}

NV_STATUS uvmfull_handle_gpu_write_fault
(
    UvmGpu       *pGpu,
    Uvm2MbVaDesc *pUvm2MbVaDesc,
    UvmGpuPde    *pGpuPde,
    UvmFaultItem *pFaultItem
)
{
    UvmPteHints *pPteHints = NULL;
    NvU64 *pPhysPages = NULL;
    NV_STATUS status = NV_OK;
    UvmGpuPte *curGpuPte = NULL;
    NvU32 curPgSize;
    UvmPbOps *pbOps = NULL;
    UvmPushbuffer *pb = NULL;
    UvmMemType memType;
    UvmFaultBufferEntry *pFault = &pFaultItem->entry;
    NvU64 vaddr = pFault->uvmFaultAddress;
    UvmVaspace *pVaspace = pUvm2MbVaDesc->pVaspace;
    UvmGpuVaspace *pGpuVaspace = &pVaspace->gpuVaspaces[pGpu->gpuId];
    UvmGpuPteUpdateParams pteParams = {0};

    curGpuPte = uvmfull_get_gpupde_current_pte(pGpuPde, vaddr, &curPgSize);
    if (!curGpuPte || !curGpuPte->swcachevalid)
    {
        WARN_ON(!curGpuPte || !curGpuPte->swcachevalid);
        status = NV_ERR_INVALID_STATE;
        goto handler_end;
    }

    // Filter any redudant or invalid faults. Faults can become redudant
    // because when we reached servicing this fault, CPU might have accessed
    // this page and page might have been moved to CPU.
    // Fault Servicing Order - PTE ->Wr->Atomic->Read
    if (!curGpuPte->valid || (curGpuPte->valid && !curGpuPte->readonly))
    {
        status = NV_OK;
        goto handler_end;
    }

    memType = curGpuPte->aperture;
    if (memType == UvmMemType_vidmem)
        if (!pGpuPde->memBackingPresent && !curGpuPte->zfodMapped)
        {
            status = NV_ERR_INVALID_STATE;
            goto handler_end;
        }


    // Make sure peer writes are supported if we have a peer mapping
    if (memType == UvmMemType_peer)
    {
        UvmGlobalState *uvm = uvmfull_get_global();
        UvmGpuPeerCaps * peerCaps = NULL;
        UvmGpu * peerGpu = uvmfull_peerid_to_gpu(pGpu->gpuId,
                                                 curGpuPte->peerid);
        UVM_PANIC_ON(!peerGpu);

        peerCaps = &uvm->uvmPeerCaps[pGpu->gpuId][peerGpu->gpuId];
        if (!peerCaps->writeSupported)
        {
            status = NV_ERR_INVALID_STATE;
            goto handler_end;
        }
    }

    // check the VA properties if this is a writable va
    pPteHints = uvmfull_2mva_get_pte_hints(pUvm2MbVaDesc, vaddr);
    UVM_PANIC_ON(pPteHints == NULL);

    // Return early with an error, this VA region does not have write
    // access
    // TODO: what is the proper error here??
    if (!pPteHints->va_prot_write && pPteHints->managedVa)
    {
        status = NV_ERR_INVALID_STATE;
        goto handler_end;
    }

    pbOps = uvm_pbops_start(pVaspace);
    if (!pbOps)
    {
        status = NV_ERR_NO_MEMORY;
        goto handler_end;
    }
    uvm_pbops_reserve_acquire_space(pbOps, &pUvm2MbVaDesc->tracker);

    // Check if in multi-GPU, do we have another GPU which has atomic mappings
    // if yes we need to mark an atomic GPU as RD-ONLY and then push pageTable
    // update for new mappings
    if (IS_MULTIGPU(pVaspace))
    {
        // There shouldn't be anything to submit at this point but initializing
        // submitmask with faultingGpu to take care of future additions. If
        // nothing is filled, redudant submits are ignored in submit call
        NvU64 gpuSubmitMask = NVBIT64(pGpuPde->hostGpuId);
        status = _downgrade_atomic_peers(pUvm2MbVaDesc, pGpuPde, vaddr,
                                         curPgSize, pbOps, &gpuSubmitMask,
                                         NV_TRUE);

        if (status != NV_OK)
        {
            WARN_ON(status);
            goto handler_end;
        }

        // Submit and complete the semaphore. Also regrab another PB
        status = uvm_pbops_submit_and_release(pbOps, gpuSubmitMask,
                                              &pUvm2MbVaDesc->tracker);
        if (status != NV_OK)
        {
            WARN_ON(status);
            goto handler_end;
        }

        // Again grab pb for next GPU
        pbOps = uvm_pbops_start(pVaspace);
        if (!pbOps)
            return NV_ERR_NO_MEMORY;
        uvm_pbops_reserve_acquire_space(pbOps, &pUvm2MbVaDesc->tracker);
    }

    pb = uvm_pbops_gpu_pb(pbOps, pGpuPde->hostGpuId);
    UVM_PANIC_ON(!pb);

    // get the current page-size and check if this needs to be split down
    // further before returning if we only need to map a sub-section of
    // of current page as write-able
    if(curGpuPte->zfodMapped)
    {
        UVM_PANIC_ON(curGpuPte->aperture != UvmMemType_vidmem);
        if (pteParams.aperture == UvmMemType_vidmem)
        {
            // Allocate newly only if memBacking was not present
            if(!pGpuPde->memBackingPresent)
            {
                status = uvmfull_alloc_backing_gpu_pages(pGpu, pGpuPde,
                                                         PAGE_SIZE_2M, 1);
                if (status != NV_OK)
                    return status;

                uvm_push_memzero_vidmem(pGpu->gpuId, pb,
                                        PFN_TO_PA(pGpuPde->gpuFirstPfn),
                                        PAGE_SIZE_2M);
            }
        }
        // Break the zfod and use implicit 2MB gpu backing or sysmem allocations
        pteParams.updateSwFlags |= UVM_GPU_SW_PTE_UPDATE_FLAGS_PFN;
        pteParams.updateSwFlags |= UVM_GPU_SW_PTE_UPDATE_FLAGS_IMPLICIT_PFN;
        pteParams.pfnvalid = NV_TRUE;
    }

    pteParams.updateSwFlags |= UVM_GPU_SW_PTE_UPDATE_FLAGS_READ_ONLY;
    pteParams.readonly = 0; // not read-only

    // Actually update the PTE
    pteParams.updateHwFlags |= UVM_GPU_HW_PTE_UPDATE_FLAGS_WRITE_PTE;

    // update the pfn in all the GPU ptes
    status = uvmfull_update_gpupde_ptes(pUvm2MbVaDesc, pGpu->gpuId,
            pGpuPde, vaddr, 1, curPgSize, &pteParams,
            pPhysPages, pbOps,
            (NvU8*)&pGpuVaspace->pteDefaults.entry,
            pGpuVaspace->pteDefaults.entrySize,
            NV_FALSE);

    if (status != NV_OK)
    {
        WARN_ON(status);
        goto handler_end;
    }

    status = _gpu_single_fault_epilogue(pGpu, pbOps, pUvm2MbVaDesc, pFaultItem);
handler_end:
    if (pbOps)
        uvm_pbops_end(pbOps);

    return status;
}

static NV_STATUS _downgrade_atomic_peers(Uvm2MbVaDesc *pUvm2MbVaDesc,
                                         UvmGpuPde    *pGpuPde,
                                         NvU64 vaddr,
                                         NvU32 pgSize,
                                         UvmPbOps * pbOps,
                                         NvU64 *gpuSubmitMask,
                                         NvBool bRevokeOnlyAtomic)
{
    NV_STATUS status = NV_OK;
    UvmGpuPde    *pGpuPeerPde = NULL;
    UvmGpuPte    *pGpuPte = NULL;
    UvmGlobalState *uvm = uvmfull_get_global();
    UvmGpuPeerCaps * peerCaps = NULL;
    UvmGpuPteUpdateParams gpteParams = {0};
    UvmVaspace *pVaspace = pUvm2MbVaDesc->pVaspace;
    NvU32 curPtePageSz = 0;
    NvU64 pteVaddr = 0, vaddrLimit = 0;

    // Walk thru every GPU PDe and if all Write mappings need to be revoked
    // revoke them. If only Atomic mappings need to be revoked, it will revoke
    // those
    list_for_each_entry(pGpuPeerPde, &pUvm2MbVaDesc->gpuPDEs, pdeNode)
    {
        // This assumes that even if we don't have any valid PTE in this GPU
        // we will still use this backing & will not allocate new backing
        if (pGpuPeerPde == pGpuPde)
            continue;

        // We might hit a case where this specific PTE is never populated due to
        // pageSize randomization but new PTE is getting mapped to a pageSize
        // which does have some PTEs populated. Consider GPU0 starting its
        // first mapping at 4K and then GPU1 mapping 64K for the first time.
        // With above explanation, do not skip checking all PTEs in any condition
        vaddrLimit = UVM_ALIGN_ADDR(vaddr, pgSize) + (pgSize - 1);
        for (pteVaddr = UVM_ALIGN_ADDR(vaddr, pgSize); pteVaddr < vaddrLimit;
             pteVaddr += curPtePageSz)
        {
            // Walk thru every PTE in this range and make all as read only
            pGpuPte = uvmfull_get_gpupde_current_pte(pGpuPeerPde,
                                                     pteVaddr,
                                                     &curPtePageSz);

            if (!pGpuPte->swcachevalid || !pGpuPte->valid)
            {
                // Take minimum increment of CPU_MIN page size in case we got "0"
                curPtePageSz = pGpuPte->swcachevalid ? curPtePageSz : PAGE_SIZE_4K;
                continue;
            }

            peerCaps = &uvm->uvmPeerCaps[pGpuPde->hostGpuId][pGpuPeerPde->hostGpuId];
            if (!peerCaps->nvlinkSupported &&
                (pGpuPte->atomic || (!pGpuPte->readonly && !bRevokeOnlyAtomic)))
            {
                gpteParams.updateSwFlags |= UVM_GPU_SW_PTE_UPDATE_FLAGS_ATOMIC;
                gpteParams.atomic = NV_FALSE;
                gpteParams.updateSwFlags |= UVM_GPU_SW_PTE_UPDATE_FLAGS_READ_ONLY;
                gpteParams.readonly = NV_TRUE;
                // Actually update the PTE
                gpteParams.updateHwFlags |= UVM_GPU_HW_PTE_UPDATE_FLAGS_WRITE_PTE;

                // This call sends a MemBar whenever it determines a downGrade
                status = uvmfull_update_gpupde_ptes(pUvm2MbVaDesc,
                                                    pGpuPeerPde->hostGpuId,
                                                    pGpuPeerPde, vaddr, 1,
                                                    curPtePageSz, &gpteParams,
                                                    NULL, pbOps,
                                (NvU8*)&pVaspace->gpuVaspaces[pGpuPeerPde->hostGpuId].pteDefaults.entry,
                                pVaspace->gpuVaspaces[pGpuPeerPde->hostGpuId].pteDefaults.entrySize,
                                                    NV_FALSE);
                if (status != NV_OK)
                    return status;

                *gpuSubmitMask |= NVBIT64(pGpuPeerPde->hostGpuId);
            }
        }
    }
    return status;

}

NV_STATUS uvmfull_handle_gpu_atomic_fault
(
    UvmGpu       *pGpu,
    Uvm2MbVaDesc *pUvm2MbVaDesc,
    UvmGpuPde    *pGpuPde,
    UvmFaultItem *pFaultItem
)
{
    NvU32 pgSize;
    UvmPteHints *pPteHints = NULL;
    NV_STATUS status = NV_OK;
    UvmMemType memType;
    UvmPbOps *pbOps = NULL;
    UvmPushbuffer *pb = NULL;
    UvmGpuPte *gpuPte = NULL;
    UvmFaultBufferEntry *pFault = &pFaultItem->entry;
    NvU64 vaddr = pFault->uvmFaultAddress;
    UvmGpuPteUpdateParams gpteParams = {0};
    UvmVaspace *pVaspace = pUvm2MbVaDesc->pVaspace;

    // We found a gpu with populated pages
    gpuPte = uvmfull_get_gpupde_current_pte(pGpuPde, vaddr, &pgSize);
    if (!gpuPte || !gpuPte->swcachevalid)
    {
        WARN_ON(!gpuPte || !gpuPte->swcachevalid);
        status = NV_ERR_INVALID_STATE;
        goto handler_end;
    }

    memType = gpuPte->aperture;
    if (memType == UvmMemType_vidmem && !pGpuPde->memBackingPresent)
    {
        // by the time we have an atomic fault the backing should be allocated
        // if its vidmem
        status = NV_ERR_INVALID_STATE;
        goto handler_end;
    }
    // check the VA properties if this is a writable va
    pPteHints = uvmfull_2mva_get_pte_hints(pUvm2MbVaDesc, vaddr);
    UVM_PANIC_ON(pPteHints == NULL);

    // Filter any redudant or invalid faults. Faults can become redudant
    // because when we reached servicing this fault, CPU/another GPU might have
    // accessed this page and page might have been moved or mappigs have changed.
    if (!gpuPte->valid ||
        (gpuPte->valid &&(gpuPte->readonly || gpuPte->atomic)))
    {
        WARN_ON(gpuPte->valid == 0x0 && (gpuPte->readonly || gpuPte->atomic));
        status = NV_OK;
        goto handler_end;
    }

    pbOps = uvm_pbops_start(pVaspace);
    if (!pbOps)
    {
        status = NV_ERR_NO_MEMORY;
        goto handler_end;
    }
    uvm_pbops_reserve_acquire_space(pbOps, &pUvm2MbVaDesc->tracker);

    // Check if in multi-GPU, do we have another GPU which has atomic mappings
    // if yes we need to unmap the same GPU and then push pageTable update for
    // new mappings
    // If gpu mappings are pointing to sysmem; no handling is needed.
    if (gpuPte->aperture == UvmMemType_sysmem_coh ||
        gpuPte->aperture == UvmMemType_sysmem_ncoh)
    {
        // We unmap the cpu and map the gpu.
        WARN_ON(!gpuPte->valid);
        WARN_ON(!gpuPte->pfnvalid);
        if (pPteHints->cpuPfnMapped == 0x1)
        {
            pPteHints->cpuPfnMapped = 0x0;
            unmap_mapping_range(
                        pVaspace->filep->f_mapping,
                        (vaddr & PAGE_MASK_4K),
                        PAGE_SIZE_4K, 1);
        }
    }

    if (IS_MULTIGPU(pVaspace))
    {
        // There shouldn't be anything to submit at this point but initializing
        // submitmask with faultingGpu to take care of future additions. If
        // nothing is filled, redudant submits are ignored in submit call
        NvU64 gpuSubmitMask = NVBIT64(pGpuPde->hostGpuId);
        status = _downgrade_atomic_peers(pUvm2MbVaDesc, pGpuPde, vaddr,
                                         pgSize, pbOps, &gpuSubmitMask,
                                         NV_FALSE);

        if (status != NV_OK)
        {
            WARN_ON(status);
            goto handler_end;
        }

        // Submit and complete the semaphore. Also regrab another PB
        status = uvm_pbops_submit_and_release(pbOps, gpuSubmitMask,
                                              &pUvm2MbVaDesc->tracker);
         if (status != NV_OK)
         {
             WARN_ON(status);
             goto handler_end;
         }

        // Again grab pb for next GPU
        pbOps = uvm_pbops_start(pVaspace);
        if (!pbOps)
            return NV_ERR_NO_MEMORY;
        uvm_pbops_reserve_acquire_space(pbOps, &pUvm2MbVaDesc->tracker);

    }

    pb = uvm_pbops_gpu_pb(pbOps, pGpuPde->hostGpuId);
    UVM_PANIC_ON(!pb);

    {
        gpteParams.updateSwFlags |= UVM_GPU_SW_PTE_UPDATE_FLAGS_ATOMIC;
        gpteParams.atomic = NV_TRUE;
        // Actually update the PTE
        gpteParams.updateHwFlags |= UVM_GPU_HW_PTE_UPDATE_FLAGS_WRITE_PTE;

        status = uvmfull_update_gpupde_ptes(pUvm2MbVaDesc, pGpu->gpuId,
                            pGpuPde, vaddr, 1, pgSize,
                            &gpteParams,
                            NULL, pbOps,
                            (NvU8*)&pVaspace->gpuVaspaces[pGpu->gpuId].pteDefaults.entry,
                            pVaspace->gpuVaspaces[pGpu->gpuId].pteDefaults.entrySize,
                            NV_FALSE);
        if (status != NV_OK)
        {
            WARN_ON(status);
            goto handler_end;
        }
    }

    status = _gpu_single_fault_epilogue(pGpu, pbOps, pUvm2MbVaDesc, pFaultItem);
handler_end:
    if (pbOps)
        uvm_pbops_end(pbOps);

    return status;
}

NV_STATUS uvmfull_handle_gpu_pde_fault
(
    UvmGpu       *pGpu,
    Uvm2MbVaDesc *pVaDesc,
    UvmGpuPde    *pGpuPde,
    UvmFaultItem *pFaultItem
)
{
    NV_STATUS status = NV_OK;
    UvmPbOps *pbOps = NULL;
    UvmVaspace *pVaspace = pVaDesc->pVaspace;

    pbOps = uvm_pbops_start(pVaspace);
    if (!pbOps)
    {
        status = NV_ERR_NO_MEMORY;
        return status;
    }
    uvm_pbops_reserve_acquire_space(pbOps, &pVaDesc->tracker);

    status = uvmfull_activate_gpupte(pVaDesc, pGpuPde,
                        NV_TRUE, NV_TRUE, pGpu->gpuId,
                        uvm_pbops_gpu_pb(pbOps, pGpu->gpuId));
    if (status != NV_OK)
    {
        pFaultItem->bFatalFault = NV_TRUE;
        if (pbOps)
            uvm_pbops_end(pbOps);
    }
    else     // Call pte fault hanlder inline
    {
        // Frees the pbOps internally.
        status = uvmfull_handle_gpu_pte_fault(pGpu, pVaDesc,
                                                pGpuPde, pFaultItem, pbOps);
    }
    return status;
}

static UvmGpuPde * _get_peer_mem_pde(Uvm2MbVaDesc *pUvm2MbVaDesc,
                                     UvmGpuPde *pSrcGpuPde)
{
    UvmGpuPde * peerBackingGpuPde = NULL;

    // Loop thru all GPUs and see if we have memory present on any GPU
    list_for_each_entry(peerBackingGpuPde, &pUvm2MbVaDesc->gpuPDEs, pdeNode)
    {
        // This assumes that even if we don't have any valid PTE in this GPU
        // we will still use this backing & will not allocate new backing
        if ((peerBackingGpuPde == pSrcGpuPde) ||
            !peerBackingGpuPde->memBackingPresent)
            continue;

        // Do we have peer accesses setup between these GPUs
        if (uvmfull_vaspace_is_peeraccess_valid(pUvm2MbVaDesc->pVaspace,
                                                pSrcGpuPde->hostGpuId,
                                                peerBackingGpuPde->hostGpuId))
            return peerBackingGpuPde;
    }
    return NULL;
}

NV_STATUS uvmfull_unmap_copy_cpu_to_gpu(Uvm2MbVaDesc *pUvm2MbVaDesc,
                                        UvmGpuPde *pGpuPde, UvmPushbuffer *pb,
                                        NvU32 offset, NvU32 cpuPages)
{
    NV_STATUS status = NV_OK;
    NvU32 i;
    UvmVaspace *pVaspace = pUvm2MbVaDesc->pVaspace;

    // Run through all the cpu pages in this gpu va range and queue
    // migration.
    for (i = 0; i < cpuPages; i++)
    {
        if (pUvm2MbVaDesc->hints[offset+i].cpuPfnValid == 0x0)
            continue;  // We copy from zero page when allocating

        // If mapped; unmap
        if (pUvm2MbVaDesc->hints[offset+i].cpuPfnMapped == 0x1)
        {
            pUvm2MbVaDesc->hints[offset+i].cpuPfnMapped = 0x0;
            unmap_mapping_range(pVaspace->filep->f_mapping,
                   pUvm2MbVaDesc->addr + (UVM_CPU_MIN_PAGE_SIZE * (offset+i)),
                   UVM_CPU_MIN_PAGE_SIZE, 1);
        }

        // if dirty; copy
        if (pUvm2MbVaDesc->hints[offset+i].cpuPfnDirty == 0x1)
        {
            NvU64 src = PFN_TO_PA(pUvm2MbVaDesc->hints[offset+i].cpuPfn);
            NvU64 dst = PFN_TO_PA(pGpuPde->gpuFirstPfn);
            dst += (UVM_CPU_MIN_PAGE_SIZE * (offset + i));
            uvm_push_cpu_to_gpu_copy(pb, src, dst, UVM_CPU_MIN_PAGE_SIZE,
                                     NV_UVM_COPY_SRC_TYPE_PHYSICAL |
                                     NV_UVM_COPY_DST_TYPE_PHYSICAL);
            pUvm2MbVaDesc->hints[offset+i].cpuPfnDirty = 0x0;

            // Fill migration event if enabled
            if (uvm_is_event_enabled(pVaspace->pEventContainer,
                                     UvmEventTypeMigration))
            {
                // FIXME: Right now we have we have no way to get precise
                // time for it we need to change it when we have a precise
                // way to get timing from task in pushbuffer
                uvm_record_migration_event(
                          pVaspace->pEventContainer,
                          UvmEventMigrationDirectionCpuToGpu,
                          pGpuPde->hostGpuId,
                          0,
                          pUvm2MbVaDesc->addr +
                          (UVM_CPU_MIN_PAGE_SIZE * (offset+i)),
                          UVM_CPU_MIN_PAGE_SIZE,
                          NV_GETTIME(), NV_GETTIME(), 0);
            }
            uvm_increment_process_counters(pGpuPde->hostGpuId,
                                           pVaspace->pCounterContainer,
                                           UvmCounterNameBytesXferHtD,
                                           UVM_CPU_MIN_PAGE_SIZE);
        }
    }

    return status;
}

static NV_STATUS _handle_gpu_peer_migrations
(
    UvmGpu       *pGpu,
    Uvm2MbVaDesc *pUvm2MbVaDesc,
    UvmGpuPde    *pGpuPde,
    UvmPbOps     *pbOps,
    NvU64         vaddr
)
{
    NV_STATUS status = NV_OK;
    UvmVaspace *pVaspace = pUvm2MbVaDesc->pVaspace;
    UvmGpuPde *pPeerGpuPde;
    UvmGpuPte *pPeerGpuPte;
    UvmGpu *pCopyGpu = NULL;
    UvmGpuPte *pCachedPeerGpuPte = NULL;
    UvmGlobalState *uvm = uvmfull_get_global();
    UvmGpu *pPeerGpu = NULL;
    NvU32 pgSize, cachedPeerId = 0;
    UvmGpuVaspace *pPeerGpuVaspace = NULL;
    NvU64 adjustedGpuPfn;
    NvU32 peerGpuMask = 0;
    NvU32 submitGpuMask = 0;
    UvmGpuPeerCaps * p2pCaps = NULL;
    UvmGpuPteUpdateParams gpteParams = { 0 };
    UvmPushbuffer *pb;
    NvU64 *pPhysPages = NULL;

    // ******* Step 1 ********
    // Unmap the page from all peer PDEs as we are changing the
    // location of a page.
    desc_for_each_gpu_pde(pPeerGpuPde, pUvm2MbVaDesc)
    {
        if (pPeerGpuPde == pGpuPde)
            continue;

        pPeerGpuPte = uvmfull_get_gpupde_current_pte(pPeerGpuPde, vaddr,
            &pgSize);
        if (!pPeerGpuPte || !pPeerGpuPte->swcachevalid)
            continue;

        // cache peerPde and Pte for future use -(peer read).
        // Find out a peer gpu with valid pfn and
        // read access enabled.
        if (!pPeerGpuPte->valid)
        {
            continue;
        }
        else
        {
            UVM_PANIC_ON(!pPeerGpuPte->pfnvalid);
            // Migration will happen only if there is a vidmem mapping present.
            if (pPeerGpuPte->aperture == UvmMemType_vidmem ||
                pPeerGpuPte->aperture == UvmMemType_peer)
            {
                p2pCaps = &uvm->uvmPeerCaps[pPeerGpuPde->hostGpuId][pGpuPde->hostGpuId];
                if (!pCachedPeerGpuPte &&
                    p2pCaps &&
                    p2pCaps->readSupported &&
                    p2pCaps->writeSupported
                    )
                {
                    // cache pte - to be used for migration
                    pPeerGpu = &uvm->uvmGpus[pPeerGpuPde->hostGpuId];
                    if (pPeerGpuPte->aperture == UvmMemType_peer)
                    {

                        pCopyGpu = uvmfull_peerid_to_gpu(pPeerGpu->gpuId,
                            pPeerGpuPte->peerid);

                        // Check if the copy gpu is p2p configured with the gpu at fault.
                        // This path will get exercised in 2+ GPU configs.
                        if (pCopyGpu != pGpu)
                        {
                            p2pCaps = &uvm->uvmPeerCaps[pCopyGpu->gpuId][pGpuPde->hostGpuId];
                            if (!p2pCaps ||
                                !p2pCaps->readSupported ||
                                !p2pCaps->writeSupported)
                                continue;
                        }
                    }
                    else
                        pCopyGpu = pPeerGpu;

                    pCachedPeerGpuPte = pPeerGpuPte;
                    cachedPeerId = p2pCaps->gpuPeerId;
                }
            }
        }

        memset(&gpteParams, 0, sizeof(gpteParams));
        gpteParams.updateSwFlags |=
            UVM_GPU_SW_PTE_UPDATE_FLAGS_PTE_VALID;
        gpteParams.valid = NV_FALSE;
        gpteParams.updateSwFlags |=
            UVM_GPU_SW_PTE_UPDATE_FLAGS_VOLATILE;
        gpteParams.vol = NV_FALSE;
        gpteParams.updateSwFlags |=
            UVM_GPU_SW_PTE_UPDATE_FLAGS_ATOMIC;
        gpteParams.atomic = NV_FALSE;
        gpteParams.updateSwFlags |=
            UVM_GPU_SW_PTE_UPDATE_FLAGS_READ_ONLY;
        gpteParams.readonly = NV_FALSE;

        // Actually update the PTE
        gpteParams.updateHwFlags |=
            UVM_GPU_HW_PTE_UPDATE_FLAGS_WRITE_PTE;

        pPeerGpu = &uvm->uvmGpus[pPeerGpuPde->hostGpuId];
        pPeerGpuVaspace =
            &pVaspace->gpuVaspaces[pPeerGpuPde->hostGpuId];

        // update the pfn in all the GPU ptes
        status = uvmfull_update_gpupde_ptes(pUvm2MbVaDesc,
                                            pPeerGpu->gpuId, pPeerGpuPde,
                                            vaddr,
                                            1,
                                            pgSize,
                                            &gpteParams,
                                            NULL, pbOps,
                                            (NvU8*)&pPeerGpuVaspace->pteDefaults.entry,
                                            pPeerGpuVaspace->pteDefaults.entrySize,
                                            NV_FALSE);

        if (status != NV_OK)
        {
            WARN_ON(status);
            goto cleanup;
        }

        peerGpuMask |= NVBIT64(pPeerGpu->gpuId);
    }

    if (peerGpuMask)
    {
        submitGpuMask = peerGpuMask;

        // ******* Step 2 ********
        // migrate using pGpuPde (peer read)
        // submit pushbuffer for all peer gpus which are being unampped.
        status = uvm_pbops_submit_and_release(pbOps, (submitGpuMask | NVBIT64(pGpu->gpuId)),
            &pUvm2MbVaDesc->tracker);

        if (status != NV_OK)
            goto cleanup;

        adjustedGpuPfn = pGpuPde->gpuFirstPfn +
            (GET_PDE_PTE_OFFSET(UVM_ALIGN_ADDR(vaddr, pgSize), PFN_SHIFT));

        // Copy only if copy gpu is diff than the gpu at fault.
        if (pCopyGpu != pGpu && pCachedPeerGpuPte)
        {
            // Again grab pb for next submissions
            pbOps = uvm_pbops_start(pVaspace);
            if (!pbOps)
            {
                status = NV_ERR_NO_MEMORY;
                goto cleanup;
            }
            uvm_pbops_reserve_acquire_space(pbOps, &pUvm2MbVaDesc->tracker);

            // get a pushbuffer
            pb = uvm_pbops_gpu_pb(pbOps, pCopyGpu->gpuId);
            if (!pb)
            {
                status = NV_ERR_NO_MEMORY;
                goto cleanup;
            }

            // Issue a peer GPU write. (Peer GPU pushes data)
            uvm_push_copy(pb,
                PFN_TO_PA(pCachedPeerGpuPte->pfn),
                GET_PEERMAP_VA_FROM_PA(cachedPeerId, PFN_TO_PA(adjustedGpuPfn)),
                pgSize,
                (NV_UVM_COPY_SRC_TYPE_PHYSICAL |
                NV_UVM_COPY_DST_TYPE_VIRTUAL),
                NV_UVM_COPY_SRC_LOCATION_FB,
                NV_UVM_COPY_DST_LOCATION_FB);

            status = uvm_pbops_submit_and_release(pbOps, NVBIT64(pCopyGpu->gpuId),
                &pUvm2MbVaDesc->tracker);

            if (status != NV_OK)
                goto cleanup;
        }

        // Again grab pb for next submissions
        pbOps = uvm_pbops_start(pVaspace);
        if (!pbOps)
        {
            status = NV_ERR_NO_MEMORY;
            goto cleanup;
        }
        uvm_pbops_reserve_acquire_space(pbOps, &pUvm2MbVaDesc->tracker);

        // ******* Step 3 ********
        // Remap whatever was unmapped in step 1 as RO.
        desc_for_each_gpu_pde(pPeerGpuPde, pUvm2MbVaDesc)
        {
            if (pPeerGpuPde == pGpuPde)
                continue;

            pPeerGpu = &uvm->uvmGpus[pPeerGpuPde->hostGpuId];

            if (!(peerGpuMask & NVBIT64(pPeerGpu->gpuId)))
                continue;

            peerGpuMask &= ~NVBIT64(pPeerGpu->gpuId);

            p2pCaps = &uvm->uvmPeerCaps[pPeerGpuPde->hostGpuId][pGpuPde->hostGpuId];
            if (p2pCaps &&
                p2pCaps->readSupported &&
                p2pCaps->writeSupported)
            {
                /*update ptes*/
                memset(&gpteParams, 0, sizeof(gpteParams));

                gpteParams.updateSwFlags |=
                    UVM_GPU_SW_PTE_UPDATE_FLAGS_PFN_AUTOINCREMENT;
                pPhysPages = &adjustedGpuPfn;

                gpteParams.pfnvalid = NV_TRUE;
                gpteParams.updateSwFlags |= UVM_GPU_SW_PTE_UPDATE_FLAGS_PFN_VALID;
                gpteParams.updateSwFlags |= UVM_GPU_SW_PTE_UPDATE_FLAGS_PFN;

                // Peer Mappings settings
                gpteParams.peerid = p2pCaps->gpuPeerId;
                gpteParams.updateSwFlags |= UVM_GPU_SW_PTE_UPDATE_FLAGS_PEERID;

                gpteParams.updateSwFlags |= UVM_GPU_SW_PTE_UPDATE_FLAGS_SWVALID;
                gpteParams.swcachevalid = NV_TRUE;

                gpteParams.updateSwFlags |= UVM_GPU_SW_PTE_UPDATE_FLAGS_APERTURE;
                gpteParams.aperture = UvmMemType_peer;

                gpteParams.updateSwFlags |= UVM_GPU_SW_PTE_UPDATE_FLAGS_PTE_VALID;
                gpteParams.valid = NV_TRUE;

                gpteParams.updateSwFlags |= UVM_GPU_SW_PTE_UPDATE_FLAGS_VOLATILE;
                gpteParams.vol = NV_TRUE;

                gpteParams.updateSwFlags |=
                    UVM_GPU_SW_PTE_UPDATE_FLAGS_READ_ONLY;
                gpteParams.readonly = NV_TRUE;

                // Actually update the PTE
                gpteParams.updateHwFlags |= UVM_GPU_HW_PTE_UPDATE_FLAGS_WRITE_PTE;

                pPeerGpuVaspace =
                    &pVaspace->gpuVaspaces[pPeerGpuPde->hostGpuId];

                // update the pfn in all the GPU ptes
                status = uvmfull_update_gpupde_ptes(pUvm2MbVaDesc,
                    pPeerGpu->gpuId, pPeerGpuPde,
                    vaddr,
                    1,
                    pgSize,
                    &gpteParams,
                    pPhysPages, pbOps,
                    (NvU8*)&pPeerGpuVaspace->pteDefaults.entry,
                    pPeerGpuVaspace->pteDefaults.entrySize,
                    NV_FALSE);

                if (status != NV_OK)
                {
                    WARN_ON(status);
                    goto cleanup;
                }
            }
        }

        // submit pushbuffer of a last gpu
        status = uvm_pbops_submit_and_release(pbOps, submitGpuMask,
            &pUvm2MbVaDesc->tracker);

        if (status != NV_OK)
            goto cleanup;

        // Again grab pb for next submissions
        pbOps = uvm_pbops_start(pVaspace);
        if (!pbOps)
        {
            status = NV_ERR_NO_MEMORY;
            goto cleanup;
        }
        uvm_pbops_reserve_acquire_space(pbOps, &pUvm2MbVaDesc->tracker);
    }

    return status;

cleanup:
    if (pbOps)
    {
        uvm_pbops_end(pbOps);
        pbOps = NULL;
    }
    return status;
}


NV_STATUS uvmfull_handle_gpu_pte_fault
(
    UvmGpu       *pGpu,
    Uvm2MbVaDesc *pUvm2MbVaDesc,
    UvmGpuPde    *pGpuPde,
    UvmFaultItem *pFaultItem,
    UvmPbOps     *pbOps
)
{
    NvU64 *pPhysPages = NULL;
    NV_STATUS status = NV_OK;
    NvU32 cpuPages = 0;
    UvmVaspace *pVaspace = pUvm2MbVaDesc->pVaspace;
    UvmMemType suggestedAperture = UvmMemType_vidmem;
    UvmFaultBufferEntry *pFault = &pFaultItem->entry;
    NvU64 vaddr = pFault->uvmFaultAddress;
    NvU32 offsetFault = GET_PDE_PTE_OFFSET(vaddr, UVM_CPU_MIN_PAGE_SHIFT);
    NvU32 offsetMapStart;
    NvU32 preferredPageSize;
    UvmGpuPteUpdateParams pteParams = {0};
    UvmGpuPte *curGpuPte;
    NvU32 curPgSize;
    UvmPushbuffer *pb;
    NvU64 zfodPfn;
    NvBool bCreateZfodMapping = NV_FALSE;
    UvmGpuPde *peerBackingGpuPde = NULL, *memBackingGpuPde = NULL;
    UvmGpuPeerCaps * peerCaps = NULL;
    UvmGpuVaspace *gpuVaspace = &pVaspace->gpuVaspaces[pGpu->gpuId];

    NvU64 curr_time = NV_GETTIME();

    _create_gpte_params_from_hints(pFault, gpuVaspace, pUvm2MbVaDesc, pGpuPde,
                 &pUvm2MbVaDesc->hints[offsetFault], &pteParams, &preferredPageSize);
    suggestedAperture = pteParams.aperture;

    if (!pUvm2MbVaDesc->hints[offsetFault].managedVa)
    {
        UVM_PANIC_ON(1 && "No support for HMM yet.");
        goto handler_end;
    }

    // If P2PMigration is set, always map a gpu at fault
    // with vidmem aperture.
    if (!gpuVaspace->policy.bForceP2PMigration &&
        IS_MULTIGPU(pVaspace))
    {
        // Check if we have a peer allocation on another GPU
        if (suggestedAperture != UvmMemType_sysmem_coh &&
            suggestedAperture != UvmMemType_sysmem_ncoh)
        {
            peerBackingGpuPde = _get_peer_mem_pde(pUvm2MbVaDesc, pGpuPde);
            if (peerBackingGpuPde)
            {
                UvmGlobalState *uvm = uvmfull_get_global();
                peerCaps = &uvm->uvmPeerCaps[pGpuPde->hostGpuId][peerBackingGpuPde->hostGpuId];
                if (peerCaps->readSupported && peerCaps->writeSupported)
                {
                    pteParams.aperture = UvmMemType_peer;
                    suggestedAperture = UvmMemType_peer;
                }
            }
        }
    }

    curGpuPte = uvmfull_get_gpupde_current_pte(pGpuPde, vaddr, &curPgSize);
    if (curGpuPte->swcachevalid)
    {
        if (curGpuPte->valid)
        {
            status = NV_OK;
            goto handler_end;
        }
    }
    else
        curPgSize = preferredPageSize;   // start with policy

    // Overwrite pageSize with availble peer page size!
    if (gpuVaspace->policy.bForceP2PMigration &&
        IS_MULTIGPU(pVaspace))
    {
        // For peer migrations always work on the existing
        // mapped paged size instead of pageSize at fault.
        // If there is any peer which is already present pick
        // its page size. This will mirror mappings on peer GPUs
        // makes migration case easy to handle.
        // TODO - Production driver can implement aggresive migration.
        // where it will parse each PTE of interest and copy data from
        // necessary PTE->aperture.
        peerBackingGpuPde = _get_peer_mem_pde(pUvm2MbVaDesc, pGpuPde);
        if (peerBackingGpuPde)
        {
            UvmGlobalState *uvm = uvmfull_get_global();
            peerCaps = &uvm->uvmPeerCaps[pGpuPde->hostGpuId][peerBackingGpuPde->hostGpuId];
            if (peerCaps->readSupported && peerCaps->writeSupported)
            {
                // Update curPageSize with peer's page size
                // Note - In migration case its safe to query peer with a backing-store.
                // because all the mapped/unmapped peer are expected to have backing-store.
                NvU32 peerPtePgSize = 0;
                UvmGpuPte *peerPte = uvmfull_get_gpupde_current_pte(peerBackingGpuPde,
                    vaddr,
                    &peerPtePgSize);
                if (peerPte && peerPte->swcachevalid)
                {
                    curPgSize = peerPtePgSize;
                }
            }
        }
    }

    if (!pbOps)
    {
        pbOps = uvm_pbops_start(pVaspace);
        if (!pbOps)
        {
            status = NV_ERR_NO_MEMORY;
            goto handler_end;
        }
        uvm_pbops_reserve_acquire_space(pbOps, &pUvm2MbVaDesc->tracker);
    }

    // Set gpuPde & PBs correctly if we have need to create peer mappings
    memBackingGpuPde = (suggestedAperture == UvmMemType_peer) ?
                        peerBackingGpuPde : pGpuPde;

    pb = uvm_pbops_gpu_pb(pbOps, memBackingGpuPde->hostGpuId);
    UVM_PANIC_ON(!pb);

    // Make sure backing pages are present, if not alloc for pte-fault
    if (suggestedAperture == UvmMemType_vidmem ||
        suggestedAperture == UvmMemType_peer)
    {
        cpuPages = curPgSize / UVM_CPU_MIN_PAGE_SIZE;
        offsetMapStart = GET_PDE_PTE_OFFSET(UVM_ALIGN_ADDR(vaddr, curPgSize),
                                       UVM_CPU_MIN_PAGE_SHIFT);

        // Notify new fault to performance heuristics subsystem
        uvmfull_notify_fault(pUvm2MbVaDesc, offsetFault, offsetMapStart, cpuPages, pGpu->gpuId,
                             pFault->uvmFaultType, pFault->uvmFaultAccessType, curr_time);

        // If the membacking is not there either on the gpu or cpu, this is a
        // first-touch, We can use zfod mapping.
        if (!memBackingGpuPde->memBackingPresent &&
            gpuVaspace->bTestZfodEnabled &&
            pFault->uvmFaultAccessType == UvmAccessType_read)
        {
            // zfod candidate
            NvU32 i;
            bCreateZfodMapping = NV_TRUE;
            UVM_PANIC_ON(pGpu->bTestZfodPfnValid == NV_FALSE);

            for (i = 0; i < cpuPages; i++)
            {
                 if (pUvm2MbVaDesc->hints[offsetMapStart+i].cpuPfnValid == 0x0 ||
                     pUvm2MbVaDesc->hints[offsetMapStart+i].cpuPfnDirty == 0x0)
                     continue;

                 bCreateZfodMapping = NV_FALSE;
                 break;
            }
        }

        if (bCreateZfodMapping)
        {
            zfodPfn = PFN_TO_PA(pGpu->testZfodPfn) +
                                (UVM_CPU_MIN_PAGE_SIZE * offsetMapStart);
            zfodPfn = PA_TO_PFN(zfodPfn);
            // Assuming numPages = 1
            pPhysPages = &zfodPfn;
        }
        else  // Generic non-zfod case.
        {
            if (!memBackingGpuPde->memBackingPresent)
            {
                status = uvmfull_alloc_backing_gpu_pages(pGpu, memBackingGpuPde,
                                                         PAGE_SIZE_2M, 1);
                if (status != NV_OK)
                {
                    goto handler_end;
                }
                uvm_push_memzero_vidmem(memBackingGpuPde->hostGpuId,
                                        pb,
                                        PFN_TO_PA(memBackingGpuPde->gpuFirstPfn),
                                        PAGE_SIZE_2M);
            }

            // Mem-backing is allocated. Update it to latest data.
            status = uvmfull_unmap_copy_cpu_to_gpu(pUvm2MbVaDesc,
                                            memBackingGpuPde, pb, offsetMapStart,
                                            cpuPages);
            if (status != NV_OK)
            {
                goto handler_end;
            }
        }
    }
    else // sysmem
    {
        cpuPages = (1 * (curPgSize / UVM_CPU_MIN_PAGE_SIZE));
        status = uvmfull_alloc_backing_sysmem_pages(pGpu, pVaspace,
                                   pUvm2MbVaDesc, vaddr, cpuPages);
        if (status != NV_OK)
        {
            WARN_ON(status);
            goto handler_end;
        }
    }

    // If suggested aperture is Peer, close PB as pageTable updates
    // will be pushed on peer GPU
    if (IS_MULTIGPU(pVaspace))
    {
        // Trigger migration
        if (gpuVaspace->policy.bForceP2PMigration)
        {
            status = _handle_gpu_peer_migrations(pGpu, pUvm2MbVaDesc,
                pGpuPde, pbOps, vaddr);
            if (status != NV_OK)
            {
                WARN_ON(status);
                goto handler_end;
            }
        }
        else
        {
            // We need to perform submissions on faulting GPU as PDE mappings might
            // have been pushed in PB as well as memBacking GPU due to CPU copies
            NvU64 gpuSubmitMask = NVBIT64(pGpuPde->hostGpuId) |
                NVBIT64(memBackingGpuPde->hostGpuId);

            // Get rid of all write/atomic mappings as we will be mapping
            // this page as atomic
            status = _downgrade_atomic_peers(pUvm2MbVaDesc,
                                            pGpuPde, vaddr,
                                            curPgSize, pbOps,
                                            &gpuSubmitMask,
                                            (pteParams.aperture == UvmMemType_vidmem ||
                                            (pteParams.aperture == UvmMemType_peer)) ?
                                            NV_FALSE : NV_TRUE);
            if (status != NV_OK)
            {
                WARN_ON(status);
                goto handler_end;
            }

            status = uvm_pbops_submit_and_release(pbOps, gpuSubmitMask,
                &pUvm2MbVaDesc->tracker);

            if (status != NV_OK)
            {
                WARN_ON(status);
                goto handler_end;
            }

            // Again grab pb for next GPU
            pbOps = uvm_pbops_start(pVaspace);
            if (!pbOps)
                return NV_ERR_NO_MEMORY;
            uvm_pbops_reserve_acquire_space(pbOps, &pUvm2MbVaDesc->tracker);
        }
    }
    {
        if (bCreateZfodMapping)
        {
            // use PFN_AUTOINCREMENT to set the appropriate pfn
            pteParams.updateSwFlags |=
                            UVM_GPU_SW_PTE_UPDATE_FLAGS_PFN_AUTOINCREMENT;
            pteParams.updateSwFlags |=
                            UVM_GPU_SW_PTE_UPDATE_FLAGS_ZFOD_MAPPED;
            pteParams.zfodMapped = NV_TRUE;
            pteParams.updateSwFlags |=
                            UVM_GPU_SW_PTE_UPDATE_FLAGS_READ_ONLY;
            pteParams.readonly = NV_TRUE;
        }
        else  // Non ZFOD case
        {
            //
            // Only vidmem or Peer allocations are contiguous in this version
            // so for vidmem allocations mark pfn autoincrement
            if ((suggestedAperture == UvmMemType_peer) && peerCaps &&
                (peerCaps->writeSupported || peerCaps->readSupported))
            {
                NvU64 firstPhysAddr = memBackingGpuPde->gpuFirstPfn +
                                      (GET_PDE_PTE_OFFSET(UVM_ALIGN_ADDR(vaddr,
                                                                         curPgSize),
                                                                         PFN_SHIFT));
                pteParams.updateSwFlags |=
                        UVM_GPU_SW_PTE_UPDATE_FLAGS_PFN_AUTOINCREMENT;
                pPhysPages = &firstPhysAddr;
                pteParams.peerid = peerCaps->gpuPeerId;
                pteParams.updateSwFlags |=
                                    UVM_GPU_SW_PTE_UPDATE_FLAGS_PEERID;
                pteParams.updateSwFlags |= UVM_GPU_SW_PTE_UPDATE_FLAGS_VOLATILE;
                pteParams.vol = NV_TRUE;
                pteParams.updateSwFlags |= UVM_GPU_SW_PTE_UPDATE_FLAGS_ATOMIC;
                pteParams.atomic = NV_TRUE;
            }
            else if (suggestedAperture == UvmMemType_vidmem)
            {
                // Implicit pfn means that the 2Mb allocation stored in the pde
                // will be used for vidmem. In sysmem; the cpuPfn stored will
                // be used.
                pteParams.updateSwFlags |=
                                    UVM_GPU_SW_PTE_UPDATE_FLAGS_IMPLICIT_PFN;

                // Also if vidmem, always mark atomic
                pteParams.updateSwFlags |= UVM_GPU_SW_PTE_UPDATE_FLAGS_ATOMIC;
                pteParams.atomic = NV_TRUE;
            }
            else if (suggestedAperture == UvmMemType_sysmem_coh ||
                     suggestedAperture == UvmMemType_sysmem_ncoh)
            {
                pteParams.updateSwFlags |=
                                    UVM_GPU_SW_PTE_UPDATE_FLAGS_IMPLICIT_PFN;
                pteParams.updateSwFlags |= UVM_GPU_SW_PTE_UPDATE_FLAGS_VOLATILE;
                pteParams.vol = NV_TRUE;
            }
            else
                UVM_PANIC_ON("Invalid Aperture");
        }

        pteParams.updateSwFlags |= UVM_GPU_SW_PTE_UPDATE_FLAGS_PFN_VALID;
        pteParams.updateSwFlags |= UVM_GPU_SW_PTE_UPDATE_FLAGS_PFN;
        pteParams.pfnvalid = NV_TRUE;

        pteParams.updateSwFlags |= UVM_GPU_SW_PTE_UPDATE_FLAGS_SWVALID;
        pteParams.swcachevalid = NV_TRUE;

        pteParams.updateSwFlags |= UVM_GPU_SW_PTE_UPDATE_FLAGS_APERTURE;
        pteParams.aperture = suggestedAperture;

        pteParams.updateSwFlags |= UVM_GPU_SW_PTE_UPDATE_FLAGS_PTE_VALID;
        pteParams.valid = NV_TRUE;

        // Actually update the PTE
        pteParams.updateHwFlags |= UVM_GPU_HW_PTE_UPDATE_FLAGS_WRITE_PTE;
    }
    // update the pfn in all the GPU ptes
    status = uvmfull_update_gpupde_ptes(pUvm2MbVaDesc, pGpu->gpuId,
                    pGpuPde, vaddr, 1, curPgSize,
                    &pteParams, pPhysPages, pbOps,
                    (NvU8*)&pVaspace->gpuVaspaces[pGpu->gpuId].pteDefaults.entry,
                    pVaspace->gpuVaspaces[pGpu->gpuId].pteDefaults.entrySize,
                    NV_FALSE);

    if (status != NV_OK)
    {
        WARN_ON(status);
        goto handler_end;
    }

    status = _gpu_single_fault_epilogue(pGpu, pbOps, pUvm2MbVaDesc, pFaultItem);

handler_end:
    if (pbOps)
        uvm_pbops_end(pbOps);

    return status;
}

NV_STATUS uvmfull_allocate_phys_sys_mem(NvU32 numPages, NvU64 **ppPhysPageList)
{
    NvU32 i;
    struct page *pg = NULL;
    NV_STATUS status = NV_OK;

    if (!ppPhysPageList || !numPages)
        return NV_ERR_INVALID_ARGUMENT;

    *ppPhysPageList = vmalloc(numPages * sizeof(NvU64));
    if (!*ppPhysPageList)
        return NV_ERR_NO_MEMORY;

    memset(*ppPhysPageList, 0, (numPages * sizeof(NvU64)));
    for (i = 0; i < numPages; i++)
    {
        pg = alloc_page(NV_UVM_GFP_FLAGS);
        if (!pg)
        {
            status = NV_ERR_NO_MEMORY;
            goto cleanup;
        }
        (*ppPhysPageList)[i] = (NvU64) page_to_pfn(pg);
    }
    return status;

    cleanup: for (i = 0; i < numPages; i++)
    {
        if ((*ppPhysPageList)[i])
            __free_page(pfn_to_page((*ppPhysPageList)[i])); //FIXIT!
    }
    vfree(*ppPhysPageList);
    *ppPhysPageList = NULL;
    return status;
}

NV_STATUS uvmfull_allocate_phys_vid_mem(UvmGpu *pGpu, NvU64 size, NvU64 *pPfn)
{
    UvmGpuAllocInfo gpuAllocPaInfo;
    UvmGpuPointer gpuPtr;
    NV_STATUS status;

    memset(&gpuAllocPaInfo, 0, sizeof(gpuAllocPaInfo));
    gpuAllocPaInfo.bContiguousPhysAlloc = NV_TRUE;

    status = nvUvmInterfaceMemoryAllocGpuPa(pGpu->gpuHandle,
                                size, &gpuPtr, &gpuAllocPaInfo);
    if (status == NV_OK)
    {
        *pPfn = PA_TO_PFN(gpuAllocPaInfo.gpuPhysOffset);
    }

    return status;
}
void uvmfull_free_phys_vid_mem(UvmGpu *pGpu, NvU64 pfn)
{
    nvUvmInterfaceMemoryFreePa(pGpu->gpuHandle, PFN_TO_PA(pfn));
}

static UvmGpuPte* _gpu_pte_adjust(UvmVaspace *vaspace, Uvm2MbVaDesc *desc2M,
                       UvmGpuPde *pde, UvmPbOps *pbOps, NvU64 faultingAddress,
                       NvU32 pgSizeIn, NvBool bWriteAccess, NvU32 *pgSizeOut)
{
    NvU32 ptePgSize;
    UvmGpuPte* gpuPte;
    NvBool bPteReadOnly;
    NV_STATUS status;

    gpuPte = uvmfull_get_gpupde_current_pte(pde, faultingAddress, &ptePgSize);
    if (!gpuPte->swcachevalid || !gpuPte->valid)
        return gpuPte;

    bPteReadOnly = (gpuPte->readonly == 1);
    if (bPteReadOnly && !bWriteAccess)
        return gpuPte;

    status = uvmfull_split_gpupde_pte(vaspace, pde->hostGpuId, desc2M, pde,
                            faultingAddress, pgSizeIn,
                            NV_UVM_SPLIT_MERGE_FLAGS_SPLIT_FIXED_PAGE_SIZE,
                            pbOps);
    UVM_PANIC_ON(status != NV_OK);

    gpuPte = uvmfull_get_gpupde_current_pte(pde, faultingAddress, pgSizeOut);
    UVM_PANIC_ON(!gpuPte);

    return gpuPte;
}

// Updates the gpu page state for the given cpu fault.
static NV_STATUS _cpu_fault_update_gpu(UvmVaspace *vaspace,
                                       Uvm2MbVaDesc *desc2M,
                                       NvU64 faultingAddress, NvU64 cpuPfn,
                                       NvBool bWriteAccess, NvBool bCpuPfnMapped)
{
    NV_STATUS status = NV_OK;
    UvmGpuPde *pde;
    UvmGpuPde *cachedPde = NULL;
    UvmGpuPte *gpuPte;
    NvU32 pgSize;
    UvmPushbuffer *pushbuffer = NULL;
    NvU64 src;
    NvU64 dst;
    UvmGpu *gpu;
    UvmGlobalState *uvm = uvmfull_get_global();
    UvmVaspace *pVaspace = desc2M->pVaspace;
    NvBool recordMigrationEvent = NV_FALSE;
    NvBool migrationOccurs = NV_FALSE;
    NvU64 beginTime = 0;
    NvU64 endTime = 0;
    NvU64 gpuId = 0;
    UvmPbOps *pbOps = NULL;
    NvU32 copyFlags = NV_UVM_COPY_SRC_TYPE_PHYSICAL |
                      NV_UVM_COPY_DST_TYPE_PHYSICAL;
    NvU64 pbAttachGpuMask = 0;
    NvBool bCopyNeeded = NV_FALSE;

    // Gather pushbuffers
    pbOps = uvm_pbops_start(vaspace);
    if (!pbOps)
        return NV_ERR_NO_MEMORY;

    uvm_pbops_reserve_acquire_space(pbOps, &desc2M->tracker);

    if (uvm_is_event_enabled(vaspace->pEventContainer, UvmEventTypeMigration))
    {
        recordMigrationEvent = NV_TRUE;
        beginTime = NV_GETTIME();
    }

    desc_for_each_gpu_pde(pde, desc2M)
    {
        UvmGpuVaspace *gpuVaspace = &vaspace->gpuVaspaces[pde->hostGpuId];
        // We found a gpu with populated pages
        gpuPte = uvmfull_get_gpupde_current_pte(pde, faultingAddress,
                                                &pgSize);
        if (!gpuPte || !gpuPte->swcachevalid)
            continue;

        gpu = &uvm->uvmGpus[pde->hostGpuId];
        pushbuffer = uvm_pbops_gpu_pb(pbOps, pde->hostGpuId);

        if (gpuPte->aperture == UvmMemType_sysmem_coh ||
            gpuPte->aperture == UvmMemType_sysmem_ncoh)
        {
            if (gpuPte->valid == 1 && gpuPte->atomic == 1)
            {
                UvmGpuPteUpdateParams gpteParams = {0};
                UVM_PANIC_ON(bCpuPfnMapped);
                gpteParams.updateSwFlags |= UVM_GPU_SW_PTE_UPDATE_FLAGS_ATOMIC;
                gpteParams.atomic = NV_FALSE;
                // Actually update the PTE
                gpteParams.updateHwFlags |= UVM_GPU_HW_PTE_UPDATE_FLAGS_WRITE_PTE;

                // update the pfn in all the GPU ptes
                uvmfull_update_gpupde_ptes(desc2M, gpu->gpuId,
                                pde, faultingAddress, 1, pgSize,
                                &gpteParams,
                                NULL, pbOps,
                                (NvU8*)&gpuVaspace->pteDefaults.entry,
                                gpuVaspace->pteDefaults.entrySize,
                                NV_FALSE);
                pbAttachGpuMask |= NVBIT64(gpu->gpuId);
            }
        }
        else
        {
            // Cache gpuPte and Pde for copying over memBacking GPU
            UvmGpuVaspace *gpuVaspace = &pVaspace->gpuVaspaces[gpu->gpuId];
            if (gpuVaspace && gpuVaspace->policy.bForceP2PMigration)
            {
                if (!cachedPde && gpuPte->aperture == UvmMemType_vidmem)
                    cachedPde = pde;
            }
            else
            {
                // TO-DO - Peer mappings can also be used but in production it's
                // better to use local GPU CE as that does physical copies and saves
                // MMU translations.
                if (!cachedPde && pde->memBackingPresent)
                    cachedPde = pde;
            }

            if (!gpuPte->valid)   // gpu is not mapped.
            {
                // Nothing to do if cpu was already mapped
                if (bCpuPfnMapped)
                    continue;

                if (gpuPte->pfnvalid)
                {
                    // copy_gpu_to_cpu  .. done below
                }
            }
            else   // gpu is mapped.
            {
                UVM_PANIC_ON(!gpuPte->pfnvalid);
                gpuPte = _gpu_pte_adjust(vaspace, desc2M, pde, pbOps,
                                    faultingAddress, UVM_CPU_MIN_PAGE_SIZE,
                                    bWriteAccess, &pgSize);
                pbAttachGpuMask |= NVBIT64(gpu->gpuId);
                bCopyNeeded = NV_TRUE;

                if (!gpuPte->readonly || bWriteAccess)
                {
                    UvmGpuPteUpdateParams gpteParams = {0};
                    gpteParams.updateSwFlags |=
                                          UVM_GPU_SW_PTE_UPDATE_FLAGS_PTE_VALID;
                    gpteParams.valid = NV_FALSE;
                    gpteParams.updateSwFlags |=
                                          UVM_GPU_SW_PTE_UPDATE_FLAGS_VOLATILE;
                    gpteParams.vol = NV_FALSE;
                    // Break zfod in any case.
                    gpteParams.updateSwFlags |=
                                        UVM_GPU_SW_PTE_UPDATE_FLAGS_ZFOD_MAPPED;
                    gpteParams.zfodMapped = NV_FALSE;
                    // Actually update the PTE
                    gpteParams.updateHwFlags |=
                                          UVM_GPU_HW_PTE_UPDATE_FLAGS_WRITE_PTE;

                    // disable READ/ATOMIC mappings if any
                    gpteParams.updateSwFlags |=
                                          UVM_GPU_SW_PTE_UPDATE_FLAGS_READ_ONLY;
                    gpteParams.readonly = NV_FALSE;
                    gpteParams.updateSwFlags |=
                                          UVM_GPU_SW_PTE_UPDATE_FLAGS_ATOMIC;
                    gpteParams.atomic = NV_FALSE;

                    // update the pfn in all the GPU ptes
                    uvmfull_update_gpupde_ptes(desc2M, gpu->gpuId,
                                    pde, faultingAddress, 1, pgSize,
                                    &gpteParams,
                                    NULL, pbOps,
                                    (NvU8*)&gpuVaspace->pteDefaults.entry,
                                    gpuVaspace->pteDefaults.entrySize,
                                    NV_FALSE);
                }
                // copy_gpu_to_cpu .. done below
            }
        }
    }


    status = uvm_pbops_submit_and_release(pbOps, pbAttachGpuMask,
                                          &desc2M->tracker);
    if (status != NV_OK)
    {
        return status;
    }

    // Again grab pb for next set of operations
    pbOps = uvm_pbops_start(vaspace);
    if (!pbOps)
        return NV_ERR_NO_MEMORY;
    uvm_pbops_reserve_acquire_space(pbOps, &desc2M->tracker);

    // Reset the submit GPU mask
    pbAttachGpuMask = 0;

     if (cachedPde && bCopyNeeded)
     {
        NvU64 pagePfn = cachedPde->gpuFirstPfn +
                        (GET_PDE_PTE_OFFSET(UVM_ALIGN_ADDR(faultingAddress, UVM_CPU_MIN_PAGE_SIZE),
                         PFN_SHIFT));
        src = PFN_TO_PA(pagePfn);
        dst = PFN_TO_PA(cpuPfn);

        // PushCopies now considering all unmap operations has completed
        migrationOccurs = NV_TRUE;
        gpuId = cachedPde->hostGpuId;
        pushbuffer = uvm_pbops_gpu_pb(pbOps, gpuId);
        pbAttachGpuMask |= NVBIT64(gpuId);
        uvm_push_gpu_to_cpu_copy(pushbuffer, src, dst,
                                 UVM_CPU_MIN_PAGE_SIZE, copyFlags);
     }

    status = uvm_pbops_submit_and_release(pbOps, pbAttachGpuMask,
                                          &desc2M->tracker);
    if (status != NV_OK)
        return status;

    status = uvm_wait_for_tracker(&desc2M->tracker);

    // Update the events and the counters
    if (NV_OK == status && migrationOccurs)
    {
        if (recordMigrationEvent)
        {
            endTime = NV_GETTIME();

            status = uvm_record_migration_event(
                  vaspace->pEventContainer,
                  UvmEventMigrationDirectionGpuToCpu,
                  gpuId,
                  0,
                  faultingAddress & PAGE_MASK_4K,
                  PAGE_SIZE_4K,
                  beginTime,
                  endTime,
                  0);
        }
        uvm_increment_process_counters(gpuId,
                                       vaspace->pCounterContainer,
                                       UvmCounterNameBytesXferDtH,
                                       PAGE_SIZE_4K);
    }
    return status;
}

/* TODO: Look at page hints to find out recommendations.
 */
int uvmfull_cpu_fault_handler(struct vm_area_struct *vma, NvU64 faultingAddress,
                              struct page **ppage, unsigned vmfFlags)
{
    UvmGlobalState *uvm;
    UvmVaspace *vaspace;
    Uvm2MbVaDesc *desc2M;
    NvU64 faultingAddress2M, cpuPfn, pte4k;
    struct page* pg;
    UvmVatContext query;
    int ret = VM_FAULT_SIGBUS;
    NvU64 curr_time;
    NvBool retrying;

    uvm = uvmfull_get_global();
    vaspace = (UvmVaspace *) (vma->vm_file->private_data);
    faultingAddress2M = faultingAddress & GPU_PDE_MASK;
    retrying = NV_FALSE;

retry:
    // curr_time needs to be updated in every retry. Otherwise we will be stuck in an infinite loop
    curr_time = NV_GETTIME();
    down_read(&uvm->lock);
    down_read(&vaspace->lock);

    query.accessType = UvmVatAccess_GenericQuery;
    query.address = faultingAddress2M;
    query.vaspace = vaspace;
    // query returns a locked descriptor if successful
    desc2M = uvmfull_vat_acquire_leaf(&vaspace->vat, &query);
    if (!desc2M)
        goto cleanup_desc;

    pte4k = (faultingAddress - faultingAddress2M) >> PAGE_SHIFT_4K;

    // Early exit if this was a write access on a read only memory
    if (!desc2M->hints[pte4k].va_prot_write && (vmfFlags & NV_FAULT_FLAG_WRITE))
        goto cleanup_desc;

    // Check if cpu pfn is allocated
    if (!desc2M->hints[pte4k].cpuPfnValid)
    {
        pg = alloc_pages(NV_UVM_GFP_FLAGS | GFP_HIGHUSER | __GFP_ZERO, 0);
        if (!pg)
        {
            ret = VM_FAULT_OOM;
            goto cleanup_desc;
        }

        UVM_PANIC_ON(desc2M->hints[pte4k].cpuPfnMapped);
        cpuPfn = page_to_pfn(pg);
        desc2M->hints[pte4k].cpuPfn = cpuPfn;
        desc2M->hints[pte4k].cpuPfnValid = 1;
        desc2M->hints[pte4k].cpuPfnDirty = 0;
        desc2M->hints[pte4k].cpuPfnMapped = 0;
        ret = VM_FAULT_MAJOR;
    }
    else  // page exists
    {
        cpuPfn = desc2M->hints[pte4k].cpuPfn;
        pg = pfn_to_page(cpuPfn);
        UVM_PANIC_ON(!pg);

        // We used to return VM_FAULT_MINOR here. However, that was deprecated
        // in the kernel, and the new guideline is to return 0 in case of a
        // minor fault. The VM_FAULT_MINOR symbol itself was removed in
        // March, 2016 with commit 0e8fb9312fbaf1a687dd731b04d8ab3121c4ff5a.
        ret = 0;
    }

    // Update the gpu state for the given cpu fault only if we have some GPU
    // assocaited with this state
    if (!list_empty(&desc2M->gpuPDEs))
    {
        UvmFaultType faultType;
        UvmAccessType accessType;

        UvmPerfPageCounters *counters;

        faultType = uvmFault_invalidPte;

        if (vmfFlags & NV_FAULT_FLAG_WRITE)
            accessType = UvmAccessType_write;
        else
            accessType = UvmAccessType_read;

        counters = &desc2M->perf[pte4k];

        if (!retrying)
            uvmfull_notify_fault(desc2M, pte4k, pte4k, 1, 0, faultType, accessType, curr_time);
        else
            uvmfull_notify_retry(desc2M, pte4k, pte4k, 1, 0, faultType, accessType, curr_time);

        if (counters->thrashing && (g_uvm_nothrashing == UvmPerfThrashingPreventionThrottleSched ||
                                    g_uvm_nothrashing == UvmPerfThrashingPreventionThrottleSleep))
        {
            ++counters->throttle.naps;

            uvmfull_vat_release_leaf(&vaspace->vat, &query);
            up_read(&vaspace->lock);
            up_read(&uvm->lock);

            if (g_uvm_nothrashing == UvmPerfThrashingPreventionThrottleSched)
                yield();
            if (g_uvm_nothrashing == UvmPerfThrashingPreventionThrottleSleep)
                msleep(g_uvm_thrashing_nap / 1000);

            retrying = NV_TRUE;
            goto retry;
        }

        if (NV_OK != _cpu_fault_update_gpu(vaspace, desc2M, faultingAddress,
                                           cpuPfn, NV_TRUE,
                                           desc2M->hints[pte4k].cpuPfnMapped == 1))
        {
            ret = VM_FAULT_SIGBUS;
            goto cleanup_desc;
        }

        // Mark page as pinned to cpu memory. On the next GPU page fault, GPU pte entries will be remapped to CPU
        // memory
        if (counters->thrashing && g_uvm_nothrashing == UvmPerfThrashingPreventionPinSysmem)
            desc2M->hints[pte4k].pinToCpu = 1;
    }

    if (desc2M->hints[pte4k].cpuPfnMapped == 0)
    {
        // install pte mappings directly in fault handler
        if (vm_insert_page(vma, faultingAddress & PAGE_MASK_4K, pg))
        {
            ret = VM_FAULT_SIGBUS;
            goto cleanup_desc;
        }
        // Tag as mapped
        desc2M->hints[pte4k].cpuPfnMapped = 1;
        // We dont control whether the cpu mapping is read-only vs read write.
        // So the best we can do here is to always mark it dirty.
        desc2M->hints[pte4k].cpuPfnDirty = 1;
    }

    // TODO BUG 1681767 figure out the way to properly support Linux 2.6.18.
    // Old linux will check if pte has been mapped after uvm full fault handler callback.
    // If not, Linux is going to install pte mapping outside of vaspace->lock.
    // This introduces a race if gpu fault handler takes vaspace lock and unmaps
    // pte before Linux kernel checks if fault has been already serviced.
#if defined(VM_FAULT_NOPAGE)
    *ppage = NULL;
    ret = VM_FAULT_NOPAGE;
#else
    get_page(pg);
    *ppage = pg;
#endif

cleanup_desc:
    if (desc2M)
        uvmfull_vat_release_leaf(&vaspace->vat, &query);
    up_read(&vaspace->lock);
    up_read(&uvm->lock);

    return ret;
}

NV_STATUS uvmfull_flush_faultbuffer_for_channel(UvmClientChannel *channel)
{
    UvmGpu *gpu;
    NV_STATUS status = NV_OK;
    UvmGpuVaspace *gpuVaspace = &channel->vaspace->gpuVaspaces[channel->gpuId];
    gpu = uvmfull_gpuid_to_gpu(channel->gpuId);

    if (!uvmfull_is_gpu_fault_capable(channel->gpuId))
        return NV_OK;

    status = _checkpoint_gpu_faultbuffer(gpu, &gpu->faultBufferInfo,
                                  gpuVaspace->gpuPDB, gpuVaspace->pdbAperture);
    return status;
}
