/*******************************************************************************
    Copyright (c) 2014-2015 NVIDIA Corporation

    Permission is hereby granted, free of charge, to any person obtaining a copy
    of this software and associated documentation files (the "Software"), to
    deal in the Software without restriction, including without limitation the
    rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
    sell copies of the Software, and to permit persons to whom the Software is
    furnished to do so, subject to the following conditions:

        The above copyright notice and this permission notice shall be
        included in all copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
    IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
    FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
    THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
    LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
    FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
    DEALINGS IN THE SOFTWARE.

*******************************************************************************/

#include "uvm_full_identity_map.h"
#include "uvm_linux.h"
#include "nv_uvm_interface.h"
#include "uvm_full_pagetbl_mgmt.h"
#include "uvm_full_fault_handler.h"

#define PAGE_DIR_LEVEL3 0x3
#define PAGE_DIR_LEVEL2 0x2
#define PAGE_DIR_LEVEL1 0x1
#define PAGE_DIR_LEVEL0 0x0

#define MAX_GPU_COUNT 0x8

#define MAX_PD0_ENTRIES 0x100 // 256
#define MAX_PD1_ENTRIES 0x200 // 512
#define MAX_PD2_ENTRIES 0x200 // 512

#define MAX_PDE0_COVERAGE 0x200000ULL      // 2MB
#define MAX_PDE1_COVERAGE 0x20000000ULL    // 512MB
#define MAX_PDE2_COVERAGE 0x4000000000ULL  // 256GB

typedef struct
{
    struct list_head list;
    NvU64            addr;
}mapList_node;

static struct kmem_cache * g_uvmMapListNodeCache __read_mostly = NULL;

static void* _alloc_sysmem_page(void)
{
    return kmalloc(PAGE_SIZE, NV_UVM_GFP_FLAGS);
}

static NvBool _identity_map_list_add(struct list_head *head, NvU64 addr)
{
    mapList_node *node = kmem_cache_alloc(g_uvmMapListNodeCache, NV_UVM_GFP_FLAGS);
    if (node)
    {
        node->addr = addr;
        list_add(&node->list, head);
        return NV_TRUE;
    }
    return NV_FALSE;
}

static NvBool _encode_pde(GMMU_FMT *pGmmuFmt, NvU64 parentPdBase,
    NvU32 parentPdIndex, NvU64 gpuPdAddr,
    NvU32 level, NvU32 levelLoBit)
{
    NvU8 *buffer = (NvU8 *)parentPdBase + parentPdIndex * sizeof(NvU64);
    NvU32 methods;

    methods = uvmfull_encode_gpu_pde(buffer,
        buffer + sizeof(NvU64),
        pGmmuFmt, level, levelLoBit,
        gpuPdAddr, UvmMemType_vidmem, 0, 0);

    if (!methods)
        return NV_FALSE;

    return NV_TRUE;
}

static NvBool _encode_pte(GMMU_FMT *pGmmuFmt,
    NvU64 parentPdBase, NvU64 addr, NvU32 peerId)
{
    NvU32 methods, i;
    UvmGpuPte pte = { 0 };
    NvU8 *buffer;

    pte.valid = 1;
    pte.aperture = UvmMemType_peer;
    pte.peerid = peerId;

    for (i = 0; i < MAX_PD0_ENTRIES; i++)
    {
        // Encode the PTE attributes
        pte.pfn = PA_TO_PFN(addr);
        pte.pfnvalid = 1;

        // Adjust region offset
        buffer = (NvU8 *)(parentPdBase + i * 2 * sizeof(NvU64));

        // User lower part for PTE encoding
        methods = uvmfull_encode_gpu_pte(
            buffer,
            buffer + sizeof(NvU64),
            pGmmuFmt,
            &pte,
            buffer + sizeof(NvU64),
            sizeof(NvU64));

        if (!methods)
            return NV_FALSE;

        addr += MAX_PDE0_COVERAGE;
    }

    return NV_TRUE;
}

static void _destroy_mappings(UvmGpu *pGpu)
{
    mapList_node *next, *pos;

    // Cleanup gpu pagelist
    list_for_each_entry_safe(pos, next, &pGpu->gpuPageList, list)
    {
        uvmfull_free_phys_vid_mem(pGpu, PA_TO_PFN(pos->addr));
        list_del(&pos->list);
        kmem_cache_free(g_uvmMapListNodeCache, pos);
    }

    // Cleanup cpu pagelist
    list_for_each_entry_safe(pos, next, &pGpu->cpuPageList, list)
    {
        kfree((void*)pos->addr);
        list_del(&pos->list);
        kmem_cache_free(g_uvmMapListNodeCache, pos);
    }
}

static NV_STATUS _setup_mappings(UvmGpu *pGpu, UvmPushbuffer *pBuffer,
    NvU64 vaStart, NvU64 vaEnd)
{
    NvU64 pfn, gpuPD3, gpuPD2, gpuPD1, gpuPD0;
    NvU64 cpuPD3, cpuPD2, cpuPD1, cpuPD0;
    NV_STATUS status;
    NvU64 i, j, k;
    NvU64 index2Start, index2End;
    NvU64 index1Start, index1End;
    NvU64 vaTemp = vaStart;
    const MMU_FMT_LEVEL *pLevelFmt;
    GMMU_FMT *pGmmuFmt;

    if (!pGpu || !pGpu->channelManager->channelPool.pGmmuFmt)
        return NV_ERR_INVALID_ARGUMENT;

    pGmmuFmt = pGpu->channelManager->channelPool.pGmmuFmt;

    // ******* setup root page directory *******
    status = uvmfull_allocate_phys_vid_mem(pGpu, PAGE_SIZE_4K, &pfn);
    if (status != NV_OK)
        return NV_ERR_NO_MEMORY;
    gpuPD3 = PFN_TO_PA(pfn);
    if (!_identity_map_list_add(&pGpu->gpuPageList, gpuPD3))
        return NV_ERR_NO_MEMORY;
    cpuPD3 = (NvU64)_alloc_sysmem_page();
    if (!cpuPD3)
        return NV_ERR_NO_MEMORY;
    if (!_identity_map_list_add(&pGpu->cpuPageList, cpuPD3))
        return NV_ERR_NO_MEMORY;
    memset((void*)cpuPD3, 0, PAGE_SIZE);

    // Call NV0080_CTRL_CMD_DMA_SET_PAGE_DIRECTORY which
    // will copy the RM PDE[0] to this new location.
    status = nvUvmInterfaceSetPageDirectory(
        pGpu->channelManager->channelPool.hVaSpace, gpuPD3,
        mmuFmtLevelEntryCount(pGmmuFmt->pRoot), NV_TRUE);
    if (status != NV_OK)
        return status;

    // ******* Encoding Level 3 *******
    for (i = mmuFmtVirtAddrToEntryIndex(pGmmuFmt->pRoot, vaStart); 
        i <= mmuFmtVirtAddrToEntryIndex(pGmmuFmt->pRoot, vaEnd); i++)
    {
        status = uvmfull_allocate_phys_vid_mem(pGpu, PAGE_SIZE_4K, &pfn);
        if (status != NV_OK)
            return NV_ERR_NO_MEMORY;
        gpuPD2 = PFN_TO_PA(pfn);
        if (!_identity_map_list_add(&pGpu->gpuPageList, gpuPD2))
            return NV_ERR_NO_MEMORY;
        cpuPD2 = (NvU64)_alloc_sysmem_page();
        if (!cpuPD2)
            return NV_ERR_NO_MEMORY;
        if (!_identity_map_list_add(&pGpu->cpuPageList, cpuPD2))
            return NV_ERR_NO_MEMORY;
        memset((void*)cpuPD2, 0, PAGE_SIZE);
        if (!_encode_pde(pGmmuFmt, cpuPD3, i, gpuPD2,
            PAGE_DIR_LEVEL3, UVM_VA_SHIFT_TRIE_LVL_3_LO))
                return NV_ERR_IN_USE;

        // ******* Encoding Level 2 *******
        pLevelFmt = mmuFmtFindLevelWithPageShift(pGmmuFmt->pRoot,
            UVM_VA_SHIFT_TRIE_LVL_2_LO);
        if (!pLevelFmt)
            return NV_ERR_INVALID_ARGUMENT;

        index2Start = mmuFmtVirtAddrToEntryIndex(pLevelFmt, vaTemp);
        index2End = mmuFmtVirtAddrToEntryIndex(pLevelFmt,
            min((NvU64)((vaTemp + MAX_PDE2_COVERAGE * MAX_PD2_ENTRIES) - 1), vaEnd));

        for (j = index2Start; j <= index2End; j++)
        {
            status = uvmfull_allocate_phys_vid_mem(pGpu, PAGE_SIZE_4K, &pfn);
            if (status != NV_OK)
                return NV_ERR_NO_MEMORY;
            gpuPD1 = PFN_TO_PA(pfn);
            if (!_identity_map_list_add(&pGpu->gpuPageList, gpuPD1))
                return NV_ERR_NO_MEMORY;
            cpuPD1 = (NvU64)_alloc_sysmem_page();
            if (!cpuPD1)
                return NV_ERR_NO_MEMORY;
            if (!_identity_map_list_add(&pGpu->cpuPageList, cpuPD1))
                return NV_ERR_NO_MEMORY;
            memset((void*)cpuPD1, 0, PAGE_SIZE);
            if (!_encode_pde(pGmmuFmt, cpuPD2, j, gpuPD1,
                PAGE_DIR_LEVEL2, UVM_VA_SHIFT_TRIE_LVL_2_LO))
                return NV_ERR_IN_USE;

            // ******* Encoding Level 1 *******
            pLevelFmt = mmuFmtFindLevelWithPageShift(pGmmuFmt->pRoot,
                UVM_VA_SHIFT_TRIE_LVL_1_LO);
            if (!pLevelFmt)
                return NV_ERR_INVALID_ARGUMENT;

            index1Start = mmuFmtVirtAddrToEntryIndex(pLevelFmt, vaTemp);
            index1End = mmuFmtVirtAddrToEntryIndex(pLevelFmt,
                min((NvU64)((vaTemp + MAX_PDE1_COVERAGE * MAX_PD1_ENTRIES) - 1), vaEnd));

            for (k = index1Start; k <= index1End; k++)
            {
                status = uvmfull_allocate_phys_vid_mem(pGpu, PAGE_SIZE_4K, &pfn);
                if (status != NV_OK)
                    return NV_ERR_NO_MEMORY;
                gpuPD0 = PFN_TO_PA(pfn);
                if (!_identity_map_list_add(&pGpu->gpuPageList, gpuPD0))
                    return NV_ERR_NO_MEMORY;
                cpuPD0 = (NvU64)_alloc_sysmem_page();
                if (!cpuPD0)
                    return NV_ERR_NO_MEMORY;
                if (!_identity_map_list_add(&pGpu->cpuPageList, cpuPD0))
                    return NV_ERR_NO_MEMORY;
                memset((void*)cpuPD0, 0, PAGE_SIZE);
                if (!_encode_pde(pGmmuFmt, cpuPD1, k, gpuPD0,
                    PAGE_DIR_LEVEL1, UVM_VA_SHIFT_TRIE_LVL_1_LO))
                    return NV_ERR_IN_USE;

                // ******* Encoding Level 0 as a PageTable *******
                if (!_encode_pte(pGmmuFmt, cpuPD0,
                    vaTemp % PASCAL_MAX_FB,
                    (vaTemp - IDENTITY_MAPPING_VA_BASE) / PASCAL_MAX_FB))
                    return NV_ERR_IN_USE;

                // Copy page table
                uvm_push_cpu_to_gpu_copy(pBuffer, __pa(cpuPD0), gpuPD0, PAGE_SIZE,
                    NV_UVM_COPY_SRC_TYPE_PHYSICAL | NV_UVM_COPY_DST_TYPE_PHYSICAL);

                vaTemp += MAX_PDE1_COVERAGE;
            }
            // copy PD1
            uvm_push_cpu_to_gpu_copy(pBuffer, __pa(cpuPD1), gpuPD1, PAGE_SIZE,
                NV_UVM_COPY_SRC_TYPE_PHYSICAL | NV_UVM_COPY_DST_TYPE_PHYSICAL);
        }
        // copy PD2
        uvm_push_cpu_to_gpu_copy(pBuffer, __pa(cpuPD2), gpuPD2, PAGE_SIZE,
            NV_UVM_COPY_SRC_TYPE_PHYSICAL | NV_UVM_COPY_DST_TYPE_PHYSICAL);
    }
    // copy PD3
    uvm_push_cpu_to_gpu_copy(pBuffer, __pa(cpuPD3 + sizeof(NvU64)),
        gpuPD3 + sizeof(NvU64), (PAGE_SIZE - sizeof(NvU64)),
        NV_UVM_COPY_SRC_TYPE_PHYSICAL | NV_UVM_COPY_DST_TYPE_PHYSICAL);

    // Push TLB invalidate
    {
        UvmTlbInvalidateVaParams invalidateParams = { 0 };
        memset(&invalidateParams, 0x0, sizeof(UvmTlbInvalidateVaParams));
        invalidateParams.invalidatelevel = UvmTlbInvalidateLevel_all;

        uvm_push_invalidate(pGpu->gpuId,
            pBuffer,
            invalidateParams,
            UvmMemType_vidmem,
            PA_TO_PFN(gpuPD3),
            UvmTlbInvalidateMemBar_local);
    }

    return NV_OK;
}

NV_STATUS uvmfull_create_identity_map(UvmGpu *pGpu)
{
    NV_STATUS status;
    UvmPbOps *pbOps = NULL;
    UvmPushbuffer *pBuffer = NULL;
    UvmTracker tracker;

    if (!pGpu || !pGpu->channelManager)
        return NV_ERR_INVALID_ARGUMENT;

    if (pGpu->bZeroFbGpu)
        return NV_OK;

    // get a pushbuffer
    pbOps = uvm_pbops_start_on_gpu(pGpu->gpuId);
    if (!pbOps)
    {
        status = NV_ERR_GENERIC;
        goto cleanup;
    }
    pBuffer = uvm_pbops_gpu_pb(pbOps, pGpu->gpuId);
    if (!pBuffer)
    {
        status = NV_ERR_NO_MEMORY;
        goto cleanup;
    }

    uvm_init_tracker(&tracker);

    // FIXME: Currently we setup mappings for all peers
    // in one go. Dynamic mapping needs a SW trie to keep
    // track of PDBs which will be implemented later.
    status = _setup_mappings(pGpu, pBuffer, IDENTITY_MAPPING_VA_BASE,
        (IDENTITY_MAPPING_VA_BASE + (MAX_GPU_COUNT * PASCAL_MAX_FB) - 1));
    if (status != NV_OK)
        goto cleanup;

    // Submit a pushbuffer
    uvm_pbops_submit(pbOps, pGpu->gpuId, pBuffer, &tracker);
    uvm_move_tracker(&tracker, pbOps->t);
    status = uvm_wait_for_tracker(&tracker);
    if (status != NV_OK)
        goto cleanup;
    uvm_pbops_end(pbOps);

    return status;

cleanup:
    if (pbOps)
        uvm_pbops_end(pbOps);
    _destroy_mappings(pGpu);
    return status;
}

NV_STATUS uvmfull_destroy_identity_map(UvmGpu *pGpu)
{
    if (!pGpu)
        return NV_ERR_INVALID_ARGUMENT;

    _destroy_mappings(pGpu);

    return NV_OK;
}

// Setup cache
NV_STATUS uvmfull_identity_map_cache_init(void)
{
    g_uvmMapListNodeCache = NV_KMEM_CACHE_CREATE("uvm_mapListNode", mapList_node);
    if (!g_uvmMapListNodeCache)
    {
        return NV_ERR_NO_MEMORY;
    }
    return NV_OK;
}

// Destroy cache
void uvmfull_identity_map_cache_destroy(void)
{
    if (g_uvmMapListNodeCache)
        kmem_cache_destroy_safe(&g_uvmMapListNodeCache);
}
