/*******************************************************************************
    Copyright (c) 2015-2015 NVIDIA Corporation

    Permission is hereby granted, free of charge, to any person obtaining a copy
    of this software and associated documentation files (the "Software"), to
    deal in the Software without restriction, including without limitation the
    rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
    sell copies of the Software, and to permit persons to whom the Software is
    furnished to do so, subject to the following conditions:

        The above copyright notice and this permission notice shall be
        included in all copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
    IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
    FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
    THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
    LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
    FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
    DEALINGS IN THE SOFTWARE.

*******************************************************************************/

#include "uvm_common.h"
#include "uvm_utils.h"
#include "nv_uvm_interface.h"
#include "uvm_full_device_mgmt.h"
#include "uvm_full_pa_mgmt.h"
#include "uvm_full_pagetbl_mgmt.h"

/**
 * ------------------------------------------------------------
 * PA Management structure initialisation API's
 * ------------------------------------------------------------
 */
static struct kmem_cache * g_uvmPaManagerCache __read_mostly = NULL;
static struct kmem_cache * g_UvmPaDescCache __read_mostly = NULL;
static struct kmem_cache * g_uvmPaToVaCache __read_mostly = NULL;

/**
 * -----------------------------------------------------------------------------
 * Static helper functions for PA Management.
 * -----------------------------------------------------------------------------
 */

/**
 * _pa_pool_init
 *
 * This function initializes the UvmPaPool structure.
 *
 * @param[in]   pPaPool     UvmPaPool object to be initialised.
 *
 * @returns void.
 */
static void _pa_pool_init(UvmPaPool *pPaPool);

/**
 * _pa_mngr_init
 *
 * This function initializes the UvmPaManager structure for a given GPU
 * device. Caller must have already allocated the object.
 *
 * @param[in]   paManager       Per device based PA Manager
 * @param[in]   bitsForPA       Represents size of PA in bits.
 * @param[in]   bitsForLeafPage Represents size of leaf chunk managed by the descriptor.
 * @param[in]   bIsGpuPaMngr    Identifies if this is GPU or CPU PA Manager.
 * @param[in]   gpuId           If bIsGpuPaMngr = true, this fields stored GPU id.
 *
 * @returns void.
 */
static void _pa_mngr_init(UvmPaManager *paManager,
                          NvU64   bitsForPA,
                          NvU64   bitsForLeafPage,
                          NvBool  bIsGpuPaMngr,
                          UvmGpuId gpuId);

/**
 * _pa_desc_init
 *
 * This function initializes the UvmPaDesc structure. Caller must have
 * already allocated the object.
 *
 * @param[in]   pPaDesc         Descriptor that keeps track of underlying PA.
 * @param[in]   pa              Represents start PA tracked by this desc.
 *
 * @return void.
 */
static void _pa_desc_init(UvmPaDesc *pPaDesc,
                          NvU64 pa);

/**
 * _pa_to_va_init
 *
 * This function initializes the UvmPaToVa structure. Caller must have
 * already allocated the object.
 *
 * @param[in]   pPaToVa         Mapping from PA to VA
 * @param[in]   va              Represents start VA corresponding to given PA.
 * @param[in]   pVaspace        Represents the VASpace corresponding to above VA.
 * @param[in]   range           Range of address from VA.
 *
 * @return void.
 */
static void _pa_to_va_init(UvmPaToVa *pPaToVa,
                           NvU64 va,
                           struct UvmVaspace *pVaspace,
                           NvU64 range);

/**
 * _gpuid_to_pma
 *
 * Retrieves the PMA object for GPU ID provided.
 *
 * @param[in]   gpuId           GPU id for which PMA object is to be retrieved.
 *
 * @return pointer to PMA object corresponding to the GPU ID provided.
 *          On 0 FB, it returns NULL.
 */
//!
static void * _gpuid_to_pma(UvmGpuId gpuId);

/**
 * ------------------------------------------------------------
 * UvmPaDesc struct manipulation APIs
 * ------------------------------------------------------------
 */

/**
 * _insert_pa_desc
 *
 * Given a root of the UvmPaManager and the PA, this function inserts the given
 * UvmPaDesc node into the radix trie with PFN as the key. Caller must perform
 * all the list manipulation in the UvmPaManager.
 *
 * @param[in]   pPaMngr         Per device based PA Manager
 * @param[in]   pPaDesc         Descriptor that keeps track of underlying PA.
 *
 * @return NV_STATUS NV_OK if succeeded; or else other specific errors.
 */
static NV_STATUS _insert_pa_desc(UvmPaManager *pPaMngr,
                                 UvmPaDesc *pPaDesc);

/**
 * _alloc_pa_desc
 *
 * Allocates an instance of UvmPaDesc struct using slab allocator.
 *
 * @return pointer to allocated UvmPaDesc if succeeded; or else returns NULL.
 */
static UvmPaDesc* _alloc_pa_desc(NvU64 pa);

/**
 * _acquire_pa_desc
 *
 * Given a PA manager and the PA of interest, this function acquires the desc
 * node having corresponding PA tracked by it. The descriptor returned has write
 * lock on it and it is callers responsibility to release the write lock.
 *
 * @param[in]   pPaMngr          Per device based PA Manager
 * @param[in]   pa               Represents start PA for which the descriptor is reqd.
 *
 * @return pointer to UvmPaDesc corresponding to PA passed.
 *         Else it returns NULL if descriptor allocation fails.
 */
static UvmPaDesc * _acquire_pa_desc(UvmPaManager *pPaMngr,
                                    NvU64 pa);

/**
 * _lock_pa_desc
 *
 * Given a PA descriptor node of interest, this function write locks it and
 * returns the node. It is callers responsibility to release the write lock.
 *
 * @param[in]   pPaDesc         Descriptor that needs to be write locked.
 *
 * @return pointer to UvmPaDesc with write lock on it.
 */
static void _lock_pa_desc(UvmPaDesc *pPaDesc);

/**
 * _release_pa_desc
 *
 * Given a PA descriptor node this function releases the write_lock on it.
 *
 * @param[in]   pPaDesc         Descriptor that keeps track of underlying PA.
 *
 * @return pointer to UvmPaDesc corresponding to PA passed.
 *         Else it returns NULL if descriptor allocation fails.
 */
static void _release_pa_desc(UvmPaDesc *pPaDesc);

/**
 * _alloc_pa_to_va
 *
 * Allocates an instance of UvmPaToVa struct using slab allocator.
 *
 * @return pointer to allocated UvmPaToVa if succeeded; or else returns NULL.
 */
static UvmPaToVa* _alloc_pa_to_va(void);

/**
 * _allocate_vid_mem_default_chunk
 *
 * Given a root of the UvmPaManager and the chunk count, this function asks PMA
 * / RM module to allocate those many memory blocks, each of default chunk size.
 * The function fills the array with PA of those allocations.
 *
 * @param[in]   pPaMngr         Per device based PA Manager.
 * @param[in]   chunkCount      Number of chunks of size default size to be allocated
 * @param[in]   allocFlags      Parameters to configure Phy mem allocation.
 * @param[out]  pPages          Fills the array with PA of allocated memory.
 * @param[out]  pOutCount       Count of number of chunks successfully allocated.
 *
 * @return NV_STATUS NV_OK if succeeded; or else other specific errors.
 */
static NV_STATUS _allocate_vid_mem_default_chunk(UvmPaManager *pPaMngr,
                                                NvU64 chunkCount,
                                                UvmPaAllocationOptions allocFlags,
                                                NvU64 *pPages,
                                                NvU64 *pOutCount);
/**
 * _allocate_default_chunk
 *
 * This function pointer invokes PMA if PMA is enabled in RM for current platform,
 * else it calls RM to allocate memory blocks, each of default chunk size.
 *
 * @param[in]   pPaMngr         Per device based PA Manager.
 * @param[in]   chunkCount      Number of chunks of size default size to be allocated
 * @param[in]   allocFlags      Parameters to configure Phy mem allocation.
 * @param[out]  pPages          Fills the array with PA of allocated memory.
 * @param[out]  pOutCount       Count of number of chunks successfully allocated.
 *
 * @return NV_STATUS NV_OK if succeeded; or else other specific errors.
 */
static NV_STATUS (*_allocate_default_chunk)(UvmPaManager *pPaMngr,
                                             NvU64 chunkCount,
                                             UvmPaAllocationOptions allocFlags,
                                             NvU64 *pPages,
                                             NvU64 *pOutCount);

/**
 * _allocate_default_chunk_pma
 *
 * Given a root of the UvmPaManager and the chunk count, this function asks PMA
 * module to allocate those many memory blocks, each of default chunk size.
 * The function fills the array with PA of those allocations.
 *
 * @param[in]   pPaMngr         Per device based PA Manager.
 * @param[in]   chunkCount      Number of chunks of size default size to be allocated
 * @param[in]   allocFlags      Parameters to configure Phy mem allocation.
 * @param[out]  pPages          Fills the array with PA of allocated memory.
 * @param[out]  pOutCount       Count of number of chunks successfully allocated.
 *
 * @return NV_STATUS NV_OK if succeeded; or else other specific errors.
 */
static NV_STATUS _allocate_default_chunk_pma(UvmPaManager *pPaMngr,
                                             NvU64 chunkCount,
                                             UvmPaAllocationOptions allocFlags,
                                             NvU64 *pPages,
                                             NvU64 *pOutCount);
/**
 * _sub_allocate_vid_mem
 *
 * Given a root of the UvmPaManager and the alloc count with each chunk size ,
 * this function asks allocates the requested memory by sub-allocating from its
 * internally managed memory pool. The function returns a list of base PA of
 * those allocations.
 *
 * @param[in]   pPaMngr         Per device based PA Manager.
 * @param[in]   chunkCount      Number of chunks of size chunkSize to be allocated.
 * @param[in]   chunkSize       Size of each allocation.
 * @param[in]   allocFlags      Parameters to configure Phy mem allocation.
 * @param[out]  pPages          Fills the array with PA of allocated memory.
 * @param[out]  pOutCount       Count of number of chunks successfully allocated.
 *
 * @return NV_STATUS NV_OK if succeeded; or else other specific errors.
 */

static NV_STATUS _sub_allocate_vid_mem(UvmPaManager *pPaMngr,
                                       NvU64 chunkCount,
                                       NvU32 chunkSize,
                                       UvmPaAllocationOptions allocFlags,
                                       NvU64 *pPages,
                                       NvU64 *pOutCount);
/**
 * -----------------------------------------------------------------------------
 * PA Mgmt Slab allocator initialization and destroy.
 * -----------------------------------------------------------------------------
 */

NV_STATUS uvmfull_pa_mgmt_slab_init(void)
{
    NV_STATUS status = NV_OK;

    g_uvmPaManagerCache = NV_KMEM_CACHE_CREATE("uvm_pa_manager", UvmPaManager);
    if (!g_uvmPaManagerCache)
    {
        status = NV_ERR_NO_MEMORY;
        goto cleanup;
    }

    g_UvmPaDescCache = NV_KMEM_CACHE_CREATE("uvm_pa_desc", UvmPaDesc);
    if (!g_UvmPaDescCache)
    {
        status = NV_ERR_NO_MEMORY;
        goto cleanup;
    }

    g_uvmPaToVaCache = NV_KMEM_CACHE_CREATE("uvm_pa_to_va", UvmPaToVa);
    if (!g_uvmPaToVaCache)
    {
        status = NV_ERR_NO_MEMORY;
        goto cleanup;
    }

cleanup:
    if (NV_OK != status)
    {
        uvmfull_pa_mgmt_slab_destroy();
    }
    return status;
}

void uvmfull_pa_mgmt_slab_destroy(void)
{
    if (g_uvmPaManagerCache)
        kmem_cache_destroy_safe(&g_uvmPaManagerCache);

    if (g_UvmPaDescCache)
        kmem_cache_destroy_safe(&g_UvmPaDescCache);

    if (g_uvmPaToVaCache)
        kmem_cache_destroy_safe(&g_uvmPaToVaCache);
}

/**
 * -----------------------------------------------------------------------------
 * PA Management structure initialisation API's
 * -----------------------------------------------------------------------------
 */

void _pa_pool_init(UvmPaPool *pPaPool)
{
    INIT_LIST_HEAD(&pPaPool->freeHead);
    INIT_LIST_HEAD(&pPaPool->usedHead);
    init_rwsem(&pPaPool->lock);
}

void _pa_mngr_init(UvmPaManager *pPaManager,
                   NvU64    bitsForPA,
                   NvU64    bitsForLeafPage,
                   NvBool   bIsGpuPaMngr,
                   UvmGpuId gpuId)
{
    init_rwsem(&pPaManager->trieLookupLock);

    pPaManager->bitsForPA = bitsForPA;
    pPaManager->bitsForLeafPage = bitsForLeafPage;
    pPaManager->bIsGpuPaMngr = bIsGpuPaMngr;

    if (bIsGpuPaMngr)
    {
        pPaManager->gpuId = gpuId;
        pPaManager->pPma = _gpuid_to_pma(pPaManager->gpuId);
    }

    if (!_allocate_default_chunk)
    {
        _allocate_default_chunk = _allocate_default_chunk_pma;
    }

    INIT_RADIX_TREE(&pPaManager->paRoot, NV_UVM_GFP_FLAGS);
    _pa_pool_init(&pPaManager->subAllocMemHead);
    _pa_pool_init(&pPaManager->defaultMemHead);
}

void _pa_desc_init(UvmPaDesc *pPaDesc,
                   NvU64      pa)
{
    init_rwsem(&pPaDesc->descLock);
    pPaDesc->pa = pa;

    INIT_LIST_HEAD(&pPaDesc->node);
    INIT_LIST_HEAD(&pPaDesc->reverseMapList);
}

void _pa_to_va_init(UvmPaToVa          *pPaToVa,
                    NvU64               va,
                    struct UvmVaspace   *pVaspace,
                    NvU64               range)
{
    pPaToVa->va = va;
    pPaToVa->range = range;
    pPaToVa->pVaspace = pVaspace;

    INIT_LIST_HEAD(&pPaToVa->paDescNode);
    INIT_LIST_HEAD(&pPaToVa->vaDescNode);
}

void * _gpuid_to_pma(UvmGpuId gpuId)
{
    NV_STATUS retCode = NV_OK;
    UvmGpu *pGpu;
    void *pPma = NULL;

    pGpu = uvmfull_gpuid_to_gpu(gpuId);

    retCode = nvUvmInterfaceGetPmaObject(&pGpu->gpuUuid, &pPma);
    if (NV_OK != retCode)
    {
        WARN_ON(1 || "Retrieving PMA object failed.");
        return NULL;
    }
    return pPma;
}

/**
 * -----------------------------------------------------------------------------
 * Object allocation using Slab allocator.
 * -----------------------------------------------------------------------------
 */

UvmPaManager* uvmfull_alloc_pa_manager(NvU64    bitsForPA,
                                       NvU64    bitsForLeafPage,
                                       NvBool   bIsGpuPaMngr,
                                       UvmGpuId gpuId)
{
    UvmPaManager *pPaMngr = NULL;

    pPaMngr = kmem_cache_zalloc(g_uvmPaManagerCache, NV_UVM_GFP_FLAGS);

    if (pPaMngr)
    {
        _pa_mngr_init(pPaMngr, bitsForPA, bitsForLeafPage,
                                bIsGpuPaMngr, gpuId);
    }
    return pPaMngr;
}

UvmPaDesc* _alloc_pa_desc(NvU64 pa)
{
    UvmPaDesc *pPaDesc = NULL;

    pPaDesc = kmem_cache_zalloc(g_UvmPaDescCache, NV_UVM_GFP_FLAGS);

    if (pPaDesc)
    {
        _pa_desc_init(pPaDesc, pa);
    }
    return pPaDesc;
}

UvmPaToVa* _alloc_pa_to_va(void)
{
    UvmPaToVa *pPaToVa = NULL;

    pPaToVa = kmem_cache_zalloc(g_uvmPaToVaCache, NV_UVM_GFP_FLAGS);

    if (pPaToVa)
    {
        _pa_to_va_init(pPaToVa, 0, NULL, 0);
    }
    return pPaToVa;
}

/**
 * ------------------------------------------------------------
 * UvmPaDesc struct manipulation APIs
 * ------------------------------------------------------------
 */

NV_STATUS _insert_pa_desc(UvmPaManager *pPaMngr,
                          UvmPaDesc *pPaDesc)
{
    NV_STATUS retCode = NV_OK;
    NvU64 pfn;

    pfn = pPaDesc->pa >> UVM_PA_MANAGER_CHUNK_SIZE_DEFAULT_SHIFT;

    down_write(&pPaMngr->trieLookupLock);
    down_write(&pPaDesc->descLock);

    retCode = radix_tree_insert(&pPaMngr->paRoot, pfn,
                          (void *)pPaDesc);

    if (NV_OK != retCode)
    {
        if (-ENOMEM == retCode)
        {
            WARN_ON(1 || "PA descriptor insertion failed. No memory available!");
            retCode = NV_ERR_NO_MEMORY;
        }
        else if (-EEXIST == retCode)
        {
            WARN_ON(1 || "PA descriptor insertion failed. Descriptor already exists!");
            retCode = NV_ERR_INVALID_REQUEST;
        }
        else
        {
            WARN_ON(1 || "PA descriptor insertion failed. Unknown error!");
            retCode = NV_ERR_GENERIC;
        }
    }

    up_write(&pPaDesc->descLock);
    up_write(&pPaMngr->trieLookupLock);

    return retCode;
}

UvmPaDesc * _acquire_pa_desc(UvmPaManager *pPaMngr,
                             NvU64         pa)
{
    NV_STATUS retCode = NV_OK;
    UvmPaDesc *pPaDesc;

    pPaDesc = uvmfull_get_pa_desc(pPaMngr, pa);

    // If descriptor is not found, then insert new descriptor into radix trie.
    if (!pPaDesc)
    {
        pPaDesc = _alloc_pa_desc(pa);
        if (!pPaDesc)
        {
            return NULL;
        }

        retCode = _insert_pa_desc(pPaMngr, pPaDesc);
        if (NV_OK != retCode)
        {
            return NULL;
        }
    }

    // return descriptor with write_lock set.
    down_write(&pPaDesc->descLock);
    return pPaDesc;
}

void _lock_pa_desc(UvmPaDesc *pPaDesc)
{
    down_write(&pPaDesc->descLock);
}

void _release_pa_desc(UvmPaDesc *pPaDesc)
{
    up_write(&pPaDesc->descLock);
}

UvmPaDesc * uvmfull_get_pa_desc(UvmPaManager *pPaMngr,
                                NvU64         pa)
{
    NvU64 pfn;
    UvmPaDesc *pPaDesc;

    down_read(&pPaMngr->trieLookupLock);

    pfn = pa >> UVM_PA_MANAGER_CHUNK_SIZE_DEFAULT_SHIFT;

    pPaDesc = (UvmPaDesc *)radix_tree_lookup(&pPaMngr->paRoot, pfn);

    up_read(&pPaMngr->trieLookupLock);

    return pPaDesc;
}


NV_STATUS _allocate_default_chunk_pma(UvmPaManager *pPaMngr,
                                      NvU64 chunkCount,
                                      UvmPaAllocationOptions allocFlags,
                                      NvU64 *pPages,
                                      NvU64 *pOutCount)
{
    NV_STATUS retCode = NV_OK;
    void *pPma = pPaMngr->pPma;
    UvmPmaAllocationOptions pmaAllocOptions = {0};

    UVM_PANIC_ON(NULL == pPma);

    if (allocFlags.bPinned)
    {
        pmaAllocOptions.flags |= UVM_PMA_ALLOCATE_PINNED;
    }
    if (allocFlags.bContiguous)
    {
        pmaAllocOptions.flags |= UVM_PMA_ALLOCATE_CONTIGUOUS;
    }

    *pOutCount = 0;
    // Get all requested memory by invoking PMA.
    retCode = nvUvmInterfacePmaAllocPages(pPma,
                                          chunkCount,
                                          UVM_PA_MANAGER_CHUNK_SIZE_DEFAULT,
                                          &pmaAllocOptions,
                                          pPages);
    if (NV_OK != retCode)
    {
        WARN_ON(1 || "Allocating memory using PMA failed. ");
        return retCode;
    }

    *pOutCount = chunkCount;
    return retCode;
}

NV_STATUS _allocate_vid_mem_default_chunk(UvmPaManager *pPaMngr,
                                          NvU64 chunkCount,
                                          UvmPaAllocationOptions allocFlags,
                                          NvU64 *pPages,
                                          NvU64 *pOutCount)
{
    NV_STATUS retCode = NV_OK;
    UvmPaDesc *pDesc = NULL;
    NvU64 chunkNo;
    NvU64 tmpCount = 0;
    NvU64 tmpOutCount = 0;

    down_write(&pPaMngr->defaultMemHead.lock);
    // Use while loop to acquire all memory from free list.
    while (tmpCount < chunkCount &&
           !list_empty(&pPaMngr->defaultMemHead.freeHead))
    {
        // acquire the paDesc - just do down_write
        pDesc = (UvmPaDesc *)list_entry(pPaMngr->defaultMemHead.freeHead.prev,
                                        UvmPaDesc, node);

        _lock_pa_desc(pDesc);

        WARN_ON(NV_TRUE == pDesc->bBackingExists);
        WARN_ON(NV_FALSE == pDesc->bIsSubAllocated);
        WARN_ON(0 == pDesc->usedBitCount);

        // move desc to used list
        // pPaMngr->defaultMemHead.freeHead = pPaMngr->defaultMemHead.freeHead.next;
        list_move(&pDesc->node,
                  &pPaMngr->defaultMemHead.usedHead);

        // Store PA to return.
        pPages[tmpCount] = pDesc->pa;

        // release the paDesc
        _release_pa_desc(pDesc);

        tmpCount++;
    }

    up_write(&pPaMngr->defaultMemHead.lock);

    // If all requested memory is found, then return from here.
    if (tmpCount >= chunkCount)
    {
        *pOutCount = tmpCount;
        return retCode;
    }

    // Request rest of the memory using PMA / RM.
    retCode = _allocate_default_chunk(pPaMngr,
                            chunkCount - tmpCount,
                            allocFlags,
                            &pPages[tmpCount],
                            &tmpOutCount);
    if (NV_OK != retCode)
    {
        WARN_ON(1 || "Default size chunk allocation failed.");
        retCode = NV_ERR_NO_MEMORY;
        goto cleanup_global;
    }

    down_write(&pPaMngr->defaultMemHead.lock);

    // Insert each default sized allocation to PA Radix trie.
    for (chunkNo = tmpCount; chunkNo < tmpCount + tmpOutCount; chunkNo++)
    {
        // Allocate descriptor.
        pDesc = _acquire_pa_desc(pPaMngr, pPages[chunkNo]);
        if (!pDesc)
        {
            WARN_ON(1 || "UvmPaDesc allocation failed.");
            retCode = NV_ERR_NO_MEMORY;
            goto cleanup;
        }

        WARN_ON(NV_FALSE == pDesc->bBackingExists);

        // Update Descriptor.
        pDesc->bBackingExists = NV_TRUE;
        pDesc->bIsSubAllocated = NV_FALSE;
        pDesc->usedBitCount = 0;

        // Insert descriptor to the non-subAllocated used list.
        list_add(&pDesc->node,
                 &pPaMngr->defaultMemHead.usedHead);

        // release the paDesc
        _release_pa_desc(pDesc);
    }

cleanup:
    if (NV_OK == retCode &&
        chunkCount != tmpCount + tmpOutCount)
    {
        retCode = NV_ERR_NO_MEMORY;
    }

    up_write(&pPaMngr->defaultMemHead.lock);
cleanup_global:
    *pOutCount = tmpCount + tmpOutCount;
    return retCode;
}

NV_STATUS uvmfull_allocate_vid_mem(UvmPaManager *pPaMngr,
                                   NvU64 chunkCount,
                                   NvU32 chunkSize,
                                   UvmPaAllocationOptions allocFlags,
                                   NvU64 *pPages,
                                   NvU64 *pOutCount)
{
    NV_STATUS retCode = NV_OK;

    if (UVM_PA_MANAGER_CHUNK_SIZE_DEFAULT < chunkSize)
    {
        UVM_PANIC_MSG("Requested chunk size is not supported!");
    }
    else if (UVM_PA_MANAGER_CHUNK_SIZE_DEFAULT == chunkSize)
    {
        // Ask PMA / RM to alloc memory.
        retCode = _allocate_vid_mem_default_chunk(pPaMngr,
                chunkCount,
                allocFlags,
                pPages,
                pOutCount);
    }
    else
    {
        retCode = _sub_allocate_vid_mem(pPaMngr,
                chunkCount,
                chunkSize,
                allocFlags,
                pPages,
                pOutCount);
    }

    return retCode;
}

NV_STATUS _sub_allocate_vid_mem(UvmPaManager *pPaMngr,
                                       NvU64 chunkCount,
                                       NvU32 chunkSize,
                                       UvmPaAllocationOptions allocFlags,
                                       NvU64 *pPages,
                                       NvU64 *pOutCount)
{
    // TODO: Update sub allocator logic here.
    return NV_OK;
}

NV_STATUS uvmfull_pa_mngr_destroy(UvmPaManager *pPaMngr)
{
    NV_STATUS retCode = NV_OK;
    struct list_head *pCurrListNode = NULL;
    struct list_head *pTempStorage = NULL;
    UvmPaDesc *pDescNode = NULL;
    NvU64 currPage[1];
    struct list_head *paMngrMem[4];
    struct rw_semaphore *pLock[2];
    NvU8 i;

    paMngrMem[0] = &pPaMngr->subAllocMemHead.usedHead;
    paMngrMem[1] = &pPaMngr->subAllocMemHead.freeHead;
    paMngrMem[2] = &pPaMngr->defaultMemHead.usedHead;
    paMngrMem[3] = &pPaMngr->defaultMemHead.freeHead;

    pLock[0] = &pPaMngr->subAllocMemHead.lock;
    pLock[1] = &pPaMngr->defaultMemHead.lock;

    for (i = 0; i < sizeof(paMngrMem)/sizeof(paMngrMem[0]); i++)
    {
        down_write(pLock[i / 2]);

        // Traverse through list and free each chunk of default size.
        list_for_each_safe(pCurrListNode, pTempStorage, paMngrMem[i])
        {
            pDescNode = (UvmPaDesc *)list_entry(pCurrListNode, UvmPaDesc, node);

            // TODO: Delete all the UvmPaToVa child nodes under this Desc.

            // Free memory by calling to PMA.
            currPage[0] = pDescNode->pa;
            nvUvmInterfacePmaFreePages(pPaMngr->pPma,
                                       currPage,
                                       1,
                                       UVM_PA_MANAGER_CHUNK_SIZE_DEFAULT,
                                       0);

            list_del(pCurrListNode);
            kmem_cache_free(g_UvmPaDescCache, pDescNode);
        }
        up_write(pLock[i / 2]);
    }

    return retCode;
}
