/*******************************************************************************
    Copyright (c) 2015-2015 NVIDIA Corporation

    Permission is hereby granted, free of charge, to any person obtaining a copy
    of this software and associated documentation files (the "Software"), to
    deal in the Software without restriction, including without limitation the
    rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
    sell copies of the Software, and to permit persons to whom the Software is
    furnished to do so, subject to the following conditions:

        The above copyright notice and this permission notice shall be
        included in all copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
    IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
    FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
    THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
    LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
    FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
    DEALINGS IN THE SOFTWARE.

*******************************************************************************/
//
// uvm_full_pa_mgmt.h
//
// This file contains data-structure and api definitions for PA management per
// device. This includes the PA to VA Trie, which is required for finding the VA
// entry in page table corresponding to the given PA.
//

#ifndef _UVM_FULL_PA_MGMT_H_
#define _UVM_FULL_PA_MGMT_H_

#include "uvm_full.h"
#include "uvm_linux.h"
#include "uvm_full_types.h"
#include "uvmtypes.h"

#include <linux/radix-tree.h>   // Linux Kernel Radix tree structure.

// Changing this size to a value bigger than 2M and then making a physical
// allocation will break the code.
#define UVM_PA_MANAGER_CHUNK_SIZE_DEFAULT                           PAGE_SIZE_2M
#define UVM_PA_MANAGER_CHUNK_SIZE_DEFAULT_SHIFT                     PAGE_SHIFT_2M

#define UVM_PA_MANAGER_SUB_ALLOCATOR_SMALLEST_PAGE_SIZE_DEFAULT     PAGE_SIZE_4K
#define UVM_PA_MANAGER_SUB_ALLOCATOR_SMALLEST_PAGE_SIZE_DEFAULT_SHIFT PAGE_SHIFT_4K

#define UVM_PA_MANAGER_SUB_ALLOCATOR_SLOT_TOTAL (UVM_PA_MANAGER_CHUNK_SIZE_DEFAULT / \
                    UVM_PA_MANAGER_SUB_ALLOCATOR_SMALLEST_PAGE_SIZE_DEFAULT)

/**
 * @struct UvmPaPool_tag
 * @brief  This structure contains the linked list to keep track of free memory
 *         as well as used memory. UvmPaManager tracks sub-allocated as well as
 *         default memory pool using this structure.
 */
struct UvmPaPool_tag
{
    //! List to track all free nodes.
    struct list_head        freeHead;

    //! List to track all free nodes.
    struct list_head        usedHead;

    struct rw_semaphore     lock;
};

/**
 * @struct UvmPaManager_tag
 * @brief  This structure describes the PA Manager which
 *         handles all the PA for a single GPU / CPU device.
 *         Create 1 instance per device.
 */
struct UvmPaManager_tag
{
    //! Lock to be taken before trie traversal.
    struct rw_semaphore     trieLookupLock;

    //! Represent the number of bits in PA.
    NvU64                   bitsForPA;

    //! This represent no. of bit for leaf page managed by the descriptor.
    //! For GPU PA Manager with default alloc size as 2MB, this represents 21 bits.
    //! Similarly for CPU PA Manager, this can be 12 bits (for 4K) or 16 bits
    //! (for 64K page used on Power PC).
    NvU64                   bitsForLeafPage;

    //! This flag helps identify whether this UvmPaManager object is for GPU or CPU.
    //! When true, it indicates that this PA Manager object tracks GPU PA.
    //! When false, it indicates that this PA Manager object tracks CPU PA.
    NvBool                  bIsGpuPaMngr;

    //! When bIsGpuPaMngr = true, this field holds the internal gpu_id for the
    //! GPU whose memory is tracked by this UvmPaManager. This is also the index
    //! into the global GPU device array.
    UvmGpuId                gpuId;

    //! When bIsGpuPaMngr = true, this field caches the PMA object corresponding
    //! to the above GPU ID.
    void                    *pPma;

    //! This represents the root node of the Radix tree. This tree keeps track
    //! of memory chunk corresponding to the pfn. Each pfn maps to UvmPaDesc.
    struct radix_tree_root  paRoot;

    //! List to track all the sub allocated memory. This helps tracking
    //! available memory for vid memory allocation request and for cleanup.
    UvmPaPool               subAllocMemHead;

    //! List to track all the non sub allocated memory. This helps tracking
    //! available memory for vid memory allocation request and for cleanup.
    UvmPaPool               defaultMemHead;
};

/**
 * @struct UvmPaDesc_tag
 * @brief This is the descriptor section of each leaf node of radix tree. Each
 *        descriptor helps track the physical memory allocation of default size
 *        on a device. If the node is subAllocated (bIsSuballocated), then this
 *        descriptor also stores the state of the sub-allocated memory. The Radix
 *        tree in the UvmPaManager stores one descriptor of this type per pfn key.
 */
struct UvmPaDesc_tag
{
    //! Lock to be taken before descriptor operations.
    struct rw_semaphore     descLock;

    //! This variable helps keep track whether the memory exists for this
    //! descriptor or its free'ed.
    NvU8                    bBackingExists         :1;

    //! Whether this descriptor is managed by sub-allocator or is allocated
    //! completely.
    NvU8                    bIsSubAllocated        :1;

    //! PA address for start of UVM_PA_MANAGER_CHUNK_SIZE_DEFAULT sized allocation.
    NvU64                   pa;

    //! Locks must always be acquired starting with PA node followed by VA node.
    //! So while VA-> PA travesring, we incremented seqNum and cache it.
    //! and then try to acquire PA lock first followed by VA lock.
    //! PA lock is acquired only if cache seqNum is same as that in the PA node.
    NvU64                   seqNum;

    //! Array to track the suballocations for this default sized memory block.
    //! If 1 bit represents one 4KB allocation then, no. if bits = 2MB / (4KB * 32)
    NvU32                   subAllocMap[UVM_PA_MANAGER_SUB_ALLOCATOR_SLOT_TOTAL /
                                        (sizeof(NvU32) * 8)];

    //! This field is used to keep track of empty slots in this allocation.
    //! 1 bit represents 1 sub-allocator slot.
    NvU32                   usedBitCount;

    //! List to track all sub allocated & nonSubAllocNode nodes.
    //! if (bIsSubAllocated = TRUE), this points to the sub allocated desc list
    //! else it points to the nonSubAllocate desc list.
    struct list_head        node;

    //! List to reverse map the global list.
    struct list_head        reverseMapList;
};

/**
 * @struct UvmPaToVa_tag
 * @brief This describes a single PA to VA mapping (4k granular)
 *        for the given physical address on a single GPU or on-sysmem.
 */
struct UvmPaToVa_tag
{
    //! VA address corresponding to the given PA.
    NvU64                   va;

    //! Range of address allocated from VA.
    NvU64                   range;

    //! UVM va space.
    struct UvmVaspace       *pVaspace;

    //! Node in the reverseMapList
    struct list_head        paDescNode;

    //! Node in mapping list in VA descriptor.
    struct list_head        vaDescNode;
};

struct UvmPaAllocationOptions_tag
{
    NvU8                    bContiguous         :1;
    NvU8                    bPinned             :1;
};

/**
 * -----------------------------------------------------------------------------
 * PA Mgmt Slab allocator initialization and destroy.
 * -----------------------------------------------------------------------------
 */

/**
 * uvmfull_pa_mgmt_slab_init
 *
 * This function initializes the slab allocator for the UvmPaManager.
 *
 * @return NV_STATUS NV_OK if succeeded; or else other specific errors.
 */
NV_STATUS uvmfull_pa_mgmt_slab_init(void);

/**
 * uvmfull_pa_mgmt_slab_destroy
 *
 * This function destroys the slab allocator for the UvmPaManager.
 *
 * @return NV_STATUS NV_OK if succeeded; or else other specific errors.
 */
void uvmfull_pa_mgmt_slab_destroy(void);

/**
 * -----------------------------------------------------------------------------
 * Object allocation using Slab allocator.
 * -----------------------------------------------------------------------------
 */

/**
 * uvmfull_alloc_pa_manager
 *
 * Allocates an instance of UvmPaManager struct using slab allocator.
 *
 * @param[in]   bitsForPA       Represents size of PA in bits.
 * @param[in]   bitsForLeafPage Represents size of leaf page managed by the descriptor.
 * @param[in]   bIsGpuPaMngr    Identifies if this is GPU or CPU PA Manager.
 * @param[in]   gpuId           If bIsGpuPaMngr = true, this fields stored GPU id.
 *
 * @return pointer to allocated UvmPaManager if succeeded; or else returns NULL.
 */
UvmPaManager* uvmfull_alloc_pa_manager(NvU64   bitsForPA,
                                       NvU64   bitsForLeafPage,
                                       NvBool  bIsGpuPaMngr,
                                       UvmGpuId gpuId);

/**
 * ------------------------------------------------------------
 * UvmPaDesc struct manipulation APIs
 * ------------------------------------------------------------
 */

/**
 * uvmfull_get_pa_desc
 *
 * Given a PA manager and the PA of interest, this function returns the desc
 * node having corresponding PA tracked by it.
 * Caller must acquire the lock on decriptor before modifying the desc pointer.
 *
 * @param[in]   pPaMngr          Per device based PA Manager
 * @param[in]   pa               Represents start PA for which the descriptor is reqd.
 *
 * @return pointer to UvmPaDesc corresponding to PA passed.
 *         Else it returns NULL if descriptor is not found.
 */
UvmPaDesc * uvmfull_get_pa_desc(UvmPaManager *pPaMngr,
                                NvU64 pa);

/**
 * uvmfull_allocate_vid_mem
 *
 * Given a root of the UvmPaManager and the chunk count with each allocation size,
 * this function asks PMA module to allocate those many pages, each of specified
 * chunkSize. The function returns a list of base PA of those allocations.
 *
 * @param[in]   pPaMngr         Per device based PA Manager.
 * @param[in]   chunkCount      Number of chunks of size chunkSize to be allocated.
 * @param[in]   chunkSize       Size of each allocation.
 * @param[in]   allocFlags      Parameters to configure Phy mem allocation.
 * @param[out]  pPages          Fills the array with PA of allocated memory.
 * @param[out]  pOutCount       Count of number of chunks successfully allocated.
 *
 * @return NV_STATUS NV_OK if succeeded; or else other specific errors.
 */
NV_STATUS uvmfull_allocate_vid_mem(UvmPaManager *pPaMngr,
                                   NvU64 chunkCount,
                                   NvU32 chunkSize,
                                   UvmPaAllocationOptions allocFlags,
                                   NvU64 *pPages,
                                   NvU64 *pOutCount);

/**
 * uvmfull_pat_destroy
 *
 * Given a root of the Trie this function destroys all entries from the PA Trie.
 *
 * @param[in]   pPaMngr         Per device based PA Manager.
 *
 * @return NV_STATUS NV_OK if succeeded; or else other specific errors.
 */
NV_STATUS uvmfull_pa_mngr_destroy(UvmPaManager *pPaMngrPerDev);

#endif   //_UVM_FULL_PA_MGMT_H_
