/*******************************************************************************
    Copyright (c) 2014-2015 NVIDIA Corporation

    Permission is hereby granted, free of charge, to any person obtaining a copy
    of this software and associated documentation files (the "Software"), to
    deal in the Software without restriction, including without limitation the
    rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
    sell copies of the Software, and to permit persons to whom the Software is
    furnished to do so, subject to the following conditions:

        The above copyright notice and this permission notice shall be
        included in all copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
    IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
    FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
    THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
    LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
    FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
    DEALINGS IN THE SOFTWARE.

*******************************************************************************/

#include "uvm_full.h"
#include "uvm_common.h"
#include "uvm_utils.h"
#include "nv_uvm_interface.h"
#include "uvm_full_pte_flags_getset.h"
#include "uvm_full_device_mgmt.h"
#include "uvm_full_pagetbl_mgmt.h"
#include "uvm_full_ctx_mgmt.h"
#include "uvm_hashmap.h"
#include "uvm_full_fault_handler.h"

#define UVM_PUSHBUFFER_INLINE_REGION_SIZE_MAX     (((1 << 13) - 1) * 4)
#define UVM_MEMSET_BUFFER_SIZE_MAX                64

static struct kmem_cache * g_uvmGpuPdeCache __read_mostly = NULL;
static struct kmem_cache * g_uvm2MVaDescCache __read_mostly = NULL;
static struct kmem_cache * g_uvmPbOpsCache __read_mostly = NULL;

#ifndef NV_MIN
#define NV_MIN(a, b) ((a)<(b)?(a):(b))
#endif

#ifndef NV_MAX
#define NV_MAX(a, b) ((a)>(b)?(a):(b))
#endif

NV_STATUS uvmfull_pagetbl_mgmt_slab_init(void)
{
    NV_STATUS status = NV_OK;
    g_uvm2MVaDescCache = NV_KMEM_CACHE_CREATE("uvm_2mvadesc", Uvm2MbVaDesc);
    if (!g_uvm2MVaDescCache)
    {
        status = NV_ERR_NO_MEMORY;
        goto cleanup;
    }
    g_uvmGpuPdeCache = NV_KMEM_CACHE_CREATE("uvm_gpupde", UvmGpuPde);
    if (!g_uvmGpuPdeCache)
    {
        status = NV_ERR_NO_MEMORY;
        goto cleanup;
    }
    g_uvmPbOpsCache = NV_KMEM_CACHE_CREATE("uvm_pbops", UvmPbOps);
    if (!g_uvmPbOpsCache)
    {
        status = NV_ERR_NO_MEMORY;
        goto cleanup;
    }
    return status;

    cleanup: uvmfull_pagetbl_mgmt_slab_destroy();
    return status;
}

void uvmfull_pagetbl_mgmt_slab_destroy(void)
{
    if (g_uvm2MVaDescCache)
        kmem_cache_destroy_safe(&g_uvm2MVaDescCache);
    if (g_uvmGpuPdeCache)
        kmem_cache_destroy_safe(&g_uvmGpuPdeCache);
    if (g_uvmPbOpsCache)
        kmem_cache_destroy_safe(&g_uvmPbOpsCache);
}

NvU32 create_pte_template(UvmGpuVaspace *gpuVaspace, NvU8 *pteTemplate,
                        NvU8 templateSize,
                        NvBool bValid, NvBool bVol, UvmMemType aperture,
                        NvBool bAtomic, NvBool bReadOnly)
{
    GMMU_FMT_PTE *pteFmt;
    GMMU_APERTURE gmmuAperture = GMMU_APERTURE_INVALID;
    GMMU_FMT *gmmuFmt = gpuVaspace->client.pGmmuFmt;
    NvU8 *buffer = pteTemplate;
    NvU32 entrySize = gpuVaspace->pteDefaults.entrySize;
    if (!gmmuFmt || !buffer || (entrySize > templateSize))
        return 0;

    pteFmt = (GMMU_FMT_PTE*)gmmuFmt->pPte;
    // Start with base template
    memset(buffer, 0x0, entrySize);

    nvFieldSetBool(&pteFmt->fldReadOnly, !!bReadOnly, buffer);
    nvFieldSetBool(&pteFmt->fldAtomicDisable, !bAtomic, buffer);
    nvFieldSetBool(&pteFmt->fldValid, !!bValid, buffer);
    nvFieldSetBool(&pteFmt->fldVolatile, !!bVol, buffer);

    gmmuAperture = uvm_memtype_to_gmmu_aperture(aperture);
    UVM_PANIC_ON(gmmuAperture == GMMU_APERTURE_INVALID);
    gmmuFieldSetAperture(&pteFmt->fldAperture, gmmuAperture, buffer);
    return entrySize;
}

static
void _create_pte_update_from_template(UvmGpuVaspace *gpuVaspace,
                                        UvmGpuPteUpdateParams *pteUpdate,
                                        NvU8 *pteTemplate)
{
    GMMU_FMT_PTE *pteFmt;
    GMMU_APERTURE aperture = GMMU_APERTURE_INVALID;
    GMMU_FMT *gmmuFmt = gpuVaspace->client.pGmmuFmt;
    pteFmt = (GMMU_FMT_PTE*)gmmuFmt->pPte;

    pteUpdate->readonly = nvFieldGetBool(&pteFmt->fldReadOnly, pteTemplate);
    pteUpdate->atomic = !nvFieldGetBool(&pteFmt->fldAtomicDisable, pteTemplate);
    pteUpdate->valid = nvFieldGetBool(&pteFmt->fldValid, pteTemplate);
    pteUpdate->vol = nvFieldGetBool(&pteFmt->fldVolatile, pteTemplate);

    aperture = gmmuFieldGetAperture(&pteFmt->fldAperture, pteTemplate);
    if (aperture == GMMU_APERTURE_VIDEO)
        pteUpdate->aperture = UvmMemType_vidmem;
    else if (aperture == GMMU_APERTURE_SYS_COH)
        pteUpdate->aperture = UvmMemType_sysmem_coh;
    else if (aperture == GMMU_APERTURE_SYS_NONCOH)
        pteUpdate->aperture = UvmMemType_sysmem_ncoh;
    else if (aperture == GMMU_APERTURE_PEER)
        pteUpdate->aperture = UvmMemType_peer;
    else
        pteUpdate->aperture = UvmMemType_invalid;
    pteUpdate->updateSwFlags |= UVM_GPU_SW_PTE_UPDATE_FLAGS_PTE_VALID;
    pteUpdate->updateSwFlags |= UVM_GPU_SW_PTE_UPDATE_FLAGS_VOLATILE;
    pteUpdate->updateSwFlags |= UVM_GPU_SW_PTE_UPDATE_FLAGS_ATOMIC;
    pteUpdate->updateSwFlags |= UVM_GPU_SW_PTE_UPDATE_FLAGS_READ_ONLY;
    pteUpdate->updateSwFlags |= UVM_GPU_SW_PTE_UPDATE_FLAGS_APERTURE;
}

static void _update_pte_hints (UvmPteHintsParams *pPteHintsParams,
                       UvmPteHints       *pPte)
{
    NvU32 pteHintFlags = pPteHintsParams->updateFlags;

    if (pteHintFlags & UVM_GPU_VA_PTE_HINTS_FLAG_SET_HOME_PROCESSOR)
        pPte->homeProcessor = pPteHintsParams->pte.homeProcessor;
    if (pteHintFlags & UVM_GPU_VA_PTE_HINTS_FLAG_SET_MANAGED_VA)
        pPte->managedVa = pPteHintsParams->pte.managedVa;
    if (pteHintFlags & UVM_GPU_VA_PTE_HINTS_FLAG_SET_NOT_MIGRATABLE)
        pPte->notMigratable = pPteHintsParams->pte.notMigratable;
    if (pteHintFlags & UVM_GPU_VA_PTE_HINTS_FLAG_SET_PROT_READ)
        pPte->va_prot_read = pPteHintsParams->pte.va_prot_read;
    if (pteHintFlags & UVM_GPU_VA_PTE_HINTS_FLAG_SET_PROT_WRITE)
        pPte->va_prot_write = pPteHintsParams->pte.va_prot_write;
    if (pteHintFlags & UVM_GPU_VA_PTE_HINTS_FLAG_SET_FLT_PLCY_RW)
        pPte->testPolicyReadWrite = pPteHintsParams->pte.testPolicyReadWrite;
    if (pteHintFlags & UVM_GPU_VA_PTE_HINTS_FLAG_SET_FLT_PLCY_ATOMIC)
        pPte->testPolicyAtomic = pPteHintsParams->pte.testPolicyAtomic;
    if (pteHintFlags & UVM_GPU_VA_PTE_HINTS_FLAG_SET_FLT_PLCY_RD_DUP)
        pPte->testPolicyRdDuplicate = pPteHintsParams->pte.testPolicyRdDuplicate;
    if (pteHintFlags & UVM_GPU_VA_PTE_HINTS_FLAG_SET_FLT_PLCY_SYSONLY)
        pPte->testPolicySysMemOnly = pPteHintsParams->pte.testPolicySysMemOnly;
    if (pteHintFlags & UVM_GPU_VA_PTE_HINTS_FLAG_SET_FLT_PLCY_CANCEL)
        pPte->testPolicyCancel = pPteHintsParams->pte.testPolicyCancel;
    if (pteHintFlags & UVM_GPU_VA_PTE_HINTS_FLAG_SET_CPU_PFN_VALID)
        pPte->cpuPfnValid = pPteHintsParams->pte.cpuPfnValid;
    if (pteHintFlags & UVM_GPU_VA_PTE_HINTS_FLAG_SET_CPU_PFN_VALUE)
        pPte->cpuPfn = pPteHintsParams->pte.cpuPfn;
    if (pteHintFlags & UVM_GPU_VA_PTE_HINTS_FLAG_SET_CPU_PFN_MAPPED)
        pPte->cpuPfnMapped = pPteHintsParams->pte.cpuPfnMapped;
    if (pteHintFlags & UVM_GPU_VA_PTE_HINTS_FLAG_SET_CPU_PFN_DIRTY)
        pPte->cpuPfnDirty = pPteHintsParams->pte.cpuPfnDirty;
    if (pteHintFlags & UVM_GPU_VA_PTE_HINTS_FLAGS_SET_RM_MEMORY)
        pPte->rmMemory = pPteHintsParams->pte.rmMemory;

    if (pteHintFlags & (UVM_GPU_VA_PTE_HINTS_FLAG_SET_FLT_PLCY_RW
                        | UVM_GPU_VA_PTE_HINTS_FLAG_SET_FLT_PLCY_ATOMIC
                        | UVM_GPU_VA_PTE_HINTS_FLAG_SET_FLT_PLCY_SYSONLY
                        | UVM_GPU_VA_PTE_HINTS_FLAG_SET_FLT_PLCY_RD_DUP
                        | UVM_GPU_VA_PTE_HINTS_FLAG_SET_FLT_PLCY_CANCEL))
        pPte->testPolicyEnabled = NV_TRUE;
}

void uvmfull_update_pte_hints(NvU64 addr, NvU64 numPages, NvU32 pgSize,
        UvmPteHintsParams *pteHintsParams, Uvm2MbVaDesc *pVaDesc2M)
{
    NvU32 i, pteoffset;

    if (pVaDesc2M)
    {
        for (i = 0; i < numPages ; i++)
        {
            pteoffset = GET_PDE_PTE_OFFSET(addr, GET_PAGE_SHIFT_MASK(pgSize));
            if (pteoffset >= GPU_PDE_MAX_PAGES)
            {
                UVM_PANIC_ON(1); // This is a bug, the numpages passed in exceeds GPU_PDE_SIZE
                break;
            }
            _update_pte_hints(pteHintsParams, &pVaDesc2M->hints[pteoffset]);

            addr += pgSize;
        }
    }
}

static NvBool
_check_pte_hints_managedva(NvU64 addr, NvU64 numPages, NvU32 pgSize,
                            Uvm2MbVaDesc *vaDesc)
{
    NvU32 i, pteoffset;
    for (i = 0; i < numPages ; i++)
    {
        pteoffset = GET_PDE_PTE_OFFSET(addr, GET_PAGE_SHIFT_MASK(pgSize));
        UVM_PANIC_ON(pteoffset >= GPU_PDE_MAX_PAGES);
        if (!vaDesc->hints[i].managedVa)
            return NV_FALSE;
        addr += pgSize;
    }
    return NV_TRUE;
}

// Initialize the 2Mb Va desc with default flags (i.e managed flag)
// Need to have a write lock on pUvm2MbVaDesc
static void _init_vadesc_flags(Uvm2MbVaDesc *vaDesc)
{
    // Nothing specific set. This is a placeholder.
}

void uvmfull_unmap_gpu_pde_pages(UvmVaspace *vaspace, UvmGpuId gpuId,
                                 UvmPbOps *pbOps, Uvm2MbVaDesc *vaDesc, 
                                 UvmGpuPde *gpuPde, NvU64 vaddr,
                                 NvU32 numPages, NvBool skipTlbInvalidate)
{
    NV_STATUS status = NV_OK;
    NvU64 gpuVa = vaddr;
    NvU32 ptePgSize;
    UvmGpuPteUpdateParams pteParams;
    UvmGpuPte *gpuPte;

    while (numPages)
    {
        memset(&pteParams, 0, sizeof(pteParams));
        gpuPte = uvmfull_get_gpupde_current_pte(gpuPde,
                gpuVa, &ptePgSize);
        if (!gpuPte->swcachevalid || !gpuPte->valid)
        {
            UVM_PANIC_ON(numPages < (ptePgSize/PAGE_SIZE_4K));
            numPages -= ptePgSize/PAGE_SIZE_4K;
            gpuVa += ptePgSize;
            continue;
        }

        pteParams.updateSwFlags |= UVM_GPU_SW_PTE_UPDATE_FLAGS_PTE_VALID;
        pteParams.valid = 0x0;
        pteParams.updateSwFlags |= UVM_GPU_SW_PTE_UPDATE_FLAGS_VOLATILE;
        pteParams.vol = 0x0;

        // disable READ/ATOMIC mappings if any
        pteParams.updateSwFlags |= UVM_GPU_SW_PTE_UPDATE_FLAGS_READ_ONLY;
        pteParams.readonly = NV_FALSE;
        pteParams.updateSwFlags |= UVM_GPU_SW_PTE_UPDATE_FLAGS_ATOMIC;
        pteParams.atomic = NV_FALSE;

        // Actually update the PTE
        pteParams.updateHwFlags |= UVM_GPU_HW_PTE_UPDATE_FLAGS_WRITE_PTE;

        status = uvmfull_update_gpupde_ptes(vaDesc, gpuId,
                gpuPde, gpuVa, 1, ptePgSize,
                &pteParams, NULL, pbOps,
                (NvU8*)&vaspace->gpuVaspaces[gpuId].pteDefaults.entry,
                vaspace->gpuVaspaces[gpuId].pteDefaults.entrySize,
                skipTlbInvalidate);
        WARN_ON(status != NV_OK);
        UVM_PANIC_ON(numPages < (ptePgSize/PAGE_SIZE_4K));
        numPages -= ptePgSize/PAGE_SIZE_4K;
        gpuVa += ptePgSize;
    }
}

NV_STATUS uvmfull_unmap_on_gpus(UvmVaspace *vaspace, NvU64 base,
                                NvU64 length, NvU8 *gpuids, NvU32 numGpus,
                                NvBool skipTlbInvalidate)
{
    Uvm2MbVaDesc *vaDesc;
    NvU64 vaddr = UVM_ALIGN_ADDR(base, PAGE_SIZE_4K);
    NvU32 remPages, i;
    NvU32 numPages = 0;
    UvmVatContext query;
    UvmPbOps *pbOps = NULL;
    NV_STATUS status = NV_OK;
    UvmGpuPde *gpuPde;
    NvBool bVatLeafAcquired = NV_FALSE;

    remPages = DIV_ROUND_UP(length, PAGE_SIZE_4K);

    // limit to PDE page size
    while (remPages)
    {
        // calculate number of pages that fall within this PDE
        numPages = GPU_PDE_MAX_PAGES - GET_PDE_PTE_OFFSET(vaddr, PAGE_SHIFT_4K);
        numPages = numPages > remPages ? remPages : numPages;

        query.accessType = UvmVatAccess_GenericQuery;
        query.address = vaddr;
        query.vaspace = vaspace;
        vaDesc = uvmfull_vat_acquire_leaf(&vaspace->vat, &query);
        if (!vaDesc)
        {
            vaddr += (numPages*PAGE_SIZE_4K);
            remPages -= numPages;
            continue;
        }
        bVatLeafAcquired = NV_TRUE;

        // Get pushbuffer.
        pbOps = uvm_pbops_start(vaspace);
        if (!pbOps)
        {
            status = NV_ERR_NO_MEMORY;
            goto cleanup;
        }
        uvm_pbops_reserve_acquire_space(pbOps, &vaDesc->tracker);

        for (i = 0; i < numGpus; i++)
        {
            UvmGpuId gpuId = (UvmGpuId)gpuids[i];

            gpuPde = uvmfull_get_gpupde_in_vadesc(vaDesc, gpuId);
            if (!gpuPde)
                continue;

            // 2MB support is not there yet. TODO: Will need to check that as well.
            //TODO skip tlb invalidate -> move it to push replay
            uvmfull_unmap_gpu_pde_pages(vaspace, gpuId, pbOps, vaDesc, gpuPde,
                                 vaddr, numPages, skipTlbInvalidate);

            // Submit the pushbuffer.
            uvm_pbops_submit(pbOps, gpuId, uvm_pbops_gpu_pb(pbOps, gpuId),
                             &vaDesc->tracker);
        }  // per gpu operation

        uvm_move_tracker(&vaDesc->tracker, pbOps->t);
        status = uvm_wait_for_tracker(&vaDesc->tracker);
        if (status != NV_OK)
            goto cleanup;

        uvm_pbops_end(pbOps);
        pbOps = NULL;
        uvmfull_vat_release_leaf(&vaspace->vat, &query);
        bVatLeafAcquired = NV_FALSE;
        vaddr += numPages * PAGE_SIZE_4K;
        remPages -= numPages;
    } // while(remPages)

cleanup:
    if (pbOps)
        uvm_pbops_end(pbOps);
    if (bVatLeafAcquired)
        uvmfull_vat_release_leaf(&vaspace->vat, &query);
    // TODO transactional cleanup path if status != NV_OK
    return status;

}

// Function to unmap leaf pageTable from PDE whenever all PTE's have become zero
NV_STATUS uvmfull_reclaim_leaf_gpu_pde(Uvm2MbVaDesc *vaDesc)
{
    NvU32 entryId;
    NV_STATUS status = NV_OK;
    UvmGpuPde *gpuPde, *tempGpuPde;
    UvmPbOps *pbOps;
    UvmGpuVaspace *gpuVaspace;
    UvmVatL0 * level0;
    NvBool bPbSubmit = NV_FALSE;

    // Grab pusbuffer on all GPU
    pbOps = uvm_pbops_start(vaDesc->pVaspace);
    if (!pbOps)
        return NV_ERR_NO_MEMORY;

    uvm_pbops_reserve_acquire_space(pbOps, &vaDesc->tracker);

    desc_for_each_gpu_pde(gpuPde, vaDesc)
    {
        UvmPushbuffer * pb;

        if (!gpuPde->bParentEntryBigValid && !gpuPde->bParentEntrySmallValid)
            continue;

        gpuVaspace = &vaDesc->pVaspace->gpuVaspaces[gpuPde->hostGpuId];
        UVM_PANIC_ON(gpuVaspace == NULL);

        pb = uvm_pbops_gpu_pb(pbOps, gpuPde->hostGpuId);

        // Unmap all pageTable mappings at PDE0, big or small
        {
            uvmfull_update_gpu_pde_entry(gpuVaspace, pb, 0,
                                         gpuPde->parentAddr,
                                         gpuPde->parentAperture,
                                         gpuPde->parentEntryId,
                                         UVM_VA_SHIFT_TRIE_LVL_0_LO,
                                         0, UvmMemType_invalid,
                                         0, UvmMemType_invalid,
                                         NV_FALSE /*skipTlbInvalidate*/);
            gpuPde->bParentEntryBigValid = 0x0;
            gpuPde->bParentEntrySmallValid = 0x0;
        }

        // Submit the pushbuffer.
        uvm_pbops_submit(pbOps, gpuPde->hostGpuId, pb, &vaDesc->tracker);
        bPbSubmit = NV_TRUE;
    }

    if (bPbSubmit)
    {
        uvm_move_tracker(&vaDesc->tracker, pbOps->t);
        status = uvm_wait_for_tracker(&vaDesc->tracker);
        if (status != NV_OK)
            goto cleanup;
    }

    //
    // Fix-Me - Move this macro to a wrapper which takes GPUID and returns
    // correct entry depending upon GPU types
    //
    entryId  = VAT_EXTRACT_INDEX(vaDesc->addr,
                                 UVM_VA_SHIFT_TRIE_LVL_0_HI,
                                 UVM_VA_SHIFT_TRIE_LVL_0_LO);
    level0  = (UvmVatL0 *)GET_LEVEL_FROM_ENTRY(vaDesc->parent,
                                               UvmVatL0, UvmVatL0Entry,
                                               entryId);
    UVM_PANIC_ON(level0 == NULL);

    // Reduce valid entry count at level0 as we have unmapped an entry here
    --level0->header.validLevelEntries;

    // Free all pageTable memory associated with different PDEs
    desc_for_each_gpu_pde_safe(gpuPde, tempGpuPde, vaDesc)
    {
        GMMU_FMT *gmmuFmt;
        const MMU_FMT_LEVEL *pFmtLevel;
        UvmGpu *pGpu;
        UvmVatLevelGpuHeader *parentGpuHeader = NULL;

        if (!gpuPde->hasBigPageTable && !gpuPde->hasSmallPageTable)
            continue;

        gpuVaspace = &vaDesc->pVaspace->gpuVaspaces[gpuPde->hostGpuId];
        gmmuFmt = gpuVaspace->client.pGmmuFmt;
        pFmtLevel = mmuFmtFindLevelWithPageShift(gmmuFmt->pRoot,
                                                 UVM_VA_SHIFT_TRIE_LVL_0_LO);

        if (gpuPde->hasBigPageTable)
        {
            uvmfull_free_pagetbl_memory(
            PA_TO_PFN(gpuPde->pageTable.pageTablePhysAddr[UVM_GPU_PT_INDEX_BIG].addr),
            PAGE_SIZE_4K, /*256Bytes not supported else pageSize will be mmFmtSize*/
            /*(NvU64)mmuFmtLevelSize(&pFmtLevel->subLevels[UVM_GPU_PT_INDEX_BIG])*/
            gpuPde->pageTable.pageTablePhysAddr[UVM_GPU_PT_INDEX_BIG].aperture,
            gpuVaspace->gpuId);
            gpuPde->hasBigPageTable = 0x0;
        }

        if (gpuPde->hasSmallPageTable)
        {
            uvmfull_free_pagetbl_memory(
            PA_TO_PFN(gpuPde->pageTable.pageTablePhysAddr[UVM_GPU_PT_INDEX_SMALL].addr),
            (NvU64)mmuFmtLevelSize(&pFmtLevel->subLevels[UVM_GPU_PT_INDEX_SMALL]),
            gpuPde->pageTable.pageTablePhysAddr[UVM_GPU_PT_INDEX_SMALL].aperture,
            gpuVaspace->gpuId);
            gpuPde->hasSmallPageTable = 0x0;
        }

        //
        // TO DO - Need to check if sysmem is also backing this region.
        // Free client's memory backing this 2MB PDE
        //
        if (gpuPde->memBackingPresent)
        {
            pGpu = uvmfull_gpuid_to_gpu(gpuPde->hostGpuId);
            uvmfull_free_backing_gpu_pages(pGpu, gpuPde);
            gpuPde->memBackingPresent = 0x0;
        }

        // Unset the mask in parent GPU entry as this entry is no more valid
        if (vaDesc->parent->pdePresentMask & NVBIT64(gpuPde->hostGpuId))
        {
            // Get the gpuHeader and reduce the gpuValidEntry count
            parentGpuHeader = uvmfull_vat_get_gpu_header(gpuPde->hostGpuId, 
                                                         &level0->header);
            --parentGpuHeader->gpuValidEntries;
            vaDesc->parent->pdePresentMask &= ~NVBIT64(gpuPde->hostGpuId);
        }

        // Make sure to destroy the data structure holding pdeInfo
        list_del(&gpuPde->pdeNode);
        kmem_cache_free(g_uvmGpuPdeCache, gpuPde);
    }

cleanup:
    if (pbOps)
        uvm_pbops_end(pbOps);

    return status;
}

// Reclaims memory in the desc for the given range. Returns valid 4k regions
// left in the desc after a reclaim was done on the desc.
// If returned 0, the desc was freed by the function.
NvU32 uvmfull_reclaim_va_desc(Uvm2MbVaDesc *vaDesc, UvmVatL0Entry **parentLevel,
                              NvU64 vaddr, NvU32 n4k)
{
    NvU32 i, ref, pteoffset;
    struct page *pg;
    NvU64 addr = vaddr;
    NV_STATUS nvStatus = NV_OK;

    UVM_PANIC_ON(addr >= vaDesc->addr + PAGE_SIZE_2M ||
                 addr < vaDesc->addr);
    i = 0;
    UVM_PANIC_ON(vaDesc->nValid4k > GPU_PDE_MAX_PAGES);

    // Fill in the parent level as it's needed to upward traversal later
    if (parentLevel)
    {
        *parentLevel = vaDesc->parent;
    }

    while (i < n4k)
    {
        pteoffset = GET_PDE_PTE_OFFSET(addr, PAGE_SHIFT_4K);
        UVM_PANIC_ON(pteoffset >= GPU_PDE_MAX_PAGES);
        if (vaDesc->hints[pteoffset].managedVa == 1)
        {
            // TODO: Add gpu memory reclaim here. (i may increment by more than 1)

            if (vaDesc->hints[pteoffset].cpuPfnValid)
            {
                pg = pfn_to_page(vaDesc->hints[pteoffset].cpuPfn);
                UVM_PANIC_ON(!pg);

                __free_page(pg);
            }
            memset(&vaDesc->hints[pteoffset], 0x0, sizeof(UvmPteHints));
            UVM_PANIC_ON(vaDesc->nValid4k == 0);
            vaDesc->nValid4k--;
        }
        addr += PAGE_SIZE_4K;
        i++;
    }

    ref = vaDesc->nValid4k;
    if (ref == 0)
    {
        // Wait for all trackers to complete
        uvm_wait_for_tracker(&vaDesc->tracker);
        uvm_reset_tracker(&(vaDesc->tracker));
        uvm_shrink_tracker(&(vaDesc->tracker));

        // unlink from parent
        vaDesc->parent->child = NULL;

        //
        // Unmap pageTable from PDE0 and destroy pageTable memory
        // This call also destroys the client backing serviced by this 2MB PDE
        //
        nvStatus = uvmfull_reclaim_leaf_gpu_pde(vaDesc);
        WARN_ON(nvStatus != NV_OK);
        kmem_cache_free(g_uvm2MVaDescCache, vaDesc);
    }
    return ref;
}

void uvmfull_delete_va_desc(Uvm2MbVaDesc *vaDesc)
{
    int pte4k;
    struct page *pg;
    for (pte4k = 0; pte4k < NUM_4K_PER_2M; ++pte4k)
    {
        if (vaDesc->hints[pte4k].cpuPfnValid)
        {
            pg = pfn_to_page(vaDesc->hints[pte4k].cpuPfn);
            UVM_PANIC_ON(!pg);

            __free_page(pg);
        }
    }
    uvm_reset_tracker(&(vaDesc->tracker));
    uvm_shrink_tracker(&(vaDesc->tracker));
    kmem_cache_free(g_uvm2MVaDescCache, vaDesc);
}

// should always be called with pVaspace->lock taken
Uvm2MbVaDesc* uvmfull_alloc_va_desc(UvmVaspace *pVaspace, NvU64 addr)
{
    Uvm2MbVaDesc *vaDesc = NULL;
    if (!pVaspace)
        return NULL;

    vaDesc = kmem_cache_zalloc(g_uvm2MVaDescCache, NV_UVM_GFP_FLAGS);
    if (!vaDesc)
        return NULL;

    INIT_LIST_HEAD(&vaDesc->gpuPDEs);
    vaDesc->addr = (addr & GPU_PDE_MASK);
    vaDesc->pVaspace = pVaspace;

    uvm_init_tracker(&(vaDesc->tracker));
    _init_vadesc_flags(vaDesc);

    return vaDesc;
}

static NV_STATUS
_adjust_surface_map_info(UvmGpuSurfaceMappingInfo *surfMapInfo, NvU64 *mapSize)
{
    NvU64 jump;
    NvU32 i;
    NvU32 nPa = surfMapInfo->pageCount;

    // The surfaceMapInfo passed by RM is list of 4k pa-s even when the
    // page size > 4k. Here we shrink the pa array so that update_pte
    // works correctly. pageCount can be 1 if the range is contiguous.
    // pageCount is really 4k-paCount here.
    *mapSize = surfMapInfo->pageSize;
    jump = *mapSize/PAGE_SIZE_4K;

    // num of pa's should be a multiple of jump
    if (nPa > 1)
    {
        if (nPa % jump)
        {
            WARN_ON(nPa % jump);
            return NV_ERR_INVALID_STATE;
        }
        nPa = nPa/jump;
    }

    // convert all pa-s to pfns
    for (i = 0; i < nPa; i++)
    {
        surfMapInfo->pteArray[i] = surfMapInfo->pteArray[i * jump];
        surfMapInfo->pteArray[i] = PA_TO_PFN(surfMapInfo->pteArray[i]);
    }

    // Map to the mapSize granularity
    surfMapInfo->mappingLength = UVM_ALIGN_UP(surfMapInfo->mappingLength,
                                              *mapSize);
    return NV_OK;
}

// GpuTouch per gpu
//    to make sure all pdes are allocated and mapped till we get to the pte.
// TODOs to make the above simpler:
//   TODO : Update va-trie to take a gpumask instead of id
NV_STATUS uvmfull_map_gpus_unmanaged_va(UvmVaspace *vaspace, NvU64 base,
                                    NvU8 *gpuids,
                                    UvmGpuSurfaceMappingInfo *surfMapInfo,
                                    UvmGpuMappingAttributes *perGpuAttribs,
                                    NvLength attribCount)
{
    Uvm2MbVaDesc *vaDesc;
    NvU64 vaddr = UVM_ALIGN_ADDR(base, PAGE_SIZE_4K);
    NvU32 i, remPages;
    NvU32 numPages = 0;
    UvmVatContext query;
    UvmPbOps *pbOps = NULL;
    UvmPteHintsParams pteHints = {0};
    NV_STATUS status = NV_OK;
    NvU64 gpuMapPageSize = PAGE_SIZE_4K; // should be the cpu page size
    UvmGpuPde *gpuPde;
    NvU64 *pfns = surfMapInfo->pteArray;
    NvBool bVatLeafAcquired = NV_FALSE;

    status = _adjust_surface_map_info(surfMapInfo, &gpuMapPageSize);
    if (status != NV_OK)
        return status;

    // Update the va type to rm memory in the desc
    pteHints.updateFlags |= UVM_GPU_VA_PTE_HINTS_FLAGS_SET_RM_MEMORY;
    pteHints.pte.rmMemory = 1;
    remPages = DIV_ROUND_UP(surfMapInfo->mappingLength, PAGE_SIZE_4K);

    WARN_ON(attribCount > 1);  // Support to be added

    // limit to PDE page size
    while (remPages)
    {
        NvU32 nGpuPtes;
        query.accessType = UvmVatAccess_GPUTouch;
        query.address = vaddr;
        query.vaspace = vaspace;
        query.gpuId = (UvmGpuId)gpuids[0];   // Update to a msk for multi gpu
        vaDesc = uvmfull_vat_acquire_leaf(&vaspace->vat, &query);
        if (!vaDesc)
        {
            status = NV_ERR_NO_MEMORY;
            goto cleanup;
        }
        bVatLeafAcquired = NV_TRUE;

        // calculate number of pages that fall within this PDE
        numPages = GPU_PDE_MAX_PAGES - GET_PDE_PTE_OFFSET(vaddr, PAGE_SHIFT_4K);
        numPages = numPages > remPages ? remPages : numPages;
        nGpuPtes = (numPages * PAGE_SIZE_4K)/gpuMapPageSize;

        // check if range is managed. TODO: cleanup previous ranges if failure.
        if (_check_pte_hints_managedva(vaddr, numPages, PAGE_SIZE_4K, vaDesc))
        {
            status = NV_ERR_INVALID_ARGUMENT;
            goto cleanup;
        }
        // Now that va is created, update with va_flags for the interested va range
        uvmfull_update_pte_hints(vaddr, numPages, PAGE_SIZE_4K, &pteHints, vaDesc);

        // Do gpu mapping now. Get pushbuffer.
        pbOps = uvm_pbops_start(vaspace);
        if (!pbOps)
        {
            status = NV_ERR_NO_MEMORY;
            goto cleanup;
        }
        uvm_pbops_reserve_acquire_space(pbOps, &vaDesc->tracker);

        for (i = 0; i < attribCount; i++)
        {
            UvmGpuId gpuId = (UvmGpuId)gpuids[i];
            UvmGpuVaspace *gpuVaspace;
            UvmGpuPteUpdateParams pteParams = {0};
            gpuVaspace = &vaspace->gpuVaspaces[gpuId];
            gpuPde = uvmfull_get_gpupde_in_vadesc(vaDesc, gpuId);
            UVM_PANIC_ON(!gpuPde);
            status = uvmfull_activate_gpupte(vaDesc, gpuPde,
                                          gpuMapPageSize != PAGE_SIZE_4K,
                                          gpuMapPageSize == PAGE_SIZE_4K,
                                          gpuId,
                                          uvm_pbops_gpu_pb(pbOps, gpuId));
            if (status != NV_OK)
            {
                status = NV_ERR_NO_MEMORY;
                goto cleanup;
            }

            // First decode the template into the update struct
            _create_pte_update_from_template(gpuVaspace, &pteParams,
                                            surfMapInfo->pteTemplate);
            pteParams.updateSwFlags |= UVM_GPU_SW_PTE_UPDATE_FLAGS_SWVALID;
            pteParams.updateSwFlags |= UVM_GPU_SW_PTE_UPDATE_FLAGS_PFN_VALID;
            pteParams.updateSwFlags |= UVM_GPU_SW_PTE_UPDATE_FLAGS_PFN;
            // Actually update the PTE
            pteParams.updateHwFlags |= UVM_GPU_HW_PTE_UPDATE_FLAGS_WRITE_PTE;
            if (surfMapInfo->pageCount == 1)
                pteParams.updateSwFlags |=
                            UVM_GPU_SW_PTE_UPDATE_FLAGS_PFN_AUTOINCREMENT;

            // The pte values that would be committed to HW
            pteParams.swcachevalid = NV_TRUE;
            pteParams.pfnvalid = NV_TRUE;

            // Always follow what user requested. So the below can overwrite the
            // template passed by the caller.
            if (perGpuAttribs[i].gpuMappingType == UvmGpuMappingTypeReadWriteAtomic)
                pteParams.valid = NV_TRUE;
            else
                pteParams.valid = NV_FALSE;

            // Same case as above. Volatility is user's decision.
            if (perGpuAttribs[i].gpuCachingType == UvmGpuCachingTypeForceCached)
                pteParams.vol = NV_FALSE;
            else
                pteParams.vol = NV_TRUE;

            /*
            if (perGpuAttribs[i].gpuMappingType == UvmGpuMappingTypeReadWriteAtomic)
            {
                pteParams.atomic = NV_TRUE;
            }
            else
                pteParams.atomic = NV_FALSE;
            */
            // For now make everything atomic enabled
            pteParams.atomic = NV_TRUE;

            // update the pfn in all the GPU ptes
            status = uvmfull_update_gpupde_ptes(vaDesc, gpuId,
                       gpuPde, vaddr, nGpuPtes, gpuMapPageSize,
                       &pteParams, pfns, pbOps,
                       surfMapInfo->pteTemplate,
                       vaspace->gpuVaspaces[gpuId].pteDefaults.entrySize,
                       NV_FALSE);
            if (status != NV_OK)
            {
                status = NV_ERR_NO_MEMORY;
                goto cleanup;
            }

            // if contiguous make sure pfn is updated in place
            if (surfMapInfo->pageCount == 1)
                pfns[0] = pfns[0] + ((nGpuPtes * gpuMapPageSize)/PAGE_SIZE_4K);
            else
                pfns += nGpuPtes;

            // Submit the pushbuffer.
            uvm_pbops_submit(pbOps, gpuId, uvm_pbops_gpu_pb(pbOps, gpuId),
                             &vaDesc->tracker);
        }  // per gpu operation

        uvm_move_tracker(&vaDesc->tracker, pbOps->t);
        status = uvm_wait_for_tracker(&vaDesc->tracker);
        if (status != NV_OK)
            goto cleanup;

        uvm_pbops_end(pbOps);
        pbOps = NULL;
        uvmfull_vat_release_leaf(&vaspace->vat, &query);
        bVatLeafAcquired = NV_FALSE;

        vaDesc->nValid4k += numPages;
        vaddr += (numPages*PAGE_SIZE_4K);

        remPages -= numPages;
    } // while(remPages)

cleanup:
    if (pbOps)
        uvm_pbops_end(pbOps);
    if (bVatLeafAcquired)
        uvmfull_vat_release_leaf(&vaspace->vat, &query);
    // TODO transactional cleanup path if status != NV_OK
    //
    // We need reverse operation, something like 
    // uvmfull_unmap_gpus_unmanaged_va(vaspace, base, length, gpuids, 
    //                                 surfMapInfo, perGpuAttribs, count)
    // that would release gpu ptes that uvmfull_map_gpus_unmanaged_va has mapped so far. 
    // In worst case, length would be surfMapInfo->mappingLength (total reverse),
    // but it could be smaller (depending on when the status was not ok).
    return status;
}

NV_STATUS uvmfull_alloc_va_range(UvmVaspace *pVaspace, NvU64 addr, NvU64 length,
                                    UvmPteHintsParams *pPteHints)
{
    Uvm2MbVaDesc *vaDesc;
    NvU64 vaddr = UVM_ALIGN_ADDR(addr, PAGE_SIZE_4K);
    NvU32 remPages;
    NvU32 numPages = 0;
    NV_STATUS status = NV_OK;
    UvmVatContext query;

    remPages = DIV_ROUND_UP(length, PAGE_SIZE_4K);

    // limit to PDE page size
    while (remPages)
    {
        query.accessType = UvmVatAccess_GenericTouch;
        query.address = vaddr;
        query.vaspace = pVaspace;
        vaDesc = uvmfull_vat_acquire_leaf(&pVaspace->vat, &query);
        if (!vaDesc)
        {
            status = NV_ERR_NO_MEMORY;
            break;
        }

        // calculate number of pages that fall within this PDE
        numPages = GPU_PDE_MAX_PAGES - GET_PDE_PTE_OFFSET(vaddr, PAGE_SHIFT_4K);
        numPages = numPages > remPages ? remPages : numPages;

        // TODO check if no ptes are set in this range!

        // Now that va is created, update with va_flags for the interested va range
        uvmfull_update_pte_hints(vaddr, numPages, PAGE_SIZE_4K, pPteHints, vaDesc);
        vaDesc->nValid4k += numPages;
        vaddr += (numPages*PAGE_SIZE_4K);

        remPages -= numPages;
        uvmfull_vat_release_leaf(&pVaspace->vat, &query);
    }

    // TODO transactional cleanup path (together with vma_close) if status != NV_OK
    // BUG 1662555
    return status;
}

UvmGpuPde *uvmfull_get_gpupde_in_vadesc(Uvm2MbVaDesc *pUvm2MbVaDesc,
        UvmGpuId gpuId)
{
    UvmGpuPde *pGpuPde = NULL;
    list_for_each_entry(pGpuPde, &pUvm2MbVaDesc->gpuPDEs, pdeNode)
    {
        if (pGpuPde->hostGpuId == gpuId)
            return pGpuPde;
    }
    return NULL;
}

static
UvmMemType _vadesc_decide_pagetbl_aperture(Uvm2MbVaDesc *vaDesc, UvmGpuId gpuid)
{
    UvmMemType memType;
    UvmGpuVaspace *gpuVaspace = &vaDesc->pVaspace->gpuVaspaces[gpuid];

    if (gpuVaspace->bZeroFb)
        memType = UvmMemType_sysmem_coh;
    else if (gpuVaspace->policy.bRandomizePteAperture)
    {
        memType = gpuVaspace->policy.bPteApertureSysmem ?
                  UvmMemType_sysmem_coh :
                  UvmMemType_vidmem;
        // Flip for next time.
        gpuVaspace->policy.bPteApertureSysmem =
                !gpuVaspace->policy.bPteApertureSysmem;
    }
    else
        memType = UvmMemType_vidmem;

    return memType;
}

static
NV_STATUS _vadesc_allocate_gpu_pagetbl(Uvm2MbVaDesc *vaDesc, UvmGpuId gpuid,
                                    NvBool bBig, NvBool bSmall,
                                    UvmPushbuffer *pb)
{
    NvU64 pfn = 0;
    UvmMemType memType;
    UvmGpuVaspace *gpuVaspace = &vaDesc->pVaspace->gpuVaspaces[gpuid];
    UvmGpuPde *pGpuPde = uvmfull_get_gpupde_in_vadesc(vaDesc, gpuid);
    if (!pGpuPde)
        return NV_ERR_INVALID_REQUEST;

    memType = _vadesc_decide_pagetbl_aperture(vaDesc, gpuid);
    if (bBig && !pGpuPde->hasBigPageTable)
    {
        // Not allocated but parent points to it!!
        UVM_PANIC_ON(pGpuPde->bParentEntryBigValid);

        if (NV_OK != uvmfull_get_pagetbl_memory(PAGE_SIZE_4K, &pfn, memType,
                                                gpuVaspace->gpuId))
        {
            return NV_ERR_NO_MEMORY;
        }
        pGpuPde->pageTable.pageTablePhysAddr[UVM_GPU_PT_INDEX_BIG].addr
                = PFN_TO_PA(pfn);
        pGpuPde->pageTable.pageTablePhysAddr[UVM_GPU_PT_INDEX_BIG].aperture
                = memType;
        pGpuPde->hasBigPageTable = 0x1;

        // if pb is provided and vidmem allocation, scrub.
        if (pb && (memType == UvmMemType_vidmem))
            uvm_push_memzero_vidmem(gpuid, pb, PFN_TO_PA(pfn), PAGE_SIZE_4K);
    }

    if (bSmall && !pGpuPde->hasSmallPageTable)
    {
        // Not allocated but parent points to it!!
        UVM_PANIC_ON(pGpuPde->bParentEntrySmallValid);

        if (NV_OK != uvmfull_get_pagetbl_memory(PAGE_SIZE_4K, &pfn, memType,
                                                gpuVaspace->gpuId))
        {
            return NV_ERR_NO_MEMORY;
        }
        pGpuPde->pageTable.pageTablePhysAddr[UVM_GPU_PT_INDEX_SMALL].addr
                = PFN_TO_PA(pfn);
        pGpuPde->pageTable.pageTablePhysAddr[UVM_GPU_PT_INDEX_SMALL].aperture
                = memType;
        pGpuPde->hasSmallPageTable = 0x1;

        // if pb is provided and vidmem allocation, scrub.
        if (pb && (memType == UvmMemType_vidmem))
            uvm_push_memzero_vidmem(gpuid, pb, PFN_TO_PA(pfn), PAGE_SIZE_4K);
    }

    return NV_OK;
}

// Allocates the gpu ptes if required and updates the pushbuffer to link them to
// the parent level. (handles both small and big if bLinkBoth is true)
NV_STATUS uvmfull_activate_gpupte(Uvm2MbVaDesc *vaDesc, UvmGpuPde* gpuPde,
                                NvBool bLinkBig, NvBool bLinkSmall,
                                UvmGpuId gpuId, UvmPushbuffer *pb)
{
    NV_STATUS status = NV_OK;
    UvmGpuPageTable *pt = &gpuPde->pageTable;
    UvmGpuVaspace *gpuVaspace = &vaDesc->pVaspace->gpuVaspaces[gpuId];

    if ((bLinkBig && !gpuPde->bParentEntryBigValid) ||
        (bLinkSmall && !gpuPde->bParentEntrySmallValid))
    {
        status = _vadesc_allocate_gpu_pagetbl(vaDesc, gpuId,
                                              bLinkBig, bLinkSmall, pb);
        if (status != NV_OK)
            return status;

        // Now link the pt-s to the parent
        // If the pte is already activated; make sure not to de-activate it.
        gpuPde->bParentEntryBigValid |= (bLinkBig ? 0x1 : 0x0);
        gpuPde->bParentEntrySmallValid |= (bLinkSmall ? 0x1 : 0x0);
        uvmfull_update_gpu_pde_entry(gpuVaspace, pb, 0,
                gpuPde->parentAddr,
                gpuPde->parentAperture,
                gpuPde->parentEntryId, UVM_VA_SHIFT_TRIE_LVL_0_LO,
                pt->pageTablePhysAddr[UVM_GPU_PT_INDEX_BIG].addr,
                gpuPde->bParentEntryBigValid?
                pt->pageTablePhysAddr[UVM_GPU_PT_INDEX_BIG].aperture:
                UvmMemType_invalid,
                pt->pageTablePhysAddr[UVM_GPU_PT_INDEX_SMALL].addr,
                gpuPde->bParentEntrySmallValid?
                pt->pageTablePhysAddr[UVM_GPU_PT_INDEX_SMALL].aperture:
                UvmMemType_invalid,
                NV_FALSE /*skipTlbInvalidate*/);
    }

    return status;
}

NV_STATUS uvmfull_vadesc_update_gpupde_info(Uvm2MbVaDesc *vaDesc,
                                            UvmGpuId gpuId,
                                            UvmVatLevelGpuHeader *gpuHeader,
                                            NvU32 entryId)
{
    UvmGpuPde *pGpuPde = NULL;

    WARN_ON(!gpuHeader);

    pGpuPde = uvmfull_get_gpupde_in_vadesc(vaDesc, gpuId);
    if (!pGpuPde)
    {
        pGpuPde = uvmfull_alloc_gpupde(vaDesc, gpuId);
        if (!pGpuPde)
            return NV_ERR_NO_MEMORY;
    }
    // 16 bytes for PDE0 entry
    pGpuPde->parentAddr = gpuHeader->levelPhyAddr;;
    pGpuPde->parentAperture = gpuHeader->aperture;
    pGpuPde->parentEntryId = entryId;
    return NV_OK;
}

UvmGpuPde* uvmfull_alloc_gpupde(Uvm2MbVaDesc *pUvm2MbVaDesc, UvmGpuId gpuId)
{
    UvmGpuPde *pGpuPde = NULL;

    pGpuPde = kmem_cache_zalloc(g_uvmGpuPdeCache, NV_UVM_GFP_FLAGS);
    if (!pGpuPde)
        return NULL;

    INIT_LIST_HEAD(&pGpuPde->pdeNode);
    pGpuPde->hostGpuId = gpuId;
    list_add(&pGpuPde->pdeNode, &pUvm2MbVaDesc->gpuPDEs);
    return pGpuPde;
}

void uvmfull_destroy_gpupde(Uvm2MbVaDesc *pUvm2MbVaDesc, UvmGpuId gpuId)
{
    UvmGpuPde *pGpuPde = NULL;

    // Make sure passed in address is 2M aligned and less than 2M
    UVM_PANIC_ON(gpuId > UVM_FULL_MAX_GPUS);
    // check if a PDE already exists, if it does this is an error!
    pGpuPde = uvmfull_get_gpupde_in_vadesc(pUvm2MbVaDesc, gpuId);
    if (pGpuPde)
    {
        list_del(&pGpuPde->pdeNode);
        kmem_cache_free(g_uvmGpuPdeCache, pGpuPde);
    }
}

inline UvmGpuPte* uvmfull_get_gpupde_pte(UvmGpuPde *pGpuPde, NvU32 pgSize,
        NvU64 addr)
{
    switch (pgSize)
    {
    case PAGE_SIZE_4K:
        return &pGpuPde->pageTable.ptes[GET_GPU_PTECACHE_OFFSET(addr, PAGE_SIZE_4K)];
        break;
    case PAGE_SIZE_64K:
        return &pGpuPde->pageTable.ptes[GET_GPU_PTECACHE_OFFSET(addr, PAGE_SIZE_64K)];
        break;
    case PAGE_SIZE_2M:
        return &pGpuPde->pageTable.ptes[0];
        break;
    default:
        return NULL;
    }
    return NULL;
}

// This function looks at the current set flags 
UvmGpuPte* uvmfull_get_gpupde_current_pte(UvmGpuPde *pGpuPde, NvU64 addr,
        NvU32 *pPageSize)
{
    NvU64 bitOffset = 0;
    NvU32 pgSize = 0;
    UvmGpuPte *pPte = NULL;

    UVM_PANIC_ON(!pGpuPde);
    if (pGpuPde->is2MbPte)
    {
        pgSize = PAGE_SIZE_2M;
        pPte = &pGpuPde->pageTable.ptes[0];
    }
    else
    {
        bitOffset = GET_PDE_PTE_OFFSET(addr, PAGE_SHIFT_64K);
        if (pGpuPde->pageTable.bitmap64KPte & NVBIT64(bitOffset))
        {
            bitOffset = GET_GPU_PTECACHE_OFFSET(addr, PAGE_SIZE_64K);
            pPte = &pGpuPde->pageTable.ptes[bitOffset];
            pgSize = PAGE_SIZE_64K;
        }
        else
        {
            bitOffset = GET_GPU_PTECACHE_OFFSET(addr, PAGE_SIZE_4K);
            pPte = &pGpuPde->pageTable.ptes[bitOffset];
            pgSize = PAGE_SIZE_4K;
        }
    }

    if (pPageSize)
        *pPageSize = pgSize;

    return pPte;
}

// This function only allocates in contiguous sizes of 4k multiples.
// Examples: 4k, 16k etc
NV_STATUS uvmfull_get_pagetbl_memory(NvU64 size, NvU64 *pfn,
                                        UvmMemType aperture, UvmGpuId gpuId)
{
    NV_STATUS status = NV_OK;
    UvmGpu *gpu = NULL;
    WARN_ON(size & ~PAGE_MASK_4K);   // only 4k multiples allowed

    if (aperture == UvmMemType_vidmem)
    {
        gpu = uvmfull_gpuid_to_gpu(gpuId);
        status = uvmfull_allocate_phys_vid_mem(gpu, PAGE_SIZE_4K, pfn);
        return status;
    }

    if (aperture == UvmMemType_sysmem_coh || aperture == UvmMemType_sysmem_ncoh)
    {
        struct page* pg;
        pg = alloc_pages(NV_UVM_GFP_FLAGS | GFP_HIGHUSER | __GFP_ZERO, 0);
        if (!pg)
            return NV_ERR_NO_MEMORY;

        *pfn = page_to_pfn(pg);
    }

    return status;
}

// Frees the physical pages acquired for page tables.
void uvmfull_free_pagetbl_memory(NvU64 pfn, NvU64 size, UvmMemType aperture,
                                UvmGpuId gpuId)
{
    UvmGpu *gpu = NULL;
    WARN_ON(size & ~PAGE_MASK_4K);   // only 4k multiples allowed

    if (aperture == UvmMemType_vidmem)
    {
        gpu = uvmfull_gpuid_to_gpu(gpuId);
        uvmfull_free_phys_vid_mem(gpu, pfn);
        return;
    }

    if (aperture == UvmMemType_sysmem_coh || aperture == UvmMemType_sysmem_ncoh)
    {
        struct page* pg;
        pg = pfn_to_page(pfn);

        WARN_ON(!pg);
        if (pg)
            __free_page(pg);
    }
}

NvLength uvmfull_encode_gpu_pte(NvU8 *bufferStart, NvU8 *bufferEnd,
                                GMMU_FMT *gmmuFmt, UvmGpuPte *pPte,
                                NvU8 *template, NvU32 entrySize)
{
    GMMU_FIELD_ADDRESS * fldAddr;
    GMMU_FMT_PTE *pteFmt;
    GMMU_APERTURE aperture = GMMU_APERTURE_INVALID;
    NvU8 *buffer = bufferStart;
    if (!gmmuFmt || !buffer || (buffer + entrySize > bufferEnd))
        return 0;

    pteFmt = (GMMU_FMT_PTE*)gmmuFmt->pPte;
    // Start with base template
    memcpy(bufferStart, template, entrySize);

    nvFieldSetBool(&pteFmt->fldReadOnly, !!pPte->readonly, buffer);
    nvFieldSetBool(&pteFmt->fldAtomicDisable, !pPte->atomic, buffer);
    nvFieldSetBool(&pteFmt->fldValid, !!pPte->valid, buffer);
    nvFieldSetBool(&pteFmt->fldVolatile, (!!pPte->valid ? !!pPte->vol : NV_FALSE),
                   buffer);

    aperture = uvm_memtype_to_gmmu_aperture(pPte->aperture);
    if (aperture == GMMU_APERTURE_PEER)
    {
        nvFieldSet32(&pteFmt->fldPeerIndex, pPte->peerid, buffer);
    }

    gmmuFieldSetAperture(&pteFmt->fldAperture, aperture, buffer);
    if (pPte->pfnvalid == 1)
    {
        fldAddr = (pPte->aperture == UvmMemType_vidmem ||
                   pPte->aperture == UvmMemType_peer) ?
                   &pteFmt->fldAddrVidmem :
                   &pteFmt->fldAddrSysmem;
        gmmuFieldSetAddress(fldAddr, PFN_TO_PA(pPte->pfn), buffer);
    }
    return entrySize;
}

// For a multilevel PDE, subLevel = 1 for 4k page. Otherwise always 0.
// Pascal multilevel PDE packs pointers/entries to 2 pagetables.
// Lower 64 bits are for BIG page table.
NvLength uvmfull_encode_gpu_pde(NvU8 *bufferStart, NvU8 *bufferEnd,
                             GMMU_FMT *gmmuFmt,
                             NvU32 level, NvU64 levelLoBit,
                             NvU64 physAddrSub0, UvmMemType apertureSub0,
                             NvU64 physAddrSub1, UvmMemType apertureSub1)
{
    const MMU_FMT_LEVEL *targetFmt;
    const GMMU_FMT_PDE *pdeFmt = NULL;
    NvU8* buffer = bufferStart;
    NvLength entrySize;
    GMMU_APERTURE gmmuAperture = GMMU_APERTURE_INVALID;
    NvU32 i;

    if (!gmmuFmt || !buffer)
        return 0;

    targetFmt = mmuFmtFindLevelWithPageShift(gmmuFmt->pRoot, levelLoBit);
    if (!targetFmt)
        return 0;

    entrySize = targetFmt->entrySize;
    if (buffer + entrySize > bufferEnd)
        return 0;

    memset(buffer, 0x0, entrySize);

    UVM_PANIC_ON(targetFmt->numSubLevels > 2);
    for (i = 0; i < targetFmt->numSubLevels; i++)
    {
        NvU64 physAddr;
        UvmMemType aperture;
        pdeFmt = gmmuFmtGetPde(gmmuFmt, targetFmt, i);
        if (!pdeFmt)
            return 0;
        physAddr = (i == 0) ? physAddrSub0 : physAddrSub1;
        aperture = (i == 0) ? apertureSub0 : apertureSub1;

        // TODO: If PEER aperture then fill the peerid info here.
        gmmuAperture = uvm_memtype_to_gmmu_aperture(aperture);

        gmmuFieldSetAperture(&pdeFmt->fldAperture, gmmuAperture, buffer);
        if (gmmuAperture != GMMU_APERTURE_INVALID)
        {
            gmmuFieldSetAddress(gmmuFmtPdePhysAddrFld(pdeFmt, gmmuAperture),
                                physAddr, buffer);

            // encode the pde as volatile(non-cached) if not vidmem else reset it
            nvFieldSetBool(&pdeFmt->fldVolatile,
                           (gmmuAperture == GMMU_APERTURE_SYS_COH ||
                            gmmuAperture == GMMU_APERTURE_SYS_NONCOH), buffer);
        }
    }
    return entrySize;
}

void uvmfull_update_gpu_pde_entry(UvmGpuVaspace *gpuVaspace, UvmPushbuffer *pb,
            NvU32 parentLevel, NvU64 parentPhyAddr, UvmMemType parentAperture,
            NvU32 entryId, NvU8 parentLoBit,
            NvU64 pointeeAddrSub0, UvmMemType pointeeApertureSub0,
            NvU64 pointeeAddrSub1, UvmMemType pointeeApertureSub1,
            NvBool skipTlbInvalidate)
{
    NvU8 *buffer;
    NvLength bytes;
    NvU64 dstAddr;
    NvU32 dstAperture;
    NvU32 entrySize;
    UvmTlbInvalidateVaParams invalParams = {0};
    UvmPbInlineRegion region = {0};

    // carveout an inline segment
    uvm_pushbuffer_inline_start(pb, &region);
    buffer = (NvU8*)region.regionStart;
    UVM_PANIC_ON(!buffer);

    bytes = uvmfull_encode_gpu_pde(buffer,
                                buffer + UVM_PUSHBUFFER_RESERVATION_SIZE,
                                gpuVaspace->client.pGmmuFmt, parentLevel,
                                parentLoBit,
                                pointeeAddrSub0, pointeeApertureSub0,
                                pointeeAddrSub1, pointeeApertureSub1);
    UVM_PANIC_ON(!bytes);
    // Close the segment. Stage a copy to parent's entry
    region.size = bytes;
    uvm_pushbuffer_inline_end(pb, &region);
    dstAperture = uvm_memtype_to_copy_dst_flags(parentAperture);

    // PD start + PD entry size * entryId
    entrySize = bytes;
    dstAddr = parentPhyAddr + entrySize * entryId;

    // end the segment with a copy request.
    region.copy.bValid = NV_TRUE;
    region.copy.copySize = bytes;
    region.copy.dstAddr = dstAddr;
    region.copy.dstAperture = dstAperture;
    region.copy.dstCopyFlags = NV_UVM_COPY_DST_TYPE_PHYSICAL;

    uvm_pushbuffer_copy_region(pb, &region);

    uvm_push_host_wfi(gpuVaspace->gpuId, pb);
    invalParams.invalidatelevel = UvmTlbInvalidateLevel_all;
    if (!skipTlbInvalidate)
    {
        // TODO: A membar needs to be appended only if a pde downgrade was
        // requested. We need a better way to find that out. Until then always
        // force a sysmembar.
        UvmTlbInvalidateMembarType membar = UvmTlbInvalidateMemBar_sys;
        uvm_push_invalidate(gpuVaspace->gpuId, pb, invalParams,
                            gpuVaspace->pdbAperture, gpuVaspace->gpuPDB >> 12,
                            membar);
    }
}

static void _update_pte_params_from_pte(UvmGpuPteUpdateParams *pParams,
                                        UvmGpuPte *pte)
{
    pParams->updateSwFlags = (UVM_GPU_SW_PTE_UPDATE_FLAGS_SWVALID     |
            UVM_GPU_SW_PTE_UPDATE_FLAGS_PTE_VALID   |
            UVM_GPU_SW_PTE_UPDATE_FLAGS_READ_ONLY   |
            UVM_GPU_SW_PTE_UPDATE_FLAGS_ATOMIC      |
            UVM_GPU_SW_PTE_UPDATE_FLAGS_APERTURE    |
            UVM_GPU_SW_PTE_UPDATE_FLAGS_PEERID      |
            UVM_GPU_SW_PTE_UPDATE_FLAGS_VOLATILE    |
            UVM_GPU_SW_PTE_UPDATE_FLAGS_ZFOD_MAPPED |
            UVM_GPU_SW_PTE_UPDATE_FLAGS_PFN         |
            UVM_GPU_SW_PTE_UPDATE_FLAGS_PFN_VALID);
    pParams->aperture = pte->aperture;
    pParams->valid = pte->valid;
    pParams->readonly = pte->readonly;
    pParams->atomic = pte->atomic;
    pParams->peerid = pte->peerid;
    pParams->vol = pte->vol;
    pParams->zfodMapped = pte->zfodMapped;
    pParams->pfnvalid = pte->pfnvalid;
    pParams->swcachevalid = pte->swcachevalid;
}

NV_STATUS uvmfull_merge_gpupde_pte(UvmVaspace *pVaspace, UvmGpuId gpuid,
        Uvm2MbVaDesc *pUvm2MbVaDesc, UvmGpuPde *pGpuPde, NvU64 addr,
        NvU32 newPgSize, UvmPbOps *pbOps)
{
    NV_STATUS status = NV_OK;
    UvmGpuPteUpdateParams pteParams = { 0 };
    UvmGpuPte     *pPte = NULL;
    UvmGpuPte     tmpPte = {0};
    UvmGpuPte     *tempPte = NULL;
    NvU32 oldPgSize = 0;
    NvU64 invStart;
    NvU64 pfn = 0;
    NvU32 prevPgSize, pgSize = 0;
    NvU32 tmpOffset, tmpSize = 0;

    UVM_CHECK_VALID_PAGE_SIZE(newPgSize, NV_ERR_INVALID_ARGUMENT);

    UVM_PANIC_ON(!pVaspace || !pGpuPde);

    pPte = uvmfull_get_gpupde_current_pte(pGpuPde, addr, &oldPgSize);
    if (!pPte)
        return NV_ERR_GENERIC;

    UVM_CHECK_VALID_PAGE_SIZE(oldPgSize, NV_ERR_GENERIC);

    if (newPgSize <= oldPgSize)
        return NV_OK; // no need to merge

    if (pPte->aperture != UvmMemType_vidmem &&
        pPte->aperture != UvmMemType_peer)
    {
        // We should not be getting a call to split on sysmem
        return NV_ERR_INVALID_ARGUMENT;
    }

    memcpy(&tmpPte, pPte, sizeof(UvmGpuPte));

    if (!uvmfull_check_page_merge(pUvm2MbVaDesc, pGpuPde, addr, pPte,
                         oldPgSize, newPgSize, &pfn))
        return NV_OK; // need to merge

    if (newPgSize != PAGE_SIZE_2M)
    {
        // make sure we have the ptes allocated for the target range within the pde
        status = uvmfull_activate_gpupte(pUvm2MbVaDesc, pGpuPde,
                                         newPgSize == PAGE_SIZE_64K,
                                         NV_FALSE,
                                         gpuid,
                                         uvm_pbops_gpu_pb(pbOps, gpuid));
        if (status != NV_OK)
            return status;
    }
    // now unmap the old page size

    invStart = UVM_ALIGN_ADDR(addr, newPgSize);
    pteParams.updateSwFlags = UVM_GPU_SW_PTE_UPDATE_FLAGS_SWVALID;
    pteParams.updateSwFlags |= UVM_GPU_SW_PTE_UPDATE_FLAGS_PTE_VALID;
    pteParams.updateHwFlags = UVM_GPU_HW_PTE_UPDATE_FLAGS_WRITE_PTE;
    pteParams.valid = 0;
    pteParams.swcachevalid = 0;

    // First unmap all the existing mappings for the large page

    tempPte =uvmfull_get_gpupde_current_pte(pGpuPde, invStart, &prevPgSize);
    tmpOffset = tmpSize = prevPgSize;
    while (tmpSize < newPgSize)
    {
        tempPte = uvmfull_get_gpupde_current_pte(pGpuPde, invStart+tmpOffset,
                                                 &pgSize);
        if (pgSize != prevPgSize)
        {
            status = uvmfull_update_gpupde_ptes(pUvm2MbVaDesc, gpuid, pGpuPde,
                                        invStart, tmpOffset/prevPgSize /*num pages*/,
                                        prevPgSize, &pteParams, NULL, pbOps,
                                        (NvU8*)&pVaspace->gpuVaspaces[gpuid].pteDefaults.entry,
                                        pVaspace->gpuVaspaces[gpuid].pteDefaults.entrySize,
                                        NV_FALSE);
            UVM_PANIC_ON(status != NV_OK);
            invStart += tmpOffset;
            prevPgSize = pgSize;
            tmpOffset = 0;
        }
        tmpOffset += pgSize;
        tmpSize += pgSize;
    }
    status = uvmfull_update_gpupde_ptes(pUvm2MbVaDesc, gpuid, pGpuPde,
                                invStart, tmpOffset/pgSize /*num pages*/,
                                pgSize, &pteParams, NULL, pbOps,
                                (NvU8*)&pVaspace->gpuVaspaces[gpuid].pteDefaults.entry,
                                pVaspace->gpuVaspaces[gpuid].pteDefaults.entrySize,
                                NV_FALSE);
    UVM_PANIC_ON(status != NV_OK);
    invStart = UVM_ALIGN_ADDR(addr, newPgSize);

    _update_pte_params_from_pte(&pteParams, &tmpPte);

    // since this is a merge operation for a single page, we will do a single page
    // update
    pteParams.updateSwFlags |= UVM_GPU_SW_PTE_UPDATE_FLAGS_PFN_VALID;
    pteParams.updateSwFlags |= UVM_GPU_SW_PTE_UPDATE_FLAGS_PFN;
    pteParams.updateSwFlags |= UVM_GPU_SW_PTE_UPDATE_FLAGS_PFN_AUTOINCREMENT;

    status = uvmfull_update_gpupde_ptes(pUvm2MbVaDesc, gpuid, pGpuPde,
            invStart, 1, newPgSize, &pteParams,
            &pfn, pbOps,
            (NvU8*)&pVaspace->gpuVaspaces[gpuid].pteDefaults.entry,
            pVaspace->gpuVaspaces[gpuid].pteDefaults.entrySize,
            NV_FALSE);
    UVM_PANIC_ON(status != NV_OK);

    return status;
}

NvU32 _uvmfull_get_next_next_lower_pg_size(UvmGpuVaspace *pGpuVaspace, NvU32 pgSize)
{
    UVM_CHECK_VALID_PAGE_SIZE(pgSize, NV_ERR_INVALID_ARGUMENT);
    // TODO: We can have better ways of deriving this data
    //       should based it on the gpu
    switch (pgSize)
    {
    case PAGE_SIZE_2M:
        return PAGE_SIZE_64K;
    case PAGE_SIZE_64K:
        if (UVM_CPU_MIN_PAGE_SIZE == PAGE_SIZE_64K)
            return PAGE_SIZE_64K;
        return PAGE_SIZE_4K;
    default:
        return PAGE_SIZE_4K;
    }
}

// This fn would work on 1 higher level page at a time
// For eg. if current/old page size is 64K and new page
// size is 4K, it will invalidate and only re-populate
// that 64K range within a pde
//
NV_STATUS uvmfull_split_gpupde_pte(UvmVaspace   *pVaspace,      UvmGpuId   gpuid,
                                    Uvm2MbVaDesc *pUvm2MbVaDesc, UvmGpuPde  *pGpuPde,
                                    NvU64        addr,           NvU32      newPgSize, 
                                    NvU32        splitFlags,     UvmPbOps   *pbOps)
{
    NV_STATUS status = NV_OK;
    UvmGpuPteUpdateParams pteParams = { 0 };
    UvmGpuVaspace *pGpuVaspace = NULL;
    UvmGpuPte     *pPte = NULL;
    UvmGpuPte     tmpPte={0};
    NvU64 rangeVa[3] = { 0 };
    NvU32 rangePgSize[3] = { 0 };
    NvU32 rangeNumPages[3] = {0};
    NvU64 invStart;
    NvU32 nextLowPgSize = 0;
    NvU64 newPfnBase = 0;
    NvU32 oldPgSize = 0;
    NvU32 i;
    NvU32 numRanges = 0;

    UVM_CHECK_VALID_PAGE_SIZE(newPgSize, NV_ERR_INVALID_ARGUMENT);

    UVM_PANIC_ON(!pVaspace || !pGpuPde);

    pPte = uvmfull_get_gpupde_current_pte(pGpuPde, addr, &oldPgSize);
    UVM_PANIC_ON(!pPte);

    UVM_CHECK_VALID_PAGE_SIZE(oldPgSize, NV_ERR_GENERIC);

    if (newPgSize > oldPgSize)
        return NV_ERR_INVALID_ARGUMENT;

    if (newPgSize == oldPgSize)
        return NV_OK; // no need to split/merge

    // We should not be getting a call to split on sysmem
    UVM_PANIC_ON(pPte->aperture != UvmMemType_vidmem &&
                 pPte->aperture != UvmMemType_peer);

    memcpy(&tmpPte, pPte, sizeof(UvmGpuPte));
    // check if the next lower page size is same as target page size
    nextLowPgSize = _uvmfull_get_next_next_lower_pg_size(pGpuVaspace,
                                                         oldPgSize);


    // Make sure we have ptes allocated for the target range within the pde
    status = uvmfull_activate_gpupte(pUvm2MbVaDesc, pGpuPde,
                                  newPgSize == PAGE_SIZE_64K,
                                  newPgSize == PAGE_SIZE_4K,
                                  gpuid,
                                  uvm_pbops_gpu_pb(pbOps, gpuid));
    if (status != NV_OK)
    {
        return status;
    }

    // Mark this larger gpu page as invalid
    invStart = UVM_ALIGN_ADDR(addr, oldPgSize);
    pteParams.updateSwFlags = UVM_GPU_SW_PTE_UPDATE_FLAGS_SWVALID;
    pteParams.updateSwFlags |= UVM_GPU_SW_PTE_UPDATE_FLAGS_PTE_VALID;
    pteParams.updateSwFlags |= UVM_GPU_SW_PTE_UPDATE_FLAGS_VOLATILE;
    pteParams.updateHwFlags = UVM_GPU_HW_PTE_UPDATE_FLAGS_WRITE_PTE;
    pteParams.valid = 0;
    pteParams.vol = 0;
    pteParams.swcachevalid = 0;

    // First unmap all the existing mappings for the large page
    status = uvmfull_update_gpupde_ptes(pUvm2MbVaDesc, gpuid, pGpuPde,
                            invStart, 1 /*num pages*/, oldPgSize, &pteParams,
                            NULL, pbOps,
                            (NvU8*)&pVaspace->gpuVaspaces[gpuid].pteDefaults.entry,
                            pVaspace->gpuVaspaces[gpuid].pteDefaults.entrySize,
                            NV_FALSE);
    UVM_PANIC_ON(status != NV_OK);

    if ((NV_UVM_SPLIT_MERGE_FLAGS_SPLIT_FIXED_PAGE_SIZE & splitFlags) ||
            nextLowPgSize == newPgSize)
    {
        rangeVa[0] = invStart;
        rangePgSize[0] = newPgSize;
        rangeNumPages[0] = oldPgSize/newPgSize;
        numRanges = 1;
    }
    else
    {
        if (invStart == UVM_ALIGN_ADDR(addr, nextLowPgSize))
        {
            numRanges = 2;
            rangeVa[0]=invStart;
            rangePgSize[0]=newPgSize;
            rangeNumPages[0]=nextLowPgSize/newPgSize;
            rangeVa[1]=invStart+UVM_ALIGN_ADDR(addr, nextLowPgSize);
            rangePgSize[1]=nextLowPgSize;
            rangeNumPages[1]=(oldPgSize/nextLowPgSize) - 1;
        }
        else if ((invStart+oldPgSize) ==
                 (UVM_ALIGN_ADDR(addr, nextLowPgSize)+nextLowPgSize))
        {
            numRanges = 2;
            rangeVa[0]=invStart;
            rangePgSize[0]=nextLowPgSize;
            rangeVa[1]=invStart+UVM_ALIGN_ADDR(addr, nextLowPgSize);
            rangePgSize[1]=newPgSize;
            rangeNumPages[0]=(oldPgSize/nextLowPgSize) - 1;
            rangeNumPages[1]=nextLowPgSize/newPgSize;
        }
        else
        {
            numRanges = 3;
            rangeVa[0]=invStart;
            rangeNumPages[0]=(rangeVa[1]-rangeVa[0])/nextLowPgSize;
            rangePgSize[0]=nextLowPgSize;

            rangeVa[1]=UVM_ALIGN_ADDR(addr, nextLowPgSize);
            rangeNumPages[1]=nextLowPgSize/newPgSize;
            rangePgSize[1]=newPgSize;

            rangeVa[2]=rangeVa[1]+nextLowPgSize;
            rangeNumPages[2] = ((invStart+oldPgSize)-rangeVa[2])/nextLowPgSize;
            rangePgSize[0]=nextLowPgSize;
        }
    }

    for (i = 0; i < numRanges; i++)
    {
        // Update the pte parameters based on the initial pte
        _update_pte_params_from_pte(&pteParams, &tmpPte);

        //
        // since we are breaking up a large page the backing pfn should be
        // contiguous
        //
        pteParams.updateSwFlags |= UVM_GPU_SW_PTE_UPDATE_FLAGS_PFN_VALID;
        pteParams.updateSwFlags |= UVM_GPU_SW_PTE_UPDATE_FLAGS_PFN;
        pteParams.updateSwFlags |= UVM_GPU_SW_PTE_UPDATE_FLAGS_PFN_AUTOINCREMENT;
        newPfnBase = PFN_TO_PA(tmpPte.pfn) + ((rangeVa[i] - invStart)/nextLowPgSize);
        newPfnBase = PA_TO_PFN(newPfnBase);

        status = uvmfull_update_gpupde_ptes(pUvm2MbVaDesc, gpuid, pGpuPde,
                rangeVa[i], rangeNumPages[i], rangePgSize[i], &pteParams,
                &newPfnBase, pbOps,
                (NvU8*)&pVaspace->gpuVaspaces[gpuid].pteDefaults.entry,
                pVaspace->gpuVaspaces[gpuid].pteDefaults.entrySize,
                NV_FALSE);


        UVM_PANIC_ON (status != NV_OK);
    }
    return status;
}

NvBool _is_pte_downgraded(UvmGpuPte *before, UvmGpuPte *after)
{
    NvBool downgrade = (before->valid && !after->valid) ||
                       (!before->readonly && after->readonly) ||
                       (before->atomic && !after->atomic);
    return downgrade;
}

NV_STATUS uvmfull_update_gpupde_ptes(Uvm2MbVaDesc *pUvm2MbVaDesc, UvmGpuId gpuid,
                                     UvmGpuPde *pGpuPde,
                                     NvU64 addr, NvU64 numPages, NvU32 pgSize,
                                     UvmGpuPteUpdateParams *pPteParams,
                                     UvmGpuPfn *pfnList, UvmPbOps *pbOps,
                                     NvU8 *pteTemplate, NvU32 entrySize,
                                     NvBool skipTlbInvalidate)
{
    UvmGpu *gpu;
    UvmGpuVaspace *gpuVaspace;
    UvmVaspace *pVaspace = pUvm2MbVaDesc->pVaspace;
    GMMU_FMT *gmmuFmt;
    NvU8 *buffer = 0;
    NvU32 offset = 0;
    UvmPushbuffer *pb;
    UvmGpuPte* pPte = NULL;
    UvmPteHints *pHint = NULL;
    NvU64 tmpaddr = UVM_ALIGN_ADDR(addr, pgSize);
    NvU32 updateSwFlags = 0;
    NvU32 updateHwFlags = 0;
    NvU32 i = 0;
    NvU64 pfn = 0;
    NV_STATUS status = NV_OK;
    NvU32 pte64mask = 0;
    NvBool bCEPteInlineUpdates;
    NvBool bCEPteUpdates;
    NvLength bytes;
    UvmMemType copyAperture;
    UvmPbInlineRegion region = {0};
    NvBool bDowngrade = NV_FALSE;
    UvmMemType membarAperture = UvmMemType_vidmem;

    gpu = uvmfull_gpuid_to_gpu(gpuid);
    bCEPteUpdates = !gpu->policies.bProviderWritesPtes;
    gpuVaspace = &pVaspace->gpuVaspaces[gpuid];
    bCEPteInlineUpdates = !gpuVaspace->policy.bForcePteMemsets;

    UVM_PANIC_ON(!pPteParams);

    if ((tmpaddr & (~GPU_PDE_MASK)) + (pgSize * numPages) > GPU_PDE_SIZE)
    {
        WARN_ON((tmpaddr & (~GPU_PDE_MASK)) + (pgSize * numPages) > GPU_PDE_SIZE);
        return NV_ERR_INVALID_ARGUMENT;
    }

    updateSwFlags = pPteParams->updateSwFlags;
    updateHwFlags = pPteParams->updateHwFlags;

    // Page size flags will be updated only if the mappings are being changed
    if ((updateSwFlags & UVM_GPU_SW_PTE_UPDATE_FLAGS_PTE_VALID) &&
         pPteParams->valid)
    {
        if (pgSize == PAGE_SIZE_2M)
        {
            pGpuPde->is2MbPte = 1;
            pGpuPde->pageTable.bitmap64KPte = 0;
        }
        else if (pgSize == PAGE_SIZE_64K)
        {
            NvU32 strBit = GET_PDE_PTE_OFFSET(addr, PAGE_SHIFT_64K);
            NvU32 endBit = strBit + numPages;

            pte64mask = ((1ul << endBit) - 1ul) & (~((1ul << strBit) - 1ul));

            pGpuPde->is2MbPte = 0;
            pGpuPde->pageTable.bitmap64KPte |= pte64mask;
        }
    }

    pb = uvm_pbops_gpu_pb(pbOps, gpuid);
    while (i < numPages)
    {
        UvmGpuPte tmpPte = {0};
        pPte = uvmfull_get_gpupde_pte(pGpuPde, pgSize, tmpaddr);
        UVM_PANIC_ON(!pPte); // This should not fail as its a static structure
        memcpy(&tmpPte, pPte, sizeof(tmpPte));

        // NOTE: 'After' the pte is copied to HW, it is reset if swcachevalid is
        // false. See at the end of the loop
        if (updateSwFlags & UVM_GPU_SW_PTE_UPDATE_FLAGS_SWVALID)
            uvmfull_set_pte_swcachevalid(pPte, pPteParams->swcachevalid);

        if (updateSwFlags & UVM_GPU_SW_PTE_UPDATE_FLAGS_ATOMIC)
            uvmfull_set_pte_atomic(pPte, pPteParams->atomic);

        if (updateSwFlags & UVM_GPU_SW_PTE_UPDATE_FLAGS_READ_ONLY)
            uvmfull_set_pte_readonly(pPte, pPteParams->readonly);

        if (updateSwFlags & UVM_GPU_SW_PTE_UPDATE_FLAGS_APERTURE)
            uvmfull_set_pte_aperture(pPte, pPteParams->aperture);

        if (updateSwFlags & UVM_GPU_SW_PTE_UPDATE_FLAGS_PEERID)
            uvmfull_set_pte_peerid(pPte, pPteParams->peerid);

        if (updateSwFlags & UVM_GPU_SW_PTE_UPDATE_FLAGS_VOLATILE)
            uvmfull_set_pte_vol(pPte, pPteParams->vol);

        if (updateSwFlags & UVM_GPU_SW_PTE_UPDATE_FLAGS_PTE_VALID)
            uvmfull_set_pte_valid(pPte, pPteParams->valid);

        if (updateSwFlags & UVM_GPU_SW_PTE_UPDATE_FLAGS_ZFOD_MAPPED)
            uvmfull_set_pte_zfod_mapped(pPte, pPteParams->zfodMapped);

        if (updateSwFlags & UVM_GPU_SW_PTE_UPDATE_FLAGS_HINTS)
            uvmfull_set_pte_hints(pPte, pPteParams->hints);

        if (updateSwFlags & UVM_GPU_SW_PTE_UPDATE_FLAGS_PFN)
        {
            if (updateSwFlags & UVM_GPU_SW_PTE_UPDATE_FLAGS_IMPLICIT_PFN)
            {
                WARN_ON(pPteParams->aperture == UvmMemType_peer);
                if (pPteParams->aperture == UvmMemType_sysmem_coh ||
                    pPteParams->aperture == UvmMemType_sysmem_ncoh) // sysmem
                {
                    UVM_PANIC_ON(pgSize != PAGE_SIZE_4K);
                    pHint = GET_VA2M_HINT(pUvm2MbVaDesc, tmpaddr, pgSize);
                    UVM_PANIC_ON(!pHint->cpuPfnValid);
                    pfn = pHint->cpuPfn;
                }
                else
                {
                    // Implicit in vidmem means auto-increment using the
                    // starting pfn cached in the gpupde
                    // Change the pfn based on the offset within the 2MB region
                    pfn = pGpuPde->gpuFirstPfn +
                            (GET_PDE_PTE_OFFSET(tmpaddr, PFN_SHIFT));
                }
            }
            else if (updateSwFlags &
                                  UVM_GPU_SW_PTE_UPDATE_FLAGS_PFN_AUTOINCREMENT)
            {
                // Here the auto increment works on the user provided pfn
                UVM_PANIC_ON(pfnList == NULL);
                pfn = pfnList[0] + ((pgSize/PFN_TO_PA(1)) * i);
            }
            else
            {
                UVM_PANIC_ON(pfnList == NULL);
                pfn = pfnList[i];
            }
            uvmfull_set_pte_pfn(pPte, pfn);
        }

        if (updateSwFlags & UVM_GPU_SW_PTE_UPDATE_FLAGS_PFN_VALID)
            uvmfull_set_pte_pfnvalid(pPte, pPteParams->pfnvalid);

        WARN_ON(!bCEPteUpdates);

        // Compare the before and after state to see if a downgrade was requested.
        bDowngrade = bDowngrade || _is_pte_downgraded(&tmpPte, pPte);

        // Also update the aperture to be used for membar type if not vidmem.
        // Check each pte that we update.
        if (bDowngrade && (tmpPte.aperture != UvmMemType_vidmem))
            membarAperture = tmpPte.aperture;

        // Fill the buffer with pte data.
        if (bCEPteUpdates)
        {
            if (bCEPteInlineUpdates)
            {
                // For inline copy, keep pushing PTE to the buffer.
                // We will initiate a single copy at the end.
                if (!buffer)    // carveout an inline pb segment
                {
                    gmmuFmt = gpuVaspace->client.pGmmuFmt;
                    if (!gmmuFmt)
                        return NV_ERR_INVALID_STATE;
                    uvm_pushbuffer_inline_start(pb, &region);
                    buffer = (NvU8*)region.regionStart;
                    if (!buffer)
                        return NV_ERR_INVALID_STATE;
                }
                bytes = uvmfull_encode_gpu_pte(buffer + offset,
                                buffer + UVM_PUSHBUFFER_INLINE_REGION_SIZE_MAX,
                                gpuVaspace->client.pGmmuFmt, pPte,
                                pteTemplate,
                                entrySize);
                UVM_PANIC_ON(!bytes);
                offset += bytes;
            }
            else
            {
                status = uvm_copy_pte_using_memset(gpuVaspace,
                                                   pGpuPde,
                                                   pPte,
                                                   tmpaddr,
                                                   pgSize,
                                                   pb);
                if (NV_OK != status)
                    return status;
            }
        }
        else // use RM to update ptes. (or maybe use the registers directly)
        {
            // This is useful for Tegra
            UVM_PANIC();
            /*
            // This code should eventually go away.
            // flush any updates to hardware
            status = uvmfull_write_gpu_pte(pVaspace, gpuid, pPte,
                                        tmpaddr, pgSize, updateHwFlags);
            UVM_PANIC_ON(status != NV_OK);
            */
        }

        // Clear the masks and the pte if sw cache is being invalidated
        if ((updateSwFlags & UVM_GPU_SW_PTE_UPDATE_FLAGS_SWVALID) &&
             pPte->swcachevalid == 0x0)
        {
            if (pgSize == PAGE_SIZE_2M)
            {
                pGpuPde->is2MbPte = 0;
                pGpuPde->pageTable.bitmap64KPte = 0;
            }
            else if (pgSize == PAGE_SIZE_64K)
            {
                NvU32 strBit = GET_PDE_PTE_OFFSET(addr, PAGE_SHIFT_64K);
                NvU32 endBit = strBit + numPages;

                pte64mask = ((1ul << endBit) - 1ul) & (~((1ul << strBit) - 1ul));

                pGpuPde->is2MbPte = 0;
                pGpuPde->pageTable.bitmap64KPte &= ~pte64mask;
            }
            memset(pPte, 0x0, sizeof(*pPte));
        }

        i++;
        tmpaddr += pgSize;
    }

    // Now that all ptes have been encoded into the buffer; lets queue inline copy.
    if (bCEPteUpdates && offset && bCEPteInlineUpdates)
    {
        NvU64 dstAddr;
        NvU32 index;
        NvU32 dstAperture;

        // Close the segment. We will stage a copy from it later.
        region.size = offset;
        uvm_pushbuffer_inline_end(pb, &region);

        if (pgSize == PAGE_SIZE_4K)
        {
            dstAddr =
            pGpuPde->pageTable.pageTablePhysAddr[UVM_GPU_PT_INDEX_SMALL].addr;
            copyAperture =
            pGpuPde->pageTable.pageTablePhysAddr[UVM_GPU_PT_INDEX_SMALL].aperture;
            index = GET_PDE_PTE_OFFSET(addr, PAGE_SHIFT_4K);
        }
        else if (pgSize == PAGE_SIZE_64K)
        {
            dstAddr =
            pGpuPde->pageTable.pageTablePhysAddr[UVM_GPU_PT_INDEX_BIG].addr;
            copyAperture =
            pGpuPde->pageTable.pageTablePhysAddr[UVM_GPU_PT_INDEX_BIG].aperture;
            index = GET_PDE_PTE_OFFSET(addr, PAGE_SHIFT_64K);
        }
        else if (pgSize == PAGE_SIZE_2M)
        {
            dstAddr  = pGpuPde->parentAddr;
            copyAperture = pGpuPde->parentAperture;
            index    = pGpuPde->parentEntryId;
        }
        else
            UVM_PANIC_MSG("Page size not supported yet!");

        dstAperture = uvm_memtype_to_copy_dst_flags(copyAperture);

        if (pgSize != PAGE_SIZE_2M)
            dstAddr += index * gpuVaspace->pteDefaults.entrySize;
        else
            dstAddr += index * gpuVaspace->pdeDefaults.pde0EntrySize;

        // end the segment with a copy request.
        region.copy.bValid = NV_TRUE;
        region.copy.copySize = offset;
        region.copy.dstAddr = dstAddr;
        region.copy.dstAperture = dstAperture;
        region.copy.dstCopyFlags = NV_UVM_COPY_DST_TYPE_PHYSICAL;

        // push a copy from the staging segment
        uvm_pushbuffer_copy_region(pb, &region);
    }

    uvm_push_host_wfi(gpuid, pb);
    // Push invalidate
    if (!skipTlbInvalidate)
    {
        UvmTlbInvalidateMembarType membar = UvmTlbInvalidateMemBar_none;
        UvmTlbInvalidateVaParams invalidateParams = {0};
        invalidateParams.invalidatelevel = UvmTlbInvalidateLevel_all;

        if (bDowngrade)
            membar = (membarAperture == UvmMemType_vidmem) ?
                      UvmTlbInvalidateMemBar_local :
                      UvmTlbInvalidateMemBar_sys;

        uvm_push_invalidate(gpuVaspace->gpuId, pb, invalidateParams,
                            gpuVaspace->pdbAperture, gpuVaspace->gpuPDB >> 12,
                            membar);
    }
    return status;
}

UvmPbOps* uvm_pbops_start(UvmVaspace *vaspace)
{
    UvmGlobalState *dev;
    UvmGpu *gpu;
    UvmPbOps *pbOps;
    NvU32 i, j;
    NV_STATUS status;

    dev = uvmfull_get_global();
    pbOps = kmem_cache_zalloc(g_uvmPbOpsCache, NV_UVM_GFP_FLAGS);
    if (!pbOps)
        return NULL;

    pbOps->nGpus = vaspace->nActiveGpus;
    if (pbOps->nGpus)
    {
        pbOps->gpuPb = vmalloc(pbOps->nGpus * sizeof(UvmGpuPb));
        if (!pbOps->gpuPb)
            goto failure;

        memset(pbOps->gpuPb, 0x0, (pbOps->nGpus * sizeof(UvmGpuPb)));
    }

    for (i = 0, j=0; j < pbOps->nGpus; i++)
    {
        if (uvmfull_gpu_vaspace_active(vaspace, i))
        {
            gpu = &dev->uvmGpus[i];
            pbOps->gpuPb[j].gpuId = i;
            status = uvm_get_pushbuffer(gpu->channelManager,
                                        &pbOps->gpuPb[j].pb);
            if (status != NV_OK)
                goto failure;
            j++;
        }
    }
    // Now allocate the temp tracker.
    pbOps->t = uvm_allocate_tracker();
    if (!pbOps->t)
        goto failure;

    status = uvm_grow_tracker(pbOps->t, pbOps->nGpus);
    if (status != NV_OK)
        goto failure;

    return pbOps;
failure:
    uvm_pbops_end(pbOps);
    return NULL;
}

UvmPbOps* uvm_pbops_start_on_gpu(UvmGpuId gpuId)
{
    UvmGlobalState *dev;
    UvmGpu *gpu;
    UvmPbOps *pbOps;
    NV_STATUS status;

    dev = uvmfull_get_global();
    gpu = &dev->uvmGpus[gpuId];
    UVM_PANIC_ON(!gpu->channelManager);

    pbOps = kmem_cache_zalloc(g_uvmPbOpsCache, NV_UVM_GFP_FLAGS);
    if (!pbOps)
        return NULL;

    pbOps->nGpus = 1;
    pbOps->gpuPb = vmalloc(pbOps->nGpus * sizeof(UvmGpuPb));
    if (!pbOps->gpuPb)
        goto failure;

    memset(pbOps->gpuPb, 0x0, (pbOps->nGpus * sizeof(UvmGpuPb)));
    pbOps->gpuPb[0].gpuId = gpuId;

    status = uvm_get_pushbuffer(gpu->channelManager, &pbOps->gpuPb->pb);
    if (status != NV_OK)
        goto failure;

    // Now allocate the temp tracker.
    pbOps->t = uvm_allocate_tracker();
    if (!pbOps->t)
        goto failure;

    status = uvm_grow_tracker(pbOps->t, pbOps->nGpus);
    if (status != NV_OK)
        goto failure;

    return pbOps;
failure:
    uvm_pbops_end(pbOps);
    return NULL;
}

void uvm_pbops_end(UvmPbOps *pbOps)
{
    NvU32 j;
    UvmGlobalState *dev;
    UvmGpu *gpu;

    if (pbOps->t)
        uvm_free_tracker(pbOps->t);

    dev = uvmfull_get_global();
    for (j=0; j < pbOps->nGpus; j++)
    {
        if (pbOps->gpuPb && pbOps->gpuPb[j].pb &&
            (pbOps->gpuPb[j].bSubmitted == NV_FALSE))
        {
            UvmGpuId id = pbOps->gpuPb[j].gpuId;
            UvmPushbuffer *pb = pbOps->gpuPb[j].pb;
            gpu = &dev->uvmGpus[id];
            uvm_cancel_pushbuffer(gpu->channelManager, pb);
        }
    }

    if (pbOps->gpuPb)
        vfree(pbOps->gpuPb);

    kmem_cache_free(g_uvmPbOpsCache, pbOps);
}

NV_STATUS uvm_pbops_submit(UvmPbOps *pbOps, UvmGpuId id,
                            UvmPushbuffer *pb, UvmTracker *tracker)
{
    NvU32 j;
    UvmGlobalState *dev;
    UvmGpu *gpu;
    UvmTrackerItem item;
    NV_STATUS status = NV_ERR_INVALID_ARGUMENT;

    dev = uvmfull_get_global();
    for (j=0; j<pbOps->nGpus; j++)
    {
        if (pbOps->gpuPb[j].gpuId == id)
        {
            pbOps->gpuPb[j].bSubmitted = NV_TRUE;
            UVM_PANIC_ON(pbOps->gpuPb[j].pb != pb);
            gpu = &dev->uvmGpus[id];
            status = uvm_submit_pushbuffer(gpu->channelManager,
                                           pb, tracker, &item);
            // Merge the item into the temporary tracker
            uvm_merge_tracker_item(pbOps->t, &item);
            return status;
        }
    }

    UVM_PANIC_MSG("Invalid PbOps\n");
    return status;
}

NV_STATUS uvm_pbops_submit_and_release(UvmPbOps *pbOps, NvU64 gpuMask,
                                       UvmTracker *tracker)
{
    NV_STATUS status = NV_OK;
    NvU64 gpuId;
    UvmPushbuffer *pushbuffer = NULL;
    NvU64 gpuSubmitMask = gpuMask;

    while (gpuSubmitMask)
    {
        gpuId = find_first_bit((unsigned long *)&gpuSubmitMask,
                                UVM_FULL_MAX_GPUS);

        // TO-DO All this could be far more optimized if we
        // maintain a mask of to-be submit PB in OPS and also
        // while opening PbOps maintain a array of all PBs/GPU.
        // Currently the lookup is O(n) and could be changed to
        // O(1)
        pushbuffer = uvm_pbops_gpu_pb(pbOps, gpuId);
        status = uvm_pbops_submit(pbOps, gpuId, pushbuffer, tracker);
        if (status != NV_OK)
            break;
        gpuSubmitMask &= ~NVBIT64(gpuId);
    }
    uvm_move_tracker(tracker, pbOps->t);
    uvm_pbops_end(pbOps);
    return status;
}

void uvm_pbops_reserve_acquire_space(UvmPbOps *pbOps, UvmTracker *tracker)
{
    size_t j;
    UvmPushbuffer *pb;
    UvmGlobalState *dev;
    UvmGpu *gpu;

    if (tracker->nUsedItems == 0)
        return;

    dev = uvmfull_get_global();
    for (j=0; j < pbOps->nGpus; j++)
    {
        gpu = &dev->uvmGpus[pbOps->gpuPb[j].gpuId];
        if (pb = pbOps->gpuPb[j].pb, pb != NULL)
            uvm_reserve_acquire_space(gpu->channelManager, pb, tracker);
    }
}

NV_STATUS
uvm_copy_pte_using_memset(UvmGpuVaspace *gpuVaspace,
                        UvmGpuPde *pGpuPde,
                        UvmGpuPte* pPte,
                        NvU64 addr,
                        NvU32 pgSize,
                        UvmPushbuffer *pb)
{
    NV_STATUS nvStatus;
    NvU8 tempBuffer[UVM_MEMSET_BUFFER_SIZE_MAX];
    NvLength bytes;
    NvU64 dstAddr;
    NvU32 index;
    UvmMemType aperture;
    NvU32 dstAperture;

    memset(tempBuffer, 0x0, UVM_MEMSET_BUFFER_SIZE_MAX);

    bytes = uvmfull_encode_gpu_pte(tempBuffer,
                    tempBuffer + UVM_MEMSET_BUFFER_SIZE_MAX,
                    gpuVaspace->client.pGmmuFmt, pPte,
                    (NvU8*)gpuVaspace->pteDefaults.entry,
                    gpuVaspace->pteDefaults.entrySize);
    UVM_PANIC_ON(!bytes);

    if (pgSize == PAGE_SIZE_4K)
    {
        dstAddr =
        (NvU64)pGpuPde->pageTable.pageTablePhysAddr[UVM_GPU_PT_INDEX_SMALL].addr;
        aperture =
        pGpuPde->pageTable.pageTablePhysAddr[UVM_GPU_PT_INDEX_SMALL].aperture;
        index = GET_PDE_PTE_OFFSET(addr, PAGE_SHIFT_4K);
    }
    else if (pgSize == PAGE_SIZE_64K)
    {
        dstAddr =
        (NvU64)pGpuPde->pageTable.pageTablePhysAddr[UVM_GPU_PT_INDEX_BIG].addr;
        aperture =
        pGpuPde->pageTable.pageTablePhysAddr[UVM_GPU_PT_INDEX_BIG].aperture;
        index = GET_PDE_PTE_OFFSET(addr, PAGE_SHIFT_64K);
    }
    else if (pgSize == PAGE_SIZE_2M)
    {
        dstAddr  = pGpuPde->parentAddr;
        aperture = pGpuPde->parentAperture;
        index    = pGpuPde->parentEntryId;
    }
    else
        UVM_PANIC_MSG("Page size not supported yet!");

    dstAperture = uvm_memtype_to_memset_dst_flags(aperture);

    if (pgSize != PAGE_SIZE_2M)
        dstAddr += index * gpuVaspace->pteDefaults.entrySize;
    else
        dstAddr += index * gpuVaspace->pdeDefaults.pde0EntrySize;

    nvStatus = uvm_push_memset(pb,
                  tempBuffer,
                  bytes,
                  dstAddr,
                  dstAperture,
                  NV_UVM_MEMSET_DST_TYPE_PHYSICAL |
                  NV_UVM_MEMSET_TRANSER_PIPELINED);

    return nvStatus;
}

void uvm_push_gpu_to_cpu_copy(UvmPushbuffer *pushbuffer, NvU64 src, NvU64 dst,
                              NvU64 size, NvU32 flags)
{
    NvU64 numBytes;
    numBytes = pushbuffer->channel->ceOps.launchDma(
                                        (unsigned **)&pushbuffer->pbOffset,
                                        (unsigned*)(pushbuffer->cpuBegin +
                                        UVM_PUSHBUFFER_RESERVATION_SIZE),
                                        src,
                                        NV_UVM_COPY_SRC_LOCATION_FB,
                                        dst,
                                        NV_UVM_COPY_DST_LOCATION_SYSMEM,
                                        size,
                                        flags
                                        /*NV_UVM_COPY_SRC_TYPE_PHYSICAL |
                                          NV_UVM_COPY_DST_TYPE_PHYSICAL*/);
    UVM_PANIC_ON(!numBytes);
    pushbuffer->curOffset += numBytes;
}

void uvm_push_cpu_to_gpu_copy(UvmPushbuffer *pushbuffer, NvU64 src, NvU64 dst,
                              NvU64 size, NvU32 flags)
{
    NvU64 numBytes;
    numBytes = pushbuffer->channel->ceOps.launchDma(
                                        (unsigned **)&pushbuffer->pbOffset,
                                        (unsigned*)(pushbuffer->cpuBegin +
                                        UVM_PUSHBUFFER_RESERVATION_SIZE),
                                        src,
                                        NV_UVM_COPY_SRC_LOCATION_SYSMEM,
                                        dst,
                                        NV_UVM_COPY_DST_LOCATION_FB,
                                        size,
                                        flags
                                        /*NV_UVM_COPY_SRC_TYPE_PHYSICAL |
                                        NV_UVM_COPY_DST_TYPE_PHYSICAL*/);
    UVM_PANIC_ON(!numBytes);
    pushbuffer->curOffset += numBytes;
}

void uvm_push_copy(UvmPushbuffer *pushbuffer, NvU64 src, NvU64 dst,
                              NvU64 size, NvU32 flags, NvU32 srcLoc,
                              NvU32 dstLoc)
{
    NvU64 numBytes;

    numBytes = pushbuffer->channel->ceOps.launchDma(
                                        (unsigned **)&pushbuffer->pbOffset,
                                        (unsigned*)(pushbuffer->cpuBegin +
                                        UVM_PUSHBUFFER_RESERVATION_SIZE),
                                        src,
                                        srcLoc,
                                        dst,
                                        dstLoc,
                                        size,
                                        flags
                                        /*NV_UVM_COPY_SRC_TYPE_PHYSICAL |
                                        NV_UVM_COPY_DST_TYPE_PHYSICAL*/);
    UVM_PANIC_ON(!numBytes);
    pushbuffer->curOffset += numBytes;
}

NV_STATUS uvm_push_memset(UvmPushbuffer *pb,
                                      NvU8 *srcStart,
                                      NvU64 copySize,
                                      NvU64 dstAdr,
                                      NvU32 dstAperture,
                                      NvU32 dstCopyFlags)
{
    size_t numBytes = 0;
    NvU32 i;
    NvU32 payload = 0;
    NvU32 dwordSize = 4;
    size_t dwordCnt = NV_CEIL(copySize, dwordSize);
    NvU32 dstFlags = dstAperture | dstCopyFlags;
    NvU32 dstFlagsNonPipelined = dstFlags & (~NV_UVM_MEMSET_TRANSER_PIPELINED);
    NvU32 flags;

    if (pb->bRegionPending || !copySize)
        return NV_ERR_GENERIC;

    // Memset allows 4 bytes payload only. So to copy 'copySize' bytes PTE,
    // we need to perform several copy operations of 4 bytes each.
    for (i = 0; i < dwordCnt; i++)
    {
        if (0 == i)
        {
            // Ensure that the first memset is not pipelined. 
            flags = dstFlagsNonPipelined;
        }
        else
        {
            flags = dstFlags;
        }

        payload = *(NvU32 *)(srcStart + i * dwordSize);
        numBytes = pb->channel->ceOps.memset(
                                       (unsigned **)&pb->pbOffset,
                                       (unsigned*)(pb->cpuBegin +
                                            UVM_PUSHBUFFER_RESERVATION_SIZE),
                                        dstAdr,
                                        dwordSize,
                                        payload,
                                        flags);
        UVM_PANIC_ON(!numBytes);
        pb->curOffset += numBytes;
        dstAdr += dwordSize;
    }

    return NV_OK;
}

void uvm_trigger_replay_priv(UvmGpuId gpuId,
                        UvmFaultBufferEntry *pFault,
                        UvmTlbInvalidateReplayType replayType,
                        UvmMemType aperture, NvU32 flags)
{
    UvmGlobalState *dev;
    UvmGpu *pGpu;
    dev = uvmfull_get_global();
    pGpu = &dev->uvmGpus[gpuId];

    pGpu->faultBufferInfo.faultOps.setReplayParamsReg(
                                    pGpu->faultBufferInfo.pReplay,
                                    pFault->uvmFaultGpcId,
                                    pFault->uvmFaultClientId,
                                    pFault->uvmFaultMmuClientType,
                                    replayType,
                                    aperture != UvmMemType_vidmem, //sysmembar?
                                    0);  // no additional flags
}

void uvm_push_memops_raw(UvmGpuId gpuId,
                         UvmPushbuffer *pushbuffer,
                         NvU32 gpcId,
                         NvU32 clientId,
                         UvmTlbInvalidateReplayType replayType,
                         UvmTlbInvalidateAckType ackType,
                         UvmTlbInvalidateVaParams invalidateParams,
                         UvmTlbInvalidateMembarType membar,
                         UvmMemType aperture,
                         NvBool disableGpcInvalidate,
                         NvU64 targetPdb)
{
    NvU64 numBytes;
    UvmTlbInvalidateMemOpsParams memopsParams = {0};
    UvmTlbInvalidatePdbAperture pdbAperture;

    if (aperture == UvmMemType_vidmem)
        pdbAperture = UvmTlbInvalidatePdbAperture_vidmem;
    else if (aperture == UvmMemType_sysmem_coh)
        pdbAperture = UvmTlbInvalidatePdbAperture_sysmemCoh;
    else if (aperture == UvmMemType_sysmem_ncoh)
        pdbAperture = UvmTlbInvalidatePdbAperture_sysmemNcoh;
    else
        UVM_PANIC_ON(aperture || "Aperture not supported");

    memopsParams.clientId = clientId;
    memopsParams.gpcId = gpcId;
    memopsParams.replayType = replayType;
    memopsParams.tlbAckType = ackType;
    memopsParams.invalidateParams = invalidateParams;
    memopsParams.membarType = membar;
    memopsParams.disableGpcInvalidate = disableGpcInvalidate;

    numBytes = pushbuffer->channel->memOps.tlbInvalidate(
                                (unsigned **)&pushbuffer->pbOffset,
                                (unsigned*)(pushbuffer->cpuBegin +
                                 UVM_PUSHBUFFER_RESERVATION_SIZE),
                                 targetPdb,
                                 pdbAperture,
                                 &memopsParams);
    UVM_PANIC_ON(!numBytes);
    pushbuffer->curOffset += numBytes;
}

void uvm_push_invalidate(UvmGpuId gpuId,
                         UvmPushbuffer *pushbuffer,
                         UvmTlbInvalidateVaParams invalidateParams,
                         UvmMemType aperture, NvU64 targetPdb,
                         UvmTlbInvalidateMembarType membarType)
{
    NvU64 numBytes;
    UvmTlbInvalidateMemOpsParams memOpsParams = {0};
    UvmTlbInvalidatePdbAperture pdbAperture;

    if (aperture == UvmMemType_vidmem)
        pdbAperture = UvmTlbInvalidatePdbAperture_vidmem;
    else if (aperture == UvmMemType_sysmem_coh)
        pdbAperture = UvmTlbInvalidatePdbAperture_sysmemCoh;
    else if (aperture == UvmMemType_sysmem_ncoh)
        pdbAperture = UvmTlbInvalidatePdbAperture_sysmemNcoh;
    else
        UVM_PANIC_ON(aperture || "Aperture not supported");

    memOpsParams.replayType = UvmTlbInvalidateReplayType_none;

    if (membarType != UvmTlbInvalidateMemBar_none)
    {
        memOpsParams.tlbAckType = UvmTlbInvalidateAckType_globally;
        memOpsParams.membarType = membarType;
    }
    else
    {
        memOpsParams.tlbAckType = UvmTlbInvalidateAckType_none;
        memOpsParams.membarType = UvmTlbInvalidateMemBar_none;
    }

    memOpsParams.invalidateParams = invalidateParams;
    numBytes = pushbuffer->channel->memOps.tlbInvalidate(
                                (unsigned **)&pushbuffer->pbOffset,
                                (unsigned*)(pushbuffer->cpuBegin +
                                 UVM_PUSHBUFFER_RESERVATION_SIZE),
                                 targetPdb,
                                 pdbAperture,
                                 &memOpsParams);
    UVM_PANIC_ON(!numBytes);
    pushbuffer->curOffset += numBytes;
}

void uvm_push_cancel(UvmGpuId gpuId,
                     UvmPushbuffer *pushbuffer,
                     NvU32 gpcId,
                     NvU32 clientId,
                     NvBool bGlobalCancel,
                     UvmTlbInvalidateVaParams invalidateParams,
                     UvmMemType aperture, NvU64 targetPdb)
{
    NvU64 numBytes;
    UvmTlbInvalidateMemOpsParams memOpsParams = {0};
    UvmTlbInvalidatePdbAperture pdbAperture;

    if (aperture == UvmMemType_vidmem)
        pdbAperture = UvmTlbInvalidatePdbAperture_vidmem;
    else if (aperture == UvmMemType_sysmem_coh)
        pdbAperture = UvmTlbInvalidatePdbAperture_sysmemCoh;
    else if (aperture == UvmMemType_sysmem_ncoh)
        pdbAperture = UvmTlbInvalidatePdbAperture_sysmemNcoh;
    else
        UVM_PANIC_ON(aperture || "Aperture not supported");

    if (bGlobalCancel)
    {
        memOpsParams.replayType = UvmTlbInvalidateReplayType_cancel_global;
        WARN_ON(gpcId || clientId);
    }
    else
    {
        memOpsParams.replayType = UvmTlbInvalidateReplayType_cancel_targeted;
        memOpsParams.clientId = clientId;
        memOpsParams.gpcId = gpcId;
    }

    memOpsParams.membarType = UvmTlbInvalidateMemBar_none;
    memOpsParams.tlbAckType = UvmTlbInvalidateAckType_none;
    memOpsParams.invalidateParams = invalidateParams;

    numBytes = pushbuffer->channel->memOps.tlbInvalidate(
                                (unsigned **)&pushbuffer->pbOffset,
                                (unsigned*)(pushbuffer->cpuBegin +
                                 UVM_PUSHBUFFER_RESERVATION_SIZE),
                                 targetPdb,
                                 pdbAperture,
                                 &memOpsParams);
    UVM_PANIC_ON(!numBytes);
    pushbuffer->curOffset += numBytes;
}

void uvm_push_replay(UvmGpuId gpuId,
                     UvmPushbuffer *pushbuffer,
                     NvBool bReplayStartAckAll,
                     UvmTlbInvalidateVaParams invalidateParams,
                     UvmMemType aperture, NvU64 targetPdb)
{
    NvU64 numBytes;
    UvmTlbInvalidateMemOpsParams memOpsParams = {0};
    UvmTlbInvalidatePdbAperture pdbAperture;

    if (aperture == UvmMemType_vidmem)
        pdbAperture = UvmTlbInvalidatePdbAperture_vidmem;
    else if (aperture == UvmMemType_sysmem_coh)
        pdbAperture = UvmTlbInvalidatePdbAperture_sysmemCoh;
    else if (aperture == UvmMemType_sysmem_ncoh)
        pdbAperture = UvmTlbInvalidatePdbAperture_sysmemNcoh;
    else
        UVM_PANIC_ON(aperture || "Aperture not supported");

    if (bReplayStartAckAll)
        memOpsParams.replayType = UvmTlbInvalidateReplayType_start_ack_all;
    else
        memOpsParams.replayType = UvmTlbInvalidateReplayType_start;

    memOpsParams.membarType = UvmTlbInvalidateMemBar_none;
    memOpsParams.tlbAckType = UvmTlbInvalidateAckType_none;
    memOpsParams.invalidateParams = invalidateParams;

    numBytes = pushbuffer->channel->memOps.tlbInvalidate(
                                (unsigned **)&pushbuffer->pbOffset,
                                (unsigned*)(pushbuffer->cpuBegin +
                                 UVM_PUSHBUFFER_RESERVATION_SIZE),
                                 targetPdb,
                                 pdbAperture,
                                 &memOpsParams);
    UVM_PANIC_ON(!numBytes);
    pushbuffer->curOffset += numBytes;
}

void uvm_push_membar(UvmGpuId gpuId,
                     UvmPushbuffer *pushbuffer,
                     UvmTlbInvalidateMembarType membarType)
{
    NvU64 numBytes;
    numBytes = pushbuffer->channel->memOps.membar(
                                (unsigned **)&pushbuffer->pbOffset,
                                (unsigned*)(pushbuffer->cpuBegin +
                                 UVM_PUSHBUFFER_RESERVATION_SIZE),
                                 membarType);
    UVM_PANIC_ON(!numBytes);
    pushbuffer->curOffset += numBytes;
}

void uvm_push_host_wfi(UvmGpuId gpuId,
                     UvmPushbuffer *pushbuffer)
{
    NvU64 numBytes;
    numBytes = pushbuffer->channel->memOps.hostwfi(
                                (unsigned **)&pushbuffer->pbOffset,
                                (unsigned*)(pushbuffer->cpuBegin +
                                 UVM_PUSHBUFFER_RESERVATION_SIZE));
    UVM_PANIC_ON(!numBytes);
    pushbuffer->curOffset += numBytes;
}

void uvm_push_memzero_vidmem(UvmGpuId gpuId, UvmPushbuffer *pb,
                        NvU64 pagePa, NvU64 size)
{
    UvmGpu *gpu = uvmfull_gpuid_to_gpu(gpuId);
    UVM_PANIC_ON(gpu->bZeroFbGpu);
    UVM_PANIC_ON(!gpu->zeroPageAllocated);
    UVM_PANIC_ON(size > PAGE_SIZE_2M);

    // TODO: Launch copy based on gpu type
    uvm_push_copy(pb, (NvU64)gpu->zeroed2MPa, pagePa, size,
                  NV_UVM_COPY_SRC_TYPE_PHYSICAL | NV_UVM_COPY_DST_TYPE_PHYSICAL,
                  UvmMemType_vidmem, UvmMemType_vidmem);
}

GMMU_APERTURE uvm_memtype_to_gmmu_aperture(UvmMemType aperture)
{
    if (aperture == UvmMemType_vidmem)
    {
        return GMMU_APERTURE_VIDEO;
    }
    else if (aperture == UvmMemType_sysmem_coh)
    {
        return GMMU_APERTURE_SYS_COH;
    }
    else if (aperture == UvmMemType_sysmem_ncoh)
    {
        return GMMU_APERTURE_SYS_NONCOH;
    }
    else if (aperture == UvmMemType_peer)
    {
        return GMMU_APERTURE_PEER;
    }
    return GMMU_APERTURE_INVALID;
}

NvU32 uvm_memtype_to_memset_dst_flags(UvmMemType aperture)
{
    if (aperture == UvmMemType_vidmem)
    {
        return NV_UVM_MEMSET_DST_LOCATION_FB;
    }
    else if (aperture == UvmMemType_sysmem_coh ||
             aperture == UvmMemType_sysmem_ncoh)
    {
        return NV_UVM_MEMSET_DST_LOCATION_SYSMEM;
    }
    else
    {
        UVM_PANIC_MSG("Invalid aperture!!\n");
        return NV_UVM_MEMSET_DST_LOCATION_FB;
    }
}

NvU32 uvm_memtype_to_copy_dst_flags(UvmMemType aperture)
{
    if (aperture == UvmMemType_vidmem)
    {
        return NV_UVM_COPY_DST_LOCATION_FB;
    }
    else if (aperture == UvmMemType_sysmem_coh ||
            aperture == UvmMemType_sysmem_ncoh)
    {
        return NV_UVM_COPY_DST_LOCATION_SYSMEM;
    }
    else
    {
        UVM_PANIC_MSG("Invalid aperture!!\n");
        return NV_UVM_COPY_DST_LOCATION_FB;
    }
}

static void uvmfull_update_bitmask(NvU64 offset, NvU64 *bitmask, NvBool *isBitmaskSet)
{
    // TODO: update to cpu page size once power PC support is added
    NvU64 bitNumber = offset / PAGE_SIZE_4K;
    bitmask[bitNumber/64] |= NVBIT64(bitNumber % 64);
    *isBitmaskSet = NV_TRUE;
}

static NvBool uvmfull_is_bitmask_set(NvU64 offset, NvU64 *bitmask)
{
    // TODO: update to cpu page size once power PC support is added
    NvU64 bitNumber = offset / PAGE_SIZE_4K;
    return (bitmask[bitNumber / 64] & NVBIT64(bitNumber % 64));
}


static void uvm_get_gpu_phy_addr_4k_aligned(Uvm2MbVaDesc *vaDesc, UvmGpuPde* gpuPde, NvU64 vaddr,
    NvBool *isPhyAddrPresent, NvU64 * phyAddr, NvBool* isApertureSysmem)
{
    NvU32 curPgSize;
    UvmPteHints *pHint = NULL;
    UvmGpuPte *gpuPte;
    *isPhyAddrPresent = NV_FALSE;

    gpuPte = uvmfull_get_gpupde_current_pte(gpuPde, vaddr, &curPgSize);
    if (gpuPte && gpuPte->swcachevalid && gpuPte->valid)
    {
        // get the phy addr mapped to current gpu virtual address
        if (gpuPte->aperture == UvmMemType_sysmem_coh ||
            gpuPte->aperture == UvmMemType_sysmem_ncoh)
        {
            *isApertureSysmem = NV_TRUE;
            pHint = GET_VA2M_HINT(vaDesc, vaddr, curPgSize);
            UVM_PANIC_ON(!pHint->cpuPfnValid);
            *phyAddr = PFN_TO_PA(pHint->cpuPfn);
        }
        else
        {
            *isApertureSysmem = NV_FALSE;

            if (curPgSize == PAGE_SIZE_4K)
            {
                *phyAddr = PFN_TO_PA(gpuPte->pfn);
            }
            else if (curPgSize == PAGE_SIZE_64K)
            {
                NvU64 offset =
                    GET_PDE_PTE_OFFSET(vaddr, PAGE_SHIFT_4K) -
                    GET_PDE_PTE_OFFSET(vaddr & PAGE_MASK_64K,
                    PAGE_SHIFT_4K);
                *phyAddr = PFN_TO_PA(gpuPte->pfn) + PAGE_SIZE_4K * offset;
            }
            else
            {
                NvU64 offset =
                    (GET_PDE_PTE_OFFSET(vaddr, PAGE_SHIFT_4K)) -
                    (GET_PDE_PTE_OFFSET(vaddr & PAGE_MASK_2M,
                    PAGE_SHIFT_4K));
                *phyAddr = PFN_TO_PA(gpuPte->pfn) + PAGE_SIZE_4K * offset;
            }
        }
        *isPhyAddrPresent = NV_TRUE;
    }
}

static NV_STATUS uvmfull_access_memory_copy_data(UvmDebugAccessType accessType, NvBool isApertureSysmem,
    UvmPushbuffer *pb, NvU64 uvmPA, NvU64 userPA, NvU64 copyBytes)
{
    // TODO: resolve below copies for kepler
    if (accessType == UvmDebugAccessTypeRead)
    {
        if (isApertureSysmem)
        {
            uvm_push_copy(pb, uvmPA, userPA, copyBytes,
                NV_UVM_COPY_SRC_TYPE_PHYSICAL | NV_UVM_COPY_DST_TYPE_PHYSICAL,
                (NvU32)NV_UVM_COPY_SRC_LOCATION_SYSMEM, (NvU32)NV_UVM_COPY_DST_LOCATION_SYSMEM);
        }
        else
        {
            uvm_push_gpu_to_cpu_copy(pb, uvmPA, userPA, copyBytes,
                NV_UVM_COPY_SRC_TYPE_PHYSICAL | NV_UVM_COPY_DST_TYPE_PHYSICAL);
        }

    }
    else if (accessType == UvmDebugAccessTypeWrite)
    {
        if (isApertureSysmem)
        {
            uvm_push_copy(pb, userPA, uvmPA, copyBytes,
                NV_UVM_COPY_SRC_TYPE_PHYSICAL | NV_UVM_COPY_DST_TYPE_PHYSICAL,
                (NvU32)NV_UVM_COPY_SRC_LOCATION_SYSMEM, (NvU32)NV_UVM_COPY_DST_LOCATION_SYSMEM);
        }
        else
        {
            uvm_push_cpu_to_gpu_copy(pb, userPA, uvmPA, copyBytes,
                NV_UVM_COPY_SRC_TYPE_PHYSICAL | NV_UVM_COPY_DST_TYPE_PHYSICAL);
        }
    }
    else
    {
        return NV_ERR_INVALID_ARGUMENT;
    }

    return NV_OK;
}

static NV_STATUS uvmfull_access_memory_per_gpu(Uvm2MbVaDesc *vaDesc, UvmGpuPde *gpuPde,
    UvmPbOps *pbOps, NvU64 userPageOffset, NvU64 currUvmVa, NvU64 numBytes, struct page* *userPageArray,
    UvmDebugAccessType accessType, NvU64 *bitmask, NvBool *pIsBitmaskSet, NvU64 bitmaskBase)
{
    NvU64 userPA;
    NvU64 uvmPA;
    NvU64 uvmOffset;
    NvBool isBitmaskSet;
    NvBool isApertureSysmem = NV_FALSE;
    UvmPushbuffer *pb;
    NvU64 vaddr = UVM_ALIGN_ADDR(currUvmVa, PAGE_SIZE_4K);
    NvU32 currUserPg = 0;
    NvU64 bytesToCopy = 0;
    NvU64 userPageRemainingBytes =0;
    NvBool isPhyAddrPresent;

    pb = uvm_pbops_gpu_pb(pbOps, gpuPde->hostGpuId);
    UVM_PANIC_ON(!pb);

    userPA = PFN_TO_PA(page_to_pfn(userPageArray[currUserPg++])) + userPageOffset;
    userPageRemainingBytes = PAGE_SIZE_4K - userPageOffset;

    while (numBytes > 0)
    {
        // check if current page is already copied
        if (uvmfull_is_bitmask_set((vaddr - (NvU64)bitmaskBase), bitmask) == NV_TRUE)
            continue;
        // 
        // We loop per gpu page and push physical copies.
        // The user buffer VA and gpu VA both can be unaligned and hence we cross
        // the page boundaries at different offsets. We need to take all cases
        // into consideration while pushing copies
        //
        uvm_get_gpu_phy_addr_4k_aligned(vaDesc, gpuPde, vaddr,&isPhyAddrPresent, &uvmPA, &isApertureSysmem);

        if (isPhyAddrPresent)
        {
            uvmOffset = currUvmVa - vaddr;
            uvmPA = uvmPA + uvmOffset;
            bytesToCopy = numBytes < (PAGE_SIZE_4K - uvmOffset) ? numBytes : (PAGE_SIZE_4K - uvmOffset);

            if (bytesToCopy < userPageRemainingBytes)
            {
                 uvmfull_access_memory_copy_data(accessType, isApertureSysmem, 
                     pb, uvmPA, userPA, bytesToCopy);

                userPA += bytesToCopy;
                userPageRemainingBytes -= bytesToCopy;
            }
            else if (bytesToCopy > userPageRemainingBytes)
            {
                NvU64 remBytesToCopy;
                uvmfull_access_memory_copy_data(accessType, isApertureSysmem, 
                    pb, uvmPA, userPA, userPageRemainingBytes);

                uvmPA += userPageRemainingBytes;

                userPA = PFN_TO_PA(page_to_pfn(userPageArray[currUserPg++]));
                remBytesToCopy = bytesToCopy- userPageRemainingBytes;

                 uvmfull_access_memory_copy_data(accessType, isApertureSysmem, 
                     pb, uvmPA, userPA, remBytesToCopy);

                userPA = userPA + remBytesToCopy;
                userPageRemainingBytes = PAGE_SIZE_4K - remBytesToCopy;
            }
            else
            {
                 uvmfull_access_memory_copy_data(accessType, isApertureSysmem, 
                     pb, uvmPA, userPA, userPageRemainingBytes);

                if ((numBytes - userPageRemainingBytes) > 0)
                {
                    userPA = PFN_TO_PA(page_to_pfn(userPageArray[currUserPg++]));
                    userPageRemainingBytes = PAGE_SIZE_4K;
                }
            }

            uvmfull_update_bitmask((vaddr - (NvU64)bitmaskBase), bitmask, &isBitmaskSet);
        }
        else
        {
            //
            // TODO: Bug 1672184: Allocate in case of write access if managed
            // fill zero copies for reads
            //
            return NV_ERR_INVALID_ADDRESS;
        }

        vaddr += PAGE_SIZE_4K;
        currUvmVa = vaddr;
        numBytes -= bytesToCopy;
    }

    if (isBitmaskSet)
    {
        uvm_pbops_submit(pbOps, gpuPde->hostGpuId, pb, &vaDesc->tracker);
        uvm_move_tracker(&vaDesc->tracker, pbOps->t);
        *pIsBitmaskSet = isBitmaskSet;
    }
    return NV_OK;
}

NV_STATUS uvmfull_access_memory(UvmVaspace *vaspace, NvU64 userCopyBuffer, NvU64 base,
                                NvU64 length, UvmDebugAccessType accessType, NvU64 *bitmask,
                                NvBool *isBitmaskSet)
{
    NvU64 vaddr;
    Uvm2MbVaDesc *vaDesc = NULL;
    UvmVatContext query;
    UvmPbOps *pbOps = NULL;
    NV_STATUS status = NV_OK;
    UvmGpuPde *gpuPde;
    struct page* *tempUserPageArray = NULL;
    NvU32 i;
    int returnVal;
    NvU64 currUvmVa = base;
    NvU64 userPageOffset;
    NvU64 userPagesReq;
    NvU64 numBytes = 0;

    *isBitmaskSet = NV_FALSE;
    //
    // temporary array to store user pinned pages per Va Desc
    // We need one more page than GPU_PDE_MAX_PAGES as the user VA is unaligned
    //
    tempUserPageArray = vmalloc(sizeof (struct page*) * GPU_PDE_MAX_PAGES + 1);
    if(!tempUserPageArray)
        return NV_ERR_NO_MEMORY;

    // 2MB loop
    while (length > 0)
    {
        userPagesReq = 0; 
        pbOps = NULL;
        vaddr = UVM_ALIGN_ADDR(currUvmVa, PAGE_SIZE_4K);

        // calculate number of pages that fall within this PDE
        query.accessType = UvmVatAccess_GenericQuery;
        query.address = vaddr;
        query.vaspace = vaspace;
        vaDesc = uvmfull_vat_acquire_leaf(&vaspace->vat, &query);
        if (!vaDesc)
        {
            status = NV_ERR_INVALID_ADDRESS;
            goto cleanup_vadesc;
        }

        // alloc push buffer for all active gpus in this range
        pbOps = uvm_pbops_start(vaspace);
        if (!pbOps)
        {
            status = NV_ERR_NO_MEMORY;
            goto cleanup_vadesc;
        }

        uvm_pbops_reserve_acquire_space(pbOps, &vaDesc->tracker);

        numBytes = length < (GPU_PDE_MAX_PAGES* PAGE_SIZE_4K) ? length : (GPU_PDE_MAX_PAGES* PAGE_SIZE_4K);
        userPageOffset = userCopyBuffer - UVM_ALIGN_ADDR(userCopyBuffer, PAGE_SIZE_4K);
        userPagesReq = UVM_ALIGN_UP((numBytes + userPageOffset), PAGE_SIZE_4K)/PAGE_SIZE_4K;

        // get userPA for this range
        down_read(&current->mm->mmap_sem);
        returnVal = get_user_pages(current, current->mm, (unsigned long)(UVM_ALIGN_ADDR(userCopyBuffer, PAGE_SIZE_4K)),
            userPagesReq, 1, 0, tempUserPageArray, NULL);
        up_read(&current->mm->mmap_sem);

        if (returnVal != userPagesReq)
        {
            status = NV_ERR_GENERIC;
            goto cleanup_vadesc;
        }

        // walk through all the pdes to find the PA backing and push copies
        desc_for_each_gpu_pde(gpuPde, vaDesc)
        {
            status = uvmfull_access_memory_per_gpu(vaDesc, gpuPde, pbOps, userPageOffset, currUvmVa, numBytes,
            tempUserPageArray, accessType, bitmask, isBitmaskSet, base);
            if (status != NV_OK)
                goto cleanup_vadesc;
        }

        status = uvm_wait_for_tracker(&vaDesc->tracker);

cleanup_vadesc:

        for (i = 0; i < userPagesReq; i++)
        {
            // mark the user page as dirty if uvm has written to those pages:
            if (accessType == UvmDebugAccessTypeRead)
                    set_page_dirty_lock(tempUserPageArray[i]);

            UVM_PANIC_ON(tempUserPageArray[i] == 0);
            page_cache_release(tempUserPageArray[i]);
            tempUserPageArray[i] = 0;
        }

        if(pbOps)
            uvm_pbops_end(pbOps);

        uvmfull_vat_release_leaf(&vaspace->vat, &query);

        if (status != NV_OK)
            break;

        userCopyBuffer += numBytes;
        currUvmVa += numBytes;
        length -= numBytes;
    } // end of 2MB loop

    vfree(tempUserPageArray);

    return status;
}
