/*******************************************************************************
    Copyright (c) 2015-2016 NVIDIA Corporation

    Permission is hereby granted, free of charge, to any person obtaining a copy
    of this software and associated documentation files (the "Software"), to
    deal in the Software without restriction, including without limitation the
    rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
    sell copies of the Software, and to permit persons to whom the Software is
    furnished to do so, subject to the following conditions:

        The above copyright notice and this permission notice shall be
        included in all copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
    IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
    FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
    THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
    LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
    FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
    DEALINGS IN THE SOFTWARE.

*******************************************************************************/

#include "uvm_linux.h"
#include "uvm_full.h"
#include "uvm_full_device_mgmt.h"
#include "uvm_full_perf.h"

void uvmfull_notify_fault(Uvm2MbVaDesc *desc2M, NvU64 pteFault, NvU64 pteBase, NvU32 cpuPages, UvmGpuId gpuId,
                          UvmFaultType faultType, UvmAccessType accessType, NvU64 tstamp)
{
    UvmPerfPageCounters *counters;
    NvU32 i;
    NvU64 elapsed;

    counters = &desc2M->perf[pteFault];

    if (accessType == UvmAccessType_atomic)
        ++counters->atomic_faults;
    else if (accessType == UvmAccessType_write)
        ++counters->write_faults;
    else if (accessType == UvmAccessType_read)
        ++counters->read_faults;

    if (counters->tstamps_tail == counters->tstamps_head)
        elapsed = tstamp;
    else if (counters->tstamps_tail == 0)
        elapsed = tstamp - counters->fault_tstamps[UVM_PERF_INFO_MAX_TSTAMPS - 1].time;
    else
        elapsed = tstamp - counters->fault_tstamps[counters->tstamps_tail - 1].time;

    // Compute elapsed time in microseconds
    elapsed /= 1000;

    if (counters->tstamps_tail == counters->tstamps_head || elapsed > g_uvm_thrashing_lapse) {
        counters->thrashing_faults = 0;
        counters->thrashing = NV_FALSE;
    }
    else {
        if (++counters->thrashing_faults > 2) {
            counters->thrashing = NV_TRUE;
            ++counters->thrashing_total;
        }
    }

    counters->fault_tstamps[counters->tstamps_tail].time = tstamp;
    counters->fault_tstamps[counters->tstamps_tail].proc = gpuId + 1; // TODO: check gpu id scheme
    counters->tstamps_tail = (counters->tstamps_tail + 1) % UVM_PERF_INFO_MAX_TSTAMPS;
    if (counters->tstamps_tail == counters->tstamps_head)
        counters->tstamps_head = (counters->tstamps_head + 1) % UVM_PERF_INFO_MAX_TSTAMPS;

    if (gpuId != 0) {
        // GPU fault
        for (i = 0; i < cpuPages; ++i) {
            counters = &desc2M->perf[pteBase + i];
            // Update number of migrations for all dirty pages in host memory that belong to the faulting GPU page
            if (desc2M->hints[pteBase+i].cpuPfnDirty == 0x1)
                ++counters->migrations;
        }
    }
    else {
        // CPU fault
        ++counters->migrations;
    }
}

void uvmfull_notify_retry(Uvm2MbVaDesc *desc2M, NvU64 pteFault, NvU64 pteBase, NvU32 cpuPages, UvmGpuId gpuId,
                          UvmFaultType faultType, UvmAccessType accessType, NvU64 tstamp)
{
    UvmPerfPageCounters *counters;
    NvU64 elapsed;

    counters = &desc2M->perf[pteFault];

    // Compute elapsed time in microseconds
    elapsed = (tstamp - counters->fault_tstamps[counters->tstamps_tail].time) / 1000;

    // Recompute thrashing conditions
    if (counters->tstamps_tail == counters->tstamps_head || elapsed > g_uvm_thrashing_lapse) {
        counters->thrashing_faults = 0;
        counters->thrashing = NV_FALSE;
    }
    else {
        if (++counters->thrashing_faults >= g_uvm_thrashing_fault_count)
            counters->thrashing = NV_TRUE;
    }
}

NV_STATUS uvmfull_perf_get_thrashing_info(UVM_PERF_GET_THRASHING_INFO_PARAMS *params, struct file *filep)
{
    UvmVaspace *vaspace = NULL;
    NvU64 vaStart, descVaStart;
    UvmGlobalState *uvm = uvmfull_get_global();
    UvmVatContext query;
    Uvm2MbVaDesc* vaDesc = NULL;
    UvmPerfPageCounters *counters;

    int pte4k;

    BUILD_BUG_ON(UVM_PERF_INFO_MAX_TSTAMPS > USHRT_MAX);

    vaStart     = UVM_ALIGN_ADDR(params->requestedVa, UVM_CPU_MIN_PAGE_SIZE);
    descVaStart = UVM_ALIGN_ADDR(vaStart, UVM_VA_DESC_SIZE);

    down_read(&uvm->lock);
    vaspace = (UvmVaspace *)filep->private_data;
    if (!vaspace)
    {
        UVM_ERR_PRINT("struct file* (0x%p)->private_data is missing a UvmVaspace pointer\n", filep);
        params->rmStatus = NV_ERR_INVALID_REQUEST;
        goto cleanup_uvm;
    }
    down_read(&vaspace->lock);

    query.address    = descVaStart;
    query.accessType = UvmVatAccess_GenericQuery;
    query.gpuId      = 0;
    query.vaspace    = vaspace;

    vaDesc = uvmfull_vat_acquire_leaf(&vaspace->vat, &query);
    if (!vaDesc) {
        UVM_ERR_PRINT("No Uvm2MbVaDesc found for page %p\n", (void *)vaStart);
        params->rmStatus = NV_ERR_INVALID_REQUEST;
        goto cleanup_vaspace;
    }
    else {
        params->rmStatus = NV_OK;
    }

    pte4k = (vaStart - descVaStart) >> PAGE_SHIFT_4K;

    counters = &vaDesc->perf[pte4k];

    params->info.atomic_faults = counters->atomic_faults;
    params->info.write_faults = counters->write_faults;
    params->info.read_faults = counters->read_faults;
    params->info.migrations = counters->migrations;
    params->info.thrashing_total = counters->thrashing_total;

    if (g_uvm_nothrashing == UvmPerfThrashingPreventionThrottleSched ||
        g_uvm_nothrashing == UvmPerfThrashingPreventionThrottleSleep) {
        // Copy stats for thrashing prevention when using throttling
        params->info.throttle.naps = counters->throttle.naps;
    }
    else if (g_uvm_nothrashing == UvmPerfThrashingPreventionPinSysmem) {
        // Copy stats for thrashing prevention when pinning the page in sysmem
        params->info.pin_sysmem.pins = counters->pin_sysmem.pins;
    }
    else {
        // Not preventing thrashing. Copy the queue of the last () page fault timestamps
        if (counters->tstamps_tail >= counters->tstamps_head) {
            unsigned short elems = counters->tstamps_tail - counters->tstamps_head;

            memcpy(params->info.none.fault_tstamps, &counters->fault_tstamps[counters->tstamps_head],
                   elems * sizeof(UvmPerfTimestamp));
            // Set the non-used entries to zero
            memset(params->info.none.fault_tstamps + elems * sizeof(UvmPerfTimestamp), 0,
                   (UVM_PERF_INFO_MAX_TSTAMPS - elems) * sizeof(UvmPerfTimestamp));
        }
        else {
            // Queue overflew and head pointer moved. Split in two copies
            memcpy(params->info.none.fault_tstamps, &counters->fault_tstamps[counters->tstamps_head],
                   (UVM_PERF_INFO_MAX_TSTAMPS - counters->tstamps_head) * sizeof(UvmPerfTimestamp));

            memcpy(params->info.none.fault_tstamps + (UVM_PERF_INFO_MAX_TSTAMPS - counters->tstamps_head),
                   &counters->fault_tstamps[0], counters->tstamps_tail * sizeof(UvmPerfTimestamp));
        }
    }

    uvmfull_vat_release_leaf(&vaspace->vat, &query);

cleanup_vaspace:
    up_read(&vaspace->lock);

cleanup_uvm:
    up_read(&uvm->lock);
    return params->rmStatus;
}
