/*******************************************************************************
    Copyright (c) 2014-2015 NVIDIA Corporation

    Permission is hereby granted, free of charge, to any person obtaining a copy
    of this software and associated documentation files (the "Software"), to
    deal in the Software without restriction, including without limitation the
    rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
    sell copies of the Software, and to permit persons to whom the Software is
    furnished to do so, subject to the following conditions:

        The above copyright notice and this permission notice shall be
        included in all copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
    IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
    FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
    THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
    LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
    FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
    DEALINGS IN THE SOFTWARE.

*******************************************************************************/

#include "uvm_full_test.h"
#include "uvm_full.h"
#include "uvm_common.h"
#include "nv_uvm_interface.h"
#include "uvmfull_test_ioctl.h"
#include "uvm_full_api.h"
#include "uvm_full_device_mgmt.h"
#include "uvm_full_pagetbl_mgmt.h"
#include "uvm_full_ctx_mgmt.h"
#include "uvm_full_fault_handler.h"

long uvmfull_test_ioctl(struct file *filp, unsigned int cmd, unsigned long arg)
{
    // Disable all test entry points if the module parameter wasn't provided.
    // These should not be enabled in a production environment.
    if (!uvm_enable_builtin_tests) {
        UVM_INFO_PRINT("ioctl %d not found. Did you mean to insmod with uvm_enable_builtin_tests=1?\n", cmd);
        return -EINVAL;
    }

    switch (cmd)
    {
        // Add test-only apis below
        UVM_FULL_ROUTE_CMD_STACK(UVM_RUN_TEST,               uvmfull_run_test);

        UVM_FULL_ROUTE_CMD_STACK(UVM_TEST_REGION_COMMIT,     uvm_test_region_commit);
        UVM_FULL_ROUTE_CMD_STACK(UVM_TEST_SET_VA_STATE,      uvm_test_set_va_state);
        UVM_FULL_ROUTE_CMD_STACK(UVM_TEST_SET_FAULT_POLICY,  uvm_test_set_fault_policy);
        UVM_FULL_ROUTE_CMD_STACK(UVM_TEST_REGISTER_DEVICE,   uvm_test_register_device);
        UVM_FULL_ROUTE_CMD_STACK(UVM_TEST_UNIT_TEST,         uvm_test_unit_test);

        UVM_FULL_ROUTE_CMD_STACK(UVM_TEST_GET_DEVICE_LIST,    uvm_test_get_device_list);
        UVM_FULL_ROUTE_CMD_STACK(UVM_TEST_REGISTER_CHANNEL,   uvm_test_register_channel);
        UVM_FULL_ROUTE_CMD_STACK(UVM_TEST_UNREGISTER_CHANNEL, uvm_test_unregister_channel);
        UVM_FULL_ROUTE_CMD_STACK(UVM_TEST_SET_PAGE_TABLE_UPDATE_POLICY,
                                 uvmfull_test_set_page_table_update_policy);
        UVM_FULL_ROUTE_CMD_STACK(UVM_TEST_GET_VA_PTE_INFO,    uvm_test_get_va_pte_info);
        UVM_FULL_ROUTE_CMD_STACK(UVM_TEST_SPLIT_PAGES,           uvm_test_split_pages);
        UVM_FULL_ROUTE_CMD_STACK(UVM_TEST_MAP_GPU_SYSMEM_PAGES, uvm_test_map_gpu_sysmem_pages);
        UVM_FULL_ROUTE_CMD_STACK(UVM_DEBUG_UNMAP,             uvm_debug_unmap);
        UVM_FULL_ROUTE_CMD_STACK(UVM_TEST_INVALIDATE_TLB,     uvm_test_invalidate_tlb);
        UVM_FULL_ROUTE_CMD_STACK(UVM_TEST_CREATE_ZFOD_PAGE,   uvm_test_create_zfod_page);
        UVM_FULL_ROUTE_CMD_STACK(UVM_TEST_CONFIGURE_ZFOD_POLICY, uvm_test_configure_zfod_policy);
        UVM_FULL_ROUTE_CMD_STACK(UVM_TEST_RESET_FAULT_BUFFER_OVERFLOW,
                                 uvmfull_test_reset_fault_buffer_overflow);
        UVM_FULL_ROUTE_CMD_STACK(UVM_TEST_CONFIGURE_FAULT_HANDLER,
                                 uvmfull_test_configure_fault_handler);
        UVM_FULL_ROUTE_CMD_STACK(UVM_TEST_SWAP_GPU_PAGES, uvm_test_swap_gpu_pages);
        UVM_FULL_ROUTE_CMD_STACK(UVM_TEST_EXPLICIT_VA_MIGRATE, uvm_test_explicit_va_migrate);
        UVM_FULL_ROUTE_CMD_STACK(UVM_TEST_CHANGE_PTE_MAPPING, uvm_test_change_pte_mapping);
        UVM_FULL_ROUTE_CMD_STACK(UVM_TEST_FORCE_PAGE_SIZE, uvm_test_force_page_size);
        UVM_FULL_ROUTE_CMD_STACK(UVM_TEST_MERGE_PAGES, uvm_test_merge_pages);
    }

    return -EINVAL;
}

NV_STATUS uvm_test_region_commit(UVM_TEST_REGION_COMMIT_PARAMS *pParams,
                                     struct file *filp)
{
    // Deprecated
    return NV_ERR_NOT_SUPPORTED;
}

NV_STATUS uvm_test_set_va_state(UVM_TEST_SET_VA_STATE_PARAMS *stateParams,
                                struct file *filp )
{
    // deprecated
    return NV_ERR_NOT_SUPPORTED;
}

static void set_test_pte_hints_params(UVM_TEST_SET_FAULT_POLICY_PARAMS *policyParams,
                               UvmPteHintsParams *pPteHintsParams)
{
    pPteHintsParams->updateFlags |= UVM_GPU_VA_PTE_HINTS_FLAG_SET_FLT_PLCY_RW
                                  | UVM_GPU_VA_PTE_HINTS_FLAG_SET_FLT_PLCY_ATOMIC
                                  | UVM_GPU_VA_PTE_HINTS_FLAG_SET_FLT_PLCY_SYSONLY
                                  | UVM_GPU_VA_PTE_HINTS_FLAG_SET_FLT_PLCY_RD_DUP
                                  | UVM_GPU_VA_PTE_HINTS_FLAG_SET_FLT_PLCY_CANCEL;

    pPteHintsParams->pte.testPolicyReadWrite  = policyParams->write;
    pPteHintsParams->pte.testPolicyAtomic     = policyParams->atomic;
    pPteHintsParams->pte.testPolicyRdDuplicate= policyParams->duplicate;
    pPteHintsParams->pte.testPolicySysMemOnly = policyParams->sysmemonly;
    pPteHintsParams->pte.testPolicyCancel     = policyParams->cancel;
}

NV_STATUS uvm_test_set_fault_policy(
                UVM_TEST_SET_FAULT_POLICY_PARAMS* policyParams,
                struct file *filp)
{
    UvmVaspace *vaspace;
    UvmGlobalState *uvm;
    Uvm2MbVaDesc *vaDesc;
    NvU64 vaddr = UVM_ALIGN_ADDR(policyParams->addr, PAGE_SIZE_4K);
    NvU32 remPages;
    NvU32 numPages = 0;
    NV_STATUS status = NV_OK;
    UvmVatContext query;
    UvmPteHintsParams pteHintsParams={0};

    set_test_pte_hints_params(policyParams, &pteHintsParams);

    remPages = DIV_ROUND_UP(policyParams->length, PAGE_SIZE_4K);

    uvm = uvmfull_get_global();
    vaspace = (UvmVaspace*)filp->private_data;

    down_read(&uvm->lock);
    down_read(&vaspace->lock);

    // limit to PDE page size
    while (remPages)
    {
        query.accessType = UvmVatAccess_GenericQuery;
        query.address = vaddr;
        query.vaspace = vaspace;
        vaDesc = uvmfull_vat_acquire_leaf(&vaspace->vat, &query);
        if (!vaDesc)
        {
            status = NV_ERR_NO_MEMORY;
            break;
        }

        // calculate number of pages that fall within this PDE
        numPages = GPU_PDE_MAX_PAGES - GET_PDE_PTE_OFFSET(vaddr, PAGE_SHIFT_4K);
        numPages = numPages > remPages ? remPages : numPages;

        // TODO check if no ptes are set in this range!

        // Now that va is created, update with va_flags for the interested va range
        uvmfull_update_pte_hints(vaddr, numPages, PAGE_SIZE_4K, &pteHintsParams, vaDesc);
        vaddr += (numPages*PAGE_SIZE_4K);

        remPages -= numPages;
        uvmfull_vat_release_leaf(&vaspace->vat, &query);
    }

    up_read(&vaspace->lock);
    up_read(&uvm->lock);
    return status;
}

NV_STATUS uvm_test_register_device(UVM_TEST_REGISTER_DEVICE_PARAMS* param,
                                   struct file* filep)
{
    UVM_DBG_PRINT_UUID("\nregister device", &param->gpuUuid);
    return uvmfull_register_gpu_vaspace(filep, &param->gpuUuid, param->hClient,
                                        param->hVaspace);
}

NV_STATUS uvm_test_get_device_list(
            UVM_TEST_GET_DEVICE_LIST_PARAMS* params, struct file* filep)
{
    NvU32 j;
    NvU32 i = 0;
    UvmGlobalState *uvm = uvmfull_get_global();
    UvmGpu *gpu;

    down_read(&uvm->lock);
    for (j = 0; j < UVM_FULL_MAX_GPUS; j++)
    {
        gpu = uvmfull_gpuid_to_active_gpu(j);
        if (gpu)
        {
            if (i >= UVM_TEST_DEVICES)
                break;
            params->device[i].bValidDevice = NV_TRUE;
            params->device[i].bCanReplayFaults =
                                    uvmfull_is_gpu_fault_capable(j);
            memcpy(&params->device[i].gpuUuid,
                    &gpu->gpuUuid,
                    sizeof(gpu->gpuUuid));
            i++;
        }
    }
    up_read(&uvm->lock);
    return NV_OK;
}

NV_STATUS uvm_test_register_channel(UVM_TEST_REGISTER_CHANNEL_PARAMS* param,
                                   struct file* filep)
{
    UVM_DBG_PRINT_UUID("\nregister channel", &param->gpuUuid);
    return NV_ERR_NOT_SUPPORTED;
}

NV_STATUS uvm_test_unregister_channel(UVM_TEST_UNREGISTER_CHANNEL_PARAMS* param,
                                   struct file* filep)
{
    UVM_DBG_PRINT_UUID("\nregister channel", &param->gpuUuid);
    return NV_ERR_NOT_SUPPORTED;
}

NV_STATUS
uvmfull_test_set_page_table_update_policy(UVM_TEST_SET_PAGE_TABLE_UPDATE_POLICY_PARAMS *pParams,
                                      struct file *filp)
{
    NV_STATUS status = NV_OK;
    UvmVaspace *vaspace = NULL;
    UvmGlobalState *uvm = uvmfull_get_global();
    UvmGpuVaspace *gpuVaspace;
    UvmGpu *gpu;
    UvmGpuId gpuId;

    down_read(&uvm->lock);
    gpu = uvmfull_uuid_to_active_gpu(&pParams->gpuUuid);
    if (!gpu || !gpu->bValid)
    {
        status = NV_ERR_INVALID_REQUEST;
        goto cleanup_global;
    }

    vaspace = (UvmVaspace*)filp->private_data;
    if (!vaspace)
    {
        status = NV_ERR_INVALID_REQUEST;
        goto cleanup_global;
    }

    down_write(&vaspace->lock);

    if (vaspace->filep != filp)
    {
        status = NV_ERR_INVALID_REQUEST;
        goto cleanup_vaspace;
    }
    gpuId = gpu->gpuId;
    gpuVaspace = &vaspace->gpuVaspaces[gpuId];
    if (!gpuVaspace->client.bValid)
    {
        status = NV_ERR_INVALID_REQUEST;
        goto cleanup_vaspace;
    }

    gpuVaspace->policy.bForceReplayStartAckAll =
        pParams->bForceReplayStartAckAll;
    gpuVaspace->policy.bForcePteMemsets = pParams->bForcePteMemsets;
    gpuVaspace->policy.bForceMigration = pParams->bForceMigration;
    gpuVaspace->policy.bForceP2PMigration = pParams->bForceP2PMigration;

    switch (pParams->mappingSize)
    {
        case PAGE_SIZE_4K:
        case PAGE_SIZE_64K:
        case PAGE_SIZE_2M:
        case PAGE_SIZE_USE_DEFAULT:
            gpuVaspace->policy.mappingSize = pParams->mappingSize;
            break;
        default:
            status = NV_ERR_INVALID_REQUEST;
            goto cleanup_vaspace;
    }

    switch (pParams->faultServiceRange)
    {
        case PAGE_SIZE_4K:
        case PAGE_SIZE_64K:
        case PAGE_SIZE_2M:
        case PAGE_SIZE_USE_DEFAULT:
            gpuVaspace->policy.faultServiceRange = pParams->faultServiceRange;
            break;
        default:
            status = NV_ERR_INVALID_REQUEST;
            goto cleanup_vaspace;
    }

cleanup_vaspace:
    up_write(&vaspace->lock);
cleanup_global:
    up_read(&uvm->lock);
    return status;
}

NV_STATUS uvm_debug_unmap(
            UVM_DEBUG_UNMAP_PARAMS* params, struct file* filep)
{
    UvmVaspace *vaspace;
    UvmGlobalState *uvm;
    NV_STATUS status = NV_OK;
    UvmGpu *pGpu;

    UVM_DBG_PRINT_UUID("\ngpu unmap range", &params->processorUuid);

    vaspace = (UvmVaspace*)filep->private_data;
    uvm = uvmfull_get_global();

    down_read(&uvm->lock);

    //TODO handle cpu unmap
    pGpu = uvmfull_uuid_to_active_gpu((UvmGpuUuid *)&params->processorUuid);
    if (pGpu == NULL)
    {
        up_read(&uvm->lock);
        return NV_ERR_INVALID_REQUEST;
    }

    down_read(&vaspace->lock);

    uvmfull_unmap_on_gpus(vaspace, params->base,
                          params->length, (NvU8 *)&pGpu->gpuId, 1,
                          params->skipTlbInvalidate);

    up_read(&vaspace->lock);
    up_read(&uvm->lock);
    return status;
}

NV_STATUS
uvmfull_test_reset_fault_buffer_overflow(UVM_TEST_RESET_FAULT_BUFFER_OVERFLOW_PARAMS *pParams,
                                         struct file *filp)
{
    NV_STATUS status = NV_OK;
    UvmGpu *pGpu;
    UvmGpuFaultBufferInfo *buf;
    UvmGlobalState *uvm;
    UvmFaultBufferRegisters testFaultBufferInfo;

    uvm = uvmfull_get_global();
    pGpu = uvmfull_intr_uuid_to_active_gpu(&pParams->gpuUuid);
    if (!pGpu || !pGpu->bValid)
    {
        // This might be a non faulting gpu afterall
        pGpu = uvmfull_uuid_to_active_gpu(&pParams->gpuUuid);
        if (!pGpu || !pGpu->bValid)
           return NV_ERR_INVALID_ARGUMENT;
    }

    buf = &pGpu->faultBufferInfo;
    // Take the faultbuffer lock as we are going to query it.
    down_write(&buf->lock);

    testFaultBufferInfo.pFaultBufferGet = buf->pFaultBufferGet;
    testFaultBufferInfo.pFaultBufferPut = buf->pFaultBufferPut;
    testFaultBufferInfo.pFaultBufferInfo = buf->pFaultBufferInfo;


    buf->faultOps.clearFaultBufferOverflow(testFaultBufferInfo);

    up_write(&buf->lock);
    pParams->rmStatus = status;
    return status;
}

NV_STATUS uvm_test_invalidate_tlb_entry(UvmVaspace *vaspace, NvU64 targetVa,
                                        UvmTargetVAMode targetVAMode,
                                        UvmTargetPdbMode targetPdb,
                                        UvmInvalidateTlbLevel tlbLevel,
                                        UvmInvalidateTlbMembarType membar,
                                        NvU32 gpcId, NvU32 clientId,
                                        UvmInvalidateTlbReplayType replayType,
                                        NvBool disableGpcInvalidate,
                                        UvmGpu *pGpu)
{
     UvmTracker tracker;
     UvmPbOps *pbOps;
     UvmPushbuffer *pb;
     NvU64 pdb = 0;
     NV_STATUS status = NV_OK;
     UvmTlbInvalidateVaParams invalidateParams = {0};
     UvmTlbInvalidateAckType tlbInvalidateAckType;
     uvm_init_tracker(&tracker);

     pbOps = uvm_pbops_start_on_gpu(pGpu->gpuId);
     UVM_PANIC_ON(!pbOps);
     if (!pbOps)
     {
         status = NV_ERR_NO_MEMORY;
         goto cleanup;
     }
     if (targetPdb == UvmTargetPdbMode_One) // if targetPdb is one, consider other params too
     {
         // Target PDB is all so ignore the targeted VA
         pdb = vaspace->gpuVaspaces[pGpu->gpuId].gpuPDB;
         invalidateParams.targetedVa = targetVa;
     } // else, invalidate all entries for all process
     invalidateParams.targetVAMode = targetVAMode;
     invalidateParams.invalidatelevel = tlbLevel;
     if (membar != UvmInvalidateTlbMemBar_none)
         tlbInvalidateAckType = UvmTlbInvalidateAckType_globally;
     else
         tlbInvalidateAckType = UvmTlbInvalidateAckType_none;

     pb = uvm_pbops_gpu_pb(pbOps, pGpu->gpuId);
     UVM_PANIC_ON(!pb);
     if (!pb)
     {
         status = NV_ERR_NO_MEMORY;
         goto cleanup;
     }

 uvm_push_memops_raw(pGpu->gpuId, pb, gpcId, clientId, replayType, tlbInvalidateAckType,
                     invalidateParams, membar, vaspace->gpuVaspaces[pGpu->gpuId].pdbAperture,
                     disableGpcInvalidate, pdb >> 12);

     uvm_pbops_submit(pbOps, pGpu->gpuId, pb, &tracker);
     uvm_move_tracker(&tracker, pbOps->t);
     status = uvm_wait_for_tracker(&tracker);
cleanup:
    if(pbOps)
        uvm_pbops_end(pbOps);
    return status;
}

NV_STATUS uvm_test_invalidate_tlb(
            UVM_TEST_INVALIDATE_TLB_PARAMS* params, struct file* filep)
{
    UvmVaspace *pVaspace;
    UvmGpuVaspace *pGpuVaspace;
    UvmGlobalState *pUvm;
    UvmGpu *pGpu;

    UVM_DBG_PRINT_UUID("\n UVM Test Invalidate TLB",&params->gpuUuid);

    pVaspace = (UvmVaspace *)filep->private_data;
    if (!pVaspace)
    {
        params->rmStatus = NV_ERR_INVALID_REQUEST;
        return params->rmStatus;
    }
    pUvm = uvmfull_get_global();

    down_read(&pUvm->lock);

    pGpu =  uvmfull_uuid_to_active_gpu(&params->gpuUuid);

    if (!pGpu || !pGpu->bValid)
    {
        params->rmStatus = NV_ERR_INVALID_REQUEST;
        goto cleanup_uvm;
    }

    down_read(&pVaspace->lock);
    if (pVaspace->filep != filep)
    {
        params->rmStatus = NV_ERR_INVALID_REQUEST;
        goto cleanup_vaspace;
    }

    pGpuVaspace = &pVaspace->gpuVaspaces[pGpu->gpuId];
    if (!pGpuVaspace->client.bValid)
    {
        params->rmStatus = NV_ERR_INVALID_REQUEST;
        goto cleanup_vaspace;
    }


    params->rmStatus = uvm_test_invalidate_tlb_entry(pVaspace, params->targetVa,
                                                     params->targetVAMode,
                                                     params->targetPdb,
                                                     params->tlbLevel,
                                                     params->membar,
                                                     params->gpcId,
                                                     params->clientId,
                                                     params->replayType,
                                                     params->disableGpcInvalidate,
                                                     pGpu
                                                    );
cleanup_vaspace:
    up_read(&pVaspace->lock);
cleanup_uvm:
    up_read(&pUvm->lock);
    return params->rmStatus;
}

NV_STATUS
uvmfull_test_configure_fault_handler(UVM_TEST_CONFIGURE_FAULT_HANDLER_PARAMS *pParams,
                                     struct file *filp)
{
    NV_STATUS status = NV_OK;
    UvmGpu *pGpu;
    UvmGpuFaultBufferInfo *buf;
    UvmGlobalState *uvm;

    uvm = uvmfull_get_global();
    pGpu = uvmfull_intr_uuid_to_active_gpu(&pParams->gpuUuid);
    if (!pGpu|| !pGpu->bValid)
    {
        // This might be a non faulting gpu afterall
        pGpu = uvmfull_uuid_to_active_gpu(&pParams->gpuUuid);
        if (!pGpu  || !pGpu->bValid)
           return NV_ERR_INVALID_ARGUMENT;
    }

    buf = &pGpu->faultBufferInfo;
    // Take the faultbuffer lock as we are going to query it.
    down_write(&buf->lock);

    buf->bOnlyReplaysNoResolves = pParams->bOnlyReplaysNoResolves;

    up_write(&buf->lock);
    pParams->rmStatus = status;
    return status;
}

NV_STATUS uvm_test_get_va_pte_info(
        UVM_TEST_GET_VA_PTE_INFO_PARAMS* params, struct file* filep)
{
    UvmVaspace         *pVaspace;
    UvmGlobalState     *pUvm;
    UvmGpuPde         *pGpuPde;
    UvmGpuPte         *pGpuPte;
    UvmVatContext     query;
    Uvm2MbVaDesc     *pUvm2MbVaDesc = NULL;
    UvmGpu             *pGpu;

    UVM_DBG_PRINT_UUID("\n UVM Test Get VA PTE Info",&params->gpuUuid);

    pVaspace = (UvmVaspace *)filep->private_data;
    if (!pVaspace)
    {
        params->rmStatus = NV_ERR_INVALID_REQUEST;
        return params->rmStatus;
    }
    pUvm = uvmfull_get_global();
    down_read(&pUvm->lock);

    pGpu =  uvmfull_uuid_to_active_gpu(&params->gpuUuid);

    if (!pGpu || !pGpu->bValid)
    {
        params->rmStatus = NV_ERR_INVALID_REQUEST;
        goto cleanup_uvm;
    }

    down_read(&pVaspace->lock);
    if (pVaspace->filep != filep)
    {
        params->rmStatus = NV_ERR_INVALID_REQUEST;
        goto cleanup_vaspace;
    }

    query.accessType = UvmVatAccess_GPUQuery;
    query.address = params->targetVa;
    query.vaspace = pVaspace;
    pUvm2MbVaDesc = uvmfull_vat_acquire_leaf(&pVaspace->vat, &query);
    if (!pUvm2MbVaDesc)
    {
        // return error
        params->rmStatus = NV_ERR_INVALID_ARGUMENT;
        goto cleanup_vaspace;
    }
    pGpuPde = uvmfull_get_gpupde_in_vadesc(pUvm2MbVaDesc, pGpu->gpuId);
    if (!pGpuPde)
    {
        // Invalid virtual address for this GPU
        params->rmStatus = NV_ERR_INVALID_REQUEST;
        goto cleanup_vaspace;
    }
    // Obtain the corresponding GPU PDE
    pGpuPte = uvmfull_get_gpupde_current_pte(pGpuPde, params->targetVa, &params->pgSize);
    if (!pGpuPte)
    {
        // Invalid virtual address for this GPU
        params->rmStatus = NV_ERR_INVALID_REQUEST;
        goto cleanup_vaspace;
    }

    params->aperture = pGpuPte->aperture;
    params->atomic   = pGpuPte->atomic;
    params->peerid     = pGpuPte->peerid;
    params->pfn         = pGpuPte->pfn;
    params->pfnvalid = pGpuPte->pfnvalid;
    params->valid     = pGpuPte->valid;
    params->readonly = pGpuPte->readonly;
    params->vol         = pGpuPte->vol;
    params->rmStatus = NV_OK;

cleanup_vaspace:
    if (pUvm2MbVaDesc)
        uvmfull_vat_release_leaf(&pVaspace->vat, &query);
    up_read(&pVaspace->lock);
cleanup_uvm:
    up_read(&pUvm->lock);
    return params->rmStatus;
}

// This function merges the pages of the same size and tags to a fixed new
// page size

NV_STATUS uvmfull_test_merge_gpupde_pte (UvmVaspace *pVaspace, UvmGpuId gpuId,
                                         NvU64 addr, NvU32 newPgSize
                                        )
{
    UvmVatContext query;
    Uvm2MbVaDesc *pUvm2MbVaDesc = NULL;
    UvmGpuPde *pGpuPde = NULL;
    UvmPbOps *pbOps = NULL;
    NV_STATUS status = NV_OK;

    query.accessType = UvmVatAccess_GPUQuery;
    query.address = addr;
    query.vaspace = pVaspace;
    query.gpuId = gpuId;
    pUvm2MbVaDesc = uvmfull_vat_acquire_leaf(&pVaspace->vat, &query);
    if (!pUvm2MbVaDesc)
    {
        // return error
        status = NV_ERR_INVALID_ARGUMENT;
        goto done;
    }

    pGpuPde = uvmfull_get_gpupde_in_vadesc(pUvm2MbVaDesc, gpuId);
    if (!pGpuPde)
    {
        status = NV_ERR_INVALID_ARGUMENT;
        goto done;
    }

    // Gather pushbuffers
    pbOps = uvm_pbops_start_on_gpu(gpuId);
    if (!pbOps)
    {
        status = NV_ERR_GENERIC;
        goto done;
    }
    uvm_pbops_reserve_acquire_space(pbOps, &pUvm2MbVaDesc->tracker);

    status = uvmfull_merge_gpupde_pte(pVaspace,
                                      gpuId,
                                      pUvm2MbVaDesc,
                                      pGpuPde,
                                      addr,
                                      newPgSize,
                                      pbOps);
    // Submit the pushbuffer.
    uvm_pbops_submit(pbOps, gpuId, uvm_pbops_gpu_pb(pbOps, gpuId),
                     &pUvm2MbVaDesc->tracker);
    uvm_move_tracker(&pUvm2MbVaDesc->tracker, pbOps->t);

    status = uvm_wait_for_tracker(&pUvm2MbVaDesc->tracker);
done:
    if (pbOps)
        uvm_pbops_end(pbOps);

    uvmfull_vat_release_leaf(&pVaspace->vat, &query);
    return status;
}
//
// This function splits the pages for the new given size to a fixed
// new page-size
//
NV_STATUS uvmfull_test_split_gpupde_pte(UvmVaspace *pVaspace, UvmGpuId gpuid,
                                   NvU64 addr, NvU32 newPgSize, NvU32 splitFlags)
{
    UvmVatContext query;
    Uvm2MbVaDesc *pUvm2MbVaDesc = NULL;
    UvmGpuPde *pGpuPde = NULL;
    UvmPbOps *pbOps = NULL;
    NV_STATUS status = NV_OK;

    query.accessType = UvmVatAccess_GPUQuery;
    query.address = addr;
    query.vaspace = pVaspace;
    query.gpuId = gpuid;
    pUvm2MbVaDesc = uvmfull_vat_acquire_leaf(&pVaspace->vat, &query);
    if (!pUvm2MbVaDesc)
    {
        // return error
        status = NV_ERR_INVALID_ARGUMENT;
        goto done;
    }

    pGpuPde = uvmfull_get_gpupde_in_vadesc(pUvm2MbVaDesc, gpuid);
    if (!pGpuPde)
    {
        status = NV_ERR_INVALID_ARGUMENT;
        goto done;
    }

    // Gather pushbuffers
    pbOps = uvm_pbops_start_on_gpu(gpuid);
    if (!pbOps)
    {
        status = NV_ERR_GENERIC;
        goto done;
    }
    uvm_pbops_reserve_acquire_space(pbOps, &pUvm2MbVaDesc->tracker);

    status = uvmfull_split_gpupde_pte(pVaspace,
                                      gpuid,
                                      pUvm2MbVaDesc,
                                      pGpuPde,
                                      addr,
                                      newPgSize,
                                      splitFlags,
                                      pbOps);
    // Submit the pushbuffer.
    uvm_pbops_submit(pbOps, gpuid, uvm_pbops_gpu_pb(pbOps, gpuid),
                     &pUvm2MbVaDesc->tracker);
    uvm_move_tracker(&pUvm2MbVaDesc->tracker, pbOps->t);

    status = uvm_wait_for_tracker(&pUvm2MbVaDesc->tracker);
done:
    if (pbOps)
        uvm_pbops_end(pbOps);

    uvmfull_vat_release_leaf(&pVaspace->vat, &query);
    return status;
}


NV_STATUS uvm_test_merge_pages(
        UVM_TEST_MERGE_PAGES_PARAMS* params, struct file* filep)
{
    UvmVaspace         *pVaspace;
    UvmGlobalState     *pUvm;
    UvmGpu             *pGpu;

    UVM_DBG_PRINT_UUID("\n UVM Test Merge Pages",&params->gpuUuid);

    params->rmStatus = NV_ERR_INVALID_REQUEST;
    UVM_CHECK_VALID_PAGE_SIZE(params->newPgSize, params->rmStatus);

    pVaspace = (UvmVaspace *)filep->private_data;
    if (!pVaspace)
    {
        params->rmStatus = NV_ERR_INVALID_REQUEST;
        return params->rmStatus;
    }
    pUvm = uvmfull_get_global();
    down_read(&pUvm->lock);

    pGpu =  uvmfull_uuid_to_active_gpu(&params->gpuUuid);

    if (!pGpu || !pGpu->bValid)
    {
        params->rmStatus = NV_ERR_INVALID_REQUEST;
        goto cleanup_uvm;
    }

    down_read(&pVaspace->lock);
    if (pVaspace->filep != filep)
    {
        params->rmStatus = NV_ERR_INVALID_REQUEST;
        goto cleanup_vaspace;
    }
    params->rmStatus =
            uvmfull_test_merge_gpupde_pte(pVaspace,
                                          pGpu->gpuId,
                                          params->targetVa,
                                          params->newPgSize
                                         );
cleanup_vaspace:
    up_read(&pVaspace->lock);
cleanup_uvm:
    up_read(&pUvm->lock);
return params->rmStatus;
}

NV_STATUS uvm_test_split_pages(
        UVM_TEST_SPLIT_PAGES_PARAMS* params, struct file* filep)
{
    UvmVaspace         *pVaspace;
    UvmGlobalState     *pUvm;
    UvmGpu             *pGpu;
    NvU32              splitFlags = 0;

    UVM_DBG_PRINT_UUID("\n UVM Test Split Pages",&params->gpuUuid);

    params->rmStatus = NV_ERR_INVALID_REQUEST;
    UVM_CHECK_VALID_PAGE_SIZE(params->newPgSize, params->rmStatus);

    pVaspace = (UvmVaspace *)filep->private_data;
    if (!pVaspace)
    {
        params->rmStatus = NV_ERR_INVALID_REQUEST;
        return params->rmStatus;
    }
    pUvm = uvmfull_get_global();
    down_read(&pUvm->lock);

    pGpu =  uvmfull_uuid_to_active_gpu(&params->gpuUuid);

    if (!pGpu || !pGpu->bValid)
    {
        params->rmStatus = NV_ERR_INVALID_REQUEST;
        goto cleanup_uvm;
    }

    down_read(&pVaspace->lock);
    if (pVaspace->filep != filep)
    {
        params->rmStatus = NV_ERR_INVALID_REQUEST;
        goto cleanup_vaspace;
    }


    if (params->bForceFixedPg)
        splitFlags = NV_UVM_SPLIT_MERGE_FLAGS_SPLIT_FIXED_PAGE_SIZE;
    else
        splitFlags = 0;

    params->rmStatus =
            uvmfull_test_split_gpupde_pte(pVaspace,
                                          pGpu->gpuId,
                                          params->targetVa,
                                          params->newPgSize,
                                          splitFlags);


cleanup_vaspace:
    up_read(&pVaspace->lock);
cleanup_uvm:
    up_read(&pUvm->lock);
    return params->rmStatus;
}

static NV_STATUS
_uvm_test_create_zfod_page_vidmem(UvmGpu *pGpu, NvU32 pattern)
{
    NV_STATUS status = NV_OK;
    UvmPbOps *pbOps = NULL;
    UvmPushbuffer *pb = NULL;
    struct page *cpuPage = NULL;
    NvU32 *zp = NULL;
    NvU64 dstAddr = 0;
    NvU32 i;
    UvmTracker tracker;
    uvm_init_tracker(&tracker);

    if (pGpu->bTestZfodPfnValid)
    {
        // Cannot overwrite the pattern that currently exists
        if(pGpu->testZfodPattern != pattern)
        {
            status = NV_ERR_INVALID_ARGUMENT;
            goto done;
        }
    }

    status = uvmfull_allocate_phys_vid_mem(pGpu, PAGE_SIZE_2M, &pGpu->testZfodPfn);
    if(status != NV_OK)
        return status;

    dstAddr = PFN_TO_PA(pGpu->testZfodPfn);
    // Now allocate a temporary cpu page
    cpuPage = alloc_page(NV_UVM_GFP_FLAGS);
    if(!cpuPage)
    {
        status = NV_ERR_NO_MEMORY;
        goto free_memory;
    }

    pGpu->testZfodPattern = pattern;
    zp = kmap(cpuPage);

    for (i = 0 ; i < UVM_CPU_MIN_PAGE_SIZE/sizeof(unsigned); i++)
        zp[i] = pGpu->testZfodPattern;


    pbOps = uvm_pbops_start_on_gpu(pGpu->gpuId);
    UVM_PANIC_ON(!pbOps);
    if (!pbOps)
    {
        status = NV_ERR_GENERIC;
        goto free_memory;
    }

    pb = uvm_pbops_gpu_pb(pbOps, pGpu->gpuId);
    UVM_PANIC_ON(!pb);
    if (!pb)
    {
        status = NV_ERR_GENERIC;
        goto free_memory;
    }

    for (i = 0; i < PAGE_SIZE_2M/UVM_CPU_MIN_PAGE_SIZE; i++)
    {
        uvm_push_cpu_to_gpu_copy(pb, PFN_TO_PA(page_to_pfn(cpuPage)),
                              dstAddr + i * UVM_CPU_MIN_PAGE_SIZE,
                              UVM_CPU_MIN_PAGE_SIZE,
                              NV_UVM_COPY_SRC_TYPE_PHYSICAL |
                              NV_UVM_COPY_DST_TYPE_PHYSICAL);
    }
    uvm_pbops_submit(pbOps, pGpu->gpuId, pb, &tracker);
    uvm_move_tracker(&tracker, pbOps->t);
    status = uvm_wait_for_tracker(&tracker);

    pGpu->bTestZfodPfnValid = NV_TRUE;
free_memory:
    if (status != NV_OK)
    {
        nvUvmInterfaceMemoryFreePa(pGpu->gpuHandle,
                                   PFN_TO_PA(pGpu->testZfodPfn));
        pGpu->bTestZfodPfnValid = NV_FALSE;
    }
done:
    if (cpuPage)
        __free_page(cpuPage);
    if (pbOps)
       uvm_pbops_end(pbOps);
    return status;
}

/**
 * @brief Test API used to create a zfod page
 *
 *        In case of 0FB we create a single CPU page
 *        In case of non-0FB, we create one GPU page per GPU
 *
 *        This function will have to be called for every GPU
 *        in both cases
 *
 * This function is serialized with global lock for now.
 */
NV_STATUS uvm_test_create_zfod_page(
                UVM_TEST_CREATE_ZFOD_PAGE_PARAMS* params,
                struct file* filep)
{
    NV_STATUS status = NV_OK;
    UvmGlobalState     *uvm;
    UvmGpu *pGpu;

    uvm = uvmfull_get_global();
    down_write(&uvm->lock);
    pGpu =  uvmfull_uuid_to_active_gpu(&params->gpuUuid);
    if (!pGpu || !pGpu->bValid)
    {
        status = NV_ERR_INVALID_ARGUMENT;
        goto done;
    }

    if (pGpu->bZeroFbGpu) // sysmem only
    {
        status = NV_ERR_INVALID_ARGUMENT;
        goto done;
    }
    else
        status = _uvm_test_create_zfod_page_vidmem(pGpu, params->pattern);

done:
    up_write(&uvm->lock);
    params->rmStatus = status;
    return status;
}

/**
 * @brief Enable testZfodPolicy for a given vaSpace
 */
NV_STATUS uvm_test_configure_zfod_policy(
                UVM_TEST_CONFIGURE_ZFOD_POLICY_PARAMS* params,
                struct file* filep)
{
    UvmGpu *pGpu;
    UvmVaspace *pVaspace = NULL;
    UvmGpuVaspace *pGpuVaspace = NULL;
    UvmGlobalState *pUvm = uvmfull_get_global();

    down_write(&pUvm->lock);
    pGpu =  uvmfull_uuid_to_active_gpu(&params->gpuUuid);
    if (!pGpu || !pGpu->bValid)
    {
        params->rmStatus = NV_ERR_INVALID_ARGUMENT;
        goto release_uvm_lock;
    }

    // Zfod page should be created prior to this call
    if(!pGpu->bTestZfodPfnValid)
    {
        params->rmStatus = NV_ERR_INVALID_ARGUMENT;
        goto release_uvm_lock;
    }

    pVaspace = (UvmVaspace *)filep->private_data;
    if (!pVaspace)
    {
        params->rmStatus = NV_ERR_INVALID_REQUEST;
        goto release_uvm_lock;
    }

    down_write(&pVaspace->lock);
    pGpuVaspace = &pVaspace->gpuVaspaces[pGpu->gpuId];
    pGpuVaspace->bTestZfodEnabled = params->bZfodEnabled;
    up_write(&pVaspace->lock);

    params->rmStatus = NV_OK;
release_uvm_lock:
    up_write(&pUvm->lock);
    return params->rmStatus;
}

NV_STATUS uvm_test_change_pte_mapping(
                UVM_TEST_CHANGE_PTE_MAPPING_PARAMS *params, struct file *filep)
{
    UvmVatContext query;
    UvmGpu *gpu = NULL;
    UvmVaspace *vaspace = NULL;
    Uvm2MbVaDesc* vaDesc = NULL;
    UvmPbOps *pbOps = NULL;
    UvmGpuPde* gpuPde;
    NvU32 pgSize;
    NvU64 vaddr;
    UvmGpuPte *pte = NULL;
    UvmPteHints *hints = NULL;
    UvmGpuVaspace *gpuVaspace = NULL;
    UvmGlobalState *uvm = uvmfull_get_global();

    vaddr = params->va;
    down_read(&uvm->lock);

    gpu =  uvmfull_uuid_to_active_gpu(&params->gpuUuid);
    if (!gpu || !gpu->bValid)
    {
        params->rmStatus = NV_ERR_INVALID_REQUEST;
        goto cleanup_uvm;
    }

    vaspace = (UvmVaspace *)filep->private_data;
    if (!vaspace)
    {
        params->rmStatus = NV_ERR_INVALID_REQUEST;
        goto cleanup_uvm;
    }

    down_read(&vaspace->lock);

    // Fetch descriptor, check if the gpu has an allocation, check if the va is
    // managed, change the pageSize.
    query.address    = vaddr;
    query.accessType = UvmVatAccess_GPUQuery;
    query.gpuId      = gpu->gpuId;
    query.vaspace    = vaspace;
    vaDesc = uvmfull_vat_acquire_leaf(&vaspace->vat, &query);
    if (!vaDesc)
    {
        params->rmStatus = NV_ERR_INVALID_REQUEST;
        goto cleanup_vaspace;
    }

    hints = &vaDesc->hints[GET_PDE_PTE_OFFSET(vaddr, UVM_CPU_MIN_PAGE_SHIFT)];
    if (!hints->managedVa)
    {
        params->rmStatus = NV_ERR_INVALID_REQUEST;
        goto cleanup_vaspace;
    }
    if ((params->targetMapping == UvmTestPteMapping_ReadWrite ||
         params->targetMapping == UvmTestPteMapping_ReadWriteAtomic) &&
        (hints->va_prot_write == 0x0))
    {
        params->rmStatus = NV_ERR_INVALID_REQUEST;
        goto cleanup_vaspace;
    }
    gpuVaspace = &vaspace->gpuVaspaces[gpu->gpuId];
    gpuPde = uvmfull_get_gpupde_in_vadesc(vaDesc, gpu->gpuId);
    if (!gpuPde)
    {
        params->rmStatus = NV_ERR_INVALID_REQUEST;
        goto cleanup_vaspace;
    }

    pte = uvmfull_get_gpupde_current_pte(gpuPde, vaddr, &pgSize);
    if (!pte || !pte->swcachevalid)
    {
        params->rmStatus = NV_OK;
        goto cleanup_vaspace;
    }

    switch(params->targetMapping)
    {
    case UvmTestPteMapping_Invalid:
        {
            UvmGpuPteUpdateParams gpteParams = {0};
            if (pte->valid == 0x0)
            {
                params->rmStatus = NV_OK;
                goto cleanup_vaspace;
            }
            pbOps = uvm_pbops_start(vaspace);
            uvm_pbops_reserve_acquire_space(pbOps, &vaDesc->tracker);
            gpteParams.updateSwFlags |= UVM_GPU_SW_PTE_UPDATE_FLAGS_PTE_VALID;
            gpteParams.valid = NV_FALSE;
            gpteParams.updateSwFlags |= UVM_GPU_SW_PTE_UPDATE_FLAGS_VOLATILE;
            gpteParams.vol = NV_FALSE;
            gpteParams.updateHwFlags |= UVM_GPU_HW_PTE_UPDATE_FLAGS_WRITE_PTE;

            // update the pfn in all the GPU ptes
            uvmfull_update_gpupde_ptes(vaDesc, gpu->gpuId,
                            gpuPde, vaddr, 1, pgSize,
                            &gpteParams,
                            NULL, pbOps,
                            (NvU8*)&gpuVaspace->pteDefaults.entry,
                            gpuVaspace->pteDefaults.entrySize,
                            NV_FALSE);
            uvm_pbops_submit_and_release(pbOps,
                                         NVBIT64(gpu->gpuId), &vaDesc->tracker);
            pbOps = NULL;
        }
        break;
    case UvmTestPteMapping_ReadOnly:
        {
            UvmPushbuffer *pb;
            UvmGpuPteUpdateParams gpteParams = {0};
            pbOps = uvm_pbops_start(vaspace);
            uvm_pbops_reserve_acquire_space(pbOps, &vaDesc->tracker);
            pb = uvm_pbops_gpu_pb(pbOps, gpu->gpuId);
            if (pte->valid == 0x1 && pte->readonly == 0x1)
            {
                params->rmStatus = NV_OK;
                goto cleanup_vaspace;
            }
            if (pte->aperture == UvmMemType_peer)
            {
                params->rmStatus = NV_ERR_INVALID_REQUEST;
                goto cleanup_vaspace;
            }
            if (pte->aperture == UvmMemType_vidmem && pte->valid == 0x0)
            {
                NvU32 cpuPages = pgSize / UVM_CPU_MIN_PAGE_SIZE;
                NvU32 offsetMapStart =
                        GET_PDE_PTE_OFFSET(UVM_ALIGN_ADDR(vaddr, pgSize),
                                               UVM_CPU_MIN_PAGE_SHIFT);
                // Mem-backing is allocated. Update it to latest data.
                params->rmStatus = uvmfull_unmap_copy_cpu_to_gpu(vaDesc, gpuPde,
                                                 pb, offsetMapStart, cpuPages);
                if (params->rmStatus != NV_OK)
                    goto cleanup_vaspace;
            }

            if (pte->aperture == UvmMemType_sysmem_coh ||
                pte->aperture == UvmMemType_sysmem_ncoh)
            {
                gpteParams.updateSwFlags |= UVM_GPU_SW_PTE_UPDATE_FLAGS_VOLATILE;
                gpteParams.vol = NV_TRUE;
            }

            gpteParams.updateSwFlags |= UVM_GPU_SW_PTE_UPDATE_FLAGS_IMPLICIT_PFN;
            gpteParams.updateSwFlags |= UVM_GPU_SW_PTE_UPDATE_FLAGS_PFN_VALID;
            gpteParams.updateSwFlags |= UVM_GPU_SW_PTE_UPDATE_FLAGS_PFN;
            gpteParams.pfnvalid = NV_TRUE;

            gpteParams.updateSwFlags |= UVM_GPU_SW_PTE_UPDATE_FLAGS_SWVALID;
            gpteParams.swcachevalid = NV_TRUE;

            gpteParams.updateSwFlags |= UVM_GPU_SW_PTE_UPDATE_FLAGS_APERTURE;
            gpteParams.aperture = pte->aperture;

            gpteParams.updateSwFlags |= UVM_GPU_SW_PTE_UPDATE_FLAGS_ATOMIC;
            gpteParams.atomic = NV_FALSE;
            gpteParams.updateSwFlags |= UVM_GPU_SW_PTE_UPDATE_FLAGS_PTE_VALID;
            gpteParams.valid = NV_TRUE;
            gpteParams.updateSwFlags |= UVM_GPU_SW_PTE_UPDATE_FLAGS_READ_ONLY;
            gpteParams.readonly = NV_TRUE;
            gpteParams.updateHwFlags |= UVM_GPU_HW_PTE_UPDATE_FLAGS_WRITE_PTE;

            // update the pfn in all the GPU ptes
            uvmfull_update_gpupde_ptes(vaDesc, gpu->gpuId,
                            gpuPde, vaddr, 1, pgSize,
                            &gpteParams,
                            NULL, pbOps,
                            (NvU8*)&gpuVaspace->pteDefaults.entry,
                            gpuVaspace->pteDefaults.entrySize,
                            NV_FALSE);
            uvm_pbops_submit_and_release(pbOps,
                                         NVBIT64(gpu->gpuId), &vaDesc->tracker);
            pbOps = NULL;
        }
        break;
    case UvmTestPteMapping_ReadWrite:
        {
            UvmPushbuffer *pb;
            UvmGpuPteUpdateParams gpteParams = {0};
            pbOps = uvm_pbops_start(vaspace);
            uvm_pbops_reserve_acquire_space(pbOps, &vaDesc->tracker);
            pb = uvm_pbops_gpu_pb(pbOps, gpu->gpuId);
            if (pte->valid == 0x1 && pte->readonly == 0x0 && pte->atomic == 0x0)
            {
                params->rmStatus = NV_OK;
                goto cleanup_vaspace;
            }
            if (pte->aperture == UvmMemType_peer)
            {
                params->rmStatus = NV_ERR_INVALID_REQUEST;
                goto cleanup_vaspace;
            }
            if (pte->aperture == UvmMemType_vidmem && pte->valid == 0x0)
            {
                NvU32 cpuPages = pgSize / UVM_CPU_MIN_PAGE_SIZE;
                NvU32 offsetMapStart =
                        GET_PDE_PTE_OFFSET(UVM_ALIGN_ADDR(vaddr, pgSize),
                                               UVM_CPU_MIN_PAGE_SHIFT);
                // Mem-backing is allocated. Update it to latest data.
                params->rmStatus = uvmfull_unmap_copy_cpu_to_gpu(vaDesc, gpuPde,
                                                  pb, offsetMapStart, cpuPages);
                if (params->rmStatus != NV_OK)
                    goto cleanup_vaspace;
            }

            if (pte->aperture == UvmMemType_sysmem_coh ||
                pte->aperture == UvmMemType_sysmem_ncoh)
            {
                gpteParams.updateSwFlags |= UVM_GPU_SW_PTE_UPDATE_FLAGS_VOLATILE;
                gpteParams.vol = NV_TRUE;
            }

            gpteParams.updateSwFlags |= UVM_GPU_SW_PTE_UPDATE_FLAGS_IMPLICIT_PFN;
            gpteParams.updateSwFlags |= UVM_GPU_SW_PTE_UPDATE_FLAGS_PFN_VALID;
            gpteParams.updateSwFlags |= UVM_GPU_SW_PTE_UPDATE_FLAGS_PFN;
            gpteParams.pfnvalid = NV_TRUE;

            gpteParams.updateSwFlags |= UVM_GPU_SW_PTE_UPDATE_FLAGS_SWVALID;
            gpteParams.swcachevalid = NV_TRUE;

            gpteParams.updateSwFlags |= UVM_GPU_SW_PTE_UPDATE_FLAGS_APERTURE;
            gpteParams.aperture = pte->aperture;

            gpteParams.updateSwFlags |= UVM_GPU_SW_PTE_UPDATE_FLAGS_READ_ONLY;
            gpteParams.readonly = NV_FALSE;
            gpteParams.updateSwFlags |= UVM_GPU_SW_PTE_UPDATE_FLAGS_ATOMIC;
            gpteParams.atomic = NV_FALSE;
            gpteParams.updateSwFlags |= UVM_GPU_SW_PTE_UPDATE_FLAGS_PTE_VALID;
            gpteParams.valid = NV_TRUE;
            gpteParams.updateHwFlags |= UVM_GPU_HW_PTE_UPDATE_FLAGS_WRITE_PTE;

            // update the pfn in all the GPU ptes
            uvmfull_update_gpupde_ptes(vaDesc, gpu->gpuId,
                            gpuPde, vaddr, 1, pgSize,
                            &gpteParams,
                            NULL, pbOps,
                            (NvU8*)&gpuVaspace->pteDefaults.entry,
                            gpuVaspace->pteDefaults.entrySize,
                            NV_FALSE);
            uvm_pbops_submit_and_release(pbOps,
                                         NVBIT64(gpu->gpuId), &vaDesc->tracker);
            pbOps = NULL;
        }
        break;
    case UvmTestPteMapping_ReadWriteAtomic:
        {
            UvmPushbuffer *pb;
            UvmGpuPteUpdateParams gpteParams = {0};
            pbOps = uvm_pbops_start(vaspace);
            uvm_pbops_reserve_acquire_space(pbOps, &vaDesc->tracker);
            pb = uvm_pbops_gpu_pb(pbOps, gpu->gpuId);
            if (pte->valid == 0x1 && pte->atomic == 0x1)
            {
                params->rmStatus = NV_OK;
                goto cleanup_vaspace;
            }
            if (pte->aperture == UvmMemType_peer)
            {
                params->rmStatus = NV_ERR_INVALID_REQUEST;
                goto cleanup_vaspace;
            }
            if (pte->aperture == UvmMemType_vidmem && pte->valid == 0x0)
            {
                NvU32 cpuPages = pgSize / UVM_CPU_MIN_PAGE_SIZE;
                NvU32 offsetMapStart =
                        GET_PDE_PTE_OFFSET(UVM_ALIGN_ADDR(vaddr, pgSize),
                                               UVM_CPU_MIN_PAGE_SHIFT);
                // Mem-backing is allocated. Update it to latest data.
                params->rmStatus = uvmfull_unmap_copy_cpu_to_gpu(vaDesc, gpuPde,
                                                pb, offsetMapStart, cpuPages);
                if (params->rmStatus != NV_OK)
                    goto cleanup_vaspace;
            }

            if (pte->aperture == UvmMemType_sysmem_coh ||
                pte->aperture == UvmMemType_sysmem_ncoh)
            {
                gpteParams.updateSwFlags |= UVM_GPU_SW_PTE_UPDATE_FLAGS_VOLATILE;
                gpteParams.vol = NV_TRUE;
            }

            gpteParams.updateSwFlags |= UVM_GPU_SW_PTE_UPDATE_FLAGS_IMPLICIT_PFN;
            gpteParams.updateSwFlags |= UVM_GPU_SW_PTE_UPDATE_FLAGS_PFN_VALID;
            gpteParams.updateSwFlags |= UVM_GPU_SW_PTE_UPDATE_FLAGS_PFN;
            gpteParams.pfnvalid = NV_TRUE;

            gpteParams.updateSwFlags |= UVM_GPU_SW_PTE_UPDATE_FLAGS_SWVALID;
            gpteParams.swcachevalid = NV_TRUE;

            gpteParams.updateSwFlags |= UVM_GPU_SW_PTE_UPDATE_FLAGS_APERTURE;
            gpteParams.aperture = pte->aperture;

            gpteParams.updateSwFlags |= UVM_GPU_SW_PTE_UPDATE_FLAGS_READ_ONLY;
            gpteParams.readonly = NV_FALSE;
            gpteParams.updateSwFlags |= UVM_GPU_SW_PTE_UPDATE_FLAGS_ATOMIC;
            gpteParams.atomic = NV_TRUE;
            gpteParams.updateSwFlags |= UVM_GPU_SW_PTE_UPDATE_FLAGS_PTE_VALID;
            gpteParams.valid = NV_TRUE;
            gpteParams.updateHwFlags |= UVM_GPU_HW_PTE_UPDATE_FLAGS_WRITE_PTE;

            // update the pfn in all the GPU ptes
            uvmfull_update_gpupde_ptes(vaDesc, gpu->gpuId,
                            gpuPde, vaddr, 1, pgSize,
                            &gpteParams,
                            NULL, pbOps,
                            (NvU8*)&gpuVaspace->pteDefaults.entry,
                            gpuVaspace->pteDefaults.entrySize,
                            NV_FALSE);
            uvm_pbops_submit_and_release(pbOps,
                                         NVBIT64(gpu->gpuId), &vaDesc->tracker);
            pbOps = NULL;
        }
        break;
    default:
        params->rmStatus = NV_OK;
        goto cleanup_vaspace;
    }

cleanup_vaspace:
    if (params->rmStatus != NV_OK && pbOps)
        uvm_pbops_end(pbOps);
    if (vaDesc)
        uvmfull_vat_release_leaf(&vaspace->vat, &query);

    up_read(&vaspace->lock);
cleanup_uvm:
    up_read(&uvm->lock);
    return params->rmStatus;
}

NV_STATUS uvm_test_force_page_size(
                UVM_TEST_FORCE_PAGE_SIZE_PARAMS *params, struct file *filep)
{
    NvU32 pageSizeNew;
    NvU32 curPageSize;
    UvmVatContext query;
    UvmGpu *gpu = NULL;
    UvmVaspace *vaspace = NULL;
    Uvm2MbVaDesc* vaDesc = NULL;
    UvmPbOps *pbOps = NULL;
    UvmGpuPde* gpuPde;
    NvU64 vaddr;
    NvU64 vaddrAligned;
    UvmGpuPte *pte = NULL;
    UvmGlobalState *uvm = uvmfull_get_global();

    vaddr = params->va;
    pageSizeNew = params->pageSize;
    if (pageSizeNew != PAGE_SIZE_4K &&
        pageSizeNew != PAGE_SIZE_64K &&
        pageSizeNew != PAGE_SIZE_2M)
    {
        params->rmStatus = NV_ERR_INVALID_REQUEST;
        return params->rmStatus;
    }

    down_read(&uvm->lock);

    gpu =  uvmfull_uuid_to_active_gpu(&params->gpuUuid);
    if (!gpu || !gpu->bValid)
    {
        params->rmStatus = NV_ERR_INVALID_REQUEST;
        goto cleanup_uvm;
    }

    vaspace = (UvmVaspace *)filep->private_data;
    if (!vaspace)
    {
        params->rmStatus = NV_ERR_INVALID_REQUEST;
        goto cleanup_uvm;
    }

    down_read(&vaspace->lock);

    // Fetch descriptor, check if the gpu has an allocation, check if the va is
    // managed, change the pageSize.
    query.address    = vaddr;
    query.accessType = UvmVatAccess_GPUQuery;
    query.gpuId      = gpu->gpuId;
    query.vaspace    = vaspace;
    vaDesc = uvmfull_vat_acquire_leaf(&vaspace->vat, &query);
    if (!vaDesc)
    {
        params->rmStatus = NV_ERR_INVALID_REQUEST;
        goto cleanup_vaspace;
    }

    gpuPde = uvmfull_get_gpupde_in_vadesc(vaDesc, gpu->gpuId);
    if (!gpuPde)
    {
        params->rmStatus = NV_ERR_INVALID_REQUEST;
        goto cleanup_vaspace;
    }

    vaddrAligned = UVM_ALIGN_ADDR(vaddr, pageSizeNew);
    pte = uvmfull_get_gpupde_current_pte(gpuPde, vaddr, &curPageSize);
    if (!pte || !pte->swcachevalid)
    {
        params->rmStatus = NV_OK;
        goto cleanup_vaspace;
    }

    if (pageSizeNew == curPageSize)
    {
        params->rmStatus = NV_OK;
        goto cleanup_vaspace;
    }
    else if (pageSizeNew > curPageSize)  // Merge required
    {
        NvU32 cpuPages, offsetMapStart;
        NvU32 suggestedPageSize;
        UvmMemType curAperture = pte->aperture;
        UvmPushbuffer *pb = NULL;
        pbOps = uvm_pbops_start(vaspace);
        uvm_pbops_reserve_acquire_space(pbOps, &vaDesc->tracker);
        pb = uvm_pbops_gpu_pb(pbOps, gpu->gpuId);

        // Check to see if hints match across the va range
        uvmfull_check_page_split(vaDesc, vaddrAligned,
                                pageSizeNew, &suggestedPageSize);
        if (suggestedPageSize < pageSizeNew)
        {
            params->rmStatus = NV_ERR_INVALID_REQUEST;
            goto cleanup_vaspace;
        }

        // For now only support vidmem
        if (curAperture != UvmMemType_vidmem)
        {
            params->rmStatus = NV_ERR_INVALID_REQUEST;
            goto cleanup_vaspace;
        }
        cpuPages = pageSizeNew / UVM_CPU_MIN_PAGE_SIZE;
        offsetMapStart = GET_PDE_PTE_OFFSET(UVM_ALIGN_ADDR(vaddr, pageSizeNew),
                                       UVM_CPU_MIN_PAGE_SHIFT);
        // Unmap cpu if required
        if (!gpuPde->memBackingPresent)
        {
            params->rmStatus = uvmfull_alloc_backing_gpu_pages(gpu, gpuPde,
                                                     PAGE_SIZE_2M, 1);
            if (params->rmStatus != NV_OK)
                goto cleanup_vaspace;

            uvm_push_memzero_vidmem(gpu->gpuId,
                                    pb,
                                    PFN_TO_PA(gpuPde->gpuFirstPfn),
                                    PAGE_SIZE_2M);
        }

        // Mem-backing is allocated. Update it to latest data.
        params->rmStatus = uvmfull_unmap_copy_cpu_to_gpu(vaDesc, gpuPde, pb,
                                                  offsetMapStart, cpuPages);
        if (params->rmStatus != NV_OK)
            goto cleanup_vaspace;
        params->rmStatus = uvmfull_merge_gpupde_pte(vaspace, gpu->gpuId, vaDesc,
                        gpuPde, vaddrAligned, pageSizeNew, pbOps);
        if (params->rmStatus != NV_OK)
            goto cleanup_vaspace;
        params->rmStatus = uvm_pbops_submit_and_release(pbOps,
                                        NVBIT64(gpu->gpuId), &vaDesc->tracker);
        pbOps = NULL;
    }
    else  // pageSizeNew < curPageSize. Split required
    {
        NvU32 splitFlags = NV_UVM_SPLIT_MERGE_FLAGS_SPLIT_FIXED_PAGE_SIZE;
        pbOps = uvm_pbops_start(vaspace);
        uvm_pbops_reserve_acquire_space(pbOps, &vaDesc->tracker);

        // Split is always possible in any aperture.
        // If the requested size is less than current, va properties match
        params->rmStatus =
                uvmfull_split_gpupde_pte(vaspace, gpu->gpuId, vaDesc, gpuPde,
                                 vaddrAligned, pageSizeNew, splitFlags, pbOps);
        if (params->rmStatus != NV_OK)
            goto cleanup_vaspace;
        params->rmStatus = uvm_pbops_submit_and_release(pbOps,
                                        NVBIT64(gpu->gpuId), &vaDesc->tracker);
        pbOps = NULL;
    }

cleanup_vaspace:
    if (params->rmStatus != NV_OK && pbOps)
        uvm_pbops_end(pbOps);
    if (vaDesc)
        uvmfull_vat_release_leaf(&vaspace->vat, &query);

    up_read(&vaspace->lock);
cleanup_uvm:
    up_read(&uvm->lock);
    return params->rmStatus;
}

NV_STATUS _test_explicit_va_migrate(UvmVaspace *vaspace, NvU64 startVa,
                                   NvU64 length, NvBool bMigrateToCpu,
                                   NvBool bRemap, UvmGpu *gpu)
{
    UvmVatContext query;
    Uvm2MbVaDesc* vaDesc = NULL;
    UvmPbOps *pbOps = NULL;
    UvmGpuPde* gpuPde;
    UvmPteHints *hints = NULL;
    NV_STATUS status = NV_OK;
    NvU64 vaddr;
    UvmGpuVaspace *gpuVaspace = NULL;
    UvmPushbuffer *pb;
    NvU64 gpuSubmitMask = 0;

    if (bMigrateToCpu)
    {
        query.address    = startVa;
        query.accessType = UvmVatAccess_GenericQuery;
        query.gpuId      = 0;
        query.vaspace    = vaspace;
    }
    else  // migrate to a gpu
    {
        UVM_PANIC_ON(!gpu || !gpu->bValid);

        // Fetch descriptor assuming this may be a first touch on the target
        // processor.
        query.address    = startVa;
        query.accessType = UvmVatAccess_GPUTouch;
        query.gpuId      = gpu->gpuId;
        query.vaspace    = vaspace;
    }

    vaDesc = uvmfull_vat_acquire_leaf(&vaspace->vat, &query);
    if (!vaDesc)
    {
        status = NV_ERR_INVALID_REQUEST;
        goto cleanup;
    }

    if (!vaDesc->nValid4k)
    {
        status = NV_ERR_INVALID_REQUEST;
        goto cleanup;
    }

    pbOps = uvm_pbops_start(vaspace);
    if (!pbOps)
    {
        status = NV_ERR_NO_MEMORY;
        goto cleanup;
    }
    uvm_pbops_reserve_acquire_space(pbOps, &vaDesc->tracker);

    vaddr = startVa;
    while (vaddr < startVa + length)
    {
        NvU32 pgSize;
        UvmGpuPte *gpuPte;
        NvU32 stride = UVM_CPU_MIN_PAGE_SIZE;
        NvU32 hintId = GET_PDE_PTE_OFFSET(vaddr, UVM_CPU_MIN_PAGE_SHIFT);
        hints = &vaDesc->hints[hintId];
        if (hints->managedVa == 0x0)
        {
            status = NV_ERR_INVALID_REQUEST;
            goto cleanup;
        }

        if (bMigrateToCpu)
        {
            desc_for_each_gpu_pde(gpuPde, vaDesc)
            {
                NvU32 i;
                struct page *pg;
                UvmGpuPteUpdateParams gpteParams = {0};
                if (!gpuPde->memBackingPresent)
                    continue;

                gpuPte = uvmfull_get_gpupde_current_pte(gpuPde, vaddr, &pgSize);
                if (!gpuPte || !gpuPte->swcachevalid)
                    continue;

                stride = pgSize;
                if (!gpuPte->valid)
                    continue;

                // Only choose the gpu that has the backing
                if (gpuPte->aperture != UvmMemType_vidmem)
                    continue;

                gpuVaspace = &vaspace->gpuVaspaces[gpuPde->hostGpuId];
                // Now unmap the gpu pde and copy the data to cpu
                gpteParams.updateSwFlags |= UVM_GPU_SW_PTE_UPDATE_FLAGS_PTE_VALID;
                gpteParams.valid = NV_FALSE;
                gpteParams.updateSwFlags |= UVM_GPU_SW_PTE_UPDATE_FLAGS_VOLATILE;
                gpteParams.vol = NV_FALSE;
                // Break zfod in any case.
                gpteParams.updateSwFlags |= UVM_GPU_SW_PTE_UPDATE_FLAGS_ZFOD_MAPPED;
                gpteParams.zfodMapped = NV_FALSE;
                // Actually update the PTE
                gpteParams.updateHwFlags |= UVM_GPU_HW_PTE_UPDATE_FLAGS_WRITE_PTE;
                // update the pfn in all the GPU ptes
                gpuSubmitMask |= NVBIT64(gpuPde->hostGpuId);
                pb = uvm_pbops_gpu_pb(pbOps, gpuPde->hostGpuId);
                uvmfull_update_gpupde_ptes(vaDesc, gpuPde->hostGpuId,
                                        gpuPde, vaddr, 1, pgSize,
                                        &gpteParams,
                                        NULL, pbOps,
                                        (NvU8*)&gpuVaspace->pteDefaults.entry,
                                        gpuVaspace->pteDefaults.entrySize,
                                        NV_FALSE);
                // Now push a copy to cpu
                if (pgSize >= UVM_CPU_MIN_PAGE_SIZE)
                {
                    for (i = 0; i < pgSize/UVM_CPU_MIN_PAGE_SIZE; i++)
                    {
                        // Check if cpu pfn is allocated
                        if (!vaDesc->hints[hintId + i].cpuPfnValid)
                        {
                            pg = alloc_pages(NV_UVM_GFP_FLAGS |
                                             GFP_HIGHUSER |
                                             __GFP_ZERO, 0);
                            if (!pg)
                            {
                                status = NV_ERR_NO_MEMORY;
                                goto cleanup;
                            }

                            vaDesc->hints[hintId + i].cpuPfn = page_to_pfn(pg);
                            vaDesc->hints[hintId + i].cpuPfnValid = 1;
                            vaDesc->hints[hintId + i].cpuPfnDirty = 0;
                            vaDesc->hints[hintId + i].cpuPfnMapped = 0;
                        }

                        uvm_push_gpu_to_cpu_copy(pb, PFN_TO_PA(gpuPte->pfn),
                                    PFN_TO_PA(vaDesc->hints[hintId + i].cpuPfn),
                                    UVM_CPU_MIN_PAGE_SIZE,
                                    NV_UVM_COPY_SRC_TYPE_PHYSICAL |
                                    NV_UVM_COPY_DST_TYPE_PHYSICAL);
                    }
                }
                else  // gpu page size is less than cpu page size
                {
                    // Not implemented
                    UVM_PANIC_ON(pgSize < UVM_CPU_MIN_PAGE_SIZE);
                }
            }
        }
        else  // migrate to a gpu
        {
            UvmGpuPte *curGpuPte;
            NvU32 curPgSize;
            UvmPushbuffer *pb;
            NvU32 cpuPages;
            NvU32 offsetMapStart;
            UvmGpuPteUpdateParams gpteParams = {0};

            gpuVaspace = &vaspace->gpuVaspaces[gpu->gpuId];
            gpuPde = uvmfull_get_gpupde_in_vadesc(vaDesc, gpu->gpuId);
            if (!gpuPde)
            {
                status = NV_ERR_INVALID_REQUEST;
                goto cleanup;
            }
            status = uvmfull_activate_gpupte(vaDesc, gpuPde,
                                NV_TRUE, NV_TRUE, gpu->gpuId,
                                uvm_pbops_gpu_pb(pbOps, gpu->gpuId));
            if (status != NV_OK)
            {
                status = NV_ERR_NO_MEMORY;
                goto cleanup;
            }

            pb = uvm_pbops_gpu_pb(pbOps, gpuPde->hostGpuId);
            UVM_PANIC_ON(!pb);
            gpuSubmitMask |= NVBIT64(gpuPde->hostGpuId);
            curGpuPte = uvmfull_get_gpupde_current_pte(gpuPde, vaddr, &curPgSize);
            if (curGpuPte->swcachevalid)
            {
                if (curGpuPte->valid)
                    stride = curPgSize;
            }
            else
            {
                curPgSize = PAGE_SIZE_64K;   // start with policy
                stride = curPgSize;
                cpuPages = curPgSize / UVM_CPU_MIN_PAGE_SIZE;
                offsetMapStart =
                            GET_PDE_PTE_OFFSET(UVM_ALIGN_ADDR(vaddr, curPgSize),
                                               UVM_CPU_MIN_PAGE_SHIFT);
                if (!gpuPde->memBackingPresent)
                {
                    status = uvmfull_alloc_backing_gpu_pages(gpu, gpuPde,
                                                             PAGE_SIZE_2M, 1);
                    if (status != NV_OK)
                    {
                        goto cleanup;
                    }
                    uvm_push_memzero_vidmem(gpuPde->hostGpuId,
                                            pb,
                                            PFN_TO_PA(gpuPde->gpuFirstPfn),
                                            PAGE_SIZE_2M);
                }

                // Mem-backing is allocated. Update it to latest data.
                status = uvmfull_unmap_copy_cpu_to_gpu(vaDesc,
                                                     gpuPde, pb, offsetMapStart,
                                                     cpuPages);
                if (status != NV_OK)
                {
                    status = NV_ERR_INVALID_STATE;
                    goto cleanup;
                }
                // Implicit pfn means that the 2Mb allocation stored in the pde
                // will be used for vidmem. In sysmem; the cpuPfn stored will
                // be used.
                gpteParams.updateSwFlags |=
                                    UVM_GPU_SW_PTE_UPDATE_FLAGS_IMPLICIT_PFN;

                 // Also if vidmem, always mark atomic
                gpteParams.updateSwFlags |= UVM_GPU_SW_PTE_UPDATE_FLAGS_ATOMIC;
                gpteParams.atomic = NV_TRUE;

                gpteParams.updateSwFlags |= UVM_GPU_SW_PTE_UPDATE_FLAGS_PFN_VALID;
                gpteParams.updateSwFlags |= UVM_GPU_SW_PTE_UPDATE_FLAGS_PFN;
                gpteParams.pfnvalid = NV_TRUE;

                gpteParams.updateSwFlags |= UVM_GPU_SW_PTE_UPDATE_FLAGS_SWVALID;
                gpteParams.swcachevalid = NV_TRUE;

                gpteParams.updateSwFlags |= UVM_GPU_SW_PTE_UPDATE_FLAGS_APERTURE;
                gpteParams.aperture = UvmMemType_vidmem;

                gpteParams.updateSwFlags |= UVM_GPU_SW_PTE_UPDATE_FLAGS_PTE_VALID;
                gpteParams.valid = NV_TRUE;

                // Actually update the PTE
                gpteParams.updateHwFlags |= UVM_GPU_HW_PTE_UPDATE_FLAGS_WRITE_PTE;
                // update the pfn in all the GPU ptes
                status = uvmfull_update_gpupde_ptes(vaDesc, gpu->gpuId,
                                gpuPde, vaddr, 1, curPgSize,
                                &gpteParams, NULL, pbOps,
                                (NvU8*)&gpuVaspace->pteDefaults.entry,
                                gpuVaspace->pteDefaults.entrySize,
                                NV_FALSE);

                if (status != NV_OK)
                {
                    WARN_ON(status);
                    goto cleanup;
                }
            }
        }

        // update the vaddr based on stride. Stride would be the gpu page size
        // in case the mapping was on gpu.
        vaddr += stride;
    }

cleanup:
    if (pbOps)
    {
        if (status != NV_OK)
            uvm_pbops_end(pbOps);
        else
            status = uvm_pbops_submit_and_release(pbOps,
                                        gpuSubmitMask, &vaDesc->tracker);
    }
    if (vaDesc)
        uvmfull_vat_release_leaf(&vaspace->vat, &query);

    return status;
}

NV_STATUS uvm_test_explicit_va_migrate(
                UVM_TEST_EXPLICIT_VA_MIGRATE_PARAMS *params, struct file *filep)
{
    UvmGpu *gpu = NULL;
    UvmVaspace *vaspace = NULL;
    NvU64 vaStart, vaEndExcl, vaddr, length;
    NV_STATUS status = NV_OK;
    UvmGlobalState *uvm = uvmfull_get_global();

    vaStart = UVM_ALIGN_ADDR(params->va, UVM_CPU_MIN_PAGE_SIZE);
    length = UVM_ALIGN_UP(params->va + params->length, UVM_CPU_MIN_PAGE_SIZE) -
                vaStart;
    vaEndExcl = vaStart + length;

    down_read(&uvm->lock);
    vaspace = (UvmVaspace *)filep->private_data;
    if (!vaspace)
    {
        params->rmStatus = NV_ERR_INVALID_REQUEST;
        goto cleanup_uvm;
    }
    down_read(&vaspace->lock);

    // Check gpu existence if gpu is dst.
    if (!params->bMigrateToCpu)
    {
        gpu =  uvmfull_uuid_to_active_gpu(&params->targetGpuUuid);
        if (!gpu || !gpu->bValid)
        {
            params->rmStatus = NV_ERR_INVALID_REQUEST;
            goto cleanup_vaspace;
        }
    }

    vaddr = vaStart;
    while (vaddr < vaEndExcl)
    {
        NvU64 end, len;
        end = UVM_ALIGN_ADDR(vaEndExcl - 1, UVM_VA_DESC_SIZE) ==
              UVM_ALIGN_ADDR(vaddr, UVM_VA_DESC_SIZE) ?
              vaEndExcl :
              UVM_ALIGN_UP(vaddr + 1, UVM_VA_DESC_SIZE);
        len = end - vaddr;
        status = _test_explicit_va_migrate(vaspace, vaddr, len,
                                            params->bMigrateToCpu,
                                            params->bRemapAfterMigration, gpu);
        if (status != NV_OK)
        {
            params->rmStatus = status;
            goto cleanup_vaspace;
        }
        vaddr += len;
    }

cleanup_vaspace:
    up_read(&vaspace->lock);

cleanup_uvm:
    up_read(&uvm->lock);
    return params->rmStatus;
}

// memzero dst followed by map based data copy from pde-backing to dstPa.
static void
_test_move_gpu_pde_data(UvmGpu* gpu, Uvm2MbVaDesc *vaDesc, UvmPbOps *pbOps,
                        UvmGpuPde *pde, unsigned long * bitmap, NvU64 dstPa,
                        UvmMemType dstAperture)
{
    NvU32 i = 0;
    UvmPushbuffer *pb = uvm_pbops_gpu_pb(pbOps, gpu->gpuId);
    UvmGpuPte *pte;
    uvm_push_memzero_vidmem(gpu->gpuId, pb, dstPa, UVM_VA_DESC_SIZE);

    while (i < NUM_4K_PER_2M)
    {
        NvU64 srcPa;
        NvU32 pgSize = PAGE_SIZE_4K;
        NvU64 vaddr = vaDesc->addr + PAGE_SIZE_4K * i;
        if (test_bit(i, bitmap))
        {
            pte = uvmfull_get_gpupde_current_pte(pde, vaddr, &pgSize);
            WARN_ON(!pte || !pte->swcachevalid || !pte->pfnvalid);
            if (!pte || !pte->swcachevalid || !pte->pfnvalid)
            {
                i += pgSize/PAGE_SIZE_4K;
                continue;
            }

            srcPa = PFN_TO_PA(pte->pfn);
            uvm_push_copy(pb, srcPa, dstPa + i*PAGE_SIZE_4K, pgSize,
                          NV_UVM_COPY_SRC_TYPE_PHYSICAL |
                          NV_UVM_COPY_DST_TYPE_PHYSICAL,
                          pte->aperture,
                          dstAperture);
        }
        i += pgSize/PAGE_SIZE_4K;
    }
}

static NV_STATUS
_test_map_gpu_pde_data(UvmGpu *gpu, Uvm2MbVaDesc *vaDesc, UvmPbOps *pbOps,
                       UvmGpuPde *gpuPde, unsigned long *bitmap)
{
    NvU32 i = 0;
    UvmGpuPte *pte;
    NV_STATUS status = NV_OK;
    UvmGpuVaspace *gpuVaspace = &vaDesc->pVaspace->gpuVaspaces[gpu->gpuId];

    while (i < NUM_4K_PER_2M)
    {
        UvmGpuPteUpdateParams pteParams;
        NvU32 pgSize = PAGE_SIZE_4K;
        NvU64 vaddr = vaDesc->addr + PAGE_SIZE_4K * i;
        if (test_bit(i, bitmap))
        {
            pte = uvmfull_get_gpupde_current_pte(gpuPde, vaddr, &pgSize);
            WARN_ON(!pte || !pte->swcachevalid || !pte->pfnvalid);
            if (!pte || !pte->swcachevalid || !pte->pfnvalid)
                continue;

            // TODO: Map the pde now
            memset(&pteParams, 0x0, sizeof(pteParams));
            pteParams.updateSwFlags |= UVM_GPU_SW_PTE_UPDATE_FLAGS_IMPLICIT_PFN;
            pteParams.updateSwFlags |= UVM_GPU_SW_PTE_UPDATE_FLAGS_PFN;

            pteParams.updateSwFlags |= UVM_GPU_SW_PTE_UPDATE_FLAGS_APERTURE;
            pteParams.aperture = UvmMemType_vidmem;

            pteParams.updateSwFlags |= UVM_GPU_SW_PTE_UPDATE_FLAGS_PTE_VALID;
            pteParams.valid = NV_TRUE;

            // Actually update the PTE
            pteParams.updateHwFlags |= UVM_GPU_HW_PTE_UPDATE_FLAGS_WRITE_PTE;
            status = uvmfull_update_gpupde_ptes(vaDesc, gpu->gpuId,
                            gpuPde, vaddr, 1, pgSize,
                            &pteParams, NULL, pbOps,
                            (NvU8*)&gpuVaspace->pteDefaults.entry,
                            gpuVaspace->pteDefaults.entrySize,
                            NV_FALSE);
            if (status != NV_OK)
            {
                WARN_ON(status);
            }
        }
        i += pgSize/PAGE_SIZE_4K;
    }
    return status;
}

NV_STATUS uvm_test_swap_gpu_pages(
                UVM_TEST_SWAP_GPU_PAGES_PARAMS *params, struct file *filep)
{
    const unsigned RANGES = 2;
    const unsigned NUM_BYTES = (NUM_4K_PER_2M/8);  //a bit for each 4k
    const unsigned SIZE_LONG = sizeof(unsigned long);
    const unsigned LONGS_2M = ((NUM_BYTES + SIZE_LONG - 1)/SIZE_LONG) * SIZE_LONG;
    NvU64           va[RANGES];
    UvmVatContext   query[RANGES];
    Uvm2MbVaDesc*   vaDesc[RANGES];
    UvmGpuPde*      gpuPde[RANGES];
    unsigned long   *pBitmap[RANGES];
    NvU64 pageSize;
    UvmGpu *gpu = NULL;
    UvmVaspace *vaspace = NULL;
    UvmPbOps *pbOps = NULL;
    NvU64 tempPfn;
    NvU32 i;
    NvU64 temp = 0;


    // num of bits for 4k mappings
    UvmGlobalState *uvm = uvmfull_get_global();

    for (i = 0; i < RANGES; i++)
        pBitmap[i] = NULL;

    vaDesc[0] = vaDesc[1] = NULL;
    // Align to smallest gpu page size
    if (params->vaA < params->vaB)
    {
        va[0] = UVM_ALIGN_ADDR(params->vaA, PAGE_SIZE_4K);
        va[1] = UVM_ALIGN_ADDR(params->vaB, PAGE_SIZE_4K);
    }
    else
    {
        va[0] = UVM_ALIGN_ADDR(params->vaB, PAGE_SIZE_4K);
        va[1] = UVM_ALIGN_ADDR(params->vaA, PAGE_SIZE_4K);
    }

    pageSize = params->pageSize;
    // Requirement that the range falls inside a 2mb desc and the Va-s should be
    // from different descriptors. Swapping the backing memory would always be
    // done at a 2mb granularity given the way backing memory is maintained.
    if (UVM_ALIGN_ADDR(va[0], UVM_VA_DESC_SIZE) !=
        UVM_ALIGN_ADDR(va[0] + pageSize - 1, UVM_VA_DESC_SIZE) ||
        UVM_ALIGN_ADDR(va[1], UVM_VA_DESC_SIZE) !=
        UVM_ALIGN_ADDR(va[1] + pageSize - 1, UVM_VA_DESC_SIZE) ||
        UVM_ALIGN_ADDR(va[0], UVM_VA_DESC_SIZE) ==
        UVM_ALIGN_ADDR(va[1], UVM_VA_DESC_SIZE))
        return NV_ERR_INVALID_ARGUMENT;

    down_read(&uvm->lock);

    gpu =  uvmfull_uuid_to_active_gpu(&params->gpuUuid);
    if (!gpu || !gpu->bValid)
    {
        params->rmStatus = NV_ERR_INVALID_REQUEST;
        goto cleanup_uvm;
    }

    vaspace = (UvmVaspace *)filep->private_data;
    if (!vaspace)
    {
        params->rmStatus = NV_ERR_INVALID_REQUEST;
        goto cleanup_uvm;
    }

    down_read(&vaspace->lock);

    // Fetch descriptors for both the va-s
    for (i = 0; i < RANGES; i++)
    {
        query[i].address    = va[i];
        query[i].accessType = UvmVatAccess_GPUQuery;
        query[i].gpuId      = gpu->gpuId;
        query[i].vaspace    = vaspace;
        vaDesc[i] = (Uvm2MbVaDesc*)
                        uvmfull_vat_acquire_leaf(&vaspace->vat, &query[i]);
        if (vaDesc[i] == NULL)
        {
            params->rmStatus = NV_ERR_INVALID_REQUEST;
            goto cleanup_vaspace;
        }

        gpuPde[i] = uvmfull_get_gpupde_in_vadesc(vaDesc[i], gpu->gpuId);
        if (!gpuPde[i])
        {
            params->rmStatus = NV_ERR_INVALID_REQUEST;
            goto cleanup_vaspace;
        }
    }

    // Step 1: Unmap both the ranges (this would also mean we unmap the full
    // 2MB range because we only support implicit mapping in a vaDesc currently).
    for (i = 0; i < RANGES; i++)
    {
        NvU32 j = 0;
        NvU64 vaddr = UVM_ALIGN_ADDR(va[i], UVM_VA_DESC_SIZE);
        UvmTracker *tracker = &(vaDesc[i]->tracker);
        NvBool bMappingExists = NV_FALSE;
        unsigned long *bitmap = NULL;
        pBitmap[i] = vmalloc(LONGS_2M * SIZE_LONG);
        if (pBitmap[i] == NULL)
        {
            params->rmStatus = NV_ERR_NO_MEMORY;
            goto cleanup_vaspace;
        }
        bitmap = pBitmap[i];
        memset(bitmap, 0x0, LONGS_2M * SIZE_LONG);

        while (j < NUM_4K_PER_2M)
        {
            NvU32 order;
            NvU32 pgSize = PAGE_SIZE_4K;
            UvmGpuPte *pte = uvmfull_get_gpupde_current_pte(gpuPde[i],
                                          vaddr + j*PAGE_SIZE_4K, &pgSize);
            if (!pte)
            {
                j++;
                continue;
            }

            if (!pte->swcachevalid || (pte->swcachevalid && !pte->valid))
            {
                j += pgSize/PAGE_SIZE_4K;
                continue;
            }

            // logbase2 of 4ks on pgSize
            order = pgSize == PAGE_SIZE_4K  ? 0 :
                    pgSize == PAGE_SIZE_64K ? 4 :
                    pgSize == PAGE_SIZE_2M  ? 21 : ~0;
            WARN_ON(order == (~0));
            // Set bitmap
            bitmap_allocate_region(bitmap, j, order);
            j += pgSize/PAGE_SIZE_4K;
            bMappingExists = NV_TRUE;
        }

        // Nothing to unmap continue to the next descriptor
        if (!bMappingExists)
            continue;

        pbOps = uvm_pbops_start(vaspace);
        if (!pbOps)
        {
            params->rmStatus = NV_ERR_NO_MEMORY;
            goto cleanup_vaspace;
        }

        uvm_pbops_reserve_acquire_space(pbOps, tracker);
        uvmfull_unmap_gpu_pde_pages(vaspace, gpu->gpuId, pbOps, vaDesc[i],
                             gpuPde[i], UVM_ALIGN_ADDR(va[i], UVM_VA_DESC_SIZE),
                             NUM_4K_PER_2M, NV_FALSE);
        uvm_pbops_submit_and_release(pbOps, NVBIT64(gpu->gpuId), tracker);
    }

    // Step 2: Allocate a temporary backing
    {
        params->rmStatus =
                    uvmfull_allocate_phys_vid_mem(gpu,
                                                  UVM_VA_DESC_SIZE, &tempPfn);
        if (params->rmStatus != NV_OK)
            goto cleanup_vaspace;
    }

    // Step 3: swap data.
    for (i = 0; i < RANGES; i++)
    {
        NvU64 dstPa;
        UvmTracker *tracker = &vaDesc[i]->tracker;
        unsigned long *bitmap = pBitmap[i];
        pbOps = uvm_pbops_start(vaspace);
        if (!pbOps)
        {
            // ok to leave the pages unmapped here. (will fault again)
            params->rmStatus = NV_ERR_NO_MEMORY;
            goto cleanup_vaspace;
        }

        uvm_pbops_reserve_acquire_space(pbOps, tracker);
        if (i == 0)
            dstPa = PFN_TO_PA(tempPfn);
        else
            dstPa = PFN_TO_PA(gpuPde[0]->gpuFirstPfn);

        _test_move_gpu_pde_data(gpu, vaDesc[i], pbOps,
                                gpuPde[i], bitmap, dstPa,
                                UvmMemType_vidmem);
        uvm_pbops_submit_and_release(pbOps, NVBIT64(gpu->gpuId), tracker);
    }

    // Step 4: Copy from tmpPfn to backing of va2 and map both descriptors
    for (i = 0; i < RANGES; i++)
    {
        UvmTracker *tracker = &vaDesc[i]->tracker;
        unsigned long *bitmap = pBitmap[i];
        pbOps = uvm_pbops_start(vaspace);
        if (!pbOps)
        {
            // ok to leave the pages unmapped here. (will fault again)
            params->rmStatus = NV_ERR_NO_MEMORY;
            goto cleanup_vaspace;
        }

        uvm_pbops_reserve_acquire_space(pbOps, tracker);
        if (i == 0)
        {
            NvU64 dstPa;
            dstPa = PFN_TO_PA(gpuPde[1]->gpuFirstPfn);
            uvm_push_copy(uvm_pbops_gpu_pb(pbOps, gpu->gpuId),
                          PFN_TO_PA(tempPfn), dstPa,
                          UVM_VA_DESC_SIZE,
                          NV_UVM_COPY_SRC_TYPE_PHYSICAL |
                          NV_UVM_COPY_DST_TYPE_PHYSICAL,
                          UvmMemType_vidmem,
                          UvmMemType_vidmem);

            temp = gpuPde[0]->gpuFirstPfn;
            gpuPde[0]->gpuFirstPfn = gpuPde[1]->gpuFirstPfn;
        }
        else
            gpuPde[1]->gpuFirstPfn = temp;

        // Now map the descs
        _test_map_gpu_pde_data(gpu, vaDesc[i], pbOps, gpuPde[i], bitmap);
        uvm_pbops_submit_and_release(pbOps, NVBIT64(gpu->gpuId), tracker);
    }

    // Step 5: Now wait for operation to complete
    for (i = 0; i < RANGES; i++)
        uvm_wait_for_tracker(&vaDesc[i]->tracker);

    // Free the temp page
    uvmfull_free_phys_vid_mem(gpu, tempPfn);

cleanup_vaspace:
    for (i = RANGES; i > 0; i--)
    {
        if (pBitmap[i])
            vfree(pBitmap[i]);
        if (vaDesc[i-1])
            uvmfull_vat_release_leaf(&vaspace->vat, &query[i-1]);
    }

    up_read(&vaspace->lock);
cleanup_uvm:
    up_read(&uvm->lock);

    return params->rmStatus;
}

NV_STATUS uvm_test_map_gpu_sysmem_pages(
        UVM_TEST_MAP_GPU_SYSMEM_PAGES_PARAMS *params, struct file* filep)
{
    struct vm_area_struct *pVma;
    UvmVaspace         *pVaspace;
    UvmGlobalState     *pUvm;
    UvmGpu             *pGpu;
    void               *pBuffer;
    UvmGpuPfn           pfn=0;
    NvU32               i;
    int                 ret;
    unsigned long long  physAddr=0;
    unsigned long long  vaddr;
    UvmGpuPteUpdateParams pteParams;
    Uvm2MbVaDesc     *pUvm2MbVaDesc = NULL;
    UvmGpuPde        *pGpuPde = NULL;
    UvmPbOps         *pbOps = NULL;
    UvmVatContext    query;
    UvmPteHintsParams pteHintsParams;

    UVM_DBG_PRINT_UUID("\n UVM Test Map 2MB CPU/GPU pages",&params->gpuUuid);

    params->rmStatus = NV_ERR_INVALID_REQUEST;
    UVM_CHECK_VALID_PAGE_SIZE(params->gpuPgSize, params->rmStatus);

    pVaspace = (UvmVaspace *)filep->private_data;
    if (!pVaspace)
    {
        params->rmStatus = NV_ERR_INVALID_REQUEST;
        return params->rmStatus;
    }

    down_write(&current->mm->mmap_sem);
    pVma = find_vma(current->mm, params->targetVa);
    if (!pVma)
    {
        UVM_ERR_PRINT("Failed to find vma for user-address %llx\n",
                                                         params->targetVa);
        params->rmStatus = NV_ERR_INVALID_REQUEST;
        up_write(&current->mm->mmap_sem);
        return params->rmStatus;
    }

    //
    // Check alignment/boundary conditions of the virtual address
    // No need to check with vm_start as find_vma will fail otherwise
    //
    if (pVma->vm_end < (params->targetVa + params->gpuPgSize) ||
        (params->targetVa & (params->gpuPgSize-1)))
    {
        UVM_ERR_PRINT("User-address %llx and gpuPgSize %x incorrect/not aligned \n",
                                                    params->targetVa, params->gpuPgSize);
        params->rmStatus = NV_ERR_INVALID_REQUEST;
        return params->rmStatus;
    }


    pUvm = uvmfull_get_global();
    down_read(&pUvm->lock);

    pGpu =  uvmfull_uuid_to_active_gpu(&params->gpuUuid);
    if (!pGpu || !pGpu->bValid)
    {
        params->rmStatus = NV_ERR_INVALID_REQUEST;
        goto cleanup_uvm;
    }

    down_read(&pVaspace->lock);
    if (pVaspace->filep != filep)
    {
        params->rmStatus = NV_ERR_INVALID_REQUEST;
        goto cleanup_vaspace;
    }

    pBuffer = kmalloc(params->gpuPgSize, NV_UVM_GFP_FLAGS | GFP_USER);
    if (!pBuffer)
    {
        params->rmStatus = NV_ERR_INVALID_REQUEST;
        goto cleanup_vaspace;
    }

    query.accessType = UvmVatAccess_GPUTouch;
    query.address = params->targetVa;
    query.vaspace = pVaspace;
    pUvm2MbVaDesc = uvmfull_vat_acquire_leaf(&pVaspace->vat, &query);
    if (!pUvm2MbVaDesc)
    {
        // return error
        params->rmStatus = NV_ERR_GENERIC;
        goto cleanup_page;
    }

    //
    // Make sure that this va is not mapped on any GPU va first, since this is a
    // controlled test, we dont want any existing mappings to be affected,
    //
    desc_for_each_gpu_pde(pGpuPde, pUvm2MbVaDesc)
    {
        NvU32 pgSize;
        UvmGpuPte *gpuPte;
        // We found a gpu with populated pages
        gpuPte = uvmfull_get_gpupde_current_pte(pGpuPde, params->targetVa,
                                                &pgSize);
        if (!gpuPte || !gpuPte->swcachevalid)
            continue;

        if (gpuPte->valid)
        {
            UVM_ERR_PRINT("Found an existing mapping on GPU %d for user-address %llx\n",
                                        pGpuPde->hostGpuId, params->targetVa);
            params->rmStatus = NV_ERR_INVALID_STATE;
            goto cleanup_page;
        }
    }


    pGpuPde = uvmfull_get_gpupde_in_vadesc(pUvm2MbVaDesc, pGpu->gpuId);
    if (!pGpuPde)
    {
        params->rmStatus = NV_ERR_GENERIC;
        goto cleanup_gpde;
    }

    physAddr = __pa(pBuffer);
    pfn = physAddr >> UVM_CPU_MIN_PAGE_SHIFT;
    memset(&pteHintsParams, 0, sizeof(pteHintsParams));

    // update all appropriate flags pte flags in hints structure
    pteHintsParams.updateFlags = UVM_GPU_VA_PTE_HINTS_FLAG_SET_CPU_PFN_VALID  |
                                 UVM_GPU_VA_PTE_HINTS_FLAG_SET_CPU_PFN_VALUE  |
                                 UVM_GPU_VA_PTE_HINTS_FLAG_SET_CPU_PFN_MAPPED |
                                 UVM_GPU_VA_PTE_HINTS_FLAG_SET_CPU_PFN_DIRTY;

    pteHintsParams.pte.cpuPfnValid  = 1;
    pteHintsParams.pte.cpuPfnDirty  = 1;
    pteHintsParams.pte.cpuPfnMapped = 1;

    for (i=0, vaddr = params->targetVa;
            i < (params->gpuPgSize/UVM_CPU_MIN_PAGE_SIZE) && vaddr < pVma->vm_end; i++)
    {
        // Map all the cpu pages to sysmem
        ret =  vm_insert_pfn(pVma, vaddr, pfn);
        UVM_ERR_PRINT("Failed to map pfn %llx to CPU user-address %llx\n", pfn, vaddr);
        UVM_PANIC_ON(1);

        // just update the pfn
        pteHintsParams.pte.cpuPfn       = pfn;

        // Update all the appropriate flags
        uvmfull_update_pte_hints(vaddr, (UVM_CPU_MIN_PAGE_SIZE/PAGE_SIZE_4K),
                                 PAGE_SIZE_4K, &pteHintsParams, pUvm2MbVaDesc);


        vaddr += UVM_CPU_MIN_PAGE_SIZE;
        pfn   += 1;
    }

    // Now map the GPU page to the same VA
    {
        // Gather pushbuffers
        pbOps = uvm_pbops_start_on_gpu(pGpu->gpuId);
        if (!pbOps)
        {
            params->rmStatus = NV_ERR_GENERIC;
            goto cleanup_gpde;
        }
        uvm_pbops_reserve_acquire_space(pbOps, &pUvm2MbVaDesc->tracker);

        memset (&pteParams, 0, sizeof(pteParams));

        // use auto-increment and pass the first pfn# as input
        pteParams.updateSwFlags |= UVM_GPU_SW_PTE_UPDATE_FLAGS_PFN_AUTOINCREMENT;
        pteParams.updateSwFlags |= UVM_GPU_SW_PTE_UPDATE_FLAGS_PFN_VALID;
        pteParams.updateSwFlags |= UVM_GPU_SW_PTE_UPDATE_FLAGS_PFN;
        pteParams.pfnvalid       = NV_TRUE;
        pteParams.updateSwFlags |= UVM_GPU_SW_PTE_UPDATE_FLAGS_SWVALID;
        pteParams.swcachevalid   = NV_TRUE;
        // force aperture to sysmem coherent
        pteParams.updateSwFlags |= UVM_GPU_SW_PTE_UPDATE_FLAGS_APERTURE;
        pteParams.aperture       = UvmMemType_sysmem_coh;
        pteParams.updateSwFlags |= UVM_GPU_SW_PTE_UPDATE_FLAGS_PTE_VALID;
        pteParams.valid          = NV_TRUE;
        // Actually update the PTE
        pteParams.updateHwFlags |= UVM_GPU_HW_PTE_UPDATE_FLAGS_WRITE_PTE;


        // GPU pfns are always saved in 4K, so make sure that we use correct offset when
        // saving internally
        pfn = PA_TO_PFN(physAddr);

        // update the pfn in all the GPU ptes
        if (NV_OK != uvmfull_update_gpupde_ptes(pUvm2MbVaDesc, pGpu->gpuId,
                                            pGpuPde,
                                            params->targetVa, 1 /*numPages*/,
                                            params->gpuPgSize, &pteParams,
                                            &pfn, pbOps,
                                            (NvU8*)&pVaspace->gpuVaspaces[pGpu->gpuId].pteDefaults.entry,
                                            pVaspace->gpuVaspaces[pGpu->gpuId].pteDefaults.entrySize,
                                            NV_FALSE))
        {
            UVM_ERR_PRINT("Failed to map pfn %llx to GPU user-address %llx\n", pfn, vaddr);
            params->rmStatus = NV_ERR_GENERIC;
            goto cleanup_gpde;
        }
        // Submit the push-buffer.
        uvm_pbops_submit(pbOps,
                         pGpu->gpuId,
                         uvm_pbops_gpu_pb(pbOps, pGpu->gpuId),
                         &pUvm2MbVaDesc->tracker);
        uvm_move_tracker(&pUvm2MbVaDesc->tracker, pbOps->t);

        params->rmStatus = uvm_wait_for_tracker(&pUvm2MbVaDesc->tracker);

cleanup_gpde:
        if (pbOps)
            uvm_pbops_end(pbOps);

        uvmfull_vat_release_leaf(&pVaspace->vat, &query);
    }


cleanup_page:
    if (pBuffer)
        kfree(pBuffer);
cleanup_vaspace:
    up_read(&pVaspace->lock);
cleanup_uvm:
    up_read(&pUvm->lock);
    up_write(&current->mm->mmap_sem);
    return params->rmStatus;
}
