/*******************************************************************************
    Copyright (c) 2014-2015 NVIDIA Corporation

    Permission is hereby granted, free of charge, to any person obtaining a copy
    of this software and associated documentation files (the "Software"), to
    deal in the Software without restriction, including without limitation the
    rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
    sell copies of the Software, and to permit persons to whom the Software is
    furnished to do so, subject to the following conditions:

        The above copyright notice and this permission notice shall be
        included in all copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
    IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
    FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
    THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
    LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
    FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
    DEALINGS IN THE SOFTWARE.

*******************************************************************************/

#include "uvm_common.h"
#include "uvm_utils.h"
#include "uvm_full.h"
#include "uvm_full_pagetbl_mgmt.h"

static struct kmem_cache * g_uvmVatGpuHeaderCache __read_mostly = NULL;
static struct kmem_cache * g_uvmVatL0Cache __read_mostly = NULL;
static struct kmem_cache * g_uvmVatL1Cache __read_mostly = NULL;
static struct kmem_cache * g_uvmVatL2Cache __read_mostly = NULL;

NV_STATUS uvmfull_va_trie_slab_init(void)
{
    NV_STATUS status = NV_OK;
    g_uvmVatL0Cache = NV_KMEM_CACHE_CREATE("uvm_vatL0Cache", UvmVatL0);
    if (!g_uvmVatL0Cache)
    {
        status = NV_ERR_NO_MEMORY;
        goto cleanup;
    }
    g_uvmVatL1Cache = NV_KMEM_CACHE_CREATE("uvm_vatL1Cache", UvmVatL1);
    if (!g_uvmVatL1Cache)
    {
        status = NV_ERR_NO_MEMORY;
        goto cleanup;
    }
    g_uvmVatL2Cache = NV_KMEM_CACHE_CREATE("uvm_vatL2Cache", UvmVatL2);
    if (!g_uvmVatL2Cache)
    {
        status = NV_ERR_NO_MEMORY;
        goto cleanup;
    }
    g_uvmVatGpuHeaderCache = NV_KMEM_CACHE_CREATE("uvm_vatGpuHeaderCache",
                                                  UvmVatLevelGpuHeader);
    if (!g_uvmVatGpuHeaderCache)
    {
        status = NV_ERR_NO_MEMORY;
        goto cleanup;
    }
    return status;

cleanup:
    uvmfull_va_trie_slab_destroy();
    return status;
}

void uvmfull_va_trie_slab_destroy(void)
{
    if (g_uvmVatL2Cache)
        kmem_cache_destroy_safe(&g_uvmVatL2Cache);
    if (g_uvmVatL1Cache)
        kmem_cache_destroy_safe(&g_uvmVatL1Cache);
    if (g_uvmVatL0Cache)
        kmem_cache_destroy_safe(&g_uvmVatL0Cache);
    if (g_uvmVatGpuHeaderCache)
        kmem_cache_destroy_safe(&g_uvmVatGpuHeaderCache);
}

static void
_vat_init_level3(UvmVat *trie, UvmVatL3 *level)
{
    NvU32               nEntries, i;
    NvU32               levelId = 3;

    memset(level, 0x0, sizeof(UvmVatL3));
    init_rwsem(&level->headerLock);
    down_write(&level->headerLock);
    level->header.id = levelId;
    level->header.loBit = trie->levelInfo[levelId].loBit;
    level->header.hiBit = trie->levelInfo[levelId].hiBit;
    INIT_LIST_HEAD(&level->header.gpuHeaders);

    nEntries = 1 << (trie->levelInfo[levelId].hiBit -
                    trie->levelInfo[levelId].loBit + 1);
    for (i = 0; i < nEntries; i++)
        init_rwsem(&level->entry[i].lock);

    up_write(&level->headerLock);
}

static void
_vat_init_level2(UvmVat *trie, UvmVatL2 *level)
{
    NvU32               levelId = 2;

    memset(level, 0x0, sizeof(UvmVatL2));

    level->header.id = levelId;
    level->header.loBit = trie->levelInfo[levelId].loBit;
    level->header.hiBit = trie->levelInfo[levelId].hiBit;
    INIT_LIST_HEAD(&level->header.gpuHeaders);
}

static void
_vat_init_level1(UvmVat *trie, UvmVatL1 *level)
{
    NvU32               levelId = 1;

    memset(level, 0x0, sizeof(UvmVatL1));

    level->header.id = levelId;
    level->header.loBit = trie->levelInfo[levelId].loBit;
    level->header.hiBit = trie->levelInfo[levelId].hiBit;
    INIT_LIST_HEAD(&level->header.gpuHeaders);
}

static void
_vat_init_level0(UvmVat *trie, UvmVatL0 *level)
{
    NvU32               nEntries, i;
    NvU32               levelId = 0;

    memset(level, 0x0, sizeof(UvmVatL0));
    init_rwsem(&level->headerLock);
    down_write(&level->headerLock);
    level->header.id = levelId;
    level->header.loBit = trie->levelInfo[levelId].loBit;
    level->header.hiBit = trie->levelInfo[levelId].hiBit;
    INIT_LIST_HEAD(&level->header.gpuHeaders);

    nEntries = 1 << (trie->levelInfo[levelId].hiBit -
                    trie->levelInfo[levelId].loBit + 1);
    for (i = 0; i < nEntries; i++)
        init_rwsem(&level->entry[i].lock);

    up_write(&level->headerLock);
}

static void _vat_push_lock(UvmVatContext *context, struct rw_semaphore *lock,
                            NvBool bWriteMode)
{
    UVM_PANIC_ON(context->private.tos >= UVM_VA_TRIE_LOCKS);
    context->private.lockStack[context->private.tos].bWriteMode = bWriteMode;
    context->private.lockStack[context->private.tos].pLock = lock;
    context->private.tos++;
}

// release all locks in stack.
static UvmVatLockState* _vat_pop_lock(UvmVatContext *context)
{
    UvmVatLockState* state = NULL;
    if (context->private.tos == 0)
        return NULL;

    context->private.tos--;
    state = &(context->private.lockStack[context->private.tos]);
    return state;
}

static void _vat_push_acquire_lock(
        UvmVatContext *context, struct rw_semaphore *lock, NvBool bWriteMode)
{
    bWriteMode ? down_write(lock) : down_read(lock);
    _vat_push_lock(context, lock, bWriteMode);
}

// release all locks in stack.
static void _vat_pop_release_lock(UvmVatContext *context)
{
    UvmVatLockState *state = _vat_pop_lock(context);
    if (state)
        state->bWriteMode ? up_write(state->pLock) : up_read(state->pLock);
}

void uvmfull_vat_release_leaf(UvmVat *trie, UvmVatContext *query)
{
    UvmVatLockState *state;

    //
    // Call vat_reclaim if query is GPUReclaim & vaDesc has been found
    // and released
    //
    if (query->accessType == UvmVatAccess_GenericReclaim &&
        query->private.level0Entry != NULL)
    {
        uvmfull_vat_reclaim_levels(trie, query);
    }

    while (state = _vat_pop_lock(query), state != NULL)
    {
        state->bWriteMode ? up_write(state->pLock) : up_read(state->pLock);
    }
}

UvmVatLevelGpuHeader*  uvmfull_vat_get_gpu_header(UvmGpuId gpuId,
                                                  UvmVatLevelHeader* header)
{
    UvmVatLevelGpuHeader *gpuHeader;
    // Loop for all gpu headers. Return the one with matching gpuid
    list_for_each_entry(gpuHeader, &header->gpuHeaders, node)
    {
        if (gpuHeader->gpuId == gpuId)
            return gpuHeader;
    }
    return NULL;
}

static UvmVatLevelGpuHeader*
_vat_alloc_gpu_level(UvmGpuVaspace *gpuVaspace, UvmVatContext* query,
                        UvmVatLevelHeader* header)
{
    NvU64 pfn;
    UvmVatLevelGpuHeader *gpuHeader =
            kmem_cache_zalloc(g_uvmVatGpuHeaderCache, NV_UVM_GFP_FLAGS);
    if (!gpuHeader)
        return NULL;

    gpuHeader->gpuId = query->gpuId;
    if (header->id == 3)
    {
        // grab it from gpuVaspace
        pfn = PA_TO_PFN(gpuVaspace->gpuPDB);
    }
    else if(NV_OK != uvmfull_get_pagetbl_memory(PAGE_SIZE_4K, &pfn,
                                            gpuVaspace->bZeroFb ?
                                            UvmMemType_sysmem_coh :
                                            UvmMemType_vidmem,
                                            gpuVaspace->gpuId))
    {
        kmem_cache_free(g_uvmVatGpuHeaderCache, gpuHeader);
        return NULL;
    }

    gpuHeader->id = header->id;
    gpuHeader->levelPhyAddr = PFN_TO_PA(pfn);
    gpuHeader->aperture = gpuVaspace->bZeroFb ?
                            UvmMemType_sysmem_coh:
                            UvmMemType_vidmem;
    list_add(&gpuHeader->node, &header->gpuHeaders);
    header->gpuAllocMask |= NVBIT64(query->gpuId);
    return gpuHeader;
}

static NV_STATUS
_vat_update_gpu_level(UvmGpuVaspace *gpuVaspace, UvmVatLevelHeader *parentLevel,
                      UvmVatLevelGpuHeader *parentGpuLevel, NvU64 *mapMask,
                      NvU32 entryId, UvmVatLevelGpuHeader* childGpuLevel,
                      NvBool bValid)
{
    UvmTracker tracker;
    UvmPbOps   *pbOps;
    UvmPushbuffer *pb;
    NV_STATUS status = NV_OK;
    UvmGpuId gpuId = gpuVaspace->gpuId;
    NvU64 childPhysAddr;
    UvmMemType childGpuAperture;;

    uvm_init_tracker(&tracker);
    pbOps = uvm_pbops_start_on_gpu(gpuId);
    if (!pbOps)
        return NV_ERR_NO_MEMORY;
    pb = uvm_pbops_gpu_pb(pbOps, gpuId);
    UVM_PANIC_ON(!pb);

    // Fill in the childPhysAddr & childGpuAperture if making an entry valid
    // For invalid PDEs we simpyl scrub the entry
    if (bValid)
    {
        childPhysAddr = childGpuLevel->levelPhyAddr;
        childGpuAperture = childGpuLevel->aperture;

        // scrub the memory before pointing to
        if (childGpuAperture == UvmMemType_vidmem)
            uvm_push_memzero_vidmem(gpuId, pb, childPhysAddr, PAGE_SIZE_4K);
    }
    else
    {
        childPhysAddr = 0;
        childGpuAperture = UvmMemType_invalid;
    }

    uvmfull_update_gpu_pde_entry(gpuVaspace, pb, parentLevel->id,
                          parentGpuLevel->levelPhyAddr,
                          parentGpuLevel->aperture, entryId, parentLevel->loBit,
                          childPhysAddr, childGpuAperture,
                          0, UvmMemType_invalid/*No sublevel*/,
                          NV_FALSE /*skipTlbInvalidate*/);

    uvm_pbops_submit(pbOps, gpuId, pb, &tracker);
    uvm_move_tracker(&tracker, pbOps->t);
    status = uvm_wait_for_tracker(&tracker);
    uvm_pbops_end(pbOps);

    if (bValid)
    {
        *mapMask = *mapMask | NVBIT64(gpuId);
        // Increase valid entry count on GPU in parent gpu header
        parentGpuLevel->gpuValidEntries++;
    }
    else
    {
        *mapMask &= ~NVBIT64(gpuId);
        // Decrease valid entry count on GPU in parent gpu header
        --parentGpuLevel->gpuValidEntries;
    }
    return status;
}

static void*
_vat_access_level3(UvmVat *trie, UvmVatL3 *level, UvmVatContext *query)
{
    UvmGpuVaspace *gpuVaspace = NULL;
    UvmVatLevelGpuHeader *gpuHeaderThis = NULL;
    NvBool bExclusive = query->private.bRootExclusive;
    NvU32 entryId = VAT_EXTRACT_INDEX(query->address, trie->levelInfo[3].hiBit,
                                      trie->levelInfo[3].loBit);
    UvmVatL3Entry *entry = &level->entry[entryId];

    _vat_push_acquire_lock(query, &level->headerLock,
                                bExclusive ? NV_TRUE : NV_FALSE);

    if (query->accessType == UvmVatAccess_GPUTouch ||
        query->accessType == UvmVatAccess_GPUQuery)
    {
        if ((query->vaspace->activeGpuMask & NVBIT64(query->gpuId)) == 0)
            goto error_release_header;

        gpuVaspace = &query->vaspace->gpuVaspaces[query->gpuId];
        // Check if level is allocated on the gpu
        if (!(NVBIT64(query->gpuId) & level->header.gpuAllocMask))
        {
            if (query->accessType == UvmVatAccess_GPUQuery)
                goto error_release_header;

            if (!bExclusive)
            {
                // drop lock and return early
                goto error_release_header;
            }
            else
            {
                gpuHeaderThis = _vat_alloc_gpu_level(gpuVaspace, query,
                                                        &level->header);
                if (gpuHeaderThis == NULL)
                {
                    goto error_release_header;
                }
            }
        }
        else   // level already allocated
        {
            gpuHeaderThis = uvmfull_vat_get_gpu_header(query->gpuId,
                                                       &level->header);
        }
    }

    // Access the entry now
    _vat_push_acquire_lock(query, &entry->lock,
                            bExclusive ? NV_TRUE : NV_FALSE);
    {
        if (entry->child == NULL)
        {
            if (query->accessType == UvmVatAccess_GenericQuery ||
                query->accessType == UvmVatAccess_GenericReclaim)
                goto error_release_entry;

            if (!bExclusive)
                goto error_release_entry;

            entry->child = kmem_cache_alloc(g_uvmVatL2Cache, NV_UVM_GFP_FLAGS);
            if (!entry->child)
            {
                UVM_PANIC();
                goto error_release_entry;
            }

            _vat_init_level2(trie, entry->child);

            //
            // Increment the valid entry count at level and update the parent
            // pointer in child
            //
            level->header.validLevelEntries++;
            entry->child->header.parentLevel = (void *)entry;
        }
    }
    if (query->accessType == UvmVatAccess_GPUTouch ||
        query->accessType == UvmVatAccess_GPUQuery)
    {
        // Check if the level entry is mapped on the gpu
        if (!(NVBIT64(query->gpuId) & entry->gpuMappedMask))
        {
            if (query->accessType == UvmVatAccess_GPUQuery)
                goto error_release_entry;

            // Map the entry now
            if (!bExclusive)
            {
                // Have to be in exclusive mode to map
                goto error_release_entry;
            }
            else
            {
                UvmVatL2 *child;
                UvmVatLevelGpuHeader *gpuHeader;
                if (entry->child == NULL)
                    goto error_release_entry;

                child = entry->child;
                if (!(NVBIT64(query->gpuId) & child->header.gpuAllocMask))
                {
                    gpuHeader = _vat_alloc_gpu_level(gpuVaspace, query,
                                                     &child->header);
                    if (gpuHeader == NULL)
                        goto error_release_entry;
                }
                else
                    gpuHeader = uvmfull_vat_get_gpu_header(query->gpuId,
                                                           &child->header);

                // allocated child level. Now point to it
                if (NV_OK != _vat_update_gpu_level(gpuVaspace,
                                                   &level->header, gpuHeaderThis,
                                                   &entry->gpuMappedMask,
                                                   entryId, gpuHeader, NV_TRUE))
                {
                    goto error_release_entry;
                }
            }
        }
    }

    return entry->child;

error_release_entry:
    _vat_pop_release_lock(query);
error_release_header:
    _vat_pop_release_lock(query);
    return NULL;
}

static void*
_vat_access_level2(UvmVat *trie, UvmVatL2 *level, UvmVatContext *query)
{
    UvmGpuVaspace *gpuVaspace = NULL;
    UvmVatLevelGpuHeader *gpuHeaderThis = NULL;
    NvBool bExclusive = query->private.bRootExclusive;
    NvU32 entryId = VAT_EXTRACT_INDEX(query->address, trie->levelInfo[2].hiBit,
                                      trie->levelInfo[2].loBit);
    UvmVatL2Entry *entry = &level->entry[entryId];

    if (query->accessType == UvmVatAccess_GPUTouch ||
        query->accessType == UvmVatAccess_GPUQuery)
    {
        if (!(NVBIT64(query->gpuId) & level->header.gpuAllocMask) &&
            query->accessType != UvmVatAccess_GPUTouch)
            goto error_release_header;

        gpuVaspace = &query->vaspace->gpuVaspaces[query->gpuId];

        // Sanity check if level is allocated on the gpu
        WARN_ON(!(NVBIT64(query->gpuId) & level->header.gpuAllocMask));

        // Grab the required header
        gpuHeaderThis = uvmfull_vat_get_gpu_header(query->gpuId, &level->header);
        WARN_ON(!gpuHeaderThis);
    }

    {
        if (entry->child == NULL)
        {
            if (query->accessType == UvmVatAccess_GenericQuery ||
                query->accessType == UvmVatAccess_GenericReclaim)
                goto error_release_entry;

            if (!bExclusive)
                goto error_release_entry;

            entry->child = kmem_cache_alloc(g_uvmVatL1Cache, NV_UVM_GFP_FLAGS);
            if (!entry->child)
            {
                UVM_PANIC();
                goto error_release_entry;
            }

            _vat_init_level1(trie, entry->child);
            
            //
            // Increment the valid entry count at level and update the parent
            // pointer in child
            //
            level->header.validLevelEntries++;
            entry->child->header.parentLevel = (void *)entry;
        }
    }
    if (query->accessType == UvmVatAccess_GPUTouch ||
        query->accessType == UvmVatAccess_GPUQuery)
    {
        // Check if level is allocated on the gpu
        if (!(NVBIT64(query->gpuId) & entry->gpuMappedMask))
        {
            if (query->accessType == UvmVatAccess_GPUQuery)
                goto error_release_entry;

            // Map the entry at this index
            if (!bExclusive)
            {
                goto error_release_entry;
            }
            else
            {
                UvmVatL1 *child;
                UvmVatLevelGpuHeader *gpuHeader;
                if (entry->child == NULL)
                    goto error_release_entry;

                child = entry->child;
                if (!(NVBIT64(query->gpuId) & child->header.gpuAllocMask))
                {
                    gpuHeader = _vat_alloc_gpu_level(gpuVaspace, query,
                                                        &child->header);
                    if (gpuHeader == NULL)
                        goto error_release_entry;
                }
                else
                    gpuHeader = uvmfull_vat_get_gpu_header(query->gpuId,
                                                           &level->header);

                // allocated child level. Now point to it
                if (NV_OK != _vat_update_gpu_level(gpuVaspace,
                                                   &level->header,gpuHeaderThis,
                                                   &entry->gpuMappedMask,
                                                   entryId, gpuHeader, NV_TRUE))
                {
                    goto error_release_entry;
                }
            }
        }
    }

    return entry->child;

error_release_entry:
error_release_header:
    return NULL;
}

static void*
_vat_access_level1(UvmVat *trie, UvmVatL1 *level, UvmVatContext *query)
{
    UvmGpuVaspace *gpuVaspace = NULL;
    UvmVatLevelGpuHeader *gpuHeaderThis = NULL;
    NvBool bExclusive = query->private.bRootExclusive;
    NvU32 entryId = VAT_EXTRACT_INDEX(query->address, trie->levelInfo[1].hiBit,
                                      trie->levelInfo[1].loBit);
    UvmVatL1Entry *entry = &level->entry[entryId];

    if (query->accessType == UvmVatAccess_GPUTouch ||
        query->accessType == UvmVatAccess_GPUQuery)
    {
        if (!(NVBIT64(query->gpuId) & level->header.gpuAllocMask) &&
            query->accessType != UvmVatAccess_GPUTouch)
            goto error_release_header;

        gpuVaspace = &query->vaspace->gpuVaspaces[query->gpuId];

        // Sanity check if level is allocated on the gpu
        WARN_ON(!(NVBIT64(query->gpuId) & level->header.gpuAllocMask));

        // Grab the required header
        gpuHeaderThis = uvmfull_vat_get_gpu_header(query->gpuId, &level->header);
        WARN_ON(!gpuHeaderThis);
    }

    {
        if (entry->child == NULL)
        {
            if (query->accessType == UvmVatAccess_GenericQuery ||
                query->accessType == UvmVatAccess_GenericReclaim)
                goto error_release_entry;

            if (!bExclusive)
                goto error_release_entry;

            entry->child = kmem_cache_alloc(g_uvmVatL0Cache, NV_UVM_GFP_FLAGS);
            if (!entry->child)
            {
                UVM_PANIC();
                goto error_release_entry;
            }

            _vat_init_level0(trie, entry->child);
            
            //
            // Increment the valid entry count at level and update the parent
            // pointer in child
            //
            level->header.validLevelEntries++;
            entry->child->header.parentLevel = (void *)entry;
        }
    }
    if (query->accessType == UvmVatAccess_GPUTouch ||
        query->accessType == UvmVatAccess_GPUQuery)
    {
        // Check if entry is mapped on the gpu
        if (!(NVBIT64(query->gpuId) & entry->gpuMappedMask))
        {
            if (query->accessType == UvmVatAccess_GPUQuery)
                goto error_release_entry;

            // Map the entry at this index
            if (!bExclusive)
            {
                goto error_release_entry;
            }
            else
            {
                UvmVatL0 *child;
                UvmVatLevelGpuHeader *gpuHeader;
                if (entry->child == NULL)
                    goto error_release_entry;

                child = entry->child;
                if (!(NVBIT64(query->gpuId) & child->header.gpuAllocMask))
                {
                    gpuHeader = _vat_alloc_gpu_level(gpuVaspace, query,
                                                    &child->header);
                    if (gpuHeader == NULL)
                        goto error_release_entry;
                }
                else
                    gpuHeader = uvmfull_vat_get_gpu_header(query->gpuId,
                                                           &child->header);

                // allocated child level. Now point to it
                if (NV_OK != _vat_update_gpu_level(gpuVaspace,
                                                   &level->header, gpuHeaderThis,
                                                   &entry->gpuMappedMask,
                                                   entryId, gpuHeader, NV_TRUE))
                {
                    goto error_release_entry;
                }
            }
        }
    }

    return entry->child;

error_release_entry:
error_release_header:
    return NULL;
}

static void*
_vat_access_level0(UvmVat *trie, UvmVatL0 *level, UvmVatContext *query)
{
    UvmVatLevelGpuHeader *gpuHeaderThis = NULL;
    Uvm2MbVaDesc *vaDesc = NULL;
    NvU32 entryId = VAT_EXTRACT_INDEX(query->address, trie->levelInfo[0].hiBit,
                                      trie->levelInfo[0].loBit);
    UvmVatL0Entry *entry = &level->entry[entryId];

    // Merging the header lock with top most level.
    //_vat_push_acquire_lock(query, &level->headerLock,
    //                            bExclusive ? NV_TRUE : NV_FALSE);

    if (query->accessType == UvmVatAccess_GPUTouch ||
        query->accessType == UvmVatAccess_GPUQuery)
    {
        if (!(NVBIT64(query->gpuId) & level->header.gpuAllocMask) &&
            query->accessType != UvmVatAccess_GPUTouch)
            goto error_release_header;

        // Sanity check if level is allocated on the gpu
        WARN_ON(!(NVBIT64(query->gpuId) & level->header.gpuAllocMask));

        // Grab the required header
        gpuHeaderThis = uvmfull_vat_get_gpu_header(query->gpuId, &level->header);
        WARN_ON(!gpuHeaderThis);
    }

    // The leaf lock is always taken in write mode
    _vat_push_acquire_lock(query, &entry->lock, NV_TRUE);
    if (entry->child == NULL)
    {
        if (query->accessType == UvmVatAccess_GenericQuery ||
            query->accessType == UvmVatAccess_GenericReclaim)
            goto error_release_entry;

        vaDesc = uvmfull_alloc_va_desc(query->vaspace, query->address);
        WARN_ON(!vaDesc);
        if (!vaDesc)
            goto error_release_entry;

        vaDesc->parent = entry;
        entry->child = (void*)vaDesc;
        
        //
        // Increment the valid entriy count at given level. There is no parent
        // pointer here as this is a 2MB entry & parent is saved in va2MbDesc
        //
        level->header.validLevelEntries++;
    }
    vaDesc = (Uvm2MbVaDesc*)entry->child;
    if (query->accessType == UvmVatAccess_GPUTouch ||
        query->accessType == UvmVatAccess_GPUQuery)
    {
        // Now check if the entry address is shared with the desc
        if (!(NVBIT64(query->gpuId) & entry->pdePresentMask))
        {
            uvmfull_vadesc_update_gpupde_info(vaDesc, query->gpuId,
                                              gpuHeaderThis,
                                              entryId);
            entry->pdePresentMask |= NVBIT64(query->gpuId);

            //
            // Increase the parent gpuValidEntry count here as this is the best place
            // to determine if a PDE will be mapped later or not. Taking it to bottom
            // level updates increase complexity as 2MB could be a PTE or PDE later
            //
            gpuHeaderThis->gpuValidEntries++;
        }
    }

    return entry->child;

error_release_entry:
    _vat_pop_release_lock(query);
error_release_header:
    return NULL;
}

// The fixed level approach
void* uvmfull_vat_acquire_leaf(UvmVat *trie, UvmVatContext *query)
{
    void *nodeIn = NULL;
    void *nodeOut = NULL;
    if (query->accessType == UvmVatAccess_Invalid)
        return NULL;

    memset(&query->private, 0x0, sizeof(query->private));
    if (trie->levels == 0)
    {
        UVM_PANIC_MSG("Va trie not initialized.\n");
        return NULL;
    }

start_traversal:
    // In a retry; make sure all locks are released before starting
    uvmfull_vat_release_leaf(trie, query);

    // For reclaim mark query as exclusive
    if (query->accessType == UvmVatAccess_GenericReclaim)
        query->private.bRootExclusive = NV_TRUE;

    // access level 3
    nodeOut = _vat_access_level3(trie, &trie->root, query);
    if (!nodeOut)
    {
        if (query->accessType == UvmVatAccess_GPUTouch ||
            query->accessType == UvmVatAccess_GenericTouch)
        {
            if (!query->private.bRootExclusive)
            {
                query->private.bRootExclusive = NV_TRUE;
                goto start_traversal;
            }
        }
        goto error_return_null;
    }

    // access level 2
    nodeIn = nodeOut;
    nodeOut = _vat_access_level2(trie, (UvmVatL2*)nodeIn, query);
    if (!nodeOut)
    {
        if (query->accessType == UvmVatAccess_GPUTouch ||
            query->accessType == UvmVatAccess_GenericTouch)
        {
            if (!query->private.bRootExclusive)
            {
                query->private.bRootExclusive = NV_TRUE;
                goto start_traversal;
            }
        }
        else
            goto error_return_null;
    }

    // access level 1
    nodeIn = nodeOut;
    nodeOut = _vat_access_level1(trie, (UvmVatL1*)nodeIn, query);
    if (!nodeOut)
    {
        if (query->accessType == UvmVatAccess_GPUTouch ||
            query->accessType == UvmVatAccess_GenericTouch)
        {
            if (!query->private.bRootExclusive)
            {
                query->private.bRootExclusive = NV_TRUE;
                goto start_traversal;
            }
        }
        else
            goto error_return_null;
    }

    // access level 0
    nodeIn = nodeOut;
    nodeOut = _vat_access_level0(trie, (UvmVatL0*)nodeIn, query);
    if (!nodeOut)
    {
        if (query->accessType == UvmVatAccess_GPUTouch ||
            query->accessType == UvmVatAccess_GenericTouch)
        {
            query->private.bL0Exclusive = NV_TRUE;
            nodeOut = _vat_access_level0(trie, (UvmVatL0*)nodeIn, query);
            if (!nodeOut)
                goto error_return_null;
        }
    }

    return nodeOut;

error_return_null:
    uvmfull_vat_release_leaf(trie, query);
    return nodeOut;
}

void uvmfull_vat_init(UvmVat *trie)
{
    trie->levels = UVM_VA_TRIE_LVLS;
    trie->levelInfo[0].loBit = UVM_VA_SHIFT_TRIE_LVL_0_LO;
    trie->levelInfo[0].hiBit = UVM_VA_SHIFT_TRIE_LVL_0_HI;
    trie->levelInfo[1].loBit = UVM_VA_SHIFT_TRIE_LVL_1_LO;
    trie->levelInfo[1].hiBit = UVM_VA_SHIFT_TRIE_LVL_1_HI;
    trie->levelInfo[2].loBit = UVM_VA_SHIFT_TRIE_LVL_2_LO;
    trie->levelInfo[2].hiBit = UVM_VA_SHIFT_TRIE_LVL_2_HI;
    trie->levelInfo[3].loBit = UVM_VA_SHIFT_TRIE_LVL_3_LO;
    trie->levelInfo[3].hiBit = UVM_VA_SHIFT_TRIE_LVL_3_HI;

    _vat_init_level3(trie, &trie->root);
}

static void _vat_destroy_level0(UvmVat *trie, UvmVatL0 *level)
{
    NvU32 i;
    down_write(&level->headerLock);
    for (i = 0; i < sizeof(level->entry)/sizeof(level->entry[0]); i++)
    {
        down_write(&level->entry[i].lock);
        if (level->entry[i].child)
        {
            // TODO: There should be an unmap and delete call for the desc
            uvmfull_delete_va_desc((Uvm2MbVaDesc*)level->entry[i].child);
            level->entry[i].child = NULL;
            // TODO: 1. For all gpus mapped; unmap the PD entry
        }
        up_write(&level->entry[i].lock);
    }
    // TODO: 2. For all gpus with allocated PDES; free the PDEs
}

static void _vat_destroy_level1(UvmVat *trie, UvmVatL1 *level)
{
    NvU32 i;
    for (i = 0; i < sizeof(level->entry)/sizeof(level->entry[0]); i++)
    {
        if (level->entry[i].child)
        {
            _vat_destroy_level0(trie, level->entry[i].child);
            kmem_cache_free(g_uvmVatL0Cache, level->entry[i].child);
            level->entry[i].child = NULL;
            // TODO: 1. For all gpus mapped; unmap the PD entry
        }
    }
    // TODO: 2. For all gpus with allocated PDES; free the PDEs
}

static void _vat_destroy_level2(UvmVat *trie, UvmVatL2 *level)
{
    NvU32 i;
    for (i = 0; i < sizeof(level->entry)/sizeof(level->entry[0]); i++)
    {
        if (level->entry[i].child)
        {
            _vat_destroy_level1(trie, level->entry[i].child);
            kmem_cache_free(g_uvmVatL1Cache, level->entry[i].child);
            level->entry[i].child = NULL;
            // TODO: 1. For all gpus mapped; unmap the PD entry
        }
    }
    // TODO: 2. For all gpus with allocated PDES; free the PDEs
}

static void _vat_destroy_level3(UvmVat *trie, UvmVatL3 *level)
{
    NvU32 i;
    down_write(&level->headerLock);
    for (i = 0; i < sizeof(level->entry)/sizeof(level->entry[0]); i++)
    {
        down_write(&level->entry[i].lock);
        if (level->entry[i].child)
        {
            _vat_destroy_level2(trie, level->entry[i].child);
            kmem_cache_free(g_uvmVatL2Cache, level->entry[i].child);
            level->entry[i].child = NULL;
            // TODO: 1. For all gpus mapped; unmap the PD entry
        }
        up_write(&level->entry[i].lock);
    }
    // TODO: 2. For all gpus with allocated PDES; free the PDEs
    up_write(&level->headerLock);
}

//
// This function should be called after 2MB Descriptor entries have been freed
// This only check levels above level0.
//
void uvmfull_vat_reclaim_levels(UvmVat * trie, UvmVatContext *query)
{
    UvmVatL0 * level0;
    UvmVatL1 * level1;
    UvmVatL2 * level2;
    UvmVatL3 * level3;
    UvmVatL1Entry * level1Entry;
    UvmVatL2Entry * level2Entry;
    UvmVatL3Entry * level3Entry;
    UvmVatLevelGpuHeader *gpuHeader, *parentGpuHeader, *tempGpuHeader;
    GMMU_FMT *gmmuFmt;
    const MMU_FMT_LEVEL *pFmtLevel;
    NvU32 entryId;
    UvmGpuVaspace * gpuVaspace;

    // leaf pageTables are reclaimed when 2MB desc is freed.
    entryId  = VAT_EXTRACT_INDEX(query->address,
                                 trie->levelInfo[UVM_VA_TRIE_LVL_0].hiBit,
                                 trie->levelInfo[UVM_VA_TRIE_LVL_0].loBit);
    level0  = (UvmVatL0 *)GET_LEVEL_FROM_ENTRY((UvmVatL0Entry *)query->private.level0Entry,
                                               UvmVatL0, UvmVatL0Entry,
                                               entryId);
    UVM_PANIC_ON(level0 == NULL);

    // Get level 1 header
    entryId  = VAT_EXTRACT_INDEX(query->address,
                                 trie->levelInfo[UVM_VA_TRIE_LVL_1].hiBit,
                                 trie->levelInfo[UVM_VA_TRIE_LVL_1].loBit);
    level1  = (UvmVatL1 *)(GET_LEVEL_FROM_ENTRY(level0->header.parentLevel,
                                                UvmVatL1, UvmVatL1Entry,
                                                entryId));
    UVM_PANIC_ON(level1 == NULL);

    //
    // Check entries on Level0 - Unmap & free all entries which are invalid on
    // specific GPU. As levelHeader validEntries is overall count for all
    // gpuHeader->validEntries, free on all GPUs will guarentee level is empty
    //
    level1Entry = (UvmVatL1Entry *)level0->header.parentLevel;
    list_for_each_entry_safe(gpuHeader, tempGpuHeader,
                             &level0->header.gpuHeaders, node)
    {
        //
        // Walk thru all GPU headers & see where this was mapped. Reduce the
        // entry cout there as it has been unmapped and if zero move to parent
        //
        if (!(level1Entry->gpuMappedMask & NVBIT64(gpuHeader->gpuId)) ||
            gpuHeader->gpuValidEntries != 0)
            continue;

        //
        // level0 gpuHeader count is "0". Unmap this level from leve1 and
        // free it.
        //
        gpuVaspace = &query->vaspace->gpuVaspaces[gpuHeader->gpuId];
        gmmuFmt = gpuVaspace->client.pGmmuFmt;
        pFmtLevel = mmuFmtFindLevelWithPageShift(gmmuFmt->pRoot,
                                                UVM_VA_SHIFT_TRIE_LVL_0_LO);

        parentGpuHeader = uvmfull_vat_get_gpu_header(gpuHeader->gpuId, 
                                                     &level1->header);
        // Unmap this level on specific GPU
        _vat_update_gpu_level(gpuVaspace, &level1->header, parentGpuHeader,
                              &level1Entry->gpuMappedMask, entryId, NULL,
                              NV_FALSE);

        // Free the memory which has been unmapped
        uvmfull_free_pagetbl_memory(PA_TO_PFN(gpuHeader->levelPhyAddr),
                                    mmuFmtLevelSize(pFmtLevel),
                                    gpuHeader->aperture,
                                    gpuVaspace->gpuId);
        level0->header.gpuAllocMask &= ~NVBIT64(gpuHeader->gpuId);
        list_del(&gpuHeader->node);
        kmem_cache_free(g_uvmVatGpuHeaderCache, gpuHeader);
    }

    // Free level0 vat dataStructure if all validEntries were zero
    if (level0->header.validLevelEntries == 0)
    {
        // One entry from level above will also become invalid
        --level1->header.validLevelEntries;

        // Break parent-child relation and free memory
        kmem_cache_free(g_uvmVatL0Cache, level0);
        level1Entry->child = NULL;
    }

    // Get level 2 header
    entryId  = VAT_EXTRACT_INDEX(query->address,
                                 trie->levelInfo[UVM_VA_TRIE_LVL_2].hiBit,
                                 trie->levelInfo[UVM_VA_TRIE_LVL_2].loBit);
    level2  = (UvmVatL2 *)(GET_LEVEL_FROM_ENTRY(level1->header.parentLevel,
                                                UvmVatL2, UvmVatL2Entry, entryId));
    UVM_PANIC_ON(level2 == NULL);

    //
    // Check entries on Level1 - Unmap & free all entries which are invalid on
    // specific GPU. As levelHeader validEntries is overall count for all
    // gpuHeader->validEntries, free on all GPUs will guarentee level is empty
    //
    level2Entry = (UvmVatL2Entry *)level1->header.parentLevel;
    list_for_each_entry_safe(gpuHeader, tempGpuHeader,
                             &level1->header.gpuHeaders, node)
    {
        //
        // Walk thru all GPU headers & see where this was mapped. Reduce the
        // entry cout there as it has been unmapped and if zero move to parent
        //
        if (!(level2Entry->gpuMappedMask & NVBIT64(gpuHeader->gpuId)) ||
            gpuHeader->gpuValidEntries != 0)
            continue;

        //
        // level1 gpuHeader count is "0". Unmap this level from level2 and
        // free it.
        //
        gpuVaspace = &query->vaspace->gpuVaspaces[gpuHeader->gpuId];
        gmmuFmt = gpuVaspace->client.pGmmuFmt;
        pFmtLevel = mmuFmtFindLevelWithPageShift(gmmuFmt->pRoot,
                                                 UVM_VA_SHIFT_TRIE_LVL_1_LO);

        parentGpuHeader = uvmfull_vat_get_gpu_header(gpuHeader->gpuId, 
                                                     &level2->header);
        // Unmap this level on specific GPU
        _vat_update_gpu_level(gpuVaspace, &level2->header, parentGpuHeader,
                              &level2Entry->gpuMappedMask, entryId, NULL,
                              NV_FALSE);

        // Free the memory which has been unmapped
        uvmfull_free_pagetbl_memory(PA_TO_PFN(gpuHeader->levelPhyAddr),
                                    mmuFmtLevelSize(pFmtLevel),
                                    gpuHeader->aperture,
                                    gpuVaspace->gpuId);
        level1->header.gpuAllocMask &= ~NVBIT64(gpuHeader->gpuId);
        list_del(&gpuHeader->node);
        kmem_cache_free(g_uvmVatGpuHeaderCache, gpuHeader);
    }

    // Free level1 vat dataStructure if all validEntries were zero
    if (level1->header.validLevelEntries == 0)
    {
        // One entry from level above will also become invalid
        --level2->header.validLevelEntries;

        // Break parent-child relation and free memory
        kmem_cache_free(g_uvmVatL1Cache, level1);
        level2Entry->child = NULL;
    }

    // Get level 3 header
    entryId  = VAT_EXTRACT_INDEX(query->address,
                                 trie->levelInfo[UVM_VA_TRIE_LVL_3].hiBit,
                                 trie->levelInfo[UVM_VA_TRIE_LVL_3].loBit);
    level3  = (UvmVatL3 *)(GET_LEVEL_FROM_ENTRY(level2->header.parentLevel,
                                                UvmVatL3, UvmVatL3Entry, entryId));
    UVM_PANIC_ON(level3 == NULL);

    //
    // Check entries on Level1 - Unmap & free all entries which are invalid on
    // specific GPU. As levelHeader validEntries is overall count for all
    // gpuHeader->validEntries, free on all GPUs will guarentee level is empty
    //
    level3Entry = (UvmVatL3Entry *)level2->header.parentLevel;
    list_for_each_entry_safe(gpuHeader, tempGpuHeader,
                             &level2->header.gpuHeaders, node)
    {
        //
        // Walk thru all GPU headers & see where this was mapped. Reduce the
        // entry cout there as it has been unmapped and if zero move to parent
        //
        if (!(level3Entry->gpuMappedMask & NVBIT64(gpuHeader->gpuId)) ||
            gpuHeader->gpuValidEntries != 0)
            continue;

        //
        // level2 gpuHeader count is "0". Unmap this level from level3 and
        // free it.
        //
        gpuVaspace = &query->vaspace->gpuVaspaces[gpuHeader->gpuId];
        gmmuFmt = gpuVaspace->client.pGmmuFmt;
        pFmtLevel = mmuFmtFindLevelWithPageShift(gmmuFmt->pRoot,
                                                 UVM_VA_SHIFT_TRIE_LVL_2_LO);

        parentGpuHeader = uvmfull_vat_get_gpu_header(gpuHeader->gpuId, 
                                                     &level3->header);
        // Unmap this level on specific GPU
        _vat_update_gpu_level(gpuVaspace, &level3->header, parentGpuHeader,
                              &level3Entry->gpuMappedMask, entryId, NULL,
                              NV_FALSE);

        // Free the memory which has been unmapped
        uvmfull_free_pagetbl_memory(PA_TO_PFN(gpuHeader->levelPhyAddr),
                                    mmuFmtLevelSize(pFmtLevel),
                                    gpuHeader->aperture,
                                    gpuVaspace->gpuId);
        level2->header.gpuAllocMask &= ~NVBIT64(gpuHeader->gpuId);
        list_del(&gpuHeader->node);
        kmem_cache_free(g_uvmVatGpuHeaderCache, gpuHeader);

        // We are done here as we never free level3
    }

    // Don't free anything at level3 as we keep level3 alive
    if (level2->header.validLevelEntries == 0)
    {
        // One entry from level above will also become invalid
        --level3->header.validLevelEntries;

        // Break parent-child relation and free memory
        kmem_cache_free(g_uvmVatL2Cache, level2);
        level3Entry->child = NULL;
    }
    return;
}

// Unmaps all PDEs at all the levels for all the gpus.
// Also destroys the leaf nodes if any.
void uvmfull_vat_destroy(UvmVat *trie)
{
    if (trie->levels == 0)
        return;

    _vat_destroy_level3(trie, &trie->root);
}
