/*******************************************************************************
    Copyright (c) 2014-2015 NVIDIA Corporation

    Permission is hereby granted, free of charge, to any person obtaining a copy
    of this software and associated documentation files (the "Software"), to
    deal in the Software without restriction, including without limitation the
    rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
    sell copies of the Software, and to permit persons to whom the Software is
    furnished to do so, subject to the following conditions:

        The above copyright notice and this permission notice shall be
        included in all copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
    IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
    FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
    THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
    LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
    FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
    DEALINGS IN THE SOFTWARE.

*******************************************************************************/

//
// uvm_full_va_trie.h
//
// This file contains data-structure definitions for VA Trie
//

#ifndef _UVM_FULL_VA_TRIE_H_
#define _UVM_FULL_VA_TRIE_H_

#include "nvtypes.h"
#include "nvstatus.h"
#include "uvm_linux.h"
#include "uvm_full_types.h"
#include "uvm_full.h"

/*
 *  VA Trie definitions:
 *  The leaf node represents a 2MB region.
 *  Levels match the Pascal PD levels.
 *  Levels also store the PA to corresponding HW levels.
 */
#define  UVM_VA_SHIFT                    (49)
#define  UVM_VA_SHIFT_4K_PAGE            (12)
#define  UVM_VA_SHIFT_64K_PAGE           (UVM_VA_SHIFT_4K_PAGE + 4)
#define  UVM_VA_SHIFT_2MB_PAGE           (UVM_VA_SHIFT_4K_PAGE + 9)

#define  UVM_VA_TRIE_LVLS                (4)
#define UVM_VA_TRIE_LVL_0                (0)
#define UVM_VA_TRIE_LVL_1                (1)
#define UVM_VA_TRIE_LVL_2                (2)
#define UVM_VA_TRIE_LVL_3                (3)

// Level 0: 8 bits
#define  UVM_VA_SHIFT_TRIE_LVL_0_LO      (UVM_VA_SHIFT_2MB_PAGE)
#define  UVM_VA_SHIFT_TRIE_LVL_0_HI      (UVM_VA_SHIFT_2MB_PAGE + 8 - 1)
#define  UVM_VA_SHIFT_TRIE_LVL_0         UVM_VA_SHIFT_TRIE_LVL_0_HI :          \
                                         UVM_VA_SHIFT_TRIE_LVL_0_LO

// Level 1: 9 bits
#define  UVM_VA_SHIFT_TRIE_LVL_1_LO      (UVM_VA_SHIFT_TRIE_LVL_0_HI + 1)
#define  UVM_VA_SHIFT_TRIE_LVL_1_HI      (UVM_VA_SHIFT_TRIE_LVL_0_HI + 9)
#define  UVM_VA_SHIFT_TRIE_LVL_1         UVM_VA_SHIFT_TRIE_LVL_1_HI :          \
                                         UVM_VA_SHIFT_TRIE_LVL_1_LO

// Level 2: 9 bits
#define  UVM_VA_SHIFT_TRIE_LVL_2_LO      (UVM_VA_SHIFT_TRIE_LVL_1_HI + 1)
#define  UVM_VA_SHIFT_TRIE_LVL_2_HI      (UVM_VA_SHIFT_TRIE_LVL_1_HI + 9)
#define  UVM_VA_SHIFT_TRIE_LVL_2         UVM_VA_SHIFT_TRIE_LVL_2_HI :          \
                                         UVM_VA_SHIFT_TRIE_LVL_2_LO

// Level 3: 2 bits
#define  UVM_VA_SHIFT_TRIE_LVL_3_LO      (UVM_VA_SHIFT_TRIE_LVL_2_HI + 1)
#define  UVM_VA_SHIFT_TRIE_LVL_3_HI      (UVM_VA_SHIFT_TRIE_LVL_2_HI + 2)
#define  UVM_VA_SHIFT_TRIE_LVL_3         UVM_VA_SHIFT_TRIE_LVL_3_HI :          \
                                         UVM_VA_SHIFT_TRIE_LVL_3_LO

#define UVM_VA_TRIE_LOCKS                (UVM_VA_TRIE_LVLS * 2)

#define GET_LEVEL_FROM_ENTRY(entryPtr, levelType, entryType, entryId) \
   (levelType *)((NvU64)((entryType *)entryPtr - entryId) - \
                         (NvU64)offsetof(levelType, entry)) 

#define VAT_EXTRACT_INDEX(addr, hiBit, loBit)                   \
            ((addr >> loBit) & (NVBIT64(hiBit - loBit + 1) - 1))

struct UvmVaTrieLevelGpuHeader_t
{
    NvU8                id;
    UvmGpuId            gpuId;
    UvmMemType          aperture;
    NvU64               levelPhyAddr;

    struct list_head    node; // Node for gpu's phys properties. head:gpuHeaders
    NvU32               gpuValidEntries; // Valid entries at give level in a GPU
};

struct UvmVaTrieLevelHeader_t
{
    NvU8                    id;
    NvU8                    loBit;
    NvU8                    hiBit;
    //! Gpus with this level allocated
    NvU64                   gpuAllocMask;

    //! List of GpuHeaders for the lvl
    struct list_head        gpuHeaders;

    void                    *parentLevel;       // Parent level pointer
    NvU32                    validLevelEntries; // Valid entries at give level
};

struct UvmVaTrieLevel0Entry_t
{
    void*                   child;    // 2MB va descriptor
    //! Whether the PDE0 entry phys address is valid in the leaf node.
    NvU64                   pdePresentMask;
    struct rw_semaphore     lock;
};

/*
 * @struct UvmVaTrieLevel0_t
 * @brief  This structure describes the level0 page directory.
 *         The index used in the array corresponds to the entry pointing to
 *         2MB va descriptor.
 */
struct UvmVaTrieLevel0_t
{
    // This header will share lock with the parent's
    struct rw_semaphore headerLock;   // To be deprecated.
    UvmVatLevelHeader   header;             // level physical information
    UvmVatL0Entry       entry[1 << (UVM_VA_SHIFT_TRIE_LVL_0_HI -
                                   UVM_VA_SHIFT_TRIE_LVL_0_LO + 1)];
};

struct UvmVaTrieLevel1Entry_t
{
    NvU64                   gpuMappedMask;
    UvmVatL0                *child;
};

/*
 * @struct UvmVaTrieLevel1_t
 * @brief  This structure describes the level1 page directory.
 *         The index used in the array corresponds to the entry pointing to
 *         level 0 directory.
 */
struct UvmVaTrieLevel1_t
{
    UvmVatLevelHeader   header;  // level physical information
    UvmVatL1Entry       entry[1 << (UVM_VA_SHIFT_TRIE_LVL_1_HI -
                                    UVM_VA_SHIFT_TRIE_LVL_1_LO + 1)];
};

struct UvmVaTrieLevel2Entry_t
{
    NvU64                   gpuMappedMask;
    UvmVatL1                *child;
};

/*
 * @struct UvmVaTrieLevel2_t
 * @brief  This structure describes the level2 page directory.
 *         The index used in the array corresponds to the entry pointing to
 *         level 1 directory.
 */
struct UvmVaTrieLevel2_t
{
    UvmVatLevelHeader   header;  // level2 physical information
    UvmVatL2Entry       entry[1 << (UVM_VA_SHIFT_TRIE_LVL_2_HI -
                                    UVM_VA_SHIFT_TRIE_LVL_2_LO + 1)];
};

struct UvmVaTrieLevel3Entry_t
{
    struct rw_semaphore     lock;
    NvU64                   gpuMappedMask;
    UvmVatL2                *child;
};

/*
 * @struct UvmVaTrieLevel3_t
 * @brief  This structure describes the level3 page directory.
 *         The index used in the array corresponds to the entry pointing to
 *         level 2 directory.
 */
struct UvmVaTrieLevel3_t
{
    struct rw_semaphore headerLock;
    UvmVatLevelHeader   header;  // level2 physical information
    UvmVatL3Entry       entry[1 << (UVM_VA_SHIFT_TRIE_LVL_3_HI -
                                    UVM_VA_SHIFT_TRIE_LVL_3_LO + 1)];
};


struct UvmVaTrie_t
{
    NvU8                   levels;
    struct
    {
        NvU8               loBit;
        NvU8               hiBit;
    } levelInfo[UVM_VA_TRIE_LVLS];

    UvmVatL3               root;
};

struct UvmVatLockState_t
{
    struct rw_semaphore*  pLock;
    NvBool                bWriteMode;
};

/*
 * @struct UvmVaTrieContext_t
 * @brief  This structure describes the state a trie traversal is in.
 */
struct UvmVaTrieContext_t
{
    NvU64               address;
    UvmVatAccess        accessType;
    UvmGpuId            gpuId;  // relevant if access type is gpu specific
    UvmVaspace          *vaspace;

    struct
    {
        // Level0 pointer needed for upward traversal
        UvmVatL0Entry          *level0Entry;
        NvBool                  bRootExclusive;
        NvBool                  bL0Exclusive;
        // we acquire locks till we reach leaf. Release in reverse.
        UvmVatLockState         lockStack[UVM_VA_TRIE_LOCKS];
        NvU32                   tos;   // top of lock stack
    } private;          // Not to be touched by the user
};

/*
 * Given a root of the Trie and the query; returns the leaf node.
 * The query will read-lock the path to the leaf. The query will write-lock the
 * leaf node and return its address.
 */
void* uvmfull_vat_acquire_leaf(UvmVat *trie, UvmVatContext *query);

void uvmfull_vat_release_leaf(UvmVat *trie, UvmVatContext *query);

void uvmfull_vat_reclaim_levels(UvmVat * trie, UvmVatContext *query);

void uvmfull_vat_init(UvmVat *trie);

void uvmfull_vat_destroy(UvmVat *trie);

NV_STATUS uvmfull_va_trie_slab_init(void);

void uvmfull_va_trie_slab_destroy(void);

UvmVatLevelGpuHeader*  uvmfull_vat_get_gpu_header(UvmGpuId gpuId,
                                                  UvmVatLevelHeader* header);

#endif   //_UVM_FULL_VA_TRIE_H_
