/*******************************************************************************
    Copyright (c) 2014-2015 NVIDIA Corporation

    Permission is hereby granted, free of charge, to any person obtaining a copy
    of this software and associated documentation files (the "Software"), to
    deal in the Software without restriction, including without limitation the
    rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
    sell copies of the Software, and to permit persons to whom the Software is
    furnished to do so, subject to the following conditions:

        The above copyright notice and this permission notice shall be
        included in all copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
    IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
    FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
    THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
    LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
    FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
    DEALINGS IN THE SOFTWARE.

*******************************************************************************/

#include "uvm_common.h"
#include "uvm_hashmap.h"

// FNV1a is a public domain hash family. While more costly than some algorithms
// that dont use multiplication, it provides good distribution.
static NvU32 _hash_func(NvU64 key)
{
    NvU8 *p = (NvU8*)&key;
    NvU32 h = 2166136261;
    size_t i;

    for(i = 0; i < sizeof(key); i++)
    {
        h = (h ^ p[i]) * 16777619;
    }
    return h;
}

static inline struct hlist_head* _hashmap_get_slot(UvmHashMap *map, NvU64 key)
{
    NvU32 slot = _hash_func(key);
    slot &= map->nSlots - 1;
    return &(map->slots[slot]);
}

void uvm_hashmap_destroy(UvmHashMap *map)
{
    size_t i;
    struct hlist_head *s;
    UvmHashEntry *entry;
    struct hlist_node *pos, *safepos;

    if (map->nItems)
    {
        for (i = 0; i < map->nSlots; i++)
        {
            s = &(map->slots[i]);
            NV_HLIST_FOR_EACH_ENTRY_SAFE(entry, pos, safepos, s, node)
            {
                uvm_hashmap_remove(map, entry);
            }
        }
        WARN_ON(map->nItems);
    }

    kfree(map->slots);
    memset(map, 0, sizeof(*map));
}

NV_STATUS uvm_hashmap_init(UvmHashMap *map, NvU8 nSlotsLog2)
{
    size_t i, nSlots;
    memset(map, 0, sizeof(*map));
    nSlots = 1 << nSlotsLog2;

    map->slots = (struct hlist_head*)(kmalloc(nSlots*sizeof(struct hlist_head),
                                         NV_UVM_GFP_FLAGS));
    if (!map->slots)
        return NV_ERR_NO_MEMORY;

    map->nSlots = nSlots;
    for (i = 0; i < nSlots; i++)
        INIT_HLIST_HEAD(&map->slots[i]);
    return NV_OK;
}

void uvm_hashmap_insert(UvmHashMap *map, UvmHashEntry *entry, NvU64 key)
{
    struct hlist_head *s;

    if (debug_mode())
    {
        // Catch duplicate keys
        if (uvm_hashmap_query(map, key))
        {
            WARN_ON(1);
            return;
        }
    }
    entry->key = key;
    s = _hashmap_get_slot(map, entry->key);
    hlist_add_head(&entry->node, s);
    map->nItems++;
}

void uvm_hashmap_remove(UvmHashMap *map, UvmHashEntry *entry)
{
    WARN_ON(map->nItems == 0);
    WARN_ON(hlist_unhashed(&entry->node));
    if (debug_mode())
    {
        if (entry != uvm_hashmap_query(map, entry->key))
        {
            WARN_ON(1);
            return;
        }
    }
    hlist_del_init(&entry->node);
    map->nItems--;
}

UvmHashEntry* uvm_hashmap_query(UvmHashMap *map, NvU64 key)
{
    UvmHashEntry *entry;
    struct hlist_head *s;
    struct hlist_node *pos;

    if (!map->nItems)
        return NULL;

    s = _hashmap_get_slot(map, key);
    NV_HLIST_FOR_EACH_ENTRY(entry, pos, s, node)
    {
        if (entry->key == key)
            return entry;
    }
    return NULL;
}
