/*******************************************************************************
    Copyright (c) 2015-2016 NVIDIA Corporation

    Permission is hereby granted, free of charge, to any person obtaining a copy
    of this software and associated documentation files (the "Software"), to
    deal in the Software without restriction, including without limitation the
    rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
    sell copies of the Software, and to permit persons to whom the Software is
    furnished to do so, subject to the following conditions:

        The above copyright notice and this permission notice shall be
        included in all copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
    IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
    FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
    THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
    LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
    FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
    DEALINGS IN THE SOFTWARE.

*******************************************************************************/

#include "uvmtypes.h"
#include "uvm_common.h"
#include "uvm_linux.h"
#include "uvm_perf.h"

#define UVM_THRASHING_FAULT_COUNT_DEFAULT 3
#define UVM_THRASHING_LAPSE_DEFAULT 60
#define UVM_THRASHING_NAP_DEFAULT 1000

// Thrashing prevention technique
int g_uvm_nothrashing;
static int uvm_nothrashing = UvmPerfThrashingPreventionNone;
module_param(uvm_nothrashing, int, S_IRUGO);

// Number of consecutive faults within a very short period of time that are considering "thrashing"
int g_uvm_thrashing_fault_count;
static int uvm_thrashing_fault_count = UVM_THRASHING_FAULT_COUNT_DEFAULT;
module_param(uvm_thrashing_fault_count, int, S_IRUGO);

// Lapse of time in microseconds between between page faults to be considered thrashing candidates
int g_uvm_thrashing_lapse;
static int uvm_thrashing_lapse = UVM_THRASHING_LAPSE_DEFAULT;
module_param(uvm_thrashing_lapse, int, S_IRUGO);

// Lapse of time in microseconds that the CPU process sleeps to avoid thrashing
int g_uvm_thrashing_nap;
static int uvm_thrashing_nap = UVM_THRASHING_NAP_DEFAULT;
module_param(uvm_thrashing_nap, int, S_IRUGO);

NV_STATUS uvm_perf_init(void)
{
    if (uvm_nothrashing < 0 || uvm_nothrashing >= UvmPerfThrashingPreventionCount) {
        UVM_ERR_PRINT("Invalid module parameter value for uvm_nothrashing\n");
        return NV_ERR_INVALID_ARGUMENT;
    }
    g_uvm_nothrashing = uvm_nothrashing;

    if (uvm_thrashing_fault_count <= 0) {
        UVM_ERR_PRINT("Invalid module parameter value for uvm_thrashing_fault_count\n");
        return NV_ERR_INVALID_ARGUMENT;
    }
    g_uvm_thrashing_fault_count = uvm_thrashing_fault_count;

    if (uvm_thrashing_lapse <= 0) {
        UVM_ERR_PRINT("Invalid module parameter value for uvm_thrashing_lapse\n");
        return NV_ERR_INVALID_ARGUMENT;
    }
    g_uvm_thrashing_lapse = uvm_thrashing_lapse;

    // With non-GPL symbols we can only sleep with ms granularity.
    // TODO: remove this restriction if we eventually have access to GPL symbols
    if (uvm_thrashing_nap < 1000) {
        UVM_ERR_PRINT("Invalid module parameter value for uvm_thrashing_nap\n");
        return NV_ERR_INVALID_ARGUMENT;
    }
    g_uvm_thrashing_nap = uvm_thrashing_nap;

    return NV_OK;
}

void uvm_perf_exit(void)
{
}

NV_STATUS uvm_perf_get_thrashing_prevention(UVM_PERF_GET_THRASHING_PREVENTION_PARAMS *params, struct file *filp)
{
    params->prevention = g_uvm_nothrashing;
    params->fault_count = g_uvm_thrashing_fault_count;
    params->lapse = g_uvm_thrashing_lapse;
    if (g_uvm_nothrashing == UvmPerfThrashingPreventionThrottleSleep)
        params->throttle_sleep.nap = g_uvm_thrashing_nap;
    params->rmStatus = NV_OK;

    return params->rmStatus;
}

NV_STATUS uvm_perf_set_thrashing_prevention(UVM_PERF_SET_THRASHING_PREVENTION_PARAMS *params, struct file *filp)
{
    params->rmStatus = NV_OK;

    if (params->prevention < 0 || params->prevention >= UvmPerfThrashingPreventionCount) {
        params->rmStatus = NV_ERR_INVALID_ARGUMENT;
        goto fail;
    }
    else
        g_uvm_nothrashing = params->prevention;

    if (params->fault_count <= 0) {
        params->rmStatus = NV_ERR_INVALID_ARGUMENT;
        goto fail;
    }
    else
        g_uvm_thrashing_fault_count = params->fault_count;

    // With non-GPL symbols we can only sleep with ms granularity.
    // TODO: remove this restriction if we eventually have access to GPL symbols
    if (params->lapse <= 1000) {
        params->rmStatus = NV_ERR_INVALID_ARGUMENT;
        goto fail;
    }
    else
        g_uvm_thrashing_lapse = params->lapse;

    if (params->prevention == UvmPerfThrashingPreventionThrottleSleep && params->throttle_sleep.nap <= 1000) {
        params->rmStatus = NV_ERR_INVALID_ARGUMENT;
        goto fail;
    }
    else
        g_uvm_thrashing_nap = params->throttle_sleep.nap;

fail:
    return params->rmStatus;
}
